/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\Program.cs
 * PURPOSE: Program initialization and deinitialization
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Ipc;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using NB.Content;
using NB.Gui;
using NB.Language;
using Support;
using Support.HashFunctions;

namespace NB
{
    static class ApplicationEntryPointObject
    {
        // This class is responsible for checking for library files.

        // An STA thread is required to access the clipboard.
        [STAThread]
        private static void Main(string[] a_Args)
        {
            const string l_SetupMutexName = (Software.c_AppName + "FileSharingSetupMutex");
            bool l_Junk;
            Mutex l_SetupMutex = new Mutex(true, l_SetupMutexName, out l_Junk);

            string l_ModulePath = Application.ExecutablePath;

            try
            {
                l_ModulePath = Path.GetDirectoryName(l_ModulePath);
            }
            catch
            {
                l_ModulePath = null;
            }

            if (null != l_ModulePath)
            {
                // It's very difficult to catch the exception that occurrs
                // when the runtime misses a library file, so do a manual
                // check. This check only works if the libraries were not
                // used so far.
                AssertThatLibraryExistsWithMsg(l_ModulePath, "Support.dll");
                AssertThatLibraryExistsWithMsg(l_ModulePath, "zlibManaged.dll");
            }

            Startup(a_Args);
        }

        private static void Startup(string[] a_Args)
        {
            Software.Startup(a_Args);
        }

        private static void AssertThatLibraryExistsWithMsg(string a_Folder,
                                                           string a_BinaryFileName)
        {
            // Do not depend on the Software class here, it would cause
            // the runtime to load the support library (constants are
            // okay though). Use framework routines instead.

            string l_FilePath;

            try
            {
                l_FilePath = Path.Combine(a_Folder, a_BinaryFileName);
            }
            catch (ArgumentException)
            {
                l_FilePath = null;
            }

            bool l_Exists = (null == l_FilePath ? false : File.Exists(l_FilePath));

            if (false == l_Exists)
            {
                MessageBox.Show(string.Format("The library \"{0}\" must be present in order to start this application.", a_BinaryFileName), Software.c_AppName, MessageBoxButtons.OK, MessageBoxIcon.Error);

                Environment.Exit(0);
            }
        }
    }

    static class Software
    {
        public const string c_AppName = "Nachtblitz";
        public const string c_AppNameLower = "nachtblitz";
        public const string c_GuiThreadName = "GUI";
        public const string c_TranslationsSubFolder = "translations";
        public const string c_ProgressBarBgFolder = "bar backgrounds";
        private const string c_PreviewSubFolder = (c_AppNameLower + " preview folder");
        public const string c_ProjectEmailAddr = (c_AppNameLower + "@mail.i2p");
        // Casing matters.
        public const string c_ProjectURL = ("http://echelon.i2p/" + c_AppNameLower + "/");
        public const string c_NodesDat = "nodes.dat";
        private const string c_CoreAppDataFilePrefix = "Core";
        private const string c_IncomplAppDataFilePrefix = "Incompl";

        public static readonly CultureInfo s_StartupCulture;
        public static readonly Time s_AppLaunchTime;
        public static readonly ReleaseInfo s_ReleaseInfo;
        public static readonly GuiToolTip s_ToolTip;
        private static string s_ModulePath;
        private static string s_ClassicAppDataPath;
        private static string s_DefaultAppDataPath;
        private static string s_DefaultTempPath;
        private static string s_DefaultPreviewFolder;
        public static TextSet s_DefaultEngTextSet;
        private static TextSet s_EffTextSet;
        public static BootConfig s_BootConfig;
        private static bool s_WindowsDetected;
        private static bool s_MonoDetected;
        private static InterAppDomainComm s_Comm;
        public static System.Drawing.Icon s_AppIcon;
        // s_IconStorage may only be used by the GUI thread.
        public static IconStorage s_IconStorage;
        private static MainForm s_MainForm;
        public static bool s_SecondLicenseAccepted;

        public static bool s_LaterVersionAvailable;
        public static uint s_NewerVersion;

        public static string s_LastOutputDir;

        public static TextSet sp_TextSet
        {
            get
            {
                return s_EffTextSet;
            }
        }

        public static MainForm sp_MainForm
        {
            // This property should only be used by the GUI.
            get { return s_MainForm; }
        }

        public static bool sp_WindowsDetected
        {
            get { return s_WindowsDetected; }
        }

        public static string sp_ModuleName
        {
            get { return Application.ExecutablePath; }
        }

        public static string sp_ModulePath
        {
            get { return s_ModulePath; }
        }

        public static string sp_ClassicAppDataPath
        {
            get { return s_ClassicAppDataPath; }
        }

        public static string sp_DefaultAppDataPath
        {
            get { return s_DefaultAppDataPath; }
        }

        public static string sp_DefaultTempPath
        {
            get { return s_DefaultTempPath; }
        }

        public static string sp_DefaultPreviewFolder
        {
            get { return s_DefaultPreviewFolder; }
        }

        static Software()
        {
            s_StartupCulture = Thread.CurrentThread.CurrentCulture;
            s_AppLaunchTime = Time.Now();
            // The revision can only be as high as 7 (eD2k protocol
            // limitation).
            s_ReleaseInfo = new ReleaseInfo(32,
                                            1, 4, 25,
                                            "Mitternacht",
                                            2015, 1, 24);
            s_NewerVersion = s_ReleaseInfo.EncodeVersion();
            s_ToolTip = new GuiToolTip();
        }

        public static void Startup(string[] a_Args)
        {
            Thread.CurrentThread.Name = c_GuiThreadName;

            // Must be done before the first control is created.
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            if (false == Environment.UserInteractive)
            {
                return;
            }

            if (false == BitConverter.IsLittleEndian)
            {
                ErrMsg(null, "Machine byte order not supported.");

                return;
            }

            if (1 == a_Args.Length &&
                "-CrashMessage" == a_Args[0])
            {
                string l_CrashLogFilePath = LongPath.Combine(LongPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), c_AppName), c_AppName + "CrashLog.txt");
                ShowCrashMessage(l_CrashLogFilePath);

                return;
            }

            if (DateTime.UtcNow.Date.AddTicks(-366 * TimeSpan.TicksPerDay) > s_ReleaseInfo.f_BuildDate)
            {
                // This version seems to have been released more than
                // one year ago, so tell the user that there is probably
                // a better version out there.

                if (DialogResult.Yes != MessageBox.Show(BuildObsoleteClientMessage(true), s_ReleaseInfo.f_InvarReleaseStr, MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2))
                {
                    return;
                }
            }

            s_MonoDetected = IsRunningOnMono();

#if MONO
            if (false == s_MonoDetected)
            {
                ErrMsg(null, "This binary is meant to run on Mono only.");

                return;
            }
#else
            if (true == s_MonoDetected)
            {
                // Abort, otherwise Mono's trouble with the
                // LayoutKind.Auto attribute will lead to a crash.

                ErrMsg(null, "This binary is meant to run on the original .NET framework.");

                return;
            }
#endif

            // Get the location of our executable.
            string l_ModuleName = Application.ExecutablePath;
            s_ModulePath = LongPath.GetDirectoryName(l_ModuleName);
            s_ClassicAppDataPath = LongPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), c_AppName);
            s_DefaultAppDataPath = LongPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), c_AppName);
            s_DefaultTempPath = GetDefaultTempPath(s_DefaultAppDataPath);
            s_DefaultPreviewFolder = BuildDefaultPreviewFolder(s_DefaultTempPath);
            string l_LibraryErrMsg;

            if (false == SupportLibrary.Init(s_ReleaseInfo.f_ReleaseNumber, s_ModulePath, out l_LibraryErrMsg))
            {
                ErrMsg(null, l_LibraryErrMsg);

                return;
            }

            s_WindowsDetected = Util.IsWindows();

            // Register global exception handlers.

            if (false == Debugger.IsAttached)
            {
                AppDomain.CurrentDomain.UnhandledException += CurrentDomain_UnhandledException;

                Application.ThreadException += Application_GuiThreadException;
                Application.SetUnhandledExceptionMode(UnhandledExceptionMode.CatchException);
            }

            Microsoft.Win32.SystemEvents.PowerModeChanged += SystemEvents_PowerModeChanged;

            // Load default text set (English).

            BinaryReader l_DefTextSetRdr = new BinaryReader(OpenResource("en.dat", true));
            {
                s_DefaultEngTextSet = TextSet.LoadTemplate(CultureInfo.GetCultureInfo(BootConfig.c_EnglishCultureLCID), s_StartupCulture, l_DefTextSetRdr);
            }
            l_DefTextSetRdr.Close();

            StorageStates l_BootConfigState;

            if (false == File.Exists(LongPath.Combine(LongPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), c_AppName), c_AppName + "Startup.dat")))
            {
                l_BootConfigState = BootConfig.Load(out s_BootConfig);

                if (s_ClassicAppDataPath != s_DefaultAppDataPath)
                {
                    bool l_DirExists = Util.EnsureDirectoryExists(s_DefaultAppDataPath);
                }
            }
            else
            {
                l_BootConfigState = StorageStates.TooAdvanced;
            }

            bool l_BootConfigLoaded;

            switch (l_BootConfigState)
            {
                case StorageStates.Success:
                    l_BootConfigLoaded = true;
                    break;

                case StorageStates.TooAdvanced:
                    ErrMsg(null, s_DefaultEngTextSet[TextKeys.MsgAppDataErrFromNewerVersion]);
                    return;

                default:
                {
                    bool l_DirExists = Util.EnsureDirectoryExists(s_ClassicAppDataPath);
                    s_BootConfig = BootConfig.CreateDefault();
                    bool l_Saved = s_BootConfig.Save();
                    l_BootConfigLoaded = false;

                    break;
                }
            }

            TextSet l_TextSet = null;
            CultureInfo l_FormatCulture = (false == s_BootConfig.ShouldUseOSFormatCulture() ? s_BootConfig.GetFormatCulture() : s_StartupCulture);

            if (false == l_BootConfigLoaded)
            {
                if (BootConfig.c_EnglishCultureLCID != s_StartupCulture.LCID)
                {
                    l_TextSet = TryLoadCustomLanguages(s_StartupCulture, true, l_FormatCulture, true);
                }
            }
            else if (false == s_BootConfig.HasDefaultLanguageCulture())
            {
                l_TextSet = TryLoadCustomLanguages(s_BootConfig.GetLanguageCulture(), true, l_FormatCulture, true);
            }

            if (null == l_TextSet)
            {
                l_TextSet = TextSet.ChangeFormatCulture(s_DefaultEngTextSet, l_FormatCulture);
            }

            s_BootConfig.SetLanguageCulture(l_TextSet.f_LanguageCulture);

            // Propagate the text set.

            l_TextSet.ApplyCulture();
            PropagateTextSet(l_TextSet);
            s_EffTextSet = l_TextSet;

            s_Comm = new InterAppDomainComm();
            string l_PendingShellStmt = null;

            if (0 == a_Args.Length)
            {
                if (false == s_Comm.p_IsServer)
                {
                    // Try to activate the running instance.

#if !DEBUG
                    bool l_Junk;
                    if (false == s_Comm.FocusMainForm(out l_Junk))
                    {
                        ErrMsg(l_TextSet, TextKeys.MsgAppAlreadyUp);
                    }

                    return;
#endif
                }
            }
            else
            {
                if (1 == a_Args.Length)
                {
                    l_PendingShellStmt = a_Args[0];
                }
                else
                {
                    ErrMsg(l_TextSet, TextKeys.MsgUnrecognizedCmdLineArgs);

                    return;
                }

                if (false == s_Comm.p_IsServer)
                {
                    if (null != l_PendingShellStmt)
                    {
                        bool l_Junk;
                        if (false == s_Comm.AcceptShellInput(l_PendingShellStmt, out l_Junk))
                        {
                            ErrMsg(l_TextSet, TextKeys.MsgFailedProcessCmdLineArgs);
                        }
                    }

                    return;
                }
            }

            // Make sure the compression library gets loaded (or fail).
            zlibManaged.zlib.compressBound(0);

            // Check if the application data dir exists.

            if (true == l_BootConfigLoaded &&
                false == Directory.Exists(s_BootConfig.p_AppDataDir))
            {
                string l_CustAppDataDir = s_BootConfig.p_AppDataDir;
                string l_EffDefAppDataDir = BootConfig.SuggestDefaultAppDataDir();

                if (l_CustAppDataDir != l_EffDefAppDataDir)
                {
                    string l_Quest = l_TextSet.Format(TextKeys.QuestAppDataDirIsMissingWantReset, l_CustAppDataDir, l_EffDefAppDataDir);

                    if (DialogResult.Yes == CustomMsgBox.ShowMsgBox(l_TextSet, null, l_TextSet.f_AppNameWithVer, l_Quest, MessageBoxIcon.Warning, MessageBoxButtons.YesNo, DialogResult.No))
                    {
                        s_BootConfig.SetNewApplicationDataDirectory(l_EffDefAppDataDir);
                        bool l_Saved = s_BootConfig.TakeOverNewApplicationDataDirectoryAndSave();
                    }
                }
            }

            // Load application data files.

            string l_CoreAppDataPath = LongPath.Combine(s_BootConfig.p_AppDataDir, c_CoreAppDataFilePrefix);
            string l_IncomplAppDataPath = LongPath.Combine(s_BootConfig.p_AppDataDir, c_IncomplAppDataFilePrefix);

            HardMemory l_CoreMem;
            HardMemory l_IncomplMem;
            if (false == LoadAppDataMem(l_TextSet, l_CoreAppDataPath, l_IncomplAppDataPath, out l_CoreMem, out l_IncomplMem))
            {
                End();
            }

            Stream l_AppIconStrm = OpenResource("Nachtblitz32.ico", true);
            s_AppIcon = new System.Drawing.Icon(l_AppIconStrm);
            l_AppIconStrm.Dispose();

            s_IconStorage = new IconStorage(IconStorage.s_SuggestedBgColor);

            // Check if the license dialog should be shown.

            const uint l_CurrentLicenseEdition = 2;

            if (false == WindowsIntegration.IsLicenseAcceptedValueSet() &&
                false == IsLicenseAccepted(l_CoreMem, l_CurrentLicenseEdition))
            {
                // First launch or sort of.

                using (LicenseDialog l_LicenseForm = new LicenseDialog(l_TextSet, LicenseDialog.DialogMode.AcceptOrRejectLicense))
                {
                    l_LicenseForm.Icon = s_AppIcon;
                    l_LicenseForm.ShowIcon = false;
                    l_LicenseForm.ShowInTaskbar = true;
                    DialogResult l_Result = l_LicenseForm.ShowDialog();

                    if (DialogResult.OK != l_Result)
                    {
                        l_CoreMem.Close(HardMemory.CompactKind.DontCompact, null);
                        l_IncomplMem.Close(HardMemory.CompactKind.DontCompact, null);

                        End();
                    }

                    s_SecondLicenseAccepted = true;
                    bool l_Set = WindowsIntegration.SetLicenseAcceptedValue();
                }
            }

            WindowsIntegration.RegisterApplication();

            using (MainForm l_MainForm = new MainForm(l_TextSet, l_PendingShellStmt, l_CoreMem, l_IncomplMem))
            {
                if (l_MainForm == s_MainForm)
                {
                    // There were no problems. So start the message loop.

                    Application.Run(l_MainForm);
                }
            }

            Microsoft.Win32.SystemEvents.PowerModeChanged -= SystemEvents_PowerModeChanged;

            s_AppIcon.Dispose();
            s_IconStorage.Dispose();
            s_ToolTip.Dispose();

#if MONO
            // Mono has a shutdown problem and stalls somehow even
            // after Environment.Exit(), so tell the user we're done.
            Console.WriteLine(c_AppName + " has shut down. You can now safely terminate the process.");
#endif

#if !DEBUG
            // Call End() in oder to kill any other threads still
            // alive. Due to the way the RSA key generation works
            // such threads may exists.

            End();
#endif
        }

        public static string BuildObsoleteClientMessage(bool a_WithQuestion)
        {
            string l_Ending = (false == a_WithQuestion ? string.Empty : "\n\nDo you really want to go on launching this obsolete version?");
            string l_WarnMsg = string.Format(Util.s_InvariantCulture, "This is software is outdated. Network functionality was suspended.\n\nA network application depends on constant maintenance to allow secure operation. According to your system's time, this version ({0}) of {1} was released more than one year ago ({2}).\n\nAs a consequence, this version of {1} is no longer considered fit for purpose and will not allow you to connect to the I2P network anymore.\n\nYou really need to update your client now. There should be a newer version of {1} available free of charge at the project's web page within I2P:\n\n{3}\n\nThe newer version will generally incorporate important fixes to better protect your anonymity and improve your file transfers. Using a recent version also allows technological progress for the entire community.{4}", s_ReleaseInfo.f_VersionStr, c_AppName, s_ReleaseInfo.f_BuildDate.ToString("MMMM d, yyyy", CultureInfo.InvariantCulture), c_ProjectURL, l_Ending);

            return l_WarnMsg;
        }

        public static string ReadAbsPathUTF16LEStringEncLen(BufferIO a_Buffer)
        {
            string l_Path = a_Buffer.ReadUTF16LEStringEncLen();
            bool l_Junk;

            if (false == LongPath.IsPathRooted(l_Path, out l_Junk))
            {
                l_Path = LongPath.Combine(sp_ModulePath, l_Path);
            }

            return l_Path;
        }

        private static TextSet TryLoadCustomLanguages(CultureInfo a_ProposedLanguageCulture,
                                                      bool a_DeriveOnFail,
                                                      CultureInfo a_FormatCulture,
                                                      bool a_Intern)
        {
            // Return can be null. s_DefaultEngTextSet must be valid
            // on entry.

            string l_Junk;
            TextSet l_TextSet = TextSet.LoadCustom(s_DefaultEngTextSet, a_ProposedLanguageCulture, a_FormatCulture, a_Intern, out l_Junk);

            if (true == a_DeriveOnFail &&
                null == l_TextSet &&
                false == a_ProposedLanguageCulture.IsNeutralCulture)
            {
                // Try the neutral culture instead.

                l_TextSet = TextSet.LoadCustom(s_DefaultEngTextSet, a_ProposedLanguageCulture.Parent, a_FormatCulture, a_Intern, out l_Junk);
            }

            // We may have failed to load any form of a_ProposedLanguageCulture.

            return l_TextSet;
        }

        private static void PropagateTextSet(TextSet a_TextSet)
        {
            SizeFormatter.ApplyTextSet(a_TextSet);
            TimeStrHelper.ApplyTextSet(a_TextSet);
            ColumnSet.ApplyTextSet(a_TextSet);
            NB.Network.Speed.ApplyTextSet(a_TextSet);
            NB.Network.eD2k.FileTypeDetector.ApplyTextSet(a_TextSet);
            NB.Network.DlPeerContext.ApplyTextSet(a_TextSet);
            TunnelControl.ApplyTextSet(a_TextSet);
            PriorityPickerControl.ApplyTextSet(a_TextSet);
            FileSearchPanel.ApplyTextSet(a_TextSet);
            FilesPanel.ApplyTextSet(a_TextSet);
            GeneralPeersPanel.ApplyTextSet(a_TextSet);
            MoveDataFilesDialog.ApplyTextSet(a_TextSet);
        }

        public static void UpdateShutDownState(uint a_State)
        {
        }

        private static bool IsLicenseAccepted(HardMemory a_CoreMem,
                                              uint a_Edition)
        {
            // a_Edition is one based.

            const uint l_Pos = 5;
            const byte l_ByteCount = sizeof(ushort);
            HardMemory.Element l_PrefsElm = a_CoreMem.GetElement(new ElmKey(Instance.c_PrefsElmKey));

            if (null == l_PrefsElm ||
                (l_Pos + l_ByteCount) > l_PrefsElm.f_MaxSize)
            {
                return false;
            }

            ByteArray l_Data = new ByteArray(l_ByteCount);

            if (false == a_CoreMem.Read(l_PrefsElm, l_Pos, ref l_Data))
            {
                return false;
            }

            uint l_Flag = (1u << (int)(a_Edition - 1));
            uint l_Combination = Util.ReadUInt16LE(l_Data.f_Arr, 0);

            return (0 != (l_Flag & l_Combination));
        }

        public static string GetRuntimeString(out string a_FrameworkStr,
                                              out string a_AdditionalPlatformInfoStr)
        {
            // a_FrameworkStr can be null on return.

            Version l_FrameworkVer;

            if (false == Util.FindFrameworkVersion(System.Reflection.Assembly.GetExecutingAssembly(), out l_FrameworkVer))
            {
                a_FrameworkStr = null;
            }
            else
            {
                a_FrameworkStr = ("Framework: v" + l_FrameworkVer);
            }

            a_AdditionalPlatformInfoStr = TryGetWineVersion();

            return string.Format(Util.s_InvariantCulture, "Runtime: v{0} ({1}-bit)", Environment.Version, IntPtr.Size * Util.c_BitsPerByte);
        }

        public static string TryGetWineVersion()
        {
            // Return can be null.

            IntPtr l_StrPtr = IntPtr.Zero;

            try
            {
                l_StrPtr = wine_get_version();
            }
            catch (BadImageFormatException)
            {
            }
            catch (TypeLoadException)
            {
            }

            if (IntPtr.Zero == l_StrPtr)
            {
                return null;
            }

            // We expect an ANSI string in the form "1.1.5".

            return ("Wine: v" + Marshal.PtrToStringAnsi(l_StrPtr));
        }

        [DllImport("ntdll.dll", CallingConvention = CallingConvention.StdCall)]
        private static extern IntPtr wine_get_version();

#region Shutdown

        private static void Application_GuiThreadException(object sender, ThreadExceptionEventArgs e)
        {
            HandleUnhandledException(e.Exception);
        }

        private static void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            Exception l_Exc;

            try
            {
                l_Exc = (Exception)e.ExceptionObject;
            }
            catch (InvalidCastException)
            {
                End();

                l_Exc = null;
            }

            HandleUnhandledException(l_Exc);
        }

        private static void HandleUnhandledException(Exception a_Exc)
        {
            // The support library is already up when we land here.

            // Unregister exception handlers to avoid recursive calls.
            Application.ThreadException -= Application_GuiThreadException;
            AppDomain.CurrentDomain.UnhandledException -= CurrentDomain_UnhandledException;

            string l_DestFileName = LongPath.Combine(sp_ClassicAppDataPath, c_AppName + "CrashLog.txt");
            string l_FrameworkStr;
            string l_AdditionalPlatformInfoStr;
            string l_RuntimeStr = Software.GetRuntimeString(out l_FrameworkStr, out l_AdditionalPlatformInfoStr);

            StringBuilder l_Bld = new StringBuilder();

            if (false == File.Exists(l_DestFileName))
            {
                l_Bld.AppendFormat("This is the {0} crash log. Please mail the content of{1}this file anonymously to the mail address {2}{1}(via the I2P mail system). It can provide invaluable help in{1}fixing the problem.", c_AppName, Environment.NewLine, c_ProjectEmailAddr);
                l_Bld.AppendLine();
                l_Bld.AppendLine();
            }

            l_Bld.AppendLine(s_ReleaseInfo.f_InvarReleaseStr);
            l_Bld.AppendLine(l_RuntimeStr);

            if (null != l_FrameworkStr)
            {
                l_Bld.AppendLine(l_FrameworkStr);
            }

            if (null != l_AdditionalPlatformInfoStr)
            {
                l_Bld.AppendLine(l_AdditionalPlatformInfoStr);
            }

            l_Bld.AppendLine("Exception: " + a_Exc.GetType());

            Exception l_InnerExc = a_Exc;

            for (uint i = 1; 4 != i; i++)
            {
                l_InnerExc = l_InnerExc.InnerException;

                if (null == l_InnerExc)
                {
                    break;
                }

                l_Bld.Append((char)0x20, (int)i);
                l_Bld.AppendLine("Inner Exception: " + l_InnerExc.GetType());
                l_Bld.AppendLine("Inner Message: " + l_InnerExc.Message);
            }

            l_Bld.AppendLine("Message: " + a_Exc.Message);
            string l_ThreadName = Thread.CurrentThread.Name;
            bool l_HasName = (null != l_ThreadName &&
                              0 != l_ThreadName.Length);
            l_Bld.AppendLine("Thread: " + (false == l_HasName ? "<unnamed>" : l_ThreadName));
            l_Bld.AppendLine("Stack trace: " + a_Exc.StackTrace);
            l_Bld.AppendLine(new string('=', 40));
            l_Bld.AppendLine();

            try
            {
                File.AppendAllText(l_DestFileName, l_Bld.ToString());
            }
            catch
            {
            }

            try
            {
                Process.Start(sp_ModuleName, "-CrashMessage");
            }
            catch (System.ComponentModel.Win32Exception)
            {
                try
                {
                    ShowCrashMessage(l_DestFileName);
                }
                catch
                {
                }
            }

            // The timeline is wrecked beyond repair.

#if DEBUG
            throw (a_Exc);
#else // DEBUG
            End();
#endif // DEBUG
        }

        private static void ShowCrashMessage(string a_CrashLogFilePath)
        {
            string l_CrashMsg = string.Format("{0} crashed!\n\nSorry for this. You might want to check out the project's webpage to find out if an improved version is already available.\n\nIf you are using the latest version, please report this problem. Details on the nature of the crash should have been written to the following file:\n\n\"{1}\"\n\nMailing the content of this file to {2} (via the I2P mail system) can provide invaluable help in fixing the problem.", c_AppName, a_CrashLogFilePath, c_ProjectEmailAddr);
            MessageBox.Show(l_CrashMsg, s_ReleaseInfo.f_InvarAppStr, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private static void End()
        {
            // This routine will not return control.

            Environment.Exit(0);
        }

        private static void SystemEvents_PowerModeChanged(object a_Sender,
                                                          Microsoft.Win32.PowerModeChangedEventArgs e)
        {
            switch (e.Mode)
            {
                case Microsoft.Win32.PowerModes.Resume:
                case Microsoft.Win32.PowerModes.Suspend:
                    Util.ReseedRandomProvider();
                    break;
            }
        }

#endregion

        public static void NotifyNewerVersionIsAvailable(uint a_Version,
                                                         DateTime a_BuildDate)
        {
            // This routine may get called with any thread.

            s_LaterVersionAvailable = true;
            s_NewerVersion = Math.Max(s_NewerVersion, a_Version);
        }

        private static bool LoadAppDataMem(TextSet a_TextSet,
                                           string a_CorePath,
                                           string a_IncomplPath,
                                           out HardMemory a_CoreMem,
                                           out HardMemory a_IncomplMem)
        {
            // Return can be null.

            string l_SubErrMsg;
            bool l_CoreLoaded;
            a_CoreMem = HardMemory.LoadOrCreate(a_TextSet, HardMemory.AppDataRole.Core, HardMemory.c_NoAppDataSetNum, a_CorePath, true, false, out l_CoreLoaded, out l_SubErrMsg);

            if (null == a_CoreMem)
            {
                a_IncomplMem = null;

                string l_Msg = a_TextSet.BuildFileErrMsg(TextKeys.MsgHardMemErrFormat, a_CorePath, l_SubErrMsg);
                ErrMsg(a_TextSet, l_Msg);

                return false;
            }

            bool l_CoreCreated = !l_CoreLoaded;
            bool l_IncomplLoaded;
            a_IncomplMem = HardMemory.LoadOrCreate(a_TextSet, HardMemory.AppDataRole.Incompl, a_CoreMem.f_AppDataSetNum, a_IncomplPath, l_CoreLoaded, l_CoreCreated, out l_IncomplLoaded, out l_SubErrMsg);

            if (null == a_IncomplMem)
            {
                a_CoreMem.Close(false, null);
                a_CoreMem = null;

                string l_Msg = a_TextSet.BuildFileErrMsg(TextKeys.MsgHardMemErrFormat, a_IncomplPath, l_SubErrMsg);
                ErrMsg(a_TextSet, l_Msg);

                return false;
            }

            return true;
        }

        public static bool DoesAnyAppDataFileExist(string a_Dir)
        {
            string[] l_FileNames = BuildAppDataFileNames(a_Dir);

            foreach (string l_FileName in l_FileNames)
            {
                if (true == System.IO.File.Exists(l_FileName))
                {
                    return true;
                }
            }

            return false;
        }

        public static string[] BuildAppDataFileNames(string a_Dir)
        {
            string[] l_FileNames = new string[4];
            l_FileNames[0] = LongPath.Combine(a_Dir, c_CoreAppDataFilePrefix + HardMemory.c_IndexFileExt);
            l_FileNames[1] = LongPath.Combine(a_Dir, c_CoreAppDataFilePrefix + HardMemory.c_BodyFileExt);
            l_FileNames[2] = LongPath.Combine(a_Dir, c_IncomplAppDataFilePrefix + HardMemory.c_IndexFileExt);
            l_FileNames[3] = LongPath.Combine(a_Dir, c_IncomplAppDataFilePrefix + HardMemory.c_BodyFileExt);

            return l_FileNames;
        }

        private static void ErrMsg(TextSet a_TextSet,
                                   TextKeys a_Key)
        {
            ErrMsg(a_TextSet, a_TextSet[a_Key]);
        }

        private static void ErrMsg(TextSet a_TextSet,
                                   string a_Text)
        {
            // a_TextSet may be null. This routine should only be used
            // before a main window was created.

            if (null == a_TextSet)
            {
                MessageBox.Show(a_Text, s_ReleaseInfo.f_InvarReleaseStr, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                CustomMsgBox.ShowMsgBox(a_TextSet, null, a_TextSet.f_AppNameWithVer, a_Text, MessageBoxIcon.Error, MessageBoxButtons.OK, DialogResult.OK);
            }
        }

        public static void SetMainForm(MainForm a_Form)
        {
            s_MainForm = a_Form;
        }

        public static bool ExternalFocusMainForm()
        {
            // Caution: This routine is not called by the GUI thread
            // that belongs to this process.

            MainForm l_Form = Util.VolatileRead(ref s_MainForm);

            if (null != l_Form)
            {
                return l_Form.f_Invoker.InvokeAsync(new ArglessHandler(l_Form.ExternalFocusMainForm), null);
            }

            return false;
        }

        public static bool ExternalAcceptShellInput(string a_Stmt)
        {
            // Caution: This routine is not called by the GUI thread
            // that belongs to this process.

            MainForm l_Form = Util.VolatileRead(ref s_MainForm);

            if (null != l_Form)
            {
                return l_Form.f_Invoker.InvokeAsync(new MainForm.ProcessShellInputHandler(l_Form.ExternalHandleShellInput), a_Stmt);
            }

            return false;
        }

        public static Stream OpenResource(string a_ResourceName,
                                          bool a_Deflated)
        {
            string l_EffName = (a_ResourceName + (false == a_Deflated ? string.Empty : ".deflated"));

            System.Reflection.Assembly l_Assembly = System.Reflection.Assembly.GetExecutingAssembly();
            Stream l_Strm = l_Assembly.GetManifestResourceStream(l_EffName);

            if (null == l_Strm)
            {
                // We may land here if we were built through Visual
                // Studio.

                System.Resources.ResourceManager l_ResMngr = new System.Resources.ResourceManager("NB.Resources", l_Assembly);
                object l_ResObj = l_ResMngr.GetObject(l_EffName);

                l_Strm = new MemoryStream((byte[])l_ResObj, false);
            }

            if (true == a_Deflated)
            {
                l_Strm = Util.InflateStream(l_Strm);
            }

            return l_Strm;
        }

        private static bool IsRunningOnMono()
        {
            return (null != Type.GetType("Mono.Runtime"));
        }

        private static string GetDefaultTempPath(string a_FallbackPath)
        {
            string l_Path;

            try
            {
                // Usually the returned path has an ending dir
                // separator.
                l_Path = Path.GetTempPath();
            }
            catch (System.Security.SecurityException)
            {
                return a_FallbackPath;
            }

            return PathNormalizer.NormalizeStatic(l_Path, false);
        }

        private static string BuildDefaultPreviewFolder(string a_DefaultTempPath)
        {
            return LongPath.Combine(a_DefaultTempPath, c_PreviewSubFolder);
        }

        public static bool HasAppDataDriveAvailBytes(bool a_IncludeOutputDir,
                                                     ulong a_Threshold)
        {
            // If the space cannot be queried, the routine returns
            // a nonzero value.

            ulong l_AvailBytes;

            return (false == GetAvailDiskSpace(a_IncludeOutputDir, out l_AvailBytes) ||
                    l_AvailBytes >= a_Threshold);
        }

        public static bool GetAvailDiskSpace(bool a_IncludeOutputDir,
                                             out ulong a_AvailBytes)
        {
            ulong l_Space;

            if (false == FHandle.GetAvailableVolumeSpace(s_BootConfig.p_AppDataDir, out l_Space))
            {
                a_AvailBytes = 0;

                return false;
            }

            if (true == a_IncludeOutputDir)
            {
                string l_LastOutputDir = Util.VolatileRead(ref s_LastOutputDir);
                ulong l_OutputDirSpace;

                if (null != l_LastOutputDir &&
                    true == FHandle.GetAvailableVolumeSpace(l_LastOutputDir, out l_OutputDirSpace))
                {
                    l_Space = Math.Min(l_Space, l_OutputDirSpace);
                }
            }

            a_AvailBytes = l_Space;

            return true;
        }

        public static StorableTime GetStorableTime()
        {
            return StorableTime.FromTime(Time.SysTimeNow());
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ReleaseInfo
        {
            // f_ReleaseNumber is one based.
            public readonly byte f_ReleaseNumber;
            public readonly byte f_MajorVersion;
            public readonly byte f_MinorVersion;
            public readonly byte f_Revision;
            public string f_ReleaseCodename;
            public readonly DateTime f_BuildDate;

            public readonly string f_VersionStr;
            public readonly string f_InvarAppStr;
            public readonly string f_InvarReleaseStr;

            public ReleaseInfo(uint a_ReleaseNumber,
                               uint a_MajorVersion,
                               uint a_MinorVersion,
                               uint a_Revision,
                               string a_ReleaseCodename,
                               int a_Year,
                               int a_Month,
                               int a_Day)
            {
                f_ReleaseNumber = (byte)a_ReleaseNumber;
                f_MajorVersion = (byte)a_MajorVersion;
                f_MinorVersion = (byte)a_MinorVersion;
                f_Revision = (byte)a_Revision;
                f_ReleaseCodename = a_ReleaseCodename;
                f_BuildDate = new DateTime(a_Year, a_Month, a_Day, 0, 0, 0, 0, Util.s_InvariantCulture.Calendar, DateTimeKind.Utc);

                CultureInfo l_InvCulture = CultureInfo.InvariantCulture;
                f_VersionStr = string.Format(l_InvCulture, "{0}.{1}.{2}", a_MajorVersion, a_MinorVersion, a_Revision);
                f_InvarAppStr = string.Format("{0} {1}", c_AppName, f_VersionStr);
                f_InvarReleaseStr = f_InvarAppStr;
#if MONO
                f_InvarReleaseStr += " (MONO)";
#endif
            }

            public int GetReleaseAgeInDays()
            {
                // The return value can be negative of course.

                return (DateTime.UtcNow - f_BuildDate).Days;
            }

            public bool IsOneYearOld()
            {
                return (366 <= GetReleaseAgeInDays());
            }

            public string BuildReleaseNotesFileName()
            {
                return string.Format(Util.s_InvariantCulture, "{0} Release Notes.pdf", f_InvarAppStr);
            }

            public uint EncodeVersion()
            {
                return (((uint)f_MajorVersion << Util.c_BitsPerUShort) |
                        ((uint)f_MinorVersion << Util.c_BitsPerByte) |
                        f_Revision);
            }
        }

        private class InterAppDomainComm
        {
            private const string c_ObjURI = "ServiceClass";
            private const string c_ChannelName = "ipc";

            private static readonly string s_IpcPortName;
            private static readonly string s_ConnectUri;

            private bool f_IsServer;

#if MONO
            private Mutex f_NamedMutex;
#endif

            public bool p_IsServer
            {
                get { return f_IsServer; }
            }

            static InterAppDomainComm()
            {
                // Hash the user name and use it as port name to avoid
                // cross user interaction.
                byte[] l_BinaryUserName = Encoding.Unicode.GetBytes(Environment.UserName);
                byte[] l_Hash = SHA_512.ComputeHash(l_BinaryUserName, 0, l_BinaryUserName.Length);

                s_IpcPortName = (Software.c_AppName + "_0x" + Util.BytesToString(l_Hash));
                s_ConnectUri = ("ipc://" + s_IpcPortName + "/" + c_ObjURI);
            }

            public InterAppDomainComm()
            {
                RegisterServer();
            }

            private void RegisterServer()
            {
                try
                {
                    IpcServerChannel l_Channel = new IpcServerChannel(c_ChannelName, s_IpcPortName);
                    ChannelServices.RegisterChannel(l_Channel, false);
                    RemotingConfiguration.RegisterWellKnownServiceType(typeof(ServiceClass), c_ObjURI, WellKnownObjectMode.SingleCall);
                    RemotingConfiguration.RegisterActivatedServiceType(typeof(ServiceClass));

                    f_IsServer = true;
                }
                catch
                {
                }

#if MONO
                // Mono seems to allow multiple ipc servers (f_IsServer
                // is always true), so we use a named mutex for actual
                // synchronization.

                f_NamedMutex = new Mutex(true, s_IpcPortName, out f_IsServer);
#endif
            }

            private ServiceClass GetRemoteInstance()
            {
                return (ServiceClass)Activator.GetObject(typeof(ServiceClass), s_ConnectUri);
            }

            public bool FocusMainForm(out bool a_ReturnValue)
            {
                // Returns true if the other process could be reached.

                try
                {
                    ServiceClass l_Obj = GetRemoteInstance();

                    if (null != l_Obj)
                    {
                        a_ReturnValue = l_Obj.ExternalFocusMainForm();

                        return true;
                    }
                }
                catch
                {
                }

                a_ReturnValue = false;

                return false;
            }

            public bool AcceptShellInput(string a_Stmt,
                                         out bool a_ReturnValue)
            {
                // Returns true if the other process could be reached.

                try
                {
                    ServiceClass l_Obj = GetRemoteInstance();

                    if (null != l_Obj)
                    {
                        a_ReturnValue = l_Obj.ExternalAcceptShellInput(a_Stmt);

                        return true;
                    }
                }
                catch
                {
                }

                a_ReturnValue = false;

                return false;
            }

            private class ServiceClass : MarshalByRefObject
            {
                public ServiceClass()
                {
                }

                public bool ExternalFocusMainForm()
                {
                    // Gets called from outside the process.

                    return Software.ExternalFocusMainForm();
                }

                public bool ExternalAcceptShellInput(string a_Stmt)
                {
                    // Gets called from outside the process.

                    if (null == a_Stmt)
                    {
                        return false;
                    }

                    return Software.ExternalAcceptShellInput(a_Stmt);
                }
            }
        }
    }

    delegate void ArglessHandler();
    delegate void OneArgHandler(object a_Obj);

    class BootConfig
    {
        private const uint c_Magic = 0x696E3705;
        private const string c_FileName = "DirConfig.dat";
        private const byte c_MyRevision = 2;
        private const byte c_FileHeaderSize = 12;
        private const byte c_EntryHeaderSize = (2 * sizeof(uint));
        // This is an arbitrary limit.
        private const uint c_MaxPlausibleEntrySize = (16 * ushort.MaxValue);

        private const int c_NoLCID = int.MaxValue;
        private const int c_InvCultureLCID = 127;
        public const int c_EnglishCultureLCID = 9;

        private CultureInfo f_LanguageCulture;
        private bool f_UseOSFormatCulture;
        // f_FormatCulture can be null.
        private CultureInfo f_FormatCulture;

        private string f_AppDataDir;
        // f_NewAppDataDir will reference the same string as
        // f_AppDataDir by default.
        private string f_NewAppDataDir;
        private string f_TempDir;

        public string p_AppDataDir
        {
            get { return f_AppDataDir; }
        }

        public string p_NewAppDataDir
        {
            get { return f_NewAppDataDir; }
        }

        public string p_TempDir
        {
            get { return f_TempDir; }
        }

        public string p_TempDirVolatile
        {
            get { return Util.VolatileRead(ref f_TempDir); }
        }

        private BootConfig(CultureInfo a_LanguageCulture,
                           CultureInfo a_FormatCulture,
                           string a_AppDataDir,
                           string a_TempDir)
        {
            // a_FormatCulture may be null.

            f_LanguageCulture = a_LanguageCulture;
            f_UseOSFormatCulture = (null == a_FormatCulture);
            f_FormatCulture = a_FormatCulture;
            f_AppDataDir = a_AppDataDir;
            f_NewAppDataDir = f_AppDataDir;
            f_TempDir = a_TempDir;
        }

        private static BootConfig Create(CultureInfo a_LanguageCulture,
                                         CultureInfo a_FormatCulture)
        {
            // a_FormatCulture may be null.

            return new BootConfig(a_LanguageCulture, a_FormatCulture, SuggestDefaultAppDataDir(), Software.sp_DefaultTempPath);
        }

        public static BootConfig CreateDefault()
        {
            return Create(CultureInfo.GetCultureInfo(c_EnglishCultureLCID), null);
        }

        public static StorageStates Load(out BootConfig a_Dest)
        {
            a_Dest = null;

            string l_FileName = BuildConfigFileName();
            Exception l_Exc;
            FHandle l_SrcFile = FHandle.OpenR(l_FileName, false, FileOptions.None, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

            if (null == l_SrcFile)
            {
                return (false == File.Exists(l_FileName) ? StorageStates.NoData : StorageStates.ReadFailed);
            }

            uint l_Rev;
            int l_LangCultLCID;
            int l_FormatCultLCID;
            string l_AppDataDir = null;
            string l_TempDir = null;

            try
            {
                FileIO l_File = FileIO.CreateReadable(l_SrcFile);
                uint l_Magic = l_File.ReadUInt32LE();
                l_Rev = l_File.ReadUInt32LE();

                if (c_Magic != l_Magic ||
                    0 == l_Rev)
                {
                    return StorageStates.Corruption;
                }

                if (c_MyRevision < l_Rev)
                {
                    return StorageStates.TooAdvanced;
                }

                uint l_Pos = l_File.ReadUInt32LE();
                l_File.SetPosition(l_Pos);
                uint l_Size;
                uint l_Checksum;

                if (1 == l_Rev)
                {
                    // This is for backward compatibility with v1.2.x,
                    // which doesn't include path information.

                    if ((6 * sizeof(uint)) != l_File.p_Length ||
                        0 != l_File.ReadUInt32LE())
                    {
                        return StorageStates.Corruption;
                    }

                    l_Size = 0;
                    l_Checksum = 0;
                }
                else
                {
                    l_Size = l_File.ReadUInt32LE();
                    l_Checksum = l_File.ReadUInt32LE();
                }

                l_LangCultLCID = (int)l_File.ReadUInt32LE();
                l_FormatCultLCID = (int)l_File.ReadUInt32LE();

                if (2 == l_Rev)
                {
                    l_AppDataDir = Software.ReadAbsPathUTF16LEStringEncLen(l_File);
                    l_TempDir = Software.ReadAbsPathUTF16LEStringEncLen(l_File);

                    ByteArray l_Data;
                    l_Pos += c_EntryHeaderSize;
                    l_File.ExposeRange(l_Pos, l_File.p_Position - l_Pos, out l_Data);

                    if (Adler32.ComputeHash(ref l_Data) != l_Checksum)
                    {
                        return StorageStates.Corruption;
                    }
                }
            }
            catch (BufferIO.BufferIOException)
            {
                return StorageStates.ReadFailed;
            }
            finally
            {
                l_SrcFile.Close();
            }

            CultureInfo l_LangCulture;
            CultureInfo l_FormatCulture;

            try
            {
                l_LangCulture = CultureInfo.GetCultureInfo(l_LangCultLCID);
                l_FormatCulture = (c_NoLCID == l_FormatCultLCID ? null : CultureInfo.GetCultureInfo(l_FormatCultLCID));
            }
            catch (ArgumentException)
            {
                return StorageStates.Corruption;
            }

            if (1 == l_Rev)
            {
                a_Dest = BootConfig.Create(l_LangCulture, l_FormatCulture);
            }
            else
            {
                PathNormalizer l_PathNormalizer;
                PathNormalizer.Create(out l_PathNormalizer);
                l_AppDataDir = l_PathNormalizer.Normalize(l_AppDataDir);
                l_TempDir = l_PathNormalizer.Normalize(l_TempDir);

                a_Dest = new BootConfig(l_LangCulture, l_FormatCulture, l_AppDataDir, l_TempDir);
            }

            return StorageStates.Success;
        }

        public bool Save()
        {
            string l_FileName = BuildConfigFileName();
            Exception l_Exc;
            FHandle l_DestFile = FHandle.OpenRWRITE(this.GetType(), l_FileName, FileMode.OpenOrCreate, FileOptions.None, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

            if (null == l_DestFile)
            {
                return false;
            }

            MemoryIO l_Mem = MemoryIO.CreateWritable();
            l_Mem.WriteUInt32LE((uint)f_LanguageCulture.LCID);
            l_Mem.WriteUInt32LE((uint)(false == f_UseOSFormatCulture ? f_FormatCulture.LCID : c_NoLCID));
            l_Mem.WriteUTF16LEStringEncLen(f_AppDataDir);
            l_Mem.WriteUTF16LEStringEncLen(f_TempDir);
            uint l_Size = (c_EntryHeaderSize + l_Mem.p_Length);

            try
            {
                FileIO l_File = FileIO.CreateWritable(l_DestFile);
                uint l_Pos = c_FileHeaderSize;

                try
                {
                    if (c_FileHeaderSize < l_File.p_Length &&
                        c_Magic == l_File.ReadUInt32LE() &&
                        2 == l_File.ReadUInt32LE())
                    {
                        uint l_OldEntryPos = l_File.ReadUInt32LE();

                        if (c_FileHeaderSize <= l_OldEntryPos &&
                            (l_OldEntryPos - c_FileHeaderSize) < l_Size)
                        {
                            // There is not enough space before the old
                            // entry to place the new entry.

                            l_File.SetPosition(l_OldEntryPos);
                            uint l_OldSize = l_File.ReadUInt32LE();

                            if (c_MaxPlausibleEntrySize >= l_OldSize)
                            {
                                // Place the new entry after the old.

                                l_Pos = (l_OldEntryPos + l_OldSize);
                            }
                        }
                    }
                }
                catch (BufferIO.InterpretBufferException)
                {
                }

                ByteArray l_Data;
                l_Mem.ExposeRange(0, l_Size - c_EntryHeaderSize, out l_Data);

                // Write the new entry.

                l_File.SetPosition(l_Pos);
                // This is the full size of the entry, including the
                // entries header.
                l_File.WriteUInt32LE(l_Size);
                l_File.WriteUInt32LE(Adler32.ComputeHash(ref l_Data));
                l_File.WriteBytes(ref l_Data);
                l_File.Flush();

                // Update the file header and reference the new entry.

                l_File.JumpToStart();
                l_File.WriteUInt32LE(c_Magic);
                l_File.WriteUInt32LE(c_MyRevision);
                l_File.WriteUInt32LE(l_Pos);
                l_File.Flush();

                // Check if the file can be shrinked.

                uint l_NewFileSize = (l_Pos + l_Size);

                if (l_File.p_Length > l_NewFileSize)
                {
                    l_File.SetLength(l_NewFileSize);
                    l_File.Flush();
                }
            }
            catch (BufferIO.FileAccessException)
            {
                return false;
            }
            finally
            {
                l_DestFile.Close();
            }

            return true;
        }

        public bool IsDirty(TextSet a_TextSet)
        {
            if (false == BootConfig.AreSameCultures(a_TextSet.f_LanguageCulture, f_LanguageCulture))
            {
                return true;
            }

            if (true == f_UseOSFormatCulture)
            {
                return (false == BootConfig.AreSameCultures(a_TextSet.f_FormatCulture, Software.s_StartupCulture));
            }

            return (false == BootConfig.AreSameCultures(a_TextSet.f_FormatCulture, f_FormatCulture));
        }

        public void SetLanguageCultureToDefault()
        {
            f_LanguageCulture = new CultureInfo(c_EnglishCultureLCID);
        }

        public void SetLanguageCulture(CultureInfo a_LanguageCulture)
        {
            f_LanguageCulture = a_LanguageCulture;
        }

        public bool HasDefaultLanguageCulture()
        {
            return (c_EnglishCultureLCID == f_LanguageCulture.LCID);
        }

        public CultureInfo GetLanguageCulture()
        {
            return f_LanguageCulture;
        }

        public void SetFormatCultureToOSSetting()
        {
            f_UseOSFormatCulture = true;
            f_FormatCulture = null;
        }

        public void SetFormatCulture(CultureInfo a_FormatCulture)
        {
            f_UseOSFormatCulture = false;
            f_FormatCulture = a_FormatCulture;
        }

        public bool ShouldUseOSFormatCulture()
        {
            return f_UseOSFormatCulture;
        }

        public CultureInfo GetFormatCulture()
        {
            return f_FormatCulture;
        }

        public bool HasApplicationDataDirectoryChanged()
        {
            return (f_NewAppDataDir != f_AppDataDir);
        }

        public void SetNewApplicationDataDirectory(string a_NewAppDataDir)
        {
            // May only be used by the GUI.

            f_NewAppDataDir = a_NewAppDataDir;
        }

        private void TakeOverNewApplicationDataDirectory()
        {
            // May only be used by the GUI.

            f_AppDataDir = f_NewAppDataDir;
        }

        public bool TakeOverNewApplicationDataDirectoryAndSave()
        {
            // May only be used by the GUI.

            TakeOverNewApplicationDataDirectory();

            return Save();
        }

        public void SetTempDirectory(string a_TempDir)
        {
            // May only be used by the GUI.

            f_TempDir = a_TempDir;
        }

        private static string BuildConfigFileName()
        {
            return LongPath.Combine(Software.sp_ClassicAppDataPath, c_FileName);
        }

        public static string BuildTranslationFolderName()
        {
            return LongPath.Combine(Software.sp_ModulePath, Software.c_TranslationsSubFolder);
        }

        public static string BuildTranslationFileName(string a_Folder,
                                                      CultureInfo a_Culture)
        {
            return LongPath.Combine(a_Folder, a_Culture.IetfLanguageTag + ".txt");
        }

        public static CultureItem[] ScanForLanguageCultures(string a_DefaultItemDisplName)
        {
            List<CultureItem> l_Ls = new List<CultureItem>(1);
            string l_Folder = BuildTranslationFolderName();

            foreach (CultureInfo l_Cult in CultureInfo.GetCultures(CultureTypes.NeutralCultures | CultureTypes.SpecificCultures))
            {
                if (c_EnglishCultureLCID == l_Cult.LCID ||
                    false == File.Exists(BuildTranslationFileName(l_Folder, l_Cult)))
                {
                    continue;
                }

                l_Ls.Add(new CultureItem(l_Cult, l_Cult.EnglishName));
            }

            return ListToArray(l_Ls, a_DefaultItemDisplName);
        }

        public static CultureItem[] BuildFormatCulturesArray(string a_DefaultItemDisplName)
        {
            CultureInfo[] l_Cultures = CultureInfo.GetCultures(CultureTypes.SpecificCultures);
            List<CultureItem> l_Ls = new List<CultureItem>(l_Cultures.Length + 1);

            foreach (CultureInfo l_Cult in l_Cultures)
            {
                if (c_InvCultureLCID == l_Cult.LCID)
                {
                    // The invariant culture isn't specific, but we
                    // check to be on the safe side.

                    continue;
                }

                l_Ls.Add(new CultureItem(l_Cult, l_Cult.EnglishName));
            }

            return ListToArray(l_Ls, a_DefaultItemDisplName);
        }

        private static CultureItem[] ListToArray(List<CultureItem> a_Ls,
                                                 string a_DefaultItemDisplName)
        {
            a_Ls.Sort();
            a_Ls.Insert(0, new CultureItem(null, a_DefaultItemDisplName));

            return a_Ls.ToArray();
        }

        public static bool AreSameCultures(CultureInfo a_A,
                                           CultureInfo a_B)
        {
            return (a_A.LCID == a_B.LCID);
        }

        public static string SuggestDefaultAppDataDir()
        {
            // The Software class must be initialized somewhat already
            // when this routine gets called.

            string l_ClassicPath = Software.sp_ClassicAppDataPath;
            string l_DefaultPath = Software.sp_DefaultAppDataPath;

            if (l_ClassicPath != l_DefaultPath &&
                true == File.Exists(LongPath.Combine(l_ClassicPath, "Core.body")) &&
                true == File.Exists(LongPath.Combine(l_ClassicPath, "Core.index")))
            {
                // Since there is already relevant data in the classic
                // dir, suggest that one.

                return l_ClassicPath;
            }

            return l_DefaultPath;
        }

        public class CultureItem : IComparable<CultureItem>
        {
            // This class can be used with e.g. a combo box.

            // f_Culture can be null.
            public readonly CultureInfo f_Culture;
            public readonly string f_DisplName;

            public CultureItem(CultureInfo a_Culture,
                               string a_DisplName)
            {
                f_Culture = a_Culture;
                f_DisplName = a_DisplName;
            }

            public bool HasCulture(CultureInfo a_Culture)
            {
                return (null != f_Culture &&
                        true == BootConfig.AreSameCultures(f_Culture, a_Culture));
            }

            public int CompareTo(CultureItem a_Other)
            {
                return this.f_DisplName.CompareTo(a_Other.f_DisplName);
            }

            public override string ToString()
            {
                return f_DisplName;
            }
        }
    }
}

// End of file.
