/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\zlib.cs
 * PURPOSE: Interface to managed ZLIB
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using Support;
using zlibManaged;

namespace NB.Compression
{
    static class zlib
    {
        public static uint GetMaxCompressedSize(uint a_UncompressedSize)
        {
            // The called routine will throw an exception for very
            // high values.
            return (uint)zlibManaged.zlib.compressBound((int)a_UncompressedSize);
        }

        public static bool Compress(ref ByteArray a_Src,
                                    bool a_ShrinkOnly,
                                    out ByteArray a_Dest)
        {
            ArraySegment<byte> l_Out;
            bool l_Succ;

            try
            {
                l_Succ = zlibManaged.zlib.compress2(out l_Out, new ArraySegment<byte>(a_Src.f_Arr, a_Src.f_Idx, a_Src.f_Len), CompressionLevels.Z_BEST_COMPRESSION);
            }
#pragma warning disable 168
            catch (Exception l_Exc)
#pragma warning restore 168
            {
                l_Succ = false;
                l_Out = new ArraySegment<byte>();

                Util.DebugBreak();
            }

            if (false == l_Succ ||
                (true == a_ShrinkOnly &&
                 l_Out.Count >= a_Src.f_Len))
            {
                a_Dest = new ByteArray();

                return false;
            }

            a_Dest = new ByteArray(l_Out.Array, l_Out.Offset, l_Out.Count);

            return true;
        }

        public static bool Uncompress(ref ByteArray a_Src,
                                      uint a_MaxDecompressedSize,
                                      out ByteArray a_Dest)
        {
            ArraySegment<byte> l_Out;
            bool l_Succ;

            try
            {
                l_Succ = zlibManaged.zlib.uncompress(out l_Out, new ArraySegment<byte>(a_Src.f_Arr, a_Src.f_Idx, a_Src.f_Len), (int)a_MaxDecompressedSize);
            }
#pragma warning disable 168
            catch (Exception l_Exc)
#pragma warning restore 168
            {
                l_Succ = false;
                l_Out = new ArraySegment<byte>();

                Util.DebugBreak();
            }

            if (false == l_Succ)
            {
                a_Dest = new ByteArray();

                return false;
            }

            a_Dest = new ByteArray(l_Out.Array, l_Out.Offset, l_Out.Count);

            return true;
        }

        public static bool ToCompressedBuffer(ref ByteArray a_SrcData,
                                              uint a_Magic,
                                              uint a_FormatRevision,
                                              BufferIO a_DestBuffer)
        {
            // This routine is abortion-safe (the global state will be
            // just fine if an exception occurs while operating on
            // a_DestBuffer).

            ByteArray l_CompressedData;

            if (false == Compress(ref a_SrcData, false, out l_CompressedData))
            {
                return false;
            }

            a_DestBuffer.WriteUInt32LE(a_Magic | (a_FormatRevision << 24));
            a_DestBuffer.WriteUInt32LE((uint)a_SrcData.f_Len);
            a_DestBuffer.WriteUInt32LE((uint)l_CompressedData.f_Len);
            a_DestBuffer.WriteBytes(ref l_CompressedData);

            return true;
        }

        public static DecompressBufferResults FromCompressedBuffer(BufferIO a_SrcBuffer,
                                                                   uint a_ExpectedMagic,
                                                                   uint a_MaxFormatRevision,
                                                                   uint a_MaxUncompressedSize,
                                                                   out uint a_ActualFormatRevision,
                                                                   out ByteArray a_DestData)
        {
            // This routine is abortion-safe.

            const byte l_MinSize = (3 * sizeof(uint));
            a_ActualFormatRevision = 0;
            a_DestData = new ByteArray();

            if (l_MinSize > a_SrcBuffer.GetRemaining())
            {
                return DecompressBufferResults.ErrSourceBufferTooShort;
            }

            uint l_MagicAndRevision = a_SrcBuffer.ReadUInt32LE();

            if ((l_MagicAndRevision & 0xFFFFFF) != a_ExpectedMagic)
            {
                return DecompressBufferResults.ErrInvalidMagic;
            }

            uint l_Revision = (l_MagicAndRevision >> 24);

            if (l_Revision > a_MaxFormatRevision)
            {
                return DecompressBufferResults.ErrFormatTooAdvanced;
            }

            uint l_RawSize = a_SrcBuffer.ReadUInt32LE();
            uint l_CompressedSize = a_SrcBuffer.ReadUInt32LE();

            if (l_RawSize > a_MaxUncompressedSize)
            {
                return DecompressBufferResults.ErrUncompressedSizeExceedsLimit;
            }

            if (a_SrcBuffer.GetRemaining() < l_CompressedSize)
            {
                return DecompressBufferResults.ErrSourceBufferTooShort;
            }

            ByteArray l_SrcData = new ByteArray(a_SrcBuffer.ReadBytes(l_CompressedSize));

            if (false == Uncompress(ref l_SrcData, a_MaxUncompressedSize, out a_DestData))
            {
                return DecompressBufferResults.ErrCompressedBufferCorrupt;
            }

            a_ActualFormatRevision = l_Revision;

            return DecompressBufferResults.Success;
        }

        public enum DecompressBufferResults : byte
        {
            Success = 0,

            ErrSourceBufferTooShort = 1,
            ErrInvalidMagic = 2,
            ErrFormatTooAdvanced = 3,
            ErrUncompressedSizeExceedsLimit = 4,
            ErrCompressedBufferCorrupt = 5,
        }
    }
}

// End of file.
