/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Native\FileIo.c
 * PURPOSE: Another file input/output interface
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

#include <malloc.h>

#define INCLUDE_WINDOWS
#include "general.h"

// The following code exposes a file io interface that works without
// the overhead of first adjusting the file pointer (using
// NtSetInformationFile) and then requiring another syscall for the
// actual read or write. So this interface doesn't adhere to the
// stream metaphor, but treats a file like a (resizable) byte array.

#define NTAPI                           __stdcall
#define NT_SUCCESS(StatCode)            (0 <= (NTSTATUS)(StatCode))
#define STATUS_SUCCESS                  ((NTSTATUS)0x00000000L)
#define STATUS_END_OF_FILE              ((NTSTATUS)0xC0000011L)

#define ASYNC_CONTEXT_DATA_ID_MAGIC     (0x6AC1D73B)

typedef PVOID PIO_APC_ROUTINE;

typedef struct _IO_STATUS_BLOCK
{
    union _NAMELESS
    {
        NTSTATUS Status;
        PVOID Pointer;
    }
    NAMELESS;

    ULONG_PTR Information;
}
IO_STATUS_BLOCK, * PIO_STATUS_BLOCK;

typedef struct _ASYNC_CONTEXT
{
    UINT32 f_DataID;
    UINT32 f_DataSize;

    IO_STATUS_BLOCK f_Block;
    HANDLE f_CompletionEventHandle;
    volatile UINT32 f_EndLock;
    NBNAT_RESULTS f_AsyncResult;
}
ASYNC_CONTEXT;

// Prototypes.

typedef NTSTATUS NTAPI NtReadWriteFileFunc(IN HANDLE FileHandle,
                                           IN OPTIONAL HANDLE Event,
                                           IN OPTIONAL PIO_APC_ROUTINE ApcRoutine,
                                           IN OPTIONAL PVOID ApcContext,
                                           METAIN OUT PIO_STATUS_BLOCK IoStatusBlock,
                                           // Buffer is input for NtWriteFile() and
                                           // output for NtReadFile().
                                           IN OUT PVOID Buffer,
                                           IN ULONG Length,
                                           IN PLARGE_INTEGER ByteOffset,
                                           IN OPTIONAL PULONG Key);

typedef NTSTATUS NTAPI NtWaitForSingleObjectFunc(IN HANDLE Handle,
                                                 IN BOOLEAN Alertable,
                                                 IN OPTIONAL PLARGE_INTEGER Timeout);

// Globals (all can be NULL).

static HMODULE g_ntdllModule = NULL;
static NtReadWriteFileFunc * g_NtReadFile = NULL;
static NtReadWriteFileFunc * g_NtWriteFile = NULL;
static NtWaitForSingleObjectFunc * g_NtWaitForSingleObject = NULL;

BOOLEAN
FileInitOrShutdownInternal(IN BOOLEAN const a_Init)
{
    // This routine is used for initialization and shutdown, regarding
    // global data. Returns a nonzero value if the desired final state
    // was reached.

    if (FALSE == a_Init)
    {
        if (NULL == g_ntdllModule)
        {
            return TRUE;
        }
    }
    else
    {
        if (NULL != g_ntdllModule)
        {
            // We're already up.

            return TRUE;
        }

        g_ntdllModule = LoadLibrary(L"ntdll.dll");

        if (NULL == g_ntdllModule)
        {
            return FALSE;
        }

        // The following three functions are exposed by ntdll.dll
        // on NT 5.0 and later.
        g_NtReadFile = (NtReadWriteFileFunc *)GetProcAddress(g_ntdllModule, "NtReadFile");
        g_NtWriteFile = (NtReadWriteFileFunc *)GetProcAddress(g_ntdllModule, "NtWriteFile");
        g_NtWaitForSingleObject = (NtWaitForSingleObjectFunc *)GetProcAddress(g_ntdllModule, "NtWaitForSingleObject");

        if (NULL != g_NtReadFile &&
            NULL != g_NtWriteFile &&
            NULL != g_NtWaitForSingleObject)
        {
            return TRUE;
        }
    }

    ASSERT(NULL != g_ntdllModule);

    {
        BOOL const l_Succ = FreeLibrary(g_ntdllModule);

        ASSERT(FALSE != l_Succ);
        UNUSED_STACK_ITEM(l_Succ);
    }

    g_ntdllModule = NULL;
    g_NtReadFile = NULL;
    g_NtWriteFile = NULL;
    g_NtWaitForSingleObject = NULL;

    return !a_Init;
}

EXPORT BOOL CALLCONV
FileCreateAsyncContext_EXPORT(METAIN OUT ASYNC_CONTEXT ** const a_AsyncContext)
{
    // This routine is thread-safe. On success, the caller must later
    // release the newly created context through FileDestroyAsyncContext().

    // We're using an auto-reset event.
    HANDLE l_EventHandle = CreateEvent((LPSECURITY_ATTRIBUTES)NULL, FALSE, FALSE, (LPCWSTR)NULL);
    ASYNC_CONTEXT * l_Cntxt;

    if (NULL == l_EventHandle)
    {
        goto FAIL;
    }

    l_Cntxt = (ASYNC_CONTEXT *)_aligned_malloc(sizeof(ASYNC_CONTEXT), sizeof(UINT64));

    if (NULL == l_Cntxt)
    {
        BOOL l_Succ = CloseHandle(l_EventHandle);

        ASSERT(FALSE != l_Succ);
        UNUSED_STACK_ITEM(l_Succ);

        goto FAIL;
    }

    UtilZeroMem(l_Cntxt, sizeof(ASYNC_CONTEXT));

    l_Cntxt->f_DataID = ASYNC_CONTEXT_DATA_ID_MAGIC;
    l_Cntxt->f_DataSize = sizeof(ASYNC_CONTEXT);
    l_Cntxt->f_CompletionEventHandle = l_EventHandle;

    *a_AsyncContext = l_Cntxt;

    return TRUE;

    FAIL:

    *a_AsyncContext = NULL;

    return FALSE;
}

EXPORT BOOL CALLCONV
FileDestroyAsyncContext_EXPORT(METAIN OUT ASYNC_CONTEXT * const a_AsyncContext)
{
    // This routine is thread-safe.

    BOOL l_Succ;

    if (ASYNC_CONTEXT_DATA_ID_MAGIC != a_AsyncContext->f_DataID)
    {
        ASSERT(FALSE);

        return FALSE;
    }

    ASSERT(sizeof(ASYNC_CONTEXT) == a_AsyncContext->f_DataSize);

    l_Succ = CloseHandle(a_AsyncContext->f_CompletionEventHandle);

    ASSERT(FALSE != l_Succ);

    a_AsyncContext->f_DataID = 0;
    _aligned_free(a_AsyncContext);

    return TRUE;
}

EXPORT NBNAT_RESULTS CALLCONV
FileReadWrite_EXPORT(IN BOOL const a_Write,
                     IN HANDLE const a_FileHandle,
                     IN UINT64 const a_FileOffset,
                     IN OUT BYTE * const a_Buf,
                     IN OUT UINT32 * const a_Size,
                     IN OPTIONAL ASYNC_CONTEXT * const a_AsyncContext)
{
    // This routine is thread-safe. It can be used with file handles
    // opened for synchronous and asynchronous access. a_AsyncContext
    // may be NULL and should be NULL for handles opened for synchronous
    // access (the operating system will treat any async access on
    // sync handles like sync access and block the thread). All threads
    // must have ceased using the passed a_AsyncContext on behalf of a
    // prior io operation before calling this routine (it's in the
    // callers responsibility, there are no checks here). Note that
    // it's not legal to modify/release a_Buf while it's used in an
    // asynchronous operation. On successful return, a_Size will
    // indicate the bytes actually read/written.

    NBNAT_RESULTS l_Result;
    LARGE_INTEGER l_Offset;
    NTSTATUS l_Status;
    IO_STATUS_BLOCK l_Block;
    IO_STATUS_BLOCK * l_EffBlock;
    BOOLEAN const l_Async = (NULL != a_AsyncContext);
    NtReadWriteFileFunc * const l_NtRWFunc = (FALSE == a_Write ? g_NtReadFile : g_NtWriteFile);

    if (NULL == l_NtRWFunc ||
        NULL == g_NtWaitForSingleObject)
    {
        ASSERT(FALSE);

        l_Result = NBNAT_FAILURE;

        goto EXIT;
    }

    if (FALSE != l_Async &&
        (ASYNC_CONTEXT_DATA_ID_MAGIC != a_AsyncContext->f_DataID ||
         0 != a_AsyncContext->f_EndLock))
    {
        ASSERT(FALSE);

        l_Result = NBNAT_FAILURE;

        goto EXIT;
    }

    l_Offset.QuadPart = (LONGLONG)a_FileOffset;
    l_EffBlock = (FALSE == l_Async ? &l_Block : &a_AsyncContext->f_Block);
    l_Status = l_NtRWFunc(a_FileHandle, FALSE == l_Async ? (HANDLE)NULL : a_AsyncContext->f_CompletionEventHandle, (PIO_APC_ROUTINE)NULL, (PVOID)NULL, l_EffBlock, a_Buf, *a_Size, &l_Offset, (PULONG)NULL);

    if (STATUS_PENDING == l_Status)
    {
        NTSTATUS l_WaitStatus;

        if (FALSE != l_Async)
        {
            // The caller will have to call FileTryEndAsyncOperation()
            // later.

            l_Result = NBNAT_PENDING;

            goto EXIT;
        }

        // We are operating sync and must block in this case, since
        // NtXXXFile() still has our l_Block referenced. We'll wait
        // on the file handle directly.

        l_WaitStatus = g_NtWaitForSingleObject(a_FileHandle, FALSE, (PLARGE_INTEGER)NULL);

        if (FALSE == NT_SUCCESS(l_WaitStatus))
        {
            l_Result = NBNAT_FAILURE;

            goto EXIT;
        }

        l_Status = l_EffBlock->NAMELESS.Status;
    }

    if (FALSE != NT_SUCCESS(l_Status))
    {
        // The io operation succeeded.

        *a_Size = (UINT32)l_EffBlock->Information;

        if (FALSE != l_Async)
        {
            a_AsyncContext->f_AsyncResult = NBNAT_SUCCESS;
        }

        return NBNAT_SUCCESS;
    }

    l_Result = NBNAT_FAILURE;

    EXIT:

    ASSERT(NBNAT_SUCCESS != l_Result);

    *a_Size = 0;

    if (FALSE != l_Async)
    {
        a_AsyncContext->f_AsyncResult = l_Result;
    }

    return l_Result;
}

#if RW_SINGLE_EXPORTS

// Not used.

EXPORT NBNAT_RESULTS CALLCONV
FileRead_EXPORT(IN HANDLE const a_FileHandle,
                IN UINT64 const a_FileOffset,
                METAIN OUT BYTE * const a_DestBuf,
                IN OUT UINT32 * const a_ReadSize,
                IN OPTIONAL ASYNC_CONTEXT * const a_AsyncContext)
{
    return FileReadWrite_EXPORT(FALSE, a_FileHandle, a_FileOffset, a_DestBuf, a_ReadSize, a_AsyncContext);
}

EXPORT NBNAT_RESULTS CALLCONV
FileWrite_EXPORT(IN HANDLE const a_FileHandle,
                 IN UINT64 const a_FileOffset,
                 IN BYTE const * const a_SrcBuf,
                 IN OUT UINT32 * const a_WriteSize,
                 IN OPTIONAL ASYNC_CONTEXT * const a_AsyncContext)
{
    return FileReadWrite_EXPORT(TRUE, a_FileHandle, a_FileOffset, (BYTE *)a_SrcBuf, a_WriteSize, a_AsyncContext);
}

#endif // RW_SINGLE_EXPORTS

EXPORT NBNAT_RESULTS CALLCONV
FileTryEndAsyncOperation_EXPORT(IN OUT ASYNC_CONTEXT * const a_AsyncContext,
                                IN BOOL const a_MayWait,
                                METAIN OUT UINT32 * const a_Size)
{
    // This routine is thread-safe. After a async io operation was
    // started using FileReadWrite() with the return value NBNAT_PENDING,
    // this routine must be called with the same async context until
    // this routine returns something other than NBNAT_PENDING. As long
    // as this routine returns NBNAT_PENDING, a_AsyncContext may not be
    // used to begin another io request. On successful return, a_Size
    // will indicate the bytes actually read/written.

    NBNAT_RESULTS l_Result;
    LARGE_INTEGER l_Timeout;
    BOOLEAN l_ReleaseLock = FALSE;
    NTSTATUS l_WaitStatus;

    if (NULL == g_NtWaitForSingleObject ||
        ASYNC_CONTEXT_DATA_ID_MAGIC != a_AsyncContext->f_DataID)
    {
        ASSERT(FALSE);

        l_Result = NBNAT_FAILURE;

        goto EXIT;
    }

    if (FALSE == a_MayWait)
    {
        // No wait shall occur if the operation cannot be completed
        // immediately.

        l_Timeout.QuadPart = 0;
    }

    // Since we're using an automatic event, we can't wait more often
    // than once. We've a lock to make sure we won't.

    if (FALSE == UtilTryAquireLock(&a_AsyncContext->f_EndLock))
    {
        // Another thread is already calling NtWaitForSingleObject().

        l_Result = NBNAT_PENDING;

        goto EXIT;
    }

    l_ReleaseLock = TRUE;

    if (NBNAT_PENDING != a_AsyncContext->f_AsyncResult)
    {
        // We've completed and will return the same status again we've
        // already returned.

        l_Result = a_AsyncContext->f_AsyncResult;

        goto EXIT;
    }

    l_WaitStatus = g_NtWaitForSingleObject(a_AsyncContext->f_CompletionEventHandle, FALSE, FALSE == a_MayWait ? &l_Timeout : (PLARGE_INTEGER)NULL);

    if (STATUS_TIMEOUT == l_WaitStatus)
    {
        ASSERT(FALSE == a_MayWait);

        l_Result = NBNAT_PENDING;

        goto EXIT;
    }

    // We've left the pending state.

    if (FALSE == NT_SUCCESS(l_WaitStatus) ||
        FALSE == NT_SUCCESS(a_AsyncContext->f_Block.NAMELESS.Status))
    {
        l_Result = NBNAT_FAILURE;
    }
    else
    {
        l_Result = NBNAT_SUCCESS;
    }

    a_AsyncContext->f_AsyncResult = l_Result;

    EXIT:

    if (FALSE != l_ReleaseLock)
    {
        UtilReleaseLock(&a_AsyncContext->f_EndLock);
    }

    *a_Size = (NBNAT_SUCCESS == l_Result ? (UINT32)a_AsyncContext->f_Block.Information : 0);

    return l_Result;
}

// End of file.
