/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Native\Library.c
 * PURPOSE: DllMain()
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

#define INCLUDE_WINDOWS
#include "general.h"

// Our file io capabilities depend on the availability of a number
// of native operating system routines, which we will evaluate at
// runtime. Introduced with 1.4.9.
#define NATIVE_CAP_FLAG_FILE_IO                 (0x1)

// Our network capabilities depend on whether WSAStartup() returned
// a success code or not. Introduced with 1.4.10.
#define NATIVE_CAP_FLAG_SOCKET_CHECKING         (0x2)

// Introduced with 1.4.14.
#define NATIVE_CAP_FLAG_HIGH_RESOLUTION_TIMER   (0x4)

// Prototypes.

void
RandInitInternalWithoutSeeding();

void
RandCleanUpInternal();

BOOLEAN
FileInitOrShutdownInternal(IN BOOLEAN const a_Init);

BOOLEAN
SocketInitOrShutdownInternal(IN BOOLEAN const a_Init);

BOOLEAN
HiResTimerInitOrShutdownInternal(IN BOOLEAN const a_Init);

// Globals.

static UINT g_CapabilityFlags = 0;
static BOOLEAN g_MultipleProcessors = FALSE;

BOOL APIENTRY
DllMain(IN HMODULE const a_Module,
        IN DWORD const a_ReasonForCall,
        IN LPVOID const a_Reserved)
{
    UNUSED_STACK_ITEM(a_Reserved);

    switch (a_ReasonForCall)
    {
        case DLL_PROCESS_ATTACH:
        {
            SYSTEM_INFO l_SysInfo;
            BOOL const l_Succ = DisableThreadLibraryCalls(a_Module);

            ASSERT(FALSE != l_Succ);
            UNUSED_STACK_ITEM(l_Succ);

            // Find out the number of processors.

            GetSystemInfo(&l_SysInfo);
            g_MultipleProcessors = (2 <= l_SysInfo.dwNumberOfProcessors);

            // Initialize other modules.

            // We won't seed the random number generator here, since
            // it's the responsibility of the RNG consumer to seed
            // before using it for the first time.
            RandInitInternalWithoutSeeding();

            if (FALSE != FileInitOrShutdownInternal(TRUE))
            {
                g_CapabilityFlags = NATIVE_CAP_FLAG_FILE_IO;
            }

            if (FALSE != SocketInitOrShutdownInternal(TRUE))
            {
                g_CapabilityFlags |= NATIVE_CAP_FLAG_SOCKET_CHECKING;
            }

            if (FALSE != HiResTimerInitOrShutdownInternal(TRUE))
            {
                g_CapabilityFlags |= NATIVE_CAP_FLAG_HIGH_RESOLUTION_TIMER;
            }

            break;
        }

        case DLL_PROCESS_DETACH:
            RandCleanUpInternal();
            FileInitOrShutdownInternal(FALSE);
            SocketInitOrShutdownInternal(FALSE);
            HiResTimerInitOrShutdownInternal(FALSE);
            break;

        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
            break;

        default:
            ASSERT(FALSE);
    }

    return TRUE;
}

EXPORT UINT64 CALLCONV
GetLibraryVersion_EXPORT()
{
    const UINT16 l_MajorVersion = 1;
    const BYTE l_MinorVersion = 4;
    const BYTE l_Revision = 25;

    return (((UINT64)g_CapabilityFlags << BITS_PER_UINT32) |
            ((UINT32)l_MajorVersion << BITS_PER_UINT16) |
            ((UINT32)l_MinorVersion << BITS_PER_BYTE) |
            l_Revision);
}

static void
UtilYieldProcessorInternal(IN UINT32 const a_Times)
{
    UINT32 i;

    for (i = a_Times; 0 != i; i--)
    {
        YieldProcessor();
    }
}

static void
UtilSleepBrieflyInternal(IN OUT UINT32 * const a_Count)
{
    // *a_Count should be zero for the first call in a series.

    const UINT32 l_InitialTries = 8;
    const UINT32 l_Sleep0Granularity = 8;
    const UINT32 l_Sleep1Granularity = 16;

    DWORD l_SleepTimeoutMs;

    if (FALSE == g_MultipleProcessors)
    {
        l_SleepTimeoutMs = (0 == (*a_Count % l_Sleep1Granularity) ? 1 : 0);
    }
    else
    {
        if (l_InitialTries > *a_Count)
        {
            UtilYieldProcessorInternal(4 << *a_Count);

            goto EXIT;
        }

        if (0 == (*a_Count % l_Sleep1Granularity))
        {
            l_SleepTimeoutMs = 1;
        }
        else if (0 == (*a_Count % l_Sleep0Granularity))
        {
            l_SleepTimeoutMs = 0;
        }
        else
        {
            YieldProcessor();
            YieldProcessor();
            YieldProcessor();
            YieldProcessor();

            goto EXIT;
        }
    }

    Sleep(l_SleepTimeoutMs);

    EXIT:

    if (MAX_UINT32 == ++(*a_Count))
    {
        *a_Count = (l_InitialTries + ((MAX_UINT32 - l_InitialTries) % l_Sleep1Granularity));
    }
}

BOOLEAN
UtilTryAquireLock(IN OUT UINT32 volatile * const a_Lock)
{
    return (0 == InterlockedCompareExchangeAcquire((LONG volatile *)a_Lock, 1, 0));
}

void
UtilAquireLock(IN OUT UINT32 volatile * const a_Lock)
{
    UINT32 l_Count = 0;

    for (;;)
    {
        if (FALSE != UtilTryAquireLock(a_Lock))
        {
            return;
        }

        UtilSleepBrieflyInternal(&l_Count);
    }
}

void
UtilReleaseLock(IN OUT UINT32 volatile * const a_Lock)
{
    // The caller is supposed to hold the lock on entry.

    ASSERT(1 == *a_Lock);

    MemoryBarrier();

    // We don't need to lock the multiprocessor bus when releasing,
    // since we are the only legal writer. Thus we don't need
    // InterlockedExchange().
    *a_Lock = 0;
}

UINT32
UtilUInt64ToUInt32(IN UINT64 const a_Num)
{
    return (UINT32)(((a_Num >> BITS_PER_UINT32) & MAX_UINT32) ^ (a_Num & MAX_UINT32));
}

UINT32
UtilUInt64ToUInt32AndRotate(IN UINT64 const a_Num,
                            IN UINT32 const a_Rotate)
{
    return ROT(UtilUInt64ToUInt32(a_Num), a_Rotate);
}

UINT32
UtilPtrToUInt32(IN OPTIONAL void const * const a_Ptr,
                IN UINT32 const a_Rotate)
{
    // This routine returns the address referenced by the pointer as
    // a 32-bit integer. a_Ptr may be NULL.

#if _M_X64
    return UtilUInt64ToUInt32AndRotate((UINT64)a_Ptr, a_Rotate);
#else // _M_X64
    return ROT((UINT32)a_Ptr, a_Rotate);
#endif // _M_X64
}

// End of file.
