/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Native\SHA256.c
 * PURPOSE: SHA256 hash implementation
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

// See: FIPS PUB 180-3.

#include "general.h"

#define BLOCK_UINT32_LENGTH             (16)
#define PROCESSING_UINT32_LENGTH        (64)

static const UINT32 g_K[PROCESSING_UINT32_LENGTH] =
{
    0x428A2F98UL, 0x71374491UL, 0xB5C0FBCFUL, 0xE9B5DBA5UL,
    0x3956C25BUL, 0x59F111F1UL, 0x923F82A4UL, 0xAB1C5ED5UL,
    0xD807AA98UL, 0x12835B01UL, 0x243185BEUL, 0x550C7DC3UL,
    0x72BE5D74UL, 0x80DEB1FEUL, 0x9BDC06A7UL, 0xC19BF174UL,
    0xE49B69C1UL, 0xEFBE4786UL, 0x0FC19DC6UL, 0x240CA1CCUL,
    0x2DE92C6FUL, 0x4A7484AAUL, 0x5CB0A9DCUL, 0x76F988DAUL,
    0x983E5152UL, 0xA831C66DUL, 0xB00327C8UL, 0xBF597FC7UL,
    0xC6E00BF3UL, 0xD5A79147UL, 0x06CA6351UL, 0x14292967UL,
    0x27B70A85UL, 0x2E1B2138UL, 0x4D2C6DFCUL, 0x53380D13UL,
    0x650A7354UL, 0x766A0ABBUL, 0x81C2C92EUL, 0x92722C85UL,
    0xA2BFE8A1UL, 0xA81A664BUL, 0xC24B8B70UL, 0xC76C51A3UL,
    0xD192E819UL, 0xD6990624UL, 0xF40E3585UL, 0x106AA070UL,
    0x19A4C116UL, 0x1E376C08UL, 0x2748774CUL, 0x34B0BCB5UL,
    0x391C0CB3UL, 0x4ED8AA4AUL, 0x5B9CCA4FUL, 0x682E6FF3UL,
    0x748F82EEUL, 0x78A5636FUL, 0x84C87814UL, 0x8CC70208UL,
    0x90BEFFFAUL, 0xA4506CEBUL, 0xBEF9A3F7UL, 0xC67178F2UL
};

#define IDX_A           (0)
#define IDX_B           (1)
#define IDX_C           (2)
#define IDX_D           (3)
#define IDX_E           (4)
#define IDX_F           (5)
#define IDX_G           (6)
#define IDX_H           (7)
#define IDX_T1          (8)
#define IDX_T2          (9)
#define TEMP_ELEMENTS   (10)

#define Ch(x, y, z)     (((x) & (y)) ^ ((~(x)) & (z)))
#define Maj(x, y, z)    (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#define Sum0(x)         (ROT_R((x), 2) ^ ROT_R((x), 13) ^ ROT_R((x), 22))
#define Sum1(x)         (ROT_R((x), 6) ^ ROT_R((x), 11) ^ ROT_R((x), 25))
#define Sigma0(x)       (ROT_R((x), 7) ^ ROT_R((x), 18) ^ ((x) >> 3))
#define Sigma1(x)       (ROT_R((x), 17) ^ ROT_R((x), 19) ^ ((x) >> 10))

void
SHA256MultiProcessInternal(IN OUT UINT32 * const a_HashValues /* 8 UINT32s */,
                           IN UINT32 const * const a_InputBlockBuf /* 16 UINT32s per block */,
                           IN UINT32 const a_InputBlockCount,
                           IN BOOLEAN const a_WipeStack)
{
    // a_HashValues and a_InputBlockBuf may reference the same memory
    // location.

    UINT32 const * l_BlockBuf;
    UINT32 l_ToGo;
    UINT32 l_W[PROCESSING_UINT32_LENGTH];
    UINT32 const * const l_wLimit16 = &l_W[BLOCK_UINT32_LENGTH];
    UINT32 const * const l_wLimit64 = &l_W[PROCESSING_UINT32_LENGTH];
    UINT32 l_Temp[TEMP_ELEMENTS];

    for (l_BlockBuf = a_InputBlockBuf, l_ToGo = a_InputBlockCount; 0 != l_ToGo; l_ToGo--)
    {
        UINT32 * w;
        UINT32 const * k;

        // Read the input block and swap the byte order of
        // each 32-bit integer.
        for (w = l_W; l_wLimit16 != w; l_BlockBuf++, w++)
        {
            *w = BYTE_SWAP32(*l_BlockBuf);
        }

        // Initialize the upper portion of the W array.
        for (; l_wLimit64 != w; w++)
        {
            *w = (Sigma1(w[-2]) + w[-7] + Sigma0(w[-15]) + w[-16]);
        }

        l_Temp[IDX_A] = a_HashValues[IDX_A];
        l_Temp[IDX_B] = a_HashValues[IDX_B];
        l_Temp[IDX_C] = a_HashValues[IDX_C];
        l_Temp[IDX_D] = a_HashValues[IDX_D];
        l_Temp[IDX_E] = a_HashValues[IDX_E];
        l_Temp[IDX_F] = a_HashValues[IDX_F];
        l_Temp[IDX_G] = a_HashValues[IDX_G];
        l_Temp[IDX_H] = a_HashValues[IDX_H];

        for (k = g_K, w = l_W; l_wLimit64 != w; k++, w++)
        {
            l_Temp[IDX_T1] = (l_Temp[IDX_H] + Sum1(l_Temp[IDX_E]) + Ch(l_Temp[IDX_E], l_Temp[IDX_F], l_Temp[IDX_G]) + *k + *w);
            l_Temp[IDX_T2] = (Sum0(l_Temp[IDX_A]) + Maj(l_Temp[IDX_A], l_Temp[IDX_B], l_Temp[IDX_C]));
            l_Temp[IDX_H] = l_Temp[IDX_G];
            l_Temp[IDX_G] = l_Temp[IDX_F];
            l_Temp[IDX_F] = l_Temp[IDX_E];
            l_Temp[IDX_E] = (l_Temp[IDX_D] + l_Temp[IDX_T1]);
            l_Temp[IDX_D] = l_Temp[IDX_C];
            l_Temp[IDX_C] = l_Temp[IDX_B];
            l_Temp[IDX_B] = l_Temp[IDX_A];
            l_Temp[IDX_A] = (l_Temp[IDX_T1] + l_Temp[IDX_T2]);
        }

        a_HashValues[IDX_A] += l_Temp[IDX_A];
        a_HashValues[IDX_B] += l_Temp[IDX_B];
        a_HashValues[IDX_C] += l_Temp[IDX_C];
        a_HashValues[IDX_D] += l_Temp[IDX_D];
        a_HashValues[IDX_E] += l_Temp[IDX_E];
        a_HashValues[IDX_F] += l_Temp[IDX_F];
        a_HashValues[IDX_G] += l_Temp[IDX_G];
        a_HashValues[IDX_H] += l_Temp[IDX_H];
    }

    if (FALSE != a_WipeStack)
    {
        UtilSecWipeBuffer(l_W, sizeof(l_W));
        UtilSecWipeBuffer(l_Temp, sizeof(l_Temp));
    }
}

// Exports.

EXPORT void CALLCONV
SHA256MultiProcess_EXPORT(IN OUT UINT32 * const a_HashValues,
                          IN UINT32 const * const a_InputBlockBuf,
                          IN UINT32 const a_InputBlockCount)
{
    SHA256MultiProcessInternal(a_HashValues, a_InputBlockBuf, a_InputBlockCount, FALSE);
}

// End of file.
