/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Support\Collections.cs
 * PURPOSE: Custom collections
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Support
{
#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct ListHead<T>
                     where T : class, INodeable<T>
    {
        private int f_Count;
        private T f_First;
        private T f_Last;

        public int p_Count
        {
            get { return f_Count; }
        }

        public T p_First
        {
            get { return f_First; }
        }

        public T p_Last
        {
            get { return f_Last; }
        }

        public void Clear()
        {
            for (T l_Node = f_First; null != l_Node;)
            {
                T l_NextNode = l_Node.p_Next;

                l_Node.SetPrevious(null);
                l_Node.SetNext(null);

                l_Node = l_NextNode;
            }

            f_Count = 0;
            f_First = null;
            f_Last = null;
        }

        public void Remove(T a_Obj)
        {
            DebugAssertReferenced(a_Obj);

            if (a_Obj == f_First)
            {
                f_First = a_Obj.p_Next;

                if (null == f_First)
                {
                    f_Last = null;
                }
                else
                {
                    f_First.SetPrevious(null);
                }

                a_Obj.SetNext(null);
            }
            else if (a_Obj == f_Last)
            {
                f_Last = a_Obj.p_Previous;

                if (null == f_Last)
                {
                    f_First = null;
                }
                else
                {
                    f_Last.SetNext(null);
                }

                a_Obj.SetPrevious(null);
            }
            else
            {
                T l_Prev = a_Obj.p_Previous;
                T l_Next = a_Obj.p_Next;

                l_Prev.SetNext(l_Next);
                l_Next.SetPrevious(l_Prev);

                a_Obj.SetPrevious(null);
                a_Obj.SetNext(null);
            }

            f_Count--;
        }

        public void AddFirst(T a_Obj)
        {
            DebugAssertUnreferenced(a_Obj);

            if (0 == f_Count)
            {
                f_Last = a_Obj;
            }
            else
            {
                a_Obj.SetNext(f_First);
                f_First.SetPrevious(a_Obj);
            }

            f_Count++;
            f_First = a_Obj;
        }

        public void AddLast(T a_Obj)
        {
            DebugAssertUnreferenced(a_Obj);

            if (0 == f_Count)
            {
                f_First = a_Obj;
            }
            else
            {
                a_Obj.SetPrevious(f_Last);
                f_Last.SetNext(a_Obj);
            }

            f_Count++;
            f_Last = a_Obj;
        }

        public void AddBefore(T a_OldObj,
                              T a_NewObj)
        {
            DebugAssertReferenced(a_OldObj);
            DebugAssertUnreferenced(a_NewObj);

            if (f_First == a_OldObj)
            {
                f_First = a_NewObj;
            }
            else
            {
                T l_Prev = a_OldObj.p_Previous;

                a_NewObj.SetPrevious(l_Prev);
                l_Prev.SetNext(a_NewObj);
            }

            a_NewObj.SetNext(a_OldObj);
            a_OldObj.SetPrevious(a_NewObj);

            f_Count++;
        }

        public void AddAfter(T a_OldObj,
                             T a_NewObj)
        {
            DebugAssertReferenced(a_OldObj);
            DebugAssertUnreferenced(a_NewObj);

            if (f_Last == a_OldObj)
            {
                f_Last = a_NewObj;
            }
            else
            {
                T l_Next = a_OldObj.p_Next;

                a_NewObj.SetNext(l_Next);
                l_Next.SetPrevious(a_NewObj);
            }

            a_NewObj.SetPrevious(a_OldObj);
            a_OldObj.SetNext(a_NewObj);

            f_Count++;
        }

        [Conditional("DEBUG")]
        public void DebugAssertReferenced(T a_Obj)
        {
            // The easiest way to find out is to walk the list.

            if (0 >= f_Count)
            {
                throw (new ApplicationException());
            }

            if (1000 <= f_Count)
            {
                // In this case do a quick check (otherwise things
                // slow down too much).

                if (null == a_Obj.p_Previous)
                {
                    if (a_Obj == f_First)
                    {
                        return;
                    }
                }
                else if (null == a_Obj.p_Next)
                {
                    if (a_Obj == f_Last)
                    {
                        return;
                    }
                }
                else
                {
                    return;
                }
            }

            for (T l_Ptr = f_First; null != l_Ptr; l_Ptr = l_Ptr.p_Next)
            {
                if (l_Ptr == a_Obj)
                {
                    return;
                }
            }

            throw (new ApplicationException());
        }

        [Conditional("DEBUG")]
        private void DebugAssertUnreferenced(T a_Obj)
        {
            if (null != a_Obj.p_Previous ||
                null != a_Obj.p_Next ||
                f_First == a_Obj)
            {
                throw (new ApplicationException());
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct ListEntry<T> where T : class, INodeable<T>
    {
        // Disable useless warning about unused fields.
#pragma warning disable 649

        // Both can be null, even if an item is in the list.
        public T f_Previous;
        public T f_Next;

#pragma warning restore 649
    }

    public interface INodeable<T> where T : class
    {
        T p_Previous
        {
            get;
        }

        T p_Next
        {
            get;
        }

        // Only to be called by the ListHead<T> class.
        void SetPrevious(T a_Obj);

        // Only to be called by the ListHead<T> class.
        void SetNext(T a_Obj);
    }

    public class ListBaseClass<T> : INodeable<T>
                         where T : class, INodeable<T>
    {
        // The class can be used for convenience, but it's not necessary
        // (any class can implement INodeable<T> itself).

        private ListEntry<T> f_ListEntry;

        public T p_Previous
        {
            get { return f_ListEntry.f_Previous; }
        }

        public T p_Next
        {
            get { return f_ListEntry.f_Next; }
        }

        public void SetPrevious(T a_Obj)
        {
            f_ListEntry.f_Previous = a_Obj;
        }

        public void SetNext(T a_Obj)
        {
            f_ListEntry.f_Next = a_Obj;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct ByteArray
    {
        public readonly byte[] f_Arr;
        public readonly int f_Idx;
        public readonly int f_Len;

        public ByteArray(byte[] a_Arr,
                         int a_Idx,
                         int a_Len)
        {
            f_Arr = a_Arr;
            f_Idx = a_Idx;
            f_Len = a_Len;
        }

        public ByteArray(int a_Len) :
                    this(new byte[a_Len], 0, a_Len)
        {
        }

        public ByteArray(byte[] a_Arr) :
                    this(a_Arr, 0, a_Arr.Length)
        {
        }

        public ByteArray(byte[] a_Arr,
                         int a_Idx) :
                    this(a_Arr, a_Idx, a_Arr.Length - a_Idx)
        {
        }

        public static bool IsContentEqual(ByteArray a_Data1,
                                          ByteArray a_Data2)
        {
            if (a_Data1.f_Len != a_Data2.f_Len)
            {
                return false;
            }

            return Buffer2.ValueEquals(a_Data1.f_Arr, a_Data1.f_Idx, a_Data2.f_Arr, a_Data2.f_Idx, a_Data1.f_Len);
        }

        public byte[] ToArray(bool a_AlwaysCreateNew)
        {
            if (false == a_AlwaysCreateNew &&
                0 == f_Idx &&
                f_Len == f_Arr.Length)
            {
                return f_Arr;
            }

            return Util.CopyBytes(f_Arr, f_Idx, f_Len);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct FlexDictionary<TKey, TVal>
    {
        // Don't use the struct default constructor.
        // This struct is a dictionary that is optimized for value
        // type values. TKey should be a reference type.

        private const int c_NoIndex = (-1);

        private static readonly IEqualityComparer<TKey> s_Comparer;

        private int f_UsedCount;
        // f_Items can be null until Grow() was called.
        private FlexDicItem<TKey, TVal>[] f_Items;

        private int f_EmptyCount;
        private int f_FirstEmptyIndex;

        public int p_Count
        {
            get { return f_UsedCount; }
        }

        public int p_Capacity
        {
            get { return (null == f_Items ? 0 : f_Items.Length); }
        }

        static FlexDictionary()
        {
            s_Comparer = EqualityComparer<TKey>.Default;
        }

        private FlexDictionary(int a_MinCapacity)
        {
            // Same that Clear() does.
            f_UsedCount = 0;
            f_Items = null;
            f_EmptyCount = 0;
            f_FirstEmptyIndex = c_NoIndex;

            SetCapacity(a_MinCapacity);
        }

        public static void Create(out FlexDictionary<TKey, TVal> a_Dest)
        {
            a_Dest = new FlexDictionary<TKey, TVal>(0);
        }

        public static void Create(int a_InitialCapacity,
                                  out FlexDictionary<TKey, TVal> a_Dest)
        {
            a_Dest = new FlexDictionary<TKey, TVal>(a_InitialCapacity);
        }

        public bool IsEmpty()
        {
            return (0 >= f_UsedCount);
        }

        public void Clone(out FlexDictionary<TKey, TVal> a_Dest)
        {
            a_Dest = new FlexDictionary<TKey,TVal>();

            a_Dest.f_UsedCount = this.f_UsedCount;
            a_Dest.f_Items = (null == this.f_Items ? null : Util.CloneArray(this.f_Items));
            a_Dest.f_EmptyCount = this.f_EmptyCount;
            a_Dest.f_FirstEmptyIndex = this.f_FirstEmptyIndex;
        }

        public void Clear()
        {
            f_UsedCount = 0;
            f_Items = null;
            f_EmptyCount = 0;
            f_FirstEmptyIndex = c_NoIndex;

            DebugNotifyChange();
        }

        private void Grow(int a_NewCapacity)
        {
            // f_Items may be null. a_NewCapacity should be a prime
            // number that is equal to or greater than the current
            // capacity.

            int l_OldCapacity = p_Capacity;

            if (l_OldCapacity >= a_NewCapacity)
            {
                return;
            }

            Array.Resize(ref f_Items, a_NewCapacity);

            // First reset the bucket item index field of each
            // element.

            for (int i = 0; i < a_NewCapacity; i++)
            {
                f_Items[i].SetBucketItemIndex(c_NoIndex);

                if (l_OldCapacity <= i)
                {
                    // Update new items. Make each point to the
                    // next one. The last assignment must be undone
                    // after the loop.

                    f_Items[i].SetNextItemIndex(i + 1);
                }
            }

            f_Items[a_NewCapacity - 1].SetNextItemIndex(f_FirstEmptyIndex);
            f_EmptyCount += (a_NewCapacity - l_OldCapacity);
            f_FirstEmptyIndex = l_OldCapacity;

            // Fill buckets with existing items.

            for (int i = 0; i < l_OldCapacity; i++)
            {
                if (false == f_Items[i].IsValid())
                {
                    continue;
                }

                int l_BucketItemIdx = HashCodeToBucketIndex(f_Items[i].p_HashCode, a_NewCapacity);
                f_Items[i].SetNextItemIndex(f_Items[l_BucketItemIdx].p_BucketItemIndex);
                f_Items[l_BucketItemIdx].SetBucketItemIndex(i);
            }

            DebugNotifyChange();
        }

        public void SetCapacity(int a_Capacity)
        {
            if (f_UsedCount > a_Capacity)
            {
                throw (new ArgumentOutOfRangeException());
            }

            Grow(PrimeNumberHelper.FindNextPrime(a_Capacity));
        }

        public FlexDicItem<TKey, TVal>[] ExposeItemArray(out FlexDicEnum<TKey, TVal> a_Enum)
        {
            FlexDicItem<TKey, TVal>[] l_Arr = ExposeItemArray();

            // Do this after the call to ExposeItemArray().
            CreateEnumeration(out a_Enum);

            return l_Arr;
        }

        public FlexDicItem<TKey, TVal>[] ExposeItemArray()
        {
            // NOTE: The caller must check the validity of an item
            // using FlexDicItem.IsValid() before using it.
            // The return value becomes invalid by any subsequent call
            // to Clear(), SetCapacity(), PrepareAdd(), PrepareSet() or
            // Remove().

            if (null == f_Items)
            {
                Grow(PrimeNumberHelper.c_LowestOddPrime);
            }

            return f_Items;
        }

        private void CreateEnumeration(out FlexDicEnum<TKey, TVal> a_Dest)
        {
            a_Dest = new FlexDicEnum<TKey, TVal>(f_Items.Length);
        }

        public TKey[] ToKeyArray(bool a_NullIfEmpty)
        {
            if (true == a_NullIfEmpty &&
                0 >= f_UsedCount)
            {
                return null;
            }

            TKey[] l_Dest = new TKey[f_UsedCount];

            if (0 < f_UsedCount)
            {
                int l_DestIdx = 0;

                FlexDicEnum<TKey, TVal> l_Enum;
                CreateEnumeration(out l_Enum);
                for (int l_SrcIdx; true == l_Enum.GetNextIndex(f_Items, out l_SrcIdx);)
                {
                    l_Dest[l_DestIdx] = f_Items[l_SrcIdx].p_Key;

                    if (++l_DestIdx == f_UsedCount)
                    {
                        break;
                    }
                }
            }

            return l_Dest;
        }

        public bool ContainsKey(TKey a_Key)
        {
            return (0 <= FindItemIndex(a_Key));
        }

        public bool TryGet(TKey a_Key,
                           out int a_Idx)
        {
            a_Idx = FindItemIndex(a_Key);

            return (0 <= a_Idx);
        }

        public bool PrepareAdd(TKey a_Key,
                               out int a_Idx,
                               out FlexDicItem<TKey, TVal>[] a_Arr)
        {
            // Returns false if the key already exists.

            bool l_Junk;
            a_Idx = PrepareInsert(a_Key, false, out l_Junk);

            if (0 > a_Idx)
            {
                a_Arr = null;

                return false;
            }

            a_Arr = f_Items;

            return true;
        }

        public int PrepareSet(TKey a_Key,
                              out FlexDicItem<TKey, TVal>[] a_Arr,
                              out bool a_Created)
        {
            int l_Idx = PrepareInsert(a_Key, true, out a_Created);

            // Query f_Items after the call to PrepareInsert().
            a_Arr = f_Items;

            return l_Idx;
        }

        private int PrepareInsert(TKey a_Key,
                                  bool a_MayOverwrite,
                                  out bool a_Created)
        {
            // f_Items may be null.

            uint l_HashCode;

            if (0 >= f_UsedCount)
            {
                if (null == f_Items)
                {
                    Grow(PrimeNumberHelper.c_LowestOddPrime);
                }

                l_HashCode = GetKeyHashCode(a_Key);
            }
            else
            {
                int l_ExistingIdx = FindItemIndexCore(a_Key, out l_HashCode);

                if (0 <= l_ExistingIdx)
                {
                    a_Created = false;

                    if (false == a_MayOverwrite)
                    {
                        // Key exists already.

                        return c_NoIndex;
                    }

                    return l_ExistingIdx;
                }
            }

            int l_Idx = ReserveEmptyItem();

            // Since HashCodeToBucketIndex() needs to know the new
            // capacity, call it after the call to ReserveEmptyItem().
            int l_BucketIdx = HashCodeToBucketIndex(l_HashCode);
            f_Items[l_Idx].InitializeItemFields(l_HashCode, f_Items[l_BucketIdx].p_BucketItemIndex, a_Key);
            f_Items[l_BucketIdx].SetBucketItemIndex(l_Idx);

            DebugNotifyChange();

            a_Created = true;

            return l_Idx;
        }

        public bool Remove(TKey a_Key)
        {
            if (0 >= f_UsedCount)
            {
                return false;
            }

            uint l_HashCode = GetKeyHashCode(a_Key);
            int l_BucketIdx = HashCodeToBucketIndex(l_HashCode);
            for (int i = f_Items[l_BucketIdx].p_BucketItemIndex, l_PrevIdx = c_NoIndex; 0 <= i; l_PrevIdx = i, i = f_Items[i].p_NextIndex)
            {
                if (false == HasItemThisKey(ref f_Items[i], a_Key, l_HashCode))
                {
                    continue;
                }

                int l_NextIdx = f_Items[i].p_NextIndex;

                if (0 > l_PrevIdx)
                {
                    f_Items[l_BucketIdx].SetBucketItemIndex(l_NextIdx);
                }
                else
                {
                    f_Items[l_PrevIdx].SetNextItemIndex(l_NextIdx);
                }

                FreeItem(i);

                DebugNotifyChange();

                return true;
            }

            return false;
        }

        private int ReserveEmptyItem()
        {
            int l_Idx;

            if (0 >= f_EmptyCount)
            {
                int l_NewCap;

                if (0 >= f_UsedCount)
                {
                    l_NewCap = PrimeNumberHelper.c_LowestOddPrime;
                }
                else
                {
                    l_NewCap = PrimeNumberHelper.FindNextPrime(checked(2 * f_UsedCount));
                }

                Grow(l_NewCap);
            }

            f_UsedCount++;
            f_EmptyCount--;
            l_Idx = f_FirstEmptyIndex;
            f_FirstEmptyIndex = f_Items[f_FirstEmptyIndex].p_NextIndex;

            return l_Idx;
        }

        private void FreeItem(int a_Index)
        {
            f_UsedCount--;
            f_Items[a_Index].ResetItemFields(f_FirstEmptyIndex);
            f_EmptyCount++;
            f_FirstEmptyIndex = a_Index;
        }

        private int FindItemIndex(TKey a_Key)
        {
            if (0 >= f_UsedCount)
            {
                return (-1);
            }

            uint l_Junk;
            return FindItemIndexCore(a_Key, out l_Junk);
        }

        private int FindItemIndexCore(TKey a_Key,
                                      out uint a_HashCode)
        {
            // a_HashCode will be set in any case.

            a_HashCode = GetKeyHashCode(a_Key);

            int l_BucketIdx = HashCodeToBucketIndex(a_HashCode);
            for (int i = f_Items[l_BucketIdx].p_BucketItemIndex; 0 <= i; i = f_Items[i].p_NextIndex)
            {
                if (true == HasItemThisKey(ref f_Items[i], a_Key, a_HashCode))
                {
                    return i;
                }
            }

            return c_NoIndex;
        }

        private int HashCodeToBucketIndex(uint a_HashCode)
        {
            // f_Items must be valid.

            return HashCodeToBucketIndex(a_HashCode, f_Items.Length);
        }

        private static int HashCodeToBucketIndex(uint a_HashCode,
                                                 int a_Capacity)
        {
            if (int.MinValue == (int)a_HashCode)
            {
                a_HashCode = (uint)int.MaxValue;
            }

            return (Math.Abs((int)a_HashCode) % a_Capacity);
        }

        private static uint GetKeyHashCode(TKey a_Key)
        {
            uint l_Digest = (uint)s_Comparer.GetHashCode(a_Key);

            if (FlexDicItem<TKey, TVal>.c_NoHashCode == l_Digest)
            {
                // This routine may not return FlexDicItem.c_NoHashCode
                // (that is zero).

                return 1;
            }

            return l_Digest;
        }

        private static bool HasItemThisKey(ref FlexDicItem<TKey, TVal> a_Item,
                                           TKey a_Key,
                                           uint a_KeyHashCode)
        {
            // a_KeyHashCode must stem from a call to
            // GetKeyHashCode() passing a_Key.

            return (a_Item.p_HashCode == a_KeyHashCode &&
                    true == s_Comparer.Equals(a_Item.p_Key, a_Key));
        }

        [Conditional("DEBUG")]
        private void DebugNotifyChange()
        {
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct FlexDicEnum<TKey, TVal>
    {
        // Don't use the struct default constructor.
        // The dictionary should not be changed while enumerating.

        private readonly int f_Count;
        private int f_Index;

        public FlexDicEnum(int a_Count)
        {
            f_Count = a_Count;
            f_Index = 0;
        }

        public void Reset()
        {
            f_Index = 0;
        }

        public bool GetNextIndex(FlexDicItem<TKey, TVal>[] a_Arr,
                                 out int a_Index)
        {
            for (; f_Index < f_Count; f_Index++)
            {
                if (true == a_Arr[f_Index].IsValid())
                {
                    a_Index = f_Index++;

                    return true;
                }
            }

            a_Index = (-1);

            return false;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct FlexDicItem<TKey, TVal>
    {
        public const uint c_NoHashCode = 0;

        // Table field.
        private int f_BucketItemIndex;

        // Item fields.
        private uint f_HashCode;
        private int f_NextIndex;
        private TKey f_Key;

        public TVal f_Val;

        public int p_BucketItemIndex
        {
            get { return f_BucketItemIndex; }
        }

        public uint p_HashCode
        {
            get { return f_HashCode; }
        }

        public int p_NextIndex
        {
            get { return f_NextIndex; }
        }

        public TKey p_Key
        {
            get { return f_Key; }
        }

        public void ResetItemFields(int a_NextEmptyIndex)
        {
            f_HashCode = c_NoHashCode;
            f_NextIndex = a_NextEmptyIndex;
            f_Key = default(TKey);
            f_Val = default(TVal);
        }

        public void InitializeItemFields(uint a_HashCode,
                                         int a_NextIndex,
                                         TKey a_Key)
        {
            f_HashCode = a_HashCode;
            f_NextIndex = a_NextIndex;
            f_Key = a_Key;
        }

        public void SetBucketItemIndex(int a_ItemIndex)
        {
            f_BucketItemIndex = a_ItemIndex;
        }

        public bool IsValid()
        {
            return (c_NoHashCode != f_HashCode);
        }

        public void SetNextItemIndex(int a_ItemIndex)
        {
            f_NextIndex = a_ItemIndex;
        }
    }

    public struct MultiMap<TKey, TVal>
                           where TVal : class
    {
        // Don't use the struct default constructor.

        private Dictionary<TKey, object> f_Dic;

        private MultiMap(Dictionary<TKey, object> a_Dic)
        {
            f_Dic = a_Dic;
        }

        public static void Create(out MultiMap<TKey, TVal> a_Dest)
        {
            a_Dest = new MultiMap<TKey, TVal>(new Dictionary<TKey, object>(0));
        }

        public static void Create(int a_InitialCapacity,
                                  out MultiMap<TKey, TVal> a_Dest)
        {
            a_Dest = new MultiMap<TKey, TVal>(new Dictionary<TKey, object>(a_InitialCapacity));
        }

        public void Clear()
        {
            f_Dic.Clear();
        }

        public bool ContainsKey(TKey a_Key)
        {
            return f_Dic.ContainsKey(a_Key);
        }

        public bool Add(TKey a_Key,
                        TVal a_Val)
        {
            // Returns true if the value was not in the dictionary yet.

            object l_Obj;

            if (false == f_Dic.TryGetValue(a_Key, out l_Obj))
            {
                f_Dic.Add(a_Key, a_Val);

                return true;
            }

            TVal[] l_Arr;

            if (false == IsTValArr(l_Obj))
            {
                // Single into multi.

                TVal l_OldVal = (TVal)l_Obj;

                if (l_OldVal == a_Val)
                {
                    return false;
                }

                l_Arr = new TVal[2];
                l_Arr[0] = l_OldVal;
                l_Arr[1] = a_Val;
            }
            else
            {
                // We've already two values.

                l_Arr = (TVal[])l_Obj;

                foreach (TVal l_Val in l_Arr)
                {
                    if (l_Val == a_Val)
                    {
                        return false;
                    }
                }

                int l_Count = l_Arr.Length;
                Array.Resize(ref l_Arr, l_Count + 1);
                l_Arr[l_Count] = a_Val;
            }

            f_Dic[a_Key] = l_Arr;

            return true;
        }

        public bool Remove(TKey a_Key,
                           TVal a_Val)
        {
            // Returns true if the value existed.

            object l_Obj;

            if (false == f_Dic.TryGetValue(a_Key, out l_Obj))
            {
                return false;
            }

            TVal[] l_Arr;

            if (false == IsTValArr(l_Obj))
            {
                // Kill single.

                bool l_OK = f_Dic.Remove(a_Key);

                return true;
            }

            // We've at least two values.

            l_Arr = (TVal[])l_Obj;
            int l_Count = l_Arr.Length;

            for (int i = (l_Count - 1); 0 <= i; i--)
            {
                if (l_Arr[i] == a_Val)
                {
                    // The value really exists.

                    if (2 == l_Count)
                    {
                        // Turn into single.

                        f_Dic[a_Key] = l_Arr[0 == i ? 1 : 0];
                    }
                    else
                    {
                        // Shorten array.

                        Array.Copy(l_Arr, i + 1, l_Arr, i, l_Count - (i + 1));
                        Array.Resize(ref l_Arr, l_Count - 1);

                        f_Dic[a_Key] = l_Arr;
                    }

                    return true;
                }
            }

            return false;
        }

        public bool TryGetFirst(TKey a_Key,
                                out TVal a_First)
        {
            object l_Obj;

            if (false == f_Dic.TryGetValue(a_Key, out l_Obj))
            {
                a_First = null;

                return false;
            }

            if (false == IsTValArr(l_Obj))
            {
                a_First = (TVal)l_Obj;
            }
            else
            {
                a_First = ((TVal[])l_Obj)[0];
            }

            return true;
        }

        public bool CreateEnumerator(TKey a_Key,
                                     out Enumerator a_Dest)
        {
            // Returns false if there is nothing to enumerate.

            object l_Obj;

            if (false == f_Dic.TryGetValue(a_Key, out l_Obj))
            {
                a_Dest = default(Enumerator);

                return false;
            }

            if (false == IsTValArr(l_Obj))
            {
                a_Dest = new MultiMap<TKey, TVal>.Enumerator((TVal)l_Obj);
            }
            else
            {
                a_Dest = new MultiMap<TKey, TVal>.Enumerator((TVal[])l_Obj);
            }

            return true;
        }

        private static bool IsTValArr(object a_Obj)
        {
            return (typeof(TVal[]) == a_Obj.GetType());
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Enumerator
        {
            // Don't use the struct default constructor.

            private readonly TVal f_Single;
            private readonly TVal[] f_Arr;
            private int f_Index;

            public Enumerator(TVal a_Single)
            {
                f_Single = a_Single;
                f_Arr = null;
                f_Index = 0;
            }

            public Enumerator(TVal[] a_Arr)
            {
                f_Single = null;
                f_Arr = a_Arr;
                f_Index = 0;
            }

            public void Reset()
            {
                f_Index = 0;
            }

            public bool GetNext(out TVal a_Next)
            {
                if (null != f_Single)
                {
                    if (0 < f_Index)
                    {
                        a_Next = null;

                        return false;
                    }

                    a_Next = f_Single;
                }
                else
                {
                    if (f_Arr.Length == f_Index)
                    {
                        a_Next = null;

                        return false;
                    }

                    a_Next = f_Arr[f_Index];
                }

                f_Index++;

                return true;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct FlexArray<T>
    {
        // Don't use the struct default constructor.

        // IMPORTANT: The last return value of p_Arr becomes completely
        // invalid whenever routines of this object get called.
        // Exceptions are Clear(), DeleteRange(), DeleteAt().

        private const byte c_DefaultAllocStep = 8;

        private static readonly T[] s_Dummy;

        private int f_Count;
        // f_Arr will never be null.
        private T[] f_Arr;

        public int p_Count
        {
            get { return f_Count; }
        }

        public T[] p_Arr
        {
            get { return f_Arr; }
        }

        static FlexArray()
        {
            s_Dummy = new T[0];
        }

        private FlexArray(int a_InitialCapacity)
        {
            f_Count = 0;
            f_Arr = (0 == a_InitialCapacity ? s_Dummy : new T[a_InitialCapacity]);
        }

        public static void Create(out FlexArray<T> a_Dest)
        {
            a_Dest = new FlexArray<T>(0);
        }

        public static void Create(int a_InitialCapacity,
                                  out FlexArray<T> a_Dest)
        {
            a_Dest = new FlexArray<T>(a_InitialCapacity);
        }

        public static void Create(T[] a_Arr,
                                  out FlexArray<T> a_Dest)
        {
            a_Dest = new FlexArray<T>();
            a_Dest.f_Count = a_Arr.Length;
            a_Dest.f_Arr = a_Arr;
        }

        public void Clear(bool a_KeepCapacity)
        {
            if (false == a_KeepCapacity)
            {
                if (0 < f_Count)
                {
                    f_Count = 0;
                    f_Arr = s_Dummy;
                }
            }
            else
            {
                Clear();
            }
        }

        public void Clear()
        {
            if (0 < f_Count)
            {
                Array.Clear(f_Arr, 0, f_Count);
                f_Count = 0;
            }
        }

        public void SetCapacity(int a_NewCapacity)
        {
            if (a_NewCapacity < f_Count)
            {
                throw(new ArgumentOutOfRangeException());
            }

            Array.Resize<T>(ref f_Arr, a_NewCapacity);
        }

        public void TrimExcess()
        {
            Array.Resize<T>(ref f_Arr, f_Count);
        }

        public void AddArray(T[] a_Src)
        {
            AddArray(a_Src, 0, a_Src.Length);
        }

        public void AddArray(T[] a_Src,
                             int a_SrcIdx,
                             int a_SrcLen)
        {
            if (0 == a_SrcLen)
            {
                return;
            }

            int l_DestIdx = f_Count;
            ReserveRange(l_DestIdx, a_SrcLen);

            Array.Copy(a_Src, a_SrcIdx, f_Arr, l_DestIdx, a_SrcLen);
        }

        public void ReserveRange(int a_StartingIndex,
                                 int a_Length)
        {
            // Think of this routine as of InsertRange(), but without
            // the store operations.

            // This check also works if a_StartingIndex is negative.
            if ((uint)a_StartingIndex > (uint)f_Count)
            {
                throw (new ArgumentOutOfRangeException());
            }

            int l_NewCount = checked(f_Count + a_Length);

            if (f_Arr.Length < l_NewCount)
            {
                Array.Resize<T>(ref f_Arr, l_NewCount + c_DefaultAllocStep);
            }

            int l_PreserveCount = (f_Count - a_StartingIndex);

            if (0 < l_PreserveCount)
            {
                Array.Copy(f_Arr, a_StartingIndex, f_Arr, a_StartingIndex + a_Length, l_PreserveCount);
            }

            f_Count = l_NewCount;
        }

        public void ReserveAt(int a_Index)
        {
            ReserveRange(a_Index, 1);
        }

        public int ReserveAtEnd()
        {
            int l_NewIdx = f_Count;

            if (f_Arr.Length > f_Count)
            {
                f_Count++;
            }
            else
            {
                ReserveRange(l_NewIdx, 1);
            }

            return l_NewIdx;
        }

        public void DeleteRange(int a_StartingIndex,
                                int a_Length)
        {
            if (false == Util.CheckRange(a_StartingIndex, a_Length, f_Count))
            {
                throw (new ArgumentOutOfRangeException());
            }

            if (0 == a_Length)
            {
                return;
            }

            int l_NewCount = (f_Count - a_Length);

            if (0 < l_NewCount)
            {
                int l_PreserveCount = (l_NewCount - a_StartingIndex);

                if (0 < l_PreserveCount)
                {
                    Array.Copy(f_Arr, a_StartingIndex + a_Length, f_Arr, a_StartingIndex, l_PreserveCount);
                }

                // Do not resize, since the caller relies on being able
                // to use the return value of a previous query of p_Arr
                // again.
            }

            Array.Clear(f_Arr, l_NewCount, a_Length);
            f_Count = l_NewCount;
        }

        public void DeleteAt(int a_Index)
        {
            DeleteRange(a_Index, 1);
        }

        public T[] ToArray(bool a_Cannibalize)
        {
            T[] l_Ret;

            if (false == a_Cannibalize)
            {
                if (0 == f_Count)
                {
                    l_Ret = s_Dummy;
                }
                else if (f_Arr.Length == f_Count)
                {
                    l_Ret = (T[])f_Arr.Clone();
                }
                else
                {
                    l_Ret = new T[f_Count];

                    Array.Copy(f_Arr, l_Ret, f_Count);
                }
            }
            else
            {
                if (0 == f_Count)
                {
                    l_Ret = s_Dummy;
                }
                else
                {
                    // We need to prepare f_Arr first.
                    TrimExcess();
                    f_Count = 0;

                    l_Ret = f_Arr;
                }

                f_Arr = s_Dummy;
            }

            return l_Ret;
        }

        public T[] ToArrayWithReset(int a_Count)
        {
            if (a_Count > f_Count)
            {
                throw(new ArgumentException());
            }

            if (0 == f_Count)
            {
                return new T[0];
            }

            T[] l_ResArr = f_Arr;

            f_Count = 0;
            f_Arr = s_Dummy;

            Array.Resize(ref l_ResArr, a_Count);

            return l_ResArr;
        }

        public void Sort(CompareByRef<T> a_Comp)
        {
            // This routine is optimized for handling structs, thus
            // it avoids unnecessary copying of T.

            if (1 >= f_Count)
            {
                return;
            }

            int[] l_Indices = new int[f_Count];

            for (int i = 0; i < f_Count; i++)
            {
                l_Indices[i] = i;
            }

            HeapSortIndices(a_Comp, l_Indices);

            // Create a new array while preserving capacity.
            T[] l_NewArr = new T[f_Arr.Length];

            for (int i = 0; i < f_Count; i++)
            {
                l_NewArr[i] = f_Arr[l_Indices[i]];
            }

            f_Arr = l_NewArr;
        }

        private void HeapSortIndices(CompareByRef<T> a_Comp,
                                     int[] a_Indices)
        {
            int l_Count = a_Indices.Length;
            bool l_PhaseTwo = false;
            for (int n = (l_Count - 1), i = ((l_Count / 2) - 1);;)
            {
                int l_Parent;
                int l_Temp;

                if (false == l_PhaseTwo)
                {
                    // n remains (count - 1)
                    // i goes [count / 2 - 1;0]

                    l_Parent = i;
                    l_Temp = a_Indices[i];

                    if (0 == i)
                    {
                        l_PhaseTwo = true;
                    }
                    else
                    {
                        i--;
                    }
                }
                else
                {
                    if (0 == n)
                    {
                        break;
                    }

                    // n goes [count - 1;0[
                    // i doesn't matter anymore

                    l_Parent = 0;
                    // Swap items at index zero and index n.
                    l_Temp = a_Indices[n];
                    a_Indices[n] = a_Indices[0];
                    a_Indices[0] = l_Temp;
                    n--;
                }

                // l_Temp is a_Indices[l_Parent].

                for (int l_Child = ((2 * l_Parent) + 1); l_Child <= n;)
                {
                    if (l_Child < n &&
                        0 > a_Comp(ref f_Arr[a_Indices[l_Child]],
                                   ref f_Arr[a_Indices[l_Child + 1]]))
                    {
                        l_Child++;
                    }

                    // l_Child is now the index of the larger of the
                    // two children.

                    if (0 <= a_Comp(ref f_Arr[l_Temp],
                                    ref f_Arr[a_Indices[l_Child]]))
                    {
                        // The largest child is not larger than the
                        // parent.

                        break;
                    }

                    a_Indices[l_Parent] = a_Indices[l_Child];
                    l_Parent = l_Child;
                    l_Child = ((2 * l_Child) + 1);
                }

                a_Indices[l_Parent] = l_Temp;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct Timeline<T>
    {
        // Don't use the struct default constructor.

        private const ushort c_MaxUnitsPerSlot = ushort.MaxValue;
        private const ushort c_MaxCount = ushort.MaxValue;

        private const uint c_TicksPerUnit = (Time.c_TicksPerSecond / 10);

        public readonly ulong f_TicksPerSlot;
        public readonly ushort f_Capacity;
        // f_Slots can be null.
        private T[] f_Slots;

        private ushort f_UsedCount;
        // While f_UsedCount is zero, f_StartingIndex and
        // f_HeadTime are undefined.
        private ushort f_StartingIndex;
        private Time f_HeadTime;

        public uint p_Count
        {
            get { return f_UsedCount; }
        }

        private Timeline(uint a_UnitsPerSlot,
                         uint a_Capacity)
        {
            // a_UnitsPerSlot and a_Capacity must not be zero.
            // a_UnitsPerSlot must not exceed c_MaxUnitsPerSlot and
            // a_Capacity must not exceed c_MaxCount.

            f_TicksPerSlot = ((ulong)a_UnitsPerSlot * c_TicksPerUnit);
            f_Capacity = (ushort)a_Capacity;
            f_Slots = null;
            f_StartingIndex = f_UsedCount = 0;
            f_HeadTime = new Time();
        }

        public static void Create(uint a_UnitsPerSlot,
                                  uint a_Capacity,
                                  out Timeline<T> a_Dest)
        {
            a_Dest = new Timeline<T>(a_UnitsPerSlot, a_Capacity);
        }

        public bool IsEmpty()
        {
            return (0 == f_UsedCount);
        }

        public bool HasEntries()
        {
            return !IsEmpty();
        }

        public void Clear()
        {
            // This routine preserves the capacity.

            Clear(true);
        }

        public void Clear(bool a_KeepCapacity)
        {
            if (false == a_KeepCapacity)
            {
                if (null == f_Slots)
                {
                    return;
                }

                f_Slots = null;
                f_UsedCount = 0;
            }
            else
            {
                if (0 == f_UsedCount)
                {
                    return;
                }

                ClearArrayWrapable(f_StartingIndex, f_UsedCount);

                f_UsedCount = 0;
            }
        }

        public T[] ExposeArray()
        {
            // Return can be null.

            return f_Slots;
        }

        public T[] ExposeArray(out uint a_StartingIndex)
        {
            // Return can be null.

            a_StartingIndex = (0 == f_UsedCount ? Util.c_NoIndex : f_StartingIndex);

            return f_Slots;
        }

        public Time GetNextTime()
        {
            return f_HeadTime;
        }

        public bool TryGetMostRecent(out T a_Value)
        {
            if (0 != f_UsedCount)
            {
                a_Value = f_Slots[f_StartingIndex];

                return true;
            }

            a_Value = default(T);

            return false;
        }

        private void CreateArray()
        {
            f_Slots = new T[f_Capacity];
        }

        public SlotResult TryCreateSlot(Time a_Time,
                                        out uint a_SlotIndex)
        {
            if (0 == f_UsedCount)
            {
                // First entry.

                if (null == f_Slots)
                {
                    CreateArray();
                }

                StartFresh(a_Time, out a_SlotIndex);

                return SlotResult.CreatedFuture;
            }

            uint l_Capacity = f_Capacity;
            SlotResult l_Res;

            if (a_Time < f_HeadTime)
            {
                // Modify historic values.

                uint l_Units = CalcUnits(a_Time, f_HeadTime, true);

                if (0 == l_Units)
                {
                    a_SlotIndex = f_StartingIndex;

                    return SlotResult.ExistsAlready;
                }

                if (l_Units >= l_Capacity)
                {
                    // We don't keep track of values this old.

                    a_SlotIndex = Util.c_NoIndex;

                    return SlotResult.Outdated;
                }

                if (l_Units >= f_UsedCount)
                {
                    f_UsedCount = (ushort)(l_Units + 1);

                    l_Res = SlotResult.CreatedPast;
                }
                else
                {
                    l_Res = SlotResult.ExistsAlready;
                }

                a_SlotIndex = IncreaseIndex(f_StartingIndex, l_Units);
            }
            else
            {
                uint l_Units = CalcUnits(f_HeadTime, a_Time, false);

                if ((l_Capacity - 1) > l_Units)
                {
                    // Move the starting index.

                    l_Units++;

                    uint l_SlotIdx = IncreaseIndex(f_StartingIndex, l_Capacity - l_Units);
                    uint l_EmptySlots = (l_Capacity - f_UsedCount);

                    if (l_EmptySlots < l_Units)
                    {
                        ClearArrayWrapable(l_SlotIdx, l_Units - l_EmptySlots);
                    }

                    f_UsedCount = (ushort)Math.Min(f_UsedCount + l_Units, l_Capacity);
                    f_StartingIndex = (ushort)l_SlotIdx;
                    f_HeadTime = RoundTimeUp(a_Time);

                    a_SlotIndex = l_SlotIdx;
                }
                else
                {
                    // Remove all previous entries.

                    ClearArrayWrapable(f_StartingIndex, f_UsedCount);

                    StartFresh(a_Time, out a_SlotIndex);
                }

                l_Res = SlotResult.CreatedFuture;
            }

            return l_Res;
        }

        private void StartFresh(Time a_Time,
                                out uint a_SlotIndex)
        {
            // f_Slots must not be null and must be clear. Other
            // variable fields are ignored and will be overwritten.

            f_UsedCount = 1;
            f_StartingIndex = (ushort)(f_Capacity - 1u);
            f_HeadTime = RoundTimeUp(a_Time);

            a_SlotIndex = f_StartingIndex;
        }

        private void ClearArrayWrapable(uint a_Idx,
                                        uint a_Len)
        {
            uint l_Len0 = Math.Min(a_Len, f_Capacity - a_Idx);
            Array.Clear(f_Slots, (int)a_Idx, (int)l_Len0);
            Array.Clear(f_Slots, 0, (int)(a_Len - l_Len0));
        }

        private uint CalcUnits(Time a_Time0,
                               Time a_Time1,
                               bool a_LessOneUnit)
        {
            ulong l_Diff64 = (a_Time1.ToBinary() - Util.BoolToByte(a_LessOneUnit) - a_Time0.ToBinary());

            return (uint)(l_Diff64 / f_TicksPerSlot);
        }

        private Time RoundTimeUp(Time a_Time)
        {
            // Rounds up to the next slice in any case.

            ulong l_Ticks = (a_Time.ToBinary() + 1);
            ulong l_Rem = (uint)(l_Ticks % f_TicksPerSlot);

            if (0 != l_Rem)
            {
                l_Ticks += (f_TicksPerSlot - l_Rem);
            }

            return Time.FromBinary(l_Ticks);
        }

        private static uint TimeToUnits(Time a_Time)
        {
            return (uint)(a_Time.ToBinary() / c_TicksPerUnit);
        }

        public uint IncreaseIndex(uint a_StartingIndex,
                                  uint a_Length)
        {
            if (0 == a_Length)
            {
                return a_StartingIndex;
            }

            uint l_Sum = (a_StartingIndex + a_Length);

            if (l_Sum < a_StartingIndex)
            {
                // Overflow.

                l_Sum += ((uint.MaxValue - f_Capacity) + 1);
            }

            if (l_Sum >= f_Capacity)
            {
                l_Sum -= f_Capacity;
            }

            return l_Sum;
        }

        public enum SlotResult : byte
        {
            None = 0,

            Outdated = 1,
            ExistsAlready = 2,
            CreatedPast = 3,
            CreatedFuture = 4,
        }
    }

    public delegate int CompareByRef<T>(ref T a_A,
                                        ref T a_B);

    public struct FlagArray<T>
    {
        private uint f_Flags;

        private void SetFlag(uint a_Flag)
        {
            f_Flags |= a_Flag;
        }

        private void ClearFlag(uint a_Flag)
        {
            f_Flags &= ~a_Flag;
        }

        private bool CheckFlag(uint a_Flag)
        {
            return (a_Flag == (f_Flags & a_Flag));
        }

        public void SetFlag(T a_Flag)
        {
            SetFlag(Convert.ToUInt32(a_Flag));
        }

        public void ClearFlag(T a_Flag)
        {
            ClearFlag(Convert.ToUInt32(a_Flag));
        }

        public bool CheckFlag(T a_Flag)
        {
            return CheckFlag(Convert.ToUInt32(a_Flag));
        }
    }

    public struct BitField
    {
        // The default constructor may be used.

        // This struct used to be generic, but the runtime
        // overhead of converting a generic enum into an integer
        // (per IConvertible or the Convert class) is unbearable.
        // The loss of type security is sad. However, it can be
        // compensated with a type-safe conversion routine on the
        // caller end.

        private uint f_Flags;

        public BitField(uint a_InitialCombination)
        {
            f_Flags = a_InitialCombination;
        }

        public uint AsUInt()
        {
            return f_Flags;
        }

        public bool CheckFlag(uint a_Flag)
        {
            return CheckAnyCombination(a_Flag);
        }

        public bool CheckAnyCombination(uint a_Flags)
        {
            return (0 != (f_Flags & a_Flags));
        }

        public bool CheckEntireCombination(uint a_Flags)
        {
            return (a_Flags == (f_Flags & a_Flags));
        }

        public void ClearFlag(uint a_Flag)
        {
            f_Flags &= ~a_Flag;
        }

        public void SetFlag(uint a_Flag)
        {
            f_Flags |= a_Flag;
        }

        public void InvertFlag(uint a_Flag)
        {
            f_Flags ^= a_Flag;
        }

        public void AdjustFlag(uint a_Flag,
                               bool a_Value)
        {
            AdjustFlagWithReturn(a_Flag, a_Value);
        }

        public bool AdjustFlagWithReturn(uint a_Flag,
                                         bool a_Value)
        {
            // Returns nonzero if the flag was flipped.

            if (CheckFlag(a_Flag) != a_Value)
            {
                InvertFlag(a_Flag);

                return true;
            }

            return false;
        }

        public void ClearCombination(uint a_Combination)
        {
            f_Flags &= ~a_Combination;
        }

        public void SetCombination(uint a_Combination)
        {
            f_Flags |= a_Combination;
        }

        public void AdjustCombination(uint a_Combination,
                                      bool a_Value)
        {
            if (false == a_Value)
            {
                ClearCombination(a_Combination);
            }
            else
            {
                SetCombination(a_Combination);
            }
        }
    }

    static class PrimeNumberHelper
    {
        // NOTE: There are no prime numbers. Only numbers that Bruce
        // Schneier does not want you to factor.

        public const byte c_LowestOddPrime = 3;
        private const byte c_PrimeNumCount = 48;

        private static readonly uint[] s_PrimeNumbers;

        static PrimeNumberHelper()
        {
            s_PrimeNumbers = new uint[c_PrimeNumCount]
            {
                c_LowestOddPrime, 7, 11, 17, 23, 31, 41, 53, 67, 83,
                103, 127, 157, 191, 233, 283, 347, 419, 509, 613,
                739, 907, 1091, 1319, 1597, 1931, 2333, 2803, 3371,
                4049, 4861, 5839, 7013, 8419, 10111, 12143, 14591,
                17519, 21031, 25243, 30307, 36373, 43651, 52387,
                62869, 75479, 90583, 108707
            };
        }

        public static int FindNextPrime(int a_Minimum)
        {
            uint l_Prime = FindNextPrimeUnsigned((uint)a_Minimum);

            if (int.MaxValue <= l_Prime)
            {
                // This is a prime, as Leonhard Euler pointed out.
                return int.MaxValue;
            }

            return (int)l_Prime;
        }

        private static uint FindNextPrimeUnsigned(uint a_Minimum)
        {
            for (int k = 0; c_PrimeNumCount > k; k++)
            {
                uint l_Prime = s_PrimeNumbers[k];

                if (a_Minimum <= l_Prime)
                {
                    return l_Prime;
                }
            }

            // Find a prime using a deterministic test. The largest
            // gap in the UInt32 range is a 336 numbers wide gap
            // after the prime 3842610773.

            uint l_MaxPrime = (uint.MaxValue - 4);

            for (uint i = (a_Minimum | 1); i <= l_MaxPrime; i += 2)
            {
                if (true == IsPrime(i))
                {
                    return i;
                }
            }

            // This is not a prime, but at least it's an odd number.
            return uint.MaxValue;
        }

        private static bool IsPrime(uint a_Number)
        {
            uint l_Sqrt = (uint)Math.Sqrt(a_Number);

            for (uint i = c_LowestOddPrime; i <= l_Sqrt; i += 2)
            {
                if (0 == (a_Number % i))
                {
                    return false;
                }
            }

            return true;
        }
    }

    public struct DummyStruct
    {
    }
}

// End of file.
