/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Support\Encodings.cs
 * PURPOSE: Encodings and hash algorithms
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Web;
using Support.Encodings;

namespace Support.HashFunctions
{
    public interface IHashFunction
    {
        ushort p_HashByteSize
        {
            get;
        }

        ushort p_BlockByteSize
        {
            get;
        }

        void Reset();

        void Update(byte[] a_Data);
        void Update(ref ByteArray a_Data);
        void Update(byte[] a_Data, int a_Idx, int a_Len);

        void Finish();

        void GetDigest(byte[] a_Dest, int a_Idx);
    }

    enum HashFunctionCondition : byte
    {
        Invalid = 0,

        Initialized = 1,
        Dirty = 2,
        Finished = 3,
    }

    public sealed class HMACWrapper : IHashFunction
    {
        private readonly byte[] f_PadAndKey;
        private IHashFunction f_Algo;
        private HashFunctionCondition f_Condition;

        public ushort p_HashByteSize
        {
            get { return f_Algo.p_HashByteSize; }
        }

        public ushort p_BlockByteSize
        {
            get { return f_Algo.p_BlockByteSize; }
        }

        public HMACWrapper(IHashFunction a_Algo,
                           byte[] a_RawKey)
        {
            f_PadAndKey = CreatePadAndKey(a_Algo, a_RawKey, 0, a_RawKey.Length);
            f_Algo = a_Algo;

            ResetCore();
        }

        public void Reset()
        {
            if (HashFunctionCondition.Initialized != f_Condition)
            {
                ResetCore();
            }
        }

        private void ResetCore()
        {
            f_Algo.Reset();
            PreparePad(true);
            f_Algo.Update(f_PadAndKey, 0, p_BlockByteSize);
            f_Condition = HashFunctionCondition.Initialized;
        }

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Src)
        {
            Update(a_Src, 0, a_Src.Length);
        }

        public void Update(byte[] a_Src,
                           int a_Idx,
                           int a_Len)
        {
            f_Condition = HashFunctionCondition.Dirty;
            f_Algo.Update(a_Src, a_Idx, a_Len);
        }

        public void Finish()
        {
            if (HashFunctionCondition.Finished == f_Condition)
            {
                return;
            }

            int l_HashSize = p_HashByteSize;
            int l_BlockSize = p_BlockByteSize;

            // Store inner hash.
            f_Algo.GetDigest(f_PadAndKey, 2 * l_BlockSize);
            f_Algo.Reset();
            PreparePad(false);
            // Hash outer padding.
            f_Algo.Update(f_PadAndKey, 0, l_BlockSize);
            // Hash inner hash.
            f_Algo.Update(f_PadAndKey, 2 * l_BlockSize, l_HashSize);
            f_Algo.Finish();

            f_Condition = HashFunctionCondition.Finished;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            Finish();

            f_Algo.GetDigest(a_Dest, a_Idx);
        }

        private void PreparePad(bool a_InnerPad)
        {
            // The result will be stored at offser zero of f_PadAndKey.

            int l_BlockSize = p_BlockByteSize;
            byte l_PadVal = (false == a_InnerPad ? (byte)0x5C : (byte)0x36);

            for (int i = 0, k = l_BlockSize; i != l_BlockSize; i++, k++)
            {
                f_PadAndKey[i] = (byte)(l_PadVal ^ f_PadAndKey[k]);
            }
        }

        private static byte[] CreatePadAndKey(IHashFunction a_Algo,
                                              byte[] a_RawKey,
                                              int a_RawKeyIdx,
                                              int a_RawKeyLen)
        {
            // This routine will change the state of a_Algo.

            int l_HashSize = a_Algo.p_HashByteSize;
            int l_BlockSize = a_Algo.p_BlockByteSize;

            byte[] l_Arr = new byte[(2 * l_BlockSize) + l_HashSize];

            if (l_BlockSize < a_RawKeyLen)
            {
                // The key is too long, so hash it.

                a_Algo.Reset();
                a_Algo.Update(a_RawKey, a_RawKeyIdx, a_RawKeyLen);
                a_Algo.GetDigest(l_Arr, l_BlockSize);
            }
            else
            {
                Buffer.BlockCopy(a_RawKey, a_RawKeyIdx, l_Arr, l_BlockSize, a_RawKeyLen);
            }

            return l_Arr;
        }

        internal static bool SelfTest()
        {
            byte[] l_DummyPlainText = new byte[3] { 0x0, 0x1, 0x2 };

            MD_4 l_MD = new MD_4();
            HMACWrapper l_HMACMD4 = new HMACWrapper(l_MD, new byte[1] { 0x01 } );
            l_HMACMD4.Update(l_DummyPlainText);

            byte[] l_Digest = new byte[MD_4.c_HashByteSize];
            l_HMACMD4.GetDigest(l_Digest, 0);

            if ("1F000373CFC4C64C4243452049AF6CCD" != Util.BytesToString(l_Digest))
            {
                return false;
            }

            l_HMACMD4 = new HMACWrapper(l_MD, Transformer.BuildTestBuffer());
            l_HMACMD4.Update(l_DummyPlainText);
            l_HMACMD4.GetDigest(l_Digest, 0);

            return ("34C605243C365093E5BE784461D434C5" == Util.BytesToString(l_Digest));
        }
    }

    public abstract class Transformer
    {
        private const ushort c_MaxBlockSize = ushort.MaxValue;

        public delegate void NativeTransformRoutine(IntPtr a_States,
                                                    IntPtr a_Block,
                                                    uint a_BlockCount);

#if FAST
        private static readonly bool s_AlignTransformBuffer;
#endif

        // f_BlockSize is guaranteed not to exceed ushort.MaxValue.
        protected readonly int f_BlockSize;
        private readonly byte[] f_Buf;

        private int f_BufUsage;
        private ulong f_MessageLength;
        private bool f_Finished;

        public ushort p_BlockByteSize
        {
            get { return (ushort)f_BlockSize; }
        }

#if FAST
        static Transformer()
        {
            s_AlignTransformBuffer = (4 != IntPtr.Size);
        }
#endif

        protected Transformer(ushort a_BlockSize)
        {
            f_BlockSize = a_BlockSize;
            f_Buf = new byte[f_BlockSize];
        }

        public void Reset()
        {
            f_BufUsage = 0;
            f_MessageLength = 0;
            f_Finished = false;

            ResetCore();
        }

        protected abstract void ResetCore();

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Data)
        {
            Update(a_Data, 0, a_Data.Length);
        }

        public virtual void Update(byte[] a_Data,
                                   int a_Idx,
                                   int a_Len)
        {
            UpdateCore(a_Data, a_Idx, a_Len);
        }

        protected void UpdateCore(byte[] a_Data,
                                  int a_Idx,
                                  int a_Len)
        {
#if FAST
            // Do a thorough check here even in release builds,
            // since we might call unsafe code.
            Util.VerifyRangeOrFail(a_Idx, a_Len, a_Data.Length);
#endif

            f_MessageLength += (uint)a_Len;

            int l_Idx = a_Idx;
            int l_Len = a_Len;

            if (0 < f_BufUsage)
            {
                int l_Missing = (f_BlockSize - f_BufUsage);

                Buffer.BlockCopy(a_Data, l_Idx, f_Buf, f_BufUsage, Math.Min(l_Missing, l_Len));

                if (l_Missing > l_Len)
                {
                    f_BufUsage += l_Len;

                    return;
                }

                TransformBlock(f_Buf);

                f_BufUsage = 0;
                l_Idx += l_Missing;
                l_Len -= l_Missing;
            }

            if (f_BlockSize <= l_Len)
            {
                int l_BlockCount = (l_Len / f_BlockSize);
                int l_BlockRangeSize = (l_BlockCount * f_BlockSize);

#if FAST
                if (true == s_AlignTransformBuffer &&
                    0 != (l_Idx % sizeof(ulong)))
                {
                    // We need to copy the data in order to align it.

                    // This byte size limit must be at least c_MaxBlockSize,
                    // so that the buffer can contain at least one whole block.
                    const int l_MaxTempBufferSize = (256 * 1024);
                    int l_TempBufBlocks = Math.Min(l_BlockCount, l_MaxTempBufferSize / f_BlockSize);

                    byte[] l_TempBuf = new byte[l_TempBufBlocks * f_BlockSize];

                    for (int l_BlocksToGo = l_BlockCount; 0 < l_BlocksToGo; )
                    {
                        int l_CurBlocks = Math.Min(l_BlocksToGo, l_TempBufBlocks);
                        int l_CurSize = (l_CurBlocks * f_BlockSize);

                        Buffer.BlockCopy(a_Data, l_Idx, l_TempBuf, 0, l_CurSize);
                        TransformBlocksCore(l_TempBuf, 0, l_CurBlocks);

                        l_Idx += l_CurSize;
                        l_BlocksToGo -= l_CurBlocks;
                    }
                }
                else
                {
#endif
                    TransformBlocksCore(a_Data, l_Idx, l_BlockCount);

                    l_Idx += l_BlockRangeSize;
#if FAST
                }
#endif

                l_Len -= l_BlockRangeSize;
            }

            if (0 < l_Len)
            {
                // Store remaining bytes for later.

                Buffer.BlockCopy(a_Data, l_Idx, f_Buf, 0, l_Len);
                f_BufUsage = l_Len;
            }
        }

        public void Finish()
        {
            if (false == f_Finished)
            {
                TransformFinalBlock(f_Buf);
                f_Finished = true;
            }
        }

        public bool IsFinished()
        {
            return f_Finished;
        }

        private void TransformBlock(byte[] a_Block)
        {
            TransformBlocksCore(a_Block, 0, 1);
        }

        private void TransformBlocksCore(byte[] a_Block,
                                         int a_ByteIndex,
                                         int a_BlockCount)
        {
            TransformBlocks(a_Block, a_ByteIndex, a_BlockCount);
        }

        // The memory location (a_Block + a_ByteIndex) is guaranteed
        // to be eight bytes aligned if s_AlignTransformBuffer is true.
        protected abstract void TransformBlocks(byte[] a_Block,
                                                int a_ByteIndex,
                                                int a_BlockCount);

        // The memory location a_Block is always aligned.
        protected abstract void TransformFinalBlock(byte[] a_Block);

        protected void DefaultTransformFinalBlock(byte[] a_Block,
                                                  PaddingKinds a_Kind)
        {
            int l_BytesBeforeLenPadding = (f_BlockSize - sizeof(ulong));
            int l_Len = f_BufUsage;

            a_Block[l_Len++] = GetPadByte(a_Kind);

            if (l_BytesBeforeLenPadding < l_Len)
            {
                // There is not enough space (64 bit) left to append
                // the size of the message right now.

                Buffer2.BlockZero(a_Block, l_Len, f_BlockSize - l_Len);

                TransformBlock(a_Block);

                l_Len = 0;
            }

            Buffer2.BlockZero(a_Block, l_Len, l_BytesBeforeLenPadding - l_Len);
            ulong l_BitLength = (f_MessageLength * Util.c_BitsPerByte);

            if (0 == (a_Kind & PaddingKinds.BitLengthInBE))
            {
                Util.WriteUInt64LE(l_BitLength, a_Block, l_BytesBeforeLenPadding);
            }
            else
            {
                Util.WriteUInt64BE(l_BitLength, a_Block, l_BytesBeforeLenPadding);
            }

            if (0 == (a_Kind & PaddingKinds.OmitLastTransform))
            {
                TransformBlock(a_Block);
            }
        }

        protected void DefaultTransformFinalKeccakBlock(byte[] a_Block,
                                                        byte a_PadByte)
        {
            uint l_LastBlockIdx = ((uint)f_BlockSize - 1);
            const byte lc_FinalBit = 0x80;

            if (l_LastBlockIdx == (uint)f_BufUsage)
            {
                f_Buf[l_LastBlockIdx] = (byte)(a_PadByte | lc_FinalBit);
            }
            else
            {
                a_Block[f_BufUsage] = a_PadByte;
                int l_ClearLen = ((int)l_LastBlockIdx - (f_BufUsage + 1));

                if (0 < l_ClearLen)
                {
                    Array.Clear(a_Block, f_BufUsage + 1, l_ClearLen);
                }

                a_Block[l_LastBlockIdx] = lc_FinalBit;
            }

            TransformBlock(a_Block);
        }

        private static byte GetPadByte(PaddingKinds a_Kind)
        {
            if (0 != (a_Kind & PaddingKinds.PadByte0x80))
            {
                return 0x80;
            }

            if (0 != (a_Kind & PaddingKinds.PadByte0x01))
            {
                return 0x1;
            }

            return 0x5A;
        }

        public static byte[] BuildTestBuffer()
        {
            byte[] l_Buf = new byte[217];

            for (int i = (l_Buf.Length - 1); 0 <= i; i--)
            {
                l_Buf[i] = (byte)i;
            }

            return l_Buf;
        }

        [Flags]
        public enum PaddingKinds : ushort
        {
            PadByte0x01 = 0x1,
            PadByte0x5A = 0x2,
            PadByte0x80 = 0x4,

            BitLengthInBE = 0x100,
            OmitLastTransform = 0x200,

            // Compositions.
            None = 0,
            MD4 = PadByte0x80,
            SHA256 = (PadByte0x80 | BitLengthInBE),
            Tiger = PadByte0x01,
            Tiger2 = PadByte0x80,
            Lookup3 = (PadByte0x5A | OmitLastTransform),
        }
    }

    public class Lookup3 : Transformer, IHashFunction
    {
        // Based on the public domain file lookup3.c,
        // by Bob Jenkins, May 2006.
        // See: <http://www.burtleburtle.net/bob/c/lookup3.c>
        // NOTE: The original algorithm handles initialization and
        // completion differently and therefore produces different
        // digests.

        // Use this algo for hashing short messages when collision
        // avoidance is especially important (where Adler32 isn't
        // suitable).

        public const byte c_HashByteSize = sizeof(ulong);
        private const byte c_BlockSize = (3 * sizeof(uint));

        private uint f_a;
        private uint f_b;
        private uint f_c;

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public Lookup3() :
                  base(c_BlockSize)
        {
            ResetCore();
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            Lookup3 l_Lookup3 = new Lookup3();

            l_Lookup3.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_Lookup3.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        protected override void ResetCore()
        {
            // Random generated.
            f_a = 0x86B3453E;
            f_b = 0xBA3B4543;
            f_c = 0xE83E537C;
        }

        protected override void TransformBlocks(byte[] a_Block,
                                                int a_ByteIndex,
                                                int a_BlockCount)
        {
            PrepStates(a_Block, a_ByteIndex, a_BlockCount, true);
        }

        protected override void TransformFinalBlock(byte[] a_Block)
        {
            base.DefaultTransformFinalBlock(a_Block, PaddingKinds.Lookup3);

            // We can't use the normal transform routine for
            // the last block, instead we use this code.

            PrepStates(a_Block, 0, 1, false);

            Final();
        }

        private
#if FAST
        unsafe
#endif
        void PrepStates(byte[] a_Block,
                        int a_ByteIndex,
                        int a_BlockCount,
                        bool a_MixAfterEachRound)
        {
            int l_Limit = (a_ByteIndex + (a_BlockCount * c_BlockSize));

#if FAST
            fixed (byte* a_Bytes = a_Block)
            {
                for (int i = a_ByteIndex; i < l_Limit; i += c_BlockSize)
                {
                    uint* a_Ptr = (uint*)&a_Bytes[i];
                    f_a += a_Ptr[0];
                    f_b += a_Ptr[1];
                    f_c += a_Ptr[2];

                    if (true == a_MixAfterEachRound)
                    {
                        Mix();
                    }
                }
            }
#else
            for (int i = a_ByteIndex; i < l_Limit; i += c_BlockSize)
            {
                f_a += Util.ReadUInt32LE(a_Block, i);
                f_b += Util.ReadUInt32LE(a_Block, i + sizeof(uint));
                f_c += Util.ReadUInt32LE(a_Block, i + (2 * sizeof(uint)));

                if (true == a_MixAfterEachRound)
                {
                    Mix();
                }
            }
#endif
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            Finish();

            Util.WriteUInt32LE(f_c, a_Dest, a_Idx);
            Util.WriteUInt32LE(f_b, a_Dest, a_Idx + sizeof(uint));
        }

        public ulong GetDigestAsULong()
        {
            Finish();

            return (((ulong)f_b << 32) | f_c);
        }

        private void Mix()
        {
            f_a -= f_c; f_a ^= Util.RotateLeft(f_c, 4); f_c += f_b;
            f_b -= f_a; f_b ^= Util.RotateLeft(f_a, 6); f_a += f_c;
            f_c -= f_b; f_c ^= Util.RotateLeft(f_b, 8); f_b += f_a;
            f_a -= f_c; f_a ^= Util.RotateLeft(f_c, 16); f_c += f_b;
            f_b -= f_a; f_b ^= Util.RotateLeft(f_a, 19); f_a += f_c;
            f_c -= f_b; f_c ^= Util.RotateLeft(f_b, 4); f_b += f_a;
        }

        private void Final()
        {
            f_c ^= f_b; f_c -= Util.RotateLeft(f_b, 14);
            f_a ^= f_c; f_a -= Util.RotateLeft(f_c, 11);
            f_b ^= f_a; f_b -= Util.RotateLeft(f_a, 25);
            f_c ^= f_b; f_c -= Util.RotateLeft(f_b, 16);
            f_a ^= f_c; f_a -= Util.RotateLeft(f_c, 4);
            f_b ^= f_a; f_b -= Util.RotateLeft(f_a, 14);
            f_c ^= f_b; f_c -= Util.RotateLeft(f_b, 24);
        }

        public static bool SelfTest()
        {
            if ("B631BE4B1945D373" !=
                Util.BytesToString(Lookup3.ComputeHash(new byte[0], 0, 0)))
            {
                return false;
            }

            byte[] l_TestBuf = BuildTestBuffer();

            Lookup3 l_L = new Lookup3();
            l_L.Reset();

            l_L.Update(Encoding.ASCII.GetBytes("abcdbcdecdefdefgefghfgh"));
            l_L.Update(Encoding.ASCII.GetBytes("ighij"));
            l_L.Update(Encoding.ASCII.GetBytes("hijkijkljklmklmnlmnomnopnopq"));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_L.GetDigest(l_Digest, 0);

            if ("3207FD38192CB0F5" != Util.BytesToString(l_Digest))
            {
                return false;
            }

            return true;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct SipHash
    {
        // The struct default constructor may be used.

        public const byte c_HashByteSize = sizeof(ulong);
        private const byte c_KeyByteSize = (2 * sizeof(ulong));

        private ulong f_v0;
        private ulong f_v1;
        private ulong f_v2;
        private ulong f_v3;

        private void SetKey(ref Key a_Key)
        {
            f_v0 = (0x736F6D6570736575UL ^ a_Key.f_Key0);
            f_v1 = (0x646F72616E646F6DUL ^ a_Key.f_Key1);
            f_v2 = (0x6C7967656E657261UL ^ a_Key.f_Key0);
            f_v3 = (0x7465646279746573UL ^ a_Key.f_Key1);
        }

        public static int CalculateHashCode(ref Key a_Key,
                                            byte[] a_SrcArr,
                                            int a_SrcIdx,
                                            int a_SrcLen)
        {
            // This routine outputs hash codes in the .NET format.

            SipHash l_SipHash = new SipHash();
            ulong l_Digest = l_SipHash.CalculateDigest(ref a_Key, a_SrcArr, a_SrcIdx, a_SrcLen);

            return DigestToDotNetHashCode(l_Digest);
        }

        public ulong CalculateDigest(ref Key a_Key,
                                     ref ByteArray a_Data)
        {
            return CalculateDigest(ref a_Key, a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public ulong CalculateDigest(ref Key a_Key,
                                     byte[] a_SrcArr,
                                     int a_SrcIdx,
                                     int a_SrcLen)
        {
            SetKey(ref a_Key);

            int l_Idx = a_SrcIdx;
            int l_Lim = (l_Idx + a_SrcLen);

            // Process whole ulongs.

            for (int l_WholeULongLim = (l_Lim - (a_SrcLen % sizeof(ulong))); l_Idx != l_WholeULongLim; l_Idx += sizeof(ulong))
            {
                SipCompress(Util.ReadUInt64LE(a_SrcArr, l_Idx));
            }

            // Process the final ulong and/or hash the input size.

            ulong l_m = (ReadUInt64LEPartial(a_SrcArr, l_Idx, l_Lim - l_Idx) |
                         ((ulong)((uint)a_SrcLen & byte.MaxValue) << (Util.c_BitsPerULong - Util.c_BitsPerByte)));
            SipCompress(l_m);

            f_v2 ^= byte.MaxValue;
            SipRound();
            SipRound();
            SipRound();
            SipRound();
            ulong l_Digest = (f_v0 ^ f_v1 ^ f_v2 ^ f_v3);

            return l_Digest;
        }

        private void SipCompress(ulong a_m)
        {
            f_v3 ^= a_m;
            SipRound();
            SipRound();
            f_v0 ^= a_m;
        }

        private void SipRound()
        {
            f_v0 += f_v1;
            f_v1 = Rotl64(f_v1, 13);
            f_v1 ^= f_v0;
            f_v0 = Rotl64(f_v0, 32);
            f_v2 += f_v3;
            f_v3 = Rotl64(f_v3, 16);
            f_v3 ^= f_v2;
            f_v0 += f_v3;
            f_v3 = Rotl64(f_v3, 21);
            f_v3 ^= f_v0;
            f_v2 += f_v1;
            f_v1 = Rotl64(f_v1, 17);
            f_v1 ^= f_v2;
            f_v2 = Rotl64(f_v2, 32);
        }

        private static ulong Rotl64(ulong x,
                                    int n)
        {
            return ((x << n) | (x >> (Util.c_BitsPerULong - n)));
        }

        private static ulong ReadUInt64LEPartial(byte[] a_Src,
                                                 int a_Idx,
                                                 int a_Len)
        {
            // a_Len may be zero.

            ulong l_Res = 0;

            for (int i = 0, k = a_Idx; i < a_Len; i++, k++)
            {
                l_Res |= ((ulong)a_Src[k] << (i * Util.c_BitsPerByte));
            }

            return l_Res;
        }

        private static int DigestToDotNetHashCode(ulong a_Digest)
        {
            uint l_LowUInt = (uint)(a_Digest & uint.MaxValue);
            uint l_HighUInt = (uint)((a_Digest >> Util.c_BitsPerUInt) & uint.MaxValue);

            return (int)(l_LowUInt ^ l_HighUInt);
        }

        public static bool SelfTest()
        {
            Key l_Key = new Key(0xA36CE4A7AEF2FAE0UL, 0xE46E7F58E55604B8UL);
            SipHash l_SipHash = new SipHash();

            if (0x37A10E6F045C1FFBUL != l_SipHash.CalculateDigest(ref l_Key, new byte[0], 0, 0))
            {
                return false;
            }

            if (0x3BC55171CFC5B538UL != l_SipHash.CalculateDigest(ref l_Key, Encoding.ASCII.GetBytes("_NACHTBLITZ ASINUS EST_"), 1, 21))
            {
                return false;
            }

            return true;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Key
        {
            // The struct default constructor may be used. In this case
            // the keys remain zero.

            public readonly ulong f_Key0;
            public readonly ulong f_Key1;

            public Key(ulong a_Key0,
                       ulong a_Key1)
            {
                f_Key0 = a_Key0;
                f_Key1 = a_Key1;
            }

            public static void CreateRandom(out Key a_Dest)
            {
                a_Dest = new Key(Util.GetRandomULong(), Util.GetRandomULong());
            }
        }
    }

    public class Tiger : Transformer, IHashFunction
    {
        public const byte c_HashByteSize = 24;
        private const byte c_ULongsPerBlock = 8;
        private const byte c_BlockSize = (c_ULongsPerBlock * sizeof(ulong));
        private const byte c_States = 3;

        private const byte c_SBoxes = 4;
        private const ushort c_SBoxElementsPerBox = 256;
        private const ushort c_SBoxesByteSize = (c_SBoxes * c_SBoxElementsPerBox * sizeof(ulong));
        // Offsets of each substitution box (in ulong elements).
        private const ushort c_SBoxT1 = (0 * c_SBoxElementsPerBox);
        private const ushort c_SBoxT2 = (1 * c_SBoxElementsPerBox);
        private const ushort c_SBoxT3 = (2 * c_SBoxElementsPerBox);
        private const ushort c_SBoxT4 = (3 * c_SBoxElementsPerBox);

        private static readonly ulong[] s_SBoxes;

        public readonly bool f_Tiger2;
        private ulong[] f_States;
#if FAST
        UnsafeBlockStruct f_bs;
#else
        private ulong[] f_x;
#endif

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        private PaddingKinds p_PaddingKind
        {
            get { return (false == f_Tiger2 ? PaddingKinds.Tiger : PaddingKinds.Tiger2); }
        }

        static Tiger()
        {
            s_SBoxes = new ulong[c_SBoxes * c_SBoxElementsPerBox];

#if FAST
            s_ExternalCompress = GetExternalCompressRoutine();
#endif
        }

        public Tiger() :
               this(false)
        {
        }

        public Tiger(bool a_Tiger2) :
                base(c_BlockSize)
        {
            f_Tiger2 = a_Tiger2;
            f_States = new ulong[c_HashByteSize / sizeof(ulong)];
#if FAST
            f_bs = new UnsafeBlockStruct();
#else
            f_x = new ulong[c_BlockSize / sizeof(ulong)];
#endif

            ResetCore();
        }

        public static bool LoadBoxBuffer(Stream a_Strm,
                                         int a_Size,
                                         uint a_Adler32,
                                         out byte[] a_Buffer)
        {
            if (a_Size == a_Strm.Length)
            {
                ByteArray l_Data = new ByteArray(a_Size);

                if (Util.ReadFromStream(a_Strm, ref l_Data) == l_Data.f_Len &&
                    Adler32.ComputeHash(ref l_Data) == a_Adler32)
                {
                    a_Buffer = l_Data.f_Arr;

                    return true;
                }
            }

            a_Buffer = null;

            return false;
        }

        public static bool LoadBoxes(Stream a_Strm)
        {
            byte[] l_Buffer;

            if (true == LoadBoxBuffer(a_Strm, c_SBoxesByteSize, 0xEEC3F0E2, out l_Buffer))
            {
                Buffer.BlockCopy(l_Buffer, 0, s_SBoxes, 0, c_SBoxesByteSize);

                return true;
            }

            return false;
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            Tiger l_Tiger = new Tiger();

            l_Tiger.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_Tiger.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        protected override void ResetCore()
        {
            f_States[0] = 0x0123456789ABCDEFUL;
            f_States[1] = 0xFEDCBA9876543210UL;
            f_States[2] = 0xF096A5B4C3B2E187UL;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            // Returns the default 24 byte tiger.

            ReturnTiger(a_Dest, a_Idx, c_HashByteSize);
        }

        public void GetDigest128(byte[] a_Dest,
                                 int a_Idx)
        {
            // Returns the 16 byte tiger.

            ReturnTiger(a_Dest, a_Idx, 16);
        }

        public void GetDigest160(byte[] a_Dest,
                                 int a_Idx)
        {
            // Returns the 20 byte tiger.

            ReturnTiger(a_Dest, a_Idx, 20);
        }

        private void ReturnTiger(byte[] a_Dest,
                                 int a_Idx,
                                 int a_Len)
        {
            Finish();

            Buffer.BlockCopy(f_States, 0, a_Dest, a_Idx, a_Len);
        }

        public
#if FAST
        unsafe
#endif
        bool VerifyBE(byte[] a_Src,
                      int a_Idx)
        {
#if FAST
            // We only use the fast track if aligned.
            if (0 == (a_Idx % sizeof(ulong)))
            {
                Util.VerifyRangeOrFail(a_Idx, c_HashByteSize, a_Src.Length);

                fixed (ulong* l_States = f_States)
                {
                    fixed (byte* l_Bytes = a_Src)
                    {
                        ulong* l_Ptr = (ulong*)(l_Bytes + a_Idx);

                        return (l_Ptr[0] == l_States[0] &&
                                l_Ptr[1] == l_States[1] &&
                                l_Ptr[2] == l_States[2]);
                    }
                }
            }
#endif

            return (Util.ReadUInt64LE(a_Src, a_Idx) == f_States[0] &&
                    Util.ReadUInt64LE(a_Src, a_Idx + 8) == f_States[1] &&
                    Util.ReadUInt64LE(a_Src, a_Idx + 16) == f_States[2]);
        }

        public override void Update(byte[] a_Data,
                                    int a_Idx,
                                    int a_Len)
        {
#if FAST
            // We'll fix the arrays and then call UpdateCore().
            FixAndCall(false, a_Data, a_Idx, a_Len);
#else
            UpdateCore(a_Data, a_Idx, a_Len);
#endif
        }

        protected override void TransformFinalBlock(byte[] a_Block)
        {
#if FAST
            // We'll fix the arrays and then call DefaultTransformFinalBlock().
            FixAndCall(true, a_Block, 0, 0);
#else
            base.DefaultTransformFinalBlock(a_Block, p_PaddingKind);
#endif
        }

        protected
#if FAST
        unsafe
#endif
        override void TransformBlocks(byte[] a_Block,
                                      int a_ByteIndex,
                                      int a_BlockCount)
        {
            // We have 4 different implementations.
            // - managed and safe C# (TigerCompress)
            // - managed and unsafe C# (TigerCompressFast)
            // - native x86 C (tiger_multicompress_x86)
            // - native x64 C (tiger_multicompress_x64)

#if FAST
            fixed (byte* l_BlockArray = a_Block)
            {
                byte* l_BlockPtr = &l_BlockArray[a_ByteIndex];

                if (null == s_ExternalCompress)
                {
                    TigerCompressFast(l_BlockPtr, a_BlockCount);
                }
                else
                {
                    s_ExternalCompress(f_States_FAST, (IntPtr)l_BlockPtr, (uint)a_BlockCount);
                }
            }
#else
            TigerCompress(a_Block, a_ByteIndex, a_BlockCount);
#endif
        }

#if FAST

        private struct UnsafeBlockStruct
        {
            public unsafe fixed ulong f_x[c_ULongsPerBlock];

            public static unsafe void CopyIn(ulong* a_Src,
                                             ulong* a_Dest)
            {
                for (int i = (c_ULongsPerBlock - 1); 0 <= i; i--)
                {
                    a_Dest[i] = a_Src[i];
                }
            }
        }

        private static readonly CompressRoutine s_ExternalCompress;

        private IntPtr f_SBoxes_FAST;
        private IntPtr f_States_FAST;
        private IntPtr f_x_FAST;

        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void tiger_multicompress_x86([In] IntPtr a_States,
                                                           [In] IntPtr a_Block,
                                                           [In] uint a_BlockCount);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void tiger_multicompress_x64([In] IntPtr a_States,
                                                           [In] IntPtr a_Block,
                                                           [In] uint a_BlockCount);

        private delegate void CompressRoutine(IntPtr a_States,
                                              IntPtr a_Block,
                                              uint a_BlockCount);

        private static unsafe CompressRoutine GetExternalCompressRoutine()
        {
            CompressRoutine l_Target = null;

            if (true == Util.s_IsWindows)
            {
                if (sizeof(uint) == IntPtr.Size)
                {
                    l_Target = tiger_multicompress_x86;
                }
                else if (sizeof(ulong) == IntPtr.Size)
                {
                    l_Target = tiger_multicompress_x64;
                }

                if (null != l_Target)
                {
                    byte* l_DummyStates = stackalloc byte[c_HashByteSize];
                    byte* l_DummyBlock = stackalloc byte[c_BlockSize];

                    try
                    {
                        l_Target((IntPtr)l_DummyStates, (IntPtr)l_DummyBlock, 1);
                    }
                    catch
                    {
                        l_Target = null;
                    }
                }
            }

            return l_Target;
        }

        private unsafe void FixAndCall(bool a_FinalBlock,
                                       byte[] a_Buffer,
                                       int a_Idx,
                                       int a_Len)
        {
            fixed (ulong* l_SBoxes = s_SBoxes)
            {
                fixed (ulong* l_States = f_States)
                {
                    f_SBoxes_FAST = (IntPtr)l_SBoxes;
                    f_States_FAST = (IntPtr)l_States;

                    if (false == a_FinalBlock)
                    {
                        base.UpdateCore(a_Buffer, a_Idx, a_Len);
                    }
                    else
                    {
                        base.DefaultTransformFinalBlock(a_Buffer, p_PaddingKind);
                    }
                }
            }
        }

        private unsafe void TigerCompressFast(byte* a_BlockSrc,
                                              int a_BlockCount)
        {
            // Check that FixAndCall() is executing.

            ulong* l_States = (ulong*)f_States_FAST;

            fixed (ulong* l_TempBlockPtr = f_bs.f_x)
            {
                // The block buffer gets written to, so we may not
                // simply take the input buffer for that.
                f_x_FAST = (IntPtr)l_TempBlockPtr;

                for (int i = 0; i < a_BlockCount; i++)
                {
                    ulong l_SavedState0 = l_States[0];
                    ulong l_SavedState1 = l_States[1];
                    ulong l_SavedState2 = l_States[2];

                    UnsafeBlockStruct.CopyIn((ulong*)a_BlockSrc, l_TempBlockPtr);

                    PassFast(0, 1, 2, 5);
                    KeyScheduleFast();
                    PassFast(2, 0, 1, 7);
                    KeyScheduleFast();
                    PassFast(1, 2, 0, 9);

                    a_BlockSrc += c_BlockSize;

                    l_States[0] ^= l_SavedState0;
                    l_States[1] -= l_SavedState1;
                    l_States[2] += l_SavedState2;
                }
            }
        }

        private void PassFast(int a,
                              int b,
                              int c,
                              uint mul)
        {
            RoundFast(a, b, c, 0, mul);
            RoundFast(b, c, a, 1, mul);
            RoundFast(c, a, b, 2, mul);
            RoundFast(a, b, c, 3, mul);
            RoundFast(b, c, a, 4, mul);
            RoundFast(c, a, b, 5, mul);
            RoundFast(a, b, c, 6, mul);
            RoundFast(b, c, a, 7, mul);
        }

        private unsafe void RoundFast(int a,
                                      int b,
                                      int c,
                                      int xi,
                                      uint mul)
        {
            ulong* l_SBoxes = (ulong*)f_SBoxes_FAST;
            ulong* l_States = (ulong*)f_States_FAST;
            byte* l_C = stackalloc byte[sizeof(ulong)];

            l_States[c] ^= ((ulong*)f_x_FAST)[xi];
            *(ulong*)l_C = l_States[c];

            l_States[a] -= (l_SBoxes[c_SBoxT1 + l_C[0]] ^ l_SBoxes[c_SBoxT2 + l_C[2]] ^ l_SBoxes[c_SBoxT3 + l_C[4]] ^ l_SBoxes[c_SBoxT4 + l_C[6]]);
            l_States[b] += (l_SBoxes[c_SBoxT1 + l_C[7]] ^ l_SBoxes[c_SBoxT2 + l_C[5]] ^ l_SBoxes[c_SBoxT3 + l_C[3]] ^ l_SBoxes[c_SBoxT4 + l_C[1]]);
            l_States[b] *= mul;
        }

        private unsafe void KeyScheduleFast()
        {
            ulong* l_x = (ulong*)f_x_FAST;

            l_x[0] -= (l_x[7] ^ 0xA5A5A5A5A5A5A5A5UL);
            l_x[1] ^= l_x[0];
            l_x[2] += l_x[1];
            l_x[3] -= (l_x[2] ^ ((~l_x[1]) << 19));
            l_x[4] ^= l_x[3];
            l_x[5] += l_x[4];
            l_x[6] -= (l_x[5] ^ ((~l_x[4]) >> 23));
            l_x[7] ^= l_x[6];
            l_x[0] += l_x[7];
            l_x[1] -= (l_x[0] ^ ((~l_x[7]) << 19));
            l_x[2] ^= l_x[1];
            l_x[3] += l_x[2];
            l_x[4] -= (l_x[3] ^ ((~l_x[2]) >> 23));
            l_x[5] ^= l_x[4];
            l_x[6] += l_x[5];
            l_x[7] -= (l_x[6] ^ 0x0123456789ABCDEFUL);
        }

#else

        private void TigerCompress(byte[] a_Block,
                                   int a_ByteIndex,
                                   int a_BlockCount)
        {
            for (int i = 0, l_SrcIdx = a_ByteIndex; i < a_BlockCount; i++, l_SrcIdx += c_BlockSize)
            {
                Buffer.BlockCopy(a_Block, l_SrcIdx, f_x, 0, c_BlockSize);

                ulong l_SavedState0 = f_States[0];
                ulong l_SavedState1 = f_States[1];
                ulong l_SavedState2 = f_States[2];

                Pass(0, 1, 2, 5);
                KeySchedule();
                Pass(2, 0, 1, 7);
                KeySchedule();
                Pass(1, 2, 0, 9);

                f_States[0] ^= l_SavedState0;
                f_States[1] -= l_SavedState1;
                f_States[2] += l_SavedState2;
            }
        }

        private void Pass(int a,
                          int b,
                          int c,
                          uint mul)
        {
            Round(a, b, c, 0, mul);
            Round(b, c, a, 1, mul);
            Round(c, a, b, 2, mul);
            Round(a, b, c, 3, mul);
            Round(b, c, a, 4, mul);
            Round(c, a, b, 5, mul);
            Round(a, b, c, 6, mul);
            Round(b, c, a, 7, mul);
        }

        private void Round(int a,
                           int b,
                           int c,
                           int xi,
                           uint mul)
        {
            f_States[c] ^= f_x[xi];

            ulong l_C = f_States[c];
            f_States[a] -= (s_SBoxes[c_SBoxT1 + (int)(l_C & 0xFF)] ^ s_SBoxes[c_SBoxT2 + ((int)(l_C >> 16) & 0xFF)] ^ s_SBoxes[c_SBoxT3 + ((int)(l_C >> 32) & 0xFF)] ^ s_SBoxes[c_SBoxT4 + ((int)(l_C >> 48) & 0xFF)]);
            f_States[b] += (s_SBoxes[c_SBoxT1 + ((int)(l_C >> 56) & 0xFF)] ^ s_SBoxes[c_SBoxT2 + ((int)(l_C >> 40) & 0xFF)] ^ s_SBoxes[c_SBoxT3 + ((int)(l_C >> 24) & 0xFF)] ^ s_SBoxes[c_SBoxT4 + ((int)(l_C >> 8) & 0xFF)]);
            f_States[b] *= mul;
        }

        private void KeySchedule()
        {
            f_x[0] -= (f_x[7] ^ 0xA5A5A5A5A5A5A5A5UL);
            f_x[1] ^= f_x[0];
            f_x[2] += f_x[1];
            f_x[3] -= (f_x[2] ^ ((~f_x[1]) << 19));
            f_x[4] ^= f_x[3];
            f_x[5] += f_x[4];
            f_x[6] -= (f_x[5] ^ ((~f_x[4]) >> 23));
            f_x[7] ^= f_x[6];
            f_x[0] += f_x[7];
            f_x[1] -= (f_x[0] ^ ((~f_x[7]) << 19));
            f_x[2] ^= f_x[1];
            f_x[3] += f_x[2];
            f_x[4] -= (f_x[3] ^ ((~f_x[2]) >> 23));
            f_x[5] ^= f_x[4];
            f_x[6] += f_x[5];
            f_x[7] -= (f_x[6] ^ 0x0123456789ABCDEFUL);
        }

#endif

        public static bool SelfTest()
        {
            if ("3293AC630C13F0245F92BBB1766E16167A4E58492DDE73F3" !=
                Util.BytesToString(Tiger.ComputeHash(new byte[0], 0, 0)))
            {
                return false;
            }

            Tiger l_T = new Tiger();

            l_T.Update(Encoding.ASCII.GetBytes("_abcdbcdecdefdefgefghfgh__"), 1, 23);
            l_T.Update(Encoding.ASCII.GetBytes("__ighij"), 2, 5);
            l_T.Update(Encoding.ASCII.GetBytes("hijkijkljklmklmnlmnomnopnopq_"), 0, 28);

            byte[] l_Digest = new byte[c_HashByteSize];

            l_T.GetDigest(l_Digest, 0);

            if ("0F7BF9A19B9C58F2B7610DF7E84F0AC3A71C631E7B53F78E" !=
                Util.BytesToString(l_Digest))
            {
                return false;
            }

            l_T.Reset();
            l_T.Update(Encoding.ASCII.GetBytes("12345678901234567890123456789012345678901234567890123456789012345678901234567890"));

            l_T.GetDigest(l_Digest, 0);

            if ("1C14795529FD9F207A958F84C52F11E887FA0CABDFD91BFD" !=
                Util.BytesToString(l_Digest))
            {
                return false;
            }

            // Test Tiger2.

            l_T = new Tiger(true);
            l_T.GetDigest(l_Digest, 0);

            return ("4441BE75F6018773C206C22745374B924AA8313FEF919F41" ==
                    Util.BytesToString(l_Digest));
        }
    }

    public class MD_4 : Transformer, IHashFunction
    {
        // See: RFC 1320 MD4 Message-Digest Algorithm April 1992.

        public const byte c_HashByteSize = 16;
        private const byte c_BlockSize = 64;

        private readonly uint[] f_Registers;
#if !FAST
        protected readonly uint[] f_Block;
#endif

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        static MD_4()
        {
#if FAST
            s_ExternalTransform = GetExternalTransformRoutine();
#endif
        }

        public MD_4() :
               base(c_BlockSize)
        {
            f_Registers = new uint[c_HashByteSize / sizeof(uint)];
#if !FAST
            f_Block = new uint[c_BlockSize / sizeof(uint)];
#endif

            ResetCore();
        }

        protected override void ResetCore()
        {
            // See RFC 1320 3.3.

            f_Registers[0] = 0x67452301;
            f_Registers[1] = 0xEFCDAB89;
            f_Registers[2] = 0x98BADCFE;
            f_Registers[3] = 0x10325476;
        }

        public override void Update(byte[] a_Data,
                                    int a_Idx,
                                    int a_Len)
        {
#if FAST
            // We'll fix the arrays and then call UpdateCore().
            FixAndCall(false, a_Data, a_Idx, a_Len);
#else
            UpdateCore(a_Data, a_Idx, a_Len);
#endif
        }

        protected override void TransformFinalBlock(byte[] a_Block)
        {
#if FAST
            // We'll fix the arrays and then call DefaultTransformFinalBlock().
            FixAndCall(true, a_Block, 0, 0);
#else
            base.DefaultTransformFinalBlock(a_Block, PaddingKinds.MD4);
#endif
        }

#if FAST

        private IntPtr f_Registers_FAST;
        private IntPtr f_Block_FAST;

        private unsafe void FixAndCall(bool a_FinalBlock,
                                       byte[] a_Buffer,
                                       int a_Idx,
                                       int a_Len)
        {
            fixed (uint* l_RegistersPtr = f_Registers)
            {
                f_Registers_FAST = (IntPtr)l_RegistersPtr;

                if (false == a_FinalBlock)
                {
                    base.UpdateCore(a_Buffer, a_Idx, a_Len);
                }
                else
                {
                    base.DefaultTransformFinalBlock(a_Buffer, PaddingKinds.MD4);
                }
            }
        }

#endif

        protected
#if FAST
        unsafe
#endif
        override void TransformBlocks(byte[] a_Block,
                                      int a_ByteIndex,
                                      int a_BlockCount)
        {
#if FAST
            fixed (byte* l_BlockArray = a_Block)
            {
                f_Block_FAST = (IntPtr)(&l_BlockArray[a_ByteIndex]);

                if (null == s_ExternalTransform)
                {
                    TransformFast(a_BlockCount);
                }
                else
                {
                    s_ExternalTransform(f_Registers_FAST, f_Block_FAST, (uint)a_BlockCount);
                }
            }
#else
            Transform(a_Block, a_ByteIndex, a_BlockCount);
#endif
        }

#if FAST

        private static readonly NativeTransformRoutine s_ExternalTransform;

        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void MD4MultiTransform_x86([In] IntPtr a_Registers,
                                                         [In] IntPtr a_Block,
                                                         [In] uint a_BlockCount);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void MD4MultiTransform_x64([In] IntPtr a_Registers,
                                                         [In] IntPtr a_Block,
                                                         [In] uint a_BlockCount);

        private static unsafe NativeTransformRoutine GetExternalTransformRoutine()
        {
            NativeTransformRoutine l_Target = null;

            if (true == Util.s_IsWindows)
            {
                if (sizeof(uint) == IntPtr.Size)
                {
                    l_Target = MD4MultiTransform_x86;
                }
                else if (sizeof(ulong) == IntPtr.Size)
                {
                    l_Target = MD4MultiTransform_x64;
                }

                if (null != l_Target)
                {
                    byte* l_DummyRegisters = stackalloc byte[c_HashByteSize];
                    byte* l_DummyBlock = stackalloc byte[c_BlockSize];

                    try
                    {
                        l_Target((IntPtr)l_DummyRegisters, (IntPtr)l_DummyBlock, 1);
                    }
                    catch
                    {
                        l_Target = null;
                    }
                }
            }

            return l_Target;
        }

        private unsafe void TransformFast(int a_BlockCount)
        {
            // See RFC 1320 3.4.

            uint* f_Registers = (uint*)f_Registers_FAST;
            uint A = f_Registers[0];
            uint B = f_Registers[1];
            uint C = f_Registers[2];
            uint D = f_Registers[3];

            for (int i = 0; i < a_BlockCount; i++)
            {
                uint l_StoredA = A;
                uint l_StoredB = B;
                uint l_StoredC = C;
                uint l_StoredD = D;

                Round1Op(ref A, B, C, D, 0, 3);
                Round1Op(ref D, A, B, C, 1, 7);
                Round1Op(ref C, D, A, B, 2, 11);
                Round1Op(ref B, C, D, A, 3, 19);
                Round1Op(ref A, B, C, D, 4, 3);
                Round1Op(ref D, A, B, C, 5, 7);
                Round1Op(ref C, D, A, B, 6, 11);
                Round1Op(ref B, C, D, A, 7, 19);
                Round1Op(ref A, B, C, D, 8, 3);
                Round1Op(ref D, A, B, C, 9, 7);
                Round1Op(ref C, D, A, B, 10, 11);
                Round1Op(ref B, C, D, A, 11, 19);
                Round1Op(ref A, B, C, D, 12, 3);
                Round1Op(ref D, A, B, C, 13, 7);
                Round1Op(ref C, D, A, B, 14, 11);
                Round1Op(ref B, C, D, A, 15, 19);

                Round2Op(ref A, B, C, D, 0, 3);
                Round2Op(ref D, A, B, C, 4, 5);
                Round2Op(ref C, D, A, B, 8, 9);
                Round2Op(ref B, C, D, A, 12, 13);
                Round2Op(ref A, B, C, D, 1, 3);
                Round2Op(ref D, A, B, C, 5, 5);
                Round2Op(ref C, D, A, B, 9, 9);
                Round2Op(ref B, C, D, A, 13, 13);
                Round2Op(ref A, B, C, D, 2, 3);
                Round2Op(ref D, A, B, C, 6, 5);
                Round2Op(ref C, D, A, B, 10, 9);
                Round2Op(ref B, C, D, A, 14, 13);
                Round2Op(ref A, B, C, D, 3, 3);
                Round2Op(ref D, A, B, C, 7, 5);
                Round2Op(ref C, D, A, B, 11, 9);
                Round2Op(ref B, C, D, A, 15, 13);

                Round3Op(ref A, B, C, D, 0, 3);
                Round3Op(ref D, A, B, C, 8, 9);
                Round3Op(ref C, D, A, B, 4, 11);
                Round3Op(ref B, C, D, A, 12, 15);
                Round3Op(ref A, B, C, D, 2, 3);
                Round3Op(ref D, A, B, C, 10, 9);
                Round3Op(ref C, D, A, B, 6, 11);
                Round3Op(ref B, C, D, A, 14, 15);
                Round3Op(ref A, B, C, D, 1, 3);
                Round3Op(ref D, A, B, C, 9, 9);
                Round3Op(ref C, D, A, B, 5, 11);
                Round3Op(ref B, C, D, A, 13, 15);
                Round3Op(ref A, B, C, D, 3, 3);
                Round3Op(ref D, A, B, C, 11, 9);
                Round3Op(ref C, D, A, B, 7, 11);
                Round3Op(ref B, C, D, A, 15, 15);

                A += l_StoredA;
                B += l_StoredB;
                C += l_StoredC;
                D += l_StoredD;

                f_Block_FAST = (IntPtr)(((byte*)f_Block_FAST) + c_BlockSize);
            }

            f_Registers[0] = A;
            f_Registers[1] = B;
            f_Registers[2] = C;
            f_Registers[3] = D;
        }

        private unsafe void Round1Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + F(b,c,d) + X[k]) <<< s;
            // F(X,Y,Z) = XY v not(X) Z.

            a += (((b & c) | ((~b) & d)) + ((uint*)f_Block_FAST)[k]);
            a = ((a << s) | (a >> (32 - s)));
        }

        private unsafe void Round2Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + G(b,c,d) + X[k] + 5A827999) <<< s;
            // G(X,Y,Z) = XY v XZ v YZ.

            a += (((b & c) | (b & d) | (c & d)) + ((uint*)f_Block_FAST)[k] + 0x5A827999);
            a = ((a << s) | (a >> (32 - s)));
        }

        private unsafe void Round3Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + H(b,c,d) + X[k] + 6ED9EBA1) <<< s;
            // H(X,Y,Z) = X xor Y xor Z.

            a += ((b ^ c ^ d) + ((uint*)f_Block_FAST)[k] + 0x6ED9EBA1);
            a = ((a << s) | (a >> (32 - s)));
        }

#else

        private void Transform(byte[] a_Block,
                               int a_ByteIndex,
                               int a_BlockCount)
        {
            // See RFC 1320 3.4.

            uint A = f_Registers[0];
            uint B = f_Registers[1];
            uint C = f_Registers[2];
            uint D = f_Registers[3];

            for (int i = 0, l_SrcIdx = a_ByteIndex; i < a_BlockCount; i++, l_SrcIdx += c_BlockSize)
            {
                Buffer.BlockCopy(a_Block, l_SrcIdx, f_Block, 0, c_BlockSize);

                uint l_StoredA = A;
                uint l_StoredB = B;
                uint l_StoredC = C;
                uint l_StoredD = D;

                Round1Op(ref A, B, C, D, 0, 3);
                Round1Op(ref D, A, B, C, 1, 7);
                Round1Op(ref C, D, A, B, 2, 11);
                Round1Op(ref B, C, D, A, 3, 19);
                Round1Op(ref A, B, C, D, 4, 3);
                Round1Op(ref D, A, B, C, 5, 7);
                Round1Op(ref C, D, A, B, 6, 11);
                Round1Op(ref B, C, D, A, 7, 19);
                Round1Op(ref A, B, C, D, 8, 3);
                Round1Op(ref D, A, B, C, 9, 7);
                Round1Op(ref C, D, A, B, 10, 11);
                Round1Op(ref B, C, D, A, 11, 19);
                Round1Op(ref A, B, C, D, 12, 3);
                Round1Op(ref D, A, B, C, 13, 7);
                Round1Op(ref C, D, A, B, 14, 11);
                Round1Op(ref B, C, D, A, 15, 19);

                Round2Op(ref A, B, C, D, 0, 3);
                Round2Op(ref D, A, B, C, 4, 5);
                Round2Op(ref C, D, A, B, 8, 9);
                Round2Op(ref B, C, D, A, 12, 13);
                Round2Op(ref A, B, C, D, 1, 3);
                Round2Op(ref D, A, B, C, 5, 5);
                Round2Op(ref C, D, A, B, 9, 9);
                Round2Op(ref B, C, D, A, 13, 13);
                Round2Op(ref A, B, C, D, 2, 3);
                Round2Op(ref D, A, B, C, 6, 5);
                Round2Op(ref C, D, A, B, 10, 9);
                Round2Op(ref B, C, D, A, 14, 13);
                Round2Op(ref A, B, C, D, 3, 3);
                Round2Op(ref D, A, B, C, 7, 5);
                Round2Op(ref C, D, A, B, 11, 9);
                Round2Op(ref B, C, D, A, 15, 13);

                Round3Op(ref A, B, C, D, 0, 3);
                Round3Op(ref D, A, B, C, 8, 9);
                Round3Op(ref C, D, A, B, 4, 11);
                Round3Op(ref B, C, D, A, 12, 15);
                Round3Op(ref A, B, C, D, 2, 3);
                Round3Op(ref D, A, B, C, 10, 9);
                Round3Op(ref C, D, A, B, 6, 11);
                Round3Op(ref B, C, D, A, 14, 15);
                Round3Op(ref A, B, C, D, 1, 3);
                Round3Op(ref D, A, B, C, 9, 9);
                Round3Op(ref C, D, A, B, 5, 11);
                Round3Op(ref B, C, D, A, 13, 15);
                Round3Op(ref A, B, C, D, 3, 3);
                Round3Op(ref D, A, B, C, 11, 9);
                Round3Op(ref C, D, A, B, 7, 11);
                Round3Op(ref B, C, D, A, 15, 15);

                A += l_StoredA;
                B += l_StoredB;
                C += l_StoredC;
                D += l_StoredD;
            }

            f_Registers[0] = A;
            f_Registers[1] = B;
            f_Registers[2] = C;
            f_Registers[3] = D;
        }

        private void Round1Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + F(b,c,d) + X[k]) <<< s;
            // F(X,Y,Z) = XY v not(X) Z.

            a += (((b & c) | ((~b) & d)) + f_Block[k]);
            a = Util.RotateLeft(a, s);
        }

        private void Round2Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + G(b,c,d) + X[k] + 5A827999) <<< s;
            // G(X,Y,Z) = XY v XZ v YZ.

            a += (((b & c) | (b & d) | (c & d)) + f_Block[k] + 0x5A827999);
            a = Util.RotateLeft(a, s);
        }

        private void Round3Op(ref uint a, uint b, uint c, uint d, uint k, int s)
        {
            // a = (a + H(b,c,d) + X[k] + 6ED9EBA1) <<< s;
            // H(X,Y,Z) = X xor Y xor Z.

            a += ((b ^ c ^ d) + f_Block[k] + 0x6ED9EBA1);
            a = Util.RotateLeft(a, s);
        }

#endif

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            MD_4 l_MD4 = new MD_4();

            l_MD4.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_MD4.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            Finish();

            Buffer.BlockCopy(f_Registers, 0, a_Dest, a_Idx, c_HashByteSize);
        }

        public uint GetDigestHighUInt()
        {
            Finish();

            return f_Registers[0];
        }

        public void GetDigestLE(byte[] a_Dest,
                                int a_Idx)
        {
            Finish();

            GetDigest(a_Dest, a_Idx);

            // BE to LE.
            UInt128.ReverseArrayBE(a_Dest, a_Idx);
        }

        public void GetDigestLE(uint[] a_Dest,
                                int a_Idx)
        {
            Finish();

            a_Dest[a_Idx + 3] = Util.ByteSwapUInt32(f_Registers[0]);
            a_Dest[a_Idx + 2] = Util.ByteSwapUInt32(f_Registers[1]);
            a_Dest[a_Idx + 1] = Util.ByteSwapUInt32(f_Registers[2]);
            a_Dest[a_Idx] = Util.ByteSwapUInt32(f_Registers[3]);
        }

        public UInt128 GetDigest()
        {
            Finish();

            return AsUInt128();
        }

        private UInt128 AsUInt128()
        {
            return new UInt128(Util.ByteSwapUInt32(f_Registers[3]),
                               Util.ByteSwapUInt32(f_Registers[2]),
                               Util.ByteSwapUInt32(f_Registers[1]),
                               Util.ByteSwapUInt32(f_Registers[0]));
        }

        public bool Verify(UInt128 a_UInt128)
        {
            // NOTE: This is a LE vs. BE operation.

            return (a_UInt128.Get32BitChunk(0) == Util.ByteSwapUInt32(f_Registers[3]) &&
                    a_UInt128.Get32BitChunk(1) == Util.ByteSwapUInt32(f_Registers[2]) &&
                    a_UInt128.Get32BitChunk(2) == Util.ByteSwapUInt32(f_Registers[1]) &&
                    a_UInt128.Get32BitChunk(3) == Util.ByteSwapUInt32(f_Registers[0]));
        }

        public bool VerifyBE(byte[] a_Src,
                             int a_Idx)
        {
            return (Util.ReadUInt32LE(a_Src, a_Idx) == f_Registers[0] &&
                    Util.ReadUInt32LE(a_Src, a_Idx + 4) == f_Registers[1] &&
                    Util.ReadUInt32LE(a_Src, a_Idx + 8) == f_Registers[2] &&
                    Util.ReadUInt32LE(a_Src, a_Idx + 12) == f_Registers[3]);
        }

        public static bool SelfTest()
        {
            if ("31D6CFE0D16AE931B73C59D7E0C089C0" !=
                Util.BytesToString(MD_4.ComputeHash(new byte[0], 0, 0)))
            {
                return false;
            }

            MD_4 l_MD4 = new MD_4();

            l_MD4.Update(ASCIIEncoding.ASCII.GetBytes("xx1234567890123456789012345678901234567xx"), 2, 37);
            l_MD4.Update(ASCIIEncoding.ASCII.GetBytes("x8901234567890123"), 1, 16);
            l_MD4.Update(ASCIIEncoding.ASCII.GetBytes("456789012345678901234567890x"), 0, 27);

            UInt128 l_Hash = l_MD4.GetDigest();

            return ("E33B4DDC9C38F2199C3E7B164FCC0536" == l_Hash.ToHexString());
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct Adler32 : IHashFunction
    {
        // Don't use the struct default constructor.

        public const byte c_HashByteSize = sizeof(uint);

        private const uint c_InitialChecksum = 1;
        private const uint c_BASE = 65521;
        private const int c_NMAX = 5552;

        private uint f_Checksum;

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public ushort p_BlockByteSize
        {
            get { return 1; }
        }

#if !FAST
        static Adler32()
        {
            s_TempBuffer = new byte[sizeof(ulong)];
        }
#endif

        public static void Create(out Adler32 a_Dest)
        {
            a_Dest = new Adler32();
            a_Dest.Reset();
        }

        public static object CreateBoxed()
        {
            Adler32 l_Temp;
            Create(out l_Temp);

            return (object)l_Temp;
        }

        public static uint GetChecksumFromBox(object a_Obj)
        {
            return ((Adler32)a_Obj).GetDigest();
        }

        public void Reset()
        {
            f_Checksum = c_InitialChecksum;
        }

        public uint GetDigest()
        {
            return f_Checksum;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            Util.WriteUInt32LE(f_Checksum, a_Dest, a_Idx);
            Util.ReverseArray(a_Dest, a_Idx, sizeof(uint));
        }

        public static uint ComputeHash(ref ByteArray a_Data)
        {
            return ComputeHash(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public static uint ComputeHash(byte[] a_Data)
        {
            return ComputeHash(a_Data, 0, a_Data.Length);
        }

        public static uint ComputeHash(byte[] a_Data,
                                       int a_Idx,
                                       int a_Len)
        {
            Adler32 l_Adler32;
            Adler32.Create(out l_Adler32);

            l_Adler32.Update(a_Data, a_Idx, a_Len);

            return l_Adler32.GetDigest();
        }

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Data)
        {
            Update(a_Data, 0, a_Data.Length);
        }

#if FAST

        public unsafe void Update(uint a_Num)
        {
            uint* l_Ptr = stackalloc uint[1];
            l_Ptr[0] = a_Num;

            UpdateCore((byte*)l_Ptr, sizeof(uint));
        }

        public unsafe void Update(ulong a_Num)
        {
            ulong* l_Ptr = stackalloc ulong[1];
            l_Ptr[0] = a_Num;

            UpdateCore((byte*)l_Ptr, sizeof(ulong));
        }

        public unsafe void Update(byte[] a_Data,
                                  int a_Idx,
                                  int a_Len)
        {
            Util.VerifyRangeOrFail(a_Idx, a_Len, a_Data.Length);

            fixed (byte* l_Ptr = a_Data)
            {
                UpdateCore(&l_Ptr[a_Idx], a_Len);
            }
        }

        private unsafe void UpdateCore(byte* a_Data,
                                       int a_Len)
        {
            uint l_Sum2 = (f_Checksum >> 16);
            f_Checksum &= ushort.MaxValue;

            byte* l_Ptr = a_Data;
            for (int l_ToGo = a_Len; 0 < l_ToGo;)
            {
                int l_Current = Math.Min(c_NMAX, l_ToGo);
                byte* l_FinalPtr = (l_Ptr + l_Current);

                do
                {
                    f_Checksum += *l_Ptr++;
                    l_Sum2 += f_Checksum;
                }
                while (l_Ptr < l_FinalPtr);

                l_ToGo -= l_Current;

                f_Checksum %= c_BASE;
                l_Sum2 %= c_BASE;
            }

            f_Checksum |= (l_Sum2 << 16);
        }

#else

        private static readonly byte[] s_TempBuffer;

        public void Update(uint a_Num)
        {
            // For performace reasons we always hold a temp
            // byte buffer ready.

            Util.MonitorEnter(s_TempBuffer);
            {
                Util.WriteUInt32LE(a_Num, s_TempBuffer, 0);

                Update(s_TempBuffer, 0, sizeof(uint));
            }
            Util.MonitorLeave(s_TempBuffer);
        }

        public void Update(ulong a_Num)
        {
            // For performace reasons we always hold a temp
            // byte buffer ready.

            Util.MonitorEnter(s_TempBuffer);
            {
                Util.WriteUInt64LE(a_Num, s_TempBuffer, 0);

                Update(s_TempBuffer, 0, sizeof(ulong));
            }
            Util.MonitorLeave(s_TempBuffer);
        }

        public void Update(byte[] a_Data,
                           int a_Idx,
                           int a_Len)
        {
            uint l_Sum2 = (f_Checksum >> 16);
            f_Checksum &= ushort.MaxValue;

            for (int l_Idx = a_Idx, l_ToGo = a_Len; 0 < l_ToGo; )
            {
                int l_Current = Math.Min(c_NMAX, l_ToGo);
                int l_Final = (l_Idx + l_Current);

                do
                {
                    f_Checksum += a_Data[l_Idx++];
                    l_Sum2 += f_Checksum;
                }
                while (l_Idx < l_Final);

                l_ToGo -= l_Current;

                f_Checksum %= c_BASE;
                l_Sum2 %= c_BASE;
            }

            f_Checksum |= (l_Sum2 << 16);
        }

#endif

        public void Finish()
        {
            // Nothing to do.
        }

        public static bool SelfTest()
        {
            Adler32 l_Adler32;
            Adler32.Create(out l_Adler32);

            l_Adler32.Update(ASCIIEncoding.ASCII.GetBytes("x123456x"), 1, 6);
            l_Adler32.Update(ASCIIEncoding.ASCII.GetBytes("7890123"));
            l_Adler32.Update(ASCIIEncoding.ASCII.GetBytes("45678901234567890"));

            uint l_Hash = l_Adler32.GetDigest();

            return (l_Hash.ToString("X8") == "5F0A0628");
        }
    }

    public struct CRC32 : IHashFunction
    {
        // Don't use the struct default constructor.

        public const byte c_HashByteSize = sizeof(uint);

        private const uint c_InitialChecksum = 0;

        private static readonly uint[] s_Table;

        private uint f_Checksum;

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public ushort p_BlockByteSize
        {
            get { return 1; }
        }

        static CRC32()
        {
            const ushort lc_TableElements = 256;
            const uint lc_Polynomial = 0xEDB88320;

            uint[] l_Table = new uint[lc_TableElements];

            for (uint i = 0; lc_TableElements != i; i++)
            {
                uint l_Val = i;

                for (uint k = Util.c_BitsPerByte; 0 != k; k--)
                {
                    if (0 == (l_Val & 1))
                    {
                        l_Val >>= 1;
                    }
                    else
                    {
                        l_Val = (lc_Polynomial ^ (l_Val >> 1));
                    }
                }

                l_Table[i] = l_Val;
            }

            s_Table = l_Table;
        }

        public CRC32(uint a_InitialState)
        {
            f_Checksum = ~a_InitialState;
        }

        public static void Create(out CRC32 a_Dest)
        {
            a_Dest = new CRC32(c_InitialChecksum);
        }

        public void Reset()
        {
            f_Checksum = ~c_InitialChecksum;
        }

        public uint GetChecksum()
        {
            return ~f_Checksum;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            // Output is in little endian notation.

            Util.WriteUInt32LE(GetChecksum(), a_Dest, a_Idx);
        }

        public static uint ComputeHash(ref ByteArray a_Data)
        {
            return ComputeHash(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public static uint ComputeHash(byte[] a_Buf)
        {
            return ComputeHash(a_Buf, 0, a_Buf.Length);
        }

        public static uint ComputeHash(byte[] a_Buf,
                                       int a_Idx,
                                       int a_Len)
        {
            CRC32 l_CRC32;
            CRC32.Create(out l_CRC32);
            l_CRC32.Update(a_Buf, a_Idx, a_Len);

            return l_CRC32.GetChecksum();
        }

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Buf)
        {
            Update(a_Buf, 0, a_Buf.Length);
        }

#if FAST
        public void Update(byte[] a_Buf,
                           int a_Idx,
                           int a_Len)
        {
            Util.VerifyRangeOrFail(a_Idx, a_Len, a_Buf.Length);

            if (0 == a_Len)
            {
                return;
            }

            unsafe
            {
                fixed (byte* l_Ptr = a_Buf)
                {
                    UpdateCore(&l_Ptr[a_Idx], (uint)a_Len);
                }
            }
        }

        private unsafe void UpdateCore(byte* a_Buf,
                                       uint a_Len)
        {
            uint l_Temp = f_Checksum;

            fixed (uint* l_TablePtr = s_Table)
            {
                byte* l_Ptr = a_Buf;
                byte* l_FinalPtr = &a_Buf[a_Len];

                for (; l_Ptr != l_FinalPtr; l_Ptr++)
                {
                    l_Temp = ((l_Temp >> Util.c_BitsPerByte) ^ l_TablePtr[(l_Temp & byte.MaxValue) ^ *l_Ptr]);
                }
            }

            f_Checksum = l_Temp;
        }
#else
        public void Update(byte[] a_Buf,
                           int a_Idx,
                           int a_Len)
        {
            uint l_Temp = f_Checksum;

            for (int i = a_Idx, lim = (a_Idx + a_Len); i != lim; i++)
            {
                l_Temp = ((l_Temp >> Util.c_BitsPerByte) ^ s_Table[(l_Temp & byte.MaxValue) ^ a_Buf[i]]);
            }

            f_Checksum = l_Temp;
        }
#endif

        public void Finish()
        {
            // Nothing to do.
        }

        internal static bool SelfTest()
        {
            CRC32 l_CRC32_2;
            CRC32.Create(out l_CRC32_2);

            l_CRC32_2.Update(ASCIIEncoding.ASCII.GetBytes("xThe qux"), 1, 6);
            l_CRC32_2.Update(ASCIIEncoding.ASCII.GetBytes("ick brown fox ju"));
            l_CRC32_2.Update(ASCIIEncoding.ASCII.GetBytes("mps over the lazy dog"));

            uint l_Hash = l_CRC32_2.GetChecksum();

            return (l_Hash == 0x414FA339);
        }
    }

    public abstract class HashAlgoWrapper<T> : IHashFunction
                                    where T : HashAlgorithm
    {
        private readonly ushort f_BlockSize;
        private readonly T f_Algo;
        private bool f_Finished;

        public abstract ushort p_HashByteSize
        {
            get;
        }

        public ushort p_BlockByteSize
        {
            get { return f_BlockSize; }
        }

        protected HashAlgoWrapper(ushort a_BlockSize,
                                  T a_Algo)
        {
            f_BlockSize = a_BlockSize;
            f_Algo = a_Algo;
        }

        public void Reset()
        {
            f_Algo.Initialize();
            f_Finished = false;
        }

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Data)
        {
            Update(a_Data, 0, a_Data.Length);
        }

        public void Update(byte[] a_Data,
                           int a_Idx,
                           int a_Len)
        {
            for (int i = 0; i < a_Len; )
            {
                i += f_Algo.TransformBlock(a_Data, a_Idx + i, a_Len - i, a_Data, i);
            }
        }

        public void Finish()
        {
            f_Algo.TransformFinalBlock(Util.s_Empty, 0, 0);

            f_Finished = true;
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            if (false == f_Finished)
            {
                Finish();
            }

            Buffer.BlockCopy(f_Algo.Hash, 0, a_Dest, a_Idx, p_HashByteSize);
        }
    }

    public class SHA_1 : HashAlgoWrapper<SHA1>
    {
        // An easy to use wrapper around the
        // System.Security.Cryptography.SHA1 class.

        // NOTE: For Bruce Schneier, SHA-1 is merely a compression
        // algorithm.

        public const byte c_HashByteSize = 20;

        // Luckily there is a fast .NET 2.0 implementation available.
        public SHA_1() :
                base(64, new SHA1CryptoServiceProvider())
        {
        }

        public override ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            SHA_1 l_SHA1 = new SHA_1();

            l_SHA1.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_SHA1.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            SHA_1 l_S = new SHA_1();

            l_S.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_S.Update(Encoding.ASCII.GetBytes("x"));
            l_S.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog"));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_S.GetDigest(l_Digest, 0);

            return ("2FD4E1C67A2D28FCED849EE1BB76E7391B93EB12" ==
                    Util.BytesToString(l_Digest));
        }
    }

    public class SHA_256 : IHashFunction
    {
        public const byte c_HashByteSize = 32;

#if FAST
        private static readonly bool s_RelyOnSelf;
#endif

        // One of f_SHA256Framework and f_SHA256Fast is null.
        // We use type specific access instead of an interface
        // to avoid a virtual call.
        private readonly SHA_256Framework f_SHA256Framework;
#if FAST
        private readonly SHA_256Fast f_SHA256Fast;
#endif

        public ushort p_BlockByteSize
        {
            get { return 64; }
        }

        static SHA_256()
        {
#if FAST
            if (true == SHA256.Create().GetType().ToString().EndsWith(".SHA256Managed"))
            {
                // This framework doesn't feature a fast implementation
                // it seems.

                s_RelyOnSelf = true;
            }
#endif
        }

        public ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public SHA_256()
        {
#if FAST
            if (true == s_RelyOnSelf)
            {
                f_SHA256Fast = new SHA_256Fast();

                return;
            }
#endif
            f_SHA256Framework = new SHA_256Framework();
        }

#if FAST
        public static bool LoadBoxesForFastCode(Stream a_Strm)
        {
            return SHA_256Fast.LoadBoxes(a_Strm);
        }
#endif

        public void Reset()
        {
#if FAST
            if (null != f_SHA256Fast)
            {
                f_SHA256Fast.Reset();

                return;
            }
#endif
            f_SHA256Framework.Reset();
        }

        public void Update(byte[] a_Data)
        {
            Update(a_Data, 0, a_Data.Length);
        }

        public void Update(ref ByteArray a_Data)
        {
            Update(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public void Update(byte[] a_Data,
                           int a_Idx,
                           int a_Len)
        {
#if FAST
            if (null != f_SHA256Fast)
            {
                f_SHA256Fast.Update(a_Data, a_Idx, a_Len);

                return;
            }
#endif
            f_SHA256Framework.Update(a_Data, a_Idx, a_Len);
        }

        public void Finish()
        {
#if FAST
            if (null != f_SHA256Fast)
            {
                f_SHA256Fast.Finish();

                return;
            }
#endif
            f_SHA256Framework.Finish();
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
#if FAST
            if (null != f_SHA256Fast)
            {
                f_SHA256Fast.GetDigest(a_Dest, a_Idx);

                return;
            }
#endif
            f_SHA256Framework.GetDigest(a_Dest, a_Idx);
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            SHA_256 l_SHA256 = new SHA_256();

            l_SHA256.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_SHA256.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            if ("E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855" != HexEncoding.ToHex(SHA_256.ComputeHash(new byte[0], 0, 0)))
            {
                return false;
            }

            SHA_256 l_S = new SHA_256();

            l_S.Reset();
            l_S.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_S.Update(Encoding.ASCII.GetBytes("x"));
            l_S.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog"));

            byte[] l_Digest = new byte[c_HashByteSize];
            l_S.GetDigest(l_Digest, 0);

            return ("D7A8FBB307D7809469CA9ABCB0082E4F8D5651E46D3CDB762D02D0BF37C9E592" == HexEncoding.ToHex(l_Digest));
        }

        private class SHA_256Framework : HashAlgoWrapper<SHA256>
        {
            public SHA_256Framework() :
                               base(64, SHA256.Create())
            {
            }

            public override ushort p_HashByteSize
            {
                get { return c_HashByteSize; }
            }
        }

#if FAST
        private class SHA_256Fast : Transformer
        {
            private const byte c_UIntsPerHash = (c_HashByteSize / sizeof(uint));
            private const byte c_UIntsPerBlock = 16;
            private const byte c_BlockSize = (c_UIntsPerBlock * sizeof(uint));
            private const byte c_UIntsPerBox = 64;
            private const ushort c_BoxSize = (c_UIntsPerBox * sizeof(uint));

            private static readonly uint[] s_K;
            private static readonly NativeTransformRoutine s_ExternalTransform;

            private readonly uint[] f_States;

            private IntPtr f_K_FAST;
            private IntPtr f_States_FAST;

            static SHA_256Fast()
            {
                s_K = new uint[c_UIntsPerBox];
                s_ExternalTransform = GetExternalTransformRoutine();
            }

            public SHA_256Fast() :
                          base(c_BlockSize)
            {
                f_States = new uint[c_UIntsPerHash];

                ResetCore();
            }

            public static bool LoadBoxes(Stream a_Strm)
            {
                byte[] l_Buffer;

                if (true == Tiger.LoadBoxBuffer(a_Strm, c_BoxSize, 0xB64E7DBF, out l_Buffer))
                {
                    Buffer.BlockCopy(l_Buffer, 0, s_K, 0, c_BoxSize);

                    return true;
                }

                return false;
            }

            protected override void ResetCore()
            {
                f_States[0] = 0x6A09E667;
                f_States[1] = 0xBB67AE85;
                f_States[2] = 0x3C6EF372;
                f_States[3] = 0xA54FF53A;
                f_States[4] = 0x510E527F;
                f_States[5] = 0x9B05688C;
                f_States[6] = 0x1F83D9AB;
                f_States[7] = 0x5BE0CD19;
            }

            public void GetDigest(byte[] a_Dest,
                                  int a_Idx)
            {
                // Result is in big endian notation.

                Finish();

                for (int i = 0, k = a_Idx; c_UIntsPerHash > i; i++, k += sizeof(uint))
                {
                    Util.WriteUInt32BE(f_States[i], a_Dest, k);
                }
            }

            public override void Update(byte[] a_Data,
                                        int a_Idx,
                                        int a_Len)
            {
                // We'll fix the arrays and then call UpdateCore().
                FixAndCall(false, a_Data, a_Idx, a_Len);
            }

            protected override void TransformFinalBlock(byte[] a_Block)
            {
                // We'll fix the arrays and then call DefaultTransformFinalBlock().
                FixAndCall(true, a_Block, 0, 0);
            }

            [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void SHA256MultiProcess_x86([In] IntPtr a_States,
                                                              [In] IntPtr a_Block,
                                                              [In] uint a_BlockCount);

            [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void SHA256MultiProcess_x64([In] IntPtr a_States,
                                                              [In] IntPtr a_Block,
                                                              [In] uint a_BlockCount);

            private static unsafe NativeTransformRoutine GetExternalTransformRoutine()
            {
                NativeTransformRoutine l_Target = null;

                if (true == Util.s_IsWindows)
                {
                    if (sizeof(uint) == IntPtr.Size)
                    {
                        l_Target = SHA256MultiProcess_x86;
                    }
                    else if (sizeof(ulong) == IntPtr.Size)
                    {
                        l_Target = SHA256MultiProcess_x64;
                    }

                    if (null != l_Target)
                    {
                        byte* l_DummyStates = stackalloc byte[c_HashByteSize];
                        byte* l_DummyBlock = stackalloc byte[c_BlockSize];

                        try
                        {
                            l_Target((IntPtr)l_DummyStates, (IntPtr)l_DummyBlock, 1);
                        }
                        catch
                        {
                            l_Target = null;
                        }
                    }
                }

                return l_Target;
            }

            private unsafe void FixAndCall(bool a_FinalBlock,
                                           byte[] a_Buffer,
                                           int a_Idx,
                                           int a_Len)
            {
                fixed (uint* l_K = s_K)
                {
                    fixed (uint* l_States = f_States)
                    {
                        f_K_FAST = (IntPtr)l_K;
                        f_States_FAST = (IntPtr)l_States;

                        if (false == a_FinalBlock)
                        {
                            base.UpdateCore(a_Buffer, a_Idx, a_Len);
                        }
                        else
                        {
                            base.DefaultTransformFinalBlock(a_Buffer, PaddingKinds.SHA256);
                        }
                    }
                }
            }

            protected unsafe override void TransformBlocks(byte[] a_Block,
                                                           int a_ByteIndex,
                                                           int a_BlockCount)
            {
                // Check that FixAndCall() is executing.

                fixed (byte* l_BlockArray = a_Block)
                {
                    byte* l_BlockPtr = &l_BlockArray[a_ByteIndex];

                    if (null == s_ExternalTransform)
                    {
                        ProcessBlocksFast(l_BlockPtr, a_BlockCount);
                    }
                    else
                    {
                        s_ExternalTransform(f_States_FAST, (IntPtr)l_BlockPtr, (uint)a_BlockCount);
                    }
                }
            }

            private unsafe void ProcessBlocksFast(byte* a_BlockSrc,
                                                  int a_BlockCount)
            {
                uint* l_KBase = (uint*)f_K_FAST;
                uint* l_States = (uint*)f_States_FAST;
                uint* l_Src = (uint*)a_BlockSrc;

                uint* l_WBase = stackalloc uint[c_UIntsPerBox];
                uint* l_WLimit16 = &l_WBase[c_UIntsPerBlock];
                uint* l_WLimit64 = &l_WBase[c_UIntsPerBox];
                uint* l_k, l_w;

                for (int l_BlockIdx = a_BlockCount; 0 < l_BlockIdx; l_BlockIdx--)
                {
                    for (l_w = l_WBase; l_WLimit16 != l_w; l_Src++, l_w++)
                    {
                        uint l_Num = *l_Src;
                        *l_w = (((l_Num & byte.MaxValue) << 24) |
                                ((l_Num & 0xFF00u) << Util.c_BitsPerByte) |
                                ((l_Num & 0xFF0000u) >> Util.c_BitsPerByte) |
                                ((l_Num >> 24) & byte.MaxValue));
                    }

                    for (; l_WLimit64 != l_w; l_w++)
                    {
                        uint l_wm15 = l_w[-15];
                        uint l_wm2 = l_w[-2];

                        *l_w = (l_w[-7] + l_w[-16]);
                        *l_w += (((l_wm15 >> 7) | (l_wm15 << 25)) ^
                                 ((l_wm15 >> 18) | (l_wm15 << 14)) ^
                                 (l_wm15 >> 3));
                        *l_w += (((l_wm2 >> 17) | (l_wm2 << 15)) ^
                                 ((l_wm2 >> 19) | (l_wm2 << 13)) ^
                                 (l_wm2 >> 10));
                    }

                    uint l_a = l_States[0];
                    uint l_b = l_States[1];
                    uint l_c = l_States[2];
                    uint l_d = l_States[3];
                    uint l_e = l_States[4];
                    uint l_f = l_States[5];
                    uint l_g = l_States[6];
                    uint l_h = l_States[7];

                    for (l_k = l_KBase, l_w = l_WBase; l_WLimit64 != l_w; l_k++, l_w++)
                    {
                        uint l_Sum0 = (((l_a >> 2) | (l_a << 30)) ^
                                       ((l_a >> 13) | (l_a << 19)) ^
                                       ((l_a >> 22) | (l_a << 10)));
                        uint l_Sum1 = (((l_e >> 6) | (l_e << 26)) ^
                                       ((l_e >> 11) | (l_e << 21)) ^
                                       ((l_e >> 25) | (l_e << 7)));
                        uint l_Ch = ((l_e & l_f) ^ ((~l_e) & l_g));
                        uint l_Maj = ((l_a & l_b) ^ (l_a & l_c) ^ (l_b & l_c));

                        uint T1 = (l_h + l_Sum1 + l_Ch + *l_k + *l_w);
                        uint T2 = (l_Sum0 + l_Maj);

                        l_h = l_g;
                        l_g = l_f;
                        l_f = l_e;
                        l_e = (l_d + T1);
                        l_d = l_c;
                        l_c = l_b;
                        l_b = l_a;
                        l_a = (T1 + T2);
                    }

                    l_States[0] += l_a;
                    l_States[1] += l_b;
                    l_States[2] += l_c;
                    l_States[3] += l_d;
                    l_States[4] += l_e;
                    l_States[5] += l_f;
                    l_States[6] += l_g;
                    l_States[7] += l_h;
                }
            }
        }
#endif
// End of #if FAST block.
    }

    public class SHA_384 : HashAlgoWrapper<SHA384>
    {
        // An easy to use wrapper around the
        // System.Security.Cryptography.SHA384 class.

        public const byte c_HashByteSize = 48;

        public SHA_384() :
                  base(128, SHA384.Create())
        {
        }

        public override ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            SHA_384 l_SHA384 = new SHA_384();

            l_SHA384.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_SHA384.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            SHA_384 l_S = new SHA_384();

            l_S.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_S.Update(Encoding.ASCII.GetBytes("x"));
            l_S.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog"));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_S.GetDigest(l_Digest, 0);

            return ("CA737F1014A48F4C0B6DD43CB177B0AFD9E5169367544C494011E3317DBF9A509CB1E5DC1E85A941BBEE3D7F2AFBC9B1" ==
                    Util.BytesToString(l_Digest));
        }
    }

    public class SHA_512 : HashAlgoWrapper<SHA512>
    {
        // An easy to use wrapper around the
        // System.Security.Cryptography.SHA512 class.

        public const byte c_HashByteSize = 64;

        public SHA_512() :
                  base(128, SHA512.Create())
        {
        }

        public override ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            SHA_512 l_SHA512 = new SHA_512();

            l_SHA512.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_SHA512.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            SHA_512 l_S = new SHA_512();

            l_S.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_S.Update(Encoding.ASCII.GetBytes("x"));
            l_S.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog"));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_S.GetDigest(l_Digest, 0);

            return ("07E547D9586F6A73F73FBAC0435ED76951218FB7D0C8D788A309D785436BBB642E93A252A954F23912547D1E8A3B5ED6E1BFD7097821233FA0538F3DB854FEE6" ==
                    Util.BytesToString(l_Digest));
        }
    }

    public class Keccak : Transformer, IHashFunction
    {
        private const byte c_StateSizeInULongs = 25;
        private const byte c_Rounds = 24;

        private static readonly ushort[] s_RhoPiOffsets;
        private static readonly ulong[] s_RoundConstants;

        public readonly Variant f_Variant;
        public readonly byte f_DigestByteSize;
        private readonly ulong[] f_States;

        public ushort p_HashByteSize
        {
            get { return f_DigestByteSize; }
        }

        public bool p_IsSHA3
        {
            get { return (Variant.SHA3_224Bits <= f_Variant); }
        }

        static Keccak()
        {
            // Initialize the rho offsets (low byte) and the pi
            // destination indices (high byte).

            s_RhoPiOffsets = new ushort[c_StateSizeInULongs]
            {
                0, 2561, 5182, 1308, 3867, 4132, 300, 2822, 5431,
                1556, 1795, 4362, 555, 3097, 5671, 5929, 2093, 4623,
                789, 3336, 3602, 6146, 2365, 4920, 1038
            };

            // Calculate the round constants.

            ulong[] l_RoundArr = new ulong[c_Rounds];
            byte l_LFSRState = 1;

            for (uint i = 0; c_Rounds != i; i++)
            {
                for (uint j = 0; (Util.c_BitsPerByte - 1) != j; j++)
                {
                    bool l_WasOdd = (0 != (l_LFSRState & 0x1));

                    if (0 == (l_LFSRState & 0x80))
                    {
                        l_LFSRState <<= 1;
                    }
                    else
                    {
                        l_LFSRState = (byte)((l_LFSRState << 1) ^ 0x71);
                    }

                    if (true == l_WasOdd)
                    {
                        uint l_BitPos = ((1u << (int)j) - 1);
                        l_RoundArr[i] ^= (1UL << (int)l_BitPos);
                    }
                }
            }

            s_RoundConstants = l_RoundArr;
        }

        public Keccak(Variant a_Variant) :
                 this(a_Variant, GetDigestByteSize(a_Variant))
        {
        }

        private Keccak(Variant a_Variant,
                       uint a_DigestByteSize) :
                  base(CalcBlockByteSize(a_DigestByteSize))
        {
            f_Variant = a_Variant;
            f_DigestByteSize = (byte)a_DigestByteSize;
            f_States = new ulong[c_StateSizeInULongs];
        }

        public static byte[] ComputeHash(Variant a_Variant,
                                         byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            Keccak l_Keccak = new Keccak(a_Variant);

            l_Keccak.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[l_Keccak.p_HashByteSize];

            l_Keccak.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        protected override void ResetCore()
        {
            Array.Clear(f_States, 0, c_StateSizeInULongs);
        }

        protected override void TransformBlocks(byte[] a_Block,
                                                int a_ByteIndex,
                                                int a_BlockCount)
        {
            uint l_ULongsPerBlock = ((uint)f_BlockSize / sizeof(ulong));
            ulong[] l_States = f_States;
            ulong[] l_TempStates = new ulong[c_StateSizeInULongs];
            ulong[] l_TempC = new ulong[5];

            for (uint i = 0, l_SrcIdx = (uint)a_ByteIndex; i != a_BlockCount; i++)
            {
                for (uint k = 0; k != l_ULongsPerBlock; k++)
                {
                    l_States[k] ^= Util.ReadUInt64LE(a_Block, (int)l_SrcIdx);
                    l_SrcIdx += sizeof(ulong);
                }

                for (uint r = 0; c_Rounds != r; r++)
                {
                    Theta(l_States, l_TempC);
                    RhoPi(l_States, l_TempStates);
                    Chi(l_TempStates, l_States);
                    Iota(l_States, r);
                }
            }
        }

        private static void Theta(ulong[] a_States,
                                  ulong[] a_TempC)
        {
            for (uint x = 0; 5 != x; x++)
            {
                a_TempC[x] = (a_States[x] ^ a_States[5 + x] ^ a_States[10 + x] ^ a_States[15 + x] ^ a_States[20 + x]);
            }

            for (uint x = 0; 5 != x; x++)
            {
                ulong l_D = (RotateLeft64(a_TempC[Mod5(x + 1)], 1) ^ a_TempC[Mod5(x + 4)]);

                for (uint i = 0; c_StateSizeInULongs != i; i += 5)
                {
                    a_States[i + x] ^= l_D;
                }
            }
        }

        private static void RhoPi(ulong[] a_SrcStates,
                                  ulong[] a_DestStates)
        {
            for (uint y = 0, i = 0; 5 != y; y++)
            {
                for (uint x = 0; 5 != x; x++, i++)
                {
                    // The low byte of l_Offset is the number of bits
                    // for the shift operation, the second byte is the
                    // destination index.
                    uint l_Offset = s_RhoPiOffsets[i];
                    uint l_Shift = (byte)l_Offset;
                    uint l_DestIdx = (byte)(l_Offset >> Util.c_BitsPerByte);

                    a_DestStates[l_DestIdx] = RotateLeft64(a_SrcStates[i], (int)l_Shift);
                }
            }
        }

        private static void Chi(ulong[] a_SrcStates,
                                ulong[] a_DestStates)
        {
            for (uint i = 0; c_StateSizeInULongs != i; i += 5)
            {
                for (uint x = 0; 5 != x; x++)
                {
                    a_DestStates[i + x] = (a_SrcStates[i + x] ^ (~a_SrcStates[i + Mod5(x + 1)] & a_SrcStates[i + Mod5(x + 2)]));
                }
            }
        }

        private static void Iota(ulong[] a_States,
                                 uint a_RoundIndex)
        {
            a_States[0] ^= s_RoundConstants[a_RoundIndex];
        }

        protected override void TransformFinalBlock(byte[] a_Block)
        {
            DefaultTransformFinalKeccakBlock(a_Block, false == p_IsSHA3 ? (byte)0x1 : (byte)0x6);
        }

        public void GetDigest(byte[] a_Dest,
                              int a_Idx)
        {
            Finish();

            Buffer.BlockCopy(f_States, 0, a_Dest, a_Idx, f_DigestByteSize);
        }

        private static ulong RotateLeft64(ulong x,
                                          int n)
        {
            return ((x << n) | (x >> (64 - n)));
        }

        private static uint Mod5(uint a_x)
        {
            return (a_x % 5);
        }

        private static ushort CalcBlockByteSize(uint a_DigestByteSize)
        {
            // Returns 1152 / 8 = 144, 1088 / 8 = 136, 832 / 8 = 104 or
            // 576 / 8 = 72.

            return (ushort)((c_StateSizeInULongs * sizeof(ulong)) - (2 * a_DigestByteSize));
        }

        public static uint GetDigestByteSize(Variant a_Variant)
        {
            switch (a_Variant)
            {
                case Variant.Keccak224Bits:
                case Variant.SHA3_224Bits:
                    return 28;

                case Variant.Keccak256Bits:
                case Variant.SHA3_256Bits:
                    return 32;

                case Variant.Keccak384Bits:
                case Variant.SHA3_384Bits:
                    return 48;

                case Variant.Keccak512Bits:
                case Variant.SHA3_512Bits:
                    return 64;
            }

            throw (new UnexpectedStateException());
        }

        public static bool SelfTest()
        {
            // Test all eight variants.

            string[] l_PlainTexts = new string[3]
            {
                string.Empty,
                "The quick brown fox jumps over the lazy dog",
                "The quick brown fox jumps over the lazy dog."
            };

            string[] l_Hashes = new string[3 * 8]
            {
                "F71837502BA8E10837BDD8D365ADB85591895602FC552B48B7390ABD",
                "C5D2460186F7233C927E7DB2DCC703C0E500B653CA82273B7BFAD8045D85A470",
                "2C23146A63A29ACF99E73B88F8C24EAA7DC60AA771780CCC006AFBFA8FE2479B2DD2B21362337441AC12B515911957FF",
                "0EAB42DE4C3CEB9235FC91ACFFE746B29C29A8C366B7C60E4E67C466F36A4304C00FA9CAF9D87976BA469BCBE06713B435F091EF2769FB160CDAB33D3670680E",
                "6B4E03423667DBB73B6E15454F0EB1ABD4597F9A1B078E3F5B5A6BC7",
                "A7FFC6F8BF1ED76651C14756A061D662F580FF4DE43B49FA82D80A4B80F8434A",
                "0C63A75B845E4F7D01107D852E4C2485C51A50AAAA94FC61995E71BBEE983A2AC3713831264ADB47FB6BD1E058D5F004",
                "A69F73CCA23A9AC5C8B567DC185A756E97C982164FE25859E0D1DCC1475C80A615B2123AF1F5F94C11E3E9402C3AC558F500199D95B6D3E301758586281DCD26",
                "310AEE6B30C47350576AC2873FA89FD190CDC488442F3EF654CF23FE",
                "4D741B6F1EB29CB2A9B9911C82F56FA8D73B04959D3D9D222895DF6C0B28AA15",
                "283990FA9D5FB731D786C5BBEE94EA4DB4910F18C62C03D173FC0A5E494422E8A0B3DA7574DAE7FA0BAF005E504063B3",
                "D135BB84D0439DBAC432247EE573A23EA7D3C9DEB2A968EB31D47C4FB45F1EF4422D6C531B5B9BD6F449EBCC449EA94D0A8F05F62130FDA612DA53C79659F609",
                "D15DADCEAA4D5D7BB3B48F446421D542E08AD8887305E28D58335795",
                "69070DDA01975C8C120C3AADA1B282394E7F032FA9CF32F4CB2259A0897DFC04",
                "7063465E08A93BCE31CD89D2E3CA8F602498696E253592ED26F07BF7E703CF328581E1471A7BA7AB119B1A9EBDF8BE41",
                "01DEDD5DE4EF14642445BA5F5B97C15E47B9AD931326E4B0727CD94CEFC44FFF23F07BF543139939B49128CAF436DC1BDEE54FCB24023A08D9403F9B4BF0D450",
                "C59D4EAEAC728671C635FF645014E2AFA935BEBFFDB5FBD207FFDEAB",
                "578951E24EFD62A3D63A86F7CD19AAA53C898FE287D2552133220370240B572D",
                "9AD8E17325408EDDB6EDEE6147F13856AD819BB7532668B605A24A2D958F88BD5C169E56DC4B2F89FFD325F6006D820B",
                "AB7192D2B11F51C7DD744E7B3441FEBF397CA07BF812CCEAE122CA4DED6387889064F8DB9230F173F6D1AB6E24B6E50F065B039F799F5592360A6558EB52D760",
                "2D0708903833AFABDD232A20201176E8B58C5BE8A6FE74265AC54DB0",
                "A80F839CD4F83F6C3DAFC87FEAE470045E4EB0D366397D5C6CE34BA1739F734D",
                "1A34D81695B622DF178BC74DF7124FE12FAC0F64BA5250B78B99C1273D4B080168E10652894ECAD5F1F4D5B965437FB9",
                "18F4F4BD419603F95538837003D9D254C26C23765565162247483F65C50303597BC9CE4D289F21D1C2F1F458828E33DC442100331B35E7EB031B5D38BA6460F8"
            };

            byte[] l_DigestBytes = new byte[1 + 64];

            for (uint i = 0, t = 0; 3 != i; i++)
            {
                string l_PlainText = l_PlainTexts[i];
                byte[] l_PlainBytes = Encoding.ASCII.GetBytes(l_PlainText);

                for (uint k = 1; (8 + 1) != k; k++, t++)
                {
                    Keccak l_Keccak = new Keccak((Variant)k);
                    l_Keccak.Update(l_PlainBytes);

                    Array.Clear(l_DigestBytes, 1, l_Keccak.p_HashByteSize);
                    l_Keccak.GetDigest(l_DigestBytes, 1);

                    string l_HashText = HexEncoding.ToHex(l_DigestBytes, 1, l_Keccak.p_HashByteSize);

                    if (false == Util.StringsEqual(l_HashText, l_Hashes[t]))
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        public enum Variant : byte
        {
            // Keccak and SHA3 differ only in the padding scheme.

            Undefined = 0,

            // 28 bytes.
            Keccak224Bits = 1,
            // 32 bytes.
            Keccak256Bits = 2,
            // 48 bytes.
            Keccak384Bits = 3,
            // 64 bytes.
            Keccak512Bits = 4,

            SHA3_224Bits = 5,
            // 32 bytes.
            SHA3_256Bits = 6,
            // 48 bytes.
            SHA3_384Bits = 7,
            // 64 bytes.
            SHA3_512Bits = 8,
        }
    }

    public class MD_5 : HashAlgoWrapper<MD5>
    {
        // An easy to use wrapper around the
        // System.Security.Cryptography.MD5 class.

        public const byte c_HashByteSize = 16;

        public MD_5() :
               base(64, new MD5CryptoServiceProvider())
        {
        }

        public override ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            MD_5 l_MD5 = new MD_5();

            l_MD5.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_MD5.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            MD_5 l_M = new MD_5();

            l_M.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_M.Update(Encoding.ASCII.GetBytes("x"));
            l_M.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog."));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_M.GetDigest(l_Digest, 0);

            return ("E4D909C290D0FB1CA068FFADDF22CBD0" ==
                    Util.BytesToString(l_Digest));
        }
    }

    public class RIPEMD_160 : HashAlgoWrapper<RIPEMD160>
    {
        // An easy to use wrapper around the
        // System.Security.Cryptography.RIPEMD160 class.

        public const byte c_HashByteSize = 20;

        public RIPEMD_160() :
                     base(64, RIPEMD160.Create())
        {
        }

        public override ushort p_HashByteSize
        {
            get { return c_HashByteSize; }
        }

        public static byte[] ComputeHash(byte[] a_Data,
                                         int a_Idx,
                                         int a_Len)
        {
            RIPEMD_160 l_RIPEMD160 = new RIPEMD_160();

            l_RIPEMD160.Update(a_Data, a_Idx, a_Len);

            byte[] l_Hash = new byte[c_HashByteSize];

            l_RIPEMD160.GetDigest(l_Hash, 0);

            return l_Hash;
        }

        public static bool SelfTest()
        {
            RIPEMD_160 l_R = new RIPEMD_160();

            l_R.Update(Encoding.ASCII.GetBytes("The quick brown fo"));
            l_R.Update(Encoding.ASCII.GetBytes("x"));
            l_R.Update(Encoding.ASCII.GetBytes(" jumps over the lazy dog"));

            byte[] l_Digest = new byte[c_HashByteSize];

            l_R.GetDigest(l_Digest, 0);

            return ("37F332F68DB77BD9D7EDD4969571AD671CF9DD3B" ==
                    Util.BytesToString(l_Digest));
        }
    }
}

namespace Support.Encodings
{
    static class ASN1
    {
        // This class is able to interpret and generate certain
        // ASN.1 blobs that are compatible with Crypto++.
        // This approach works with up to 792 bit keys (the ASN.1
        // pubkey will be 129 bytes long at most).

        public const byte c_SupposedFirstASN1Byte = 0x30;

        private static readonly byte[] s_iMuleASN1KeyHeader = new byte[]
        {
            // The 0xFF bytes will be filled with meaningful values
            // when the header is written.

            0x30, 0xFF, 0x30, 0x0D, 0x06, 0x09, 0x2A, 0x86, 0x48, 0x86,
            0xF7, 0x0D, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0xFF, 0x00,
            0x30, 0xFF, 0x02, 0xFF
        };

        public static bool ToPubKeyParams(ref ByteArray a_In,
                                          out byte[] a_ModBE,
                                          out byte[] a_ExpBE)
        {
            byte[] l_TightArr = a_In.ToArray(false);

            return ToPubKeyParams(l_TightArr, out a_ModBE, out a_ExpBE);
        }

        private static bool ToPubKeyParams(byte[] a_In,
                                           out byte[] a_ModBE,
                                           out byte[] a_ExpBE)
        {
            a_ModBE = null;
            a_ExpBE = null;

            int l_Len = a_In.Length;

            if (s_iMuleASN1KeyHeader.Length > l_Len ||
                0 != (a_In[1] & 0x80) ||
                (l_Len - 2) != a_In[1] ||
                (l_Len - 19) != a_In[18] ||
                (l_Len - 22) != a_In[21])
            {
                return false;
            }

            for (int i = 0; i < s_iMuleASN1KeyHeader.Length; i++)
            {
                byte l_By = s_iMuleASN1KeyHeader[i];

                if (a_In[i] != l_By &&
                    0xFF != l_By)
                {
                    return false;
                }
            }

            int l_ModLen = a_In[23];

            if ((s_iMuleASN1KeyHeader.Length + l_ModLen + 2) > l_Len ||
                0x02 != a_In[s_iMuleASN1KeyHeader.Length + l_ModLen])
            {
                return false;
            }

            int l_ExpLen = a_In[s_iMuleASN1KeyHeader.Length + l_ModLen + 1];

            if ((s_iMuleASN1KeyHeader.Length + l_ModLen + 2 + l_ExpLen) != l_Len)
            {
                return false;
            }

            a_ModBE = CopyBytes(a_In, s_iMuleASN1KeyHeader.Length, l_ModLen);
            a_ExpBE = CopyBytes(a_In, s_iMuleASN1KeyHeader.Length + l_ModLen + 2, l_ExpLen);

            return true;
        }

        private static byte[] CopyBytes(byte[] a_Src,
                                        int a_Idx,
                                        int a_Len)
        {
            // Crypto++ sometimes encodes at least one leading
            // zero that we would like to have removed.

            int l_ActIdx = a_Idx;
            int l_ActLen = a_Len;

            for (; 0 < l_ActLen; l_ActIdx++, l_ActLen--)
            {
                if (0x00 != a_Src[l_ActIdx])
                {
                    break;
                }
            }

            return Util.CopyBytes(a_Src, l_ActIdx, l_ActLen);
        }

        public static byte[] ToiMuleReadablePubKey(byte[] a_ModBE,
                                                   byte[] a_ExpBE)
        {
            // Generates a certain type of ASN.1 blob that can be read by
            // Crypto++. Returns null on error.

            // NOTE: There must be an extra zero before the mod value.

            int l_Len = (s_iMuleASN1KeyHeader.Length + 1 + a_ModBE.Length + 2 + a_ExpBE.Length);

            if (0x80 <= (l_Len - 2))
            {
                // We don't support length values that don't fit
                // into a byte.

                return null;
            }

            byte[] l_Key = new byte[l_Len];

            Buffer.BlockCopy(s_iMuleASN1KeyHeader, 0, l_Key, 0, s_iMuleASN1KeyHeader.Length);
            Buffer.BlockCopy(a_ModBE, 0, l_Key, 1 + s_iMuleASN1KeyHeader.Length, a_ModBE.Length);
            Buffer.BlockCopy(a_ExpBE, 0, l_Key, l_Len - a_ExpBE.Length, a_ExpBE.Length);

            // Copy sizes in.
            l_Key[1] = (byte)(l_Len - 2);
            l_Key[18] = (byte)(l_Len - 19);
            l_Key[21] = (byte)(l_Len - 22);
            l_Key[23] = (byte)(1 + a_ModBE.Length);
            // 0x02 is the integer tag.
            l_Key[l_Len - a_ExpBE.Length - 2] = (byte)0x02;
            l_Key[l_Len - a_ExpBE.Length - 1] = (byte)a_ExpBE.Length;

            return l_Key;
        }
    }

    public static class HexEncoding
    {
        private const byte c_HexCharsPerByte = Util.c_HexCharsPerByte;
        private const byte c_HexCharsPerUInt = Util.c_HexCharsPerUInt;

        public static string ToHex(byte[] a_Bytes)
        {
            return ToHex(a_Bytes, 0, a_Bytes.Length);
        }

        public static string ToHex(byte[] a_Bytes,
                                   int a_Idx,
                                   int a_Len)
        {
            StringBuilder l_Bld = new StringBuilder(a_Len * c_HexCharsPerByte);

            int l_Idx = a_Idx;
            int l_Lim = (l_Idx + a_Len);

            // Process whole uints.
            for (l_Lim -= sizeof(uint); l_Idx <= l_Lim; l_Idx += sizeof(uint))
            {
                uint l_Num = Util.ReadUInt32LE(a_Bytes, l_Idx);
                l_Bld.Append(Util.ByteSwapUInt32(l_Num).ToString("X8", CultureInfo.InvariantCulture));
            }

            // Process remaining bytes.
            for (l_Lim += sizeof(uint); l_Idx < l_Lim; l_Idx++)
            {
                l_Bld.Append(a_Bytes[l_Idx].ToString("X2", CultureInfo.InvariantCulture));
            }

            return l_Bld.ToString();
        }

        public static byte[] FromHex(string a_Src)
        {
            return FromHex(a_Src, 0, a_Src.Length);
        }

        public static byte[] FromHex(string a_Src,
                                     int a_Idx,
                                     int a_Len)
        {
            byte[] l_Dest = new byte[a_Len / c_HexCharsPerByte];

            FromHex(a_Src, a_Idx, a_Len, l_Dest, 0);

            return l_Dest;
        }

        public static void FromHex(string a_Src,
                                   int a_SrcIdx,
                                   int a_SrcLen,
                                   byte[] a_Dest,
                                   int a_DestIdx)
        {
            // The second loop will throw an exception on
            // odd input length.

            int l_SrcIdx = a_SrcIdx;
            int l_DestIdx = a_DestIdx;
            int l_Lim = (l_SrcIdx + a_SrcLen);

            // Process whole uints.
            for (l_Lim -= c_HexCharsPerUInt; l_SrcIdx <= l_Lim; l_SrcIdx += c_HexCharsPerUInt, l_DestIdx += sizeof(uint))
            {
                uint l_UInt = ParseUInt(a_Src, l_SrcIdx, c_HexCharsPerUInt);

                Util.WriteUInt32LE(Util.ByteSwapUInt32(l_UInt), a_Dest, l_DestIdx);
            }

            // Process remaining bytes.
            for (l_Lim += c_HexCharsPerUInt; l_SrcIdx < l_Lim; l_SrcIdx += c_HexCharsPerByte, l_DestIdx++)
            {
                a_Dest[l_DestIdx] = (byte)ParseUInt(a_Src, l_SrcIdx, c_HexCharsPerByte);
            }
        }

        public static uint ParseUInt(string a_Str,
                                     int a_Idx,
                                     int a_Chars)
        {
            uint l_Num;
            if (true == TryParseUInt(a_Str, a_Idx, a_Chars, out l_Num))
            {
                return l_Num;
            }

            throw (new FormatException());
        }

        public static bool TryParseUInt(string a_Str,
                                        int a_Idx,
                                        int a_Chars,
                                        out uint a_Num)
        {
            // This routine does not requite to create a
            // substring like uint.Parse(..., System.Globalization.NumberStyles.HexNumber).

            a_Num = 0;

            for (int i = ((a_Chars - 1) * Util.c_BitsPerNibble); 0 <= i; i -= Util.c_BitsPerNibble)
            {
                char l_Char = a_Str[a_Idx++];

                if ('0' <= l_Char &&
                    '9' >= l_Char)
                {
                    a_Num |= ((uint)(l_Char - '0') << i);
                }
                else if ('A' <= l_Char &&
                         'F' >= l_Char)
                {
                    a_Num |= (((uint)(l_Char - 'A') + 10) << i);
                }
                else
                {
                    return false;
                }
            }

            return true;
        }
    }

    public static class UrlEncoding
    {
        // We need to reference the System.Web assembly for this.
        // The following routines will never fail, illegal input
        // will just be copied through.

        public static string UrlEncode(string a_Text)
        {
            string l_EncodedText = HttpUtility.UrlEncode(a_Text);

            // Spaces got converted into '+', so change.
            l_EncodedText = l_EncodedText.Replace("+", "%20");

            return l_EncodedText;
        }

        public static string UrlEncode(byte[] a_Bytes)
        {
            return HttpUtility.UrlEncode(a_Bytes);
        }

        public static string UrlEncode(ref ByteArray a_Data)
        {
            return HttpUtility.UrlEncode(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);
        }

        public static string UrlDecode(string a_Text)
        {
            return HttpUtility.UrlDecode(a_Text);
        }

        public static string UrlDecode(string a_Src,
                                       int a_SrcIdx,
                                       int a_SrcLen)
        {
            return UrlDecode(a_Src.Substring(a_SrcIdx, a_SrcLen));
        }
    }

    public static class HtmlEncoding
    {
        // We need to reference the System.Web assembly for this.
        // The following routines will never fail, illegal input
        // will just be copied through.

        public static string HtmlEncode(string a_Text)
        {
            string l_EncodedStr = HttpUtility.HtmlEncode(a_Text);

            if (0 <= l_EncodedStr.IndexOf("  "))
            {
                // If there are several spaces in a row, convert
                // all but the first into protected spaces.

                StringBuilder l_Bld = new StringBuilder(l_EncodedStr);

                bool l_Run = false;
                for (int i = 0; i < l_Bld.Length; i++)
                {
                    if (' ' == l_Bld[i])
                    {
                        if (false == l_Run)
                        {
                            l_Run = true;
                        }
                        else
                        {
                            l_Bld.Replace(" ", "&nbsp;", i, 1);
                            i += 5;
                        }
                    }
                    else
                    {
                        l_Run = false;
                    }
                }

                l_EncodedStr = l_Bld.ToString();
            }

            return l_EncodedStr;
        }

        public static string HtmlDecode(string a_Text)
        {
            return HttpUtility.HtmlDecode(a_Text);
        }
    }

    public static class Base32
    {
        // The ToBase32() and FromBase32() routines are based on the
        // GPLed code of the imule 1.4.5 client (OtherFunctions.cpp).
        // Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
        // Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )

        private const string c_Base32Chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
        private const byte c_BitsPerBase32Char = 5; // 2^5 = 32
        private const byte c_BitsPerByte = Util.c_BitsPerByte;

        public static string ToBase32(byte[] a_Data)
        {
            return ToBase32(a_Data, 0, a_Data.Length);
        }

        public static string ToBase32(byte[] a_Data,
                                      int a_Idx,
                                      int a_Len)
        {
            // This formula gives us the rough size needed.
            StringBuilder l_Bld = new StringBuilder(((a_Len * c_BitsPerByte) / c_BitsPerBase32Char) + 1);

            for (int i = a_Idx, lim = (a_Idx + a_Len - 1), index = 0; i <= lim; )
            {
                byte l_Out;

                // Is the current word going to span a byte boundary?
                if (index > (c_BitsPerByte - c_BitsPerBase32Char))
                {
                    l_Out = (byte)(a_Data[i] & (0xFF >> index));
                    index = ((index + c_BitsPerBase32Char) % c_BitsPerByte);
                    l_Out = (byte)(l_Out << index);

                    if (i < lim)
                    {
                        l_Out = (byte)(l_Out | (a_Data[i + 1] >> (c_BitsPerByte - index)));
                    }

                    i++;
                }
                else
                {
                    l_Out = (byte)((a_Data[i] >> (c_BitsPerByte - (index + c_BitsPerBase32Char))) & 0x1F);
                    index = ((index + c_BitsPerBase32Char) % c_BitsPerByte);

                    if (0 == index)
                    {
                        i++;
                    }
                }

                l_Bld.Append(c_Base32Chars[l_Out]);
            }

            return l_Bld.ToString();
        }

        public static byte[] FromBase32(string a_Str)
        {
            return FromBase32(a_Str, 0, a_Str.Length);
        }

        public static byte[] FromBase32(string a_Str,
                                        int a_StartingIdx,
                                        int a_Chars)
        {
            // Return can be null if the format is invalid.

            // The expected output length.
            int l_DecodeLen = ((a_Chars * c_BitsPerBase32Char) / c_BitsPerByte);
            byte[] l_Out = new byte[l_DecodeLen];

            uint l_Bits = 0;
            int l_Count = 0;
            int l_OutIdx = 0;
            for (int i = a_StartingIdx, lim = (a_StartingIdx + a_Chars); i < lim; i++)
            {
                char l_Char = a_Str[i];

                if (l_Char >= 'A' &&
                    l_Char <= 'Z')
                {
                    l_Bits |= (uint)(l_Char - 'A');
                }
                else if (l_Char >= '2' &&
                         l_Char <= '7')
                {
                    l_Bits |= (uint)(l_Char - '2' + 26);
                }
                // We accept lowecase chars.
                else if (l_Char >= 'a' &&
                         l_Char <= 'z')
                {
                    l_Bits |= (uint)(l_Char - 'a');
                }
                else
                {
                    return null;
                }

                l_Count += c_BitsPerBase32Char;

                if (l_Count >= c_BitsPerByte)
                {
                    l_Out[l_OutIdx] = (byte)(l_Bits >> (l_Count - c_BitsPerByte));
                    l_OutIdx++;

                    l_Count -= c_BitsPerByte;
                }

                l_Bits <<= c_BitsPerBase32Char;
            }

            return l_Out;
        }
    }

    public static class Base64
    {
        // I2P uses the following base64 alphabet (and so do we).
        // [A-Z][a-z][0-9]-~

        public static string ToBase64(byte[] a_Bytes)
        {
            return ToBase64(a_Bytes, 0, a_Bytes.Length);
        }

        public static string ToBase64(byte[] a_Bytes,
                                      int a_Idx,
                                      int a_Len)
        {
            string l_Str = Convert.ToBase64String(a_Bytes, a_Idx, a_Len);

            StringBuilder l_Bld = new StringBuilder(l_Str);
            l_Bld.Replace('+', '-'); // 62
            l_Bld.Replace('/', '~'); // 63

            return l_Bld.ToString();
        }

        public static byte[] FromBase64(string a_Str,
                                        int a_Idx)
        {
            return FromBase64(a_Str, a_Idx, a_Str.Length - a_Idx);
        }

        public static byte[] FromBase64(string a_Str,
                                        int a_Idx,
                                        int a_Len)
        {
            // Return can be null if the format is invalid.

            StringBuilder l_Bld = new StringBuilder(a_Str, a_Idx, a_Len, a_Len);
            l_Bld.Replace('-', '+'); // 62
            l_Bld.Replace('~', '/'); // 63

            string l_EncodedStr = l_Bld.ToString();

            try
            {
                return Convert.FromBase64String(l_EncodedStr);
            }
            catch (FormatException)
            {
                return null;
            }
        }
    }
}

// End of file.
