/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\Content.cs
 * PURPOSE: Content management and related code
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using NB.Content.HashSets;
using NB.Content.Internal;
using NB.Content.Internal.IncomplInput;
using NB.Import;
using NB.Language;
using NB.Network;
using NB.Network.eD2k;
using Support;
using Support.HashFunctions;

namespace NB.Content
{
    class TemporaryFile
    {
        private const FHandle.ThreadSafetyMode c_DefaultAccessMode = FHandle.ThreadSafetyMode.NeedSynchronization;

        private readonly FHandle f_File;
        public readonly string f_FullFileName;
        private DeletionKind f_DeletionKind;

        private TemporaryFile(FHandle a_File,
                              string a_FullFileName,
                              DeletionKind a_DeletionKind)
        {
            f_File = a_File;
            f_FullFileName = a_FullFileName;
            f_DeletionKind = a_DeletionKind;

            if (DeletionKind.DeleteOnFinalize != f_DeletionKind)
            {
                GC.SuppressFinalize(this);
            }
        }

        ~TemporaryFile()
        {
            Close();
        }

        public void AdjustDeleteOnFinalize(bool a_Set)
        {
            if (DeletionKind.DeletionAlreadyHandled != f_DeletionKind)
            {
                if (false == a_Set)
                {
                    f_DeletionKind = DeletionKind.DontDeleteOnFinalize;

                    GC.SuppressFinalize(this);
                }
                else
                {
                    f_DeletionKind = DeletionKind.DeleteOnFinalize;

                    GC.ReRegisterForFinalize(this);
                }
            }
        }

        public void Close()
        {
            f_File.Close();

            if (DeletionKind.DeleteOnFinalize == f_DeletionKind)
            {
                if (true == FHandle.Delete(f_FullFileName))
                {
                    f_DeletionKind = DeletionKind.DeletionAlreadyHandled;
                }

                GC.SuppressFinalize(this);
            }
        }

        public void CancelDeletionAndClose()
        {
            AdjustDeleteOnFinalize(false);
            Close();
        }

        public static TemporaryFile Create()
        {
            return Create(DeletionKind.DeleteThroughOSOnClose);
        }

        public static TemporaryFile Create(DeletionKind a_DeletionKind)
        {
            string l_TempDir = Software.s_BootConfig.p_TempDirVolatile;
            TemporaryFile l_TempFile = Create(l_TempDir, a_DeletionKind);

            if (null == l_TempFile &&
                false == Directory.Exists(l_TempDir))
            {
                // Try to use the default temp dir.

                l_TempFile = Create(Software.sp_DefaultTempPath, a_DeletionKind);
            }

            return l_TempFile;
        }

        private static TemporaryFile Create(string a_TempDir,
                                            DeletionKind a_DeletionKind)
        {
            string l_FileName = string.Format(System.Globalization.CultureInfo.InvariantCulture, "temp{0:00000}.tmp", Util.GetRandomUIntFast() % 100000);
            l_FileName = LongPath.Combine(a_TempDir, l_FileName);

            return Create(l_FileName, FileMode.CreateNew, a_DeletionKind);
        }

        public static TemporaryFile Create(string a_FileName,
                                           FileMode a_Mode,
                                           DeletionKind a_DeletionKind)
        {
            // The DeleteOnClose flag ensures that the temp file is
            // deleted even if this process ends ungraceful (as long
            // as the machine isn't powered off or the OS crashes).
            FileOptions l_FileOptions = (DeletionKind.DeleteThroughOSOnClose == a_DeletionKind ? FileOptions.DeleteOnClose : FileOptions.None);
            Exception l_Junk;
            FHandle l_File = FHandle.OpenRWRITE(typeof(TemporaryFile), a_FileName, a_Mode, l_FileOptions, FHandle.c_SuggestedBufferSize, c_DefaultAccessMode, out l_Junk);

            if (null != l_File)
            {
                return new TemporaryFile(l_File, a_FileName, a_DeletionKind);
            }

            return null;
        }

        public FHandle ExposeFHandle()
        {
            return f_File;
        }

        public enum DeletionKind : byte
        {
            DeleteThroughOSOnClose = 0,
            DeleteOnFinalize = 1,
            DontDeleteOnFinalize = 2,

            // Internal alias.
            DeletionAlreadyHandled = DeleteThroughOSOnClose,
        }
    }

    class HardMemory
    {
        private const uint c_IndexBufferSize = (128 * 1024);
        public const uint c_BodyBufferSize = (512 * 1024);

        private const byte c_HeaderSize = 12;
        // The aligment requirement may not be smaller than the header.
        public const ushort c_Aligment = 512;
        public const string c_IndexFileExt = ".index";
        private const string c_IndexFileBackupExt = ".index.backup";
        public const string c_BodyFileExt = ".body";
        private const ushort c_ReadBufferElementCount = 512;
        private const ushort c_WriteBufferElementCount = 512;
        // c_MaximumBodyFilePos must be aligned.
        public const ulong c_MaximumBodyFilePos = (FHandle.c_MaxFileSize - uint.MaxValue);

        private const uint c_IndexHeaderMagic = 0xB46ACD5F;
        private const uint c_BodyHeaderMagic = 0x9EA8147B;
        private const ushort c_MyRevision = 8;
        private const ushort c_FirstRevisionWithAppDataSetInfo = 3;
        public const uint c_NoAppDataSetNum = 0;

        private static ByteArray s_KillBuffer;

        public readonly AppDataRole f_AppDataRole;
        public readonly uint f_AppDataSetNum;
        private readonly string f_FileNamePrefix;
        private readonly FHandle f_IndexFile;
        private readonly FHandle f_BodyFile;

        private readonly Dictionary<ElmKey, Element> f_Elements;
        private ListHead<Element> f_List;

        private ulong f_NextElementKey;
        private ulong f_NextIndexSlot;
        private ulong f_LargestPossibleGap;
        // f_GapSearchElement can be null.
        private Element f_GapSearchElement;
        private IndexDirtyness f_IndexDirty;
        private bool f_AllowFlush;

        // f_Resource can be null.
        private Resource f_Resource;
        private bool f_ConcurrentElementWritesPossible;

        public bool p_AllowFlush
        {
            get { return f_AllowFlush; }
            set { f_AllowFlush = value; }
        }

        public Resource p_Resource
        {
            get { return f_Resource; }
        }

        static HardMemory()
        {
            s_KillBuffer = new ByteArray(Element.c_Size);
        }

        private HardMemory(AppDataRole a_AppDataRole,
                           uint a_AppDataSetNum,
                           string a_FileNamePrefix,
                           FHandle a_IndexFile,
                           FHandle a_BodyFile,
                           Dictionary<ElmKey, Element> a_Elements,
                           ref ListHead<Element> a_List)
        {
            f_AppDataRole = a_AppDataRole;
            f_AppDataSetNum = a_AppDataSetNum;
            f_FileNamePrefix = a_FileNamePrefix;
            f_IndexFile = a_IndexFile;
            f_BodyFile = a_BodyFile;
            f_Elements = a_Elements;
            f_List = a_List;

            f_AllowFlush = true;
        }

        public bool IsEmpty()
        {
            return (0 == f_List.p_Count);
        }

        public void SetResource(Resource a_Resource,
                                bool a_ConcurrentElementWritesPossible)
        {
            // a_ConcurrentElementIOPossible indicates if the Write()
            // routine will be used by several threads operating on the
            // same element at the same time.

            f_Resource = a_Resource;
            f_ConcurrentElementWritesPossible = a_ConcurrentElementWritesPossible;
        }

        [Conditional("DEBUG")]
        private void DebugAssertAccess(bool a_ExclRequired)
        {
            if (null != f_Resource &&
                false == (false == a_ExclRequired ? f_Resource.HasSharedOrExclusiveAccess() : f_Resource.HasExclusiveAccess()))
            {
                // Resource not acquired properly.

                throw (new ApplicationException());
            }
        }

        [Conditional("DEBUG")]
        private void DebugAssertValidElement(Element a_Element)
        {
            if (true == a_Element.p_Deleted)
            {
                throw (new ApplicationException());
            }

            f_List.DebugAssertReferenced(a_Element);
        }

        private static bool OpenFiles(string a_Path,
                                      FileMode a_Mode,
                                      out FHandle a_IndexFile,
                                      out FHandle a_BodyFile,
                                      out string a_ErrMsg)
        {
            Exception l_Exc;
            string l_IndexFileName = (a_Path + c_IndexFileExt);
            a_IndexFile = FHandle.OpenRWRITE(typeof(HardMemory), l_IndexFileName, a_Mode, FileOptions.None, c_IndexBufferSize, FHandle.ThreadSafetyMode.NeedSynchronization, out l_Exc);

            if (null == a_IndexFile)
            {
                if (FileMode.Open == a_Mode &&
                    false == File.Exists(l_IndexFileName))
                {
                    // There might be a backup file from an index
                    // compact operation that was aborted.

                    string l_IndexBackupName = (a_Path + c_IndexFileBackupExt);

                    if (true == File.Exists(l_IndexBackupName) &&
                        true == FHandle.Move(l_IndexBackupName, l_IndexFileName))
                    {
                        a_IndexFile = FHandle.OpenRWRITE(typeof(HardMemory), l_IndexFileName, FileMode.Open, FileOptions.None, c_IndexBufferSize, FHandle.ThreadSafetyMode.NeedSynchronization, out l_Exc);
                    }
                }

                if (null == a_IndexFile)
                {
                    a_BodyFile = null;
                    a_ErrMsg = l_Exc.Message;

                    return false;
                }
            }

            a_BodyFile = FHandle.OpenRWRITE(typeof(HardMemory), a_Path + c_BodyFileExt, a_Mode, FileOptions.None, c_BodyBufferSize, FHandle.ThreadSafetyMode.NeedSynchronization, out l_Exc);

            if (null == a_BodyFile)
            {
                a_IndexFile.Close();
                a_ErrMsg = l_Exc.Message;

                return false;
            }

            a_ErrMsg = null;

            return true;
        }

        public static HardMemory Create(TextSet a_TextSet,
                                        AppDataRole a_AppDataRole,
                                        uint a_ExpectedAppDataSetNum,
                                        string a_Path,
                                        out string a_ErrMsg)
        {
            // a_ExpectedAppDataSetNum may be c_NoAppDataSetNum. This
            // routine will overwrite existing data.

            FHandle l_IndexFile;
            FHandle l_BodyFile;

            if (false == OpenFiles(a_Path, FileMode.Create, out l_IndexFile, out l_BodyFile, out a_ErrMsg))
            {
                return null;
            }

            // If there is no set number yet, create one.
            uint l_EffAppDataSetNum = (c_NoAppDataSetNum == a_ExpectedAppDataSetNum ? Util.GetRandomUInt() : a_ExpectedAppDataSetNum);

            if (false == WriteHeader(l_IndexFile, c_IndexHeaderMagic, a_AppDataRole, l_EffAppDataSetNum, true) ||
                false == WriteHeader(l_BodyFile, c_BodyHeaderMagic, a_AppDataRole, l_EffAppDataSetNum, true))
            {
                l_IndexFile.Close();
                l_BodyFile.Close();

                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrWriteFailed];

                return null;
            }

            a_ErrMsg = null;

            ListHead<Element> l_List = new ListHead<Element>();
            HardMemory l_HardMem = new HardMemory(a_AppDataRole,
                                                  l_EffAppDataSetNum,
                                                  a_Path,
                                                  l_IndexFile,
                                                  l_BodyFile,
                                                  new Dictionary<ElmKey, Element>(),
                                                  ref l_List);
            l_HardMem.f_NextElementKey = 1;

            return l_HardMem;
        }

        private static bool CalcMaxCountFromIndexFileSize(ulong a_IndexFileSize,
                                                          out ulong a_Count)
        {
            a_Count = 0;

            if (c_Aligment > a_IndexFileSize)
            {
                // No entries. Accept any file with a size in that
                // range, since header sizes can vary across versions.

                return true;
            }

            ulong l_Diff = (a_IndexFileSize - c_Aligment);

            if (0 == (l_Diff % Element.c_Size))
            {
                a_Count = (l_Diff / Element.c_Size);

                return true;
            }

            return false;
        }

        public static HardMemory Load(TextSet a_TextSet,
                                      AppDataRole a_AppDataRole,
                                      uint a_ExpectedAppDataSetNum,
                                      string a_Path,
                                      out string a_ErrMsg)
        {
            // a_ExpectedAppDataSetNum may be c_NoAppDataSetNum.

            FHandle l_IndexFile;
            FHandle l_BodyFile;

            if (false == OpenFiles(a_Path, FileMode.Open, out l_IndexFile, out l_BodyFile, out a_ErrMsg))
            {
                return null;
            }

            HardMemory l_HardMem = LoadCore(a_TextSet, a_AppDataRole, a_ExpectedAppDataSetNum, a_Path, l_IndexFile, l_BodyFile, out a_ErrMsg);

            if (null == l_HardMem)
            {
                l_IndexFile.Close();
                l_BodyFile.Close();
            }

            return l_HardMem;
        }

        private static HardMemory LoadCore(TextSet a_TextSet,
                                           AppDataRole a_AppDataRole,
                                           uint a_ExpectedAppDataSetNum,
                                           string a_Path,
                                           FHandle a_IndexFile,
                                           FHandle a_BodyFile,
                                           out string a_ErrMsg)
        {
            uint l_IndexAppDataSetNum;
            uint l_BodyAppDataSetNum;
            bool l_IndexHasOldRev;
            bool l_BodyHasOldRev;
            if (false == CheckMagicAndRevision(a_TextSet, a_IndexFile, c_IndexHeaderMagic, a_AppDataRole, out l_IndexAppDataSetNum, out l_IndexHasOldRev, out a_ErrMsg) ||
                false == CheckMagicAndRevision(a_TextSet, a_BodyFile, c_BodyHeaderMagic, a_AppDataRole, out l_BodyAppDataSetNum, out l_BodyHasOldRev, out a_ErrMsg))
            {
                return null;
            }

            uint l_EffAppDataSetNum;
            bool l_NeedUpdateHeadersDueToAppDataSetNumber;
            if (false == InterpretAppDataSetNumber(a_ExpectedAppDataSetNum, l_IndexAppDataSetNum, l_BodyAppDataSetNum, out l_EffAppDataSetNum, out l_NeedUpdateHeadersDueToAppDataSetNumber))
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrSetMismatch];

                return null;
            }

            ulong l_IndexFileSize;

            if (false == a_IndexFile.GetSize(out l_IndexFileSize))
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrAccessFailed];

                return null;
            }

            ulong l_MaxElms;

            if (false == CalcMaxCountFromIndexFileSize(l_IndexFileSize, out l_MaxElms))
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrWrongFileSize];

                return null;
            }

            int l_Capacity = (int)Math.Min((uint)int.MaxValue, l_MaxElms);
            Dictionary<ElmKey, Element> l_Elements = new Dictionary<ElmKey, Element>(l_Capacity);
            List<Element> l_List = new List<Element>(l_Capacity);
            ulong l_HighestKey = 0;

            int l_BufSize = ((int)Math.Min(l_MaxElms, c_ReadBufferElementCount) * Element.c_Size);
            byte[] l_Buf = new byte[l_BufSize];
            ulong l_SlotIndex = 0;

            while (0 != l_MaxElms)
            {
                uint l_CurCount = (uint)Math.Min(c_ReadBufferElementCount, l_MaxElms);
                uint l_CurBlockSize = (l_CurCount * Element.c_Size);
                ByteArray l_ReadDest = new ByteArray(l_Buf, 0, (int)l_CurBlockSize);

                if (false == a_IndexFile.Read(Element.CalcIndexPos(l_SlotIndex), ref l_ReadDest))
                {
                    a_ErrMsg = a_TextSet[TextKeys.MsgFileErrReadFailed];

                    return null;
                }

                int l_BufPos = 0;
                for (ulong lim = (l_SlotIndex + l_CurCount); l_SlotIndex < lim; l_SlotIndex++)
                {
                    Element l_Element;
                    if (false == Element.Read(l_Buf, l_BufPos, l_SlotIndex, out l_Element))
                    {
                        a_ErrMsg = a_TextSet[TextKeys.MsgFileErrCorruption];

                        return null;
                    }
                    else if (null != l_Element)
                    {
                        try
                        {
                            l_Elements.Add(l_Element.f_ElmKey, l_Element);
                        }
                        catch (ArgumentException l_Exc)
                        {
                            // Key exists already.

                            a_ErrMsg = l_Exc.Message;

                            return null;
                        }

                        l_List.Add(l_Element);
                        l_HighestKey = Math.Max(l_HighestKey, l_Element.f_ElmKey.f_Val);
                    }

                    l_BufPos += Element.c_Size;
                }

                l_MaxElms -= l_CurCount;
            }

            ulong l_HighestIndexSlot = (0 == l_List.Count ? 0 : l_List[l_List.Count - 1].f_IndexSlot);
            ListHead<Element> l_ListHead;
            uint l_LargestPossibleGap;

            if (false == VerifyAndReorderElementList(ref l_List, out l_ListHead, out l_LargestPossibleGap))
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrCorruption];

                return null;
            }

            HardMemory l_HardMem = new HardMemory(a_AppDataRole,
                                                  l_EffAppDataSetNum,
                                                  a_Path,
                                                  a_IndexFile,
                                                  a_BodyFile,
                                                  l_Elements,
                                                  ref l_ListHead);
            l_HardMem.f_NextElementKey = (l_HighestKey + 1);
            l_HardMem.f_NextIndexSlot = (l_HighestIndexSlot + 1);
            l_HardMem.f_LargestPossibleGap = l_LargestPossibleGap;
            l_HardMem.TrimBodyFile(false);

            if (true == l_IndexHasOldRev ||
                true == l_BodyHasOldRev ||
                true == l_NeedUpdateHeadersDueToAppDataSetNumber)
            {
                // The header is obsolete.

                bool l_HeadersWritten = (true == l_HardMem.WriteHeader(a_IndexFile, c_IndexHeaderMagic, true) &&
                                         true == l_HardMem.WriteHeader(a_BodyFile, c_BodyHeaderMagic, true));
            }

            a_ErrMsg = null;

            return l_HardMem;
        }

        private static bool InterpretAppDataSetNumber(uint a_ExpectedAppDataSetNum,
                                                      uint a_IndexAppDataSetNum,
                                                      uint a_BodyAppDataSetNum,
                                                      out uint a_EffAppDataSetNum,
                                                      out bool a_NeedUpdateHeaders)
        {
            bool l_HasExpAppDataSetNum = (c_NoAppDataSetNum != a_ExpectedAppDataSetNum);
            a_NeedUpdateHeaders = false;

            if (c_NoAppDataSetNum != a_IndexAppDataSetNum)
            {
                a_EffAppDataSetNum = a_IndexAppDataSetNum;

                if (true == l_HasExpAppDataSetNum &&
                    a_IndexAppDataSetNum != a_ExpectedAppDataSetNum)
                {
                    return false;
                }

                if (c_NoAppDataSetNum == a_BodyAppDataSetNum)
                {
                    a_NeedUpdateHeaders = true;
                }
                else if (a_IndexAppDataSetNum != a_BodyAppDataSetNum)
                {
                    return false;
                }
            }
            else if (c_NoAppDataSetNum != a_BodyAppDataSetNum)
            {
                a_EffAppDataSetNum = a_BodyAppDataSetNum;

                if (true == l_HasExpAppDataSetNum &&
                    a_BodyAppDataSetNum != a_ExpectedAppDataSetNum)
                {
                    return false;
                }

                a_NeedUpdateHeaders = true;
            }
            else
            {
                a_EffAppDataSetNum = (false == l_HasExpAppDataSetNum ? Util.GetRandomUInt() : a_ExpectedAppDataSetNum);
                a_NeedUpdateHeaders = true;
            }

            return true;
        }

        public static HardMemory LoadOrCreate(TextSet a_TextSet,
                                              AppDataRole a_AppDataRole,
                                              uint a_ExpectedAppDataSetNum,
                                              string a_Path,
                                              bool a_MayLoad,
                                              bool a_MayOverwrite,
                                              out bool a_Loaded,
                                              out string a_ErrMsg)
        {
            // a_ExpectedAppDataSetNum may be c_NoAppDataSetNum.

            HardMemory l_HardMemory;

            if (false == a_MayLoad)
            {
                l_HardMemory = null;
                a_Loaded = false;
                a_ErrMsg = null;
            }
            else
            {
                l_HardMemory = Load(a_TextSet, a_AppDataRole, a_ExpectedAppDataSetNum, a_Path, out a_ErrMsg);
                a_Loaded = (null != l_HardMemory);
            }

            if (null == l_HardMemory)
            {
                if (false == a_MayOverwrite &&
                    (true == File.Exists(a_Path + c_IndexFileExt) ||
                     true == File.Exists(a_Path + c_BodyFileExt)))
                {
                    // We may not overwrite, so return the error that
                    // occurred when trying to load.

                    return null;
                }

                l_HardMemory = Create(a_TextSet, a_AppDataRole, a_ExpectedAppDataSetNum, a_Path, out a_ErrMsg);
            }

            return l_HardMemory;
        }

        private static bool VerifyAndReorderElementList(ref List<Element> a_SrcList,
                                                        out ListHead<Element> a_DestList,
                                                        out uint a_LargestPossibleGap)
        {
            a_DestList = new ListHead<Element>();

            if (0 == a_SrcList.Count)
            {
                a_LargestPossibleGap = 0;
            }
            else
            {
                // We don't calculate this now, so just max it so that
                // Allocate() will check things.
                a_LargestPossibleGap = uint.MaxValue;
                ulong l_FinalPos = c_Aligment;

                foreach (Element l_Candid in a_SrcList)
                {
                    if (l_Candid.f_BodyFilePos >= l_FinalPos)
                    {
                        a_DestList.AddLast(l_Candid);

                        l_FinalPos = l_Candid.f_BodyFinalPos;
                    }
                    else
                    {
                        for (Element l_Elm = a_DestList.p_First; null != l_Elm; l_Elm = l_Elm.p_Next)
                        {
                            if (l_Candid.f_BodyFinalPos <= l_Elm.f_BodyFinalPos)
                            {
                                if (l_Candid.f_BodyFinalPos > l_Elm.f_BodyFilePos)
                                {
                                    // Overlap (file corrupt).

                                    Util.DebugBreak();

                                    return false;
                                }

                                a_DestList.AddBefore(l_Elm, l_Candid);

                                break;
                            }
                        }
                    }
                }
            }

            return true;
        }

        private ulong GetWastedBodySize()
        {
            // This routine is not thread safe.

            ulong l_WastedSize = 0;

            if (0 < f_List.p_Count)
            {
                ulong l_FinalPos = c_Aligment;
                for (Element l_Elm = f_List.p_First; null != l_Elm; l_Elm = l_Elm.p_Next)
                {
                    l_WastedSize += (l_Elm.f_BodyFilePos - l_FinalPos);
                    l_FinalPos = l_Elm.f_BodyFinalPos;
                }
            }

            return l_WastedSize;
        }

        private ulong GetWastedIndexSize()
        {
            // This routine is not thread safe.

            int l_Count = f_List.p_Count;
            ulong l_IndexFileSize;

            if (0 < l_Count &&
                true == f_IndexFile.GetSize(out l_IndexFileSize))
            {
                ulong l_ExpectedSize = (((ulong)l_Count * Element.c_Size) - c_Aligment);

                if (l_IndexFileSize > l_ExpectedSize)
                {
                    return (l_IndexFileSize - l_ExpectedSize);
                }
            }

            return 0;
        }

        private bool FindGap(uint a_MaxSize,
                             out ulong a_IndexSlot,
                             out Element a_BodyPred)
        {
            if (0 == f_Elements.Count)
            {
                a_IndexSlot = 0;
                a_BodyPred = null;

                f_NextIndexSlot = 1;
            }
            else
            {
                a_IndexSlot = FindIndexGap();
                a_BodyPred = FindBodyGap(a_MaxSize);

                if (null != a_BodyPred &&
                    c_MaximumBodyFilePos <= a_BodyPred.f_BodyFinalPos)
                {
                    return false;
                }
            }

            return true;
        }

        private ulong FindIndexGap()
        {
            // An overflow of this value would be a serious issue,
            // since we don't check if a slot is really not occupied
            // before we return. But it's not imaginable that an
            // overflow ever happens.

            return f_NextIndexSlot++;
        }

        private ulong UndoFindIndexGap()
        {
            // Not wasting slots can safe some disk space during longer
            // sessions.

            return f_NextIndexSlot--;
        }

        private Element FindBodyGap(uint a_RequestedSize)
        {
            // Returns the preceding node.

            if (a_RequestedSize <= f_LargestPossibleGap)
            {
                // However, there isn't necessarily a gap as large as
                // f_LargestPossibleGap.

                ulong l_MaxGapSize = 0;

                if (null == f_GapSearchElement ||
                    true == f_GapSearchElement.p_Deleted)
                {
                    f_GapSearchElement = f_List.p_First;
                }

                Element l_LoopElm = f_GapSearchElement;

                while (true)
                {
                    Element l_PrevElm = l_LoopElm.p_Previous;
                    ulong l_Gap;

                    if (null == l_PrevElm)
                    {
                        l_Gap = (l_LoopElm.f_BodyFilePos - c_Aligment);
                    }
                    else
                    {
                        l_Gap = (l_LoopElm.f_BodyFilePos - l_PrevElm.f_BodyFinalPos);
                    }

                    if (l_Gap >= a_RequestedSize)
                    {
                        return l_PrevElm;
                    }

                    l_MaxGapSize = Math.Max(l_MaxGapSize, l_Gap);

                    l_LoopElm = l_LoopElm.p_Next;

                    if (null == l_LoopElm)
                    {
                        l_LoopElm = f_List.p_First;
                    }

                    if (l_LoopElm == f_GapSearchElement)
                    {
                        // We have looped through all elements and now
                        // know the maximum gap size for sure.
                        f_LargestPossibleGap = l_MaxGapSize;

                        break;
                    }
                }

                f_GapSearchElement = f_GapSearchElement.p_Next;
            }

            return f_List.p_Last;
        }

        private static Element FindSmallestElement(Element a_ScanStartElm)
        {
            // Input may not be null and output never will be.

            Element l_ResElm = a_ScanStartElm;

            for (Element l_Elm = l_ResElm.p_Next; null != l_Elm; l_Elm = l_Elm.p_Next)
            {
                if (l_ResElm.f_MaxSize > l_Elm.f_MaxSize)
                {
                    l_ResElm = l_Elm;
                }
            }

            return l_ResElm;
        }

        private bool CompactBodyFile(bool a_OnlyIfWorthwhile,
                                     ICompactProgressCollector a_Collector)
        {
            // a_Collector may be null. Returns false if an error
            // occurred.

            if (0 >= f_List.p_Count)
            {
                return true;
            }

            if (true == a_OnlyIfWorthwhile)
            {
                ulong l_FullSize = f_List.p_Last.f_BodyFinalPos;
                ulong l_WastedSize = GetWastedBodySize();

                if ((8 * 1024 * 1024) > l_WastedSize ||
                    (l_FullSize / 12) > l_WastedSize)
                {
                    return true;
                }
            }

            Thread l_Thread = Thread.CurrentThread;
            ThreadPriority l_StoredPrio = l_Thread.Priority;
            l_Thread.Priority = ThreadPriority.BelowNormal;

            bool l_Succ = CompactBodyFileCore(a_Collector);

            l_Thread.Priority = l_StoredPrio;

            return l_Succ;
        }

        private bool CompactBodyFileCore(ICompactProgressCollector a_Collector)
        {
            ulong l_GapPos = c_Aligment;
            // l_Elm is the first item after the gap.
            Element l_Elm = f_List.p_First;
            Element l_SmallestElmLeft = null;
            uint l_Pos = 0;
            uint l_Count = (uint)f_List.p_Count;
            bool l_NeedCompactBody = false;
            byte[] l_TempBuffer = new byte[48 * 1024 * 1024];

            while (false == a_Collector.p_Aborted)
            {
                if (null != a_Collector)
                {
                    a_Collector.NotifyProgress(l_Pos, l_Count);
                }

                ulong l_Gap = (l_Elm.f_BodyFilePos - l_GapPos);

                if (0 != l_Gap &&
                    null == l_SmallestElmLeft)
                {
                    l_SmallestElmLeft = FindSmallestElement(l_Elm);
                }

                if (0 == l_Gap ||
                    l_SmallestElmLeft.f_MaxSize > l_Gap)
                {
                    // This gap can't be filled.

                    if (l_SmallestElmLeft == l_Elm)
                    {
                        l_SmallestElmLeft = null;
                    }

                    l_GapPos = l_Elm.f_BodyFinalPos;
                    l_Elm = l_Elm.p_Next;
                    l_Pos++;

                    if (null == l_Elm)
                    {
                        break;
                    }

                    continue;
                }

                // Copy the smallest element into our gap.

                bool l_OK = this.CopyElementToBodyPos(l_SmallestElmLeft, l_GapPos, l_SmallestElmLeft.p_ValidLen, true, l_TempBuffer);

                if (false == l_OK)
                {
                    return false;
                }

                Element l_TempElm = new Element(l_SmallestElmLeft.f_ElmKey, l_SmallestElmLeft.f_IndexSlot, l_GapPos, l_SmallestElmLeft.f_MaxSize, l_SmallestElmLeft.p_ValidLen);

                if (false == UpdateElement(l_TempElm) &&
                    false == f_IndexFile.Flush())
                {
                    return false;
                }

                l_NeedCompactBody = true;
                f_List.AddBefore(l_Elm, l_TempElm);
                l_GapPos = l_TempElm.f_BodyFinalPos;

                // Do not remove l_SmallestElmLeft earlier than here,
                // since it can be identical to l_Elm.
                f_List.Remove(l_SmallestElmLeft);
                l_SmallestElmLeft.SetDeleted();

                if (l_SmallestElmLeft == l_Elm)
                {
                    l_Elm = l_TempElm.p_Next;
                    l_Pos++;

                    if (null == l_Elm)
                    {
                        break;
                    }
                }

                l_SmallestElmLeft = null;
            }

            if (true == l_NeedCompactBody)
            {
                TrimBodyFile(false);
            }

            return true;
        }

        private bool CompactAndCloseIndexFile()
        {
            // The index file will always be closed when this routine
            // returns.

            bool l_IndexClosed = false;

            try
            {
                ulong l_IndexFileSize;

                if (false == f_IndexFile.GetSize(out l_IndexFileSize))
                {
                    return false;
                }

                ulong l_EstimatedCount;
                if (true == CalcMaxCountFromIndexFileSize(l_IndexFileSize, out l_EstimatedCount) &&
                    l_EstimatedCount <= (uint)f_Elements.Count)
                {
                    // Nothing to do.

                    return true;
                }

                string l_BackupFileName = (f_FileNamePrefix + c_IndexFileBackupExt);
                Exception l_Exc;
                // Open in a way that allows us to overwrite an old
                // backup file (if any).
                FHandle l_BackupFile = FHandle.OpenRWRITE(typeof(HardMemory), l_BackupFileName, FileMode.Create, FileOptions.None, c_IndexBufferSize, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

                if (null == l_BackupFile)
                {
                    return false;
                }

                bool l_DeleteBackup = false;

                try
                {
                    if (false == WriteHeader(l_BackupFile, c_IndexHeaderMagic, false))
                    {
                        l_DeleteBackup = true;

                        return false;
                    }

                    byte[] l_Buf = new byte[c_WriteBufferElementCount * Element.c_Size];
                    int l_BufItems = 0;
                    ulong l_FilePos = c_Aligment;

                    for (Element l_Elm = f_List.p_First; true; l_Elm = l_Elm.p_Next)
                    {
                        if (null == l_Elm ||
                            l_BufItems == c_WriteBufferElementCount)
                        {
                            // We reached the end of the list or the
                            // buffer should be emptied (or both).

                            if (0 != l_BufItems)
                            {
                                ByteArray l_Data = new ByteArray(l_Buf, 0, l_BufItems * Element.c_Size);

                                if (false == l_BackupFile.Write(l_FilePos, ref l_Data))
                                {
                                    l_DeleteBackup = true;

                                    return false;
                                }

                                l_BufItems = 0;
                                l_FilePos += (uint)l_Data.f_Len;
                            }

                            if (null == l_Elm)
                            {
                                break;
                            }
                        }

                        l_Elm.Write(l_Buf, l_BufItems * Element.c_Size);
                        l_BufItems++;
                    }
                }
                finally
                {
                    l_BackupFile.Close(true);

                    if (true == l_DeleteBackup)
                    {
                        bool l_Deleted = FHandle.Delete(l_BackupFileName);
                    }
                }

                // We now have a tight backup of the index that we can
                // use to substitute the bloated index file. If we are
                // interrupted while doing this, the loader code is able
                // to fix things.

                f_IndexFile.Close(true);
                l_IndexClosed = true;

                string l_IndexFileName = (f_FileNamePrefix + c_IndexFileExt);

                return (true == FHandle.Delete(l_IndexFileName) &&
                        true == FHandle.Move(l_BackupFileName, l_IndexFileName));
            }
            finally
            {
                if (false == l_IndexClosed)
                {
                    f_IndexFile.Close(true);
                }
            }
        }

        public void Close(bool a_MayCompact,
                          ICompactProgressCollector a_Collector)
        {
            Close(false == a_MayCompact ? CompactKind.DontCompact : CompactKind.CompactIfWorthwhile, a_Collector);
        }

        public void Close(CompactKind a_CompactKind,
                          ICompactProgressCollector a_Collector)
        {
            // a_Collector may be null.

            if (CompactKind.DontCompact != a_CompactKind)
            {
                bool l_BodyDone = CompactBodyFile(CompactKind.CompactIfWorthwhile == a_CompactKind, a_Collector);
            }

            f_BodyFile.Close(true);

            if (CompactKind.DontCompact == a_CompactKind)
            {
                f_IndexFile.Close(true);
            }
            else
            {
                bool l_IndexDone = CompactAndCloseIndexFile();
            }
        }

        public bool Flush()
        {
            // In theory no special access rights are required for
            // this operation, but we demand shared access anyway.
            DebugAssertAccess(false);

            // Flush the body first.
            return (true == f_BodyFile.Flush() &&
                    true == f_IndexFile.Flush());
        }

        public ElmKey FindUnoccupiedKey()
        {
            return FindUnoccupiedKey(0);
        }

        public ElmKey FindUnoccupiedKey(uint a_Reserved)
        {
            DebugAssertAccess(true);

            while (true)
            {
                if (a_Reserved > f_NextElementKey)
                {
                    f_NextElementKey = a_Reserved;
                }

                if (false == f_Elements.ContainsKey(new ElmKey(f_NextElementKey)))
                {
                    return new ElmKey(f_NextElementKey++);
                }

                f_NextElementKey++;
            }
        }

        public static bool Delete(string a_Path)
        {
            FHandle.Delete(a_Path + c_IndexFileExt);
            FHandle.Delete(a_Path + c_BodyFileExt);

            return true;
        }

        private bool CopyElementToBodyPos(Element a_Element,
                                          ulong a_BodyFileDestOffset,
                                          uint a_Length,
                                          bool a_Flush,
                                          byte[] a_TempBuffer)
        {
            // a_TempBuffer may be null.

            return FHandle.CopyFileToFile(f_BodyFile,
                                          a_Element.f_BodyFilePos,
                                          f_BodyFile,
                                          a_BodyFileDestOffset,
                                          a_Length,
                                          a_Flush,
                                          a_TempBuffer);
        }

        private bool CopyFileToElement(FHandle a_Src,
                                       ulong a_BodyFileDestOffset,
                                       uint a_Length,
                                       bool a_Flush)
        {
            return FHandle.CopyFileToFile(a_Src,
                                          0,
                                          f_BodyFile,
                                          a_BodyFileDestOffset,
                                          a_Length,
                                          a_Flush,
                                          null);
        }

        public bool SetAtomic(ElmKey a_ElmKey,
                              FHandle a_NewContent)
        {
            ByteArray l_Empty = new ByteArray();

            return SetAtomicInternal(a_ElmKey, ref l_Empty, a_NewContent);
        }

        public bool SetAtomic(ElmKey a_ElmKey,
                              ref ByteArray a_Data)
        {
            return SetAtomicInternal(a_ElmKey, ref a_Data, null);
        }

        private bool SetAtomicInternal(ElmKey a_ElmKey,
                                       ref ByteArray a_SrcData,
                                       FHandle a_SrcFile)
        {
            // Substitutes the content of a_Element with the content
            // of the buffer OR file as an atomic operation (on failure
            // or interruption nothing will have changed).

            DebugAssertAccess(true);

            uint l_Size;

            if (null == a_SrcFile)
            {
                l_Size = (uint)a_SrcData.f_Len;
            }
            else
            {
                ulong l_Size64;

                if (false == a_SrcFile.GetSize(out l_Size64) ||
                    0 == l_Size64 ||
                    uint.MaxValue < l_Size64)
                {
                    return false;
                }

                l_Size = (uint)l_Size64;
            }

            ulong l_IndexSlot;
            Element l_BodyPred;

            if (false == FindGap(l_Size, out l_IndexSlot, out l_BodyPred))
            {
                return false;
            }

            ulong l_BodyFilePos;
            if (true == SecureDisk(f_BodyFile, l_BodyPred, l_Size, out l_BodyFilePos))
            {
                bool l_Copied;

                if (null == a_SrcFile)
                {
                    l_Copied = (true == f_BodyFile.Write(l_BodyFilePos, ref a_SrcData) &&
                                (false == f_AllowFlush ||
                                 true == f_BodyFile.Flush()));
                }
                else
                {
                    l_Copied = CopyFileToElement(a_SrcFile, l_BodyFilePos, l_Size, f_AllowFlush);
                }

                if (true == l_Copied)
                {
                    Element l_SubstElement;
                    Element l_OldElement = GetElement(a_ElmKey);

                    if (null == l_OldElement)
                    {
                        l_SubstElement = new Element(a_ElmKey,
                                                     l_IndexSlot,
                                                     l_BodyFilePos,
                                                     l_Size,
                                                     l_Size);
                    }
                    else
                    {
                        // We don't need the extra slot, so undo.
                        UndoFindIndexGap();

                        l_SubstElement = new Element(l_OldElement.f_ElmKey,
                                                     l_OldElement.f_IndexSlot,
                                                     l_BodyFilePos,
                                                     l_Size,
                                                     l_Size);
                    }

                    // Overwrite the index with our new data, we expect
                    // this to happen atomic (since index slots are
                    // aligned).
                    if (true == UpdateElement(l_SubstElement))
                    {
                        f_IndexDirty |= IndexDirtyness.DirtyFromAlloc;

                        AddNodeAfter(l_BodyPred, l_SubstElement);

                        if (null != l_OldElement)
                        {
                            bool l_Compact = (f_List.p_Last == l_OldElement);

                            f_List.Remove(l_OldElement);
                            l_OldElement.SetDeleted();

                            if (true == l_Compact)
                            {
                                // We are able to release unneeded
                                // space at the end of the file.

                                TrimBodyFile(f_AllowFlush);
                            }
                        }

                        f_Elements[l_SubstElement.f_ElmKey] = l_SubstElement;

                        return true;
                    }
                }

                // We failed, but nothing has changed (except the file
                // size maybe).

                if (f_List.p_Last == l_BodyPred)
                {
                    // Restore original body file size.

                    TrimBodyFile(false);
                }
            }

            return false;
        }

        public bool Resize(ref Element a_Element,
                           uint a_NewMaxSize,
                           bool a_PreserveOldData)
        {
            // Don't use the original reference anymore after a
            // successful call. On failure nothing will have changed.

            DebugAssertAccess(true);
            DebugAssertValidElement(a_Element);

            if (a_Element.f_MaxSize == a_NewMaxSize)
            {
                return true;
            }

            if (0 != (f_IndexDirty & IndexDirtyness.DirtyFromKill) &&
                true == f_AllowFlush)
            {
                if (false == ProtectIntegrity())
                {
                    return false;
                }
            }

            bool l_HandleInPlace = false;
            bool l_CompactOnFailure = false;
            bool l_FlushAndCompactOnSucc = false;
            Element l_NextElm = a_Element.p_Next;

            if (null == l_NextElm)
            {
                // We are at the end of the file and will try to just
                // modify the file's size, so that we don't have to
                // copy any data.

                if (a_Element.f_MaxSize > a_NewMaxSize)
                {
                    // Shrink.

                    l_HandleInPlace = true;
                    l_FlushAndCompactOnSucc = true;
                }
                else
                {
                    // Expand. Don't use SecureDisk() here, since it's
                    // only suitable for fresh allocation.

                    ulong l_ReqSize = (a_Element.f_BodyFilePos + a_NewMaxSize);

                    if (true == f_BodyFile.SetSize(l_ReqSize))
                    {
                        l_HandleInPlace = true;
                        l_CompactOnFailure = true;
                    }
                }
            }
            else if ((l_NextElm.f_BodyFilePos - a_Element.f_BodyFilePos) >= a_NewMaxSize)
            {
                // We can handle the resize in place with no note
                // (ideal case).

                l_HandleInPlace = true;
            }

            uint l_NewValidSize = (false == a_PreserveOldData ?
                                   0 : Math.Min(a_NewMaxSize, a_Element.p_ValidLen));
            Element l_BodyPred;
            ulong l_BodyFilePos;
            IndexDirtyness l_Dirtyness;

            if (true == l_HandleInPlace)
            {
                l_BodyPred = a_Element;
                l_BodyFilePos = a_Element.f_BodyFilePos;
                l_Dirtyness = (a_Element.f_MaxSize > a_NewMaxSize ?
                               IndexDirtyness.DirtyFromKill : IndexDirtyness.DirtyFromAlloc);
            }
            else
            {
                // Expand to a new place. We'll have to find that new
                // place and copy the old data in.

                ulong l_Junk;
                if (false == FindGap(a_NewMaxSize, out l_Junk, out l_BodyPred))
                {
                    goto COMPACT_IF_REQUIRED_AND_FAIL;
                }

                UndoFindIndexGap();

                if (false == SecureDisk(f_BodyFile, l_BodyPred, a_NewMaxSize, out l_BodyFilePos))
                {
                    goto COMPACT_IF_REQUIRED_AND_FAIL;
                }

                l_CompactOnFailure = (f_List.p_Last == l_BodyPred);

                if (true == a_PreserveOldData &&
                    false == CopyElementToBodyPos(a_Element, l_BodyFilePos, l_NewValidSize, f_AllowFlush, null))
                {
                    goto COMPACT_IF_REQUIRED_AND_FAIL;
                }

                l_Dirtyness = IndexDirtyness.DirtyFromAlloc;
            }

            Element l_NewElm = new Element(a_Element.f_ElmKey,
                                           a_Element.f_IndexSlot,
                                           l_BodyFilePos,
                                           a_NewMaxSize,
                                           l_NewValidSize);

            if (false == UpdateElement(l_NewElm))
            {
                goto COMPACT_IF_REQUIRED_AND_FAIL;
            }

            f_IndexDirty |= l_Dirtyness;

            AddNodeAfter(l_BodyPred, l_NewElm);
            f_List.Remove(a_Element);
            a_Element.SetDeleted();

            f_Elements[l_NewElm.f_ElmKey] = l_NewElm;
            a_Element = l_NewElm;

            if (true == l_FlushAndCompactOnSucc)
            {
                // We may now be able to release unneeded space.
                TrimBodyFile(f_AllowFlush);
            }

            return true;

            COMPACT_IF_REQUIRED_AND_FAIL:

            if (true == l_CompactOnFailure)
            {
                // We have increased the file size, so undo it.
                TrimBodyFile(false);
            }

            return false;
        }

        public Element Allocate(ElmKey a_ElmKey,
                                uint a_MaxSize)
        {
            // Return can be null (out of disk space). The caller might
            // wish to call Flush() after the call.

            DebugAssertAccess(true);

            ulong l_IndexSlot;
            Element l_BodyPred;

            if (false == FindGap(a_MaxSize, out l_IndexSlot, out l_BodyPred))
            {
                return null;
            }

            Element l_Element = AllocateDeep(a_ElmKey, l_IndexSlot, l_BodyPred, a_MaxSize);

            if (null == l_Element)
            {
                return null;
            }

            f_Elements.Add(l_Element.f_ElmKey, l_Element);
            AddNodeAfter(l_BodyPred, l_Element);

            return l_Element;
        }

        private void AddNodeAfter(Element a_Predecessor,
                                  Element a_NewElm)
        {
            // a_Predecessor can be null.

            if (null == a_Predecessor)
            {
                f_List.AddFirst(a_NewElm);
            }
            else
            {
                f_List.AddAfter(a_Predecessor, a_NewElm);
            }
        }

        private Element AllocateDeep(ElmKey a_ElmKey,
                                     ulong a_IndexSlot,
                                     Element a_BodyPred,
                                     uint a_MaxSize)
        {
            // Secures the required disk space and indexes the element
            // or returns null on failure.

            if (0 != (f_IndexDirty & IndexDirtyness.DirtyFromKill) &&
                true == f_AllowFlush)
            {
                if (false == ProtectIntegrity())
                {
                    return null;
                }
            }

            ulong l_BodyFilePos;
            if (true == SecureDisk(f_BodyFile, a_BodyPred, a_MaxSize, out l_BodyFilePos))
            {
                Element l_Element = new Element(a_ElmKey,
                                                a_IndexSlot,
                                                l_BodyFilePos,
                                                a_MaxSize,
                                                0);

                // This will (of course) also increase the index file's
                // size if necessary.
                if (true == UpdateElement(l_Element))
                {
                    f_IndexDirty |= IndexDirtyness.DirtyFromAlloc;

                    return l_Element;
                }

                // We failed to index.

                if (0 == f_Elements.Count ||
                    f_List.p_Last == a_BodyPred)
                {
                    // Restore original body file size.

                    TrimBodyFile(false);
                }
            }

            return null;
        }

        private bool SecureDisk(FHandle a_File,
                                Element a_Predecessor,
                                uint a_Size,
                                out ulong a_FilePos)
        {
            // Makes sure the requested disk space is secured.
            // a_Predecessor may be null.

            if (null == a_Predecessor)
            {
                a_FilePos = c_Aligment;
            }
            else
            {
                a_FilePos = a_Predecessor.f_BodyFinalPos;
            }

            ulong l_OldFileSize;

            if (false == a_File.GetSize(out l_OldFileSize))
            {
                return false;
            }

            ulong l_ReqFileSize = (a_FilePos + a_Size);

            if (l_OldFileSize < l_ReqFileSize)
            {
                return a_File.SetSize(l_ReqFileSize);
            }

            return true;
        }

        public bool Delete(ElmKey a_ElmKey)
        {
            Element l_Element = GetElement(a_ElmKey);

            if (null != l_Element)
            {
                return Delete(l_Element);
            }

            return false;
        }

        public bool Delete(Element a_Element)
        {
            DebugAssertAccess(true);
            DebugAssertValidElement(a_Element);

            if (false == f_IndexFile.Write(a_Element.CalcIndexPos(), ref s_KillBuffer))
            {
                return false;
            }

            f_IndexDirty |= IndexDirtyness.DirtyFromKill;

            bool l_FlushIndexBeforeResize = true;

            if (0 != (f_IndexDirty & IndexDirtyness.DirtyFromAlloc) &&
                true == f_AllowFlush)
            {
                if (false == ProtectIntegrity())
                {
                    // It's to late to abort, but we never expect to
                    // land here anyway.

                    Util.DebugBreak();
                }
                else
                {
                    l_FlushIndexBeforeResize = false;
                }
            }

            bool l_OK = f_Elements.Remove(a_Element.f_ElmKey);

            if (1 == f_List.p_Count)
            {
                f_LargestPossibleGap = 0;
            }
            else
            {
                ulong l_NewGap;

                if (f_List.p_First == a_Element)
                {
                    // Calc the size of the new starting gap.

                    l_NewGap = (a_Element.p_Next.f_BodyFilePos - c_Aligment);
                }
                else if (f_List.p_Last != a_Element)
                {
                    // Calc the size of the new gap.

                    l_NewGap = (a_Element.p_Next.f_BodyFilePos -
                                a_Element.p_Previous.f_BodyFinalPos);
                }
                else
                {
                    l_NewGap = 0;
                }

                f_LargestPossibleGap = Math.Max(f_LargestPossibleGap, l_NewGap);
            }

            if (a_Element == f_List.p_Last)
            {
                TrimBodyFile(l_FlushIndexBeforeResize);
            }

            if (a_Element == f_GapSearchElement)
            {
                f_GapSearchElement = f_GapSearchElement.p_Next;
            }

            f_List.Remove(a_Element);
            a_Element.SetDeleted();

            return true;
        }

        private bool ProtectIntegrity()
        {
            // This routine flushes all previous index changes to make
            // sure a file range isn't referenced multiple times in
            // the index file (it can happen if allocations and deletes
            // are mixed without flushing).
            if (true == f_IndexFile.Flush())
            {
                f_IndexDirty = IndexDirtyness.NotDirty;

                return true;
            }

            return false;
        }

        private void TrimBodyFile(bool a_FlushIndexBefore)
        {
            // Flush to make sure all references to the slack of the
            // file are gone.
            if (true == a_FlushIndexBefore)
            {
                if (false == f_IndexFile.Flush())
                {
                    // Should never happen.

                    Util.DebugBreak();

                    return;
                }

                f_IndexDirty = IndexDirtyness.NotDirty;
            }

            ulong l_BodyFileSize = (0 == f_List.p_Count ?
                                    c_HeaderSize : f_List.p_Last.f_BodyFinalPos);

            f_BodyFile.SetSize(l_BodyFileSize);
        }

        private bool UpdateElement(Element a_Element)
        {
            // a_Element doesn't have to be in any list to use this
            // routine.

            ByteArray f_Data = new ByteArray(Element.c_Size);

            a_Element.Write(f_Data.f_Arr, 0);

            return f_IndexFile.Write(a_Element.CalcIndexPos(), ref f_Data);
        }

        public Element GetElement(ElmKey a_ElmKey)
        {
            // Return can be null.

            DebugAssertAccess(false);

            Element l_Element;
            bool l_OK = f_Elements.TryGetValue(a_ElmKey, out l_Element);

            return l_Element;
        }

        public bool CopyToFile(Element a_Element,
                               FHandle a_DestFile,
                               bool a_NeedToZero,
                               bool a_Flush)
        {
            DebugAssertAccess(false);
            DebugAssertValidElement(a_Element);

            a_DestFile.SetSize(a_Element.f_MaxSize);

            if (a_Element.f_MaxSize > a_Element.p_ValidLen &&
                true == a_NeedToZero)
            {
                throw (new NotImplementedException());
            }

            return FHandle.CopyFileToFile(f_BodyFile,
                                          a_Element.f_BodyFilePos,
                                          a_DestFile,
                                          0,
                                          a_Element.p_ValidLen,
                                          a_Flush,
                                          null);
        }

        public bool Read(Element a_Element,
                         uint a_StartingIndex,
                         ref ByteArray a_Data)
        {
            // We need shared access for reads in order to ensure no
            // one deletes the element while we are reading.
            DebugAssertAccess(false);
            DebugAssertValidElement(a_Element);

            if (a_Element.p_ValidLen > a_StartingIndex)
            {
                uint l_FinalPos = (a_StartingIndex + (uint)a_Data.f_Len);

                if (l_FinalPos > a_Element.p_ValidLen)
                {
                    // Won't overflow.
                    int l_Valid = (int)(a_Element.p_ValidLen - a_StartingIndex);
                    // 'Read' some non-intialized bytes (simply zero
                    // the buffer).
                    Buffer2.BlockZero(a_Data.f_Arr, a_Data.f_Idx + l_Valid, a_Data.f_Len - l_Valid);
                    ByteArray l_ReadDest = new ByteArray(a_Data.f_Arr, a_Data.f_Idx, l_Valid);

                    return f_BodyFile.Read(a_Element.f_BodyFilePos + a_StartingIndex, ref l_ReadDest);
                }

                return f_BodyFile.Read(a_Element.f_BodyFilePos + a_StartingIndex, ref a_Data);
            }

            // 'Read' non-intialized bytes.
            Buffer2.BlockZero(ref a_Data);

            return true;
        }

        public bool Write(Element a_Element,
                          uint a_ElementStartingIndex,
                          FHandle a_File)
        {
            DebugAssertAccess(f_ConcurrentElementWritesPossible);
            DebugAssertValidElement(a_Element);

            ulong l_FileSize;

            if (false == a_File.GetSize(out l_FileSize) ||
                uint.MaxValue < l_FileSize)
            {
                return false;
            }

            uint l_BytesToGo = (uint)l_FileSize;
            byte[] l_TempBuf = new byte[(int)Math.Min(256u * 1024u, l_BytesToGo)];

            uint l_ElementPos = a_ElementStartingIndex;
            ulong l_FilePos = 0;
            while (0 != l_BytesToGo)
            {
                uint l_CurBytes = Math.Min(l_BytesToGo, (uint)l_TempBuf.Length);
                ByteArray l_Data = new ByteArray(l_TempBuf, 0, (int)l_CurBytes);

                if (false == a_File.Read(l_FilePos, ref l_Data) ||
                    false == Write(a_Element, l_ElementPos, ref l_Data))
                {
                    return false;
                }

                l_ElementPos += l_CurBytes;
                l_FilePos += l_CurBytes;
                l_BytesToGo -= l_CurBytes;
            }

            return true;
        }

        public bool Zero(Element a_Element,
                         uint a_KeepStartingBytes)
        {
            DebugAssertAccess(f_ConcurrentElementWritesPossible);
            DebugAssertValidElement(a_Element);

            if (a_Element.p_ValidLen > a_KeepStartingBytes)
            {
                a_Element.SetValidLength(a_KeepStartingBytes);

                return UpdateElement(a_Element);
            }

            return true;
        }

        public bool Write(Element a_Element,
                          uint a_StartingIndex,
                          ref ByteArray a_Data)
        {
            // We need only shared access if there is one thread at
            // most writing on a certain element at one point in time,
            // otherwise we need exclusive access.
            DebugAssertAccess(f_ConcurrentElementWritesPossible);
            DebugAssertValidElement(a_Element);

            if (0 == a_Data.f_Len)
            {
                return true;
            }

            if (true == f_BodyFile.Write(a_Element.f_BodyFilePos + a_StartingIndex, ref a_Data))
            {
                uint l_NewValidLen = (a_StartingIndex + (uint)a_Data.f_Len);

                if (a_Element.p_ValidLen < l_NewValidLen)
                {
                    if (a_Element.p_ValidLen < a_StartingIndex)
                    {
                        // We need to prepend zeros.
                        uint l_Zeros = (a_StartingIndex - a_Element.p_ValidLen);

                        if (false == f_BodyFile.Zero(a_Element.f_BodyFilePos + a_Element.p_ValidLen, l_Zeros))
                        {
                            return false;
                        }
                    }

                    a_Element.SetValidLength(l_NewValidLen);

                    return UpdateElement(a_Element);
                }

                return true;
            }

            return false;
        }

        private bool WriteHeader(FHandle a_File,
                                 uint a_Magic,
                                 bool a_AlsoFlush)
        {
            return WriteHeader(a_File, a_Magic, f_AppDataRole, f_AppDataSetNum, a_AlsoFlush);
        }

        private static bool WriteHeader(FHandle a_File,
                                        uint a_Magic,
                                        AppDataRole a_AppDataRole,
                                        uint a_AppDataSetNum,
                                        bool a_AlsoFlush)
        {
            ByteArray l_Data = new ByteArray(c_HeaderSize);
            Util.WriteUInt32LE(a_Magic, l_Data.f_Arr, 0);
            Util.WriteUInt16LE(c_MyRevision, l_Data.f_Arr, 4);
            l_Data.f_Arr[6] = (byte)a_AppDataRole;
            // The eighth byte is unused.
            Util.WriteUInt32LE(a_AppDataSetNum, l_Data.f_Arr, 8);

            return (true == a_File.Write(0, ref l_Data) &&
                    (false == a_AlsoFlush ||
                     true == a_File.Flush()));
        }

        private static bool CheckMagicAndRevision(TextSet a_TextSet,
                                                  FHandle a_File,
                                                  uint a_ExpectedMagic,
                                                  AppDataRole a_ExpectedAppDataRole,
                                                  out uint a_AppDataSetNum,
                                                  out bool a_OldRevision,
                                                  out string a_ErrMsg)
        {
            // a_ErrMsg will be valid in case of failure.

            a_AppDataSetNum = c_NoAppDataSetNum;
            a_OldRevision = false;

            ByteArray l_Data = new ByteArray(c_HeaderSize);
            ulong l_Size;

            if (false == a_File.GetSize(out l_Size))
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrAccessFailed];

                return false;
            }

            bool l_HeaderRead;

            if (c_HeaderSize > l_Size)
            {
                // This file is probably from an older version
                // (pre v1.3.0).

                if (sizeof(ulong) > l_Size)
                {
                    a_ErrMsg = a_TextSet[TextKeys.MsgFileErrWrongFileSize];

                    return false;
                }

                ByteArray l_ShortData = new ByteArray(l_Data.f_Arr, 0, sizeof(ulong));
                l_HeaderRead = a_File.Read(0, ref l_ShortData);
            }
            else
            {
                l_HeaderRead = a_File.Read(0, ref l_Data);
            }

            if (false == l_HeaderRead)
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrReadFailed];

                return false;
            }

            if (Util.ReadUInt32LE(l_Data.f_Arr, l_Data.f_Idx) != a_ExpectedMagic)
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgFileErrCorruption];

                return false;
            }

            uint l_Revision = Util.ReadUInt16LE(l_Data.f_Arr, l_Data.f_Idx + 4);

            if (c_MyRevision < l_Revision)
            {
                a_ErrMsg = a_TextSet[TextKeys.MsgAppDataErrFromNewerVersion];

                return false;
            }

            if (c_FirstRevisionWithAppDataSetInfo <= l_Revision)
            {
                if (l_Data.f_Arr[l_Data.f_Idx + 6] != (byte)a_ExpectedAppDataRole)
                {
                    a_ErrMsg = a_TextSet[TextKeys.MsgFileErrSetMismatch];

                    return false;
                }

                a_AppDataSetNum = Util.ReadUInt32LE(l_Data.f_Arr, l_Data.f_Idx + 8);
            }

            if (c_MyRevision > l_Revision)
            {
                a_OldRevision = true;
            }

            a_ErrMsg = null;

            return true;
        }

        [Flags]
        private enum IndexDirtyness : byte
        {
            NotDirty = 0,
            DirtyFromAlloc = 0x1,
            DirtyFromKill = 0x2,
        }

        public class Element : ListBaseClass<Element>
        {
            public const byte c_Size = (4 * sizeof(ulong));

            public readonly ElmKey f_ElmKey;
            public readonly ulong f_IndexSlot;
            public readonly ulong f_BodyFilePos;
            // f_BodyFinalPos is rounded up to the next aligned
            // position (c_Aligment).
            public readonly ulong f_BodyFinalPos;
            public readonly uint f_MaxSize;

            private uint f_ValidLen;
            private bool f_Deleted;

            public uint p_ValidLen
            {
                get { return f_ValidLen; }
            }

            public bool p_Deleted
            {
                get { return f_Deleted; }
            }

            public Element(ElmKey a_ElmKey,
                           ulong a_IndexSlot,
                           ulong a_BodyFilePos,
                           uint a_MaxSize,
                           uint a_ValidLen)
            {
                f_ElmKey = a_ElmKey;
                f_IndexSlot = a_IndexSlot;
                f_BodyFilePos = a_BodyFilePos;
                f_BodyFinalPos = (f_BodyFilePos + a_MaxSize);

                uint l_Rest = (a_MaxSize % c_Aligment);

                if (0 != l_Rest)
                {
                    f_BodyFinalPos += (c_Aligment - l_Rest);
                }

                f_MaxSize = a_MaxSize;
                f_ValidLen = a_ValidLen;
            }

            public static bool Read(byte[] a_Src,
                                    int a_Index,
                                    ulong a_IndexSlot,
                                    out Element a_Element)
            {
                ulong l_MaxSize64 = Util.ReadUInt64LE(a_Src, a_Index);

                if (0 == l_MaxSize64)
                {
                    // Entry invalidated, return true.

                    a_Element = null;

                    return true;
                }

                ulong l_ValidLen64 = Util.ReadUInt64LE(a_Src, a_Index + 8);
                ulong l_BodyFilePos = Util.ReadUInt64LE(a_Src, a_Index + 16);

                if (// We don't yet allow >32 bit elements.
                    uint.MaxValue < l_MaxSize64 ||
                    l_ValidLen64 > l_MaxSize64 ||
                    0 == l_BodyFilePos ||
                    HardMemory.c_MaximumBodyFilePos < l_BodyFilePos ||
                    0 != (l_BodyFilePos % c_Aligment))
                {
                    // Corruption.

                    a_Element = null;

                    return false;
                }

                ElmKey l_ElmKey = new ElmKey(Util.ReadUInt64LE(a_Src, a_Index + 24));

                a_Element = new Element(l_ElmKey, a_IndexSlot, l_BodyFilePos, (uint)l_MaxSize64, (uint)l_ValidLen64);

                return true;
            }

            public void Write(byte[] a_Dest,
                              int a_Index)
            {
                // Zero the first half of the buffer, since we will
                // write 32 bit values into 64 bit slots.
                Buffer2.BlockZero(a_Dest, a_Index, 16);

                Util.WriteUInt32LE(f_MaxSize, a_Dest, a_Index);
                Util.WriteUInt32LE(f_ValidLen, a_Dest, a_Index + 8);
                Util.WriteUInt64LE(f_BodyFilePos, a_Dest, a_Index + 16);
                Util.WriteUInt64LE(f_ElmKey.f_Val, a_Dest, a_Index + 24);
            }

            public ulong CalcIndexPos()
            {
                return CalcIndexPos(f_IndexSlot);
            }

            public static ulong CalcIndexPos(ulong a_IndexSlot)
            {
                return (c_Aligment + (a_IndexSlot * Element.c_Size));
            }

            public void SetValidLength(uint a_NewLen)
            {
                // May only get called by the HardMemory class.

                f_ValidLen = a_NewLen;
            }

            public void SetDeleted()
            {
                // May only get called by the HardMemory class.

                f_Deleted = true;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct LeakFinder
        {
            // Don't use the struct default constructor.

            private readonly HardMemory f_HardMem;
            private readonly Dictionary<ElmKey, Element> f_Remaining;

            public int p_RemainingCount
            {
                get { return f_Remaining.Count; }
            }

            public LeakFinder(HardMemory a_HardMem)
            {
                f_HardMem = a_HardMem;
                f_Remaining = new Dictionary<ElmKey, Element>(f_HardMem.f_Elements);
            }

            public LeakStates TouchIfNotZero(ElmKey a_ElmKey)
            {
                if (true == a_ElmKey.IsZero())
                {
                    return LeakStates.NoSuchElement;
                }

                return Touch(a_ElmKey);
            }

            public LeakStates Touch(ElmKey a_ElmKey)
            {
                if (true == f_Remaining.Remove(a_ElmKey))
                {
                    return LeakStates.NoError;
                }

                bool l_RemovedAlready = f_HardMem.f_Elements.ContainsKey(a_ElmKey);

                return (false == l_RemovedAlready ? LeakStates.NoSuchElement : LeakStates.RedundantElement);
            }

            public void CollectGarbage()
            {
                foreach (Element l_Elm in f_Remaining.Values)
                {
                    bool l_Deleted = f_HardMem.Delete(l_Elm);
                }

                f_Remaining.Clear();
            }
        }

        public enum LeakStates : byte
        {
            NoError = 0,
            NoSuchElement = 1,
            RedundantElement = 2,
        }

        public enum CompactKind : byte
        {
            DontCompact = 1,
            CompactIfWorthwhile = 2,
            CompactInAnyCase = 3,
        }

        public enum AppDataRole : byte
        {
            Core = 1,
            Incompl = 2,
        }

        public interface ICompactProgressCollector
        {
            bool p_Aborted
            {
                get;
            }

            void NotifyProgress(uint a_Done,
                                uint a_Total);
        }
    }

    [Flags]
    enum P2PHashSets : ushort
    {
        None = 0,

        // Not a hash set.
        eD2kBitmap = 0x1,

        // eD2k hash set.
        eD2kHashSet = 0x2,
        // 'Advanced Intelligent Corruption Handling'.
        AICHTree = 0x4,
        // Tigerized AICH.
        TAICHTree = 0x8,
        // Upper8MiBTAICHTree represents a subset of the TAICHTree.
        // This flag should be used only for availability indication
        // with the file descriptor. Hash queries should use TAICHTree.
        Upper8MiBTAICHTree = 0x10,

        // Not a hash set. From 1.4.7 on.
        ClientNativeBitmap = 0x80,

        _ALL_HASHSETS_ = (eD2kHashSet | AICHTree | TAICHTree),
        _ALL_HASHSETS_INCL_SUBSETS_ = (_ALL_HASHSETS_ | Upper8MiBTAICHTree),
    }

    enum ComplStatus : byte
    {
        // That is CompletenessStatus.

        Dropped = 0,
        Incomplete = 1,
        // Missing files are files that were once completely available
        // on local drive but are now missing for some reason.
        Missing = 2,
        // Complete files are currently completely available on the local
        // drive.
        Complete = 3,
    }

    partial class KnownFile : ITransferable
    {
        private const ushort c_Magic = 0x5809;
        public const byte c_DefaultSlotCount = byte.MaxValue;

        // Where not otherwise noted, the fields of this part of the
        // class are under the f_Lock of the ContentManager class.

        public readonly FileNum f_FileNum;
        public readonly ulong f_FileSize;

        // f_Dic is only a reference to the dictionary this item belongs
        // to (the dictionary isn't owned by this class).
        private Dictionary<KnownFile, KnownFile> f_Dic;
        private LocalFile f_LocalFile;

        private ComplStatus f_Status;
        private FileFlags f_Flags;
        public MasterHashes f_MasterHashes;
        private Category f_Category;
        private FileNameKeywordCache f_KeywordCache;

        // Incomplete file fields.
        public string f_IncomplFileName;
        // The following fields are under a special lock of the
        // incomplete file workers.
        public bool f_IncomplWorkerOccupied;
        public ElmKey f_IncomplKey;
        public IncompleteFile f_Incompl;
        // Read by the GUI, volatile.
        public IncomplFileProgress f_IncomplProgress;

        // Complete file fields.
        private Time f_LastLinkTime;
        // f_ExpectedFileName is only valid for missing files. This file
        // name preserves the casing and is not necessarily a full file
        // name.
        private string f_ExpectedFileName;
        // f_ExpectedLastWriteTime is the master switch for fast
        // detection.
        private Time f_ExpectedLastWriteTime;
        // f_Hashes is only valid while this instace is not linked to
        // a local file.
        private byte[] f_Hashes;
        private bool f_CheckedQuickHash;
        // f_HashSetKey and f_HashElmLayout are under the sole control
        // of the hash worker.
        public ElmKey f_HashSetKey;
        public HashElmLayout f_HashElmLayout;

        public string f_CompatFileName;

        public LocalFile p_LocalFile
        {
            get { return f_LocalFile; }
        }

        public LocalFile p_LocalFileVolatile
        {
            get { return Util.VolatileRead(ref f_LocalFile); }
        }

        public ComplStatus p_Status
        {
            get { return f_Status; }
        }

        public ComplStatus p_StatusVolatile
        {
            get { return Util.VolatileRead(ref f_Status); }
        }

        public Category p_Category
        {
            get { return f_Category; }
        }

        public bool p_CheckedQuickHashVolatile
        {
            get { return Util.VolatileRead(ref f_CheckedQuickHash); }
        }

        public IncomplFileProgress p_IncomplProgressVolatile
        {
            get { return Util.VolatileRead(ref f_IncomplProgress); }
        }

        public uint p_MaxUlSlots
        {
            get { return byte.MaxValue; }
        }

        private KnownFile(FileNum a_FileNum,
                          ulong a_FileSize)
        {
            f_FileNum = a_FileNum;
            f_FileSize = a_FileSize;

            f_Flags = FileFlags.SingleFile;
            f_Category = Category.s_Unassigned;

            uint l_Div = (uint)(f_FileSize / eD2kHashSet.c_PartSize);
            f_PartCount = (l_Div + (0 == (f_FileSize % eD2kHashSet.c_PartSize) ? 0u : 1u));
            // Historical errors of the eD2k protocol OR how would
            // someone guess the part count, if he didn't know the
            // correct way of calculating it?
            // Example #1: 'Always add one extra part.'
            f_PartCountFileStatus = (l_Div + 1);
            // Example #2: 'Always add one extra, if there is at least
            // one full part.'
            f_PartCountHashSet = (l_Div + (eD2kHashSet.c_PartSize > f_FileSize ? 0u : 1u));
            f_eD2kBitArrayByteSize = ((f_PartCount / Util.c_BitsPerByte) + (0 == (f_PartCount % Util.c_BitsPerByte) ? 0u : 1u));
        }

        public static KnownFile CreateFileFromLink(FileNum a_FileNum,
                                                   FileLink a_Link,
                                                   FileID a_FileID,
                                                   bool a_Started)
        {
            return CreateFileFromLinkOrAlike(a_FileNum,
                                             a_Link.f_FileSize,
                                             a_Link.f_FileName,
                                             a_Link.f_MasterHashes,
                                             a_FileID,
                                             a_Started);
        }

        public static KnownFile CreateFileFromSearch(FileNum a_FileNum,
                                                     Gui.FileSearchPanel.FoundFile[] a_Files,
                                                     int a_Idx,
                                                     FileID a_FileID,
                                                     bool a_Started)
        {
            return CreateFileFromLinkOrAlike(a_FileNum,
                                             a_Files[a_Idx].f_FileSize,
                                             a_Files[a_Idx].f_FileName,
                                             new MasterHashes(a_FileID),
                                             a_FileID,
                                             a_Started);
        }

        public static KnownFile CreateFileFromCollection(FileNum a_FileNum,
                                                         FileCollection.FileInfo[] a_Infos,
                                                         int a_Idx,
                                                         FileID a_FileID,
                                                         bool a_Started)
        {
            return CreateFileFromLinkOrAlike(a_FileNum,
                                             a_Infos[a_Idx].f_FileSize,
                                             a_Infos[a_Idx].f_FileName,
                                             a_Infos[a_Idx].f_MasterHashes,
                                             a_FileID,
                                             a_Started);
        }

        public static KnownFile CreateFileByImporting(FileNum a_FileNum,
                                                      ImportableFile a_ImpFile,
                                                      bool a_Started)
        {
            return CreateFileFromLinkOrAlike(a_FileNum,
                                             a_ImpFile.f_FileSize,
                                             a_ImpFile.f_FileName,
                                             new MasterHashes(a_ImpFile.f_FileID),
                                             a_ImpFile.f_FileID,
                                             a_Started);
        }

        private static KnownFile CreateFileFromLinkOrAlike(FileNum a_FileNum,
                                                           ulong a_FileSize,
                                                           string a_FileName,
                                                           MasterHashes a_MasterHashes,
                                                           FileID a_FileID,
                                                           bool a_Started)
        {
            KnownFile l_Known = new KnownFile(a_FileNum, a_FileSize);
            l_Known.f_Status = ComplStatus.Incomplete;
            a_MasterHashes.Clone(out l_Known.f_MasterHashes);
            l_Known.f_FileID = a_FileID;
            l_Known.f_IncomplFileName = a_FileName;
            l_Known.f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(a_FileName);
            l_Known.InitSpeedRelatedParams(a_Started);
            l_Known.InitializeTimeFields();

            return l_Known;
        }

        public static KnownFile CreateFileFromLocal(FileNum a_FileNum,
                                                    LocalFile a_Local)
        {
            KnownFile l_Known = new KnownFile(a_FileNum, a_Local.f_FileSize);
            l_Known.InitSpeedRelatedParams(false);
            l_Known.InitializeTimeFields();
            l_Known.SetLocalFile(a_Local, false);

            return l_Known;
        }

        private void InitializeTimeFields()
        {
            f_AddedOn = Time.SysTimeNow();
            f_LastReception = new Time();
            f_LastSeenComplete = new Time();
        }

        public static KnownFile FromBuffer(BufferIO a_Buffer)
        {
            // The caller must hold the appropriate lock. Returns null
            // if an error occurred.

            if (c_Magic != a_Buffer.ReadUInt16LE())
            {
                return null;
            }

            FileNum l_FileNum = new FileNum(a_Buffer.ReadUInt64LE());
            ulong l_FileSize = a_Buffer.ReadUInt64LE();

            if (false == ContentManager.IsValidFileSize(l_FileSize))
            {
                // Don't treat this as a fatal error, though it's
                // quite bad and should never happen.

                Util.DebugBreak();

                l_FileSize = 1;
            }

            KnownFile l_Known = new KnownFile(l_FileNum, l_FileSize);

            l_Known.f_Status = (ComplStatus)a_Buffer.ReadByte();

            if (ComplStatus.Incomplete != l_Known.f_Status &&
                ComplStatus.Missing != l_Known.f_Status)
            {
                return null;
            }

            l_Known.f_Flags = (FileFlags)a_Buffer.ReadByte();

            if (0 == (FileFlags.SingleFile & l_Known.f_Flags) ||
                0 != (~FileFlags._ALL_ & l_Known.f_Flags))
            {
                // Repair.

                Util.DebugBreak();

                l_Known.f_Flags |= FileFlags.SingleFile;
                l_Known.f_Flags &= FileFlags._ALL_;
            }

            if (false == MasterHashes.FromBuffer(false, a_Buffer, out l_Known.f_MasterHashes))
            {
                return null;
            }

            if (true == l_Known.f_MasterHashes.p_HaseD2kHash)
            {
                l_Known.f_FileID = l_Known.f_MasterHashes.CreateFileID();
            }

            l_Known.f_Category = new Category(a_Buffer.ReadUInt16LE());

            if (false == l_Known.f_Category.IsUnassignedOrNonGeneric())
            {
                // Repair.

                Util.DebugBreak();

                l_Known.f_Category = Category.s_Unassigned;
            }

            if (true == a_Buffer.ReadBool())
            {
                // This file has an incomplete file key (even complete
                // files can).

                l_Known.f_IncomplKey = new ElmKey(a_Buffer.ReadUInt64LE());
            }

            if (ComplStatus.Incomplete == l_Known.f_Status)
            {
                string l_IncomplFileName = a_Buffer.ReadUTF16LEStringEncLen();

                l_Known.f_CompatFileName = l_Known.f_IncomplFileName = FileNames.MakeFileNameLegalFailsafe(l_IncomplFileName);

                if (true == a_Buffer.ReadBool())
                {
                    l_Known.f_IncomplProgress = IncomplFileProgress.FromBuffer(a_Buffer, l_FileSize);

                    if (null == l_Known.f_IncomplProgress)
                    {
                        return null;
                    }
                }
            }
            else
            {
                l_Known.f_LastLinkTime = a_Buffer.ReadTime();
                l_Known.f_ExpectedFileName = a_Buffer.ReadUTF16LEStringEncLen();
                l_Known.f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(l_Known.f_ExpectedFileName);
                l_Known.f_ExpectedLastWriteTime = a_Buffer.ReadFormerDateTime();
                l_Known.f_Hashes = a_Buffer.ReadBytes(2 * SHA_256.c_HashByteSize);
                l_Known.f_HashSetKey = new ElmKey(a_Buffer.ReadUInt64LE());

                if (0 >= l_Known.f_ExpectedFileName.Length ||
                    FileNames.c_MaxPathChars < l_Known.f_ExpectedFileName.Length)
                {
                    l_Known.f_CompatFileName = l_Known.f_ExpectedFileName = FileNames.c_FallbackFileName;
                    l_Known.f_ExpectedLastWriteTime = new Time();
                }
            }

            if (false == l_Known.ReadNetworkParamsFromBuffer(a_Buffer))
            {
                return null;
            }

            return l_Known;
        }

        public void ToBuffer(Time a_Time,
                             BufferIO a_Buffer)
        {
            // The caller must hold the appropriate lock.

            a_Buffer.WriteUInt16LE(c_Magic);
            a_Buffer.WriteUInt64LE(f_FileNum.f_Val);
            a_Buffer.WriteUInt64LE(f_FileSize);

            a_Buffer.WriteByte((byte)(ComplStatus.Complete == f_Status ? ComplStatus.Missing : f_Status));
            a_Buffer.WriteByte((byte)f_Flags);
            f_MasterHashes.ToBuffer(a_Buffer);
            a_Buffer.WriteUInt16LE(f_Category.f_Val);

            if (true == f_IncomplKey.IsNonzero())
            {
                // This file has an incomplete file key (even complete
                // files can).

                a_Buffer.WriteBool(true);
                a_Buffer.WriteUInt64LE(f_IncomplKey.f_Val);
            }
            else
            {
                a_Buffer.WriteBool(false);
            }

            if (ComplStatus.Incomplete == f_Status)
            {
                a_Buffer.WriteUTF16LEStringEncLen(f_IncomplFileName);

                IncomplFileProgress l_IncomplProgress = null;

                if (true == f_IncomplKey.IsNonzero())
                {
                    l_IncomplProgress = Util.VolatileRead(ref f_IncomplProgress);
                }

                if (null == l_IncomplProgress)
                {
                    a_Buffer.WriteBool(false);
                }
                else
                {
                    a_Buffer.WriteBool(true);
                    l_IncomplProgress.Write(a_Buffer);
                }
            }
            else
            {
                if (null == f_LocalFile)
                {
                    a_Buffer.WriteTime(f_LastLinkTime);
                    a_Buffer.WriteUTF16LEStringEncLen(f_ExpectedFileName);
                    a_Buffer.WriteTime(f_ExpectedLastWriteTime);
                    a_Buffer.WriteBytes(f_Hashes);
                }
                else
                {
                    // This file is still linked.

                    a_Buffer.WriteTime(a_Time);
                    a_Buffer.WriteUTF16LEStringEncLen(f_LocalFile.BuildFullFileName());
                    a_Buffer.WriteTime(f_LocalFile.f_LastWriteTime);
                    a_Buffer.WriteBytes(f_LocalFile.ExposeHashes());
                }

                a_Buffer.WriteUInt64LE(f_HashSetKey.f_Val);
            }

            WriteNetworkParamsToBuffer(a_Buffer);
        }

        public bool IsInDic(Dictionary<KnownFile, KnownFile> a_Dic)
        {
            return (a_Dic == f_Dic);
        }

        public void AddToDic(Dictionary<KnownFile, KnownFile> a_Dic)
        {
            ChangeDic(this, f_Dic, a_Dic);
            f_Dic = a_Dic;
        }

        public void TurnMissingIntoIncompl()
        {
            // The caller must hold the appropriate lock.

            f_Status = ComplStatus.Incomplete;

            f_IncomplFileName = Path.GetFileName(f_ExpectedFileName);
            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(f_IncomplFileName);

            f_LastLinkTime = new Time();
            MakeFastDetectionImpossible();
            f_Hashes = null;
            f_CheckedQuickHash = false;
        }

        public void TurnIncomplIntoMissing(string a_ExpectedFullFileName,
                                           Time a_ExpectedLastWriteTime,
                                           byte[] a_Hashes)
        {
            // The caller must hold the appropriate lock.

            f_Status = ComplStatus.Missing;
            f_KeywordCache = new FileNameKeywordCache();

            f_IncomplFileName = null;

            // Initialize fast detect params.
            f_ExpectedFileName = a_ExpectedFullFileName;
            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(Path.GetFileName(a_ExpectedFullFileName));
            f_ExpectedLastWriteTime = a_ExpectedLastWriteTime;
            f_Hashes = a_Hashes;

            // We don't need to release f_IncomplProgress here.
        }

        public void ClearLocalFile(Time a_Time)
        {
            // The caller must hold the appropriate lock.

            // Save away parameters needed for later recognition.

            f_LastLinkTime = a_Time;
            f_ExpectedFileName = f_LocalFile.BuildFullFileName();
            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(Path.GetFileName(f_ExpectedFileName));
            f_ExpectedLastWriteTime = f_LocalFile.f_LastWriteTime;
            f_Hashes = f_LocalFile.ExposeHashes();
            f_CheckedQuickHash = false;

            f_LocalFile.ClearKnownFile();
            f_LocalFile.ClearHashes();
            f_LocalFile = null;
            f_Status = ComplStatus.Missing;
            f_KeywordCache = new FileNameKeywordCache();
        }

        public void SetLocalFile(LocalFile a_Local,
                                 bool a_CopyHashesToLocal)
        {
            // The caller must hold the appropriate lock.

            f_Status = ComplStatus.Complete;
            f_LocalFile = a_Local;
            f_LocalFile.SetKnownFile(this);
            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(f_LocalFile.p_FileName);

            if (true == a_CopyHashesToLocal)
            {
                f_LocalFile.SetHashes(f_Hashes);
            }

            // Clear parameters needed for recognition in order to
            // release unneeded memory.

            MakeFastDetectionImpossible();
            f_Hashes = null;

            f_KeywordCache = new FileNameKeywordCache();
        }

        public void NotifyLocalFileNameChanged()
        {
            // The caller must hold the appropriate lock.

            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(f_LocalFile.p_FileName);
            ClearKeywordCache();
        }

        public void RenameMissing(string a_NewFileName)
        {
            // NOTE: Renaming a missing files will also deactivate fast
            // detection. The new name gets replaced by the name of
            // the local file that this known file is linked to next.

            // We don't assign a full file name here, but it's okay
            // since we disable fast detection anyway.
            f_ExpectedFileName = a_NewFileName;
            f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(a_NewFileName);
            f_ExpectedLastWriteTime = new Time();
        }

        public void MakeFastDetectionImpossible()
        {
            // The caller must hold the appropriate lock.

            f_ExpectedFileName = null;
            f_ExpectedLastWriteTime = new Time();
        }

        public bool TryToFastDetectLocalFile(LocalFile a_Local,
                                             string a_FullFileName)
        {
            // The caller must hold the appropriate lock.

            // NOTE: Fast detection will only work if f_ExpectedFileName
            // is a full file name (which is not always so).
            return (f_ExpectedLastWriteTime == a_Local.f_LastWriteTime &&
                    a_FullFileName == f_ExpectedFileName);
        }

        public bool HasTheseHashes(byte[] a_Hashes)
        {
            // The caller must hold the appropriate lock.

            byte[] l_Hashes = (null == f_LocalFile ? f_Hashes : f_LocalFile.ExposeHashes());

            return Buffer2.ValueEquals(l_Hashes, a_Hashes);
        }

        public bool TryToSetHashedLocalFile(LocalFile a_Local)
        {
            // The caller must hold the appropriate lock.

            if (true == Buffer2.ValueEquals(f_Hashes, a_Local.ExposeHashes()))
            {
                SetLocalFile(a_Local, false);
                f_CheckedQuickHash = true;

                return true;
            }

            return false;
        }

        public void SetStatusToDropped()
        {
            f_Status = ComplStatus.Dropped;
        }

        public bool HasTrustedHashes()
        {
            // The caller should hold the appropriate lock.

            return (0 != (FileFlags.AllHashesSelfCalculated & f_Flags));
        }

        public void AdjustAllHashesSelfCalculatedFlag(bool a_Set)
        {
            // The caller must hold the appropriate lock. Should only
            // be used synchronized with the hash worker.

            if (HasTrustedHashes() == a_Set)
            {
                return;
            }

            if (false == a_Set)
            {
                f_Flags &= ~FileFlags.AllHashesSelfCalculated;
            }
            else
            {
                f_Flags |= FileFlags.AllHashesSelfCalculated;
            }
        }

        public void SetCategory(Category a_Category)
        {
            // May be used without holding a lock.

            f_Category = a_Category;
        }

        public void ClearHashSetKey()
        {
            // May only be used by the hash worker (or otherwise
            // synchronized).

            f_HashSetKey = new ElmKey();
        }

        public void CalcHashSetLayout()
        {
            // May only be used by the hash worker. And it should not
            // be done while holding a lock, since it can be quite
            // expensive.

            if (false == f_HashElmLayout.IsValid())
            {
                f_HashElmLayout = new KnownFile.HashElmLayout(f_FileSize);
            }
        }

        public void SetQuickHashChecked()
        {
            // The caller must hold the appropriate lock.

            f_CheckedQuickHash = true;
        }

        public static void ChangeDic<T>(T a_Item,
                                        Dictionary<T, T> a_From,
                                        Dictionary<T, T> a_To)
        {
            // a_From and a_To may be null.

            if (null != a_From)
            {
                bool l_OK = a_From.Remove(a_Item);
            }

            if (null != a_To)
            {
                a_To.Add(a_Item, a_Item);
            }
        }

        public void ClearKeywordCache()
        {
            // The caller must hold the appropriate lock.

            if (true == f_KeywordCache.IsValid())
            {
                f_KeywordCache = new FileNameKeywordCache();
            }
        }

        public void TryAddKeywords(PublishCookie a_PubCookie)
        {
            // The caller must hold the appropriate lock.

            if (false == f_KeywordCache.IsValid())
            {
                if (ComplStatus.Incomplete == f_Status ||
                    ComplStatus.Complete == f_Status)
                {
                    f_KeywordCache = new FileNameKeywordCache(a_PubCookie.f_MD4, f_CompatFileName);
                }
                else
                {
                    return;
                }
            }

            f_KeywordCache.TryAddKeywords(a_PubCookie, this);
        }

        public string GetFileName(bool a_NeedFileSystemName)
        {
            // The caller must hold the appropriate lock. Returns the
            // file name without path.

            if (false == a_NeedFileSystemName)
            {
                return f_CompatFileName;
            }

            if (ComplStatus.Incomplete == f_Status)
            {
                return f_IncomplFileName;
            }

            if (ComplStatus.Missing == f_Status)
            {
                return Path.GetFileName(f_ExpectedFileName);
            }

            return f_LocalFile.p_FileName;
        }

        public string GetFileNameVolatile(bool a_NeedFileSystemName,
                                          out ComplStatus a_Status,
                                          out string a_ParentPath)
        {
            // Same as GetFileName(), but no lock is required.
            // a_ParentPath can be null on return.

            a_Status = f_Status;
            LocalFile l_Local = f_LocalFile;
            string l_IncomplFileName = f_IncomplFileName;
            string l_ExpectedFileName = f_ExpectedFileName;
            string l_CompatFileName = f_CompatFileName;
            Util.MemoryBarrier();

            bool l_Linked = (ComplStatus.Complete == a_Status &&
                             null != l_Local);

            if (false == l_Linked)
            {
                a_ParentPath = null;
            }
            else
            {
                ParentFolder l_Parent = l_Local.p_ParentVolatile;
                a_ParentPath = (null == l_Parent ? null : l_Parent.p_Path);

                if (true == a_NeedFileSystemName)
                {
                    return l_Local.p_FileName;
                }
            }

            if (false == a_NeedFileSystemName)
            {
                return l_CompatFileName;
            }

            if (ComplStatus.Incomplete == a_Status)
            {
                return (null == l_IncomplFileName ? string.Empty : l_IncomplFileName);
            }

            return (null == l_ExpectedFileName ? string.Empty : Path.GetFileName(l_ExpectedFileName));
        }

        public string GetFileNameVolatile(bool a_NeedFileSystemName)
        {
            ComplStatus l_Junk0;
            string l_Junk1;
            return GetFileNameVolatile(a_NeedFileSystemName, out l_Junk0, out l_Junk1);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct FileNameKeywordCache
        {
            // f_Buffer holds the UInt128s of all keywords in little
            // endian notation.
            public readonly uint[] f_Buffer;

            public FileNameKeywordCache(MD_4 a_MD4,
                                        string a_FileName)
            {
                string[] l_Keywords = SearchManager.IdentifyKeywords(a_FileName, false, true);

                if (null == l_Keywords)
                {
                    f_Buffer = new uint[0];

                    return;
                }

                f_Buffer = new uint[l_Keywords.Length * UInt128.c_UIntSize];

                for (int i = (l_Keywords.Length - 1); 0 <= i; i--)
                {
                    SearchManager.CalcKeywordHash(a_MD4, l_Keywords[i]);

                    // NOTE: We don't check if a hash value occurs
                    // multiple times, since that is extremely unlikely.

                    a_MD4.GetDigestLE(f_Buffer, i * UInt128.c_UIntSize);
                }
            }

            public bool IsValid()
            {
                return (null != f_Buffer);
            }

            public void TryAddKeywords(PublishCookie a_PubCookie,
                                       KnownFile a_Known)
            {
                for (int i = 0; f_Buffer.Length > i; i += UInt128.c_UIntSize)
                {
                    uint l_HighestUInt = f_Buffer[i + UInt128.c_UIntSize - 1];

                    if (true == a_PubCookie.IsRelevant(l_HighestUInt))
                    {
                        UInt128 l_KeywordID = new UInt128(f_Buffer, i);
                        List<KnownFile> l_Files;

                        if (false == a_PubCookie.f_Keywords.TryGetValue(l_KeywordID, out l_Files))
                        {
                            l_Files = new List<KnownFile>(1);
                            l_Files.Add(a_Known);

                            a_PubCookie.f_Keywords.Add(l_KeywordID, l_Files);
                        }
                        else
                        {
                            if (false == l_Files.Contains(a_Known))
                            {
                                l_Files.Add(a_Known);
                            }
                        }
                    }
                }
            }
        }

        [Flags]
        public enum FileFlags : byte
        {
            SingleFile = 0x1,
            AllHashesSelfCalculated = 0x40,

            _ALL_ = (SingleFile | AllHashesSelfCalculated),
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct HashElmLayout
        {
            // Don't use the struct default constructor.

            // 12 TiB. We use this limit to make sure we remain 32 bit
            // friendly (element size < 4 GiB).
            public const ulong c_MaxFileSize = (12UL * 1024 * 1024 * 1024 * 1024);
            private const uint c_CommonMagic = 0x96329A24;
            public const ushort c_CommonHeaderSize = HardMemory.c_Aligment;
            private const byte c_UsedHeaderSize = 24;
            private const ushort c_Alignment = HardMemory.c_Aligment;

            public readonly uint f_AICHOffset;
            public readonly uint f_TAICHOffset;
            public readonly uint f_ElementSize;

            public uint p_eD2kOffset
            {
                get { return c_CommonHeaderSize; }
            }

            public HashElmLayout(ulong a_FileSize)
            {
                uint f_Junk;
                uint f_eD2kSize = eD2kHashSet.CalcHashSetSize(a_FileSize, out f_Junk);
                f_AICHOffset = CalcNextOffset(c_CommonHeaderSize, f_eD2kSize);
                uint f_AICHSize = AICHTree.CalcHashSetSize(a_FileSize, out f_Junk);
                f_TAICHOffset = CalcNextOffset(f_AICHOffset, f_AICHSize);
                uint f_TAICHSize = TAICHTree.CalcHashSetSize(a_FileSize, out f_Junk);
                f_ElementSize = (f_TAICHOffset + f_TAICHSize);
            }

            public bool IsValid()
            {
                return (0 != f_ElementSize);
            }

            public ulong CalcMinAvailDiskSpaceForCreation()
            {
                // Returns the disk space that should be available at
                // least when the element is created.

                return ((ulong)ContentManager.c_MinAvailDiskSpace + f_ElementSize);
            }

            private static uint CalcNextOffset(uint a_Offset,
                                               uint a_Size)
            {
                uint l_Padding = (a_Size % c_Alignment);

                if (0 != l_Padding)
                {
                    l_Padding = (c_Alignment - l_Padding);
                }

                return (a_Offset + a_Size + l_Padding);
            }

            public static bool CheckHeader(KnownFile a_Known,
                                           HardMemory a_HardMem,
                                           HardMemory.Element a_Elm)
            {
                byte[] l_Header = new byte[c_UsedHeaderSize];
                ByteArray l_Data = new ByteArray(l_Header);

                if (false == a_HardMem.Read(a_Elm, 0, ref l_Data))
                {
                    return false;
                }

                if (Util.ReadUInt32LE(l_Header, 0) != c_CommonMagic ||
                    Util.ReadUInt32LE(l_Header, 4) != 0 ||
                    Util.ReadUInt64LE(l_Header, 8) != a_Known.f_FileNum.f_Val ||
                    Util.ReadUInt64LE(l_Header, 16) != a_Known.f_FileSize)
                {
                    return false;
                }

                return true;
            }

            public static bool WriteHeader(KnownFile a_Known,
                                           HardMemory a_HardMem,
                                           HardMemory.Element a_Elm)
            {
                byte[] l_Header = new byte[c_UsedHeaderSize];
                Util.WriteUInt32LE(c_CommonMagic, l_Header, 0);
                Util.WriteUInt64LE(a_Known.f_FileNum.f_Val, l_Header, 8);
                Util.WriteUInt64LE(a_Known.f_FileSize, l_Header, 16);

                ByteArray l_Data = new ByteArray(l_Header);

                return a_HardMem.Write(a_Elm, 0, ref l_Data);
            }

            public static bool DeleteOrResetChecked(KnownFile a_Known,
                                                    HardMemory a_HardMem,
                                                    ElmKey a_ElmKey,
                                                    bool a_JustReset)
            {
                // Does not clear a_Known.f_HashSetKey.

                if (false == ContentManager.IsValidHashElmKey(a_ElmKey))
                {
                    return false;
                }

                HardMemory.Element l_Elm = a_HardMem.GetElement(a_ElmKey);

                if (null == l_Elm ||
                    c_CommonHeaderSize >= l_Elm.f_MaxSize)
                {
                    return false;
                }

                if (false == CheckHeader(a_Known, a_HardMem, l_Elm))
                {
                    return false;
                }

                // Target confirmed, so begin with the actual task.

                if (false == a_JustReset)
                {
                    return a_HardMem.Delete(l_Elm);
                }

                // Keep only the header.
                return a_HardMem.Zero(l_Elm, KnownFile.HashElmLayout.c_CommonHeaderSize);
            }
        }
    }

    class ReadCookie : APCCookie
    {
        public readonly KnownFile f_Known;
        public readonly ulong f_Offset;
        public readonly int f_Size;

        public ByteArray f_Bytes;

        public ReadCookie(KnownFile a_Known,
                          ulong a_Offset,
                          int a_Size)
        {
            f_Known = a_Known;
            f_Offset = a_Offset;
            f_Size = a_Size;
        }

        public void Fail()
        {
            SetState(CookieStates.Failure);
        }

        public void Succeed(ref ByteArray a_Bytes)
        {
            f_Bytes = a_Bytes;
            SetState(CookieStates.Success);
        }
    }

    enum HashQueryKinds : byte
    {
        Indices = 0,
        Range = 1,
        AICHPartRecoveryData = 2,
    }

    class HashQueryCookie : GeneralAPCCookie
    {
        public const uint c_SuggestedMaxSize = (1024 * 1024);
        public const uint c_MaxSpecificIndices = (c_SuggestedMaxSize / MD_4.c_HashByteSize);

        public readonly KnownFile f_Known;
        public readonly P2PHashSets f_HashType;
        public readonly Opcodes f_Opcode;
        public readonly HashQueryKinds f_Kind;
        // f_MasterHash has big endian notation.
        public readonly byte[] f_MasterHash;

        // For f_Kind Indices only.
        public readonly uint[] f_HashIndices;
        // For f_Kind Range only.
        public readonly uint f_StartingIndex;
        public readonly uint f_Count;
        // For f_Kind AICHPartRecoveryData only.
        public readonly uint f_AICHTrgtPartIndex;

        public readonly ulong f_EstimatedResultSize;

        public uint f_TickedID;
        private bool f_MalformedInput;
        private byte[] f_Data;

        public bool p_MalformedInput
        {
            get { return f_MalformedInput; }
        }

        public byte[] p_Data
        {
            get { return f_Data; }
        }

        public HashQueryCookie(ContentManager a_ContentMngr,
                               KnownFile a_Known,
                               P2PHashSets a_HashType,
                               Opcodes a_Opcode,
                               HashQueryKinds a_Kind,
                               byte[] a_MasterHash,
                               uint[] a_HashIndices,
                               uint a_StartingIndex,
                               uint a_Count,
                               uint a_AICHTrgtPartIndex) :
                          base(a_ContentMngr.HandleHashQuery)
        {
            f_Known = a_Known;
            f_HashType = a_HashType;
            f_Opcode = a_Opcode;
            f_Kind = a_Kind;
            f_MasterHash = a_MasterHash;
            f_HashIndices = a_HashIndices;
            f_StartingIndex = a_StartingIndex;
            f_Count = a_Count;
            f_AICHTrgtPartIndex = a_AICHTrgtPartIndex;

            switch (a_Kind)
            {
                case HashQueryKinds.Indices:
                    f_EstimatedResultSize = CalcSize(f_HashType, (uint)f_HashIndices.Length);
                    break;

                case HashQueryKinds.Range:
                    f_EstimatedResultSize = CalcSize(f_HashType, f_Count);
                    break;

                case HashQueryKinds.AICHPartRecoveryData:
                    f_EstimatedResultSize = (16 * 1024);
                    break;

                default: throw(new ApplicationException());
            }
        }

        public void Fail()
        {
            Fail(false);
        }

        public void Fail(bool a_MalformedInput)
        {
            // Used by the content manager.

            f_MalformedInput = a_MalformedInput;

            SetState(CookieStates.Failure);
        }

        public void Succeed(byte[] a_Data)
        {
            // Used by the content manager.

            f_Data = a_Data;

            SetState(CookieStates.Success);
        }

        private static ulong CalcSize(P2PHashSets a_HashType,
                                      uint a_HashCount)
        {
            ulong l_Size = a_HashCount;

            switch (a_HashType)
            {
                case P2PHashSets.eD2kBitmap:
                    // Nothing to do.
                    break;

                case P2PHashSets.eD2kHashSet:
                    l_Size *= MD_4.c_HashByteSize;
                    break;

                case P2PHashSets.AICHTree:
                    l_Size *= SHA_1.c_HashByteSize;
                    break;

                case P2PHashSets.TAICHTree:
                case P2PHashSets.Upper8MiBTAICHTree:
                    l_Size *= Tiger.c_HashByteSize;
                    break;

                case P2PHashSets.ClientNativeBitmap:
                    l_Size += sizeof(uint);
                    break;

                default: throw (new ApplicationException());
            }

            return l_Size;
        }
    }

    class FileDescriptorCookie : GeneralAPCCookie
    {
        public readonly KnownFile f_Known;
        private byte[] f_Data;

        public byte[] p_Data
        {
            get { return f_Data; }
        }

        public FileDescriptorCookie(ContentManager a_ContentMngr,
                                    KnownFile a_Known) :
                               base(a_ContentMngr.BuildFileDescriptor)
        {
            f_Known = a_Known;
        }

        public void Fail()
        {
            SetState(CookieStates.Failure);
        }

        public void Succeed(byte[] a_Data)
        {
            f_Data = a_Data;

            SetState(CookieStates.Success);
        }
    }

    class QueryUsersCookie : APCCookie
    {
        private UserNum[] f_Users;

        public UserNum[] p_Users
        {
            get { return f_Users; }
        }

        public void Fail()
        {
            // Called by the IncompleFile class.

            SetState(CookieStates.Failure);
        }

        public void Succeed(UserNum[] a_Users)
        {
            // Called by the IncompleFile class.

            f_Users = a_Users;

            SetState(CookieStates.Success);
        }
    }

    [Flags]
    enum ShareKinds : byte
    {
        None = 0,
        eD2k_on_I2P = 0x1,

        FriendsOnly = 0x80,
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct PathNormalizer
    {
        // NOTE: This has nothing to do with Unicode normalization.

        private readonly StringBuilder f_Bld;

        private PathNormalizer(StringBuilder a_Bld)
        {
            f_Bld = a_Bld;
        }

        public static void Create(out PathNormalizer a_Dest)
        {
            a_Dest = new PathNormalizer(new StringBuilder());
        }

        public string Normalize(string a_Path)
        {
            return Normalize(a_Path, false);
        }

        public string Normalize(string a_Path,
                                bool a_NeedPreprocessing)
        {
            f_Bld.Length = 0;
            f_Bld.Append(a_Path);

            if (true == a_NeedPreprocessing)
            {
                RemovePossibleQuotes(f_Bld);
                MergeDirSeparators(f_Bld);
            }

            Normalize(f_Bld);

            string l_NewPath = f_Bld.ToString();
            f_Bld.Length = 0;

            return l_NewPath;
        }

        public static bool IsNormalized(string a_Path)
        {
            return (false == a_Path.Contains(Path.AltDirectorySeparatorChar.ToString()) &&
                    0 < a_Path.Length &&
                    Path.DirectorySeparatorChar != a_Path[a_Path.Length - 1]);
        }

        public static string NormalizeStatic(string a_Path,
                                             bool a_NeedPreprocessing)
        {
            PathNormalizer l_PathNormalizer;
            Create(out l_PathNormalizer);

            return l_PathNormalizer.Normalize(a_Path, a_NeedPreprocessing);
        }

        private static void Normalize(StringBuilder a_Bld)
        {
            // We don't trim paths here, because not every OS might do
            // this/do it the same way.

            a_Bld.Replace(Path.AltDirectorySeparatorChar, Path.DirectorySeparatorChar);
            int l_Len = a_Bld.Length;

            if (0 < l_Len &&
                Path.DirectorySeparatorChar == a_Bld[l_Len - 1])
            {
                a_Bld.Length = (l_Len - 1);
            }
        }

        private static void RemovePossibleQuotes(StringBuilder a_Bld)
        {
            string l_Path = a_Bld.ToString().Trim();
            int l_Len = l_Path.Length;

            if (2 <= l_Len &&
                '"' == l_Path[0] &&
                '"' == l_Path[l_Len - 1])
            {
                a_Bld.Length = 0;
                a_Bld.Append(l_Path, 1, l_Len - 2);
            }
        }

        private static void MergeDirSeparators(StringBuilder a_Bld)
        {
            ReplaceDoubleCharWithSingleChar(a_Bld, Path.DirectorySeparatorChar);
            ReplaceDoubleCharWithSingleChar(a_Bld, Path.AltDirectorySeparatorChar);
        }

        private static void ReplaceDoubleCharWithSingleChar(StringBuilder a_Bld,
                                                            char a_Char)
        {
            a_Bld.Replace(new string(a_Char, 2), new string(a_Char, 1));
        }
    }

    class SharedDirectories
    {
        public const ShareKinds c_DefaultShareKind = ShareKinds.eD2k_on_I2P;
        public const byte c_MaxSharedDirs = 128;
        public const int c_NoTrgtDirIdx = (-1);
        private const ushort c_Magic = 0xF729;

        public delegate void FileChangedHandler(string a_Path);
        public delegate void FileNameChanged(string a_OldPath,
                                             string a_NewPath);

        private readonly Dir[] f_Dirs;
        private readonly int f_ComplTrgtDirIdx;

        private uint f_ScanGen;

        // For browsing.
        private int f_DirIdx;
        private FileFinder f_Finder;

        public uint p_ScanGen
        {
            get { return f_ScanGen; }
        }

        public SharedDirectories() :
                            this(new Dir[0])
        {
        }

        public SharedDirectories(Dir[] a_Dirs) :
                            this(a_Dirs, c_NoTrgtDirIdx)
        {
        }

        public SharedDirectories(Dir[] a_Dirs,
                                 int a_ComplTrgtDirIdx)
        {
            // a_CompletionTrgtDirIdx may be subzero to indicate no
            // dir.

            f_ScanGen = 1;
            f_Dirs = a_Dirs;
            f_ComplTrgtDirIdx = a_ComplTrgtDirIdx;
            FileFinder.Create(out f_Finder);

            ResetFileBrowser();
        }

        public static SharedDirectories FromBuffer(BufferIO a_Buffer)
        {
            // Returns null on error.

            if (c_Magic != a_Buffer.ReadUInt16LE())
            {
                return null;
            }

            int l_Count = (int)a_Buffer.ReadUInt32LE();
            int l_ComplTrgtDirIdx = (int)a_Buffer.ReadUInt32LE();

            if (c_MaxSharedDirs < (uint)l_Count ||
                l_Count <= l_ComplTrgtDirIdx)
            {
                return null;
            }

            Dir[] l_Dirs = new Dir[l_Count];

            for (int i = 0; i < l_Count; i++)
            {
                if (false == Dir.FromBuffer(a_Buffer, out l_Dirs[i]))
                {
                    return null;
                }
            }

            return new SharedDirectories(l_Dirs, l_ComplTrgtDirIdx);
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt16LE(c_Magic);

            int l_Count = f_Dirs.Length;
            a_Buffer.WriteUInt32LE((uint)l_Count);
            a_Buffer.WriteUInt32LE((uint)f_ComplTrgtDirIdx);

            for (int i = 0; i < l_Count; i++)
            {
                f_Dirs[i].ToBuffer(a_Buffer);
            }
        }

        public SharedDirectories RecreateWithEnforcedIntegrity()
        {
            FlexArray<Dir> l_FlexDirs;
            FlexArray<Dir>.Create(this.f_Dirs.Length, out l_FlexDirs);

            bool l_EntriesDroppedDueToLimit;
            Dir.CopyWithEnforcedIntegrity(this.f_Dirs, ref l_FlexDirs, out l_EntriesDroppedDueToLimit);

            Dir[] l_Dirs = l_FlexDirs.ToArray(true);
            int l_ComplTrgtDirIdx = FindCompletionTargetPath(l_Dirs, this.GetCompletionTargetPath());

            return new SharedDirectories(l_Dirs, l_ComplTrgtDirIdx);
        }

        public Dir[] Expose(out int a_ComplTrgtDirIdx)
        {
            a_ComplTrgtDirIdx = f_ComplTrgtDirIdx;

            return f_Dirs;
        }

        public void SetScanGeneration(uint a_ScanGen)
        {
            f_ScanGen = a_ScanGen;
        }

        public void IncScanGeneration()
        {
            if (0 == ++f_ScanGen)
            {
                f_ScanGen = 1;
            }
        }

        public string GetCompletionTargetPath()
        {
            // Return can be null.

            if (0 > f_ComplTrgtDirIdx)
            {
                return null;
            }

            return f_Dirs[f_ComplTrgtDirIdx].f_Path;
        }

        public void ResetFileBrowser()
        {
            f_DirIdx = (-1);
        }

        [Conditional("DEBUG")]
        public void DebugAssertIntegrity()
        {
            if (false == Dir.CheckArrayIntegrity(f_Dirs) ||
                f_Dirs.Length <= f_ComplTrgtDirIdx)
            {
                throw (new ApplicationException());
            }
        }

        public bool ValueEquals(SharedDirectories a_Other)
        {
            // Compares are done case sensitive in any case.

            int l_Count = this.f_Dirs.Length;
            Dir[] l_OtherDirs = a_Other.f_Dirs;

            if (l_OtherDirs.Length != l_Count)
            {
                return false;
            }

            if (0 > this.f_ComplTrgtDirIdx)
            {
                if (0 <= a_Other.f_ComplTrgtDirIdx)
                {
                    return false;
                }
            }
            else
            {
                if (0 > a_Other.f_ComplTrgtDirIdx ||
                    false == this.f_Dirs[this.f_ComplTrgtDirIdx].ValueEqualsMindCase(ref l_OtherDirs[a_Other.f_ComplTrgtDirIdx]))
                {
                    return false;
                }
            }

            List<int> l_IndicesDone = new List<int>(l_Count);
            for (int i = 0; i < l_Count; i++)
            {
                bool l_Hit = false;

                for (int k = 0; k < l_Count; k++)
                {
                    if (true == this.f_Dirs[i].ValueEqualsMindCase(ref l_OtherDirs[k]))
                    {
                        if (true == l_IndicesDone.Contains(k))
                        {
                            continue;
                        }

                        l_IndicesDone.Add(k);
                        l_Hit = true;

                        break;
                    }
                }

                if (false == l_Hit)
                {
                    return false;
                }
            }

            return true;
        }

        public FileSystemWatcher[] CreateWatchers(FileChangedHandler a_FileAppearedOrChanged,
                                                  FileChangedHandler a_FileDisappeared,
                                                  FileNameChanged a_FileRenamed,
                                                  ArglessHandler a_NeedCompleteScan)
        {
            // Return can be null (in case there are no dirs or the dirs
            // cannot be watched for some reason).

            int l_DirCount = f_Dirs.Length;

            if (0 == l_DirCount)
            {
                return null;
            }

            List<FileSystemWatcher> l_Ls = new List<FileSystemWatcher>(2 * l_DirCount);
            Redirector l_Red = new Redirector(a_FileAppearedOrChanged, a_FileDisappeared, a_FileRenamed, a_NeedCompleteScan);

            for (int i = 0; i < l_DirCount; i++)
            {
                string l_WatchedPath = f_Dirs[i].f_Path;
                string l_ParentDir;
                string l_DirName = LongPath.SplitFullPath(l_WatchedPath, out l_ParentDir);

                // Register two different file system watchers for each dir.

                FileSystemWatcher l_Watcher = new FileSystemWatcher();
                l_Watcher.Changed += l_Red.FileAppearedOrChanged;
                l_Watcher.Created += l_Red.FileAppearedOrChanged;
                l_Watcher.Deleted += l_Red.FileDisappeared;
                l_Watcher.Renamed += l_Red.FileRenamed;
                l_Watcher.Error += l_Red.WatcherBufferOverflow;
                l_Watcher.NotifyFilter = (NotifyFilters.FileName | NotifyFilters.Size);
                l_Watcher.IncludeSubdirectories = f_Dirs[i].f_Recursive;

                try
                {
                    // Throws an argument exception if the dir doesn't
                    // exist. We can never be sure that it wasn't
                    // removed in the meantime.
                    l_Watcher.Path = l_WatchedPath;
                    l_Watcher.EnableRaisingEvents = true;
                }
                catch
                {
                    l_Watcher.Dispose();
                    l_Watcher = null;
                }

                if (null != l_Watcher)
                {
                    l_Ls.Add(l_Watcher);
                }

                FileSystemWatcher l_DirGuard = new FileSystemWatcher();
                l_DirGuard.Created += l_Red.SharedDirAppeared;
                l_DirGuard.Deleted += l_Red.SharedDirDisappeared;
                l_DirGuard.Renamed += l_Red.SharedDirRenamed;
                l_DirGuard.Error += l_Red.WatcherBufferOverflow;
                l_DirGuard.Filter = l_DirName;
                l_DirGuard.NotifyFilter = NotifyFilters.DirectoryName;
                l_DirGuard.IncludeSubdirectories = false;

                try
                {
                    l_DirGuard.Path = l_ParentDir;
                    l_DirGuard.EnableRaisingEvents = true;
                }
                catch
                {
                    l_DirGuard.Dispose();
                    l_DirGuard = null;
                }

                if (null != l_DirGuard)
                {
                    l_Ls.Add(l_DirGuard);
                }
            }

            return (0 == l_Ls.Count ? null : l_Ls.ToArray());
        }

        public int FindNextFiles(string[] a_Dest,
                                 int a_MaxCount)
        {
            // Returns the number of elements filled.

            if (0 > f_DirIdx)
            {
                if (0 == f_Dirs.Length)
                {
                    return 0;
                }

                f_DirIdx = 0;
                f_Finder.Setup(ref f_Dirs[0]);
            }

            int l_TotalCount = 0;
            int l_RemCount = a_MaxCount;

            while (0 < l_RemCount)
            {
                int l_Count = f_Finder.FindNextFiles(a_Dest, l_TotalCount, l_RemCount);

                if (0 == l_Count)
                {
                    if (f_Dirs.Length > (f_DirIdx + 1))
                    {
                        f_DirIdx++;
                        f_Finder.Setup(ref f_Dirs[f_DirIdx]);

                        continue;
                    }

                    break;
                }

                l_TotalCount += l_Count;
                l_RemCount -= l_Count;
            }

            return l_TotalCount;
        }

        public ShareKinds GetShareKind(string a_ContainingFolder)
        {
            for (int i = (f_Dirs.Length - 1); 0 <= i; i--)
            {
                if (true == f_Dirs[i].IsFolderShared(a_ContainingFolder))
                {
                    // Since a folder can only be shared through no
                    // more than one entry of f_Dirs, we can abort as
                    // soon as we have found the one.

                    return f_Dirs[i].f_ShareKind;
                }
            }

            return ShareKinds.None;
        }

        public static void DestroyWatchers(FileSystemWatcher[] a_Watchers)
        {
            foreach (FileSystemWatcher l_Watcher in a_Watchers)
            {
                l_Watcher.Dispose();
            }
        }

        public static int FindCompletionTargetPath(Dir[] a_Dirs,
                                                   string a_Path)
        {
            // a_Path may be null. If no such path is found, a
            // negative value is returned.

            if (null != a_Path)
            {
                for (int i = (a_Dirs.Length - 1); 0 <= i; i--)
                {
                    if (true == FileNames.AreNamesIdentical(a_Dirs[i].f_Path, a_Path))
                    {
                        return i;
                    }
                }
            }

            return c_NoTrgtDirIdx;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Dir
        {
            public const ushort c_MaxPathChars = (ushort.MaxValue / sizeof(char));

            public readonly string f_Path;
            public bool f_Recursive;
            public ShareKinds f_ShareKind;

            public bool p_FriendsOnly
            {
                get { return (0 != (ShareKinds.FriendsOnly & f_ShareKind)); }
            }

            public Dir(string a_Path,
                       bool a_Recursive,
                       ShareKinds a_ShareKind)
            {
                f_Path = a_Path;
                f_Recursive = a_Recursive;
                f_ShareKind = a_ShareKind;
            }

            public static bool FromBuffer(BufferIO a_Buffer,
                                          out Dir a_Dir)
            {
                string l_Path = Software.ReadAbsPathUTF16LEStringEncLen(a_Buffer);

                if (c_MaxPathChars < l_Path.Length)
                {
                    a_Dir = new Dir();

                    return false;
                }

                bool l_Recursive = a_Buffer.ReadBool();
                ShareKinds l_ShareKind = (ShareKinds)a_Buffer.ReadByte();

                a_Dir = new Dir(l_Path, l_Recursive, l_ShareKind);

                return true;
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteUTF16LEStringEncLen(f_Path);
                a_Buffer.WriteBool(f_Recursive);
                a_Buffer.WriteByte((byte)f_ShareKind);
            }

            public bool ValueEqualsMindCase(ref Dir a_Dir)
            {
                return (true == FileNames.AreNamesIdenticalMindCase(this.f_Path, a_Dir.f_Path) &&
                        true == NonStringFieldsEqual(ref a_Dir));
            }

            private bool NonStringFieldsEqual(ref Dir a_Dir)
            {
                return (this.f_Recursive == a_Dir.f_Recursive &&
                        this.f_ShareKind == a_Dir.f_ShareKind);
            }

            public bool IsFolderShared(string a_Folder)
            {
                if (true == FileNames.AreNamesIdentical(f_Path, a_Folder))
                {
                    return true;
                }

                return (true == this.f_Recursive &&
                        true == FileNames.IsSubdir(this.f_Path, a_Folder));
            }

            public static bool DoOverlap(ref Dir a_A,
                                         ref Dir a_B)
            {
                return (CongruenceKinds.NoOverlap != CheckCongruence(ref a_A, ref a_B));
            }

            public static CongruenceKinds CheckCongruence(ref Dir a_A,
                                                          ref Dir a_B)
            {
                // By the means of Unicode equivalence, a string with
                // more code points can be contained in a shorter
                // string, so we can't really optimize this based on
                // a length check.

                if (true == FileNames.AreNamesIdentical(a_A.f_Path, a_B.f_Path))
                {
                    if (false == a_A.f_Recursive)
                    {
                        return (false == a_B.f_Recursive ? CongruenceKinds.Equality : CongruenceKinds.AContainedInB);
                    }

                    return (false == a_B.f_Recursive ? CongruenceKinds.BContainedInA : CongruenceKinds.Equality);
                }

                if (true == a_A.f_Recursive &&
                    true == FileNames.IsSubdir(a_A.f_Path, a_B.f_Path))
                {
                    return CongruenceKinds.BContainedInA;
                }

                if (true == a_B.f_Recursive &&
                    true == FileNames.IsSubdir(a_B.f_Path, a_A.f_Path))
                {
                    return CongruenceKinds.AContainedInB;
                }

                return CongruenceKinds.NoOverlap;
            }

            public static void CopyWithEnforcedIntegrity(Dir[] a_Src,
                                                         ref FlexArray<Dir> a_Dest,
                                                         out bool a_EntriesDroppedDueToLimit)
            {
                CopyWithEnforcedIntegrity(a_Src, 0, a_Src.Length, ref a_Dest, out a_EntriesDroppedDueToLimit);
            }

            public static void CopyWithEnforcedIntegrity(Dir[] a_Src,
                                                         int a_Index,
                                                         int a_Length,
                                                         ref FlexArray<Dir> a_Dest,
                                                         out bool a_EntriesDroppedDueToLimit)
            {
                // The content of a_Src may be inconsistent and
                // a_Length may be greater than c_MaxSharedDirs.

                a_EntriesDroppedDueToLimit = false;
                int l_Count = a_Dest.p_Count;

                Dir[] l_Arr = a_Dest.p_Arr;
                for (int i = a_Index, lim = (a_Index + a_Length); i < lim; i++)
                {
                    bool l_AddEvenIfLimitReached = false;

                    for (int k = 0; k < l_Count; k++)
                    {
                        CongruenceKinds l_Kind = CheckCongruence(ref a_Src[i], ref l_Arr[k]);

                        switch (l_Kind)
                        {
                            case CongruenceKinds.NoOverlap:
                                break;

                            case CongruenceKinds.Equality:
                            {
                                // Try to preserve entries with the
                                // friends only flag (be conservative).

                                if (false == a_Src[i].p_FriendsOnly)
                                {
                                    goto case CongruenceKinds.AContainedInB;
                                }

                                goto case CongruenceKinds.BContainedInA;
                            }

                            case CongruenceKinds.AContainedInB:
                                goto SKIP;

                            case CongruenceKinds.BContainedInA:
                            {
                                a_Dest.DeleteAt(k);
                                l_Count--;
                                l_Arr = a_Dest.p_Arr;

                                l_AddEvenIfLimitReached = true;

                                break;
                            }

                            default: throw (new ApplicationException());
                        }
                    }

                    if (false == l_AddEvenIfLimitReached &&
                        c_MaxSharedDirs <= l_Count)
                    {
                        a_EntriesDroppedDueToLimit = true;

                        continue;
                    }

                    int l_DestIdx = a_Dest.ReserveAtEnd();
                    l_Count++;
                    l_Arr = a_Dest.p_Arr;

                    l_Arr[l_DestIdx] = a_Src[i];

                    SKIP:;
                }
            }

            public static bool CheckArrayIntegrity(Dir[] a_Dirs)
            {
                int l_Count = a_Dirs.Length;
                for (int i = 1; i < l_Count; i++)
                {
                    for (int k = 0; k < i; k++)
                    {
                        if (true == Dir.DoOverlap(ref a_Dirs[i], ref a_Dirs[k]))
                        {
                            return false;
                        }
                    }
                }

                return true;
            }

            public enum CongruenceKinds : byte
            {
                NoOverlap = 0,
                // Equality indicates same path and recursion state.
                Equality = 1,
                // In case B is recursive:
                AContainedInB = 2,
                // In case A is recursive:
                BContainedInA = 3,
            }
        }

        private class Redirector
        {
            // This class helps dispatching file system events.

            private readonly FileChangedHandler f_FileAppearedOrChanged;
            private readonly FileChangedHandler f_FileDisappeared;
            private readonly FileNameChanged f_FileRenamed;
            private readonly ArglessHandler f_NeedCompleteScan;

            public Redirector(FileChangedHandler a_FileAppearedOrChanged,
                              FileChangedHandler a_FileDisappeared,
                              FileNameChanged a_FileRenamed,
                              ArglessHandler a_NeedCompleteScan)
            {
                f_FileAppearedOrChanged = a_FileAppearedOrChanged;
                f_FileDisappeared = a_FileDisappeared;
                f_FileRenamed = a_FileRenamed;
                f_NeedCompleteScan = a_NeedCompleteScan;
            }

            // NOTE: The framework can give us paths with shortened
            // element names (prepended tilde) if the path is very
            // long.

            public void FileAppearedOrChanged(object sender, FileSystemEventArgs e)
            {
                f_NeedCompleteScan();

                CallFileChangedHandler(e, f_FileAppearedOrChanged);
            }

            public void FileDisappeared(object sender, FileSystemEventArgs e)
            {
                CallFileChangedHandler(e, f_FileDisappeared);
            }

            private void CallFileChangedHandler(FileSystemEventArgs e,
                                                FileChangedHandler a_Handler)
            {
                string l_Path;

                try
                {
                    // This can throw a PathTooLong exception, although
                    // it's not documented.

                    l_Path = e.FullPath;
                }
                catch
                {
                    f_NeedCompleteScan();

                    return;
                }

                a_Handler(l_Path);
            }

            public void FileRenamed(object sender, RenamedEventArgs e)
            {
                string l_OldName;
                string l_NewName;

                try
                {
                    // This can throw a PathTooLong exception, although
                    // it's not documented.

                    l_OldName = e.OldFullPath;
                    l_NewName = e.FullPath;
                }
                catch
                {
                    f_NeedCompleteScan();

                    return;
                }

                f_FileRenamed(l_OldName, l_NewName);
            }

            public void SharedDirAppeared(object sender, FileSystemEventArgs e)
            {
                f_NeedCompleteScan();
            }

            public void SharedDirDisappeared(object sender, FileSystemEventArgs e)
            {
                f_NeedCompleteScan();
            }

            public void SharedDirRenamed(object sender, RenamedEventArgs e)
            {
                f_NeedCompleteScan();
            }

            public void WatcherBufferOverflow(object sender, ErrorEventArgs e)
            {
                f_NeedCompleteScan();
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct FileFinder
        {
            // The root folder is not counted, so a limit of one will
            // allow branching into the first subfolder level.
            private const byte c_MaxRecursionDepth = 128;

            private static readonly string[] s_Empty;

            private FlexArray<BrowseableDir> f_Stack;
            private bool f_Recursive;

            static FileFinder()
            {
                s_Empty = new string[0];
            }

            public static void Create(out FileFinder a_Finder)
            {
                a_Finder = new FileFinder();
                FlexArray<BrowseableDir>.Create(out a_Finder.f_Stack);
                a_Finder.f_Recursive = false;
            }

            public void Setup(ref Dir a_Dir)
            {
                Setup(a_Dir.f_Path, a_Dir.f_Recursive);
            }

            public void Setup(string a_Path,
                              bool a_Recursive)
            {
                f_Stack.Clear();
                f_Recursive = a_Recursive;

                Enter(a_Path);
            }

            private bool Enter(string a_Path)
            {
                // Returns true if a new dir was put onto the stack.
                // This routine modifies f_Stack.

                if (c_MaxRecursionDepth < f_Stack.p_Count)
                {
                    return false;
                }

                string[] l_DirNames = null;
                string[] l_FileNames = null;

                try
                {
                    if (true == f_Recursive)
                    {
                        l_DirNames = Directory.GetDirectories(a_Path);
                    }

                    l_FileNames = Directory.GetFiles(a_Path);
                }
                catch (ArgumentNullException l_Exc)
                {
                    throw (l_Exc);
                }
                catch
                {
                }

                FileNames.FilterPathArray(ref l_DirNames);
                FileNames.FilterPathArray(ref l_FileNames);

                if (null == l_DirNames)
                {
                    if (null == l_FileNames)
                    {
                        return false;
                    }

                    l_DirNames = s_Empty;
                }
                else if (null == l_FileNames)
                {
                    l_FileNames = s_Empty;
                }

                int l_Idx = f_Stack.ReserveAtEnd();
                f_Stack.p_Arr[l_Idx] = new BrowseableDir(l_DirNames, l_FileNames);

                return true;
            }

            public int FindNextFiles(string[] a_Dest,
                                     int a_Index,
                                     int a_MaxCount)
            {
                // Returns the number of elements filled.

                int l_StackCount = f_Stack.p_Count;

                if (0 >= l_StackCount ||
                    0 >= a_MaxCount)
                {
                    // End reached or nothing to do.

                    return 0;
                }

                int l_Idx = a_Index;
                int l_ToGo = a_MaxCount;
                int l_StackIdx = (l_StackCount - 1);
                BrowseableDir[] l_Arr = f_Stack.p_Arr;

                while (true)
                {
                    // Find next files.
                    int l_Progress = l_Arr[l_StackIdx].CopyFiles(a_Dest, l_Idx, l_ToGo);

                    if (0 < l_Progress)
                    {
                        l_Idx += l_Progress;
                        l_ToGo -= l_Progress;

                        if (0 == l_ToGo)
                        {
                            // Max count reached.

                            return (l_Idx - a_Index);
                        }
                    }

                    // Find next dir.
                    while (true)
                    {
                        string l_NextPath = l_Arr[l_StackIdx].GetNextDir();

                        if (null == l_NextPath)
                        {
                            if (0 == l_StackIdx)
                            {
                                // End reached.

                                return (l_Idx - a_Index);
                            }

                            f_Stack.DeleteAt(l_StackIdx--);

                            continue;
                        }

                        if (true == Enter(l_NextPath))
                        {
                            // Branching into the new dir.

                            l_StackIdx++;
                            l_Arr = f_Stack.p_Arr;

                            break;
                        }
                    }
                }
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct BrowseableDir
            {
                private readonly string[] f_Dirs;
                private readonly string[] f_Files;

                private int f_DirIdx;
                private int f_FileIdx;

                public BrowseableDir(string[] a_Dirs,
                                     string[] a_Files)
                {
                    f_Dirs = a_Dirs;
                    f_Files = a_Files;
                    f_DirIdx = 0;
                    f_FileIdx = 0;
                }

                public string GetNextDir()
                {
                    // Returns null if there are no dirs left.

                    if (f_Dirs.Length <= f_DirIdx)
                    {
                        return null;
                    }

                    return f_Dirs[f_DirIdx++];
                }

                public int CopyFiles(string[] a_Dest,
                                     int a_Idx,
                                     int a_MaxCount)
                {
                    // Returns the number of file names copied.

                    if (f_Files.Length <= f_FileIdx)
                    {
                        return 0;
                    }

                    int l_Count = Math.Min(f_Files.Length - f_FileIdx, a_MaxCount);
                    Array.Copy(f_Files, f_FileIdx, a_Dest, a_Idx, l_Count);
                    f_FileIdx += l_Count;

                    return l_Count;
                }
            }
        }
    }

    class ContentManager
    {
        private const string c_ScanThreadName = "Scan";
        private const string c_HashThreadName = "Hash";
        private const string c_IdentThreadName = "Ident";
        private const string c_ReadThreadName = "Read";
        private const string c_IncomplThreadsNameBeginning = "Incompl_";

        private const uint c_StoredFilesMagic = 0x7D46D0B0;
        private const uint c_StoredFilesHeaderSize = 512;

        // In seconds.
        public const uint c_ScanDiskInterval = (7 * Time.c_SecsPerMinute);
        public const uint c_RescanForFinishedFilesInterval = ((3 * Time.c_SecsPerMinute) + Time.c_SecsPerHalfMinute);
#if MONO
        // Mono can cause hardware noise when checking, so check less
        // often.
        public const uint c_CheckFreeDiskSpaceInterval = 120;
#else
        public const uint c_CheckFreeDiskSpaceInterval = 3;
#endif
        private const uint c_CheckFreeDiskSpaceHashSetSizeThreshold = (8 * 1024 * 1024);

        public const ulong c_MaxSharedFileSize = KnownFile.HashElmLayout.c_MaxFileSize;
        public const ulong c_MinAvailDiskSpace = (512 * 1024 * 1024);
        public const ulong c_RootElementKey = 0;
        private const byte c_StartingFileNum = 1;
        private const byte c_MaxFinishedIncomplFilesCount = 32;
        private const uint c_DefaultBufferSize = (256 * 1024);
        private const uint c_DefaultHashItemSize = c_DefaultBufferSize;
        private const uint c_DefaultIdentItemSize = (4 * c_DefaultBufferSize);

        private readonly Instance f_Inst;
        private APCQueue<GeneralAPCCookie> f_ScanQueue;
        private APCQueue<GeneralAPCCookie> f_HashQueue;
        private APCQueue<GeneralAPCCookie> f_IdentQueue;
        private APCQueue<ReadCookie> f_ReadComplQueue;
        private readonly object f_Lock;

        // The suggested number of workers for processing operations
        // on incomplete files.
        private readonly byte f_IncomplWorkers;
        private readonly Resource f_IncomplMemResource;
        private readonly HardMemory f_IncomplMem;

#region f_LockDomain

        // Under the f_Lock lock.
        private readonly Dictionary<FileNum, KnownFile> f_KnownFilesByFileNum;
        private readonly MultiMap<ulong, KnownFile> f_KnownFilesBySize;
        private readonly MultiMap<FileID, KnownFile> f_KnownFilesByeD2kFileID;
        // Known files that are not yet linked to a local file and
        // can't therefore be hashed.
        private readonly Dictionary<KnownFile, KnownFile> f_UnlinkedKnw;
        // Known files that are linked, but haven't had all of their
        // hash sets created yet (KnownFile.HasTrustedHashes()).
        private readonly Dictionary<KnownFile, KnownFile> f_UnhashedKnw;
        // Known files that are linked and hashed.
        private readonly Dictionary<KnownFile, KnownFile> f_HashedKnw;

        // Under the f_Lock lock.
        private readonly FileRegistry f_FileReg;
        // Local files that aren't linked to a know file yet and are
        // not ready for hashing ('quick' and 'total' hash).
        private readonly Dictionary<LocalFile, LocalFile> f_UnprocessedLoc;
        // Same as f_UnprocessedLoc, but those files are ready for
        // hashing.
        private readonly Dictionary<LocalFile, LocalFile> f_UnhashedLoc;
        // Local files that are linked to a known file.
        private readonly Dictionary<LocalFile, LocalFile> f_RecognizedLoc;

        // Under the f_Lock lock.
        private IOPrefs f_IOPrefs;
        private ScanState f_ScanState;
        private Time f_NextAutoScanTime;
        // f_Watchers can be null.
        private FileSystemWatcher[] f_Watchers;
        private ulong f_NextFileNum;
        // The following two strings can be null.
        private string f_DontOpenExtractionFileName;
        private string f_DontOpenRemovalFileName;
        private string f_DontOpenMoveSrcFileName;
        private string f_DontOpenMoveDestFileName;

#endregion

        private uint f_HashingIdleTicks;

        // Self locked fields.
        private ListHead<IncompleteFile> f_IncomplFilesList;
        private readonly Dictionary<KnownFile, Queue<IncomplOperationContext>> f_IncomplOps;
        private readonly List<KnownFile> f_IncomplOrder;
        private uint f_IncomplWorkersSpinning;

        // Exclusive to the hash worker.
        private HashingContext f_HashCntxt;
        private FlexArray<KnownFile> f_FinishedIncompl;
        private Time f_NextRescanForFinishedFilesTime;
        // f_RecycleComplFilesCntxt can be null.
        private MoveCompleteFilesToRecycleBinContext f_RecycleComplFilesCntxt;
        private Time f_LastHashingTime;

        // Exclusive to the ident worker.
        private readonly SHA_256 f_IdentWorkerSHA256;
        private readonly Queue<MultiFileImportContext> f_ImportContexts;
        private bool f_IdentWorkerIoBusy;
        private Time f_LastIdentTotalHashingTime;

        private bool f_RunScanWorker;
        private bool f_RunHashWorker;
        private bool f_RunIdentWorker;
        private bool f_RunReadComplWorker;
        private bool f_RunIncomplWorker;
        private ThreadSafeCounter32 f_WatchersThreadsPresent;

        private readonly MonitorEvent f_ScanWorkerFinished;
        private readonly MonitorEvent f_HashWorkerFinished;
        private readonly MonitorEvent f_IdentWorkerFinished;
        private readonly MonitorEvent f_ReadComplWorkerFinished;
        private readonly MonitorEvent[] f_IncomplWorkerFinished;

        // Updated by the identification worker.
        private bool f_CriticalDiskSpaceLevelReached;

        public bool p_CriticalDiskSpaceLevelReachedVolatile
        {
            get { return Util.VolatileRead(ref f_CriticalDiskSpaceLevelReached); }
        }

        public ContentManager(Instance a_Inst,
                              HardMemory a_IncomplMem,
                              ref IOPrefs a_IOPrefs)
        {
            f_Inst = a_Inst;
            APCQueue<GeneralAPCCookie>.Create(out f_ScanQueue);
            APCQueue<GeneralAPCCookie>.Create(out f_HashQueue);
            APCQueue<GeneralAPCCookie>.Create(out f_IdentQueue);
            APCQueue<ReadCookie>.Create(out f_ReadComplQueue);
            f_Lock = new object();
            f_IncomplWorkers = (byte)CalcIncomplWorkerCount();
            f_IncomplMemResource = Resource.Create(f_IncomplWorkers);
            f_IncomplMem = a_IncomplMem;
            f_IncomplMem.SetResource(f_IncomplMemResource, false);

            f_KnownFilesByFileNum = new Dictionary<FileNum, KnownFile>();
            MultiMap<ulong, KnownFile>.Create(out f_KnownFilesBySize);
            MultiMap<FileID, KnownFile>.Create(out f_KnownFilesByeD2kFileID);
            f_UnlinkedKnw = new Dictionary<KnownFile, KnownFile>();
            f_UnhashedKnw = new Dictionary<KnownFile, KnownFile>();
            f_HashedKnw = new Dictionary<KnownFile, KnownFile>();

            f_FileReg = new FileRegistry();
            f_UnprocessedLoc = new Dictionary<LocalFile, LocalFile>();
            f_UnhashedLoc = new Dictionary<LocalFile, LocalFile>();
            f_RecognizedLoc = new Dictionary<LocalFile, LocalFile>();

            f_IOPrefs = a_IOPrefs;
            Software.s_LastOutputDir = f_IOPrefs.f_SharedDirs.GetCompletionTargetPath();

            f_IncomplFilesList = new ListHead<IncompleteFile>();
            f_IncomplOps = new Dictionary<KnownFile, Queue<IncomplOperationContext>>();
            f_IncomplOrder = new List<KnownFile>();

            FlexArray<KnownFile>.Create(out f_FinishedIncompl);
            f_ImportContexts = new Queue<MultiFileImportContext>();
            f_IdentWorkerSHA256 = new SHA_256();

            f_ScanWorkerFinished = new MonitorEvent(true);
            f_HashWorkerFinished = new MonitorEvent(true);
            f_IdentWorkerFinished = new MonitorEvent(true);
            f_ReadComplWorkerFinished = new MonitorEvent(true);
            f_IncomplWorkerFinished = new MonitorEvent[f_IncomplWorkers];

            for (int i = 0; f_IncomplWorkers > i; i++)
            {
                f_IncomplWorkerFinished[i] = new MonitorEvent(true);
            }

            f_NextFileNum = c_StartingFileNum;
        }

        private static int CalcIncomplWorkerCount()
        {
            // Gives us between 1 and 6 threads.

            int l_Count = Math.Min(6, Math.Max(1, Environment.ProcessorCount - 1));

            return l_Count;
        }

        private void LockEnter()
        {
            Util.MonitorEnter(f_Lock);
        }

        private void LockLeave()
        {
            Util.MonitorLeave(f_Lock);
        }

        public void Start()
        {
            const uint l_RescanForFinishedFilesInitialWaitSecs = Time.c_SecsPerHalfMinute;

            f_ScanState = ScanState.NotStarted;
            f_NextAutoScanTime = new Time();
            f_NextRescanForFinishedFilesTime = f_Inst.GetTime().AddSeconds(l_RescanForFinishedFilesInitialWaitSecs);

            UpdatePreferences(ref f_IOPrefs);
            StartWorkers();
        }

        public void Shutdown(bool a_FindLeaks,
                             bool a_Compact,
                             HardMemory.ICompactProgressCollector a_Collector)
        {
            StopWorkers();

            if (true == a_FindLeaks)
            {
                FindLeaksAndFix();
            }

            Software.UpdateShutDownState(0x4000);

            StorageStates l_State = f_Inst.f_ContentMngr.SaveFiles();

            Software.UpdateShutDownState(0x8000);

            CloseAllLocalFiles();
            RemoveAllKnownFiles();

            Software.UpdateShutDownState(0x10000);

            f_IncomplMem.Close(a_Compact, a_Collector);

            Software.UpdateShutDownState(0x80000);
        }

        private void FindLeaksAndFix()
        {
            // The caller must have stopped the workers.

            Software.UpdateShutDownState(0x800);

            f_Inst.f_MemResource.AcquireExclusive();
            {
                HardMemory.LeakFinder l_Finder = new HardMemory.LeakFinder(f_Inst.f_Mem);

                l_Finder.Touch(new ElmKey(Instance.c_PrefsElmKey));
                l_Finder.Touch(new ElmKey(Instance.c_NodesElmKey));
                l_Finder.Touch(new ElmKey(Instance.c_UserCntxtsElmKey));
                l_Finder.Touch(new ElmKey(Instance.c_KnownFilesElmKey));
                l_Finder.Touch(new ElmKey(Instance.c_SearchCacheElmKey));

                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    if (false == l_Known.f_HashSetKey.IsNonzero())
                    {
                        continue;
                    }

                    if (false == IsValidHashElmKey(l_Known))
                    {
                        Util.DebugBreak();

                        l_Known.ClearHashSetKey();

                        continue;
                    }

                    HardMemory.LeakStates l_State = l_Finder.Touch(l_Known.f_HashSetKey);

                    if (HardMemory.LeakStates.NoError != l_State)
                    {
                        Util.DebugBreak();

                        l_Known.AdjustAllHashesSelfCalculatedFlag(false);
                        l_Known.ClearHashSetKey();
                    }
                }

                if (0 < l_Finder.p_RemainingCount)
                {
                    Util.DebugBreak();
                }

                l_Finder.CollectGarbage();
            }
            f_Inst.f_MemResource.ReleaseExclusive();

            Software.UpdateShutDownState(0x1000);

            bool l_Aborted = false;

            f_IncomplMemResource.AcquireExclusive();
            {
                HardMemory.LeakFinder l_Finder = new HardMemory.LeakFinder(f_IncomplMem);
                IncompleteFile.TempBuffers l_TempBuffers = new IncompleteFile.TempBuffers();

                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    if (true == l_Known.f_IncomplKey.IsNonzero())
                    {
                        if (false == IncompleteFile.TouchElements(f_IncomplMem, l_Known, ref l_Finder, l_TempBuffers))
                        {
                            // A possible io error prevents us from
                            // finishing.

                            l_Aborted = true;

                            break;
                        }
                        else if (true == l_Known.f_IncomplKey.IsZero())
                        {
                            l_Known.f_IncomplProgress = null;
                        }
                    }
                }

                if (false == l_Aborted)
                {
                    if (0 < l_Finder.p_RemainingCount)
                    {
                        Util.DebugBreak();
                    }

                    l_Finder.CollectGarbage();
                }
            }
            f_IncomplMemResource.ReleaseExclusive();

            Software.UpdateShutDownState(0x2000);
        }

        private void CloseAllLocalFiles()
        {
            // The caller must hold f_Lock. This routine doesn't update
            // the dictionaries of the known files involved.

            Time l_Time = f_Inst.GetTime();

            foreach (LocalFile l_Local in f_UnprocessedLoc.Values)
            {
                l_Local.Close();
            }

            f_UnprocessedLoc.Clear();

            foreach (LocalFile l_Local in f_UnhashedLoc.Values)
            {
                l_Local.Close();
            }

            f_UnhashedLoc.Clear();

            foreach (LocalFile l_Local in f_RecognizedLoc.Values)
            {
                l_Local.Close();
            }

            f_RecognizedLoc.Clear();
            f_FileReg.Clear();
        }

        private StorageStates SaveFiles()
        {
            return f_Inst.WriteToElement(new ElmKey(Instance.c_KnownFilesElmKey), SaveFilesCore);
        }

        private StorageStates SaveFilesCore(BufferIO a_Buffer)
        {
            Time l_Time = f_Inst.GetTime();
            uint l_Count = 0;
            BufferIO.CountWriter l_CountWtr;

            a_Buffer.WriteUInt32LE(c_StoredFilesMagic);
            BufferIO.CountWriter.InitAndMoveBuffer(a_Buffer, sizeof(uint), out l_CountWtr);
            a_Buffer.WriteUInt16LE((ushort)Path.VolumeSeparatorChar);
            a_Buffer.WriteUInt16LE((ushort)Path.DirectorySeparatorChar);
            a_Buffer.SetPosition(c_StoredFilesHeaderSize);

            Software.UpdateShutDownState(0x20000);

            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    l_Count++;
                    l_Known.ToBuffer(l_Time, a_Buffer);
                }
            }
            LockLeave();

            Software.UpdateShutDownState(0x40000);

            l_CountWtr.Finish(l_Count);

            return StorageStates.Success;
        }

        public StorageStates LoadFiles()
        {
            // This routine can be called from outside at any time
            // (also before Start() was called), but there may not be
            // files attached yet.

            return f_Inst.ReadFromElement(new ElmKey(Instance.c_KnownFilesElmKey), LoadFilesCore);
        }

        public StorageStates LoadFilesCore(BufferIO a_Buffer)
        {
            if (c_StoredFilesHeaderSize > a_Buffer.GetRemaining() ||
                c_StoredFilesMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            StorageStates l_State = StorageStates.Success;
            uint l_Count = a_Buffer.ReadUInt32LE();
            a_Buffer.SetPosition(c_StoredFilesHeaderSize);
            ulong l_HighestFileNum = c_StartingFileNum;

            LockEnter();
            {
                for (uint i = 0; i < l_Count; i++)
                {
                    KnownFile l_Known;

                    try
                    {
                        l_Known = KnownFile.FromBuffer(a_Buffer);
                    }
                    catch (BufferIO.InterpretBufferException)
                    {
                        l_Known = null;
                    }

                    if (null == l_Known ||
                        true == f_KnownFilesByFileNum.ContainsKey(l_Known.f_FileNum))
                    {
                        Util.DebugBreak();

                        RemoveAllKnownFiles();

                        l_State = StorageStates.Corruption;

                        break;
                    }

                    ReferenceKnownFile(l_Known, f_UnlinkedKnw);
                    l_HighestFileNum = Math.Max(l_HighestFileNum, l_Known.f_FileNum.f_Val);
                }

                if (StorageStates.Success == l_State)
                {
                    if (false == a_Buffer.EndReached())
                    {
                        Util.DebugBreak();

                        RemoveAllKnownFiles();

                        l_State = StorageStates.Corruption;
                    }
                    else
                    {
                        f_NextFileNum = (l_HighestFileNum + 1);

                        l_State = StorageStates.Success;
                    }
                }
            }
            LockLeave();

            return l_State;
        }

        private void RemoveAllKnownFiles()
        {
            // The caller must hold f_Lock.

            f_KnownFilesByFileNum.Clear();
            f_KnownFilesBySize.Clear();
            f_KnownFilesByeD2kFileID.Clear();
            f_UnlinkedKnw.Clear();
            f_UnhashedKnw.Clear();
            f_HashedKnw.Clear();

            f_NextFileNum = c_StartingFileNum;
        }

        private void ResolveRedundantKeyProblem(Dictionary<ElmKey, KnownFile> a_UsedKeys,
                                                KnownFile a_Known)
        {
            Util.DebugBreak();
        }

        private void NotifyFileAppearedOrChanged(string a_FileName)
        {
            // a_FileName is treated case sensitive and should be
            // rooted. We watch for changed files mainly to get to
            // know when an empty file turns filled (this doesn't
            // work if the file is opened with exclusive access
            // by others).

            f_WatchersThreadsPresent.Inc();
            {
                if (true == f_RunScanWorker &&
                    true == FileNames.IsPathProcessable(a_FileName))
                {
                    AddOrUpdateSingleFile(a_FileName, false);
                }
            }
            f_WatchersThreadsPresent.Dec();
        }

        private void NotifyFileDisappeared(string a_FileName)
        {
            // a_FileName is treated case sensitive and should be
            // rooted.

            f_WatchersThreadsPresent.Inc();
            {
                if (true == f_RunScanWorker &&
                    true == FileNames.IsPathProcessable(a_FileName))
                {
                    CloseLocalFile(a_FileName, false);
                }
            }
            f_WatchersThreadsPresent.Dec();
        }

        private void NotifyFileRenamed(string a_OldPath,
                                       string a_NewPath)
        {
            // a_OldPath and a_NewPath are treated case sensitive
            // and should be rooted. NOTE: We only get in-place renames
            // here, but we could also handle file moves.

            f_WatchersThreadsPresent.Inc();
            {
                if (true == f_RunScanWorker &&
                    true == FileNames.IsPathProcessable(a_OldPath))
                {
                    UpdateFileName(a_OldPath, a_NewPath, false);
                }
            }
            f_WatchersThreadsPresent.Dec();
        }

        private void NotifyCompleteScanIndicated()
        {
            // Gets called if a shared directory has somehow changed
            // or there were problems tracking changes.

            // This will delay the start in many situations. We want
            // that, since there is often a series of changes going on.
            const byte c_MaxSecsUntilStart = 5;

            f_WatchersThreadsPresent.Inc();
            {
                if (true == f_RunScanWorker &&
                    ScanState.Finished == f_ScanState)
                {
                    Time l_NextTime = f_Inst.GetTime().AddSeconds(c_MaxSecsUntilStart);

                    LockEnter();
                    {
                        if (ScanState.Finished == f_ScanState)
                        {
                            f_NextAutoScanTime = Time.Min(f_NextAutoScanTime, l_NextTime);
                        }
                    }
                    LockLeave();
                }
            }
            f_WatchersThreadsPresent.Dec();
        }

        private void UpdatePreferences(ref IOPrefs a_NewPrefs)
        {
            FileSystemWatcher[] l_Watchers = null;

            LockEnter();
            {
                bool l_Change = !f_IOPrefs.f_SharedDirs.ValueEquals(a_NewPrefs.f_SharedDirs);

                if (false == l_Change)
                {
                    if (f_IOPrefs.f_WatchFileSystem != a_NewPrefs.f_WatchFileSystem)
                    {
                        // The watcher state changes.

                        if (false == a_NewPrefs.f_WatchFileSystem)
                        {
                            l_Watchers = f_Watchers;
                            f_Watchers = null;
                        }
                    }
                }
                else
                {
                    // Take over the existing scan generation.
                    a_NewPrefs.f_SharedDirs.SetScanGeneration(f_IOPrefs.f_SharedDirs.p_ScanGen);

                    if (true == f_IOPrefs.f_WatchFileSystem)
                    {
                        l_Watchers = f_Watchers;
                        f_Watchers = null;
                    }
                }

                f_IOPrefs = a_NewPrefs;
                Software.s_LastOutputDir = f_IOPrefs.f_SharedDirs.GetCompletionTargetPath();

                if (true == l_Change)
                {
                    TriggerScan();
                }
            }
            LockLeave();

            // Do havy io outside a locked region.

            if (null != l_Watchers)
            {
                SharedDirectories.DestroyWatchers(l_Watchers);
                l_Watchers = null;
            }

            if (true == a_NewPrefs.f_WatchFileSystem &&
                false == Util.s_IsRunningOnWine)
            {
                l_Watchers = a_NewPrefs.f_SharedDirs.CreateWatchers(this.NotifyFileAppearedOrChanged,
                                                                    this.NotifyFileDisappeared,
                                                                    this.NotifyFileRenamed,
                                                                    this.NotifyCompleteScanIndicated);
            }

            if (null != l_Watchers)
            {
                LockEnter();
                {
                    // If things are still as they were, we can
                    // now update f_Watcher.

                    if (null == f_Watchers &&
                        true == f_IOPrefs.f_WatchFileSystem &&
                        f_IOPrefs.f_SharedDirs == a_NewPrefs.f_SharedDirs)
                    {
                        f_Watchers = l_Watchers;
                        l_Watchers = null;
                    }
                }
                LockLeave();
            }

            if (null != l_Watchers)
            {
                SharedDirectories.DestroyWatchers(l_Watchers);
            }

            byte l_Percentage = a_NewPrefs.f_HashingHDLoadThresholdPercentage;
            const uint lc_LowIdleTicks = (Util.c_MiniSleepTimeout * Time.c_TicksPerMillisecond);
            const uint lc_MaxIdleTicks = (2 * Time.c_TicksPerSecond);
            uint l_IdleTicks;

            if (1 >= l_Percentage)
            {
                l_IdleTicks = lc_MaxIdleTicks;
            }
            else if (100 <= l_Percentage)
            {
                l_IdleTicks = 0;
            }
            else
            {
                if (50 <= l_Percentage)
                {
                    l_Percentage++;
                }

                l_IdleTicks = (((100u - l_Percentage) * ((lc_MaxIdleTicks - lc_LowIdleTicks) / 100)) + lc_LowIdleTicks);
            }

            f_HashingIdleTicks = l_IdleTicks;
        }

        public AccessLevel CheckUlAccess(KnownFile a_Known,
                                         ShareKinds a_Caller,
                                         UserContext a_UserCntxt)
        {
            // NOTE: This is only a snapshot test that expires
            // immediately.

            ComplStatus l_Status = a_Known.p_StatusVolatile;
            ShareKinds l_ShareKind;

            if (ComplStatus.Complete == l_Status)
            {
                LocalFile l_Local = a_Known.p_LocalFileVolatile;

                if (null == l_Local)
                {
                    return AccessLevel.SourceMissing;
                }

                l_ShareKind = l_Local.p_ShareKind;
            }
            else if (ComplStatus.Incomplete == l_Status)
            {
                l_ShareKind = ShareKinds.eD2k_on_I2P;
            }
            else if (ComplStatus.Missing == l_Status)
            {
                return AccessLevel.SourceMissing;
            }
            else
            {
                return AccessLevel.Impossible;
            }

            if (a_Caller != (a_Caller & l_ShareKind))
            {
                return AccessLevel.AccessDenied;
            }

            if (0 != (ShareKinds.FriendsOnly & l_ShareKind) &&
                false == a_UserCntxt.IsFriend())
            {
                return AccessLevel.AccessDenied;
            }

            return AccessLevel.AccessGranted;
        }

#region IncompleteFileHandling

        public bool RegisterSourceAsync(KnownFile a_KnownFile,
                                        SourceCookie a_Cookie)
        {
            bool l_Registered = AddSingleIncomplFileOp(new IncomplOperationContext(a_KnownFile, true, a_Cookie));

            if (false == l_Registered)
            {
                a_Cookie.RejectSource();
            }

            return l_Registered;
        }

        public bool UnregisterSourceAsync(KnownFile a_KnownFile,
                                          SourceCookie a_Cookie)
        {
            if (false == a_Cookie.ShouldUnregister())
            {
                return true;
            }

            return AddSingleIncomplFileOp(new IncomplOperationContext(a_KnownFile, false, a_Cookie));
        }

        public bool UpdateSourceDataAsync(KnownFile a_KnownFile,
                                          SourceCookie a_Cookie,
                                          P2PHashSets a_ProvidableHashSets,
                                          ulong a_FileSize,
                                          ref MasterHashes a_MasterHashes)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_KnownFile, a_Cookie, a_ProvidableHashSets, a_FileSize, ref a_MasterHashes));
        }

        public bool DeliverHashesAsync(KnownFile a_Known,
                                      SourceCookie a_Cookie,
                                      byte[] a_Buffer,
                                      TemporaryFile a_File)
        {
            // a_Buffer and a_File may be null.

            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie, a_Buffer, a_File));
        }

        public bool QueryUsersAsync(KnownFile a_Known,
                                    QueryUsersCookie a_Cookie)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie));
        }

        public bool RegisterObtainCookieAsync(KnownFile a_Known,
                                              ObtainCookie a_Cookie,
                                              uint a_MaxBytes,
                                              BitMap a_CompletedByUser,
                                              NetworkAvailability a_VolatileAvail)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie, a_MaxBytes, a_CompletedByUser, a_VolatileAvail));
        }

        public bool WriteAsync(KnownFile a_Known,
                               ObtainCookie a_Cookie,
                               ref ByteArray a_Data)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie, ref a_Data));
        }

        public bool ExtractAsync(KnownFile a_Known,
                                 ExtractionCookie a_Cookie)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie));
        }

        private bool ImportAsync(KnownFile a_Known,
                                 ImportCookie a_Cookie)
        {
            return AddSingleIncomplFileOp(new IncomplOperationContext(a_Known, a_Cookie));
        }

        public bool DeleteIncomplDataAsync(List<KnownFile> a_Files)
        {
            int l_Count = a_Files.Count;

            if (0 >= l_Count)
            {
                return true;
            }

            IncomplOperationContext[] l_Ops = new IncomplOperationContext[l_Count];

            int l_DestIdx = 0;
            foreach (KnownFile l_Known in a_Files)
            {
                l_Ops[l_DestIdx++] = IncomplOperationContext.CreateForPermanentDeletion(l_Known);
            }

            return AddIncomplFileOpRange(l_Ops);
        }

        private bool AddSingleIncomplFileOp(IncomplOperationContext a_Context)
        {
            return InternalAddIncomplFileOps(a_Context);
        }

        private bool AddIncomplFileOpRange(IncomplOperationContext[] a_Contexts)
        {
            return (0 >= a_Contexts.Length ||
                    true == InternalAddIncomplFileOps(a_Contexts));
        }

        private bool InternalAddIncomplFileOps(object a_Obj)
        {
            // May only be used by AddSingleIncomplFileOp() and
            // AddIncomplFileOpRange().

            bool l_IsArr = (typeof(IncomplOperationContext[]) == a_Obj.GetType());

            Util.MonitorEnter(f_IncomplOps);
            {
                if (false == l_IsArr)
                {
                    InternalAddSingleIncomplFileOpCore((IncomplOperationContext)a_Obj);
                }
                else
                {
                    foreach (IncomplOperationContext l_Cntxt in (IncomplOperationContext[])a_Obj)
                    {
                        InternalAddSingleIncomplFileOpCore(l_Cntxt);
                    }
                }

                Monitor.Pulse(f_IncomplOps);
            }
            Util.MonitorLeave(f_IncomplOps);

            return true;
        }

        private void InternalAddSingleIncomplFileOpCore(IncomplOperationContext a_Context)
        {
            // May only be used by InternalAddIncomplFileOps(). The
            // caller must hold the f_IncomplOps lock.

            KnownFile l_Known = a_Context.f_Known;
            Queue<IncomplOperationContext> l_OpsQueue;

            if (false == f_IncomplOps.TryGetValue(l_Known, out l_OpsQueue))
            {
                l_OpsQueue = new Queue<IncomplOperationContext>(1);

                f_IncomplOps.Add(l_Known, l_OpsQueue);
                f_IncomplOrder.Add(l_Known);
            }

            l_OpsQueue.Enqueue(a_Context);
        }

        private void IncomplWorker(object a_Obj)
        {
            int l_WorkerIndex = (int)a_Obj;
            object l_IncomplLock = f_IncomplOps;
            IncompleteFile.TempBuffers l_IncomplBuffers = new IncompleteFile.TempBuffers();

            Util.MonitorEnter(l_IncomplLock);
            f_IncomplWorkersSpinning++;

            while (true)
            {
                START:

                int l_Count = f_IncomplOrder.Count;

                if (0 == l_Count)
                {
                    if (false == f_RunIncomplWorker)
                    {
                        // Shutdown in progress. Keep only one thread
                        // and only if there are open files left.

                        if (2 <= f_IncomplWorkersSpinning ||
                            0 == f_IncomplFilesList.p_Count)
                        {
                            f_IncomplWorkersSpinning--;

                            bool l_Last = (0 == f_IncomplWorkersSpinning);

                            Util.MonitorLeave(l_IncomplLock);

                            f_IncomplWorkerFinished[l_WorkerIndex].Set();

                            if (true == l_Last)
                            {
                                Software.UpdateShutDownState(0x8000000);
                            }

                            return;
                        }

                        // Close what's left. Normally this shouldn't be
                        // necessary, but it's legal if it happens.

                        for (IncompleteFile l_Incompl = f_IncomplFilesList.p_First; null != l_Incompl; l_Incompl = l_Incompl.p_Next)
                        {
                            if (true == l_Incompl.f_Known.f_IncomplWorkerOccupied)
                            {
                                continue;
                            }

                            Util.DebugBreak();

                            f_IncomplFilesList.Remove(l_Incompl);

                            l_Incompl.f_Known.f_IncomplWorkerOccupied = true;
                            Util.MonitorLeave(l_IncomplLock);
                            l_Incompl.AttachThread(l_IncomplBuffers);
                            {
                                OpStatus l_CloseStatus = l_Incompl.Close();
                            }
                            l_Incompl.DetachThread();
                            Util.MonitorEnter(l_IncomplLock);
                            l_Incompl.f_Known.f_IncomplWorkerOccupied = false;
                            l_Incompl.f_Known.f_Incompl = null;

                            goto START;
                        }
                    }
                }
                else
                {
                    // Process the queued incomplete file operations,
                    // in the order they were enqueued (per file).

                    for (int i = 0; i < l_Count; i++)
                    {
                        if (true == f_IncomplOrder[i].f_IncomplWorkerOccupied)
                        {
                            continue;
                        }

                        KnownFile l_Known = f_IncomplOrder[i];
                        Queue<IncomplOperationContext> l_OpsQueue = f_IncomplOps[l_Known];

                        f_IncomplOrder.RemoveAt(i);
                        f_IncomplOps.Remove(l_Known);

                        IncompleteFile l_IncomplFileBefore = l_Known.f_Incompl;
                        IncompleteFile l_IncomplFileAfter;

                        l_Known.f_IncomplWorkerOccupied = true;
                        Util.MonitorLeave(l_IncomplLock);
                        {
                            ProcessIncomplOps(l_Known, l_OpsQueue, l_IncomplBuffers);

                            l_IncomplFileAfter = l_Known.f_Incompl;

                            if (null == l_IncomplFileAfter)
                            {
                                IncomplFileProgress l_IncomplProgress = l_Known.f_IncomplProgress;

                                if (null != l_IncomplProgress)
                                {
                                    Util.VolatileWrite(out l_Known.f_IncomplProgress, l_IncomplProgress.RecreateWithClearedHotChunks(true));
                                }
                            }
                            else
                            {
                                l_IncomplFileAfter.Service(f_Inst.GetTime());

                                Util.VolatileWrite(out l_Known.f_IncomplProgress, l_IncomplFileAfter.GetProgress());

                                l_IncomplFileAfter.DetachThread();
                            }
                        }
                        Util.MonitorEnter(l_IncomplLock);
                        l_Known.f_IncomplWorkerOccupied = false;

                        // Update f_IncomplFilesList.

                        if (l_IncomplFileBefore != l_IncomplFileAfter)
                        {
                            if (null != l_IncomplFileBefore)
                            {
                                f_IncomplFilesList.Remove(l_IncomplFileBefore);
                            }

                            if (null != l_IncomplFileAfter)
                            {
                                f_IncomplFilesList.AddLast(l_IncomplFileAfter);
                            }
                        }

                        goto START;
                    }
                }

                Monitor.Wait(l_IncomplLock, Util.c_MiniSleepTimeout);
            }
        }

        private void ProcessIncomplOps(KnownFile a_Known,
                                       Queue<IncomplOperationContext> a_OpsQueue,
                                       IncompleteFile.TempBuffers a_IncomplBuffers)
        {
            // l_Incompl must be kept up to date throughout this
            // routine.
            IncompleteFile l_Incompl = a_Known.f_Incompl;

            if (null != l_Incompl)
            {
                l_Incompl.AttachThread(a_IncomplBuffers);
            }

            // Queued items must always be processed in the order they
            // were enqueued.

            foreach (IncomplOperationContext l_Cntxt in a_OpsQueue)
            {
                OpStatus l_Status = OpStatus.Error;

                switch (l_Cntxt.f_Type)
                {
                    case IncomplOperationContext.OperationTypes.RegisterSource:
                    {
                        SourceCookie l_SourceCookie = (SourceCookie)l_Cntxt.f_Param;
                        bool l_Loaded = false;

                        if (null == l_Incompl)
                        {
                            l_Status = IncompleteFile.Initialize(f_Inst, a_Known, f_IncomplMem, ref a_Known.f_IncomplKey, a_IncomplBuffers, out a_Known.f_Incompl);

                            if (OpStatus.Done != l_Status)
                            {
                                l_SourceCookie.RejectSource();
                            }
                            else
                            {
                                l_Loaded = true;
                                l_Incompl = a_Known.f_Incompl;

                                MasterHashes l_MasterHashes;

                                LockEnter();
                                {
                                    a_Known.f_MasterHashes.Clone(out l_MasterHashes);
                                }
                                LockLeave();

                                l_Incompl.SetOwnFileKnowledge(a_Known.f_FileSize, ref l_MasterHashes);
                            }
                        }

                        if (null != l_Incompl)
                        {
                            l_Status = l_Incompl.RegisterSource(l_SourceCookie);

                            if (OpStatus.Done != l_Status &&
                                true == l_Loaded)
                            {
                                // Unload again.

                                OpStatus l_CloseStatus = l_Incompl.Close();

                                ReleaseIncompl(l_Cntxt.f_Known, ref l_Incompl);
                            }
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.UnregisterSource:
                    {
                        if (null != l_Incompl)
                        {
                            SourceCookie l_SourceCookie = (SourceCookie)l_Cntxt.f_Param;
                            l_Status = l_Incompl.UnregisterSource(l_SourceCookie);

                            if (false == l_Incompl.HasActiveSources())
                            {
                                // Auto close.

                                OpStatus l_CloseStatus = l_Incompl.Close();

                                if (OpStatus.Done != l_CloseStatus)
                                {
                                    Util.DebugBreak();
                                }

                                ReleaseIncompl(l_Cntxt.f_Known, ref l_Incompl);
                            }
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.UpdateSource:
                    {
                        if (null != l_Incompl)
                        {
                            UpdateSourceContext l_UpdateSourceCntxt = (UpdateSourceContext)l_Cntxt.f_Param;
                            l_Status = l_Incompl.UpdateSource(ref l_UpdateSourceCntxt);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.Read:
                    {
                        ReadCookie l_ReadCookie = (ReadCookie)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_ReadCookie.Fail();
                        }
                        else
                        {
                            ByteArray l_Data = new ByteArray(l_ReadCookie.f_Size);

                            l_Status = l_Incompl.Read(l_ReadCookie.f_Offset, ref l_Data);

                            if (OpStatus.Done == l_Status)
                            {
                                l_ReadCookie.Succeed(ref l_Data);
                            }
                            else
                            {
                                l_ReadCookie.Fail();
                            }
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.DeliverHashes:
                    {
                        if (null != l_Incompl)
                        {
                            DeliverHashContext l_DeliverHashCntxt = (DeliverHashContext)l_Cntxt.f_Param;
                            l_Status = l_Incompl.DeliverHashes(ref l_DeliverHashCntxt);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.QueryUsers:
                    {
                        QueryUsersCookie l_QueryUsersCookie = (QueryUsersCookie)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_QueryUsersCookie.Fail();
                        }
                        else
                        {
                            l_Status = l_Incompl.QueryUsers(l_QueryUsersCookie);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.RegisterObtaining:
                    {
                        ObtainContext l_ObtainCookieCntxt = (ObtainContext)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_ObtainCookieCntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.IncomplFileNotLoaded);
                        }
                        else
                        {
                            if (false == f_CriticalDiskSpaceLevelReached)
                            {
                                l_Status = l_Incompl.RegisterObtainCookie(ref l_ObtainCookieCntxt);
                            }
                            else
                            {
                                l_ObtainCookieCntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.CriticalDiskSpaceLevel);
                            }
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.Write:
                    {
                        if (true == CanWriteToIncomplFile(l_Incompl))
                        {
                            WriteContext l_WriteCntxt = (WriteContext)l_Cntxt.f_Param;
                            l_Status = l_Incompl.Write(ref l_WriteCntxt);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.Extract:
                    {
                        ExtractionCookie l_ExtractionCookie = (ExtractionCookie)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_ExtractionCookie.Fail(ExtractionCookie.Result.FileNotActive);
                        }
                        else
                        {
                            l_Status = l_Incompl.Extraction(l_ExtractionCookie);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.BuildFileDescriptor:
                    {
                        FileDescriptorCookie l_FileDescCookie = (FileDescriptorCookie)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_FileDescCookie.Fail();
                        }
                        else
                        {
                            l_Status = l_Incompl.BuildFileDescriptor(l_FileDescCookie);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.QueryHashes:
                    {
                        HashQueryCookie l_HashQueryCookie = (HashQueryCookie)l_Cntxt.f_Param;

                        if (null == l_Incompl)
                        {
                            l_HashQueryCookie.Fail();
                        }
                        else
                        {
                            l_Status = l_Incompl.HandleHashQuery(l_HashQueryCookie);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.Import:
                    {
                        ImportCookie l_ImportCookie = (ImportCookie)l_Cntxt.f_Param;

                        if (false == CanWriteToIncomplFile(l_Incompl))
                        {
                            l_ImportCookie.Fail(ImportCookie.Result.FileNotActive);
                        }
                        else
                        {
                            l_Status = l_Incompl.ContinueImport(l_ImportCookie);
                        }
                    }
                    break;

                    case IncomplOperationContext.OperationTypes.Delete:
                    {
                        // Prepare deletion.

                        if (null == l_Incompl)
                        {
                            if (true == a_Known.f_IncomplKey.IsZero())
                            {
                                l_Status = OpStatus.NothingToDo;
                            }
                            else
                            {
                                l_Status = IncompleteFile.Initialize(f_Inst, a_Known, f_IncomplMem, ref a_Known.f_IncomplKey, a_IncomplBuffers, out a_Known.f_Incompl);

                                if (OpStatus.Done == l_Status)
                                {
                                    l_Incompl = a_Known.f_Incompl;
                                }
                            }
                        }
                        else
                        {
                            // Delete the file even if it's in use.

                            l_Status = OpStatus.Done;
                        }

                        if (OpStatus.Done == l_Status)
                        {
                            // Commit deletion.

                            l_Status = l_Incompl.DestroyForGood(ref a_Known.f_IncomplKey);

                            if (OpStatus.Done == l_Status)
                            {
                                ReleaseIncompl(a_Known, ref l_Incompl);
                                a_Known.f_IncomplProgress = null;
                            }
                            else
                            {
                                Util.DebugBreak();
                            }
                        }
                    }
                    break;

                    default: throw (new ApplicationException());
                }
            }
        }

        private bool CanWriteToIncomplFile(IncompleteFile a_Incompl)
        {
            // Should only be used by ProcessIncomplOps().

            if (null != a_Incompl)
            {
                KnownFile l_Known = a_Incompl.f_Known;

                if (ComplStatus.Incomplete == l_Known.p_Status ||
                    ComplStatus.Incomplete == l_Known.p_StatusVolatile)
                {
                    return true;
                }
            }

            return false;
        }

        private void ReleaseIncompl(KnownFile a_Known,
                                    ref IncompleteFile a_Incompl)
        {
            a_Incompl.DetachThread();

            a_Known.f_Incompl = null;
            a_Incompl = null;
        }

#endregion

        private void AddPossibleFileID(KnownFile a_Known)
        {
            AddOrRemovePossibleFileID(a_Known, true);
        }

        private void RemovePossibleFileID(KnownFile a_Known)
        {
            AddOrRemovePossibleFileID(a_Known, false);
        }

        private void AddOrRemovePossibleFileID(KnownFile a_Known,
                                               bool a_Add)
        {
            // The caller must hold f_Lock.

            if (null != a_Known.f_FileID.f_Val)
            {
                bool l_OK;

                if (false == a_Add)
                {
                    l_OK = f_KnownFilesByeD2kFileID.Remove(a_Known.f_FileID, a_Known);
                }
                else
                {
                    l_OK = f_KnownFilesByeD2kFileID.Add(a_Known.f_FileID, a_Known);
                }
            }
        }

        private bool HasFileWitheD2kFileIDAndSize(FileID a_FileID,
                                                  ulong a_FileSize)
        {
            // The caller must hold f_Lock.

            return (null != GetAnyFileWitheD2kFileIDAndSize(a_FileID, a_FileSize));
        }

        private KnownFile GetAnyFileWitheD2kFileIDAndSize(FileID a_FileID,
                                                          ulong a_FileSize)
        {
            // The caller must hold f_Lock.

            MultiMap<FileID, KnownFile>.Enumerator l_Enum;
            if (true == f_KnownFilesByeD2kFileID.CreateEnumerator(a_FileID, out l_Enum))
            {
                KnownFile l_Known;
                while (true == l_Enum.GetNext(out l_Known))
                {
                    if (l_Known.f_FileSize == a_FileSize)
                    {
                        // There is a file with that file id and size.

                        return l_Known;
                    }
                }
            }

            return null;
        }

        private KnownFile[] GetFilesByCategory(Category a_Category)
        {
            Dictionary<FileNum, KnownFile> l_DicClone;

            LockEnter();
            {
                l_DicClone = new Dictionary<FileNum, KnownFile>(f_KnownFilesByFileNum);
            }
            LockLeave();

            List<KnownFile> l_Ls = new List<KnownFile>(l_DicClone.Count);
            l_Ls.AddRange(l_DicClone.Values);

            if (Category.s_All != a_Category)
            {
                if (Category.s_Incomplete == a_Category)
                {
                    l_Ls.RemoveAll(IsNotIncomplPredicate);
                }
                else if (Category.s_Missing == a_Category)
                {
                    l_Ls.RemoveAll(IsNotMissingPredicate);
                }
                else if (Category.s_Complete == a_Category)
                {
                    l_Ls.RemoveAll(IsNotComplPredicate);
                }
                else if (Category.s_Active == a_Category)
                {
                    l_Ls.RemoveAll(IsNotActivePredicate);
                }
                else if (Category.s_Downloads == a_Category)
                {
                    l_Ls.RemoveAll(IsNotDownloadedPredicate);
                }
                else if (Category.s_Uploads == a_Category)
                {
                    l_Ls.RemoveAll(IsNotUploadedPredicate);
                }
                else
                {
                    // Non-generic categories go here as well as
                    // Category.s_Unassigned.
                    l_Ls.RemoveAll(a_Category.FilterRoutine);
                }
            }

            return l_Ls.ToArray();
        }

        private static bool IsNotIncomplPredicate(KnownFile a_Known)
        {
            return (ComplStatus.Incomplete != a_Known.p_Status);
        }

        private static bool IsNotMissingPredicate(KnownFile a_Known)
        {
            return (ComplStatus.Missing != a_Known.p_Status);
        }

        private static bool IsNotComplPredicate(KnownFile a_Known)
        {
            return (ComplStatus.Complete != a_Known.p_Status);
        }

        private static bool IsNotActivePredicate(KnownFile a_Known)
        {
            return !a_Known.IsActive();
        }

        private static bool IsNotDownloadedPredicate(KnownFile a_Known)
        {
            return !a_Known.IsDownloadedVolatile();
        }

        private static bool IsNotUploadedPredicate(KnownFile a_Known)
        {
            return !a_Known.IsUploadedVolatile();
        }

#region AsyncCalls

        public void EnqueueAPC(GeneralAPCCookie a_Cookie)
        {
            // By default the scan worker will handle this.
            f_ScanQueue.Enqueue(a_Cookie);
        }

        public bool EnqueueAPCAndWaitAsGUI(GeneralAPCCookie a_Cookie)
        {
            // Do not continue processing if false is returned.

            EnqueueAPC(a_Cookie);

            return a_Cookie.GuiWaitForCompletion();
        }

        public GeneralAPCCookie IncomplNotifyFinishedAsync(KnownFile a_Known)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(NotifyFinishedIncomplFile, a_Known);

            // This can only be done by the hash worker.
            f_HashQueue.Enqueue(l_Cookie);

            return l_Cookie;
        }

        public void GuiBeginImport(MultiFileImportContext a_Cntxt)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(AddImportJob, a_Cntxt);

            // This is done by the ident worker.
            f_IdentQueue.Enqueue(l_Cookie);
        }

        public void GuiBeginFileRemoval(Gui.FilesPanel.FileRemovalContext a_Cntxt)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(RemoveFiles, a_Cntxt);

            // This can only be done by the hash worker (since missing
            // files might need their hash sets removed).
            f_HashQueue.Enqueue(l_Cookie);
        }

        public GeneralAPCCookie GuiBeginToTurnMissingIntoIncomplFiles(KnownFile[] a_Known)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(TurnMissingIntoIncomplFiles, a_Known);

            // This can only be done by the hash worker (since missing
            // files might need their hash sets removed).
            f_HashQueue.Enqueue(l_Cookie);

            return l_Cookie;
        }

        public GeneralAPCCookie GuiBeginRecheckOfCompleteFiles(KnownFile[] a_Known)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(RecheckCompleteFiles, a_Known);

            // This can only be done by the hash worker.
            f_HashQueue.Enqueue(l_Cookie);

            return l_Cookie;
        }

        public bool NetworkQueryHashes(HashQueryCookie a_Cookie)
        {
            f_HashQueue.Enqueue(a_Cookie);

            return true;
        }

        public bool NetworkBuildFileDescriptor(FileDescriptorCookie a_Cookie)
        {
            f_HashQueue.Enqueue(a_Cookie);

            return true;
        }

        public ReadCookie ReadAsync(KnownFile a_Known,
                                    bool a_Incompl,
                                    ShareKinds a_Caller,
                                    ulong a_Offset,
                                    int a_Size)
        {
            ReadCookie l_Cookie = new ReadCookie(a_Known, a_Offset, a_Size);

            if (false == a_Incompl)
            {
                f_ReadComplQueue.Enqueue(l_Cookie);
            }
            else
            {
                if (false == AddSingleIncomplFileOp(new IncomplOperationContext(l_Cookie)))
                {
                    return null;
                }
            }

            return l_Cookie;
        }

        public void NetworkGetFilesForDownloading(GeneralAPCCookie a_Cookie)
        {
            List<KnownFile> l_Ls;

            LockEnter();
            {
                l_Ls = new List<KnownFile>(f_KnownFilesByFileNum.Count);

                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    if (ComplStatus.Incomplete == l_Known.p_Status &&
                        true == l_Known.IsDownloadAllowed() &&
                        null != l_Known.f_FileID.f_Val)
                    {
                        IncomplFileProgress l_IncomplProgress = l_Known.p_IncomplProgressVolatile;

                        if (null == l_IncomplProgress ||
                            false == l_IncomplProgress.f_DoneMap.p_AllSet)
                        {
                            l_Ls.Add(l_Known);
                        }
                    }
                }
            }
            LockLeave();

            l_Ls.Sort(CompareKnownFilesForDownloading);

            a_Cookie.SetReturnObjectAndState(l_Ls.ToArray(), CookieStates.Success);
        }

        private static int CompareKnownFilesForDownloading(KnownFile a_A,
                                                           KnownFile a_B)
        {
            // Puts files well suited for downloading at the beginning
            // of the sort order.

            int l_Res = a_B.f_DlPriority.f_Value.CompareTo(a_A.f_DlPriority.f_Value);

            if (0 != l_Res)
            {
                return l_Res;
            }

            IncomplFileProgress l_ProgressA = a_A.f_IncomplProgress;
            IncomplFileProgress l_ProgressB = a_B.f_IncomplProgress;
            Util.MemoryBarrier();

            if (null == l_ProgressA)
            {
                return (null == l_ProgressB ? 0 : 1);
            }

            return (null == l_ProgressB ? (-1) : l_ProgressB.f_CompletedBytesInclUnverified.CompareTo(l_ProgressA.f_CompletedBytesInclUnverified));
        }

        public void NetworkGetFilesForeD2kPublishing(GeneralAPCCookie a_Cookie)
        {
            PublishCookie l_PubCookie = (PublishCookie)a_Cookie;

            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    ComplStatus l_Status = l_Known.p_Status;

                    if (ComplStatus.Missing == l_Status ||
                        true == l_Known.f_UlLimit.IsNearZero() ||
                        null == l_Known.f_FileID.f_Val)
                    {
                        // File is missing, may not be uploaded or has
                        // an unknown file id.

                        continue;
                    }

                    if (ComplStatus.Incomplete == l_Status)
                    {
                        IncomplFileProgress l_IncomplProgress = l_Known.p_IncomplProgressVolatile;

                        if (null == l_IncomplProgress ||
                            false == l_IncomplProgress.f_DoneMap.Has8MiBBitmapeD2kRelevance(l_Known.f_FileSize))
                        {
                            // This file is not yet uploadable.

                            continue;
                        }
                    }

                    uint l_HighUInt = l_Known.f_FileID.f_Val.GetHighUInt();

                    if (true == l_PubCookie.IsRelevant(l_HighUInt))
                    {
                        l_PubCookie.f_Files.Add(l_Known);
                    }

                    if (false == l_PubCookie.f_FilesOnly)
                    {
                        l_Known.TryAddKeywords(l_PubCookie);
                    }
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void NetworkReleasePublishingResources(GeneralAPCCookie a_Cookie)
        {
            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    l_Known.ClearKeywordCache();
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiSaveFiles(GeneralAPCCookie a_Cookie)
        {
            StorageStates l_State = SaveFiles();

            a_Cookie.SetReturnObjectAndState(l_State, CookieStates.Success);
        }

        public void GuiReloadSharedFiles(GeneralAPCCookie a_Cookie)
        {
            LockEnter();
            {
                TriggerScan();
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiRescanForFinishedFiles(GeneralAPCCookie a_Cookie)
        {
            f_NextRescanForFinishedFilesTime = new Time();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiUpdatePreferences(GeneralAPCCookie a_Cookie)
        {
            Preferences l_Prefs = (Preferences)a_Cookie.GetArg();

            UpdatePreferences(ref l_Prefs.f_IOPrefs);

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiClearCategory(GeneralAPCCookie a_Cookie)
        {
            Category l_Category = (Category)a_Cookie.GetArg();

            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    if (l_Category == l_Known.p_Category)
                    {
                        l_Known.SetCategory(Category.s_Unassigned);
                    }
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiChangeFilesCategory(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile[] l_Files = (KnownFile[])l_Args[0];
            Category l_Category = (Category)l_Args[1];

            LockEnter();
            {
                foreach (KnownFile l_Known in l_Files)
                {
                    l_Known.SetCategory(l_Category);
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiGetFileLink(GeneralAPCCookie a_Cookie)
        {
            KnownFile l_Known = a_Cookie.GetArgAsKnownFile();
            string l_FileName;
            MasterHashes l_MasterHashes;
            bool l_Dropped;

            // We use the lock for that, since that is the only way
            // to produce reliable results.

            LockEnter();
            {
                l_Dropped = (ComplStatus.Dropped == l_Known.p_Status);

                if (false == l_Dropped)
                {
                    l_FileName = l_Known.GetFileName(false);
                    l_Known.f_MasterHashes.Clone(out l_MasterHashes);
                }
                else
                {
                    l_FileName = null;
                    l_MasterHashes = new MasterHashes();
                }
            }
            LockLeave();

            if (false == l_Dropped &&
                true == l_MasterHashes.p_HaseD2kHash)
            {
                FileLink l_Link = new FileLink(l_FileName, l_Known.f_FileSize, ref l_MasterHashes);
                a_Cookie.SetReturnObjectAndState(l_Link, CookieStates.Success);
            }
            else
            {
                a_Cookie.SetState(CookieStates.Failure);
            }
        }

        public void GuiAddFileFromLink(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            FileLink l_Link = (FileLink)l_Args[0];
            Category l_Category = (Category)l_Args[1];

            if (false == IsValidFileSize(l_Link.f_FileSize) ||
                false == l_Link.f_MasterHashes.p_HaseD2kHash)
            {
                a_Cookie.SetState(CookieStates.Failure);

                return;
            }

            FileID l_FileID = l_Link.f_MasterHashes.CreateFileID();
            KnownFile l_Known;

            LockEnter();
            {
                if (false == HasFileWitheD2kFileIDAndSize(l_FileID, l_Link.f_FileSize))
                {
                    l_Known = KnownFile.CreateFileFromLink(FindUnoccupiedFileNum(), l_Link, l_FileID, f_IOPrefs.f_InsertIncomplFilesStarted);
                    l_Known.SetCategory(l_Category);

                    ReferenceKnownFile(l_Known, f_UnlinkedKnw);
                }
                else
                {
                    l_Known = null;
                }
            }
            LockLeave();

            if (null == l_Known)
            {
                a_Cookie.SetState(CookieStates.Failure);
            }
            else
            {
                a_Cookie.SetReturnObjectAndState(l_Known, CookieStates.Success);
            }
        }

        public void GuiAddFilesFromCollection(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            FileCollection.FileInfo[] l_Infos = (FileCollection.FileInfo[])l_Args[0];
            Category l_Category = (Category)l_Args[1];
            uint l_Added = 0;

            LockEnter();
            {
                for (int i = (l_Infos.Length - 1); 0 <= i; i--)
                {
                    if (false == l_Infos[i].f_Enabled ||
                        false == IsValidFileSize(l_Infos[i].f_FileSize) ||
                        false == l_Infos[i].f_MasterHashes.p_HaseD2kHash)
                    {
                        continue;
                    }

                    FileID l_FileID = l_Infos[i].f_MasterHashes.CreateFileID();

                    if (true == HasFileWitheD2kFileIDAndSize(l_FileID, l_Infos[i].f_FileSize))
                    {
                        continue;
                    }

                    l_Added++;
                    KnownFile l_Known = KnownFile.CreateFileFromCollection(FindUnoccupiedFileNum(), l_Infos, i, l_FileID, f_IOPrefs.f_InsertIncomplFilesStarted);
                    l_Known.SetCategory(l_Category);

                    ReferenceKnownFile(l_Known, f_UnlinkedKnw);
                }
            }
            LockLeave();

            a_Cookie.SetReturnObjectAndState(l_Added, CookieStates.Success);
        }

        public void GuiAddFoundFiles(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            Gui.FileSearchPanel.FoundFile[] l_Files = (Gui.FileSearchPanel.FoundFile[])l_Args[0];
            Category l_Category = (Category)l_Args[1];
            uint l_Added = 0;

            LockEnter();
            {
                for (int i = (l_Files.Length - 1); 0 <= i; i--)
                {
                    if (false == IsValidFileSize(l_Files[i].f_FileSize))
                    {
                        continue;
                    }

                    FileID l_FileID = new FileID(l_Files[i].f_FileID.ToRef());

                    if (true == HasFileWitheD2kFileIDAndSize(l_FileID, l_Files[i].f_FileSize))
                    {
                        continue;
                    }

                    l_Added++;
                    KnownFile l_Known = KnownFile.CreateFileFromSearch(FindUnoccupiedFileNum(), l_Files, i, l_FileID, f_IOPrefs.f_InsertIncomplFilesStarted);
                    l_Known.SetCategory(l_Category);

                    ReferenceKnownFile(l_Known, f_UnlinkedKnw);
                }
            }
            LockLeave();

            a_Cookie.SetReturnObjectAndState(l_Added, CookieStates.Success);
        }

        public void GuiMoveCompleteFile(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile[] l_Files = (KnownFile[])l_Args[0];
            string l_TargetDirPath = (string)l_Args[1];
            Gui.UnabortableControllerContext l_ControllerCntxt = (Gui.UnabortableControllerContext)l_Args[2];

            uint l_Count = 0;

            if (true == Directory.Exists(l_TargetDirPath))
            {
                foreach (KnownFile l_Known in l_Files)
                {
                    ComplStatus l_Status;
                    string l_ParentPath;
                    string l_FileName = l_Known.GetFileNameVolatile(true, out l_Status, out l_ParentPath);

                    if (ComplStatus.Complete != l_Status ||
                        null == l_ParentPath)
                    {
                        continue;
                    }

                    FileNameParts l_OldNameParts = new FileNameParts(LongPath.Combine(l_ParentPath, l_FileName));

                    if (null == l_OldNameParts.f_ParentDir ||
                        false == FileNames.IsPathProcessable(l_OldNameParts.f_Path))
                    {
                        continue;
                    }

                    FileNameParts l_NewNameParts = new FileNameParts(LongPath.Combine(l_TargetDirPath, l_FileName));

                    if (null == l_NewNameParts.f_ParentDir ||
                        false == FileNames.IsPathProcessable(l_NewNameParts.f_Path))
                    {
                        continue;
                    }

                    if (true == TryMoveCompleteFile(l_Known, ref l_OldNameParts, ref l_NewNameParts))
                    {
                        l_Count++;
                    }
                }
            }

            a_Cookie.SetReturnObjectAndState(l_Count, CookieStates.Success);

            l_ControllerCntxt.SetFinished();
        }

        private bool TryMoveCompleteFile(KnownFile a_Known,
                                         ref FileNameParts a_OldNameParts,
                                         ref FileNameParts a_NewNameParts)
        {
            bool l_Succ = false;
            LocalFile l_LinkingProtectedLocal = null;

            LockEnter();
            {
                if (ComplStatus.Complete != a_Known.p_Status)
                {
                    goto LEAVE;
                }

                LocalFile l_Local = a_Known.p_LocalFile;

                if (false == f_FileReg.HasFile(l_Local, a_OldNameParts.f_ParentDir, a_OldNameParts.f_FileName))
                {
                    goto LEAVE;
                }

                LocalFile l_Existing = f_FileReg.GetFile(a_NewNameParts.f_ParentDir, a_NewNameParts.f_FileName);

                if (null != l_Existing &&
                    l_Existing != l_Local)
                {
                    goto LEAVE;
                }

                f_DontOpenMoveSrcFileName = a_OldNameParts.f_Path;
                f_DontOpenMoveDestFileName = a_NewNameParts.f_Path;
                l_Local.AdjustLinkProtection(true);

                l_Succ = true;
                l_LinkingProtectedLocal = l_Local;

                LEAVE:;
            }
            LockLeave();

            if (true == l_Succ)
            {
                bool l_LinkingProtectionInced = true;
                l_Succ = TryMoveCompleteFileCore(a_Known, l_LinkingProtectedLocal, ref l_LinkingProtectionInced, ref a_OldNameParts, ref a_NewNameParts);

                f_DontOpenMoveDestFileName = f_DontOpenMoveSrcFileName = null;

                if (true == l_LinkingProtectionInced)
                {
                    LockEnter();
                    {
                        l_LinkingProtectedLocal.AdjustLinkProtection(false);
                    }
                    LockLeave();
                }
            }

            return l_Succ;
        }

        private bool TryMoveCompleteFileCore(KnownFile a_Known,
                                             LocalFile a_Local,
                                             ref bool a_LinkingProtectionInced,
                                             ref FileNameParts a_OldNameParts,
                                             ref FileNameParts a_NewNameParts)
        {
            // a_LinkingProtectionInced must be nonzero on entry.
            // This routine may clear f_DontOpenMoveDestFileName and
            // f_DontOpenMoveSrcFileName.

            if (true == FHandle.Move(a_OldNameParts.f_Path, a_NewNameParts.f_Path))
            {
                bool l_Succ = false;
                bool l_CloseLocal = false;

                LockEnter();
                {
                    if (true == a_LinkingProtectionInced)
                    {
                        a_LinkingProtectionInced = false;
                        a_Local.AdjustLinkProtection(false);
                    }

                    ComplStatus l_Status = a_Known.p_Status;

                    if (ComplStatus.Complete == l_Status)
                    {
                        if (a_Known.p_LocalFile == a_Local &&
                            true == f_FileReg.HasFile(a_Local, a_OldNameParts.f_ParentDir, a_OldNameParts.f_FileName))
                        {
                            ShareKinds l_ShareKind = f_IOPrefs.f_SharedDirs.GetShareKind(a_NewNameParts.f_ParentDir);

                            if (ShareKinds.None == l_ShareKind)
                            {
                                DereferenceLocalFile(a_Local);
                                l_CloseLocal = true;
                                l_Succ = true;

                                goto LEAVE;
                            }

                            string l_StoredLocation = a_Local.p_Parent.p_Path;
                            string l_StoredFileName = a_Local.p_FileName;

                            f_FileReg.DerefFile(a_Local);

                            a_Local.UpdateFileName(a_NewNameParts.f_FileName);
                            LocalFile l_Existing;

                            if (true == f_FileReg.RefFile(a_Local, a_NewNameParts.f_ParentDir, out l_Existing))
                            {
                                a_Local.SetScanGenAndShareKind(f_IOPrefs.f_SharedDirs.p_ScanGen, l_ShareKind);
                                l_Succ = true;

                                goto LEAVE;
                            }

                            // Set the previous directory again.

                            a_Local.UpdateFileName(l_StoredFileName);

                            if (false == f_FileReg.RefFile(a_Local, l_StoredLocation, out l_Existing))
                            {
                                DereferenceLocalFile(a_Local, false);
                                l_CloseLocal = true;

                                goto LEAVE;
                            }
                        }
                    }
                    else if (ComplStatus.Missing == l_Status)
                    {
                        f_DontOpenMoveDestFileName = f_DontOpenMoveSrcFileName = null;

                        LockLeave();

                        AddOrUpdateSingleFile(ref a_NewNameParts, true);

                        return true;
                    }

                    LEAVE:;
                }
                LockLeave();

                if (true == l_CloseLocal)
                {
                    a_Local.Close();
                }

                return l_Succ;
            }

            // Try to close the file temporarily, in case our own
            // access blocks the move.

            bool l_SuccessfullyDereferenced = false;

            LockEnter();
            {
                if (ComplStatus.Complete == a_Known.p_Status &&
                    a_Known.p_LocalFile == a_Local &&
                    true == f_FileReg.HasFile(a_Local, a_OldNameParts.f_ParentDir, a_OldNameParts.f_FileName))
                {
                    if (true == a_LinkingProtectionInced)
                    {
                        a_LinkingProtectionInced = false;
                        a_Local.AdjustLinkProtection(false);
                    }

                    DereferenceLocalFile(a_Local);
                    l_SuccessfullyDereferenced = true;
                }
            }
            LockLeave();

            if (false == l_SuccessfullyDereferenced)
            {
                return false;
            }

            a_Local.Close();

            if (true == FHandle.Move(a_OldNameParts.f_Path, a_NewNameParts.f_Path))
            {
                f_DontOpenMoveDestFileName = f_DontOpenMoveSrcFileName = null;

                AddOrUpdateSingleFile(ref a_NewNameParts, true);

                return true;
            }

            // Try to repair.

            f_DontOpenMoveDestFileName = f_DontOpenMoveSrcFileName = null;

            AddOrUpdateSingleFile(ref a_OldNameParts, true);

            return false;
        }

        public void GuiUpdateFoundFilesCompletionStatus(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            Gui.FileSearchPanel.FoundFile[] l_Files = (Gui.FileSearchPanel.FoundFile[])l_Args[0];
            int l_Count = (int)l_Args[1];

            // NOTE: l_Files may not be accessed beyond the count passed.

            LockEnter();
            {
                for (int i = 0; i < l_Count; i++)
                {
                    l_Files[i].f_ComplStatus = ComplStatus.Dropped;

                    FileID l_FileID = new FileID(l_Files[i].f_FileID.ToRef());
                    MultiMap<FileID, KnownFile>.Enumerator l_Enum;

                    if (true == f_KnownFilesByeD2kFileID.CreateEnumerator(l_FileID, out l_Enum))
                    {
                        for (KnownFile l_Known; true == l_Enum.GetNext(out l_Known);)
                        {
                            if (l_Files[i].f_FileSize != l_Known.f_FileSize)
                            {
                                continue;
                            }

                            ComplStatus l_Status = l_Known.p_Status;

                            if (l_Files[i].f_ComplStatus > l_Status)
                            {
                                continue;
                            }

                            l_Files[i].f_ComplStatus = l_Status;
                        }
                    }
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiCanAddFileWithID(GeneralAPCCookie a_Cookie)
        {
            // The returned value is only a snapshot.

            object[] l_Args = a_Cookie.GetArgAsObjArray();
            FileID l_FileID = new FileID((UInt128)l_Args[0]);
            ulong l_FileSize = (ulong)l_Args[1];

            bool l_Yes;

            LockEnter();
            {
                l_Yes = !HasFileWitheD2kFileIDAndSize(l_FileID, l_FileSize);
            }
            LockLeave();

            a_Cookie.SetReturnObjectAndState(l_Yes, CookieStates.Success);
        }

        public void NetworkGetKnownFileCache(GeneralAPCCookie a_Cookie)
        {
            Dictionary<UInt128, KnownFile> l_Dic = new Dictionary<UInt128, KnownFile>(f_KnownFilesByFileNum.Count);

            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    KnownFile l_Junk;
                    UInt128 l_FileID = l_Known.f_FileID.f_Val;

                    if (null != l_FileID &&
                        false == l_Dic.TryGetValue(l_FileID, out l_Junk))
                    {
                        l_Dic.Add(l_FileID, l_Known);
                    }
                }
            }
            LockLeave();

            a_Cookie.SetReturnObjectAndState(new KnownFileCache(l_Dic), CookieStates.Success);
        }

        public void NetworkGetFileByeD2kHash(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            FileID l_FileID = new FileID((UInt128)l_Args[0]);
            ulong l_FileSize = (ulong)l_Args[1];
            P2PHashSets l_MustNotDifferHashType = (P2PHashSets)l_Args[2];
            byte[] l_MustNotDifferHash = (byte[])l_Args[3];

            // l_FileSize may be zero, if file size is irrelevant.

            LockEnter();
            {
                MultiMap<FileID, KnownFile>.Enumerator l_Enum;
                if (true == f_KnownFilesByeD2kFileID.CreateEnumerator(l_FileID, out l_Enum))
                {
                    KnownFile l_Ret = null;

                    KnownFile l_Known;
                    while (true == l_Enum.GetNext(out l_Known))
                    {
                        // Find the file with the lowest file num to
                        // produce consistent output (in case the call
                        // is repeated).

                        if (0 != l_FileSize &&
                            l_Known.f_FileSize != l_FileSize)
                        {
                            continue;
                        }

                        if (P2PHashSets.None != l_MustNotDifferHashType &&
                            true == l_Known.f_MasterHashes.HasAllHashes(l_MustNotDifferHashType) &&
                            false == l_Known.f_MasterHashes.IsHashEqual(l_MustNotDifferHashType, l_MustNotDifferHash, 0))
                        {
                            continue;
                        }

                        if (null == l_Ret ||
                            l_Ret.f_FileNum.f_Val > l_Known.f_FileNum.f_Val)
                        {
                            l_Ret = l_Known;
                        }
                    }

                    a_Cookie.SetReturnObject(l_Ret);
                }
            }
            LockLeave();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiRenameFile(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile l_Known = (KnownFile)l_Args[0];
            string l_NewFileName = (string)l_Args[1];

            // l_NewFileName is treated case sensitive and must not
            // contain path information. This routine does only
            // rename within the same directory.

            bool l_Succ;

            LockEnter();
            {
                switch (l_Known.p_Status)
                {
                    case ComplStatus.Incomplete:
                        l_Known.f_IncomplFileName = l_NewFileName;
                        l_Known.f_CompatFileName = FileNames.MakeFileNameLegalFailsafe(l_NewFileName);
                        l_Known.ClearKeywordCache();
                        l_Succ = true;
                        break;

                    case ComplStatus.Complete:
                    {
                        LocalFile l_Local = l_Known.p_LocalFile;
                        string l_DirPath = l_Local.p_Parent.p_Path;
                        string l_OldFileName = l_Local.p_FileName;

                        LockLeave();

                        if (true == Util.StringsEqual(l_OldFileName, l_NewFileName))
                        {
                            l_Succ = true;

                            goto EXIT;
                        }

                        FileNameParts l_OldNameParts = new FileNameParts(LongPath.Combine(l_DirPath, l_OldFileName));
                        FileNameParts l_NewNameParts = new FileNameParts(LongPath.Combine(l_DirPath, l_NewFileName));

                        if (false == FileNames.IsPathProcessable(l_OldNameParts.f_Path) ||
                            false == FileNames.IsPathProcessable(l_NewNameParts.f_Path))
                        {
                            l_Succ = false;

                            goto EXIT;
                        }

                        l_Succ = TryMoveCompleteFile(l_Known, ref l_OldNameParts, ref l_NewNameParts);

                        goto EXIT;
                    }

                    case ComplStatus.Missing:
                        l_Known.RenameMissing(l_NewFileName);
                        l_Succ = true;
                        break;

                    case ComplStatus.Dropped:
                        l_Succ = false;
                        break;

                    default: throw (new ApplicationException());
                }
            }
            LockLeave();

            EXIT:

            a_Cookie.SetState(false == l_Succ ? CookieStates.Failure : CookieStates.Success);
        }

#endregion

#region FileRegistration

        private bool HasFileForbiddenAttributes(string a_Path)
        {
            // If we can't get the attributes, we assume the file has
            // no forbidden attributes.

            FileAttributes l_Attrs;
            Exception l_Exc;
            return (true == FHandle.GetAttributes(a_Path, out l_Attrs, out l_Exc) &&
                    0 != (l_Attrs & f_IOPrefs.f_ExclusionFileAttrbs));
        }

        private void UpdateFileName(string a_OldPath,
                                    string a_NewPath,
                                    bool a_MayIgnoreProtection)
        {
            // Use this if the file name has changed (it doesn't harm
            // if the old and the new name are the same).

            string l_OldDirName;
            string l_OldFileName = LongPath.SplitFullPath(a_OldPath, out l_OldDirName);

            if (null == l_OldDirName)
            {
                Util.DebugBreak();

                return;
            }

            string l_NewDirName;
            string l_NewFileName = LongPath.SplitFullPath(a_NewPath, out l_NewDirName);
            bool l_NewNameIsLegal = FileNames.IsLegalFileName(l_NewFileName);

            LocalFile l_FileThatMustBeClosed = null;

            LockEnter();
            {
                if (true == MayAddOrUpdateFile(a_NewPath, true))
                {
                    LocalFile l_Local = f_FileReg.GetFile(l_OldDirName, l_OldFileName);

                    if (null != l_Local)
                    {
                        if (true == a_MayIgnoreProtection ||
                            false == l_Local.p_IsLinkProtected)
                        {
                            if (false == UpdateFileNameCore(l_Local, l_NewDirName, l_NewFileName, l_NewNameIsLegal))
                            {
                                l_FileThatMustBeClosed = l_Local;
                            }
                        }
                    }
                }
            }
            LockLeave();

            if (null != l_FileThatMustBeClosed)
            {
                l_FileThatMustBeClosed.Close();
            }
        }

        private bool UpdateFileNameCore(LocalFile a_Local,
                                        string a_NewDirName,
                                        string a_NewFileName,
                                        bool a_NewNameIsLegal)
        {
            // The caller must hold f_Lock. a_NewDirName may be null.

            // We may not exit from here on while leaving the file
            // referenced, if we've failed to rename it. Close the file
            // if zero is returned.

            bool l_IsStillFileRegReferenced = true;

            if (null != a_NewDirName &&
                true == a_NewNameIsLegal)
            {
                ShareKinds l_ShareKind = f_IOPrefs.f_SharedDirs.GetShareKind(a_NewDirName);

                if (ShareKinds.None != l_ShareKind)
                {
                    // The new folder is shared.

                    f_FileReg.DerefFile(a_Local);
                    l_IsStillFileRegReferenced = false;

                    a_Local.UpdateFileName(a_NewFileName);

                    LocalFile l_Junk;

                    if (true == f_FileReg.RefFile(a_Local, a_NewDirName, out l_Junk))
                    {
                        a_Local.SetScanGenAndShareKind(f_IOPrefs.f_SharedDirs.p_ScanGen, l_ShareKind);

                        return true;
                    }
                }
            }

            DereferenceLocalFile(a_Local, l_IsStillFileRegReferenced);

            return false;
        }

        private void AddOrUpdateFiles(string[] a_Paths,
                                      int a_Count,
                                      bool a_MayIgnoreProtection)
        {
            FileNameParts[] l_NameParts = new FileNameParts[a_Count];

            for (int i = 0; i < a_Count; i++)
            {
                l_NameParts[i] = new FileNameParts(a_Paths[i]);

                if (true == HasFileForbiddenAttributes(l_NameParts[i].f_Path))
                {
                    // This file should not be added or updated.

                    l_NameParts[i].f_Handled = true;
                }
            }

            LockEnter();
            {
                for (int i = 0; i < a_Count; i++)
                {
                    if (true == l_NameParts[i].f_Handled)
                    {
                        continue;
                    }

                    LocalFile l_Existing = f_FileReg.GetFile(l_NameParts[i].f_ParentDir, l_NameParts[i].f_FileName);

                    if (null != l_Existing)
                    {
                        // We already have this file, so simply update it.

                        bool l_Updated = UpdateLocalFile(l_Existing, true, ref l_NameParts[i]);
                        l_NameParts[i].f_Handled = true;
                    }
                }
            }
            LockLeave();

            for (int i = 0; i < a_Count; i++)
            {
                if (true == l_NameParts[i].f_Handled)
                {
                    continue;
                }

                AddOrUpdateSingleFile(ref l_NameParts[i], a_MayIgnoreProtection);
            }
        }

        private void AddOrUpdateSingleFile(string a_FullFileName,
                                           bool a_MayIgnoreProtection)
        {
            FileNameParts l_NameParts = new FileNameParts(a_FullFileName);
            AddOrUpdateSingleFile(ref l_NameParts, a_MayIgnoreProtection);
        }

        private void AddOrUpdateSingleFile(ref FileNameParts a_NameParts,
                                           bool a_MayIgnoreProtection)
        {
            // Adds a new file and and/or - if there is already a
            // valid file - update the ref number.

            if (false == a_NameParts.IsPathProcessable())
            {
                Util.DebugBreak();

                return;
            }

            SharedDirectories l_SharedDirs = f_IOPrefs.f_SharedDirs;
            Util.MemoryBarrier();

            if (ShareKinds.None == l_SharedDirs.GetShareKind(a_NameParts.f_ParentDir) ||
                false == MayAddOrUpdateFile(a_NameParts.f_Path, false))
            {
                return;
            }

            Exception l_Exc;
            LocalFile l_Local = LocalFile.OpenR(a_NameParts.f_Path, f_IOPrefs.f_OpenFilesAsMoveable, out l_Exc);

            if (null == l_Local)
            {
                return;
            }

            if (true == HasFileForbiddenAttributes(a_NameParts.f_Path))
            {
                l_Local.Close();

                return;
            }

            bool l_CloseLocalFile = true;
            LocalFile l_OtherFileThatMustBeClosed = null;

            LockEnter();
            {
                if (true == MayAddOrUpdateFile(a_NameParts.f_Path, true))
                {
                    l_Local.UpdateFileName(a_NameParts.f_FileName);

                    RETRY:

                    LocalFile l_Existing;
                    bool l_Succ = f_FileReg.RefFile(l_Local, a_NameParts.f_ParentDir, out l_Existing);

                    if (false == l_Succ)
                    {
                        if (null != l_Existing)
                        {
                            // There is already a file that goes by that name.

                            if (l_Existing.f_FileSize == l_Local.f_FileSize)
                            {
                                // It seems to be the same file.

                                if (false == UpdateLocalFile(l_Existing, true, ref a_NameParts))
                                {
                                    if (null == l_OtherFileThatMustBeClosed)
                                    {
                                        if (true == a_MayIgnoreProtection ||
                                            false == l_Existing.p_IsLinkProtected)
                                        {
                                            DereferenceLocalFile(l_Existing);
                                            l_OtherFileThatMustBeClosed = l_Existing;
                                        }
                                    }
                                }
                            }
                            else
                            {
                                // l_Existing has a false name in our
                                // registry, so close it and try to give
                                // that name to l_Local.

                                if (null == l_OtherFileThatMustBeClosed)
                                {
                                    if (true == a_MayIgnoreProtection ||
                                        false == l_Existing.p_IsLinkProtected)
                                    {
                                        DereferenceLocalFile(l_Existing);
                                        l_OtherFileThatMustBeClosed = l_Existing;

                                        goto RETRY;
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        if (false == UpdateLocalFile(l_Local, false, ref a_NameParts))
                        {
                            f_FileReg.DerefFile(l_Local);
                        }
                        else
                        {
                            l_Local.AddToDic(f_UnprocessedLoc);
                            l_CloseLocalFile = false;
                        }
                    }
                }
            }
            LockLeave();

            if (true == l_CloseLocalFile)
            {
                l_Local.Close();
            }

            if (null != l_OtherFileThatMustBeClosed)
            {
                l_OtherFileThatMustBeClosed.Close();
            }
        }

        private bool MayAddOrUpdateFile(string a_Path,
                                        bool a_LockHeld)
        {
            string l_P0 = f_DontOpenExtractionFileName;
            string l_P1 = f_DontOpenRemovalFileName;
            string l_P2 = f_DontOpenMoveSrcFileName;
            string l_P3 = f_DontOpenMoveDestFileName;

            if (false == a_LockHeld)
            {
                Util.MemoryBarrier();
            }

            return (false == FileNames.AreNamesIdentical(l_P0, a_Path) &&
                    false == FileNames.AreNamesIdentical(l_P1, a_Path) &&
                    false == FileNames.AreNamesIdentical(l_P2, a_Path) &&
                    false == FileNames.AreNamesIdentical(l_P3, a_Path));
        }

        private bool UpdateLocalFile(LocalFile a_Local,
                                     bool a_AlsoUpdateRegistrationCasing,
                                     ref FileNameParts a_NameParts)
        {
            // The caller must hold f_Lock.

            string l_ParentDirName = a_Local.p_Parent.p_Path;
            ShareKinds l_ShareKind = f_IOPrefs.f_SharedDirs.GetShareKind(l_ParentDirName);

            if (ShareKinds.None == l_ShareKind)
            {
                return false;
            }

            a_Local.SetScanGenAndShareKind(f_IOPrefs.f_SharedDirs.p_ScanGen, l_ShareKind);

            if (true == a_AlsoUpdateRegistrationCasing)
            {
                if (false == FileNames.sp_TreatFileNamesCaseSensitive &&
                    false == a_Local.p_IsLinkProtected)
                {
                    ParentFolder l_Folder = a_Local.p_Parent;
                    f_FileReg.TryUpdateDirPath(l_Folder, a_NameParts.f_ParentDir);
                    l_Folder.TryUpdateFileName(a_Local, a_NameParts.f_FileName);
                }
            }

            return true;
        }

        private void CloseLocalFile(string a_Path,
                                    bool a_MayIgnoreProtection)
        {
            string l_DirName;
            string l_FileName = LongPath.SplitFullPath(a_Path, out l_DirName);

            if (null == l_DirName)
            {
                return;
            }

            LocalFile l_Local;

            LockEnter();
            {
                l_Local = f_FileReg.GetFile(l_DirName, l_FileName);

                if (null != l_Local)
                {
                    if (true == a_MayIgnoreProtection ||
                        false == l_Local.p_IsLinkProtected)
                    {
                        DereferenceLocalFile(l_Local);
                    }
                    else
                    {
                        l_Local = null;
                    }
                }
            }
            LockLeave();

            if (null != l_Local)
            {
                l_Local.Close();
            }
        }

        private uint RemoveCompleteFiles(bool a_IsHashWorker,
                                         Gui.FilesPanel.FileRemovalContext a_Cntxt,
                                         List<KnownFile> a_Files,
                                         bool a_JustMoveToRecycleBin)
        {
            // This routine may be called with a thread other than the
            // hash woker, but the call must be synchronized with the
            // hash worker in that case. Returns the number of files
            // successfully deleted. If a_JustMoveToRecycleBin is true,
            // this routine may block.

            uint l_Count = (uint)a_Files.Count;
            uint l_ElapsedCount = 0;
            uint l_SuccCount = 0;

            List<KnownFile> l_DelIncomplDataLs = new List<KnownFile>();
            List<KnownFile> l_DropHashSetsLs = new List<KnownFile>();

            foreach (KnownFile l_Known in a_Files)
            {
                if (true == a_Cntxt.p_Aborted)
                {
                    break;
                }

                if (true == RemoveSingleCompleteFile(l_Known, a_JustMoveToRecycleBin, l_DelIncomplDataLs, l_DropHashSetsLs))
                {
                    l_SuccCount++;
                }

                a_Cntxt.SetProgress(++l_ElapsedCount, l_Count);
            }

            DeleteIncomplDataAsync(l_DelIncomplDataLs);

            if (0 < l_DropHashSetsLs.Count)
            {
                // This can only be done by the hash worker.

                if (false == a_IsHashWorker)
                {
                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(DeleteHashSetsAsAPC, l_DropHashSetsLs);
                    f_HashQueue.Enqueue(l_Cookie);
                }
                else
                {
                    DeleteHashSets(l_DropHashSetsLs);
                }
            }

            return l_SuccCount;
        }

        private bool RemoveSingleCompleteFile(KnownFile a_Known,
                                              bool a_JustMoveToRecycleBin,
                                              List<KnownFile> a_DeleteIncomplDataLs,
                                              List<KnownFile> a_DropHashSetsLs)
        {
            // This routine may only be called by the hash worker.

            LocalFile l_Local;
            string l_FileName;

            LockEnter();
            {
                if (ComplStatus.Complete != a_Known.p_Status)
                {
                    LockLeave();

                    return false;
                }

                l_Local = a_Known.p_LocalFile;
                l_FileName = l_Local.BuildFullFileName();

                DereferenceLocalFile(l_Local);

                f_DontOpenRemovalFileName = l_FileName;
            }
            LockLeave();

            l_Local.Close();
            l_Local = null;
            bool l_Deleted = FHandle.Delete(l_FileName, a_JustMoveToRecycleBin);

            // Lockless zeroing.
            f_DontOpenRemovalFileName = null;

            if (false == l_Deleted)
            {
                // We failed to delete, so try to re-add the file
                // (it's probably a missing file currently).

                AddOrUpdateSingleFile(l_FileName, true);

                return false;
            }

            // The file was deleted.

            bool l_Dropped;
            LocalFile l_SecondFile = null;

            LockEnter();
            {
                l_Dropped = (ComplStatus.Dropped == a_Known.p_Status);

                if (false == l_Dropped)
                {
                    if (ComplStatus.Complete == a_Known.p_Status)
                    {
                        // The known file got reassociated with a
                        // local file in the meantime.

                        l_SecondFile = a_Known.p_LocalFile;
                        f_DontOpenRemovalFileName = l_SecondFile.BuildFullFileName();

                        DereferenceLocalFile(l_SecondFile);
                    }

                    CollectKeysOfFile(a_Known, a_DeleteIncomplDataLs, a_DropHashSetsLs);
                    DereferenceKnownFile(a_Known);
                }
            }
            LockLeave();

            if (true == l_Dropped)
            {
                return false;
            }

            if (null != l_SecondFile)
            {
                l_SecondFile.Close();
                f_DontOpenRemovalFileName = null;
            }

            return true;
        }

        private FileNum FindUnoccupiedFileNum()
        {
            // The caller must hold f_Lock.

            while (true == f_KnownFilesByFileNum.ContainsKey(new FileNum(f_NextFileNum)))
            {
                f_NextFileNum++;
            }

            return new FileNum(f_NextFileNum++);
        }

        private void ReferenceKnownFile(KnownFile a_Known,
                                        Dictionary<KnownFile, KnownFile> a_InitialDic)
        {
            // The caller must hold f_Lock.

            f_KnownFilesByFileNum.Add(a_Known.f_FileNum, a_Known);
            bool l_Added = f_KnownFilesBySize.Add(a_Known.f_FileSize, a_Known);

            AddPossibleFileID(a_Known);

            a_Known.AddToDic(a_InitialDic);
        }

        private void DereferenceKnownFile(KnownFile a_Known)
        {
            // The caller must hold f_Lock.

            bool l_Removed = f_KnownFilesByFileNum.Remove(a_Known.f_FileNum);

            l_Removed = f_KnownFilesBySize.Remove(a_Known.f_FileSize, a_Known);

            RemovePossibleFileID(a_Known);

            a_Known.AddToDic(null);
            a_Known.SetStatusToDropped();
        }

        private void DereferenceLocalFile(LocalFile a_Local)
        {
            DereferenceLocalFile(a_Local, true);
        }

        private void DereferenceLocalFile(LocalFile a_Local,
                                          bool a_AlsoDerefFromFileReg)
        {
            // The caller must hold f_Lock. NOTE: This routine does
            // not close the file.

            KnownFile l_Known = a_Local.p_KnownFile;

            if (null != l_Known)
            {
                l_Known.ClearLocalFile(f_Inst.GetTime());
                l_Known.AddToDic(f_UnlinkedKnw);
            }

            a_Local.AddToDic(null);

            if (true == a_AlsoDerefFromFileReg)
            {
                f_FileReg.DerefFile(a_Local);
            }
        }

        private bool DropLocalFile(LocalFile a_Local,
                                   KnownFile a_ConditionKnown)
        {
            // a_ConditionKnown can be null. Otherwise the local file
            // will only be dropped if it is linked to the known file
            // specified.

            bool l_Close;

            LockEnter();
            {
                if (null == a_Local.p_Parent ||
                    (null != a_ConditionKnown &&
                     a_Local.p_KnownFile != a_ConditionKnown))
                {
                    // This local file was already dropped or is not
                    // linked to the known file the caller specified,
                    // so abort.

                    l_Close = false;
                }
                else
                {
                    DereferenceLocalFile(a_Local);
                    l_Close = true;
                }
            }
            LockLeave();

            if (true == l_Close)
            {
                a_Local.Close();
            }

            return l_Close;
        }

        private void CollectLocalFileIfScanGenDiffers(uint a_ScanGen,
                                                      Dictionary<LocalFile, LocalFile> a_SrcDic,
                                                      List<LocalFile> a_DestList)
        {
            // The caller must hold f_Lock.

            foreach (LocalFile l_Local in a_SrcDic.Keys)
            {
                if (l_Local.p_ScanGen != a_ScanGen)
                {
                    a_DestList.Add(l_Local);
                }
            }
        }

#endregion

        [Conditional("DEBUG")]
        private void DebugAssertThread(string a_ExpectedName)
        {
            if (Thread.CurrentThread.Name != a_ExpectedName)
            {
                throw(new InvalidOperationException());
            }
        }

        private void StartWorkers()
        {
            f_RunScanWorker = true;
            f_RunHashWorker = true;
            f_RunIdentWorker = true;
            f_RunReadComplWorker = true;
            f_RunIncomplWorker = true;
            Util.MemoryBarrier();

            f_ScanWorkerFinished.Reset();
            f_HashWorkerFinished.Reset();
            f_IdentWorkerFinished.Reset();
            f_ReadComplWorkerFinished.Reset();

            Util.StartThread(c_ScanThreadName, ScanWorker);
            Util.StartThread(c_HashThreadName, HashWorker, ThreadPriority.BelowNormal);
            Util.StartThread(c_IdentThreadName, IdentificationWorker, ThreadPriority.BelowNormal);
            Util.StartThread(c_ReadThreadName, ReadComplWorker);

            for (int i = 0; f_IncomplWorkers > i; i++)
            {
                f_IncomplWorkerFinished[i].Reset();

                Util.StartThreadWithParam(c_IncomplThreadsNameBeginning + i, IncomplWorker, ThreadPriority.Normal, i);
            }
        }

        private void StopWorkers()
        {
            Software.UpdateShutDownState(0x80);

            f_RunScanWorker = false;
            f_RunHashWorker = false;
            f_RunIdentWorker = false;
            f_RunReadComplWorker = false;
            f_RunIncomplWorker = false;
            Util.MemoryBarrier();

            MonitorEvent[] l_Events = new MonitorEvent[4 + f_IncomplWorkers];

            l_Events[0] = f_ScanWorkerFinished;
            l_Events[1] = f_HashWorkerFinished;
            l_Events[2] = f_IdentWorkerFinished;
            l_Events[3] = f_ReadComplWorkerFinished;

            for (int i = 0; f_IncomplWorkers > i; i++)
            {
                l_Events[4 + i] = f_IncomplWorkerFinished[i];
            }

            Software.UpdateShutDownState(0x100);

            MonitorEvent.WaitAll(l_Events);
            FileSystemWatcher[] l_Watchers;

            LockEnter();
            {
                l_Watchers = f_Watchers;
                f_Watchers = null;
            }
            LockLeave();

            if (null != l_Watchers)
            {
                SharedDirectories.DestroyWatchers(l_Watchers);
            }

            Software.UpdateShutDownState(0x200);

            while (0 != f_WatchersThreadsPresent.p_Count)
            {
                Util.MiniSleep();
            }

            Software.UpdateShutDownState(0x400);
        }

        private void TriggerScan()
        {
            // The caller must hold f_Lock.

            f_ScanState = ScanState.NotStarted;
            f_IOPrefs.f_SharedDirs.IncScanGeneration();
            f_IOPrefs.f_SharedDirs.ResetFileBrowser();
        }

        private void ScanWorker()
        {
            f_ScanWorkerFinished.Reset();

            uint l_FilesGen = 0;
            const int l_PathsBufElements = 50;
            string[] l_PathsBuffer = new string[l_PathsBufElements];
            NB.Gui.MainForm.FormInput l_FormInput = f_Inst.f_MainForm.f_Input;
            Time l_IgnoreTime = new Time();
            GeneralAPCCookie[] l_APCCookies = null;

            while (true == Util.VolatileRead(ref f_RunScanWorker))
            {
                if (null == l_APCCookies)
                {
                    l_APCCookies = f_ScanQueue.GetNow();
                }

                Time l_Time = f_Inst.UpdateTime();

                // First see if the GUI demands an update.

                Time l_NextFileUpdateTime = Time.FromBinary((ulong)System.Threading.Interlocked.Read(ref l_FormInput.f_NextFileUpdateTime));

                if (false == l_NextFileUpdateTime.Equals(l_IgnoreTime) &&
                    true == l_NextFileUpdateTime.IsEarlierThan(l_Time))
                {
                    Category l_Category = Util.VolatileRead(ref l_FormInput.f_SubscribedFileCategory);
                    KnownFile[] l_FilesForGui = GetFilesByCategory(l_Category);

                    NB.Gui.MainForm.FormInput.Files f_Files = new NB.Gui.MainForm.FormInput.Files(++l_FilesGen, l_Category, l_FilesForGui);
                    Util.VolatileWrite(out l_FormInput.f_Files, f_Files);

                    l_IgnoreTime = l_NextFileUpdateTime;
                }

                // Now handle any async tasks.

                if (null != l_APCCookies)
                {
                    APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
                }

                // Now scan for files.

                SharedDirectories l_SharedDirs = null;

                LockEnter();
                {
                    if (true == f_IOPrefs.f_WatchFileSystem &&
                        ScanState.Finished == f_ScanState &&
                        l_Time >= f_NextAutoScanTime)
                    {
                        // Time for an automatic scan.

                        TriggerScan();
                    }

                    if (ScanState.Finished != f_ScanState)
                    {
                        f_ScanState = ScanState.Started;

                        l_SharedDirs = f_IOPrefs.f_SharedDirs;
                    }
                }
                LockLeave();

                if (null != l_SharedDirs)
                {
                    // Continue scanning.
                    int l_Count = l_SharedDirs.FindNextFiles(l_PathsBuffer, l_PathsBufElements);

                    if (0 < l_Count)
                    {
                        AddOrUpdateFiles(l_PathsBuffer, l_Count, false);

                        continue;
                    }
                    else
                    {
                        Array.Clear(l_PathsBuffer, 0, l_PathsBufElements);

                        // We finished scanning, so remove left over
                        // files that once we're shared but aren't any
                        // longer.

                        List<LocalFile> l_RemLs = new List<LocalFile>();

                        LockEnter();
                        {
                            // Proceed if there wasn't a reset in the
                            // meantime.
                            if (ScanState.Started == f_ScanState)
                            {
                                uint l_ScanGen = l_SharedDirs.p_ScanGen;
                                CollectLocalFileIfScanGenDiffers(l_ScanGen, f_UnprocessedLoc, l_RemLs);
                                CollectLocalFileIfScanGenDiffers(l_ScanGen, f_UnhashedLoc, l_RemLs);
                                CollectLocalFileIfScanGenDiffers(l_ScanGen, f_RecognizedLoc, l_RemLs);

                                foreach (LocalFile l_Local in l_RemLs)
                                {
                                    if (false == l_Local.p_IsLinkProtected)
                                    {
                                        DereferenceLocalFile(l_Local);
                                    }
                                }

                                uint l_Interval = (false == Util.s_IsRunningOnWine ? c_ScanDiskInterval : (c_ScanDiskInterval / 2));
                                f_ScanState = ScanState.Finished;
                                f_NextAutoScanTime = l_Time.AddSeconds(l_Interval);
                            }
                        }
                        LockLeave();

                        foreach (LocalFile l_Local in l_RemLs)
                        {
                            if (null == l_Local.p_Parent)
                            {
                                l_Local.Close();
                            }
                        }
                    }
                }

                l_APCCookies = f_ScanQueue.WaitForInput();
            }

            if (null != l_APCCookies)
            {
                APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
            }

            f_ScanQueue.ShutDown();

            f_ScanWorkerFinished.Set();

            Software.UpdateShutDownState(0x800000);
        }

#region Hashing

        private bool HandleHashing(ref byte[] a_TempBuffer,
                                   uint a_MaxBytes)
        {
            // Returns true if a subsequent call to this routine might
            // progress things. a_TempBuffer may be null. This routine
            // may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            if (false == f_HashCntxt.IsValid())
            {
                KnownFile l_Known;
                LocalFile l_Local;

                LockEnter();
                {
                    if (false == Util.GetAny<KnownFile>(f_UnhashedKnw.Keys, out l_Known))
                    {
                        // There are no files left to hash anymore.

                        l_Local = null;
                    }
                    else
                    {
                        l_Local = l_Known.p_LocalFile;
                    }
                }
                LockLeave();

                if (null == l_Local)
                {
                    return false;
                }

                l_Known.CalcHashSetLayout();

                bool l_Succ;
                HardMemory.Element l_Elm;

                f_Inst.f_MemResource.AcquireExclusive();
                {
                    l_Succ = LoadOrCreateHashElement(l_Known, out l_Elm);
                }
                f_Inst.f_MemResource.ReleaseExclusive();

                if (false == l_Succ)
                {
                    return false;
                }

                if (null == a_TempBuffer)
                {
                    a_TempBuffer = new byte[c_DefaultBufferSize];
                }

                l_Known.f_HashSetKey = l_Elm.f_ElmKey;
                HashSet.HashSetData l_SetData = new HashSet.HashSetData(f_Inst.f_Mem, l_Elm, l_Local, a_TempBuffer);
                f_HashCntxt = new HashingContext(l_Known, l_Local, l_SetData);
                f_HashCntxt.f_HashSetType = P2PHashSets.eD2kHashSet;
            }

            bool l_NeedToInitHashSet = (null == f_HashCntxt.f_HashSet);

            if (true == l_NeedToInitHashSet)
            {
                f_HashCntxt.f_HashSet = BuildHashSet(f_HashCntxt.f_HashSetType, f_HashCntxt.f_SetData, f_HashCntxt.f_Known);
            }

            HashingStatus l_Status;
            byte[] l_MasterHash = null;

            f_Inst.f_MemResource.AcquireShared();
            {
                if (true == l_NeedToInitHashSet)
                {
                    bool l_Initialized = f_HashCntxt.f_HashSet.InitializeFromDisk();

                    // Don't mind if this failed, since we will
                    // overwrite in that case.
                }

                l_Status = f_HashCntxt.f_HashSet.ContinueHashing(ref a_MaxBytes);

                if (HashingStatus.Done == l_Status)
                {
                    if (false == f_HashCntxt.f_HashSet.SaveProgress())
                    {
                        Util.DebugBreak();

                        l_Status = HashingStatus.DestFileAccessFailed;
                    }

                    l_MasterHash = new byte[f_HashCntxt.f_HashSet.f_HashSize];

                    if (false == f_HashCntxt.f_HashSet.GetMasterHash(l_MasterHash, 0))
                    {
                        l_Status = HashingStatus.DestFileAccessFailed;
                        l_MasterHash = null;
                    }
                }
            }
            f_Inst.f_MemResource.ReleaseShared();

            if (HashingStatus.NeedFurtherCalls == l_Status)
            {
                return true;
            }

            if (HashingStatus.Done == l_Status)
            {
                KnownFile l_Known = f_HashCntxt.f_Known;

                LockEnter();
                {
                    if (ComplStatus.Complete != l_Known.p_Status)
                    {
                        // This can happen if the status was changed
                        // by another thread than the hash worker.

                        f_HashCntxt = new HashingContext();

                        LockLeave();

                        return true;
                    }

                    l_Known.f_MasterHashes.SetHash(f_HashCntxt.f_HashSetType, l_MasterHash, 0);

                    switch (f_HashCntxt.f_HashSetType)
                    {
                        case P2PHashSets.eD2kHashSet:
                        {
                            RemovePossibleFileID(l_Known);

                            l_Known.f_FileID = l_Known.f_MasterHashes.CreateFileID();
                            AddPossibleFileID(l_Known);
                        }
                        break;

                        case P2PHashSets.TAICHTree:
                        {
                            // We're done.

                            l_Known.AdjustAllHashesSelfCalculatedFlag(true);

                            if (true == l_Known.IsInDic(f_UnhashedKnw))
                            {
                                l_Known.AddToDic(f_HashedKnw);
                            }
                        }
                        break;
                    }
                }
                LockLeave();

                f_HashCntxt.f_HashSet = null;

                switch (f_HashCntxt.f_HashSetType)
                {
                    case P2PHashSets.eD2kHashSet:
                        f_HashCntxt.f_HashSetType = P2PHashSets.AICHTree;
                        break;

                    case P2PHashSets.AICHTree:
                        f_HashCntxt.f_HashSetType = P2PHashSets.TAICHTree;
                        break;

                    case P2PHashSets.TAICHTree:
                        f_HashCntxt = new HashingContext();
                        break;

                    default: throw (new ApplicationException());
                }

                return true;
            }

            if (HashingStatus.SrcFileAccessFailed == l_Status)
            {
                DropLocalFile(f_HashCntxt.f_Local, f_HashCntxt.f_Known);
            }

            f_HashCntxt = new HashingContext();

            // Take a break.

            return false;
        }

        private bool LoadOrCreateHashElement(KnownFile a_Known,
                                             out HardMemory.Element a_Elm)
        {
            if (false == IsValidHashElmKey(a_Known))
            {
                a_Elm = null;
            }
            else
            {
                // Try to load the existing element.

                a_Elm = f_Inst.f_Mem.GetElement(a_Known.f_HashSetKey);

                if (null != a_Elm)
                {
                    if (a_Known.f_HashElmLayout.f_ElementSize != a_Elm.f_MaxSize ||
                        false == KnownFile.HashElmLayout.CheckHeader(a_Known, f_Inst.f_Mem, a_Elm))
                    {
                        Util.DebugBreak();

                        // Do not delete the element in this case, it
                        // might belong to another file.

                        a_Elm = null;
                    }
                }
            }

            if (null == a_Elm)
            {
                if (true == a_Known.f_HashSetKey.IsNonzero())
                {
                    a_Known.ClearHashSetKey();
                }

                // Create a new element.

                if (true == f_CriticalDiskSpaceLevelReached)
                {
                    return false;
                }

                if (c_CheckFreeDiskSpaceHashSetSizeThreshold <= a_Known.f_HashElmLayout.f_ElementSize &&
                    false == Software.HasAppDataDriveAvailBytes(false, a_Known.f_HashElmLayout.CalcMinAvailDiskSpaceForCreation()))
                {
                    // This is a large hash set, so we checked the
                    // available disk space right here and found out
                    // that it's too small.

                    return false;
                }

                a_Elm = f_Inst.f_Mem.Allocate(f_Inst.f_Mem.FindUnoccupiedKey(Instance.c_ReservedMemElmKeys), a_Known.f_HashElmLayout.f_ElementSize);

                if (null == a_Elm)
                {
                    // Out of disk or so.

                    return false;
                }

                if (false == KnownFile.HashElmLayout.WriteHeader(a_Known, f_Inst.f_Mem, a_Elm))
                {
                    // Remove the element just created.
                    bool l_Deleted = f_Inst.f_Mem.Delete(a_Elm);

                    a_Elm = null;

                    return false;
                }
            }

            return true;
        }

        public void BuildFileDescriptor(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            FileDescriptorCookie l_Cookie = (FileDescriptorCookie)a_Cookie;
            KnownFile l_Known = l_Cookie.f_Known;

            ComplStatus l_Status;
            bool l_HasAll;
            MasterHashes l_MasterHashes;

            LockEnter();
            {
                l_Status = l_Known.p_Status;

                if (ComplStatus.Complete == l_Status)
                {
                    l_HasAll = l_Known.HasTrustedHashes();
                    l_Known.f_MasterHashes.Clone(out l_MasterHashes);
                }
                else
                {
                    l_HasAll = false;
                    l_MasterHashes = new MasterHashes();
                }
            }
            LockLeave();

            switch (l_Status)
            {
                case ComplStatus.Incomplete:
                    // Pass this to a possible incomplete file.
                    AddSingleIncomplFileOp(new IncomplOperationContext(l_Cookie));
                    break;

                case ComplStatus.Complete:
                case ComplStatus.Missing:
                    {
                        ByteArray l_Data;

                        if (false == l_MasterHashes.ExposeHash(P2PHashSets.eD2kHashSet, out l_Data))
                        {
                            goto case ComplStatus.Dropped;
                        }

                        MemoryIO l_Buffer = MemoryIO.CreateWritable();
                        l_Buffer.WriteBytes(ref l_Data);
                        l_Buffer.WriteUInt64LE(l_Known.f_FileSize);

                        if (false == l_HasAll)
                        {
                            l_Buffer.ZeroBytes(sizeof(ushort));
                        }
                        else
                        {
                            l_Buffer.WriteUInt16LE((ushort)P2PHashSets._ALL_HASHSETS_INCL_SUBSETS_);
                        }

                        l_MasterHashes.ToBuffer(l_Buffer);

                        l_Cookie.Succeed(l_Buffer.ToArray(true));
                    }
                    break;

                case ComplStatus.Dropped:
                    l_Cookie.Fail();
                    break;

                default: throw (new ApplicationException());
            }
        }

        public void HandleHashQuery(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            HashQueryCookie l_Cookie = (HashQueryCookie)a_Cookie;
            KnownFile l_Known = l_Cookie.f_Known;
            ComplStatus l_Status;

            LockEnter();
            {
                l_Status = l_Known.p_Status;

                if (ComplStatus.Dropped != l_Status &&
                    false == l_Known.f_MasterHashes.IsHashEqual(l_Cookie.f_HashType, l_Cookie.f_MasterHash, 0))
                {
                    // The caller is thinking of a different master
                    // hash and therefore we can't satisfy the request.

                    l_Status = ComplStatus.Dropped;
                }
            }
            LockLeave();

            if (ComplStatus.Dropped == l_Status)
            {
                l_Cookie.Fail();

                return;
            }

            if (ComplStatus.Incomplete == l_Status)
            {
                // Try to pass this down to the IncompleteFile class.

                if (P2PHashSets.AICHTree == l_Cookie.f_HashType ||
                    false == AddSingleIncomplFileOp(new IncomplOperationContext(l_Cookie)))
                {
                    l_Cookie.Fail();
                }

                return;
            }

            HardMemory.Element l_Elm;

            if (false == IsValidHashElmKey(l_Known))
            {
                l_Cookie.Fail();

                return;
            }

            l_Known.CalcHashSetLayout();

            bool l_MalformedInput = false;

            f_Inst.f_MemResource.AcquireShared();

            try
            {
                // Try to load the existing element.

                l_Elm = f_Inst.f_Mem.GetElement(l_Known.f_HashSetKey);

                if (null == l_Elm ||
                    l_Known.f_HashElmLayout.f_ElementSize != l_Elm.f_MaxSize ||
                    false == KnownFile.HashElmLayout.CheckHeader(l_Known, f_Inst.f_Mem, l_Elm))
                {
                    Util.DebugBreak();

                    l_Known.ClearHashSetKey();

                    return;
                }

                // We don't need to hold the lock during tree creation
                // (it can be expensive).

                HashSet.HashSetData l_SetData;
                HashSet l_HashSet;

                f_Inst.f_MemResource.ReleaseShared();
                {
                    l_SetData = new HashSet.HashSetData(f_Inst.f_Mem, l_Elm, l_Known.f_FileSize);
                    l_HashSet = BuildHashSet(l_Cookie.f_HashType, l_SetData, l_Known);
                }
                f_Inst.f_MemResource.AcquireShared();

                if (false == l_HashSet.InitializeFromDisk())
                {
                    Util.DebugBreak();

                    return;
                }

                bool l_OK;
                byte[] l_Dest;

                switch (l_Cookie.f_Kind)
                {
                    case HashQueryKinds.Indices:
                        l_OK = l_HashSet.CheckInputAndGetHashes(l_Cookie.f_HashIndices, out l_Dest, out l_MalformedInput);
                        break;

                    case HashQueryKinds.Range:
                        l_OK = l_HashSet.CheckInputAndGetHashes(l_Cookie.f_StartingIndex, l_Cookie.f_Count, out l_Dest, out l_MalformedInput);
                        break;

                    case HashQueryKinds.AICHPartRecoveryData:
                    {
                        AICHTree l_AICHTree = (AICHTree)l_HashSet;

                        l_AICHTree.StartRecoveryOperations();
                        {
                            // Use f_PartCount here, not f_PartCountHashSet.
                            l_OK = l_AICHTree.CheckInputAndCreatePartRecoveryData(l_Cookie.f_AICHTrgtPartIndex, l_Cookie.f_Known.f_PartCount, out l_Dest, out l_MalformedInput);
                        }
                        l_AICHTree.EndRecoveryOperations();

                        break;
                    }

                    default: throw (new ApplicationException());
                }

                if (true == l_OK)
                {
                    l_Cookie.Succeed(l_Dest);
                }
            }
            finally
            {
                f_Inst.f_MemResource.ReleaseShared();

                if (CookieStates.Pending == l_Cookie.p_State)
                {
                    l_Cookie.Fail(l_MalformedInput);
                }
            }
        }

        private HashSet BuildHashSet(P2PHashSets a_HashSetType,
                                     HashSet.HashSetData a_SetData,
                                     KnownFile a_Known)
        {
            switch (a_HashSetType)
            {
                case P2PHashSets.eD2kHashSet:
                    return eD2kHashSet.CreateeD2kHashSet(a_SetData, a_Known.f_HashElmLayout.p_eD2kOffset);

                case P2PHashSets.AICHTree:
                    return AICHTree.CreateAICHTree(a_SetData, a_Known.f_HashElmLayout.f_AICHOffset);

                case P2PHashSets.TAICHTree:
                    return TAICHTree.CreateTAICHTree(a_SetData, a_Known.f_HashElmLayout.f_TAICHOffset);

                default: throw (new ApplicationException());
            }
        }

        private void RemoveFiles(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            Gui.FilesPanel.FileRemovalContext l_Cntxt = (Gui.FilesPanel.FileRemovalContext)a_Cookie.GetArg();
            KnownFile[] l_InputFiles = l_Cntxt.f_Files;
            FileRemovalModes l_RemovalMode = l_Cntxt.f_RemovalMode;

            uint l_SuccCount = 0;
            List<KnownFile> l_DelIncomplDataLs = new List<KnownFile>();
            List<KnownFile> l_DropHashSetsLs = new List<KnownFile>();
            List<KnownFile> l_RemoveCompletedLs = new List<KnownFile>();

            // NOTE: The progress of l_Cntxt will only be updated when
            // complete files are deleted from disk (see
            // RemoveCompleteFiles()).

            LockEnter();
            {
                foreach (KnownFile l_Known in l_InputFiles)
                {
                    if (true == l_Cntxt.p_Aborted)
                    {
                        break;
                    }

                    ComplStatus l_Status = l_Known.p_Status;

                    if (ComplStatus.Dropped == l_Status)
                    {
                        continue;
                    }

                    if (ComplStatus.Incomplete == l_Status ||
                        ComplStatus.Missing == l_Status)
                    {
                        CollectKeysOfFile(l_Known, l_DelIncomplDataLs, l_DropHashSetsLs);
                        DereferenceKnownFile(l_Known);
                        l_Known.AdjustAllHashesSelfCalculatedFlag(false);

                        l_SuccCount++;
                    }
                    else
                    {
                        if (FileRemovalModes.DoNothing != l_RemovalMode)
                        {
                            l_RemoveCompletedLs.Add(l_Known);
                        }
                    }
                }
            }
            LockLeave();

            bool l_AsyncOpStarted = false;

            if (0 < l_RemoveCompletedLs.Count &&
                false == l_Cntxt.p_Aborted)
            {
                // Handle complete files.

                TryReleaseRecycleComplFilesCntxt();

                // We refuse to do this if there is a thread active
                // in parallel, so check.
                if (null == f_RecycleComplFilesCntxt)
                {
                    if (FileRemovalModes.DeletePermanently == l_RemovalMode)
                    {
                        l_SuccCount += RemoveCompleteFiles(true, l_Cntxt, l_RemoveCompletedLs, false);
                    }
                    else
                    {
                        // We have to do this with the help of an
                        // own thread, since the operation can take
                        // infinite time to complete.

                        f_RecycleComplFilesCntxt = new MoveCompleteFilesToRecycleBinContext(this, l_Cntxt, l_RemoveCompletedLs, l_SuccCount);
                        f_RecycleComplFilesCntxt.Start();
                        l_AsyncOpStarted = true;
                    }
                }
            }

            // Release adjacent data of non-complete files (do this
            // even if we were aborted).
            DeleteIncomplDataAsync(l_DelIncomplDataLs);
            DeleteOrResetHashSetsCore(l_DropHashSetsLs, false);

            if (false == l_AsyncOpStarted)
            {
                // Finish right away.

                l_Cntxt.SetFinished(l_SuccCount);
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        private void TryReleaseRecycleComplFilesCntxt()
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            if (null != f_RecycleComplFilesCntxt &&
                true == f_RecycleComplFilesCntxt.f_ParentCntxt.p_Finished)
            {
                f_RecycleComplFilesCntxt = null;
            }
        }

        private void CollectKeysOfFile(KnownFile a_Known,
                                       List<KnownFile> a_DelIncomplDataLs,
                                       List<KnownFile> a_DropHashSetsLs)
        {
            // The caller must hold f_Lock.

            if (true == a_Known.f_IncomplKey.IsNonzero())
            {
                // This complete file still has it's
                // incomplete data. Try to delete it now.

                a_DelIncomplDataLs.Add(a_Known);
            }

            // NOTE: It's not completly reliable if this is done by a
            // different thread than the hash worker, but it's deemed
            // okay.
            if (true == a_Known.f_HashSetKey.IsNonzero())
            {
                a_DropHashSetsLs.Add(a_Known);
            }
        }

        private void TurnMissingIntoIncomplFiles(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            KnownFile[] l_InputFiles = a_Cookie.GetArgAsKnownFileArray();
            List<KnownFile> l_DropHashSetsLs = new List<KnownFile>();
            uint l_Count = 0;

            LockEnter();
            {
                foreach (KnownFile l_Known in l_InputFiles)
                {
                    if (ComplStatus.Missing == l_Known.p_Status)
                    {
                        l_Count++;
                        l_Known.TurnMissingIntoIncompl();
                        l_Known.AdjustAllHashesSelfCalculatedFlag(false);

                        if (true == l_Known.f_HashSetKey.IsNonzero())
                        {
                            l_DropHashSetsLs.Add(l_Known);
                        }
                    }
                }
            }
            LockLeave();

            DeleteOrResetHashSetsCore(l_DropHashSetsLs, false);

            a_Cookie.SetReturnObjectAndState(l_Count, CookieStates.Success);
        }

        private void RecheckCompleteFiles(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            KnownFile[] l_InputFiles = a_Cookie.GetArgAsKnownFileArray();
            Time l_Time = f_Inst.GetTime();
            List<KnownFile> l_ResetHashSetsLs = new List<KnownFile>(l_InputFiles.Length);

            LockEnter();
            {
                foreach (KnownFile l_Known in l_InputFiles)
                {
                    if (ComplStatus.Complete == l_Known.p_Status)
                    {
                        LocalFile l_Local = l_Known.p_LocalFile;

                        l_ResetHashSetsLs.Add(l_Known);

                        l_Known.ClearLocalFile(l_Time);
                        l_Known.MakeFastDetectionImpossible();

                        l_Known.AddToDic(f_UnlinkedKnw);
                        l_Local.AddToDic(f_UnprocessedLoc);

                        RemovePossibleFileID(l_Known);

                        l_Known.AdjustAllHashesSelfCalculatedFlag(false);
                        l_Known.f_MasterHashes = new MasterHashes();
                        l_Known.f_FileID = new FileID();
                    }
                }
            }
            LockLeave();

            DeleteOrResetHashSetsCore(l_ResetHashSetsLs, true);

            a_Cookie.SetReturnObjectAndState((uint)l_ResetHashSetsLs.Count, CookieStates.Success);
        }

        private void RescanForFinishedFiles(Time a_Time)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            List<KnownFile> l_DelIncomplDataLs = new List<KnownFile>();

            LockEnter();
            {
                foreach (KnownFile l_Known in f_KnownFilesByFileNum.Values)
                {
                    if (ComplStatus.Incomplete == l_Known.p_Status)
                    {
                        if (c_MaxFinishedIncomplFilesCount <= f_FinishedIncompl.p_Count)
                        {
                            continue;
                        }

                        IncomplFileProgress l_IncomplProgress;
                        l_IncomplProgress = Util.VolatileRead(ref l_Known.f_IncomplProgress);

                        if (null != l_IncomplProgress &&
                            true == l_IncomplProgress.f_DoneMap.p_AllSet)
                        {
                            AddFinishedIncomplFile(l_Known);
                        }
                    }
                    else
                    {
                        if (true == l_Known.f_IncomplKey.IsNonzero())
                        {
                            // This complete file still has it's
                            // incomplete data. Try to delete it now.

                            l_DelIncomplDataLs.Add(l_Known);
                        }
                    }
                }
            }
            LockLeave();

            DeleteIncomplDataAsync(l_DelIncomplDataLs);
        }

        private void DeleteHashSetsAsAPC(GeneralAPCCookie a_Cookie)
        {
            // An apc wrapper around DeleteHashSets().

            List<KnownFile> l_Ls = (List<KnownFile>)a_Cookie.GetArg();
            DeleteHashSets(l_Ls);

            a_Cookie.SetState(CookieStates.Success);
        }

        private void DeleteHashSets(List<KnownFile> a_Files)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            // We don't update the dictionary the passed files are in.

            if (0 == a_Files.Count)
            {
                return;
            }

            // First reset the flag.

            LockEnter();
            {
                foreach (KnownFile l_Known in a_Files)
                {
                    l_Known.AdjustAllHashesSelfCalculatedFlag(false);
                }
            }
            LockLeave();

            DeleteOrResetHashSetsCore(a_Files, false);
        }

        private void DeleteOrResetHashSetsCore(List<KnownFile> a_Files,
                                               bool a_JustReset)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            if (0 == a_Files.Count)
            {
                return;
            }

            f_Inst.f_MemResource.AcquireExclusive();
            {
                foreach (KnownFile l_Known in a_Files)
                {
                    if (f_HashCntxt.f_Known == l_Known)
                    {
                        // Stop hashing this file.

                        f_HashCntxt = new HashingContext();
                    }

                    // The flag should be clear.

                    if (true == l_Known.f_HashSetKey.IsNonzero())
                    {
                        bool l_Succ = KnownFile.HashElmLayout.DeleteOrResetChecked(l_Known, f_Inst.f_Mem, l_Known.f_HashSetKey, a_JustReset);

                        if (false == a_JustReset)
                        {
                            l_Known.ClearHashSetKey();
                        }
                    }
                }
            }
            f_Inst.f_MemResource.ReleaseExclusive();
        }

        private bool ContinueExtraction(ref ExtractionContext a_Cntxt)
        {
            // Returns true if a subsequent call to this routine
            // might progress things. This routine may only be called
            // by the hash worker.
            DebugAssertThread(c_HashThreadName);

            if (true == a_Cntxt.IsValid())
            {
                if (false == ContinueExtractionCore(ref a_Cntxt))
                {
                    AbortExtraction(ref a_Cntxt);

                    return (0 < f_FinishedIncompl.p_Count);
                }

                return true;
            }

            int l_Count = f_FinishedIncompl.p_Count;

            if (0 == l_Count)
            {
                return false;
            }

            KnownFile l_Known = f_FinishedIncompl.p_Arr[0];
            string l_TrgtDirName;
            string l_TrgtFileName;

            LockEnter();
            {
                l_TrgtDirName = f_IOPrefs.f_SharedDirs.GetCompletionTargetPath();

                if (null == l_TrgtDirName)
                {
                    l_TrgtFileName = null;
                }
                else
                {
                    if (ComplStatus.Incomplete == l_Known.p_Status)
                    {
                        l_TrgtFileName = l_Known.GetFileName(true);
                    }
                    else
                    {
                        l_TrgtFileName = null;
                    }
                }
            }
            LockLeave();

            if (null == l_TrgtDirName)
            {
                // We can't extract due to lack of a target dir. Inform
                // the GUI.

                f_Inst.InvokeMainFormAsync(new ArglessHandler(f_Inst.f_MainForm.NotifyMissingOutputDirectory));

                return false;
            }

            f_FinishedIncompl.DeleteAt(0);

            if (null == l_TrgtFileName)
            {
                // The file wasn't incomplete it seems, so return soon
                // to pick another.

                return true;
            }

            string l_FullFileName;

            if (false == FileNames.FindUnoccupiedFileName(l_TrgtDirName, l_TrgtFileName, out l_FullFileName))
            {
                return true;
            }

            TemporaryFile l_TrgtFile;
            ExtractionCookie.Result l_Result = ExtractionCookie.CreateTempFileForCompletion(l_FullFileName, l_Known.f_FileSize, out l_TrgtFile);

            if (ExtractionCookie.Result.Success != l_Result)
            {
                if (ExtractionCookie.Result.NotEnoughDiskSpace == l_Result ||
                    ExtractionCookie.Result.LowDiskSpace == l_Result)
                {
                    f_Inst.InvokeMainFormAsync(new ArglessHandler(f_Inst.f_MainForm.NotifyFailedToExtractDueToInsufficientDiskSpace));
                }
                // Else the file name was rejected by the OS maybe
                // (Result.FailedToCreateFile).

                return false;
            }

            // Load the incomplete file.

            SourceCookie l_SourceCookie = new SourceCookie(UserNum.s_ComplFileExtraction);

            if (false == RegisterSourceAsync(l_Known, l_SourceCookie))
            {
                l_TrgtFile.Close();

                return false;
            }

            a_Cntxt = new ExtractionContext(l_Known, l_SourceCookie, l_TrgtFile, l_FullFileName);

            return true;
        }

        private bool ContinueExtractionCore(ref ExtractionContext a_Cntxt)
        {
            // Returns false if the extraction should be aborted.

            if (ExtractionContext.Stages.Registering == a_Cntxt.f_Stage)
            {
                CookieStates l_SourceState = a_Cntxt.f_SourceCookie.p_ContentMngrStateVolatile;

                if (CookieStates.Pending == l_SourceState)
                {
                    return true;
                }

                if (CookieStates.Success != l_SourceState)
                {
                    return false;
                }

                a_Cntxt.f_Stage = ExtractionContext.Stages.Extracting;
            }

            if (ExtractionContext.Stages.Extracting == a_Cntxt.f_Stage)
            {
                if (null != a_Cntxt.f_ExtractionCookie)
                {
                    CookieStates l_ExtractionState = a_Cntxt.f_ExtractionCookie.p_StateVolatile;

                    if (CookieStates.Pending == l_ExtractionState)
                    {
                        return true;
                    }

                    if (CookieStates.Success != l_ExtractionState)
                    {
                        return false;
                    }
                }

                if (false == a_Cntxt.BuildNextExtractionCookie())
                {
                    // Extraction is done.

                    if (false == a_Cntxt.f_TempFile.ExposeFHandle().Flush())
                    {
                        return false;
                    }

                    if (true == a_Cntxt.f_PreserveWarnMessage)
                    {
                        ByteArray l_FirstSectorData = new ByteArray(a_Cntxt.f_FirstSector);

                        if (false == a_Cntxt.f_TempFile.ExposeFHandle().Write(0, ref l_FirstSectorData) ||
                            false == a_Cntxt.f_TempFile.ExposeFHandle().Flush())
                        {
                            return false;
                        }
                    }

                    a_Cntxt.f_Stage = ExtractionContext.Stages.Hashing;
                }
                else
                {
                    // Extract the next chunk.

                    ExtractionCookie l_ExtCookie = a_Cntxt.f_ExtractionCookie;

                    if (false == ExtractAsync(a_Cntxt.f_Known, l_ExtCookie))
                    {
                        return false;
                    }

                    if (true == a_Cntxt.f_PreserveWarnMessage &&
                        0 == l_ExtCookie.f_FilePos)
                    {
                        a_Cntxt.f_FirstSector = l_ExtCookie.f_StartBuffer;
                    }

                    return true;
                }
            }

            if (ExtractionContext.Stages.Hashing == a_Cntxt.f_Stage)
            {
                // We're done extracting and will now ident-hash the file.

                HashingStatus l_HashingStatus = a_Cntxt.Hash(c_DefaultIdentItemSize);

                if (HashingStatus.NeedFurtherCalls == l_HashingStatus)
                {
                    return true;
                }

                if (HashingStatus.Done != l_HashingStatus)
                {
                    return false;
                }

                a_Cntxt.Unregister(this);

                bool l_KeepLocal = true;
                bool l_KeptKnown = true;
                string l_KnownFileName;
                string l_ExistingFullFileName = null;

                LockEnter();
                {
                    if (ComplStatus.Incomplete != a_Cntxt.f_Known.p_Status)
                    {
                        LockLeave();

                        return false;
                    }

                    l_KnownFileName = a_Cntxt.f_Known.GetFileName(true);

                    // Check if we already know this file.

                    MultiMap<ulong, KnownFile>.Enumerator l_Enum;
                    if (true == f_KnownFilesBySize.CreateEnumerator(a_Cntxt.f_Known.f_FileSize, out l_Enum))
                    {
                        KnownFile l_Other;
                        while (true == l_Enum.GetNext(out l_Other))
                        {
                            if ((ComplStatus.Complete == l_Other.p_Status ||
                                 ComplStatus.Missing == l_Other.p_Status) &&
                                true == l_Other.HasTheseHashes(a_Cntxt.f_Hashes))
                            {
                                // A duplicate, destroy a_Cntxt.f_Known.

                                if (ComplStatus.Complete == l_Other.p_Status)
                                {
                                    // This file is already on the hard
                                    // drive (but with a different name).

                                    l_KeepLocal = false;
                                    l_ExistingFullFileName = l_Other.p_LocalFile.BuildFullFileName();
                                }

                                l_KeptKnown = false;
                                DereferenceKnownFile(a_Cntxt.f_Known);

                                break;
                            }
                        }
                    }
                }
                LockLeave();

                AddSingleIncomplFileOp(IncomplOperationContext.CreateForPermanentDeletion(a_Cntxt.f_Known));
                CloseExtractionTempFile(ref a_Cntxt, l_KeepLocal);

                if (true == l_KeptKnown)
                {
                    // Try to set the exctracted file up for fast
                    // detection.

                    // Query the file time. Do this after the file was,
                    // closed, otherwise the result is useless.
                    Time l_LastWriteTime;
                    Exception l_Junk;
                    if (false == FHandle.GetLastWriteTimeByFileName(a_Cntxt.f_TempFileName, out l_LastWriteTime, out l_Junk))
                    {
                        // If we didn't get the file time, still go on.

                        l_LastWriteTime = new Time();
                    }

                    bool l_Succ = false;

                    LockEnter();
                    {
                        if (ComplStatus.Incomplete == a_Cntxt.f_Known.p_Status)
                        {
                            l_Succ = true;
                            a_Cntxt.f_Known.TurnIncomplIntoMissing(a_Cntxt.f_TempFileName, l_LastWriteTime, a_Cntxt.f_Hashes);
                        }
                    }
                    LockLeave();

                    if (false == l_Succ)
                    {
                        // Should happen very rarely. If it happens,
                        // it's still okay to go on.

                        Util.DebugBreak();
                    }
                }

                if (false == l_KeepLocal)
                {
                    f_Inst.InvokeMainFormAsync(new Gui.MainForm.NotifyDownloadCompletionHandler(f_Inst.f_MainForm.NotifyDownloadOfDuplicate), l_KnownFileName, l_ExistingFullFileName);
                }
                else
                {
                    // Manually add the local file, now that it's not
                    // locked anymore and inform the GUI.

                    AddOrUpdateSingleFile(a_Cntxt.f_TempFileName, true);
                    f_Inst.InvokeMainFormAsync(new Gui.MainForm.NotifyDownloadCompletionHandler(f_Inst.f_MainForm.NotifyDownloadCompletion), l_KnownFileName, a_Cntxt.f_TempFileName);
                }

                a_Cntxt = new ExtractionContext();

                return true;
            }

            // Unknown stage.

            throw (new ApplicationException());
        }

        private void AbortExtraction(ref ExtractionContext a_Cntxt)
        {
            if (false == a_Cntxt.IsValid())
            {
                return;
            }

            a_Cntxt.Unregister(this);
            CloseExtractionTempFile(ref a_Cntxt, false);

            a_Cntxt = new ExtractionContext();
        }

        private void CloseExtractionTempFile(ref ExtractionContext a_Cntxt,
                                             bool a_KeepTempFile)
        {
            if (true == a_KeepTempFile)
            {
                a_Cntxt.f_TempFile.CancelDeletionAndClose();

                return;
            }

            // We have to make sure the file isn't opened by
            // AddOrUpdateSingleFile() in the moment we drop the
            // handle and before we delete the file.

            LockEnter();
            {
                f_DontOpenExtractionFileName = a_Cntxt.f_TempFile.f_FullFileName;
            }
            LockLeave();

            // This will delete the file.
            a_Cntxt.f_TempFile.Close();

            // Lockless zeroing.
            f_DontOpenExtractionFileName = null;
        }

        private void NotifyFinishedIncomplFile(GeneralAPCCookie a_Cookie)
        {
            // This routine is invoked by the hash worker when the
            // corresponding incomplete file demands it.
            DebugAssertThread(c_HashThreadName);

            KnownFile l_Known = a_Cookie.GetArgAsKnownFile();
            bool l_Added = AddFinishedIncomplFile(l_Known);

            a_Cookie.SetState(false == l_Added ? CookieStates.Failure : CookieStates.Success);
        }

        private bool AddFinishedIncomplFile(KnownFile a_Known)
        {
            // This routine may only be called by the hash worker.
            DebugAssertThread(c_HashThreadName);

            // The file will be ignored if the list is already filled
            // well.

            int l_Count = f_FinishedIncompl.p_Count;

            if (c_MaxFinishedIncomplFilesCount > l_Count)
            {
                KnownFile[] l_Arr = f_FinishedIncompl.p_Arr;
                for (int i = 0; i < l_Count; i++)
                {
                    if (a_Known == l_Arr[i])
                    {
                        // Duplicate.

                        return false;
                    }
                }

                int l_Idx = f_FinishedIncompl.ReserveAtEnd();
                f_FinishedIncompl.p_Arr[l_Idx] = a_Known;

                return true;
            }

            return false;
        }

        public static bool IsValidHashElmKey(KnownFile a_Known)
        {
            return IsValidHashElmKey(a_Known.f_HashSetKey);
        }

        public static bool IsValidHashElmKey(ElmKey a_Key)
        {
            return (Instance.c_ReservedMemElmKeys <= a_Key.f_Val);
        }

        private void HashWorker()
        {
            f_HashWorkerFinished.Reset();

            GeneralAPCCookie[] l_APCCookies = null;
            ExtractionContext l_ExtractionCntxt = new ExtractionContext();
            Time l_NextExtractionTime = new Time();
            byte[] l_TempBuffer = null;

            while (true == Util.VolatileRead(ref f_RunHashWorker))
            {
                if (null == l_APCCookies)
                {
                    l_APCCookies = f_HashQueue.GetNow();
                }

                // We must do this somewhere in order to release
                // unneeded memory.
                TryReleaseRecycleComplFilesCntxt();

                Time l_Time = f_Inst.UpdateTime();

                if (l_Time >= f_NextRescanForFinishedFilesTime)
                {
                    RescanForFinishedFiles(l_Time);

                    f_NextRescanForFinishedFilesTime = l_Time.AddSeconds(c_RescanForFinishedFilesInterval);
                }

                // Now handle any async tasks.

                if (null != l_APCCookies)
                {
                    APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
                }

                if (false == f_IdentWorkerIoBusy)
                {
                    // Handle creating complete files.

                    if (false == l_Time.IsEarlierThan(l_NextExtractionTime) &&
                        false == ContinueExtraction(ref l_ExtractionCntxt))
                    {
                        // Pause for a short while.

                        l_NextExtractionTime = l_Time.AddSeconds(5);
                    }

                    // Handle hashing.

                    if (0 == f_HashingIdleTicks ||
                        f_LastHashingTime.AddTicks(f_HashingIdleTicks) <= l_Time)
                    {
                        f_LastHashingTime = l_Time;

                        if (true == HandleHashing(ref l_TempBuffer, c_DefaultHashItemSize) &&
                            0 == f_HashingIdleTicks)
                        {
                            continue;
                        }
                    }
                }

                l_TempBuffer = null;
                l_APCCookies = f_HashQueue.WaitForInput();
            }

            if (null != l_APCCookies)
            {
                APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
            }

            f_HashQueue.ShutDown();
            AbortExtraction(ref l_ExtractionCntxt);

            f_HashWorkerFinished.Set();

            Software.UpdateShutDownState(0x1000000);
        }

#endregion

#region Identification

        private void IdentifyUnprocessedFile(LocalFile a_Local)
        {
            // The caller must hold f_Lock.

            string l_FullFileName = a_Local.BuildFullFileName();
            MultiMap<ulong, KnownFile>.Enumerator l_Enum;
            if (true == f_KnownFilesBySize.CreateEnumerator(a_Local.f_FileSize, out l_Enum))
            {
                KnownFile l_Known;
                while (true == l_Enum.GetNext(out l_Known))
                {
                    if (null == l_Known.p_LocalFile &&
                        true == l_Known.TryToFastDetectLocalFile(a_Local, l_FullFileName))
                    {
                        l_Known.SetLocalFile(a_Local, true);
                        FinishLinkingKnownFile(l_Known);

                        return;
                    }
                }
            }

            // If the fast track fails, we need quick and total hash to
            // complete the identification.

            a_Local.AddToDic(f_UnhashedLoc);
        }

        private void IdentifyHashedFile(LocalFile a_Local)
        {
            // The caller must hold f_Lock. 'Hashed' in this context
            // means that we have the quick and total hash of the file
            // (it doesn't say anything about the P2P hash sets).

            MultiMap<ulong, KnownFile>.Enumerator l_Enum;
            if (true == f_KnownFilesBySize.CreateEnumerator(a_Local.f_FileSize, out l_Enum))
            {
                KnownFile l_Known;
                while (true == l_Enum.GetNext(out l_Known))
                {
                    if (ComplStatus.Missing == l_Known.p_Status &&
                        true == l_Known.TryToSetHashedLocalFile(a_Local))
                    {
                        // Local file identified.

                        FinishLinkingKnownFile(l_Known);

                        return;
                    }
                }
            }

            // Create a known file for this new and unknown local file.

            KnownFile l_NewKnown = KnownFile.CreateFileFromLocal(FindUnoccupiedFileNum(), a_Local);

            ReferenceKnownFile(l_NewKnown, f_UnhashedKnw);
            a_Local.AddToDic(f_RecognizedLoc);
        }

        private void FinishLinkingKnownFile(KnownFile a_Known)
        {
            // The caller must hold f_Lock.

            if (false == a_Known.HasTrustedHashes())
            {
                a_Known.AddToDic(f_UnhashedKnw);
            }
            else
            {
                a_Known.AddToDic(f_HashedKnw);
            }

            a_Known.p_LocalFile.AddToDic(f_RecognizedLoc);
        }

        private bool HandleIdentification(Time a_Time,
                                          ref IdentificationContext a_Cntxt,
                                          ref byte[] a_TempBuffer,
                                          uint a_MaxBytes)
        {
            // Returns true if a subsequent call to this routine might
            // progress things. a_TempBuffer may be null.
            DebugAssertThread(c_IdentThreadName);

            bool l_NeedToEnter = true;

            if (true == a_Cntxt.IsValid())
            {
                // Peek to find out if we need to enter at all.

                l_NeedToEnter = (0 < f_UnprocessedLoc.Count);
                Util.MemoryBarrier();
            }

            if (true == l_NeedToEnter)
            {
                bool l_HasUnprocessed;

                LockEnter();
                {
                    // Take a break after ten files.
                    for (int i = 0; 10 > i; i++)
                    {
                        LocalFile l_Local;
                        if (false == Util.GetAny<LocalFile>(f_UnprocessedLoc.Keys, out l_Local))
                        {
                            break;
                        }

                        IdentifyUnprocessedFile(l_Local);
                    }

                    l_HasUnprocessed = (0 < f_UnprocessedLoc.Count);

                    if (false == l_HasUnprocessed &&
                        false == a_Cntxt.IsValid())
                    {
                        LocalFile l_Local;
                        if (true == Util.GetAny<LocalFile>(f_UnhashedLoc.Keys, out l_Local))
                        {
                            a_Cntxt = new IdentificationContext(l_Local);
                            f_IdentWorkerIoBusy = true;
                        }
                    }
                }
                LockLeave();

                if (true == l_HasUnprocessed)
                {
                    // Identification shall go on.

                    return true;
                }

                if (false == a_Cntxt.IsValid())
                {
                    // No more files to identify.

                    return false;
                }
            }

            bool l_Remove = false;

            if (false == a_Cntxt.f_HasQuickHash)
            {
                if (false == HashSet.CalcQuickHash(a_Cntxt.f_Local.ExposeFHandle(), f_IdentWorkerSHA256))
                {
                    l_Remove = true;
                }
                else
                {
                    a_Cntxt.f_HasQuickHash = true;

                    LockEnter();
                    {
                        a_Cntxt.f_Local.SetQuickHash(f_IdentWorkerSHA256);
                    }
                    LockLeave();
                }
            }

            if (false == l_Remove)
            {
                HashingStatus l_HashStatus;

                if (false == a_Cntxt.f_HasTotalHash)
                {
                    if (0 != f_HashingIdleTicks &&
                        f_LastIdentTotalHashingTime.AddTicks(f_HashingIdleTicks) > a_Time)
                    {
                        return true;
                    }

                    f_LastIdentTotalHashingTime = a_Time;

                    if (null == a_TempBuffer)
                    {
                        a_TempBuffer = new byte[c_DefaultBufferSize];
                    }

                    l_HashStatus = HashSet.CalcFileHash(a_Cntxt.f_Local.ExposeFHandle(), a_Cntxt.f_Local.f_FileSize, f_IdentWorkerSHA256, ref a_MaxBytes, ref a_Cntxt.f_TotalHashFilePos, a_TempBuffer);
                }
                else
                {
                    l_HashStatus = HashingStatus.Done;
                }

                if (HashingStatus.Done == l_HashStatus)
                {
                    LockEnter();
                    {
                        if (null != a_Cntxt.f_Local.p_Parent)
                        {
                            // The file is still valid.

                            if (false == a_Cntxt.f_HasTotalHash)
                            {
                                a_Cntxt.f_Local.SetTotalHash(f_IdentWorkerSHA256);
                            }

                            if (null == a_Cntxt.f_Local.p_KnownFile)
                            {
                                IdentifyHashedFile(a_Cntxt.f_Local);
                            }
                        }
                    }
                    LockLeave();

                    a_Cntxt = new IdentificationContext();

                    return true;
                }
                else if (HashingStatus.NeedFurtherCalls != l_HashStatus)
                {
                    // There was an error.

                    l_Remove = true;
                }
            }

            if (true == l_Remove)
            {
                DropLocalFile(a_Cntxt.f_Local, null);

                a_Cntxt = new IdentificationContext();
            }

            return true;
        }

        private void AddImportJob(GeneralAPCCookie a_Cookie)
        {
            // This routine may only be called by the ident worker.
            DebugAssertThread(c_IdentThreadName);

            MultiFileImportContext l_Cntxt = (MultiFileImportContext)a_Cookie.GetArg();
            f_ImportContexts.Enqueue(l_Cntxt);

            a_Cookie.SetState(CookieStates.Success);
        }

        private bool HandleImports(ref ImportData a_ImportData)
        {
            // This routine may only be called by the ident worker.
            DebugAssertThread(c_IdentThreadName);

            // Returns true if an operation was initialized that may
            // decrease the available disk space (as a hint).

            if (false == a_ImportData.IsValid())
            {
                if (0 >= f_ImportContexts.Count)
                {
                    // Nothing to do.

                    return false;
                }

                MultiFileImportContext l_Cntxt = f_ImportContexts.Dequeue();
                a_ImportData = new ImportData(this, l_Cntxt);
                f_IdentWorkerIoBusy = true;
            }

            ImportableFile l_ImpFile;
            ImportData.ProcessResult l_Res = a_ImportData.Process(out l_ImpFile);

            SWITCH:

            switch (l_Res)
            {
                case ImportData.ProcessResult.Pending:
                    return false;

                case ImportData.ProcessResult.PendingWatchSpace:
                    return true;

                case ImportData.ProcessResult.NextFile:
                {
                    // We need to find or create a file.

                    KnownFile l_Known;
                    bool l_Created;

                    LockEnter();
                    {
                        l_Known = GetAnyFileWitheD2kFileIDAndSize(l_ImpFile.f_FileID, l_ImpFile.f_FileSize);

                        if (null == l_Known)
                        {
                            l_Known = KnownFile.CreateFileByImporting(FindUnoccupiedFileNum(), l_ImpFile, f_IOPrefs.f_InsertIncomplFilesStarted);
                            ReferenceKnownFile(l_Known, f_UnlinkedKnw);

                            l_Created = true;
                        }
                        else
                        {
                            if (ComplStatus.Incomplete != l_Known.p_Status)
                            {
                                // The files exists already and isn't
                                // suitable for receiving imports.

                                l_Known = null;
                            }

                            l_Created = false;
                        }
                    }
                    LockLeave();

                    if (null == l_Known)
                    {
                        a_ImportData.SkipFile();

                        return false;
                    }

                    l_Res = a_ImportData.SetFile(l_Known, l_Created);

                    goto SWITCH;
                }

                case ImportData.ProcessResult.EndReached:
                    a_ImportData = new ImportData();
                    return false;
            }

            // We should never land here.

            throw (new ApplicationException());
        }

        private void CancelImportsOnShutdown(ref ImportData a_ImportData)
        {
            // This routine may only be called by the ident worker.
            DebugAssertThread(c_IdentThreadName);

            // Abort current import (if any).

            if (true == a_ImportData.IsValid())
            {
                a_ImportData.Abort(MultiFileImportContext.ResultCodes.Shutdown);
                a_ImportData = new ImportData();
            }

            // Clear global queue.

            foreach (MultiFileImportContext l_Cntxt in f_ImportContexts)
            {
                l_Cntxt.ContentManagerAbort(MultiFileImportContext.ResultCodes.Shutdown);
            }

            f_ImportContexts.Clear();
        }

        private void IdentificationWorker()
        {
            f_IdentWorkerFinished.Reset();

            Time l_NextCheckDiskSpaceTime = new Time();
            Time l_NextReseedRandomProviderTime = new Time();
            IdentificationContext l_IdentCntxt = new IdentificationContext();
            byte[] l_TempBuffer = null;
            ImportData l_ImportData = new ImportData();
            GeneralAPCCookie[] l_APCCookies = null;

            while (true == Util.VolatileRead(ref f_RunIdentWorker))
            {
                // Check available disk space.

                Time l_Time = f_Inst.GetTime();

                if (l_Time >= l_NextCheckDiskSpaceTime)
                {
                    bool l_Critical = !Software.HasAppDataDriveAvailBytes(true, c_MinAvailDiskSpace);

                    if (l_Critical != f_CriticalDiskSpaceLevelReached)
                    {
                        f_CriticalDiskSpaceLevelReached = l_Critical;

                        if (false == l_Critical)
                        {
                            f_Inst.InvokeMainFormAsync(new ArglessHandler(f_Inst.f_MainForm.NotifyResumedDownloadsAndHashingDueToAvailDiskSpace));
                        }
                    }

                    if (true == l_Critical)
                    {
                        f_Inst.InvokeMainFormAsync(new ArglessHandler(f_Inst.f_MainForm.NotifySuspendedDownloadsAndHashingDueToInsuffDiskSpace));
                    }

                    l_NextCheckDiskSpaceTime = l_Time.AddSeconds(c_CheckFreeDiskSpaceInterval);
                }

                if (l_Time >= l_NextReseedRandomProviderTime)
                {
                    uint l_Rand = Util.GetRandomUInt();

                    if (true == l_NextReseedRandomProviderTime.IsSet())
                    {
                        Util.ReseedRandomProvider();
                    }

                    l_Rand -= Util.GetRandomUInt();
                    uint l_Secs = ((7 * Time.c_SecsPerMinute) + (l_Rand % (5 * Time.c_SecsPerMinute)));
                    l_NextReseedRandomProviderTime = l_Time.AddSeconds(l_Secs);
                }

                // Now handle any async tasks.

                if (null != l_APCCookies)
                {
                    APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
                }

                // Handle imports.

                bool l_IsImportBusy = false;

                if (true == HandleImports(ref l_ImportData))
                {
                    // We need to have an eye on the disk space.

                    Time l_Soon = l_Time.AddMilliseconds(3 * Util.c_MiniSleepTimeout);
                    l_NextCheckDiskSpaceTime = Time.Min(l_NextCheckDiskSpaceTime, l_Soon);

                    l_IsImportBusy = true;
                }

                // Handle identification of files.

                bool l_IsIdentificationBusy = HandleIdentification(l_Time, ref l_IdentCntxt, ref l_TempBuffer, c_DefaultIdentItemSize);

                if (true == l_IsImportBusy ||
                    true == l_IsIdentificationBusy)
                {
                    if (true == l_IsImportBusy ||
                        0 == f_HashingIdleTicks)
                    {
                        continue;
                    }
                }
                else
                {
                    f_IdentWorkerIoBusy = false;
                }

                l_TempBuffer = null;
                l_APCCookies = f_IdentQueue.WaitForInput();
            }

            if (null != l_APCCookies)
            {
                APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
            }

            f_IdentQueue.ShutDown();
            CancelImportsOnShutdown(ref l_ImportData);

            f_IdentWorkerFinished.Set();

            Software.UpdateShutDownState(0x2000000);
        }

#endregion

        private void ReadComplWorker()
        {
            // This routine handles reading from complete files.

            f_ReadComplWorkerFinished.Reset();

            SHA_256 l_SHA256 = new SHA_256();

            while (true)
            {
                ReadCookie[] l_ReadCookies = f_ReadComplQueue.WaitForInput(ref f_RunReadComplWorker);

                if (null == l_ReadCookies)
                {
                    break;
                }

                foreach (ReadCookie l_Cookie in l_ReadCookies)
                {
                    ReadCompl(l_Cookie, l_SHA256);
                }
            }

            f_ReadComplQueue.ShutDown();

            f_ReadComplWorkerFinished.Set();

            Software.UpdateShutDownState(0x4000000);
        }

        private void ReadCompl(ReadCookie a_Cookie,
                               SHA_256 a_SHA256)
        {
            KnownFile l_Known = a_Cookie.f_Known;
            LocalFile l_Local = l_Known.p_LocalFileVolatile;

            if (null == l_Local)
            {
                goto FAIL_EXIT;
            }

            if (false == l_Known.p_CheckedQuickHashVolatile)
            {
                // We haven't read from this file in this session,
                // so now hash minor portions and see if they fit
                // to what we know about this file.

                if (false == HashSet.CalcQuickHash(l_Local.ExposeFHandle(), a_SHA256))
                {
                    DropLocalFile(l_Local, null);

                    goto FAIL_EXIT;
                }

                byte[] l_QuickHash = new byte[SHA_256.c_HashByteSize];
                a_SHA256.GetDigest(l_QuickHash, 0);
                bool l_Fail = false;

                LockEnter();
                {
                    if (l_Local.p_KnownFile != l_Known ||
                        null == l_Local.p_Parent)
                    {
                        // Things have changed in the meantime.

                        l_Fail = true;
                    }
                    else
                    {
                        if (false == l_Local.VerifyQuickHash(l_QuickHash, 0))
                        {
                            // This local file doesn't seem to be the known
                            // file we thought it was.

                            l_Known.ClearLocalFile(f_Inst.GetTime());
                            l_Known.MakeFastDetectionImpossible();

                            l_Known.AddToDic(f_UnlinkedKnw);
                            l_Local.AddToDic(f_UnhashedLoc);

                            l_Fail = true;
                        }
                        else
                        {
                            l_Known.SetQuickHashChecked();
                        }
                    }
                }
                LockLeave();

                if (true == l_Fail)
                {
                    goto FAIL_EXIT;
                }
            }

            ByteArray l_Dest = new ByteArray(a_Cookie.f_Size);

            if (true == l_Local.Read(a_Cookie.f_Offset, ref l_Dest))
            {
                a_Cookie.Succeed(ref l_Dest);

                return;
            }

            DropLocalFile(l_Local, null);

            FAIL_EXIT:

            if (CookieStates.Pending == a_Cookie.p_State)
            {
                a_Cookie.Fail();
            }
        }

        public static bool IsValidFileSize(ulong a_FileSize)
        {
            return (0 != a_FileSize &&
                    c_MaxSharedFileSize >= a_FileSize);
        }

        private enum ScanState : byte
        {
            NotStarted = 0,
            Started = 1,
            Finished = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct IdentificationContext
        {
            public readonly LocalFile f_Local;
            public bool f_HasQuickHash;
            public bool f_HasTotalHash;
            public ulong f_TotalHashFilePos;

            public IdentificationContext(LocalFile a_Local)
            {
                // The caller should hold the appropriate lock to
                // a_Local.

                f_Local = a_Local;
                f_HasQuickHash = a_Local.HasQuickHash();
                f_HasTotalHash = a_Local.HasTotalHash();
                f_TotalHashFilePos = 0;
            }

            public bool IsValid()
            {
                return (null != f_Local);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ImportData
        {
            // Don't use the struct default constructor.

            private readonly ContentManager f_CntntMngr;
            private readonly MultiFileImportContext f_Cntxt;

            private InternalState f_State;
            // The following two pointers can be null.
            private KnownFile f_Known;
            private SourceCookie f_SourceCookie;

            public ImportData(ContentManager a_CntntMngr,
                              MultiFileImportContext a_Cntxt)
            {
                f_CntntMngr = a_CntntMngr;
                f_Cntxt = a_Cntxt;

                f_State = InternalState.Initialized;
                f_Known = null;
                f_SourceCookie = null;
            }

            public bool IsValid()
            {
                return (null != f_Cntxt);
            }

            public ProcessResult Process(out ImportableFile a_ImpFile)
            {
                a_ImpFile = null;

                switch (f_State)
                {
                    case InternalState.ImpCookiePending:
                    case InternalState.Initialized:
                    {
                        if (false == f_Cntxt.CanQueryNextCookie())
                        {
                            return ProcessResult.Pending;
                        }

                        bool l_NewFile;
                        ImportCookie l_ImpCookie = f_Cntxt.GetNextCookie(out l_NewFile);

                        if (null == l_ImpCookie)
                        {
                            ClearFileCore();
                            f_State = InternalState.Finished;

                            return ProcessResult.EndReached;
                        }

                        bool l_LackOfSpace = f_CntntMngr.f_CriticalDiskSpaceLevelReached;

                        if (true == l_NewFile &&
                            false == l_LackOfSpace)
                        {
                            a_ImpFile = l_ImpCookie.f_ImportFile;

                            // NOTE: An estimation that will always be
                            // too low, since it doesn't account for
                            // storage overhead.
                            ulong l_MinSize = (c_MinAvailDiskSpace + a_ImpFile.f_Done);

                            if (true == Software.HasAppDataDriveAvailBytes(false, l_MinSize))
                            {
                                ClearFileCore();
                                f_State = InternalState.WaitingForFile;

                                return ProcessResult.NextFile;
                            }

                            a_ImpFile = null;
                            l_LackOfSpace = true;
                        }

                        if (true == l_LackOfSpace)
                        {
                            return AbortCore(MultiFileImportContext.ResultCodes.LackOfAvailDiskSpace);
                        }

                        return ImportAsync();
                    }

                    case InternalState.RegisteringSource:
                    {
                        CookieStates l_Status = f_SourceCookie.p_ContentMngrStateVolatile;

                        if (CookieStates.Pending == l_Status)
                        {
                            return ProcessResult.Pending;
                        }

                        if (CookieStates.Success == l_Status)
                        {
                            return ImportAsync();
                        }

                        // Failed to register.

                        f_Cntxt.SkipFile(false);

                        ClearFileCore(false);
                        f_State = InternalState.Initialized;

                        goto case InternalState.Initialized;
                    }

                    case InternalState.Finished:
                        return ProcessResult.EndReached;

                    // Do not handle InternalState.WaitingForFile,
                    // since it's illegal here.
                }

                // We should never land here.

                throw (new ApplicationException());
            }

            public void SkipFile()
            {
                SkipFileCore(false);
            }

            private void SkipFileCore(bool a_TreatAsSuccess)
            {
                f_Cntxt.SkipFile(a_TreatAsSuccess);
                f_State = InternalState.Initialized;
            }

            public ProcessResult SetFile(KnownFile a_Known,
                                         bool a_JustCreated)
            {
                // UnregisterFile() or Abort() must be called
                // later for any successful call.

                ImportCookie l_ImpCookie = f_Cntxt.GetCurrentCookie();
                ImportableFile l_ImpFile = l_ImpCookie.f_ImportFile;

                if (true == a_JustCreated)
                {
                    a_Known.SetCategory(f_Cntxt.f_DestCategory);

                    if (true == f_Cntxt.f_CopyAttributes)
                    {
                        a_Known.f_DlControl = l_ImpFile.f_DlControl;
                        a_Known.f_DlPriority = l_ImpFile.f_DlPriority;
                        a_Known.f_UlPriority = l_ImpFile.f_UlPriority;
                    }

                    f_Cntxt.NotifyFileCreation();
                }

                ulong l_SrcFileSize;

                if (false == l_ImpFile.HasHashSetOrContent() ||
                    false == l_ImpCookie.f_SrcFile.GetSize(out l_SrcFileSize) ||
                    l_ImpFile.GetExpectedMinFileSize() > l_SrcFileSize)
                {
                    // There is nothing to import here or the source
                    // file is too small. In any case we wanted the
                    // file to be created (if it didn't exist).

                    SkipFileCore(true);

                    // NOTE: In fact the caller could call Process()
                    // immediately (no need to wait).
                    return ProcessResult.Pending;
                }

                f_Known = a_Known;
                f_SourceCookie = new SourceCookie(UserNum.s_Import);

                if (false == f_CntntMngr.RegisterSourceAsync(f_Known, f_SourceCookie))
                {
                    f_Known = null;
                    f_SourceCookie = null;

                    return AbortCore(MultiFileImportContext.ResultCodes.AsyncOpFailed);
                }

                f_State = InternalState.RegisteringSource;

                return ProcessResult.Pending;
            }

            private ProcessResult ImportAsync()
            {
                if (false == f_CntntMngr.ImportAsync(f_Known, f_Cntxt.GetCurrentCookie()))
                {
                    return AbortCore(MultiFileImportContext.ResultCodes.AsyncOpFailed);
                }

                f_State = InternalState.ImpCookiePending;

                return ProcessResult.PendingWatchSpace;
            }

            public void Abort(MultiFileImportContext.ResultCodes a_ResultCode)
            {
                // For external use.

                AbortCore(a_ResultCode);
            }

            private ProcessResult AbortCore(MultiFileImportContext.ResultCodes a_ResultCode)
            {
                f_Cntxt.ContentManagerAbort(a_ResultCode);

                ClearFileCore();
                f_State = InternalState.Finished;

                return ProcessResult.EndReached;
            }

            private void ClearFileCore()
            {
                ClearFileCore(true);
            }

            private void ClearFileCore(bool a_AlsoUnregister)
            {
                // If there is no file it's legal to call, but nothing
                // happens.

                if (null != f_Known)
                {
                    if (true == a_AlsoUnregister)
                    {
                        bool l_OK = f_CntntMngr.UnregisterSourceAsync(f_Known, f_SourceCookie);
                    }

                    f_Known = null;
                    f_SourceCookie = null;
                }
            }

            private enum InternalState : byte
            {
                Initialized = 0,
                ImpCookiePending = 1,
                // WaitingForFile means SetFile() or SkipFile() must
                // be called next.
                WaitingForFile = 2,
                RegisteringSource = 3,
                Finished = 4,
            }

            public enum ProcessResult : byte
            {
                Pending = 0,
                PendingWatchSpace = 1,
                NextFile = 2,
                EndReached = 3,
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct HashingContext
        {
            public readonly KnownFile f_Known;
            public readonly LocalFile f_Local;
            public readonly HashSet.HashSetData f_SetData;

            public HashSet f_HashSet;
            public P2PHashSets f_HashSetType;

            public HashingContext(KnownFile a_Known,
                                  LocalFile a_Local,
                                  HashSet.HashSetData a_SetData)
            {
                f_Known = a_Known;
                f_Local = a_Local;
                f_SetData = a_SetData;

                f_HashSet = null;
                f_HashSetType = P2PHashSets.None;
            }

            public bool IsValid()
            {
                return (null != f_Known);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ExtractionContext
        {
            public readonly KnownFile f_Known;
            public readonly SourceCookie f_SourceCookie;
            public readonly TemporaryFile f_TempFile;
            public readonly string f_TempFileName;
            public readonly bool f_PreserveWarnMessage;

            public Stages f_Stage;
            public byte[] f_FirstSector;

            // Extraction specific.
            public ExtractionCookie f_ExtractionCookie;

            // Hash specific.
            private SHA_256 f_SHA256;
            private byte[] f_TempBuffer;
            private ulong f_HashFilePos;
            public byte[] f_Hashes;

            public ExtractionContext(KnownFile a_Known,
                                     SourceCookie a_SourceCookie,
                                     TemporaryFile a_TempFile,
                                     string a_TempFileName)
            {
                f_Known = a_Known;
                f_SourceCookie = a_SourceCookie;
                f_TempFile = a_TempFile;
                f_TempFileName = a_TempFileName;
                f_PreserveWarnMessage = (HardMemory.c_Aligment < a_Known.f_FileSize);

                f_Stage = Stages.Registering;
                f_FirstSector = null;
                f_ExtractionCookie = null;

                f_SHA256 = null;
                f_TempBuffer = null;
                f_HashFilePos = 0;
                f_Hashes = null;
            }

            public bool IsValid()
            {
                return (null != f_Known);
            }

            public void Unregister(ContentManager a_ContentMngr)
            {
                bool l_OK = a_ContentMngr.UnregisterSourceAsync(f_Known, f_SourceCookie);
            }

            public bool BuildNextExtractionCookie()
            {
                // Returns false if the extraction is finished,
                // otherwise a new cookie was created.

                ulong l_FilePos = (null == f_ExtractionCookie ? 0 : (f_ExtractionCookie.f_FilePos + f_ExtractionCookie.f_Size));
                ulong l_RemainingSize = (f_Known.f_FileSize - l_FilePos);

                if (0 == l_RemainingSize)
                {
                    // Done.

                    f_ExtractionCookie = null;

                    return false;
                }

                uint l_NewSize = (uint)Math.Min(l_RemainingSize, IncompleteFile.c_ChunkSize);
                f_ExtractionCookie = new ExtractionCookie(l_FilePos, l_NewSize, f_TempFile, false);

                if (true == f_PreserveWarnMessage &&
                    0 == l_FilePos)
                {
                    f_ExtractionCookie.f_StartBuffer = new byte[HardMemory.c_Aligment];
                }

                return true;
            }

            public HashingStatus Hash(uint a_MaxBytes)
            {
                if (null == f_SHA256)
                {
                    f_SHA256 = new SHA_256();
                    f_TempBuffer = new byte[c_DefaultBufferSize];
                }

                HashingStatus l_HashStatus = HashSet.CalcFileHash(f_TempFile.ExposeFHandle(), f_Known.f_FileSize, f_SHA256, ref a_MaxBytes, ref f_HashFilePos, f_TempBuffer);

                if (HashingStatus.Done == l_HashStatus)
                {
                    f_Hashes = new byte[2 * SHA_256.c_HashByteSize];
                    f_SHA256.GetDigest(f_Hashes, LocalFile.c_TotalHashPos);

                    if (false == HashSet.CalcQuickHash(f_TempFile.ExposeFHandle(), f_SHA256))
                    {
                        return HashingStatus.SrcFileAccessFailed;
                    }

                    f_SHA256.GetDigest(f_Hashes, LocalFile.c_QuickHashPos);
                }

                return l_HashStatus;
            }

            public enum Stages : byte
            {
                Registering = 0,
                Extracting = 1,
                Hashing = 2,
            }
        }

        public enum AccessLevel : byte
        {
            AccessGranted = 0,

            AccessDenied = 1,
            SourceMissing = 2,
            Impossible = 3,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct FileNameParts
        {
            // Don't use the struct default constructor.

            public readonly string f_Path;
            public readonly string f_ParentDir;
            public readonly string f_FileName;
            public bool f_Handled;

            public FileNameParts(string a_Path)
            {
                f_Path = a_Path;
                f_FileName = LongPath.SplitFullPath(f_Path, out f_ParentDir);
                f_Handled = false;
            }

            public bool IsPathProcessable()
            {
                return FileNames.IsPathProcessable(f_Path);
            }
        }

        private class IncomplOperationContext
        {
            public readonly OperationTypes f_Type;
            public readonly KnownFile f_Known;
            public readonly object f_Param;

            private IncomplOperationContext(OperationTypes a_Type,
                                            KnownFile a_Known)
            {
                f_Type = a_Type;
                f_Known = a_Known;
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           bool a_Register,
                                           SourceCookie a_SourceCookie) :
                                      this(false == a_Register ? OperationTypes.UnregisterSource : OperationTypes.RegisterSource, a_Known)
            {
                f_Param = a_SourceCookie;
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           SourceCookie a_SourceCookie,
                                           P2PHashSets a_ProvidableHashSets,
                                           ulong a_FileSize,
                                           ref MasterHashes a_MasterHashes) :
                                      this(OperationTypes.UpdateSource, a_Known)
            {
                f_Param = new UpdateSourceContext(a_SourceCookie, a_ProvidableHashSets, a_FileSize, ref a_MasterHashes);
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           SourceCookie a_SourceCookie,
                                           byte[] a_Buffer,
                                           TemporaryFile a_File) :
                                      this(OperationTypes.DeliverHashes, a_Known)
            {
                f_Param = new DeliverHashContext(a_SourceCookie, a_Buffer, a_File);
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           QueryUsersCookie a_QueryUsersCookie) :
                                      this(OperationTypes.QueryUsers, a_Known)
            {
                f_Param = a_QueryUsersCookie;
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           ObtainCookie a_ObtainCookie,
                                           uint a_MaxBytes,
                                           BitMap a_CompletedByUser,
                                           NetworkAvailability a_VolatileAvail) :
                                      this(OperationTypes.RegisterObtaining, a_Known)
            {
                f_Param = new ObtainContext(a_ObtainCookie, a_MaxBytes, a_CompletedByUser, a_VolatileAvail);
            }

            public IncomplOperationContext(ReadCookie a_ReadCookie) :
                                      this(OperationTypes.Read, a_ReadCookie.f_Known)
            {
                f_Param = a_ReadCookie;
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           ObtainCookie a_ObtainCookie,
                                           ref ByteArray a_Data) :
                                      this(OperationTypes.Write, a_Known)
            {
                f_Param = new WriteContext(a_ObtainCookie, ref a_Data);
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           ExtractionCookie a_ExtractionCookie) :
                                      this(OperationTypes.Extract, a_Known)
            {
                f_Param = a_ExtractionCookie;
            }

            public IncomplOperationContext(FileDescriptorCookie a_FileDescCookie) :
                                      this(OperationTypes.BuildFileDescriptor, a_FileDescCookie.f_Known)
            {
                f_Param = a_FileDescCookie;
            }

            public IncomplOperationContext(HashQueryCookie a_HashQueryCookie) :
                                      this(OperationTypes.QueryHashes, a_HashQueryCookie.f_Known)
            {
                f_Param = a_HashQueryCookie;
            }

            public IncomplOperationContext(KnownFile a_Known,
                                           ImportCookie a_ImportCookie) :
                                      this(OperationTypes.Import, a_Known)
            {
                f_Param = a_ImportCookie;
            }

            public static IncomplOperationContext CreateForPermanentDeletion(KnownFile a_Known)
            {
                return new IncomplOperationContext(OperationTypes.Delete, a_Known);
            }

            public enum OperationTypes : byte
            {
                RegisterSource = 0,
                UnregisterSource = 1,
                UpdateSource = 2,
                DeliverHashes = 3,
                QueryUsers = 4,
                RegisterObtaining = 5,
                Read = 6,
                Write = 7,
                Extract = 8,
                BuildFileDescriptor = 9,
                QueryHashes = 10,
                Import = 11,
                Delete = 12,
            }
        }

        private class MoveCompleteFilesToRecycleBinContext
        {
            private const string c_MoveToRecycleBinThreadName = "Recycler";

            private readonly ContentManager f_ContentMngr;
            public readonly Gui.FilesPanel.FileRemovalContext f_ParentCntxt;
            private readonly List<KnownFile> f_Files;
            private readonly uint f_InitialCount;

            public MoveCompleteFilesToRecycleBinContext(ContentManager a_ContentMngr,
                                                        Gui.FilesPanel.FileRemovalContext a_ParentCntxt,
                                                        List<KnownFile> a_Files,
                                                        uint a_InitialCount)
            {
                f_ContentMngr = a_ContentMngr;
                f_ParentCntxt = a_ParentCntxt;
                f_Files = a_Files;
                f_InitialCount = a_InitialCount;
            }

            public void Start()
            {
                Util.StartThread(c_MoveToRecycleBinThreadName, MoveCompleteFilesToRecycleBinWorker);
            }

            private void MoveCompleteFilesToRecycleBinWorker()
            {
                uint l_MovedCount = f_ContentMngr.RemoveCompleteFiles(false, f_ParentCntxt, f_Files, true);

                f_ParentCntxt.SetFinished(f_InitialCount + l_MovedCount);
            }
        }

        public class MultiFileImportContext : Gui.AbortableControllerContext
        {
            private readonly string f_PartFilesDir;
            private readonly ImportableFile[] f_Files;
            private readonly ulong f_TotalImportSize;
            public readonly bool f_CopyAttributes;
            public readonly Category f_DestCategory;

            private int f_FileIndex;
            private ImportCookie f_CurCookie;
            private ulong f_ElapsedSizeAfterCurFile;

            private ResultCodes f_ResultCode;
            // Stats:
            private ulong f_ElapsedSize;
            // The size of bytes actually imported into the content
            // manager.
            private ulong f_TranslatedSize;
            private uint f_FilesCreated;
            // The number of files that were imported as much as
            // possible.
            private uint f_FilesSuccessfullyImported;
            private uint f_FilesWithSrcAccessError;

            public MultiFileImportContext(string a_PartFilesDir,
                                          ImportableFile[] a_Files,
                                          ulong a_TotalImportSize,
                                          bool a_CopyAttributes,
                                          Category a_DestCategory)
            {
                f_PartFilesDir = a_PartFilesDir;
                f_Files = a_Files;
                f_TotalImportSize = a_TotalImportSize;
                f_CopyAttributes = a_CopyAttributes;
                f_DestCategory = a_DestCategory;
            }

            public ResultCodes QueryStats(out ulong a_ElapsedSize,
                                          out ulong a_TranslatedSize,
                                          out uint a_FilesCreated,
                                          out uint a_FilesSuccessfullyImported)
            {
                a_ElapsedSize = f_ElapsedSize;
                a_TranslatedSize = f_TranslatedSize;
                a_FilesCreated = f_FilesCreated;
                a_FilesSuccessfullyImported = f_FilesSuccessfullyImported;

                return f_ResultCode;
            }

            public bool CanQueryNextCookie()
            {
                // Returns false if there are operations running that
                // need to before going on GetNextCookie().

                return (null == f_CurCookie ||
                        CookieStates.Pending != f_CurCookie.p_StateVolatile);
            }

            public ImportCookie GetCurrentCookie()
            {
                return f_CurCookie;
            }

            public ImportCookie GetNextCookie(out bool a_NewFile)
            {
                // Returns null when the end is reached.

                bool l_Aborted = base.p_Aborted;
                CookieStates l_CookieState = (null == f_CurCookie ? CookieStates.Pending : f_CurCookie.p_State);
                Util.MemoryBarrier();

                a_NewFile = false;

                if (true == l_Aborted)
                {
                    CloseCookie();
                    Finish(ResultCodes.AbortedThroughController);

                    return null;
                }

                if (null != f_CurCookie)
                {
                    if (CookieStates.Pending == l_CookieState)
                    {
                        // CanQueryNextCookie() wasn't honored.

                        throw (new ApplicationException());
                    }

                    f_CurCookie.QueryCounts(ref f_ElapsedSize, ref f_TranslatedSize);
                    base.SetPercentage(f_ElapsedSize, f_TotalImportSize);

                    ImportCookie.Result l_Res = f_CurCookie.p_Result;
                    bool l_Succ = false;

                    if (CookieStates.Success == l_CookieState)
                    {
                        if (ImportCookie.Result.SuccessCallAgain == l_Res)
                        {
                            f_CurCookie.Recycle();

                            return f_CurCookie;
                        }

                        if (ImportCookie.Result.SuccessFinished == l_Res)
                        {
                            l_Succ = true;
                        }
                    }
                    else
                    {
                        if (ImportCookie.Result.FailedToAccessSrcFile == l_Res)
                        {
                            f_FilesWithSrcAccessError++;
                        }
                    }

                    // We are through with this file.
                    SkipFile(l_Succ);
                }

                for (int lim = f_Files.Length; f_FileIndex < lim; f_FileIndex++)
                {
                    ImportableFile l_ImpFile = f_Files[f_FileIndex];

                    if (false == l_ImpFile.f_ImportFlag)
                    {
                        continue;
                    }

                    // A new candidate.

                    ulong l_ElapsedSizeAfter = (f_ElapsedSize + l_ImpFile.f_Done);

                    if (false == IsValidFileSize(l_ImpFile.f_FileSize))
                    {
                        // Should never happen.

                        f_ElapsedSize = l_ElapsedSizeAfter;

                        continue;
                    }

                    // Process a file even if l_ImpFile.HasHashSetOrContent()
                    // returns false, since the user might want to
                    // import the file id and file size alone, even if
                    // no data is present.

                    string l_ContentFileName = l_ImpFile.f_ContentFileName;
                    bool l_IsValidPath;

                    if (false == LongPath.IsPathRooted(l_ContentFileName, out l_IsValidPath))
                    {
                        if (false == l_IsValidPath)
                        {
                            l_ContentFileName = string.Empty;
                        }
                        else
                        {
                            l_ContentFileName = LongPath.Combine(f_PartFilesDir, l_ContentFileName);
                        }
                    }

                    Exception l_Junk;
                    FHandle l_SrcFile = FHandle.OpenR(l_ContentFileName, true, FileOptions.None, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Junk);

                    if (null == l_SrcFile)
                    {
                        f_ElapsedSize = l_ElapsedSizeAfter;
                        f_FilesWithSrcAccessError++;

                        continue;
                    }

                    // Start processing this file.

                    a_NewFile = true;
                    f_CurCookie = new ImportCookie(l_ImpFile, l_SrcFile);
                    f_ElapsedSizeAfterCurFile = l_ElapsedSizeAfter;

                    return f_CurCookie;
                }

                Finish(ResultCodes.Done);

                return null;
            }

            public void SkipFile(bool a_TreatAsSuccess)
            {
                if (true == a_TreatAsSuccess)
                {
                    f_FilesSuccessfullyImported++;
                }

                // Rectify the counter if necessary.
                f_ElapsedSize = f_ElapsedSizeAfterCurFile;

                CloseCookie();
                f_FileIndex++;
            }

            public void NotifyFileCreation()
            {
                f_FilesCreated++;
            }

            public void ContentManagerAbort(ResultCodes a_ResultCode)
            {
                // Used by the content manager when a series of calls
                // of GetNextCookie() is aborted.

                CloseCookie();
                Finish(a_ResultCode);
            }

            private void Finish(ResultCodes a_ResultCode)
            {
                f_ResultCode = a_ResultCode;

                base.SetFinishedBase();
            }

            private void CloseCookie()
            {
                if (null != f_CurCookie)
                {
                    f_CurCookie.f_SrcFile.Close();
                    f_CurCookie = null;
                    f_ElapsedSizeAfterCurFile = 0;
                }
            }

            public enum ResultCodes : byte
            {
                NotFinished = 0,

                // Finish codes.
                Done = 1,
                AbortedThroughController = 2,
                LackOfAvailDiskSpace = 3,
                Shutdown = 4,
                AsyncOpFailed = 5,
            }
        }
    }

    class KnownFileCache
    {
        private readonly Dictionary<UInt128, KnownFile> f_Dic;

        public KnownFileCache(Dictionary<UInt128, KnownFile> a_Dic)
        {
            f_Dic = a_Dic;
        }

        public KnownFile TryGetFile(FileID a_FileID,
                                    ulong a_FileSize)
        {
            // a_FileSize can be zero if the file size does not matter.
            // Returns zero if there is no such file.

            KnownFile l_Known;

            if (true == f_Dic.TryGetValue(a_FileID.f_Val, out l_Known) &&
                (0 == a_FileSize ||
                 l_Known.f_FileSize == a_FileSize))
            {
                if (ComplStatus.Dropped != l_Known.p_Status)
                {
                    return l_Known;
                }
            }

            return null;
        }
    }

    static class FileNames
    {
        // Limit is in UTF-16 code units.
        public const ushort c_MaxPathChars = (ushort)short.MaxValue;
        public const ushort c_MaxFileNameChars = 255;
        private const ushort c_MaxFileNameCharsWithNumberSpace = (c_MaxFileNameChars - Util.c_HexCharsPerUInt - 1);
        public const string c_FallbackFileName = "File Name";
        private const char c_MinFileNameChar = (char)0x20;

        private static readonly char[] s_TrimWhitespaceChars;
        private static readonly char[] s_InvFileNameChars;
        private static readonly char[] s_ClearableFileNameChars;

        private static bool s_TreatFileNamesCaseSensitive;
        private static StringComparison s_NameCompMode;

        public static bool sp_TreatFileNamesCaseSensitive
        {
            get { return s_TreatFileNamesCaseSensitive; }
        }

        static FileNames()
        {
            s_TrimWhitespaceChars = new char[]
            {
                // .NET chars used by the string.Trim() routine
                // (including space):
                '\x0009', '\x000A', '\x000B', '\x000C', '\x000D', '\x0020',
                '\x0085', '\x00A0', '\x1680', '\x2000', '\x2001',
                '\x2002', '\x2003', '\x2004', '\x2005', '\x2006',
                '\x2007', '\x2008', '\x2009', '\x200A', '\x200B',
                '\x2028', '\x2029', '\x3000', '\xFEFF'
            };

            // See http://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits.
            s_InvFileNameChars = new char[] { '/', '\\', ':', '*', '?', '\"', '<', '>', '|' };
            s_ClearableFileNameChars = new char[] { '.', '_' };

            // The applications starts case insensitive.
            AdjustCaseSensitivity(false);
        }

        private static void AdjustCaseSensitivity(bool a_CaseSensitive)
        {
            s_TreatFileNamesCaseSensitive = a_CaseSensitive;
            s_NameCompMode = (false == s_TreatFileNamesCaseSensitive ? StringComparison.InvariantCultureIgnoreCase : StringComparison.InvariantCulture);
        }

        public static void FilterPathArray(ref string[] a_Paths)
        {
            // a_Paths may be null. Return can be null, if a_Paths
            // is null, empty or does not contain any valid path.

            string[] l_Arr = a_Paths;
            uint l_Len = (null == l_Arr ? 0 : (uint)l_Arr.Length);

            if (0 == l_Len)
            {
                goto NO_OUTPUT;
            }

            uint l_FinalValidIdx = 0;

            for (uint i = 0; i != l_Len; i++)
            {
                string l_Path = l_Arr[i];

                if (false == IsPathProcessable(l_Path))
                {
                    continue;
                }

                if (i != l_FinalValidIdx)
                {
                    l_Arr[l_FinalValidIdx] = l_Path;
                }

                l_FinalValidIdx++;
            }

            if (l_FinalValidIdx != l_Len)
            {
                // We won't take over everything.

                if (0 == l_FinalValidIdx)
                {
                    goto NO_OUTPUT;
                }

                Array.Resize(ref l_Arr, (int)l_FinalValidIdx);

                a_Paths = l_Arr;
            }

            return;

            NO_OUTPUT:

            a_Paths = null;
        }

        public static bool IsPathProcessable(string a_AbsPath)
        {
            return IsPathProcessable(a_AbsPath, 0, (uint)a_AbsPath.Length);
        }

        public static bool IsPathProcessable(string a_AbsPath,
                                             uint a_Idx,
                                             uint a_Len)
        {
            // The input path can be absolute or just one file name. It should
            // not be something in between.

            return (0 != a_Len &&
                    c_MaxPathChars >= a_Len);
        }

        private static int GetFileNameCore(string a_FileName,
                                           out int a_Len)
        {
            // Returns an index into a_FileName.

            a_Len = 0;

            int l_Count = a_FileName.Length;
            int l_Idx = 0;

            for (; l_Idx < l_Count; l_Idx++)
            {
                if (false == Util.IsCharInArray(a_FileName[l_Idx], s_TrimWhitespaceChars))
                {
                    break;
                }
            }

            for (int i = (l_Count - 1); l_Idx <= i; i--)
            {
                if (false == Util.IsCharInArray(a_FileName[i], s_TrimWhitespaceChars))
                {
                    a_Len = (i - l_Idx + 1);

                    break;
                }
            }

            return l_Idx;
        }

        public static bool IsLegalFileNameChar(char a_Char)
        {
            if (c_MinFileNameChar > a_Char)
            {
                return false;
            }

            foreach (char l_Char in s_InvFileNameChars)
            {
                if (l_Char == a_Char)
                {
                    return false;
                }
            }

            return true;
        }

        public static bool IsLegalFileName(string a_FileName)
        {
            int l_Len;
            int l_Idx = GetFileNameCore(a_FileName, out l_Len);

            if (0 == l_Len ||
                c_MaxFileNameChars < l_Len)
            {
                return false;
            }

            int l_LastIdx = (l_Idx + l_Len - 1);

            if ('.' == a_FileName[l_Idx] ||
                '.' == a_FileName[l_LastIdx])
            {
                return false;
            }

            if (0 <= a_FileName.IndexOfAny(s_InvFileNameChars, l_Idx, l_Len))
            {
                return false;
            }

            for (int i = l_Idx; i <= l_LastIdx; i++)
            {
                if (c_MinFileNameChar > a_FileName[i])
                {
                    return false;
                }
            }

            return true;
        }

        private static int FindLastExtensionIndex(string a_FileName,
                                                  int a_Idx,
                                                  int a_Len)
        {
            return a_FileName.LastIndexOf('.', a_Idx + a_Len - 1, a_Len);
        }

        public static string MakeFileNameLegalFailsafe(string a_RawFileName)
        {
            // a_RawFileName must not be null.

            string l_Res = TryToMakeFileNameLegal(a_RawFileName);

            return (null == l_Res ? c_FallbackFileName : l_Res);
        }

        public static string TryToMakeFileNameLegal(string a_RawFileName)
        {
            // Return can be null.

            if (false == IsLegalFileName(a_RawFileName) ||
                false == a_RawFileName.IsNormalized(NormalizationForm.FormC))
            {
                return CleanUpFileName(a_RawFileName);
            }

            return a_RawFileName;
        }

        public static string CleanUpFileName(string a_RawFileName)
        {
            // Return can be null.

            string l_FileName = a_RawFileName.Normalize(NormalizationForm.FormC);
            int l_Len;
            int l_Idx = GetFileNameCore(l_FileName, out l_Len);

            if (0 == l_Len)
            {
                return null;
            }

            if (c_MaxFileNameChars < l_Len)
            {
                l_Len = c_MaxFileNameChars;
            }

            int l_LastPeriodIdx = FindLastExtensionIndex(l_FileName, l_Idx, l_Len);

            if (0 > l_LastPeriodIdx)
            {
                return CleanUpFileNameCore(l_FileName, l_Idx, l_Len);
            }

            int l_FirstLen = (l_LastPeriodIdx - l_Idx);
            string l_Part1 = CleanUpFileNameCore(l_FileName, l_Idx, l_FirstLen);
            string l_Part2 = CleanUpFileNameCore(l_FileName, l_LastPeriodIdx + 1, l_Len - l_FirstLen - 1);

            if (null == l_Part1)
            {
                return l_Part2;
            }

            if (null == l_Part2)
            {
                return l_Part1;
            }

            return (l_Part1 + '.' + l_Part2);
        }

        public static bool FindUnoccupiedFileName(string a_Path,
                                                  string a_RawFileName,
                                                  out string a_Res)
        {
            string l_FileName = a_RawFileName;
            int l_Len;
            int l_Idx = GetFileNameCore(l_FileName, out l_Len);

            if (0 == l_Len)
            {
                l_FileName = c_FallbackFileName;
                l_Idx = 0;
                l_Len = c_FallbackFileName.Length;
            }
            else if (c_MaxFileNameCharsWithNumberSpace < l_Len)
            {
                l_Len = c_MaxFileNameCharsWithNumberSpace;
            }

            if (3 <= l_Len)
            {
                string l_OtherFileName = l_FileName.Substring(l_Idx, l_Len);

                if (true == StartsWithDOSDeviceName(l_OtherFileName))
                {
                    l_FileName = ('_' + l_OtherFileName);
                    l_Idx = 0;
                    l_Len++;
                }
            }

            char l_LastChar = l_FileName[l_Idx + l_Len - 1];

            if (' ' == l_LastChar ||
                '.' == l_LastChar)
            {
                l_FileName = (l_FileName.Substring(0, l_Len - 1) + '_');
            }

            int l_InsertIdx = FindLastExtensionIndex(l_FileName, l_Idx, l_Len);

            if (0 > l_InsertIdx)
            {
                l_InsertIdx = (l_Idx + l_Len);
            }

            StringBuilder l_Bld = new StringBuilder(a_Path);
            l_Bld.Append(Path.DirectorySeparatorChar);
            l_InsertIdx += l_Bld.Length;
            l_Bld.Append(l_FileName, l_Idx, l_Len);

            a_Res = l_Bld.ToString();
            if (false == File.Exists(a_Res))
            {
                return true;
            }

            for (int i = 0; i < 999; i++)
            {
                string l_NumStr = Util.NumberToInvString(i);
                l_Bld.Insert(l_InsertIdx, l_NumStr);

                a_Res = l_Bld.ToString();
                if (false == File.Exists(a_Res))
                {
                    return true;
                }

                l_Bld.Remove(l_InsertIdx, l_NumStr.Length);
            }

            l_Bld.Insert(l_InsertIdx, Util.GetRandomUIntFast().ToString("X8"));

            a_Res = l_Bld.ToString();
            if (false == File.Exists(a_Res))
            {
                return true;
            }

            // This should never happen.

            Util.DebugBreak();

            a_Res = null;

            return false;
        }

        public static bool StartsWithDOSDeviceName(string a_Name)
        {
            // Checks if the portion before the first period (if any)
            // is a DOS device name, ignoring any trailing space chars.

            int l_Len = a_Name.Length;

            if (3 > l_Len)
            {
                return false;
            }

            int l_ActLen = a_Name.IndexOf('.');

            if (0 > l_ActLen)
            {
                l_ActLen = l_Len;
            }

            while (2 < l_ActLen)
            {
                if (' ' != a_Name[l_ActLen - 1])
                {
                    break;
                }

                l_ActLen--;
            }

            if (3 == l_ActLen)
            {
                if (true == a_Name.StartsWith("AUX", StringComparison.InvariantCultureIgnoreCase) ||
                    true == a_Name.StartsWith("CON", StringComparison.InvariantCultureIgnoreCase) ||
                    true == a_Name.StartsWith("NUL", StringComparison.InvariantCultureIgnoreCase) ||
                    true == a_Name.StartsWith("PRN", StringComparison.InvariantCultureIgnoreCase))
                {
                    return true;
                }
            }
            else if (4 == l_ActLen &&
                     true == Util.IsDigitInvariant(a_Name[3]))
            {
                if (true == a_Name.StartsWith("COM", StringComparison.InvariantCultureIgnoreCase) ||
                    true == a_Name.StartsWith("LPT", StringComparison.InvariantCultureIgnoreCase))
                {
                    return true;
                }
            }

            return false;
        }

        private static string CleanUpFileNameCore(string a_Str,
                                                  int a_Idx,
                                                  int a_Len)
        {
            // Return can be null.

            StringBuilder l_Bld = new StringBuilder(a_Len);

            bool l_Space = false;
            for (int i = a_Idx, lim = (a_Idx + a_Len); i < lim; i++)
            {
                char l_Char = a_Str[i];

                if (true == Util.IsCharInArray(l_Char, s_TrimWhitespaceChars) ||
                    true == Util.IsCharInArray(l_Char, s_ClearableFileNameChars))
                {
                    l_Space = true;
                }
                else
                {
                    if (c_MinFileNameChar > l_Char ||
                        true == Util.IsCharInArray(l_Char, s_InvFileNameChars))
                    {
                        continue;
                    }

                    if (true == l_Space)
                    {
                        if (0 < l_Bld.Length)
                        {
                            l_Bld.Append(' ');
                        }

                        l_Space = false;
                    }

                    l_Bld.Append(l_Char);
                }
            }

            if (0 == l_Bld.Length)
            {
                return null;
            }

            string l_Res = l_Bld.ToString();

            return l_Res;
        }

        public static string ListIllegalFileNameChars()
        {
            return Util.Join(' ', s_InvFileNameChars);
        }

        public static bool AreNamesIdentical(string a_Name1,
                                             string a_Name2)
        {
            // a_Name1 and a_Name2 can be null.

            return string.Equals(a_Name1, a_Name2, s_NameCompMode);
        }

        public static bool AreNamesIdenticalMindCase(string a_Name1,
                                                     string a_Name2)
        {
            // This routine always compares case sensitive.

            return Util.StringsEqual(a_Name1, a_Name2);
        }

        public static bool IsSubdir(string a_Parent,
                                    string a_Subdir)
        {
            // This routine is able to handle Unicode equivalence. Both
            // input strings may not have an ending dir separator.

            char l_SepChar = Path.DirectorySeparatorChar;
            int l_Count = Util.CountOccurrencesOfChar(a_Parent, l_SepChar);
            int l_Idx = Util.FindCharIndex(a_Subdir, l_SepChar, l_Count + 1);

            if (0 > l_Idx ||
                (a_Subdir.Length - 1) <= l_Idx)
            {
                return false;
            }

            if (l_Idx == a_Parent.Length)
            {
                return (0 == string.Compare(a_Parent, 0, a_Subdir, 0, l_Idx, s_NameCompMode));
            }

            // This is a complicated case, where string portions might
            // be identical, but differ in length. We lack a routine
            // like string.Compare() with two length arguments.

            string l_TempStr = a_Subdir.Substring(0, l_Idx);

            return AreNamesIdentical(a_Parent, l_TempStr);
        }

        public static string ToDicName(string a_Name)
        {
            return (false == s_TreatFileNamesCaseSensitive ? a_Name.ToLowerInvariant() : a_Name);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct IOPrefs
    {
        private const bool c_DefaultInsertIncomplFilesStarted = true;
        private const bool c_DefaultTreatFileNamesCaseSenstive = false;
        public const FileAttributes c_NoFileAttributes = (FileAttributes.Directory | FileAttributes.Device | FileAttributes.ReparsePoint);
        public const FileAttributes c_DefaultExclFileAttrbs = (c_NoFileAttributes | FileAttributes.Hidden | FileAttributes.System);

        public readonly bool f_InsertIncomplFilesStarted;
        public readonly bool f_OpenFilesAsMoveable;
        public readonly bool f_WatchFileSystem;
        public readonly SharedDirectories f_SharedDirs;
        // From v1.4.0 on.
        public readonly bool f_TreatFileNamesCaseSenstive;
        public readonly FileAttributes f_ExclusionFileAttrbs;
        // From v1.4.25 on.
        public readonly byte f_HashingHDLoadThresholdPercentage;

        public IOPrefs(bool a_InsertIncomplFilesStarted,
                       bool a_OpenFilesAsMoveable,
                       bool a_WatchFileSystem,
                       SharedDirectories a_SharedDirs,
                       bool a_TreatFileNamesCaseSenstive,
                       FileAttributes a_ExclusionFileAttrbs,
                       byte a_HashingHDLoadThresholdPercentage)
        {
            a_SharedDirs.DebugAssertIntegrity();

            f_InsertIncomplFilesStarted = a_InsertIncomplFilesStarted;
            f_OpenFilesAsMoveable = a_OpenFilesAsMoveable;
            f_WatchFileSystem = a_WatchFileSystem;
            f_SharedDirs = a_SharedDirs;
            f_TreatFileNamesCaseSenstive = false;
            f_ExclusionFileAttrbs = a_ExclusionFileAttrbs;
            f_HashingHDLoadThresholdPercentage = a_HashingHDLoadThresholdPercentage;
        }

        public static void CreateDefault(out IOPrefs a_Dest)
        {
            SharedDirectories l_SharedDirs = new SharedDirectories();

            a_Dest = new IOPrefs(c_DefaultInsertIncomplFilesStarted,
                                 true,
                                 true,
                                 l_SharedDirs,
                                 c_DefaultTreatFileNamesCaseSenstive,
                                 c_DefaultExclFileAttrbs,
                                 100);
        }

        public void RecreateWithNewSharedDirs(SharedDirectories a_SharedDirs,
                                              out IOPrefs a_Dest)
        {
            a_Dest = new IOPrefs(f_InsertIncomplFilesStarted, f_OpenFilesAsMoveable, f_WatchFileSystem, a_SharedDirs, f_TreatFileNamesCaseSenstive, f_ExclusionFileAttrbs, f_HashingHDLoadThresholdPercentage);
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      uint a_Revision,
                                      out IOPrefs a_Dest)
        {
            // NOTE: For IOPrefs revision 1 and 0 are the same.

            bool l_InsertIncomplFilesStarted = a_Buffer.ReadBool();
            bool l_OpenFilesAsMoveable = a_Buffer.ReadBool();
            bool l_WatchFileSystem = a_Buffer.ReadBool();
            SharedDirectories l_SharedDirs = SharedDirectories.FromBuffer(a_Buffer);

            if (null == l_SharedDirs)
            {
                Util.DebugBreak();

                a_Dest = new IOPrefs();

                return false;
            }

            // Read revision 2 fields.

            bool l_TreatFileNamesCaseSenstive;
            FileAttributes l_ExclusionFileAttrbs;

            if (2 > a_Revision)
            {
                l_TreatFileNamesCaseSenstive = c_DefaultTreatFileNamesCaseSenstive;
                l_ExclusionFileAttrbs = c_DefaultExclFileAttrbs;
            }
            else
            {
                l_TreatFileNamesCaseSenstive = a_Buffer.ReadBool();
                l_ExclusionFileAttrbs = ((FileAttributes)a_Buffer.ReadUInt32LE() | c_NoFileAttributes);
            }

            byte l_HashingHDLoadThresholdPercentage;

            if (3 > a_Revision)
            {
                l_HashingHDLoadThresholdPercentage = 100;
            }
            else
            {
                l_HashingHDLoadThresholdPercentage = (byte)Util.MinMax(1, 100, a_Buffer.ReadByte());
            }

            a_Dest = new IOPrefs(l_InsertIncomplFilesStarted, l_OpenFilesAsMoveable, l_WatchFileSystem, l_SharedDirs, l_TreatFileNamesCaseSenstive, l_ExclusionFileAttrbs, l_HashingHDLoadThresholdPercentage);

            return true;
        }

        public uint ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteBool(f_InsertIncomplFilesStarted);
            a_Buffer.WriteBool(f_OpenFilesAsMoveable);
            a_Buffer.WriteBool(f_WatchFileSystem);
            f_SharedDirs.ToBuffer(a_Buffer);
            a_Buffer.WriteBool(f_TreatFileNamesCaseSenstive);
            a_Buffer.WriteUInt32LE((uint)f_ExclusionFileAttrbs);
            a_Buffer.WriteByte(f_HashingHDLoadThresholdPercentage);

            // Increase with every change.
            return 3;
        }

        public void DetermineChangeKind(ref IOPrefs a_ChangeTo,
                                        ref PrefChangeKinds a_ChangeKind)
        {
            if (PrefChangeKinds.AppRestartNeededToFullyTakeOver == a_ChangeKind ||
                this.f_TreatFileNamesCaseSenstive != a_ChangeTo.f_TreatFileNamesCaseSenstive)
            {
                a_ChangeKind = PrefChangeKinds.AppRestartNeededToFullyTakeOver;

                return;
            }

            if (PrefChangeKinds.ReconnectNeededToFullyTakeOver == a_ChangeKind)
            {
                return;
            }

            if (PrefChangeKinds.ApplicableWithoutReconnect == a_ChangeKind ||
                this.f_InsertIncomplFilesStarted != a_ChangeTo.f_InsertIncomplFilesStarted ||
                this.f_OpenFilesAsMoveable != a_ChangeTo.f_OpenFilesAsMoveable ||
                this.f_WatchFileSystem != a_ChangeTo.f_WatchFileSystem ||
                false == this.f_SharedDirs.ValueEquals(a_ChangeTo.f_SharedDirs) ||
                this.f_ExclusionFileAttrbs != a_ChangeTo.f_ExclusionFileAttrbs ||
                this.f_HashingHDLoadThresholdPercentage != a_ChangeTo.f_HashingHDLoadThresholdPercentage)
            {
                a_ChangeKind = PrefChangeKinds.ApplicableWithoutReconnect;
            }
        }
    }
}

namespace NB.Content.Internal
{
    class FileRegistry
    {
        // The casing of the keys of f_Folders depends on the global
        // setting (FileNames class).
        private readonly Dictionary<string, ParentFolder> f_Folders;

        public FileRegistry()
        {
            f_Folders = new Dictionary<string, ParentFolder>();
        }

        public void Clear()
        {
            f_Folders.Clear();
        }

        public LocalFile GetFile(string a_FolderPath,
                                 string a_FileName)
        {
            // Return can be null, if no such file exists.

            ParentFolder l_Folder;
            if (true == f_Folders.TryGetValue(FileNames.ToDicName(a_FolderPath), out l_Folder))
            {
                return l_Folder.GetFile(a_FileName);
            }

            return null;
        }

        public bool HasFile(LocalFile a_Local,
                            string a_FolderPath,
                            string a_FileName)
        {
            return (GetFile(a_FolderPath, a_FileName) == a_Local);
        }

        public void TryUpdateDirPath(ParentFolder a_Folder,
                                     string a_NewDirPath)
        {
            if (false == Util.StringsEqual(a_Folder.p_Path, a_NewDirPath))
            {
                a_Folder.SetPath(a_NewDirPath);
            }
        }

        public bool RefFile(LocalFile a_File,
                            string a_FolderPath,
                            out LocalFile a_ExistingLocal)
        {
            // In case of failure, a_ExistingFile might be set.

            ParentFolder l_Folder;
            string l_DicFolderPath = FileNames.ToDicName(a_FolderPath);

            if (false == f_Folders.TryGetValue(l_DicFolderPath, out l_Folder))
            {
                a_ExistingLocal = null;

                l_Folder = new ParentFolder(a_FolderPath);
                f_Folders.Add(l_DicFolderPath, l_Folder);
            }
            else
            {
                a_ExistingLocal = l_Folder.GetFile(a_File.p_FileName);

                if (null != a_ExistingLocal)
                {
                    // We already have that file.

                    return false;
                }
            }

            l_Folder.AddFile(a_File);

            return true;
        }

        public void DerefFile(LocalFile a_Local)
        {
            ParentFolder l_Parent = a_Local.p_Parent;

            l_Parent.RemoveFile(a_Local);

            if (false == l_Parent.HasFiles())
            {
                bool l_OK = f_Folders.Remove(FileNames.ToDicName(l_Parent.p_Path));
            }
        }
    }

    class ParentFolder
    {
        // f_Path preserves the casing.
        private string f_Path;
        // The casing of the keys of f_Locals depends on the global
        // setting (FileNames class).
        public readonly Dictionary<string, LocalFile> f_Locals;

        public string p_Path
        {
            get { return f_Path; }
        }

        public ParentFolder(string a_Path)
        {
            f_Path = a_Path;
            f_Locals = new Dictionary<string, LocalFile>();
        }

        public void SetPath(string a_Path)
        {
            f_Path = a_Path;
        }

        public bool HasFiles()
        {
            return (0 < f_Locals.Count);
        }

        public LocalFile GetFile(string a_FileName)
        {
            // Return can be null, if no such file exists.

            LocalFile l_Local;
            if (true == f_Locals.TryGetValue(FileNames.ToDicName(a_FileName), out l_Local))
            {
                return l_Local;
            }

            return null;
        }

        public void TryUpdateFileName(LocalFile a_Local,
                                      string a_FileName)
        {
            if (false == Util.StringsEqual(a_Local.p_FileName, a_FileName))
            {
                a_Local.UpdateFileName(a_FileName);
            }
        }

        public void AddFile(LocalFile a_Local)
        {
            f_Locals.Add(FileNames.ToDicName(a_Local.p_FileName), a_Local);

            a_Local.SetParent(this);
        }

        public void RemoveFile(LocalFile a_Local)
        {
            bool l_OK = f_Locals.Remove(FileNames.ToDicName(a_Local.p_FileName));

            a_Local.ClearParent();
        }
    }

    class LocalFile
    {
        public const byte c_QuickHashPos = 0;
        public const byte c_TotalHashPos = SHA_256.c_HashByteSize;

        private readonly FHandle f_File;
        public readonly Time f_LastWriteTime;
        public readonly ulong f_FileSize;

        // f_FileName preserves the casing.
        private string f_FileName;
        private ParentFolder f_Parent;

        private HashKind f_HashLevel;
        private byte[] f_Hashes;

        private uint f_ScanGen;
        private ShareKinds f_ShareKind;

        private KnownFile f_KnownFile;

        // f_Dic is only a reference to the dictionary this item belongs
        // to (the dictionary isn't owned by this class).
        private Dictionary<LocalFile, LocalFile> f_Dic;

        private byte f_LinkProtectionCount;

        public string p_FileName
        {
            get { return f_FileName; }
        }

        public ParentFolder p_Parent
        {
            get { return f_Parent; }
        }

        public ParentFolder p_ParentVolatile
        {
            get { return Util.VolatileRead(ref f_Parent); }
        }

        public uint p_ScanGen
        {
            get { return f_ScanGen; }
        }

        public KnownFile p_KnownFile
        {
            get { return f_KnownFile; }
        }

        public ShareKinds p_ShareKind
        {
            get { return f_ShareKind; }
        }

        public bool p_IsLinkProtected
        {
            get
            {
                return (null != f_KnownFile &&
                        0 != f_LinkProtectionCount);
            }
        }

        private LocalFile(FHandle a_File,
                          Time a_LastWriteTime,
                          ulong a_FileSize)
        {
            f_File = a_File;
            f_LastWriteTime = a_LastWriteTime;
            f_FileSize = a_FileSize;
        }

        public static LocalFile OpenR(string a_FullFileName,
                                      bool a_CanBeMovedOrDeleted,
                                      out Exception a_Exc)
        {
            // Return can be null on error.

            FHandle l_File = FHandle.OpenR(a_FullFileName, a_CanBeMovedOrDeleted, FileOptions.None, 0, FHandle.ThreadSafetyMode.NeedSynchronization, out a_Exc);

            if (null != l_File)
            {
                Time l_LastWriteTime;
                ulong l_FileSize;

                if (false == l_File.GetLastWriteTime(a_FullFileName, out l_LastWriteTime, out a_Exc) ||
                    false == l_File.GetSize(out l_FileSize) ||
                    false == ContentManager.IsValidFileSize(l_FileSize))
                {
                    l_File.Close();
                }
                else
                {
                    return new LocalFile(l_File, l_LastWriteTime, l_FileSize);
                }
            }

            return null;
        }

        public void Close()
        {
            f_File.Close();
        }

        public FHandle ExposeFHandle()
        {
            return f_File;
        }

        public bool Read(ulong a_Offset,
                         ref ByteArray a_Data)
        {
            return f_File.Read(a_Offset, ref a_Data, true);
        }

        public void ClearKnownFile()
        {
            // Only the KnownFile class may use this routine.

            f_KnownFile = null;
        }

        public void SetKnownFile(KnownFile a_KnownFile)
        {
            // Only the KnownFile class may use this routine.

            f_KnownFile = a_KnownFile;
        }

        public void AddToDic(Dictionary<LocalFile, LocalFile> a_Dic)
        {
            KnownFile.ChangeDic(this, f_Dic, a_Dic);
            f_Dic = a_Dic;
        }

        public void UpdateFileName(string a_FileName)
        {
            // Does not rename the file in the file system. Make sure
            // the file registry is updated as well.

            f_FileName = a_FileName;

            if (null != f_KnownFile)
            {
                f_KnownFile.NotifyLocalFileNameChanged();
            }
        }

        public string BuildFullFileName()
        {
            return LongPath.Combine(f_Parent.p_Path, f_FileName);
        }

        public string BuildFullFileNameVolatile()
        {
            ParentFolder l_Parent = Util.VolatileRead(ref f_Parent);

            if (null == l_Parent)
            {
                return null;
            }

            return LongPath.Combine(l_Parent.p_Path, f_FileName);
        }

        public void ClearParent()
        {
            f_Parent = null;
        }

        public void SetParent(ParentFolder a_Parent)
        {
            f_Parent = a_Parent;
        }

        public void SetScanGenAndShareKind(uint a_ScanGen,
                                           ShareKinds a_ShareKind)
        {
            f_ScanGen = a_ScanGen;
            f_ShareKind = a_ShareKind;
        }

        public void AdjustLinkProtection(bool a_Inc)
        {
            if (false == a_Inc)
            {
                f_LinkProtectionCount--;
            }
            else
            {
                f_LinkProtectionCount++;
            }
        }

#region QuickAndTotalHash

        private void EnsureHashesAllocation()
        {
            if (null == f_Hashes)
            {
                f_Hashes = new byte[2 * SHA_256.c_HashByteSize];
            }
        }

        public void ClearHashes()
        {
            // The caller must hold the appropriate lock.

            f_HashLevel = HashKind.None;
            f_Hashes = null;
        }

        public byte[] ExposeHashes()
        {
            // The caller must hold the appropriate lock.

            return f_Hashes;
        }

        public void SetHashes(byte[] a_Hashes)
        {
            // The caller must hold the appropriate lock.

            f_HashLevel = HashKind.QuickAndTotal;
            f_Hashes = a_Hashes;
        }

        public bool HasQuickHash()
        {
            return (0 != (f_HashLevel & HashKind.QuickHash));
        }

        public bool HasTotalHash()
        {
            return (0 != (f_HashLevel & HashKind.TotalHash));
        }

        public bool HasQuickAndTotalHash()
        {
            return (HashKind.QuickAndTotal == (f_HashLevel & HashKind.QuickAndTotal));
        }

        public void SetQuickHash(SHA_256 a_SHA256)
        {
            // The caller must hold the appropriate lock.

            EnsureHashesAllocation();

            f_HashLevel |= HashKind.QuickHash;
            a_SHA256.GetDigest(f_Hashes, c_QuickHashPos);
        }

        public void SetTotalHash(SHA_256 a_SHA256)
        {
            // The caller must hold the appropriate lock.

            EnsureHashesAllocation();

            f_HashLevel |= HashKind.TotalHash;
            a_SHA256.GetDigest(f_Hashes, c_TotalHashPos);
        }

        public bool VerifyQuickHash(byte[] a_Hash,
                                    int a_Idx)
        {
            // The caller must hold the appropriate lock.

            return Buffer2.ValueEquals(f_Hashes, c_QuickHashPos, a_Hash, a_Idx, SHA_256.c_HashByteSize);
        }

        public bool VerifyTotalHash(byte[] a_Hash,
                                    int a_Idx)
        {
            // The caller must hold the appropriate lock.

            return Buffer2.ValueEquals(f_Hashes, c_TotalHashPos, a_Hash, a_Idx, SHA_256.c_HashByteSize);
        }

#endregion

        [Flags]
        private enum HashKind : byte
        {
            None = 0,
            QuickHash = 0x1,
            TotalHash = 0x2,

            QuickAndTotal = (HashKind.QuickHash | HashKind.TotalHash),
        }
    }

    sealed class BlockCache
    {
        // 512 KiB.
        public const uint c_ViewByteSize = 0x80000;

        public delegate bool DeviceReadWriteHandler(uint a_ByteOffset,
                                                    ref ByteArray a_Data);

        public readonly uint f_DeviceSize;
        private readonly DeviceReadWriteHandler f_DeviceRead;
        private readonly DeviceReadWriteHandler f_DeviceWrite;
        private readonly CacheView[] f_Views;

        // f_ViewCount determines how many elements of f_Views are
        // valid (not f_Views.Length).
        private int f_ViewCount;
        private uint f_OpCount;

        public BlockCache(uint a_MaxViews,
                          uint a_DeviceSize,
                          DeviceReadWriteHandler a_DeviceRead,
                          DeviceReadWriteHandler a_DeviceWrite)
        {
            f_DeviceSize = a_DeviceSize;
            f_DeviceRead = a_DeviceRead;
            f_DeviceWrite = a_DeviceWrite;
            f_Views = new CacheView[a_MaxViews];
        }

        public bool FlushCache()
        {
            // Returns true if all pending writes were successful.

            bool l_Succ = true;

            for (int i = 0; i < f_ViewCount; i++)
            {
                if (false == FlushViewIfDirty(i))
                {
                    l_Succ = false;
                }
            }

            return l_Succ;
        }

        private bool FlushViewIfDirty(int a_ViewIdx)
        {
            // Returns true if no flush is necessary.

            return (false == f_Views[a_ViewIdx].p_Dirty ||
                    true == FlushView(a_ViewIdx));
        }

        private bool FlushView(int a_ViewIdx)
        {
            ByteArray l_Data;
            f_Views[a_ViewIdx].BuildDirtyRange(out l_Data);

            if (false == f_DeviceWrite(f_Views[a_ViewIdx].f_ViewNumber * c_ViewByteSize, ref l_Data))
            {
                return false;
            }

            f_Views[a_ViewIdx].ClearDirtyFlag();

            return true;
        }

        private int FindViewIndex(uint a_ViewNumber)
        {
            // Return can be negative if there is no such view.

            for (int i = 0; i < f_ViewCount; i++)
            {
                if (f_Views[i].f_ViewNumber == a_ViewNumber)
                {
                    return i;
                }
            }

            return (-1);
        }

        private int FindLeastUsedViewIndex()
        {
            int l_TrgtIdx = 0;
            uint l_HighestLackOfUsage = f_Views[0].CalcLackOfUsage(f_OpCount);

            for (int i = 1; i < f_ViewCount; i++)
            {
                uint l_LackOfUsage = f_Views[i].CalcLackOfUsage(f_OpCount);

                if (l_LackOfUsage > l_HighestLackOfUsage)
                {
                    l_TrgtIdx = i;
                    l_HighestLackOfUsage = l_LackOfUsage;
                }
            }

            return l_TrgtIdx;
        }

        private void IncOpCount(int a_ViewIndex)
        {
            f_OpCount++;

            if (0 == f_OpCount)
            {
                // Overflow.

                for (int i = 0; i < f_ViewCount; i++)
                {
                    f_Views[i].ResetOpCounts();
                }

                f_OpCount = 1;
            }

            f_Views[a_ViewIndex].IncOpCount();
        }

        private bool AccessView(uint a_ViewNumber,
                                out int a_ViewIdx)
        {
            // Prepares a view for access by an operation routine.

            a_ViewIdx = FindViewIndex(a_ViewNumber);

            if (0 <= a_ViewIdx)
            {
                IncOpCount(a_ViewIdx);

                return true;
            }

            uint l_ByteOffset = (a_ViewNumber * c_ViewByteSize);
            int l_ViewByteSize = (int)Math.Min(f_DeviceSize - l_ByteOffset, c_ViewByteSize);

            if (f_Views.Length > f_ViewCount)
            {
                // Create a new view.

                a_ViewIdx = f_ViewCount++;
                f_Views[a_ViewIdx] = new CacheView(a_ViewNumber, l_ViewByteSize);
            }
            else
            {
                // Substitute an existing view.

                a_ViewIdx = FindLeastUsedViewIndex();

                if (false == FlushViewIfDirty(a_ViewIdx))
                {
                    return false;
                }

                CacheView.RecreateWithBufferReuse(a_ViewNumber, l_ViewByteSize, ref f_Views[a_ViewIdx]);
            }

            ByteArray l_Data;
            f_Views[a_ViewIdx].BuildWholeRange(out l_Data);

            if (true == f_DeviceRead(l_ByteOffset, ref l_Data))
            {
                // Set the base count before incrementing the global
                // op count.
                f_Views[a_ViewIdx].SetBaseOpCount(f_OpCount);
                IncOpCount(a_ViewIdx);

                return true;
            }

            // We failed to read the block into the cache, so remove
            // the cache view.

            f_ViewCount--;

            if (a_ViewIdx < f_ViewCount)
            {
                Array.Copy(f_Views, a_ViewIdx + 1, f_Views, a_ViewIdx, f_ViewCount - a_ViewIdx);
            }

            // Always clear the last element.
            f_Views[f_ViewCount] = new CacheView();

            return false;
        }

        private bool Operate(uint a_ViewNumber,
                             uint a_InnerByteIdx,
                             uint a_FinalInnerByteIdx,
                             ref OperationContext a_Cntxt)
        {
            int l_ViewIdx;
            if (false == AccessView(a_ViewNumber, out l_ViewIdx))
            {
                return false;
            }

            ByteArray l_Data;
            f_Views[l_ViewIdx].BuildRange(a_InnerByteIdx, a_FinalInnerByteIdx, out l_Data);
            a_Cntxt.f_Operate(ref a_Cntxt, ref l_Data);

            if (true == a_Cntxt.p_WriteAccess)
            {
                f_Views[l_ViewIdx].MarkAsDirty(a_InnerByteIdx, a_FinalInnerByteIdx);
            }

            return true;
        }

        private bool HandleOperation(uint a_ByteOffset,
                                     uint a_ByteSize,
                                     ref OperationContext a_Cntxt)
        {
            uint l_FinalByteIdx = (a_ByteOffset + a_ByteSize);

            uint l_InnerByteIdx;
            uint l_FinalInnerByteIdx;
            uint l_ViewNumber = CalcViewNumberAndInnerHashIndex(a_ByteOffset, out l_InnerByteIdx);
            uint l_LastViewNumber = CalcViewNumberAndInnerHashIndex(l_FinalByteIdx, out l_FinalInnerByteIdx);

            if (l_ViewNumber == l_LastViewNumber)
            {
                if (false == Operate(l_ViewNumber, l_InnerByteIdx, l_FinalInnerByteIdx, ref a_Cntxt))
                {
                    return false;
                }
            }
            else
            {
                // The operation spans multiple views.

                if (false == Operate(l_ViewNumber, l_InnerByteIdx, c_ViewByteSize, ref a_Cntxt))
                {
                    return false;
                }

                a_Cntxt.ProgressAdditionalDataIdx((int)(c_ViewByteSize - l_InnerByteIdx));

                for (++l_ViewNumber; l_ViewNumber < l_LastViewNumber; l_ViewNumber++)
                {
                    if (false == Operate(l_ViewNumber, 0, c_ViewByteSize, ref a_Cntxt))
                    {
                        return false;
                    }

                    a_Cntxt.ProgressAdditionalDataIdx((int)c_ViewByteSize);
                }

                if (0 != l_FinalInnerByteIdx &&
                    false == Operate(l_ViewNumber, 0, l_FinalInnerByteIdx, ref a_Cntxt))
                {
                    return false;
                }

                a_Cntxt.ResetAdditionalDataIdx();
            }

            return true;
        }

        public bool CachedHash(uint a_ByteOffset,
                               uint a_ByteSize,
                               IHashFunction a_HashAlgo)
        {
            // This is a get routine.

            OperationContext l_Cntxt;
            OperationContext.CreateForHashing(a_HashAlgo, out l_Cntxt);
            return HandleOperation(a_ByteOffset, a_ByteSize, ref l_Cntxt);
        }

        public bool ReadCached(uint a_ByteOffset,
                               ref ByteArray a_Dest)
        {
            OperationContext l_Cntxt;
            OperationContext.CreateForReading(ref a_Dest, out l_Cntxt);
            return HandleOperation(a_ByteOffset, (uint)a_Dest.f_Len, ref l_Cntxt);
        }

        public bool WriteCached(uint a_ByteOffset,
                                ref ByteArray a_Src)
        {
            OperationContext l_Cntxt;
            OperationContext.CreateForWriting(ref a_Src, out l_Cntxt);
            return HandleOperation(a_ByteOffset, (uint)a_Src.f_Len, ref l_Cntxt);
        }

        private static uint CalcViewNumberAndInnerHashIndex(uint a_ByteOffset,
                                                            out uint a_InnerByteIdx)
        {
            // Returns the view number.

            uint l_ViewNumber = (a_ByteOffset / c_ViewByteSize);
            a_InnerByteIdx = (a_ByteOffset % c_ViewByteSize);

            return l_ViewNumber;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct CacheView
        {
            // Don't use the struct default constructor.

            public readonly uint f_ViewNumber;
            public readonly byte[] f_Buffer;

            private bool f_Dirty;
            private uint f_StartingDirtyByteIdx;
            private uint f_FinalDirtyByteIdx;

            private uint f_BaseOpCount;
            private uint f_OpCount;

            public bool p_Dirty
            {
                get { return f_Dirty; }
            }

            private CacheView(uint a_ViewNumber,
                              byte[] a_Buffer)
            {
                f_ViewNumber = a_ViewNumber;
                f_Buffer = a_Buffer;

                f_Dirty = false;
                f_StartingDirtyByteIdx = 0;
                f_FinalDirtyByteIdx = 0;

                f_BaseOpCount = 0;
                f_OpCount = 0;
            }

            public CacheView(uint a_ViewNumber,
                             int a_ByteSize) :
                        this(a_ViewNumber, new byte[(int)a_ByteSize])
            {
            }

            public static void RecreateWithBufferReuse(uint a_ViewNumber,
                                                       int a_ByteSize,
                                                       ref CacheView a_View)
            {
                // This routine tries to reuse the existing buffer (if
                // the size matches).

                if (a_View.f_Buffer.Length == a_ByteSize)
                {
                    a_View = new CacheView(a_ViewNumber, a_View.f_Buffer);
                }
                else
                {
                    a_View = new CacheView(a_ViewNumber, a_ByteSize);
                }
            }

            public void ClearDirtyFlag()
            {
                f_Dirty = false;
            }

            public void MarkAsDirty(uint a_InnerByteIdx,
                                    uint a_FinalInnerByteIdx)
            {
                if (false == f_Dirty)
                {
                    f_Dirty = true;
                    f_StartingDirtyByteIdx = a_InnerByteIdx;
                    f_FinalDirtyByteIdx = a_FinalInnerByteIdx;
                }
                else
                {
                    f_StartingDirtyByteIdx = Math.Min(f_StartingDirtyByteIdx, a_InnerByteIdx);
                    f_FinalDirtyByteIdx = Math.Max(f_FinalDirtyByteIdx, a_FinalInnerByteIdx);
                }
            }

            public void BuildWholeRange(out ByteArray a_Data)
            {
                a_Data = new ByteArray(f_Buffer);
            }

            public void BuildRange(uint a_InnerByteIdx,
                                   uint a_FinalInnerByteIdx,
                                   out ByteArray a_Data)
            {
                a_Data = new ByteArray(f_Buffer, (int)a_InnerByteIdx, (int)(a_FinalInnerByteIdx - a_InnerByteIdx));
            }

            public void BuildDirtyRange(out ByteArray a_Data)
            {
                BuildRange(f_StartingDirtyByteIdx, f_FinalDirtyByteIdx, out a_Data);
            }

            public void ResetOpCounts()
            {
                f_BaseOpCount = 0;
                f_OpCount = 0;
            }

            public void SetBaseOpCount(uint a_BaseOpCount)
            {
                f_BaseOpCount = a_BaseOpCount;
            }

            public void IncOpCount()
            {
                f_OpCount++;
            }

            public uint CalcLackOfUsage(uint a_GlobalOpCount)
            {
                return ((a_GlobalOpCount - f_BaseOpCount) - f_OpCount);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct OperationContext
        {
            // Don't use the struct default constructor.

            public delegate void Operate(ref OperationContext a_Cntxt,
                                         ref ByteArray a_Data);

            private readonly static Operate s_StaticHash;
            private readonly static Operate s_StaticRead;
            private readonly static Operate s_StaticWrite;

            public readonly Operate f_Operate;

            // Only one of the following two fields is expected to
            // be valid.
            private readonly IHashFunction f_HashAlgo;
            private readonly ByteArray f_Data;
            // f_AdditionalDataIdx belongs to f_Data.
            private int f_AdditionalDataIdx;

            public bool p_WriteAccess
            {
                get { return object.ReferenceEquals(s_StaticWrite, f_Operate); }
            }

            private int p_EffIdx
            {
                get
                {
                    return (f_Data.f_Idx + f_AdditionalDataIdx);
                }
            }

            static OperationContext()
            {
                s_StaticHash = new Operate(StaticHash);
                s_StaticRead = new Operate(StaticRead);
                s_StaticWrite = new Operate(StaticWrite);
            }

            private OperationContext(IHashFunction a_HashAlgo)
            {
                // Constructor for hashing.

                f_Operate = s_StaticHash;
                f_HashAlgo = a_HashAlgo;
                f_Data = new ByteArray();
                f_AdditionalDataIdx = 0;
            }

            private OperationContext(Operate a_Operate,
                                     ref ByteArray a_Data)
            {
                // Constructor for copying.

                f_Operate = a_Operate;
                f_HashAlgo = null;
                f_Data = a_Data;
                f_AdditionalDataIdx = 0;
            }

            public static void CreateForHashing(IHashFunction a_HashAlgo,
                                                out OperationContext a_Dest)
            {
                a_Dest = new OperationContext(a_HashAlgo);
            }

            public static void CreateForReading(ref ByteArray a_DestData,
                                                out OperationContext a_Dest)
            {
                a_Dest = new OperationContext(s_StaticRead, ref a_DestData);
            }

            public static void CreateForWriting(ref ByteArray a_SrcData,
                                                out OperationContext a_Dest)
            {
                a_Dest = new OperationContext(s_StaticWrite, ref a_SrcData);
            }

            public void ResetAdditionalDataIdx()
            {
                f_AdditionalDataIdx = 0;
            }

            public void ProgressAdditionalDataIdx(int a_Length)
            {
                f_AdditionalDataIdx += a_Length;
            }

            private static void StaticHash(ref OperationContext a_Cntxt,
                                           ref ByteArray a_Data)
            {
                a_Cntxt.f_HashAlgo.Update(ref a_Data);
            }

            private static void StaticRead(ref OperationContext a_Cntxt,
                                           ref ByteArray a_Src)
            {
                // a_Cntxt.f_Data is the destination.

                Buffer.BlockCopy(a_Src.f_Arr, a_Src.f_Idx, a_Cntxt.f_Data.f_Arr, a_Cntxt.p_EffIdx, a_Src.f_Len);
            }

            private static void StaticWrite(ref OperationContext a_Cntxt,
                                            ref ByteArray a_Dest)
            {
                // a_Cntxt.f_Data is the source.

                Buffer.BlockCopy(a_Cntxt.f_Data.f_Arr, a_Cntxt.p_EffIdx, a_Dest.f_Arr, a_Dest.f_Idx, a_Dest.f_Len);
            }
        }
    }
}

namespace NB.Content.HashSets
{
    abstract class HashSet
    {
        protected const byte c_HeaderSize = 32;
        private const int c_QuickHashSegmentSize = 512;
        private const byte c_MaxCacheViews = 8;
        private const uint c_CacheThreshold = HardMemory.c_BodyBufferSize;

        protected readonly HashSetData f_SetData;
        protected readonly uint f_DestOffset;
        public readonly byte f_HashSize;
        public readonly uint f_HashCount;
        public readonly uint f_MasterHashIndex;

        // f_Buffer is just a shortcut to the much used
        // f_SetData.f_Buffer.
        protected readonly byte[] f_Buffer;

        private uint f_HashesDone;
        private bool f_Ready;
        // f_Cache can be null.
        private BlockCache f_Cache;

        protected abstract uint p_Magic
        {
            get;
        }

        protected HashSet(HashSetData a_SetData,
                          uint a_DestOffset,
                          uint a_HashSize,
                          uint a_HashCount,
                          uint a_MasterHashIndex) :
                     base()
        {
            f_SetData = a_SetData;
            f_DestOffset = (a_DestOffset + c_HeaderSize);
            f_HashSize = (byte)a_HashSize;
            f_HashCount = a_HashCount;
            f_MasterHashIndex = a_MasterHashIndex;

            f_Buffer = a_SetData.f_TempBuffer;
        }

        public bool IsComplete()
        {
            return (f_HashCount == f_HashesDone);
        }

        public bool InitializeFromDisk()
        {
            // Resets the count on error.

            byte[] l_Header = new byte[c_HeaderSize];
            ByteArray l_Data = new ByteArray(l_Header);

            if (true == f_SetData.ReadDest(f_DestOffset - c_HeaderSize, ref l_Data))
            {
                uint l_Magic = Util.ReadUInt32LE(l_Header, 0);
                ulong l_Count = Util.ReadUInt64LE(l_Header, 8);
                ulong l_Done = Util.ReadUInt64LE(l_Header, 16);

                if (p_Magic == l_Magic &&
                    0 == Util.ReadUInt32LE(l_Header, 4) &&
                    f_HashCount == l_Count &&
                    l_Done <= l_Count)
                {
                    f_HashesDone = (uint)l_Done;

                    return true;
                }
            }

            f_HashesDone = 0;

            return false;
        }

        public bool SaveProgress()
        {
            // First flush the changes, then update the header.

            if (null != f_Cache)
            {
                if (false == f_Cache.FlushCache())
                {
                    return false;
                }

                if (false == f_Ready)
                {
                    f_Cache = null;
                }
            }

            if (false == f_SetData.FlushDest())
            {
                return false;
            }

            byte[] l_Header = new byte[c_HeaderSize];
            Util.WriteUInt32LE(p_Magic, l_Header, 0);
            Util.WriteUInt64LE(f_HashCount, l_Header, 8);
            Util.WriteUInt64LE(f_HashesDone, l_Header, 16);

            ByteArray l_Data = new ByteArray(l_Header);

            return f_SetData.WriteDest(f_DestOffset - c_HeaderSize, ref l_Data);
        }

        protected static uint CalcHashSetSizeCore(uint a_HashCount,
                                                  uint a_HashSize)
        {
            return (c_HeaderSize + (a_HashCount * a_HashSize));
        }

        private void AdjustHashingBase(bool a_State)
        {
            if (a_State == f_Ready)
            {
                return;
            }

            f_Ready = a_State;

            if (true == f_Ready &&
                null == f_Cache)
            {
                uint l_Size = (f_HashCount * f_HashSize);

                if (c_CacheThreshold <= l_Size)
                {
                    f_Cache = new BlockCache(c_MaxCacheViews, l_Size, DeviceRead, DeviceWrite);
                }
            }

            AdjustHashingChild(f_Ready);
        }

        protected abstract void AdjustHashingChild(bool a_State);

        public HashingStatus ContinueHashing(ref uint a_MaxBytes)
        {
            if (f_HashCount <= f_HashesDone)
            {
                return HashingStatus.Done;
            }

            if (false == f_Ready)
            {
                AdjustHashingBase(true);
            }

            uint l_Elapsed;
            HashingStatus l_Status = Hash(f_HashesDone, ref a_MaxBytes, out l_Elapsed);

            bool l_NeedFurther = (HashingStatus.NeedFurtherCalls == l_Status);

            if (true == l_NeedFurther ||
                HashingStatus.Done == l_Status)
            {
                f_HashesDone += l_Elapsed;
            }

            if (false == l_NeedFurther)
            {
                // Release resources.

                AdjustHashingBase(false);
            }

            return l_Status;
        }

        protected abstract HashingStatus Hash(uint a_HashIndex,
                                              ref uint a_MaxBytes,
                                              out uint a_HashesElapsed);

        protected uint GetHashPos(uint a_HashIndex)
        {
            return (f_DestOffset + (a_HashIndex * f_HashSize));
        }

        [Conditional("DEBUG")]
        private void DebugAssertValidBufferBounds(uint a_HashIndex,
                                                  uint a_MaxHash,
                                                  int a_BufferSize)
        {
            if (a_MaxHash <= a_HashIndex ||
                0 != (a_BufferSize % f_HashSize) ||
                ((long)(a_MaxHash - a_HashIndex) * f_HashSize) < a_BufferSize)
            {
                throw (new ArgumentOutOfRangeException());
            }
        }

        public bool GetHashes(uint a_HashIndex,
                              ref ByteArray a_Dest)
        {
            DebugAssertValidBufferBounds(a_HashIndex, f_HashesDone, a_Dest.f_Len);

            return GetHashesCore(a_HashIndex, ref a_Dest);
        }

        protected bool GetHashesCore(uint a_HashIndex,
                                     ref ByteArray a_Dest)
        {
            // Doesn't enforce f_HashesDone as a limit like GetHashes(),
            // but allows access to the whole range of hashes.
            DebugAssertValidBufferBounds(a_HashIndex, f_HashCount, a_Dest.f_Len);

            if (null == f_Cache)
            {
                return f_SetData.ReadDest(GetHashPos(a_HashIndex), ref a_Dest);
            }

            return f_Cache.ReadCached(a_HashIndex * f_HashSize, ref a_Dest);
        }

        protected bool SetHashes(uint a_HashIndex,
                                 ref ByteArray a_Src)
        {
            DebugAssertValidBufferBounds(a_HashIndex, f_HashCount, a_Src.f_Len);

            if (null == f_Cache)
            {
                return f_SetData.WriteDest(GetHashPos(a_HashIndex), ref a_Src);
            }

            return f_Cache.WriteCached(a_HashIndex * f_HashSize, ref a_Src);
        }

        protected bool SetHash(uint a_HashIndex,
                               IHashFunction a_HashAlgo,
                               byte[] a_Buffer,
                               int a_BufferIndex)
        {
            a_HashAlgo.GetDigest(a_Buffer, a_BufferIndex);

            ByteArray l_Data = new ByteArray(a_Buffer, a_BufferIndex, a_HashAlgo.p_HashByteSize);

            return SetHashes(a_HashIndex, ref l_Data);
        }

        private bool DeviceRead(uint a_Offset,
                                ref ByteArray a_Dest)
        {
            // Used by the BlockCache class.

            return f_SetData.ReadDest(f_DestOffset + a_Offset, ref a_Dest);
        }

        private bool DeviceWrite(uint a_Offset,
                                 ref ByteArray a_Src)
        {
            // Used by the BlockCache class.

            return f_SetData.WriteDest(f_DestOffset + a_Offset, ref a_Src);
        }

        public bool CheckInputAndGetHashes(uint a_StartingHashIndex,
                                           uint a_Count,
                                           out byte[] a_Dest,
                                           out bool a_MalformedInput)
        {
            // Passed indices may be invalid (indicated by
            // a_MalformedInput on return).

            uint l_FinalIndex = (a_StartingHashIndex + a_Count);

            if (l_FinalIndex > f_HashesDone ||
                // Check for overflow.
                l_FinalIndex < a_StartingHashIndex)
            {
                a_Dest = null;
                a_MalformedInput = true;

                return false;
            }

            // We never expect this to overflow.
            a_Dest = new byte[(int)a_Count * f_HashSize];
            a_MalformedInput = false;

            ByteArray l_Data = new ByteArray(a_Dest);
            if (true == GetHashesCore(a_StartingHashIndex, ref l_Data))
            {
                return true;
            }

            a_Dest = null;

            return false;
        }

        public bool CheckInputAndGetHashes(uint[] a_HashIndices,
                                           out byte[] a_Dest,
                                           out bool a_MalformedInput)
        {
            // Passed indices may be invalid (indicated by
            // a_MalformedInput on return).

            a_Dest = null;

            int l_HashSize = (int)f_HashSize;
            byte[] l_Buf = new byte[a_HashIndices.Length * l_HashSize];

            int l_BufPos = 0;
            for (int i = 0, lim = (a_HashIndices.Length - 1); i <= lim; i++)
            {
                uint l_Index = a_HashIndices[i];

                if (l_Index >= f_HashesDone)
                {
                    a_MalformedInput = true;

                    return false;
                }

                uint l_Count = 1;

                while (lim > i)
                {
                    uint l_PeekedIndex = a_HashIndices[i + 1];

                    if ((l_Index + l_Count) != l_PeekedIndex)
                    {
                        break;
                    }

                    if (l_PeekedIndex >= f_HashesDone)
                    {
                        a_MalformedInput = true;

                        return false;
                    }

                    i++;
                    l_Count++;
                }

                int l_ByteLen = ((int)l_Count * l_HashSize);
                ByteArray l_Data = new ByteArray(l_Buf, l_BufPos, l_ByteLen);

                if (false == GetHashesCore(l_Index, ref l_Data))
                {
                    a_MalformedInput = false;

                    return false;
                }

                l_BufPos += l_ByteLen;
            }

            a_Dest = l_Buf;
            a_MalformedInput = false;

            return true;
        }

        public bool GetMasterHash(byte[] a_Dest,
                                  int a_Idx)
        {
            // NOTE: The return value for the eD2kHashSet is in big
            // endian notation and must be reversed to form a file id.

            uint l_MasterHashIndex = f_MasterHashIndex;

            if (l_MasterHashIndex < f_HashesDone)
            {
                ByteArray l_Data = new ByteArray(a_Dest, a_Idx, (int)f_HashSize);

                if (true == GetHashes(l_MasterHashIndex, ref l_Data))
                {
                    return true;
                }
            }

            a_Dest = null;

            return false;
        }

        protected HashingStatus HashFile(uint a_HashIndex,
                                         ref uint a_MaxBytes,
                                         out uint a_HashesElapsed,
                                         ref uint a_Pos,
                                         uint a_ChunkSize,
                                         IHashFunction a_HashAlgo)
        {
            // When this routines returns HashingStatus.Done,
            // a_HashAlgo will be reset and a_Pos zero.

            // The number of hashs we complete during this call.
            a_HashesElapsed = 0;

            // We have to make sure we use 64 arithmetics here.
            ulong l_FilePos = (((ulong)a_HashIndex * a_ChunkSize) + a_Pos);

            while (f_SetData.f_FileSize > l_FilePos)
            {
                // Calculate the bytes left of this chunk.
                uint l_BytesLeft = (uint)Math.Min(f_SetData.f_FileSize - l_FilePos,
                                                  a_ChunkSize - a_Pos);

                a_Pos = 0;

                while (0 < l_BytesLeft)
                {
                    if (0 == a_MaxBytes)
                    {
                        a_Pos = (uint)(l_FilePos % a_ChunkSize);

                        return HashingStatus.NeedFurtherCalls;
                    }

                    uint l_BytesToDo = Math.Min((uint)f_Buffer.Length,
                                                Math.Min(a_MaxBytes, l_BytesLeft));

                    if (false == HashFileChunk(l_FilePos, (int)l_BytesToDo, a_HashAlgo))
                    {
                        return HashingStatus.SrcFileAccessFailed;
                    }

                    l_FilePos += l_BytesToDo;
                    l_BytesLeft -= l_BytesToDo;
                    a_MaxBytes -= l_BytesToDo;
                }

                if (false == SetHash(a_HashIndex + a_HashesElapsed, a_HashAlgo, f_Buffer, 0))
                {
                    return HashingStatus.DestFileAccessFailed;
                }

                a_HashAlgo.Reset();

                a_HashesElapsed++;
            }

            return HashingStatus.Done;
        }

        private bool HashFileChunk(ulong a_FilePos,
                                   int a_Count,
                                   IHashFunction a_HashAlgo)
        {
            ByteArray l_Data = new ByteArray(f_Buffer, 0, a_Count);

            if (false == f_SetData.ReadSrc(a_FilePos, ref l_Data))
            {
                return false;
            }

            a_HashAlgo.Update(ref l_Data);

            return true;
        }

        public static bool CalcQuickHash(FHandle a_File,
                                         SHA_256 a_SHA256)
        {
            // Hashes the first and the last 512 bytes of the file (if
            // present) with the SHA256 algo.

            ulong l_FileSize;

            if (false == a_File.GetSize(out l_FileSize))
            {
                return false;
            }

            a_SHA256.Reset();
            ByteArray l_Data;

            if ((2 * c_QuickHashSegmentSize) >= l_FileSize)
            {
                // Read the whole file.

                l_Data = new ByteArray((int)l_FileSize);

                if (false == a_File.Read(0, ref l_Data))
                {
                    return false;
                }
            }
            else
            {
                // Read the first and the last 512 bytes.

                l_Data = new ByteArray(c_QuickHashSegmentSize);

                if (false == a_File.Read(0, ref l_Data))
                {
                    return false;
                }

                a_SHA256.Update(ref l_Data);

                if (false == a_File.Read(l_FileSize - c_QuickHashSegmentSize, ref l_Data))
                {
                    return false;
                }
            }

            a_SHA256.Update(ref l_Data);
            a_SHA256.Finish();

            return true;
        }

        public static HashingStatus CalcFileHash(FHandle a_File,
                                                 ulong a_FileSize,
                                                 IHashFunction a_HashAlgo,
                                                 ref uint a_MaxBytes,
                                                 ref ulong a_FilePos,
                                                 byte[] a_TempBuffer)
        {
            uint l_BytesToDo = (uint)Math.Min(a_FileSize - a_FilePos, a_MaxBytes);

            if (0 == a_FilePos &&
                0 != l_BytesToDo)
            {
                // We are starting.

                a_HashAlgo.Reset();
            }

            while (0 != l_BytesToDo)
            {
                int l_Read = (int)Math.Min(l_BytesToDo, (uint)a_TempBuffer.Length);
                ByteArray l_Data = new ByteArray(a_TempBuffer, 0, l_Read);

                if (false == a_File.Read(a_FilePos, ref l_Data))
                {
                    return HashingStatus.SrcFileAccessFailed;
                }

                a_HashAlgo.Update(ref l_Data);

                a_FilePos += (uint)l_Read;
                l_BytesToDo -= (uint)l_Read;
            }

            if (a_FileSize == a_FilePos)
            {
                a_HashAlgo.Finish();

                return HashingStatus.Done;
            }

            return HashingStatus.NeedFurtherCalls;
        }

        public class HashSetData
        {
            // f_Element is the destination (where the the hash set is
            // written to).
            private readonly HardMemory f_HardMem;
            private readonly HardMemory.Element f_Element;
            public readonly ulong f_FileSize;
            // f_Local is the source file (the file that is hashed).
            // f_Local can be null.
            private readonly LocalFile f_Local;
            // f_Buffer can be null.
            public readonly byte[] f_TempBuffer;

            public HashSetData(HardMemory a_HardMem,
                               HardMemory.Element a_Element,
                               ulong a_FileSize)
            {
                // This routine won't access a_Element.

                f_HardMem = a_HardMem;
                f_Element = a_Element;
                f_FileSize = a_FileSize;
            }

            public HashSetData(HardMemory a_HardMem,
                               HardMemory.Element a_Element,
                               LocalFile a_Local,
                               byte[] a_TempBuffer)
            {
                f_HardMem = a_HardMem;
                f_Element = a_Element;
                f_FileSize = a_Local.f_FileSize;
                f_Local = a_Local;
                f_TempBuffer = a_TempBuffer;
            }

            public bool HasDestRange(uint a_Offset,
                                     uint a_Size)
            {
                return Util.CheckRangeUnsigned(a_Offset, a_Size, f_Element.f_MaxSize);
            }

            public bool FlushDest()
            {
                return f_HardMem.Flush();
            }

            public bool ReadDest(uint a_Offset,
                                 ref ByteArray a_Data)
            {
                return f_HardMem.Read(f_Element, a_Offset, ref a_Data);
            }

            public bool WriteDest(uint a_Offset,
                                  ref ByteArray a_Data)
            {
                return f_HardMem.Write(f_Element, a_Offset, ref a_Data);
            }

            public bool ReadSrc(ulong a_Offset,
                                ref ByteArray a_Data)
            {
                return f_Local.Read(a_Offset, ref a_Data);
            }
        }
    }

    enum HashingStatus : byte
    {
        Done,
        DestFileAccessFailed,
        SrcFileAccessFailed,
        NeedFurtherCalls,
    }

    class eD2kHashSet : HashSet
    {
        public const uint c_PartSize = (9500 * 1024);
#pragma warning disable 429
        public const uint c_PartLimitFromMaxFileSize = (uint)((ContentManager.c_MaxSharedFileSize / c_PartSize) -
                                                              (0 == (ContentManager.c_MaxSharedFileSize % c_PartSize) ? 0u : 1u));
#pragma warning restore 429
        // See CalcHashCount() (roughly 37.1 PiB).
        public const ulong c_MaxFileSize = ((ulong)(uint.MaxValue - 2) * c_PartSize);

        private MD_4 f_MD4;
        private uint f_Pos;

        /* HASH VALUES OF FILES CREATED WITH FHandle.CreateDebugHashFile()
         * (HASHS AS HEXADECIMALS), VERIFIED WITH iMule AND eMule.
         *
         * SIZE = 1 BYTE
         * PARTHASHES = 1 (THIS IS ALSO THE MASTERHASH)
         * MASTERHASH = 33F9A9FBD93ECCA018B0387A4A6AEC5D
         *
         * SIZE = 1 * c_PartSize
         * HASHES = 1 + MASTER = 2
         * MASTERHASH = 21F0DBF4189F145BA95D988F3D0C90FD
         *
         * SIZE = (1 * c_PartSize) + 1 BYTE
         * HASHES = 1 + SPECIAL + MASTER = 3
         * MASTERHASH = 36CA3A30C514E233E7467988EB484811
         *
         * SIZE = 2 * c_PartSize
         * HASHES = 2 + MASTER = 3
         * MASTERHASH = FDFFEDAEDCDB5FD3EFDEF14F71CC4BA1
         *
         * SIZE = (2 * c_PartSize) + 1 BYTE
         * HASHES = 2 + SPECIAL + MASTER = 4
         * MASTERHASH = EB0596458FB5ADCD4BEBEAD2F04EDD04
         *
         * SIZE = (2^32 + 1023)
         * MASTERHASH = F2EA26C7DC79AD955728604F855FD7B5
         */

        protected override uint p_Magic
        {
            get { return 0x4B740839; }
        }

        private eD2kHashSet(HashSetData a_SetData,
                            uint a_DestOffset,
                            uint a_HashCount) :
                       base(a_SetData, a_DestOffset, MD_4.c_HashByteSize, a_HashCount, a_HashCount - 1)
        {
        }

        public static eD2kHashSet CreateeD2kHashSet(HashSetData a_SetData,
                                                    uint a_DestOffset)
        {
            uint l_HashCount = eD2kHashSet.CalcHashCount(a_SetData.f_FileSize);

            return new eD2kHashSet(a_SetData, a_DestOffset, l_HashCount);
        }

        protected override void AdjustHashingChild(bool a_State)
        {
            if (false == a_State)
            {
                f_MD4 = null;
            }
            else
            {
                f_MD4 = new MD_4();
                f_Pos = 0;
            }
        }

        public static uint CalcHashSetSize(ulong a_FileSize,
                                           out uint a_HashCount)
        {
            a_HashCount = CalcHashCount(a_FileSize);
            return HashSet.CalcHashSetSizeCore(a_HashCount, MD_4.c_HashByteSize);
        }

        public static uint CalcHashCount(ulong a_FileSize)
        {
            if (c_PartSize > a_FileSize)
            {
                return 1;
            }

            // Regardless of the slack we need to have one 'special'
            // part hash (since this is the convention) and the master
            // hash.
            return ((uint)(a_FileSize / c_PartSize) + 2);
        }

        protected override HashingStatus Hash(uint a_HashIndex,
                                              ref uint a_MaxBytes,
                                              out uint a_HashesElapsed)
        {
            uint l_HashIndex = a_HashIndex;
            uint l_MasterHashIdx = (f_HashCount - 1);

            if (l_MasterHashIdx > l_HashIndex ||
                1 == f_HashCount)
            {
                HashingStatus l_Status = HashFile(l_HashIndex,
                                                  ref a_MaxBytes,
                                                  out a_HashesElapsed,
                                                  ref f_Pos,
                                                  c_PartSize,
                                                  f_MD4);

                if (HashingStatus.Done != l_Status)
                {
                    return l_Status;
                }

                if (1 == f_HashCount)
                {
                    // We already have the sole part hash that works
                    // as our the master hash.
                    return HashingStatus.Done;
                }

                l_HashIndex += a_HashesElapsed;

                if (l_MasterHashIdx > l_HashIndex)
                {
                    // In this case we need to add a 'special hash'
                    // which is nothing but the MD4 hash value of
                    // an empty string.

                    if (false == SetHash(l_HashIndex, f_MD4, f_Buffer, 0))
                    {
                        return HashingStatus.DestFileAccessFailed;
                    }

                    f_MD4.Reset();

                    l_HashIndex++;
                    a_HashesElapsed++;
                }
            }
            else
            {
                a_HashesElapsed = 0;
            }

            return CalcMasterHash(ref a_MaxBytes, ref a_HashesElapsed);
        }

        private HashingStatus CalcMasterHash(ref uint a_MaxBytes,
                                             ref uint a_HashesElapsed)
        {
            // We use f_Pos to store the number of hashed hashs.

            uint l_Missing = (f_MasterHashIndex - f_Pos);
            uint l_Count = Math.Min(l_Missing, a_MaxBytes / MD_4.c_HashByteSize);

            int l_BytesToGo = (int)(l_Count * MD_4.c_HashByteSize);
            // Make sure l_MaxBytesPerTurn is a multiple of the hash
            // size.
            int l_MaxBytesPerTurn = ((f_Buffer.Length / MD_4.c_HashByteSize) * MD_4.c_HashByteSize);

            while (0 < l_BytesToGo)
            {
                int l_Bytes = Math.Min(l_MaxBytesPerTurn, l_BytesToGo);
                ByteArray l_Data = new ByteArray(f_Buffer, 0, l_Bytes);

                if (false == base.GetHashesCore(f_Pos, ref l_Data))
                {
                    return HashingStatus.DestFileAccessFailed;
                }

                f_MD4.Update(ref l_Data);

                f_Pos += ((uint)l_Bytes / MD_4.c_HashByteSize);
                a_MaxBytes -= (uint)l_Bytes;
                l_BytesToGo -= l_Bytes;
            }

            if (l_Count == l_Missing)
            {
                // We now have the master hash.

                if (false == base.SetHash(f_MasterHashIndex, f_MD4, f_Buffer, 0))
                {
                    return HashingStatus.DestFileAccessFailed;
                }

                a_HashesElapsed++;

                return HashingStatus.Done;
            }

            return HashingStatus.NeedFurtherCalls;
        }

        public static uint GeteD2kHashSetElementCount(ulong a_FileSize)
        {
            if (c_PartSize > a_FileSize)
            {
                return 0;
            }

            return ((uint)(a_FileSize / c_PartSize) + 1);
        }

        public static bool ConfirmeD2kUriPartHashes(ref ByteArray a_MasterHash,
                                                    byte[] a_PartHashes)
        {
            MD_4 l_MD4 = new MD_4();
            l_MD4.Update(a_PartHashes, 0, a_PartHashes.Length);
            l_MD4.Finish();

            return l_MD4.VerifyBE(a_MasterHash.f_Arr, a_MasterHash.f_Idx);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct eD2kHashSetAccessor
    {
        // Don't use the struct default constructor.

        // f_HashCount can be zero.
        private readonly uint f_HashCount;
        private readonly uint f_SrcByteBaseOffset;

        // Only one of those two should be nonzero.
        private readonly byte[] f_SrcBuffer;
        private readonly HardMemory f_SrcHardMem;

        // Only valid if f_SrcHardMem is nonzero.
        private readonly HardMemory.Element f_SrcElement;

        public eD2kHashSetAccessor(uint a_HashCount,
                                   ref ByteArray a_Data)
        {
            f_HashCount = a_HashCount;
            f_SrcByteBaseOffset = (uint)a_Data.f_Idx;

            f_SrcBuffer = a_Data.f_Arr;
            f_SrcHardMem = null;
            f_SrcElement = null;
        }

        public eD2kHashSetAccessor(uint a_HashCount,
                                   HardMemory a_HardMem,
                                   HardMemory.Element a_Element,
                                   uint a_StartingOffset)
        {
            f_HashCount = a_HashCount;
            f_SrcByteBaseOffset = a_StartingOffset;

            f_SrcBuffer = null;
            f_SrcHardMem = a_HardMem;
            f_SrcElement = a_Element;
        }

        public bool IsValid()
        {
            return (null != f_SrcBuffer ||
                    null != f_SrcHardMem);
        }

        public void AcquireSharedIfDiskBased()
        {
            if (null != f_SrcHardMem)
            {
                f_SrcHardMem.p_Resource.AcquireShared();
            }
        }

        public void ReleaseSharedIfDiskBased()
        {
            if (null != f_SrcHardMem)
            {
                f_SrcHardMem.p_Resource.ReleaseShared();
            }
        }

        public bool GetPartHashes(uint a_PartIndex,
                                  uint a_PartCount,
                                  byte[] a_Dest,
                                  int a_DestIndex)
        {
            uint l_Offset = (f_SrcByteBaseOffset + (a_PartIndex * MD_4.c_HashByteSize));
            uint l_Length = (a_PartCount * MD_4.c_HashByteSize);

            if (null != f_SrcBuffer)
            {
                Buffer.BlockCopy(f_SrcBuffer, (int)l_Offset, a_Dest, a_DestIndex, (int)l_Length);

                return true;
            }
            else
            {
                // We need to have shared access to
                // f_SrcHardMem.p_Resource (if any).
                ByteArray l_Data = new ByteArray(a_Dest, a_DestIndex, (int)l_Length);

                if (true == f_SrcHardMem.Read(f_SrcElement, l_Offset, ref l_Data))
                {
                    return true;
                }
            }

            return false;
        }

        public byte[] CheckInputAndBuildHashArray(uint a_StartingIndex,
                                                  uint a_Count,
                                                  out bool a_MalformedInput)
        {
            if (null != f_SrcBuffer)
            {
                return TAICHTreeAccessor.BuildHashArray(a_StartingIndex, a_Count, f_HashCount, MD_4.c_HashByteSize, f_SrcBuffer, f_SrcByteBaseOffset, out a_MalformedInput);
            }

            return TAICHTreeAccessor.BuildHashArray(a_StartingIndex, a_Count, f_HashCount, MD_4.c_HashByteSize, f_SrcHardMem, f_SrcElement, f_SrcByteBaseOffset, out a_MalformedInput);
        }
    }

    class AICHTree : HashSet
    {
        public const uint c_PartSize = (9500 * 1024);
        public const uint c_BlockSize = (180 * 1024);
        // 128 TiB is the file size limit in order to make hash indices
        // fit an uint.
        public const ulong c_MaxFileSize = 0x800000000000;

        private AICHTreeLayout f_TreeOpsLayout;
        private SHA_1 f_TreeOpsHasher;
        private AICHTreeLayout.HashingContext f_Cntxt;

        /* HASH VALUES OF FILES CREATED WITH FHandle.CreateDebugHashFile()
         * (HASHS IN BASE32 ENCODING), VERIFIED WITH iMule AND eMule.
         *
         * SIZE = 1 BYTE
         * TOPHASH = X6FUKMGY2JDN25FMKOQTI4N3UF4UDX7X
         *
         * SIZE = 1 * c_BlockSize
         * TOPHASH = 5DN4YNS3TGMVH67OJOQSFTTTH6P4N4TJ
         *
         * SIZE = (1 * c_BlockSize) + 1 BYTE
         * TOPHASH = OWXLCQGQWYW3O62ZDJWHAF43R4PAHJ7Q
         *
         * SIZE = 2 * c_BlockSize
         * TOPHASH = CY2F7RLCEAP3C7GVPWROHXZCWXURPK6C
         *
         * SIZE = 1 * c_PartSize
         * TOPHASH = 4LRSUJ2KJ22JVCF6WYVWLKALYQZTQTFU
         *
         * SIZE = (1 * c_PartSize) + c_BlockSize + 1
         * TOPHASH = 2RRX4UQHVEFXFVDGELVPMKRDSNU7C6S3
         *
         * SIZE = (11 * c_PartSize) + c_BlockSize + 1
         * TOPHASH = BR7ADCL4BFVRJTVT7HILD24IMRWGTE37
         *
         * SIZE = (2^32 + 1023)
         * TOPHASH = WA25Z6Z2NFYYXIV7LE7Z7M23L6RLNGV3
         */

        protected override uint p_Magic
        {
            get { return 0x3EECBC16; }
        }

        private AICHTree(HashSetData a_SetData,
                         uint a_DestOffset,
                         uint a_HashCount) :
                    base(a_SetData, a_DestOffset, SHA_1.c_HashByteSize, a_HashCount, 0)
        {
        }

        public static AICHTree CreateAICHTree(HashSetData a_SetData,
                                              uint a_DestOffset)
        {
            uint l_HashCount = AICHTreeLayout.CalcCount(a_SetData.f_FileSize);

            return new AICHTree(a_SetData, a_DestOffset, l_HashCount);
        }

        protected override void AdjustHashingChild(bool a_State)
        {
            if (false == a_State)
            {
                f_Cntxt = new AICHTreeLayout.HashingContext();
            }
            else
            {
                f_Cntxt = new AICHTreeLayout.HashingContext(this, f_Buffer, f_SetData.f_FileSize);
            }
        }

        public void StartRecoveryOperations()
        {
            f_TreeOpsLayout = new AICHTreeLayout(f_SetData.f_FileSize);
            f_TreeOpsHasher = new SHA_1();
        }

        public void EndRecoveryOperations()
        {
            f_TreeOpsLayout = null;
            f_TreeOpsHasher = null;
        }

        public static uint CalcHashSetSize(ulong a_FileSize,
                                           out uint a_HashCount)
        {
            a_HashCount = AICHTreeLayout.CalcCount(a_FileSize);
            return HashSet.CalcHashSetSizeCore(a_HashCount, SHA_1.c_HashByteSize);
        }

        public bool CheckInputAndCreatePartRecoveryData(uint a_PartIndex,
                                                        uint a_PartCount,
                                                        out byte[] a_Dest,
                                                        out bool a_MalformedInput)
        {
            if (a_PartCount > a_PartIndex)
            {
                byte[] l_MasterHash = new byte[SHA_1.c_HashByteSize];

                if (true == GetMasterHash(l_MasterHash, 0))
                {
                    MemoryIO l_Mem = MemoryIO.CreateWritable();

                    // Write the AICH master hash (since OP_AICHANSWER
                    // needs it).
                    l_Mem.WriteBytes(l_MasterHash);

                    if (true == CreatePartRecoveryData(a_PartIndex, l_Mem))
                    {
                        a_Dest = l_Mem.ToArray(true);
                        a_MalformedInput = false;

                        return true;
                    }
                }
            }

            a_Dest = null;
            a_MalformedInput = false;

            return false;
        }

        public bool CreatePartRecoveryData(uint a_PartIndex,
                                           BufferIO a_Dest)
        {
            if (1 == f_HashCount)
            {
                return false;
            }

            HashTreeNode l_Junk;
            bool l_Write32BitHashIdents;
            RecoveryEssential[] l_Arr = CreateEssentials(a_PartIndex, out l_Junk, out l_Write32BitHashIdents);

            if (true == l_Write32BitHashIdents)
            {
                a_Dest.ZeroBytes(sizeof(ushort));
            }

            uint l_Count = (uint)l_Arr.Length;
            a_Dest.WriteUInt16LE((ushort)l_Count);

            ByteArray l_Data = new ByteArray(SHA_1.c_HashByteSize);

            for (uint i = 0; i != l_Count; i++)
            {
                if (false == GetHashes(l_Arr[i].f_Node.f_HashIndex, ref l_Data))
                {
                    return false;
                }

                uint l_HashIdent = l_Arr[i].f_HashIdent;

                if (false == l_Write32BitHashIdents)
                {
                    a_Dest.WriteUInt16LE((ushort)l_HashIdent);
                }
                else
                {
                    a_Dest.WriteUInt32LE(l_HashIdent);
                }

                a_Dest.WriteBytes(ref l_Data);
            }

            if (false == l_Write32BitHashIdents)
            {
                // The protocol requires this.

                a_Dest.ZeroBytes(sizeof(ushort));
            }

            return true;
        }

        private RecoveryEssential[] CreateEssentials(uint a_PartIndex,
                                                     out HashTreeNode a_Master,
                                                     out bool a_Requires32BitHashIdents)
        {
            if (null == f_TreeOpsLayout)
            {
                throw (new UnexpectedStateException());
            }

            int l_UpperLevels;
            HashTreeNode l_Node = f_TreeOpsLayout.GetFamilyTree(a_PartIndex, out l_UpperLevels);

            a_Master = l_Node;
            a_Requires32BitHashIdents = false;

            // l_MaxLowLevelHashs is 53.
            const uint l_MaxLowLevelHashs = ((c_PartSize + (c_BlockSize - 1)) / c_BlockSize);
            RecoveryEssential[] l_Arr = new RecoveryEssential[l_UpperLevels + l_MaxLowLevelHashs];
            // The top node is always considered to be on the left,
            // so set the lowest bit.
            uint l_HashIdent = 0x1;

            // Include the upper nodes needed for verification (the
            // other sibling can be calculated from lowest level
            // and is not included).

            for (int i = 0; i < l_UpperLevels; i++)
            {
                bool l_GoLeft = l_Node.p_FirstChild.p_HasAnyChild;
                HashTreeNode l_IncludeNode;

                if (false == l_GoLeft)
                {
                    // Include the left child and proceed right.
                    l_IncludeNode = l_Node.p_FirstChild;
                    l_Node = l_Node.p_SecondChild;
                }
                else
                {
                    // Include the right child and proceed left.
                    l_IncludeNode = l_Node.p_SecondChild;
                    l_Node = l_Node.p_FirstChild;
                }

                // Place l_IncludeNode into the array.

                bool l_IncludeNodeIsLeft = !l_GoLeft;
                uint l_EffHashIdent = ((l_HashIdent << 1) | Util.BoolToByte(l_IncludeNodeIsLeft));
                l_Arr[i] = new RecoveryEssential(l_EffHashIdent, l_IncludeNode);

                if (ushort.MaxValue < l_EffHashIdent)
                {
                    a_Requires32BitHashIdents = true;
                }

                l_HashIdent <<= 1;
                l_HashIdent |= Util.BoolToByte(l_GoLeft);
            }

            // Include the lowest level nodes.

            Stack<HashTreeNode> l_Stack = new Stack<HashTreeNode>();
            int l_ArrIdx = l_UpperLevels;
            l_Node.SetNum(2);

            for (int k = 0;;)
            {
                int l_Num = l_Node.GetNum();

                if (0 == l_Num)
                {
                    if (false == l_Node.p_HasAnyChild)
                    {
                        l_Arr[l_ArrIdx] = new RecoveryEssential(l_HashIdent, l_Node);
                        l_ArrIdx++;

                        if (ushort.MaxValue < l_HashIdent)
                        {
                            a_Requires32BitHashIdents = true;
                        }
                    }

                    // Return.

                    if (0 == k)
                    {
                        break;
                    }

                    l_Node = l_Stack.Pop();
                    k--;
                    l_HashIdent >>= 1;

                    continue;
                }

                HashTreeNode l_Child = null;
                bool l_WentLeft = false;

                if (2 == l_Num)
                {
                    l_Node.SetNum(1);
                    l_Child = l_Node.p_FirstChild;

                    l_WentLeft = true;
                }

                if (null == l_Child)
                {
                    l_Node.SetNum(0);
                    l_Child = l_Node.p_SecondChild;
                }

                if (null != l_Child)
                {
                    l_Stack.Push(l_Node);
                    k++;
                    l_HashIdent <<= 1;
                    l_HashIdent |= Util.BoolToByte(l_WentLeft);

                    l_Child.SetNum(2);
                    l_Node = l_Child;
                }
            }

            // Shrink (only needed with small arrays).
            Array.Resize(ref l_Arr, l_ArrIdx);

            return l_Arr;
        }

        protected override HashingStatus Hash(uint a_HashIndex,
                                              ref uint a_MaxBytes,
                                              out uint a_HashesElapsed)
        {
            HashingStatus l_Status = f_Cntxt.Continue(ref a_MaxBytes);

            // Due to the 'chaotic' layout of the AICH tree, we can't
            // progress gradually, so it's all or nothing.
            a_HashesElapsed = (HashingStatus.Done == l_Status ? f_HashCount : 0);

            return l_Status;
        }

        public HashingStatus ContextHashFile(ulong a_FilePos,
                                             uint a_Size,
                                             ref uint a_Pos,
                                             ref uint a_MaxBytes,
                                             SHA_1 a_HashAlgo,
                                             byte[] a_TempBuffer)
        {
            ulong l_FilePos = (a_FilePos + a_Pos);
            uint l_BytesLeft = (a_Size - a_Pos);
            bool l_NeedFurther = false;

            if (l_BytesLeft > a_MaxBytes)
            {
                l_BytesLeft = a_MaxBytes;
                l_NeedFurther = true;
            }

            while (0 != l_BytesLeft)
            {
                int l_Len = (int)Math.Min(l_BytesLeft, (uint)a_TempBuffer.Length);
                ByteArray l_Data = new ByteArray(a_TempBuffer, 0, l_Len);

                if (false == f_SetData.ReadSrc(l_FilePos, ref l_Data))
                {
                    return HashingStatus.SrcFileAccessFailed;
                }

                a_HashAlgo.Update(a_TempBuffer, 0, l_Len);

                a_Pos += (uint)l_Len;
                a_MaxBytes -= (uint)l_Len;
                l_BytesLeft -= (uint)l_Len;
            }

            if (true == l_NeedFurther)
            {
                return HashingStatus.NeedFurtherCalls;
            }

            a_Pos = 0;

            return HashingStatus.Done;
        }

        public bool ContextSetHash(uint a_HashIndex,
                                   byte[] a_Src,
                                   int a_SrcIndex)
        {
            ByteArray l_Data = new ByteArray(a_Src, a_SrcIndex, SHA_1.c_HashByteSize);

            return SetHashes(a_HashIndex, ref l_Data);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct RecoveryEssential
        {
            public readonly uint f_HashIdent;
            public readonly HashTreeNode f_Node;

            public RecoveryEssential(uint a_HashIdent,
                                     HashTreeNode a_Node)
            {
                f_HashIdent = a_HashIdent;
                f_Node = a_Node;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    class AICHTreeLayout
    {
        // We use 128 TiB as an arbitrary limit, the real limit is
        // available memory (not 32 bit integers). We will use about
        // 221 MiB at most.
        public const ulong c_MaxFileSize = 0x800000000000;

        private ulong f_FileSize;
        public readonly uint f_Count;
        // We need the nodes array to quickly find the hash index of
        // a certain hash node in the tree (since the AICH tree is
        // not easy to grasp without creating it).
        private AICHTreeNode[] f_Nodes;

        public AICHTreeLayout(ulong a_FileSize)
        {
            f_FileSize = a_FileSize;
            f_Nodes = Create(a_FileSize, false, out f_Count);
        }

        public void Release()
        {
            f_Nodes = null;
        }

        public static uint CalcCount(ulong a_FileSize)
        {
            // Calculates the exact number of hashs.

            uint l_Count;
            Create(a_FileSize, true, out l_Count);

            return l_Count;
        }

        private static int EstimateArraySize(ulong a_FileSize)
        {
            // This formula is based on regression analysis.
            return ((int)(a_FileSize / 4863999) + 1);
        }

        private static int SuggestStackSize(ulong a_FileSize)
        {
            return (uint.MaxValue >= a_FileSize ? 16 : 31);
        }

        private static Level[] CreateStack(int a_Slots,
                                           bool a_NeedBuffer)
        {
            Level[] l_Stack = new Level[a_Slots];

            // Initialize each object in the array.
            for (int i = 0; a_Slots > i; i++)
            {
                l_Stack[i] = new Level(a_NeedBuffer);
            }

            return l_Stack;
        }

        private static AICHTreeNode[] Create(ulong a_FileSize,
                                             bool a_JustCalcCount,
                                             out uint a_Count)
        {
            // If a_JustCalcCount is true we won't return anything.

            if (AICHTree.c_BlockSize >= a_FileSize)
            {
                a_Count = 1;

                if (false == a_JustCalcCount)
                {
                    return new AICHTreeNode[0];
                }

                return null;
            }

            Level[] l_Stack = CreateStack(SuggestStackSize(a_FileSize), false);
            AICHTreeNode[] l_Nodes = null;

            if (false == a_JustCalcCount)
            {
                l_Nodes = new AICHTreeNode[EstimateArraySize(a_FileSize)];
            }

            l_Stack[0].InitTop(a_FileSize);

            uint l_UpperIndex = 0;
            bool l_NodeWork = false;

            if (false == a_JustCalcCount &&
                // Include all nodes that have at least one child with
                // a size that is as large as one part or larger.
                false == l_Stack[0].p_SubPartChildren)
            {
                AICHTreeNode.Initialize(out l_Nodes[0], 0);
                l_Stack[0].f_Tag = 0;

                l_UpperIndex = 1;
                l_NodeWork = true;
            }

            uint l_TotalIndex = 1;

            for (int l_Lev = 0; 0 <= l_Lev; )
            {
                int l_RefLev = l_Lev;

                if (0 == l_Stack[l_RefLev].p_Life)
                {
                    l_Lev--;

                    continue;
                }

                l_Lev++;
                l_Stack[l_RefLev].Age();

                if (0 != l_Stack[l_RefLev].p_Life)
                {
                    l_Stack[l_Lev].Init(l_TotalIndex, l_Stack[l_RefLev].f_LeftSize, true);
                }
                else
                {
                    l_Stack[l_Lev].Init(l_TotalIndex, l_Stack[l_RefLev].f_RightSize, false);
                }

                if (true == l_NodeWork)
                {
                    if (false == l_Stack[l_Lev].p_SubPartChildren)
                    {
                        l_Stack[l_Lev].f_Tag = l_UpperIndex;
                    }
                    else
                    {
                        l_Stack[l_Lev].f_Tag = uint.MaxValue;
                    }

                    if (uint.MaxValue != l_Stack[l_RefLev].f_Tag)
                    {
                        if (false == l_Stack[l_Lev].f_IsLeft)
                        {
                            uint l_Tag = l_Stack[l_RefLev].f_Tag;
                            l_Nodes[l_Tag].SetRightChildArrayIndex(l_UpperIndex);
                        }

                        AICHTreeNode.Initialize(out l_Nodes[l_UpperIndex], l_TotalIndex);
                        l_UpperIndex++;
                    }
                }

                l_TotalIndex++;
            }

            a_Count = l_TotalIndex;

            if (false == a_JustCalcCount)
            {
                Array.Resize<AICHTreeNode>(ref l_Nodes, (int)l_UpperIndex);

                return l_Nodes;
            }

            return null;
        }

        public HashTreeNode GetFamilyTree(uint a_PartIndex,
                                          out int a_UpperLevels)
        {
            ulong l_FilePos = ((ulong)a_PartIndex * AICHTree.c_PartSize);

            HashTreeNode l_Master = new HashTreeNode(0, 0);

            if (1 == f_Count)
            {
                a_UpperLevels = 0;

                return l_Master;
            }

            Level l_Ref = new Level(false);
            l_Ref.InitTop(f_FileSize);

            if (0 == f_Nodes.Length)
            {
                DebugAssertPartSize(a_PartIndex, ref l_Ref);
                SetWholeSubTree(l_Master, ref l_Ref);

                a_UpperLevels = 0;

                return l_Master;
            }

            HashTreeNode l_Cur = l_Master;
            uint l_UpperIndex = 0;

            for (int i = 0; true; i++)
            {
                AICHTreeNode l_Node = f_Nodes[l_UpperIndex];

                if (false == l_Node.p_HasChildren)
                {
                    // We have reached the part level.

                    if (0 == l_Ref.f_LeftSize)
                    {
                        // In this case we are dealing with a part at
                        // the very end of the file where the tree has
                        // no child nodes any more.

                        a_UpperLevels = (i - 1);

                        return l_Master;
                    }

                    DebugAssertPartSize(a_PartIndex, ref l_Ref);
                    SetWholeSubTree(l_Cur, ref l_Ref);

                    a_UpperLevels = i;

                    return l_Master;
                }

                uint l_LeftUpperIdx = l_Node.GetLeftChildArrayIndex(l_UpperIndex);
                uint l_RightUpperIdx = l_Node.p_RightChildArrayIndex;

                HashTreeNode l_LeftChild = new HashTreeNode(f_Nodes[l_LeftUpperIdx].f_HashIndex, i + 1);
                HashTreeNode l_RightChild = new HashTreeNode(f_Nodes[l_RightUpperIdx].f_HashIndex, i + 1);

                // Select the right path and only include nodes needed
                // for validation.

                l_Cur.SetFirstChild(l_LeftChild);
                l_Cur.SetSecondChild(l_RightChild);

                if (l_Ref.f_LeftSize > l_FilePos)
                {
                    l_Cur = l_LeftChild;
                    l_Ref.Init(l_Cur.f_HashIndex, l_Ref.f_LeftSize, true);
                    l_UpperIndex = l_LeftUpperIdx;
                }
                else
                {
                    // Go right.

                    l_FilePos -= l_Ref.f_LeftSize;

                    l_Cur = l_RightChild;
                    l_Ref.Init(l_Cur.f_HashIndex, l_Ref.f_RightSize, false);
                    l_UpperIndex = l_RightUpperIdx;
                }
            }
        }

        [Conditional("DEBUG")]
        private void DebugAssertPartSize(uint a_PartIndex,
                                         ref Level a_Ref)
        {
            // Checks if the part level is exactly reached.

            ulong l_LeftSize = (f_FileSize - (a_PartIndex * AICHTree.c_PartSize));
        }

        private static void SetWholeSubTree(HashTreeNode a_Node,
                                            ref Level a_Ref)
        {
            // We'll set up a stack large enough to process the size
            // of a part (seven levels).
            Level[] l_Stack = CreateStack(7, false);
            Stack<HashTreeNode> l_NodeStack = new Stack<HashTreeNode>(6);

            HashTreeNode l_Node = a_Node;
            l_Stack[0] = a_Ref;

            uint l_Index = (l_Node.f_HashIndex + 1);

            for (int i = 0; true; )
            {
                int l_RefLev = i;

                if (0 == l_Stack[l_RefLev].p_Life)
                {
                    if (0 == i)
                    {
                        break;
                    }

                    i--;

                    l_Node = l_NodeStack.Pop();

                    continue;
                }

                i++;
                l_Stack[l_RefLev].Age();

                HashTreeNode l_New = new HashTreeNode(l_Index, l_Node.f_Level + 1);

                if (0 != l_Stack[l_RefLev].p_Life)
                {
                    l_Stack[i].Init(l_Index, l_Stack[l_RefLev].f_LeftSize, true);

                    l_Node.SetFirstChild(l_New);
                }
                else
                {
                    l_Stack[i].Init(l_Index, l_Stack[l_RefLev].f_RightSize, false);

                    l_Node.SetSecondChild(l_New);
                }

                l_NodeStack.Push(l_Node);
                l_Node = l_New;

                l_Index++;
            }

            return;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Level
        {
            public uint f_HashIndex;
            public bool f_IsLeft;
            public ulong f_LeftSize;
            public ulong f_RightSize;
            private int f_Life;

            // Used by Create() to store the upper index.
            public uint f_Tag;

            // Used by HashingContext.
            public readonly byte[] f_ChildHashes;

            public int p_Life
            {
                get { return f_Life; }
            }

            public bool p_SubPartChildren
            {
                // Checks if both children are each below the part size.
                // We only check the left size, since it's never smaller
                // than the right size.
                get { return (AICHTree.c_PartSize > f_LeftSize); }
            }

            public Level(bool a_NeedBuffer)
            {
                f_HashIndex = 0;
                f_IsLeft = false;
                f_LeftSize = 0;
                f_RightSize = 0;
                f_Life = 0;

                f_Tag = 0;

                if (false == a_NeedBuffer)
                {
                    f_ChildHashes = null;
                }
                else
                {
                    f_ChildHashes = new byte[HashingContext.c_SizeOfPair];
                }
            }

            public void InitTop(ulong a_FileSize)
            {
                Init(0, a_FileSize, true);
            }

            public void Init(uint a_HashIndex,
                             ulong a_Size,
                             bool a_IsLeft)
            {
                // This routine ought to initialize each core field.

                f_HashIndex = a_HashIndex;
                f_IsLeft = a_IsLeft;

                uint l_BlockSize = (AICHTree.c_PartSize < a_Size ? AICHTree.c_PartSize : AICHTree.c_BlockSize);

                if (l_BlockSize < a_Size)
                {
                    // This node has children, so calculate the sizes.

                    uint l_Blocks = ((uint)(a_Size / l_BlockSize) + (0 == (a_Size % l_BlockSize) ? 0u : 1u));

                    f_LeftSize = ((ulong)((true == a_IsLeft ? (l_Blocks + 1) : l_Blocks) / 2) * l_BlockSize);
                    f_RightSize = (a_Size - f_LeftSize);
                    f_Life = 2;
                }
                else
                {
                    // This node has no children.

                    f_LeftSize = 0;
                    f_RightSize = 0;
                    f_Life = 0;
                }
            }

            public void Age()
            {
                f_Life--;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct AICHTreeNode
        {
            // This structure should be as small as possible, to save
            // memory (memory consuption can be colossal with large
            // files).

            public const uint c_NoIndex = uint.MaxValue;

            public readonly uint f_HashIndex;
            private uint f_RightChildArrayIndex;

            public bool p_HasChildren
            {
                get { return (c_NoIndex != f_RightChildArrayIndex); }
            }

            public uint p_RightChildArrayIndex
            {
                get
                {
                    return f_RightChildArrayIndex;
                }
            }

            private AICHTreeNode(uint a_HashIndex,
                                 uint a_RightChildArrayIndex)
            {
                f_HashIndex = a_HashIndex;
                f_RightChildArrayIndex = a_RightChildArrayIndex;
            }

            public static void Initialize(out AICHTreeNode a_Node,
                                          uint a_HashIndex)
            {
                a_Node = new AICHTreeNode(a_HashIndex, c_NoIndex);
            }

            public void SetRightChildArrayIndex(uint a_ChildArrayIndex)
            {
                f_RightChildArrayIndex = a_ChildArrayIndex;
            }

            public uint GetLeftChildArrayIndex(uint a_ParentArrayIndex)
            {
                // Only valid if SetRightChildArrayIndex() was used before.

                // NOTE: The left child (if any) has always the same
                // hash as the parent plus one.
                return (a_ParentArrayIndex + 1);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct HashingContext
        {
            public const byte c_SizeOfPair = (SHA_1.c_HashByteSize * 2);

            private readonly AICHTree f_Tree;
            private readonly byte[] f_Buffer;
            private readonly ulong f_FileSize;
            private readonly SHA_1 f_SHA1;
            private readonly Level[] f_Stack;

            private int f_Lev;
            private uint f_Index;
            private ulong f_FilePos;
            private uint f_InnerChunkPos;

            public HashingContext(AICHTree a_Creator,
                                  byte[] a_Buffer,
                                  ulong a_FileSize)
            {
                f_Tree = a_Creator;
                f_Buffer = a_Buffer;
                f_FileSize = a_FileSize;
                f_SHA1 = new SHA_1();
                f_Stack = CreateStack(SuggestStackSize(f_FileSize), true);
                f_Stack[0].InitTop(f_FileSize);

                f_Lev = 0;
                f_Index = 1;
                f_FilePos = 0;
                f_InnerChunkPos = 0;
            }

            public HashingStatus Continue(ref uint a_MaxBytes)
            {
                // See AICHTreeLayout.Create().

                for (; true; )
                {
                    int l_RefLev = f_Lev;

                    if (0 == f_Stack[l_RefLev].p_Life)
                    {
                        if (0 == f_Stack[l_RefLev].f_LeftSize)
                        {
                            // There are no child nodes any more, so
                            // hash the file content.

                            ulong l_ChunkSize;

                            if (0 == f_Lev)
                            {
                                // We'll only get here for files that
                                // are one block large (c_BlockSize)
                                // at most.
                                l_ChunkSize = f_FileSize;
                            }
                            else
                            {
                                int l_ParentLev = (f_Lev - 1);
                                l_ChunkSize = (true == f_Stack[l_RefLev].f_IsLeft ? f_Stack[l_ParentLev].f_LeftSize : f_Stack[l_ParentLev].f_RightSize);
                            }

                            HashingStatus l_Status = this.f_Tree.ContextHashFile(f_FilePos,
                                                                                 (uint)l_ChunkSize,
                                                                                 ref f_InnerChunkPos,
                                                                                 ref a_MaxBytes,
                                                                                 f_SHA1,
                                                                                 f_Buffer);

                            if (HashingStatus.Done != l_Status)
                            {
                                return l_Status;
                            }
                        }
                        else
                        {
                            // This node has children and we'll now
                            // collect their hashs.

                            if (c_SizeOfPair > a_MaxBytes)
                            {
                                return HashingStatus.NeedFurtherCalls;
                            }

                            f_SHA1.Update(f_Stack[l_RefLev].f_ChildHashes);

                            a_MaxBytes -= c_SizeOfPair;
                        }

                        if (false == SetHash(ref f_Stack[l_RefLev]))
                        {
                            return HashingStatus.DestFileAccessFailed;
                        }

                        if (0 > --f_Lev)
                        {
                            break;
                        }

                        f_FilePos -= f_Stack[l_RefLev].f_LeftSize;

                        continue;
                    }

                    f_Lev++;
                    f_Stack[l_RefLev].Age();

                    if (0 != f_Stack[l_RefLev].p_Life)
                    {
                        f_Stack[f_Lev].Init(f_Index, f_Stack[l_RefLev].f_LeftSize, true);
                    }
                    else
                    {
                        f_FilePos += f_Stack[l_RefLev].f_LeftSize;
                        f_Stack[f_Lev].Init(f_Index, f_Stack[l_RefLev].f_RightSize, false);
                    }

                    f_Index++;
                }

                return HashingStatus.Done;
            }

            private bool SetHash(ref Level a_Ref)
            {
                byte[] l_Hash;
                int l_Idx;

                if (0 == f_Lev)
                {
                    // Create a temporary buffer.

                    l_Hash = new byte[SHA_1.c_HashByteSize];
                    l_Idx = 0;
                }
                else
                {
                    l_Hash = f_Stack[f_Lev - 1].f_ChildHashes;
                    l_Idx = (true == a_Ref.f_IsLeft ? 0 : (int)SHA_1.c_HashByteSize);
                }

                f_SHA1.GetDigest(l_Hash, l_Idx);
                f_SHA1.Reset();

                return f_Tree.ContextSetHash(a_Ref.f_HashIndex, l_Hash, l_Idx);
            }
        }
    }

    class TAICHTree : HashSet
    {
        public const uint c_BlockSize = (512 * 1024);
        public const byte c_PairSize = (2 * Tiger.c_HashByteSize);

        private TAICHTreeLayout f_HashingLayout;
        private Tiger f_Tiger;
        private uint f_Pos;

        /* HASH VALUES OF FILES CREATED WITH FHandle.CreateDebugHashFile()
         * (HASHS AS HEXADECIMALS)
         *
         * SIZE = 1 BYTE
         * TOPHASH = BDE9ABFBFF763F7B6C90F5C90486DC5508E9B8968A864EBE
         *
         * SIZE = 1 * c_BlockSize
         * TOPHASH = F94AC17EA044FBA0153FA27CA60F1F615B29765B231220B9
         *
         * SIZE = (1 * c_BlockSize) + 1 BYTE
         * TOPHASH = 56181FDE912E9CE48CFBD20017521F243B322D2158456A3E
         *
         * SIZE = 2 * c_BlockSize
         * TOPHASH = 4E6C8788F07A2B4509F073D88689B918293AD50BDC30D41D
         *
         * SIZE = (2 * c_BlockSize) + 1 BYTE
         * TOPHASH = BC927250E2CEA174BB9B30E94205E03E4CF06D45C933ABB7
         *
         * SIZE = (2^32 + 1023)
         * TOPHASH = 62FD68C80309C758F2149C0BA9299F1344F9D699331D3BA4
         */

        protected override uint p_Magic
        {
            get { return 0xDBD0FDB4; }
        }

        private TAICHTree(HashSetData a_SetData,
                          uint a_DestOffset,
                          uint a_HashCount) :
                     base(a_SetData, a_DestOffset, Tiger.c_HashByteSize, a_HashCount, a_HashCount - 1)
        {
        }

        public static TAICHTree CreateTAICHTree(HashSetData a_SetData,
                                                uint a_DestOffset)
        {
            uint l_HashCount = TAICHTreeLayout.CalcTotal(a_SetData.f_FileSize);

            return new TAICHTree(a_SetData, a_DestOffset, l_HashCount);
        }

        protected override void AdjustHashingChild(bool a_State)
        {
            if (false == a_State)
            {
                f_HashingLayout = new TAICHTreeLayout();
                f_Tiger = null;
            }
            else
            {
                f_HashingLayout = new TAICHTreeLayout(f_SetData.f_FileSize);
                f_Tiger = new Tiger();
                f_Pos = 0;
            }
        }

        public static uint CalcHashSetSize(ulong a_FileSize,
                                           out uint a_HashCount)
        {
            a_HashCount = TAICHTreeLayout.CalcTotal(a_FileSize);
            return HashSet.CalcHashSetSizeCore(a_HashCount, Tiger.c_HashByteSize);
        }

        protected override HashingStatus Hash(uint a_StartingHashIndex,
                                              ref uint a_MaxBytes,
                                              out uint a_HashesElapsed)
        {
            uint l_InnerIndex;
            int l_Level = f_HashingLayout.GetLevel(a_StartingHashIndex, out l_InnerIndex);

            if (f_HashingLayout.p_MaxLevel == l_Level)
            {
                HashingStatus l_Status = HashFile(a_StartingHashIndex,
                                                  ref a_MaxBytes,
                                                  out a_HashesElapsed,
                                                  ref f_Pos,
                                                  c_BlockSize,
                                                  f_Tiger);

                if (HashingStatus.Done != l_Status)
                {
                    return l_Status;
                }

                f_Pos = uint.MaxValue;

                l_InnerIndex = 0;
                l_Level--;
            }
            else
            {
                a_HashesElapsed = 0;
            }

            uint l_HashIndex = (a_StartingHashIndex + a_HashesElapsed);

            bool l_BufferPrepared = false;
            for (; 0 <= l_Level; l_Level--)
            {
                if (c_PairSize > a_MaxBytes)
                {
                    a_HashesElapsed = (l_HashIndex - a_StartingHashIndex);

                    return HashingStatus.NeedFurtherCalls;
                }

                if (false == HashLevel(l_Level, l_InnerIndex, ref l_HashIndex, ref a_MaxBytes, ref l_BufferPrepared))
                {
                    return HashingStatus.DestFileAccessFailed;
                }

                l_InnerIndex = 0;
            }

            a_HashesElapsed = (l_HashIndex - a_StartingHashIndex);

            return HashingStatus.Done;
        }

        private bool HashLevel(int a_Level,
                               uint a_InnerIndex,
                               ref uint a_HashIndex,
                               ref uint a_MaxBytes,
                               ref bool a_BufferPrepared)
        {
            bool l_BufferSpoiled = (false == a_BufferPrepared);

            a_BufferPrepared = false;

            uint l_InnerCount = f_HashingLayout.GetLevelCount(a_Level);
            uint l_HashesLeft = (l_InnerCount - a_InnerIndex);

            uint l_ExtraOrigin;
            bool l_HasExtra = f_HashingLayout.GetExtraOrigin(a_Level, out l_ExtraOrigin);

            uint l_PairsToGo = (false == l_HasExtra ? l_HashesLeft : (l_HashesLeft - 1));
            l_PairsToGo = Math.Min(l_PairsToGo, a_MaxBytes / c_PairSize);
            uint l_PairsPerBuffer = ((uint)f_Buffer.Length / c_PairSize);

            if (0 == l_PairsToGo)
            {
                l_BufferSpoiled = true;
            }
            else
            {
                uint l_SrcHashIndex = (a_HashIndex - f_HashingLayout.GetLevelCount(a_Level + 1));
                // For each inner index unit: subtract one and add two.
                // So in effect add one.
                l_SrcHashIndex += a_InnerIndex;

                do
                {
                    uint l_PairsThisRound = Math.Min(l_PairsToGo, l_PairsPerBuffer);

                    ByteArray l_SrcData = new ByteArray(f_Buffer, 0, (int)(l_PairsThisRound * c_PairSize));

                    if (true == l_BufferSpoiled &&
                        false == GetHashesCore(l_SrcHashIndex, ref l_SrcData))
                    {
                        return false;
                    }

                    HashPairs(l_PairsThisRound);
                    a_MaxBytes -= (uint)l_SrcData.f_Len;

                    ByteArray l_DestData = new ByteArray(f_Buffer, 0, l_SrcData.f_Len / 2);

                    if (false == SetHashes(a_HashIndex, ref l_DestData))
                    {
                        return false;
                    }

                    l_BufferSpoiled = ((false == l_HasExtra ? l_PairsThisRound : (l_PairsThisRound + 1)) != l_InnerCount);

                    a_HashIndex += l_PairsThisRound;
                    l_HashesLeft -= l_PairsThisRound;
                    l_PairsToGo -= l_PairsThisRound;
                    l_SrcHashIndex += (l_PairsThisRound * 2);
                }
                while (0 < l_PairsToGo);
            }

            if (1 == l_HashesLeft &&
                true == l_HasExtra)
            {
                int l_BufPos = (int)(l_InnerCount * Tiger.c_HashByteSize);

                if (true == l_BufferSpoiled ||
                    f_Buffer.Length < l_BufPos)
                {
                    l_BufPos = 0;
                    l_BufferSpoiled = true;
                }
                else
                {
                    l_BufPos -= Tiger.c_HashByteSize;
                }

                ByteArray l_Data = new ByteArray(f_Buffer, l_BufPos, Tiger.c_HashByteSize);

                if (false == GetHashesCore(l_ExtraOrigin, ref l_Data) ||
                    false == SetHashes(a_HashIndex, ref l_Data))
                {
                    return false;
                }

                a_HashIndex++;
            }

            if (false == l_BufferSpoiled)
            {
                a_BufferPrepared = true;
            }

            return true;
        }

        private void HashPairs(uint a_Pairs)
        {
            int l_SrcIdx = 0;
            int l_DestIdx = 0;
            for (uint i = 0; i < a_Pairs; i++)
            {
                f_Tiger.Reset();
                f_Tiger.Update(f_Buffer, l_SrcIdx, c_PairSize);
                f_Tiger.GetDigest(f_Buffer, l_DestIdx);

                l_SrcIdx += c_PairSize;
                l_DestIdx += Tiger.c_HashByteSize;
            }
        }

        public static TAICHVerifyResult VerifyUpperTree(ref TAICHTreeLayout a_Layout,
                                                        ref TAICHTreeAccessor a_UpperTree,
                                                        Tiger a_HashAlgo)
        {
            // This routine is thread safe. a_UpperTree.f_GreatestUpperLevel
            // may be any level index from zero to count minus one.

            uint l_BaseSrcPos = 0;
            uint l_UpperSrcPos;
            bool l_Extra = false;
            uint l_ExtraOriginPos = 0;
            for (int l_Lev = a_UpperTree.f_GreatestUpperLevel; 0 < l_Lev; l_Lev--)
            {
                uint l_HashesToGo = a_Layout.GetLevelCount(l_Lev);
                uint l_NextBaseSrcPos = (l_BaseSrcPos + (l_HashesToGo * Tiger.c_HashByteSize));
                l_UpperSrcPos = l_NextBaseSrcPos;

                uint l_Pairs = (l_HashesToGo / 2);

                for (uint i = 0; i < l_Pairs; i++)
                {
                    TAICHVerifyResult l_VerRes = a_UpperTree.CalcAndVerifyHash(a_HashAlgo, l_BaseSrcPos, l_UpperSrcPos);

                    if (TAICHVerifyResult.Valid != l_VerRes)
                    {
                        return l_VerRes;
                    }

                    l_BaseSrcPos += c_PairSize;
                    l_UpperSrcPos += Tiger.c_HashByteSize;
                }

                if (true == l_Extra)
                {
                    if (true == a_Layout.HasExtra(l_Lev))
                    {
                        TAICHVerifyResult l_VerRes = a_UpperTree.VerifyHash(l_ExtraOriginPos, l_BaseSrcPos - Tiger.c_HashByteSize);

                        if (TAICHVerifyResult.Valid != l_VerRes)
                        {
                            return l_VerRes;
                        }

                        l_Extra = false;
                    }
                }
                else if (0 != (l_HashesToGo & 1))
                {
                    // What we are looking for is any level with an
                    // odd count, except the top level.

                    l_Extra = true;
                    l_ExtraOriginPos = l_BaseSrcPos;
                }

                l_BaseSrcPos = l_NextBaseSrcPos;
            }

            return TAICHVerifyResult.Valid;
        }

        public static TAICHVerifyResult VerifySubTree(ref TAICHTreeLayout a_Layout,
                                                      byte a_Exponent,
                                                      uint a_Chunk,
                                                      ref TAICHTreeAccessor a_UpperTree,
                                                      ref TigerHashHive a_ProvidedHashes,
                                                      Tiger a_HashAlgo)
        {
            // This routine is thread safe. a_ProvidedHashes should
            // contain data according to the result of
            // CalcSubTreeIndices().

            uint l_FirstHashIndex;
            uint l_BaseCount;
            if (false == a_Layout.GetLowLevelHashesOfChunk(a_Exponent, a_Chunk, out l_FirstHashIndex, out l_BaseCount))
            {
                return TAICHVerifyResult.InvalidInput;
            }

            byte[] l_Buf = new byte[Tiger.c_HashByteSize];
            List<uint> l_CurLs = new List<uint>();

            for (uint i = l_FirstHashIndex, lim = (l_FirstHashIndex + l_BaseCount); i < lim; i++)
            {
                l_CurLs.Add(i);
            }

            List<uint> l_NextLs = new List<uint>();
            TigerHashHive l_CalculatedHashes;
            TigerHashHive.CreateExpandable(0, out l_CalculatedHashes);

            for (int l_Lev = a_Layout.p_MaxLevel; a_UpperTree.f_GreatestUpperLevel < l_Lev; l_Lev--)
            {
                bool l_LastRound = ((a_UpperTree.f_GreatestUpperLevel + 1) == l_Lev);

                foreach (uint l_HashIndex in l_CurLs)
                {
                    uint l_ParentHash;
                    uint l_SiblingHash;
                    bool l_ExtraForward;
                    int l_ParentLev = a_Layout.GetHashInfo(l_HashIndex, out l_ParentHash, out l_SiblingHash, out l_ExtraForward);

                    if (false == l_ExtraForward)
                    {
                        bool l_GotIt = l_CalculatedHashes.Get(l_ParentHash, l_Buf, 0);

                        if (false == l_GotIt)
                        {
                            if (false == HashPair(ref l_CalculatedHashes, ref a_ProvidedHashes, l_HashIndex, l_SiblingHash, a_HashAlgo))
                            {
                                return TAICHVerifyResult.RelevantHashesNotProvided;
                            }

                            a_HashAlgo.GetDigest(l_Buf, 0);
                        }

                        if (false == l_LastRound)
                        {
                            Util.AddSorted<uint>(l_NextLs, l_ParentHash);

                            if (false == l_GotIt)
                            {
                                l_CalculatedHashes.Add(l_ParentHash, l_Buf);
                            }
                        }
                        else
                        {
                            TAICHVerifyResult l_VerRes = a_UpperTree.VerifyHash(l_ParentHash, l_Buf);

                            if (TAICHVerifyResult.Valid != l_VerRes)
                            {
                                return l_VerRes;
                            }
                        }
                    }
                    else
                    {
                        uint l_ExtraHashIndex = (l_SiblingHash + 1);
                        if (false == l_CalculatedHashes.Get(l_HashIndex, l_Buf, 0) &&
                            false == a_ProvidedHashes.Get(l_HashIndex, l_Buf, 0))
                        {
                            return TAICHVerifyResult.RelevantHashesNotProvided;
                        }

                        if (l_ParentLev >= a_UpperTree.f_GreatestUpperLevel)
                        {
                            l_CalculatedHashes.Add(l_ExtraHashIndex, l_Buf);
                        }
                        else
                        {
                            TAICHVerifyResult l_VerRes = a_UpperTree.VerifyHash(l_ExtraHashIndex, l_Buf);

                            if (TAICHVerifyResult.Valid != l_VerRes)
                            {
                                return l_VerRes;
                            }
                        }
                    }
                }

                Util.Swap<List<uint>>(ref l_CurLs, ref l_NextLs);
                l_NextLs.Clear();
            }

            return TAICHVerifyResult.Valid;
        }

        private static bool HashPair(ref TigerHashHive a_PrimarySrc,
                                     ref TigerHashHive a_SecondarySrc,
                                     uint a_SiblingIndex1,
                                     uint a_SiblingIndex2,
                                     Tiger a_HashAlgo)
        {
            if (a_SiblingIndex1 > a_SiblingIndex2)
            {
                // We need to sort the indices first, hashs are
                // calculated from left to right.
                a_SiblingIndex1 = a_SiblingIndex2;
                a_SiblingIndex2 = (a_SiblingIndex1 + 1);
            }

            a_HashAlgo.Reset();

            return ((true == a_PrimarySrc.UpdateHashFunction(a_HashAlgo, a_SiblingIndex1) ||
                     true == a_SecondarySrc.UpdateHashFunction(a_HashAlgo, a_SiblingIndex1)) &&
                    (true == a_PrimarySrc.UpdateHashFunction(a_HashAlgo, a_SiblingIndex2) ||
                     true == a_SecondarySrc.UpdateHashFunction(a_HashAlgo, a_SiblingIndex2)));
        }

        public static List<uint> CalcSubTreeIndices(ref TAICHTreeLayout a_Layout,
                                                    byte a_Exponent,
                                                    uint a_Chunk,
                                                    int a_GreatestUpperLevel)
        {
            // This routine is thread safe.

            uint l_FirstHashIndex;
            uint l_BaseCount;
            if (false == a_Layout.GetLowLevelHashesOfChunk(a_Exponent, a_Chunk, out l_FirstHashIndex, out l_BaseCount))
            {
                // The chunk specified by a_Chunk doesn't exist.

                return null;
            }

            int l_Span = (a_Layout.p_MaxLevel - a_GreatestUpperLevel);
            List<uint>[] l_Ls = new List<uint>[l_Span];

            for (int i = 0; i < l_Span; i++)
            {
                l_Ls[i] = new List<uint>();
            }

            List<uint> l_RequiredHashes = new List<uint>((int)l_BaseCount);

            for (uint i = l_FirstHashIndex, lim = (l_FirstHashIndex + l_BaseCount); i < lim; i++)
            {
                l_RequiredHashes.Add(i);
                l_Ls[0].Add(i);
            }

            List<ExtraInfo> l_ExtraLs = new List<ExtraInfo>();

            for (int y = 0; y < l_Span; y++)
            {
                List<uint> l_CurLs = l_Ls[y];
                List<uint> l_ParentLs = ((y + 1) >= l_Span ? null : l_Ls[y + 1]);

                for (int x = 0, lim = (l_CurLs.Count - 1); x <= lim; x++)
                {
                    uint l_HashIndex = l_CurLs[x];

                    uint l_ParentHash;
                    uint l_SiblingHash;
                    bool l_ExtraForward;
                    int l_ParentLev = a_Layout.GetHashInfo(l_HashIndex, out l_ParentHash, out l_SiblingHash, out l_ExtraForward);

                    if (false == l_ExtraForward)
                    {
                        bool l_IsUnavailable = ((0 == x &&
                                                 l_SiblingHash < l_HashIndex) ||
                                                (lim == x &&
                                                 l_SiblingHash > l_HashIndex));

                        if (true == l_IsUnavailable)
                        {
                            Util.AddSorted<uint>(l_RequiredHashes, l_SiblingHash);
                        }

                        if (null != l_ParentLs)
                        {
                            Util.AddSorted<uint>(l_ParentLs, l_ParentHash);
                        }
                    }
                    else
                    {
                        if (a_GreatestUpperLevel <= l_ParentLev)
                        {
                            l_ExtraLs.Add(new ExtraInfo(l_SiblingHash + 1, l_ParentLev + 1));
                        }
                    }
                }
            }

            foreach (ExtraInfo l_Info in l_ExtraLs)
            {
                // Make sure we haven't put the extra hash on the list
                // of required hashs.
                l_RequiredHashes.Remove(l_Info.f_ExtraIndex);

                int l_ListIdx = (a_Layout.p_MaxLevel - l_Info.f_Level);
                if (false == l_Ls[l_ListIdx].Contains(l_Info.p_SiblingIndex))
                {
                    l_RequiredHashes.Add(l_Info.p_SiblingIndex);
                }
            }

            l_RequiredHashes.TrimExcess();

            return l_RequiredHashes;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ExtraInfo
        {
            public readonly uint f_ExtraIndex;
            public readonly int f_Level;

            public uint p_SiblingIndex
            {
                get { return (f_ExtraIndex - 1); }
            }

            public ExtraInfo(uint a_ExtraIndex,
                             int a_Level)
            {
                f_ExtraIndex = a_ExtraIndex;
                f_Level = a_Level;
            }
        }
    }

    enum TAICHVerifyResult : int
    {
        Valid = 0,

        // Error codes.
        InvalidInput = 1,
        RelevantHashesNotProvided = 2,
        // AccessFailed happens if the TAICHTreeAccessor fails.
        AccessFailed = 3,
        HashMismatch = 4,
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TAICHTreeLayout
    {
        // Don't use the struct default constructor.

        // Exactly one PiB.
        public const ulong c_MaxFileSize = 0x4000000000000;

        public readonly uint f_Total;
        private readonly uint[] f_Counts;

        public int p_Levels
        {
            get { return f_Counts.Length; }
        }

        public int p_MaxLevel
        {
            get { return (f_Counts.Length - 1); }
        }

        public uint p_TopHashIndex
        {
            get { return (f_Total - 1); }
        }

        public TAICHTreeLayout(ulong a_FileSize)
        {
            List<uint> l_Ls = new List<uint>();

            f_Total = Calc(a_FileSize, l_Ls);
            f_Counts = l_Ls.ToArray();
        }

        public bool IsValid()
        {
            return (null != f_Counts);
        }

        public static uint CalcTotal(ulong a_FileSize)
        {
            return Calc(a_FileSize, null);
        }

        private static uint Calc(ulong a_FileSize,
                                 List<uint> a_List)
        {
            // a_List can be null.

            uint l_Total = 0;
            uint l_Left = (uint)((a_FileSize / TAICHTree.c_BlockSize) +
                                 (0 == (a_FileSize % TAICHTree.c_BlockSize) ? 0u : 1u));

            bool l_Extra = false;
            while (true)
            {
                if (true == l_Extra)
                {
                    if (1 == (l_Left & 1))
                    {
                        l_Left++;
                        l_Extra = false;
                    }
                }

                if (null != a_List)
                {
                    a_List.Add(l_Left);
                }

                l_Total = (l_Total + l_Left);

                if (false == l_Extra)
                {
                    if (1 == l_Left)
                    {
                        break;
                    }

                    l_Extra = (0 != (l_Left & 1));
                }

                l_Left /= 2;
            }

            if (null != a_List)
            {
                a_List.Reverse();
            }

            return l_Total;
        }

        public uint GetLevelCount(int a_Level)
        {
            return f_Counts[a_Level];
        }

        public uint SumUpLevelCount(int a_StartingLevel,
                                    int a_EndingLevel)
        {
            // If a_EndingLevel is greater than a_StartingLevel this
            // routine will return zero.

            uint l_Count = 0;

            for (int i = a_StartingLevel; a_EndingLevel >= i; i++)
            {
                l_Count += f_Counts[i];
            }

            return l_Count;
        }

        public bool HasExtra(int a_Level)
        {
            uint l_Count = f_Counts[a_Level];

            return (// If the level in question has an even count.
                    0 == (l_Count & 1) &&
                    // If a lower level exists and if it's too small
                    // to account for the size of count.
                    (a_Level + 1) < f_Counts.Length &&
                    (2 * l_Count) > f_Counts[a_Level + 1]);
        }

        public bool GetExtraOrigin(int a_PossibleExtraLevel,
                                   out uint a_ExtraOriginHashIndex)
        {
            // Returns the hash index of the origin of the extra of
            // the passed level.

            if (false == HasExtra(a_PossibleExtraLevel))
            {
                a_ExtraOriginHashIndex = 0;

                return false;
            }

            for (int i = (a_PossibleExtraLevel + 1); i < f_Counts.Length; i++)
            {
                if (0 != (f_Counts[i] & 1))
                {
                    uint l_Idx = 0;

                    for (int k = p_MaxLevel; k >= i; k--)
                    {
                        l_Idx += f_Counts[k];
                    }

                    a_ExtraOriginHashIndex = (l_Idx - 1);

                    return true;
                }
            }

            throw (new ApplicationException());
        }

        public int GetLevel(uint a_HashIndex,
                            out uint a_InnerIndex)
        {
            uint l_Junk;
            return GetLevel(a_HashIndex, out a_InnerIndex, out l_Junk);
        }

        private int GetLevel(uint a_HashIndex,
                             out uint a_InnerIndex,
                             out uint a_Sum)
        {
            a_Sum = 0;

            for (int i = p_MaxLevel; 0 <= i; i--)
            {
                uint l_NextSum = (a_Sum + f_Counts[i]);

                if (l_NextSum > a_HashIndex)
                {
                    a_InnerIndex = (a_HashIndex - a_Sum);

                    return i;
                }

                a_Sum = l_NextSum;
            }

            throw (new ApplicationException());
        }

        public int GetHashInfo(uint a_HashIndex,
                               out uint a_ParentHashIndex,
                               out uint a_SiblingHashIndex,
                               out bool a_ExtraForward)
        {
            // Returns the parent's level.

            // Don't call this routine to find out about the top hash,
            // since it has no parent.

            uint l_InnerIndex;
            uint l_Sum;
            int l_Level = GetLevel(a_HashIndex, out l_InnerIndex, out l_Sum);

            uint l_Count = f_Counts[l_Level];

            if (0 == (l_InnerIndex & 1))
            {
                if (l_InnerIndex == (l_Count - 1))
                {
                    // Find the next level that has an extra.

                    uint l_OldCount = l_Count;
                    for (l_Level--; true; l_Level--)
                    {
                        l_Sum += l_OldCount;
                        l_Count = f_Counts[l_Level];

                        if (0 == (l_Count & 1) &&
                            (2 * l_Count) > l_OldCount)
                        {
                            l_InnerIndex = (l_Count - 1);
                            a_SiblingHashIndex = (l_Sum + l_InnerIndex - 1);
                            a_ExtraForward = true;

                            break;
                        }

                        l_OldCount = l_Count;
                    }
                }
                else
                {
                    // Join right neighbour.
                    a_SiblingHashIndex = (a_HashIndex + 1);
                    a_ExtraForward = false;
                }
            }
            else
            {
                // Join left neighbour.
                a_SiblingHashIndex = (a_HashIndex - 1);
                a_ExtraForward = false;
            }

            a_ParentHashIndex = (l_Sum + l_Count + (l_InnerIndex / 2));

            return (l_Level - 1);
        }

        public HashTreeNode GetFamilyTree(uint a_HashIndex,
                                          int a_StopLevel)
        {
            uint l_InnerIndex;
            uint l_Sum;
            int l_Level = GetLevel(a_HashIndex, out l_InnerIndex, out l_Sum);

            HashTreeNode l_Node = new HashTreeNode(a_HashIndex, l_Level);

            bool l_Extra = false;
            uint l_ExpectedCount = 0;
            for (; l_Node.f_Level > a_StopLevel; l_Level--)
            {
                l_Sum += f_Counts[l_Level];

                if (false == l_Extra)
                {
                    if (0 == (l_InnerIndex & 1))
                    {
                        if ((f_Counts[l_Level] - 1) == l_InnerIndex)
                        {
                            // There is no right neighbour.

                            l_Extra = true;
                            l_ExpectedCount = (f_Counts[l_Level] / 2);
                        }
                        else
                        {
                            // Join right neighbour.

                            l_InnerIndex /= 2;
                            l_Node = l_Node.CreateParent(l_Sum + l_InnerIndex, l_Node.f_HashIndex + 1, true);
                        }
                    }
                    else
                    {
                        // Join left neighbour.

                        l_InnerIndex /= 2;
                        l_Node = l_Node.CreateParent(l_Sum + l_InnerIndex, l_Node.f_HashIndex - 1, false);
                    }
                }
                else
                {
                    if (l_ExpectedCount == f_Counts[l_Level])
                    {
                        l_ExpectedCount /= 2;
                    }
                    else
                    {
                        l_InnerIndex = (f_Counts[l_Level - 1] - 1);
                        // Create a forward node.
                        l_Node = new HashTreeNode(l_Sum + l_InnerIndex,
                                                  l_Level - 1,
                                                  new HashTreeNode(l_Sum - 2, l_Level),
                                                  new HashTreeNode(l_Sum - 1, l_Level, l_Node, l_Node));

                        l_Extra = false;
                    }
                }
            }

            return l_Node;
        }

        public bool GetLowLevelHashesOfChunk(byte a_Exponent,
                                            uint a_Chunk,
                                            out uint a_HashIndex,
                                            out uint a_Count)
        {
            uint l_LowLevHashes = f_Counts[p_MaxLevel];
            a_HashIndex = (a_Chunk << a_Exponent);

            if (l_LowLevHashes > a_HashIndex)
            {
                uint l_HashesPerChunk = (1u << a_Exponent);
                a_Count = Math.Min(l_LowLevHashes - a_HashIndex, l_HashesPerChunk);

                return true;
            }

            a_HashIndex = 0;
            a_Count = 0;

            return false;
        }

        public int GetUpperTreeBoundaries(byte a_Exponent)
        {
            uint l_HashIndex;
            uint l_Count;
            return GetUpperTreeBoundaries(a_Exponent, out l_HashIndex, out l_Count);
        }

        public int GetUpperTreeBoundaries(byte a_Exponent,
                                          out uint a_HashIndex,
                                          out uint a_Count)
        {
            // Returns the greatest level affected.

            int l_MaxLevel = p_MaxLevel;

            if (l_MaxLevel <= a_Exponent)
            {
                // This goes for all files with a length of up to
                // (2^a_Exponent * TAICHTree.c_BlockSize). Simply
                // describe the top hash.

                a_HashIndex = p_TopHashIndex;
                a_Count = 1;

                return 0;
            }

            int l_Lev = (l_MaxLevel - a_Exponent);

            a_Count = SumUpLevelCount(0, l_Lev);
            a_HashIndex = (f_Total - a_Count);

            return l_Lev;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TigerHashHive
    {
        // Don't use the struct default constructor.

        public const byte c_AllocGranularity = 16;

        public readonly bool f_ReadOnly;
        private readonly List<uint> f_Indices;
        private byte[] f_Buffer;

        private TigerHashHive(bool a_ReadOnly,
                              List<uint> a_Indices,
                              byte[] a_Buffer)
        {
            f_ReadOnly = a_ReadOnly;
            f_Indices = a_Indices;
            f_Buffer = a_Buffer;
        }

        public static void CreateExpandable(int a_InitCapacity,
                                            out TigerHashHive a_Dest)
        {
            // Set a_InitCapacity to zero if unsure.

            // Don't exaggerate.

            List<uint> l_Indices = new List<uint>(a_InitCapacity);
            byte[] l_Buffer = (0 == a_InitCapacity ? null : new byte[a_InitCapacity * Tiger.c_HashByteSize]);

            a_Dest = new TigerHashHive(false, l_Indices, l_Buffer);
        }

        public static void CreateReadOnly(uint[] a_Indices,
                                          byte[] a_Buffer,
                                          out TigerHashHive a_Dest)
        {
            CreateReadOnly(new List<uint>(a_Indices), a_Buffer, out a_Dest);
        }

        public static void CreateReadOnly(List<uint> a_Indices,
                                          byte[] a_Buffer,
                                          out TigerHashHive a_Dest)
        {
            // This instance takes ownership of a_Indices.

            a_Dest = new TigerHashHive(true, a_Indices, a_Buffer);
        }

        public static void CreateReadOnly(uint a_StartingHashIndex,
                                          ByteArray a_Src,
                                          Tiger a_Tiger,
                                          out TigerHashHive a_Dest)
        {
            int l_Count = ((a_Src.f_Len / (int)TAICHTree.c_BlockSize) +
                           (0 == (a_Src.f_Len % (int)TAICHTree.c_BlockSize) ? 0 : 1));

            List<uint> l_Ls = new List<uint>(l_Count);
            byte[] l_Bytes = new byte[l_Count * Tiger.c_HashByteSize];

            uint l_HashIndex = a_StartingHashIndex;
            int l_Pos = a_Src.f_Idx;
            for (int i = 0; i < l_Count; i++, l_HashIndex++, l_Pos += (int)TAICHTree.c_BlockSize)
            {
                int l_Len = Math.Min((int)TAICHTree.c_BlockSize, a_Src.f_Len - l_Pos);

                a_Tiger.Reset();
                a_Tiger.Update(a_Src.f_Arr, l_Pos, l_Len);
                a_Tiger.GetDigest(l_Bytes, i * Tiger.c_HashByteSize);

                l_Ls.Add(l_HashIndex);
            }

            a_Dest = new TigerHashHive(false, l_Ls, l_Bytes);
        }

        public bool IsValid()
        {
            // Returns true if this instance was initialized properly.

            return (null != f_Indices);
        }

        public int GetCount()
        {
            return f_Indices.Count;
        }

        private int GetListIndex(uint a_HashIndex)
        {
            for (int i = (f_Indices.Count - 1); 0 <= i; i--)
            {
                if (a_HashIndex == f_Indices[i])
                {
                    return i;
                }
            }

            return (-1);
        }

        public bool HasHash(uint a_HashIndex)
        {
            return (0 <= GetListIndex(a_HashIndex));
        }

        public bool Get(uint a_HashIndex,
                        byte[] a_Dest,
                        int a_DestIndex)
        {
            int l_ListIdx = GetListIndex(a_HashIndex);

            if (0 <= l_ListIdx)
            {
                Buffer.BlockCopy(f_Buffer, l_ListIdx * Tiger.c_HashByteSize, a_Dest, a_DestIndex, Tiger.c_HashByteSize);

                return true;
            }

            return false;
        }

        public int WriteAll(byte[] a_Dest,
                            int a_DestIndex)
        {
            // Returns the number of bytes written.

            Buffer.BlockCopy(f_Buffer, 0, a_Dest, a_DestIndex, f_Buffer.Length);

            return f_Buffer.Length;
        }

        public bool Verify(uint a_HashIndex,
                           Tiger a_Tiger)
        {
            int l_ListIdx = GetListIndex(a_HashIndex);

            if (0 <= l_ListIdx)
            {
                return a_Tiger.VerifyBE(f_Buffer, l_ListIdx * Tiger.c_HashByteSize);
            }

            return false;
        }

        public bool UpdateHashFunction(IHashFunction a_HashAlgo,
                                       uint a_HashIndex)
        {
            int l_ListIdx = GetListIndex(a_HashIndex);

            if (0 <= l_ListIdx)
            {
                a_HashAlgo.Update(f_Buffer, l_ListIdx * Tiger.c_HashByteSize, Tiger.c_HashByteSize);

                return true;
            }

            return false;
        }

        public void Add(uint a_HashIndex,
                        Tiger a_Tiger)
        {
            int l_ListIdx = PrepareAdd(a_HashIndex);

            a_Tiger.GetDigest(f_Buffer, l_ListIdx * Tiger.c_HashByteSize);
        }

        public void Add(uint a_HashIndex,
                        byte[] a_Src)
        {
            Add(a_HashIndex, a_Src, 0);
        }

        public void Add(uint a_HashIndex,
                        byte[] a_Src,
                        int a_SrcIdx)
        {
            int l_ListIdx = PrepareAdd(a_HashIndex);

            Buffer.BlockCopy(a_Src, a_SrcIdx, f_Buffer, l_ListIdx * Tiger.c_HashByteSize, Tiger.c_HashByteSize);
        }

        private int PrepareAdd(uint a_HashIndex)
        {
            // Returns the list index of the slot.

            int l_ListIdx = GetListIndex(a_HashIndex);

            if (0 > l_ListIdx)
            {
                l_ListIdx = f_Indices.Count;

                if (null == f_Buffer ||
                    ((l_ListIdx + 1) * Tiger.c_HashByteSize) > f_Buffer.Length)
                {
                    int l_ReqBufSize = ((l_ListIdx + c_AllocGranularity) * Tiger.c_HashByteSize);

                    Array.Resize(ref f_Buffer, l_ReqBufSize);
                }

                f_Indices.Add(a_HashIndex);
            }

            return l_ListIdx;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TAICHTreeAccessor
    {
        // Don't use the struct default constructor.

        public readonly int f_GreatestUpperLevel;
        private readonly uint f_HashBaseIndex;
        private readonly uint f_HashCount;
        private readonly uint f_SrcByteBaseOffset;

        // Only one of those three should be nonzero.
        private readonly byte[] f_SrcBuffer;
        private readonly FHandle f_SrcFile;
        private readonly HardMemory f_SrcHardMem;

        // Only valid if f_SrcHardMem is nonzero.
        private readonly HardMemory.Element f_SrcElement;

        private readonly byte[] f_TempBuf;

        private TAICHTreeAccessor(ref TAICHTreeLayout a_Layout,
                                  int a_GreatestUpperLevel)
        {
            f_GreatestUpperLevel = a_GreatestUpperLevel;
            f_HashBaseIndex = a_Layout.SumUpLevelCount(a_GreatestUpperLevel + 1, a_Layout.p_MaxLevel);
            f_HashCount = (a_Layout.f_Total - f_HashBaseIndex);
            f_SrcByteBaseOffset = 0;

            f_SrcBuffer = null;
            f_SrcFile = null;
            f_SrcHardMem = null;
            f_SrcElement = null;

            f_TempBuf = new byte[TAICHTree.c_PairSize];
        }

        public TAICHTreeAccessor(ref TAICHTreeLayout a_Layout,
                                 int a_GreatestUpperLevel,
                                 ref ByteArray a_Data) :
                            this(ref a_Layout, a_GreatestUpperLevel)
        {
            DebugAssertSize((uint)a_Data.f_Len);

            f_SrcBuffer = a_Data.f_Arr;
            f_SrcByteBaseOffset = (uint)a_Data.f_Idx;
        }

        public TAICHTreeAccessor(ref TAICHTreeLayout a_Layout,
                                 int a_GreatestUpperLevel,
                                 uint a_BufferThreshold,
                                 FHandle a_File,
                                 uint a_StartingOffset,
                                 ulong a_FileSize) :
                            this(ref a_Layout, a_GreatestUpperLevel)
        {
            a_FileSize -= a_StartingOffset;
            DebugAssertSize(a_FileSize);

            if (int.MaxValue >= a_BufferThreshold &&
                a_FileSize <= a_BufferThreshold)
            {
                ByteArray l_Data = new ByteArray((int)a_FileSize);

                if (true == a_File.Read(a_StartingOffset, ref l_Data))
                {
                    f_SrcBuffer = l_Data.f_Arr;
                    f_SrcByteBaseOffset = 0;

                    return;
                }
            }

            f_SrcFile = a_File;
            f_SrcByteBaseOffset = a_StartingOffset;
        }

        public TAICHTreeAccessor(ref TAICHTreeLayout a_Layout,
                                 int a_GreatestUpperLevel,
                                 HardMemory a_HardMem,
                                 HardMemory.Element a_Element,
                                 uint a_StartingOffset) :
                            this(ref a_Layout, a_GreatestUpperLevel)
        {
            DebugAssertSize(a_Element.f_MaxSize - a_StartingOffset);

            f_SrcHardMem = a_HardMem;
            f_SrcElement = a_Element;
            f_SrcByteBaseOffset = a_StartingOffset;
        }

        public bool IsValid()
        {
            return (null != f_TempBuf);
        }

        public void AcquireSharedIfDiskBased()
        {
            if (null != f_SrcHardMem)
            {
                f_SrcHardMem.p_Resource.AcquireShared();
            }
        }

        public void ReleaseSharedIfDiskBased()
        {
            if (null != f_SrcHardMem)
            {
                f_SrcHardMem.p_Resource.ReleaseShared();
            }
        }

        [Conditional("DEBUG")]
        public void DebugAssertSize(ulong a_AvailLength)
        {
            if ((f_HashCount * Tiger.c_HashByteSize) > a_AvailLength)
            {
                throw (new ApplicationException());
            }
        }

        public bool CopyCore(uint a_RawSrcPos,
                             ByteArray a_Dest)
        {
            // Returns false only if an io error occurs.

            uint l_SrcPos = (f_SrcByteBaseOffset + a_RawSrcPos);

            if (null != f_SrcBuffer)
            {
                Buffer.BlockCopy(f_SrcBuffer, (int)l_SrcPos, a_Dest.f_Arr, a_Dest.f_Idx, a_Dest.f_Len);

                return true;
            }
            else if (null != f_SrcFile)
            {
                if (true == f_SrcFile.Read(l_SrcPos, ref a_Dest))
                {
                    return true;
                }
            }
            else
            {
                // We need to have shared access to
                // f_SrcHardMem.p_Resource (if any).

                if (true == f_SrcHardMem.Read(f_SrcElement, l_SrcPos, ref a_Dest))
                {
                    return true;
                }
            }

            return false;
        }

        public TAICHVerifyResult VerifyTopHash(byte[] a_Buf)
        {
            uint l_TopHashIndex = (f_HashBaseIndex + f_HashCount - 1);

            return VerifyHash(l_TopHashIndex, a_Buf, 0);
        }

        public TAICHVerifyResult VerifyHash(uint a_HashIndex,
                                            byte[] a_Buf)
        {
            return VerifyHash(a_HashIndex, a_Buf, 0);
        }

        public TAICHVerifyResult VerifyHash(uint a_HashIndex,
                                            Tiger a_Tiger)
        {
            a_Tiger.GetDigest(f_TempBuf, Tiger.c_HashByteSize);

            return VerifyHash(a_HashIndex, f_TempBuf, Tiger.c_HashByteSize);
        }

        public TAICHVerifyResult VerifyHash(uint a_HashIndex,
                                            byte[] a_Buf,
                                            int a_BufIndex)
        {
            // Since one overload depends on it, make sure this routine
            // can work with a_Buf being f_TempBuf while a_BufIndex is
            // Tiger.c_HashByteSize.

            uint l_RawPos = ((a_HashIndex - f_HashBaseIndex) * Tiger.c_HashByteSize);
            if (false == CopyCore(l_RawPos, new ByteArray(f_TempBuf, 0, Tiger.c_HashByteSize)))
            {
                return TAICHVerifyResult.AccessFailed;
            }

            return VerifyEquality(f_TempBuf, 0, a_Buf, a_BufIndex);
        }

        public TAICHVerifyResult VerifyHash(uint a_RawPos1,
                                            uint a_RawPos2)
        {
            if (false == CopyCore(a_RawPos1, new ByteArray(f_TempBuf, 0, Tiger.c_HashByteSize)) ||
                false == CopyCore(a_RawPos2, new ByteArray(f_TempBuf, Tiger.c_HashByteSize, Tiger.c_HashByteSize)))
            {
                return TAICHVerifyResult.AccessFailed;
            }

            return VerifyEquality(f_TempBuf, 0, f_TempBuf, Tiger.c_HashByteSize);
        }

        public TAICHVerifyResult CalcAndVerifyHash(Tiger a_HashAlgo,
                                                   uint a_RawCalcPos,
                                                   uint a_RawVerifyPos)
        {
            // Calculates a hash from the two hashs at a_RawCalcPos and
            // verifies it against the hash at a_RawVerifyPos.

            if (false == CopyCore(a_RawCalcPos, new ByteArray(f_TempBuf, 0, TAICHTree.c_PairSize)))
            {
                return TAICHVerifyResult.AccessFailed;
            }

            a_HashAlgo.Reset();
            a_HashAlgo.Update(f_TempBuf, 0, TAICHTree.c_PairSize);
            a_HashAlgo.GetDigest(f_TempBuf, 0);

            if (false == CopyCore(a_RawVerifyPos, new ByteArray(f_TempBuf, Tiger.c_HashByteSize, Tiger.c_HashByteSize)))
            {
                return TAICHVerifyResult.AccessFailed;
            }

            return VerifyEquality(f_TempBuf, 0, f_TempBuf, Tiger.c_HashByteSize);
        }

        private TAICHVerifyResult VerifyEquality(byte[] a_A,
                                                 int a_AIdx,
                                                 byte[] a_B,
                                                 int a_BIdx)
        {
            if (false == Buffer2.ValueEquals(a_A, a_AIdx, a_B, a_BIdx, Tiger.c_HashByteSize))
            {
                return TAICHVerifyResult.HashMismatch;
            }

            return TAICHVerifyResult.Valid;
        }

        public byte[] CheckInputAndBuildHashArray(uint a_StartingIndex,
                                                  uint a_Count,
                                                  out bool a_MalformedInput)
        {
            if (f_HashBaseIndex > a_StartingIndex)
            {
                // The requested range does not belong to the tree
                // section we have.

                a_MalformedInput = true;

                return null;
            }

            uint l_EffStartingIndex = (a_StartingIndex - f_HashBaseIndex);

            if (null != f_SrcBuffer)
            {
                return BuildHashArray(l_EffStartingIndex, a_Count, f_HashCount, Tiger.c_HashByteSize, f_SrcBuffer, f_SrcByteBaseOffset, out a_MalformedInput);
            }

            if (null != f_SrcHardMem)
            {
                return BuildHashArray(l_EffStartingIndex, a_Count, f_HashCount, Tiger.c_HashByteSize, f_SrcHardMem, f_SrcElement, f_SrcByteBaseOffset, out a_MalformedInput);
            }

            // Files are not supported here.

            Util.DebugBreak();

            a_MalformedInput = false;

            return null;
        }

        public static byte[] BuildHashArray(uint a_StartingElement,
                                            uint a_ElementCount,
                                            uint a_MaxElements,
                                            int a_HashSize,
                                            byte[] a_SrcBuffer,
                                            uint a_SrcByteBaseOffset,
                                            out bool a_MalformedInput)
        {
            // The unit of the first three arguments is hash elements,
            // the other sizes are in bytes. The first two arguments
            // are treated as untrusted. Return can be null, in which
            // case a_MalformedInput may be set.

            if (false == Util.CheckRangeUnsigned(a_StartingElement, a_ElementCount, a_MaxElements))
            {
                a_MalformedInput = true;

                return null;
            }

            a_MalformedInput = false;

            return Util.CopyBytes(a_SrcBuffer, (int)a_SrcByteBaseOffset + ((int)a_StartingElement * a_HashSize), (int)a_ElementCount * a_HashSize);
        }

        public static byte[] BuildHashArray(uint a_StartingElement,
                                            uint a_ElementCount,
                                            uint a_MaxElements,
                                            int a_HashSize,
                                            HardMemory a_SrcHardMem,
                                            HardMemory.Element a_SrcElement,
                                            uint a_SrcByteBaseOffset,
                                            out bool a_MalformedInput)
        {
            // See above comment.

            if (false == Util.CheckRangeUnsigned(a_StartingElement, a_ElementCount, a_MaxElements))
            {
                a_MalformedInput = true;

                return null;
            }

            a_MalformedInput = false;

            // We need to have shared access to f_SrcHardMem.p_Resource
            // (if any).

            uint l_SrcPos = (a_SrcByteBaseOffset + (a_StartingElement * (uint)a_HashSize));
            ByteArray l_Data = new ByteArray((int)a_ElementCount * a_HashSize);
            if (true == a_SrcHardMem.Read(a_SrcElement, l_SrcPos, ref l_Data))
            {
                return l_Data.f_Arr;
            }

            return null;
        }
    }

    class HashTreeNode
    {
        // We can't call this TreeNode since that name conflicts with
        // a winforms name.

        public readonly uint f_HashIndex;
        public readonly int f_Level;

        // One or both can be null.
        private HashTreeNode f_FirstChild;
        private HashTreeNode f_SecondChild;

        private int f_Num;
        private byte[] f_Digest;

        public bool p_HasAnyChild
        {
            get
            {
                return (null != f_FirstChild ||
                        null != f_SecondChild);
            }
        }

        public HashTreeNode p_FirstChild
        {
            get { return f_FirstChild; }
        }

        public HashTreeNode p_SecondChild
        {
            get { return f_SecondChild; }
        }

        public bool p_IsForwardNode
        {
            get
            {
                return (f_FirstChild == f_SecondChild);
            }
        }

        public HashTreeNode(uint a_HashIndex,
                            int a_Level) :
            this(a_HashIndex, a_Level, null, null)
        {
        }

        public HashTreeNode(uint a_HashIndex,
                            int a_Level,
                            HashTreeNode a_FirstChild,
                            HashTreeNode a_SecondChild)
        {
            f_HashIndex = a_HashIndex;
            f_Level = a_Level;
            f_FirstChild = a_FirstChild;
            f_SecondChild = a_SecondChild;
        }

        public HashTreeNode CreateParent(uint a_ParentHashIndex,
                                         uint a_SiblingHashIndex,
                                         bool a_JoinRight)
        {
            HashTreeNode l_Sibling = new HashTreeNode(a_SiblingHashIndex, this.f_Level);

            return new HashTreeNode(a_ParentHashIndex,
                                    this.f_Level - 1,
                                    false == a_JoinRight ? l_Sibling : this,
                                    false == a_JoinRight ? this : l_Sibling);
        }

        public void SetFirstChild(HashTreeNode a_Child)
        {
            f_FirstChild = a_Child;
        }

        public void SetSecondChild(HashTreeNode a_Child)
        {
            f_SecondChild = a_Child;
        }

        public int GetNum()
        {
            return f_Num;
        }

        public void SetNum(int a_Num)
        {
            f_Num = a_Num;
        }

        public byte[] GetDigest()
        {
            return f_Digest;
        }

        public void SetDigest(byte[] a_Digest)
        {
            f_Digest = a_Digest;
        }

        public bool CalcSubHashes(IHashFunction a_HashAlgo)
        {
            // This routine overwrites the f_Num field (we use it to
            // indicate the life of a node).

            Stack<HashTreeNode> l_Stack = new Stack<HashTreeNode>();

            HashTreeNode l_Node = this;
            l_Node.f_Num = 2;

            for (int i = 0;;)
            {
                if (0 == l_Node.f_Num)
                {
                    if (null == l_Node.f_Digest)
                    {
                        if (null != l_Node.p_FirstChild &&
                            null != l_Node.p_SecondChild)
                        {
                            byte[] l_Digest = new byte[a_HashAlgo.p_HashByteSize];

                            a_HashAlgo.Reset();
                            a_HashAlgo.Update(l_Node.p_FirstChild.f_Digest);
                            a_HashAlgo.Update(l_Node.p_SecondChild.f_Digest);
                            a_HashAlgo.GetDigest(l_Digest, 0);

                            l_Node.f_Digest = l_Digest;
                        }
                        else
                        {
                            // There are essential hash values missing
                            // that we need in order to calculate the
                            // top hash.

                            return false;
                        }
                    }

                    if (0 == i)
                    {
                        break;
                    }

                    l_Node = l_Stack.Pop();
                    i--;

                    continue;
                }

                HashTreeNode l_Child = null;

                if (2 == l_Node.f_Num)
                {
                    l_Node.f_Num = 1;
                    l_Child = l_Node.p_FirstChild;
                }

                if (null == l_Child)
                {
                    l_Node.f_Num = 0;
                    l_Child = l_Node.p_SecondChild;
                }

                if (null != l_Child)
                {
                    l_Stack.Push(l_Node);
                    i++;

                    l_Child.f_Num = 2;
                    l_Node = l_Child;
                }
            }

            return true;
        }
    }
}

// End of file.
