/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\Content2.cs
 * PURPOSE: Handling of incomplete files
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using NB.Content.HashSets;
using NB.Content.Internal;
using NB.Content.Internal.IncomplInput;
using NB.Import;
using NB.Network;
using NB.Network.eD2k;
using Support;
using Support.Encodings;
using Support.HashFunctions;

namespace NB.Content
{
    class BitMap
    {
        public const uint c_eD2kGranularity = Content.HashSets.eD2kHashSet.c_PartSize;
        public const uint c_NativeGranularity = Content.Internal.IncompleteFile.c_ChunkSize;
        private const byte c_BitsPerByte = Util.c_BitsPerByte;
        private const byte c_BitsPerUInt = Util.c_BitsPerUInt;

        public readonly uint f_Granularity;
        public readonly uint f_BitCount;
        private readonly uint[] f_UIntArray;
        private uint f_SetBitCount;

        public uint p_SetBitCount
        {
            get { return f_SetBitCount; }
        }

        public bool p_AllClear
        {
            get { return (0 == f_SetBitCount); }
        }

        public bool p_AllSet
        {
            get { return (f_BitCount == f_SetBitCount); }
        }

        public BitMap(uint a_Granularity,
                      uint a_BitCount)
        {
            f_Granularity = a_Granularity;
            f_BitCount = a_BitCount;
            f_UIntArray = new uint[CalcUIntCount(f_BitCount)];
        }

        private BitMap(uint a_Granularity,
                       uint a_BitCount,
                       uint[] a_UIntArray,
                       uint a_SetBitCount)
        {
            f_Granularity = a_Granularity;
            f_BitCount = a_BitCount;
            f_UIntArray = a_UIntArray;
            f_SetBitCount = a_SetBitCount;
        }

        public bool HasSameFormat(BitMap a_Other)
        {
            return (this.f_Granularity == a_Other.f_Granularity &&
                    this.f_BitCount == a_Other.f_BitCount);
        }

        public void Write(BufferIO a_Buffer)
        {
            // May be used even if the bitmap is modified in parallel.

            a_Buffer.WriteUInt32LE(f_Granularity);
            a_Buffer.WriteUInt32LE(f_BitCount);
            a_Buffer.WriteUIntsLE(f_UIntArray);
        }

        public void WriteBodyToBuffer(BufferIO a_Buffer)
        {
            WriteBodyToBuffer(0, (int)CalcByteCount(), a_Buffer);
        }

        public void WriteBodyToBuffer(int a_ByteIndex,
                                      int a_ByteCount,
                                      BufferIO a_Buffer)
        {
            byte[] l_Temp = new byte[Math.Min(32 * 1024, a_ByteCount)];

            for (int i = a_ByteIndex, lim = (a_ByteIndex + a_ByteCount); i < lim;)
            {
                int l_Cur = Math.Min(lim - i, l_Temp.Length);
                Buffer.BlockCopy(f_UIntArray, i, l_Temp, 0, l_Cur);
                a_Buffer.WriteBytes(l_Temp, 0, l_Cur);

                i += l_Cur;
            }
        }

        public static BitMap FromBuffer(BufferIO a_Buffer,
                                        ulong a_FileSize)
        {
            // Return can be null on error.

            uint l_Granularity = a_Buffer.ReadUInt32LE();

            if (0 == l_Granularity)
            {
                return null;
            }

            uint l_BitCount = a_Buffer.ReadUInt32LE();

            if (CalcBitCount(l_Granularity, a_FileSize) != l_BitCount)
            {
                return null;
            }

            uint[] l_UInts = a_Buffer.ReadUIntsLE(CalcUIntCount(l_BitCount));
            ZeroSurplusBits(l_BitCount, l_UInts);

            return new BitMap(l_Granularity, l_BitCount, l_UInts, CountSetBits(l_UInts));
        }

        public BitMap Clone()
        {
            return new BitMap(this.f_Granularity, this.f_BitCount, Util.CloneArray(this.f_UIntArray), this.f_SetBitCount);
        }

        public BitMap CloneVolatile()
        {
            uint[] l_UIntArray = Util.CloneArray(this.f_UIntArray);

            return new BitMap(this.f_Granularity, this.f_BitCount, l_UIntArray, CountSetBits(l_UIntArray));
        }

        public void CopyVolatile(BitMap a_Dest)
        {
            Buffer.BlockCopy(this.f_UIntArray, 0, a_Dest.f_UIntArray, 0, this.f_UIntArray.Length * sizeof(uint));
            a_Dest.f_SetBitCount = CountSetBits(a_Dest.f_UIntArray);
        }

        public BitMap RecreateEmpty()
        {
            BitMap l_Empty = new BitMap(this.f_Granularity, this.f_BitCount);

            return l_Empty;
        }

        public static BitMap CreateFromFileSize(uint a_Granularity,
                                                ulong a_FileSize)
        {
            ulong l_BitCount = CalcBitCount(a_Granularity, a_FileSize);

            return new BitMap(a_Granularity, checked((uint)l_BitCount));
        }

        public uint CalcByteCount()
        {
            return CalcByteCount(f_BitCount);
        }

        public int GetUIntCount()
        {
            return f_UIntArray.Length;
        }

        public bool CopyToBufferWithCheck(uint a_ByteIndex,
                                          uint a_ByteCount,
                                          ref byte[] a_Dest,
                                          int a_DestIndex)
        {
            if (false == Util.CheckRangeUnsigned(a_ByteIndex, a_ByteCount, CalcByteCount()))
            {
                return false;
            }

            CopyToBuffer(a_ByteIndex, a_ByteCount, ref a_Dest, a_DestIndex);

            return true;
        }

        public void CopyToBuffer(uint a_ByteIndex,
                                 uint a_ByteCount,
                                 ref byte[] a_Dest,
                                 int a_DestIndex)
        {
            // If a_Dest is null, a buffer will be created.

            if (null == a_Dest)
            {
                a_Dest = new byte[a_DestIndex + (int)a_ByteCount];
            }

            Buffer.BlockCopy(f_UIntArray, (int)a_ByteIndex, a_Dest, a_DestIndex, (int)a_ByteCount);
        }

        public void SetAll(bool a_Value)
        {
            if (false == a_Value)
            {
                if (false == this.p_AllClear)
                {
                    Buffer2.BlockZero(f_UIntArray, 0, f_UIntArray.Length);
                    f_SetBitCount = 0;
                }
            }
            else
            {
                if (false == this.p_AllSet)
                {
                    Buffer2.SetUIntArray(f_UIntArray, 0, f_UIntArray.Length, uint.MaxValue);
                    ZeroSurplusBits(f_BitCount, f_UIntArray);
                    f_SetBitCount = f_BitCount;
                }
            }
        }

        public void EnumerateSetRuns(ref EnumerateRunsContext a_Cntxt)
        {
            // A subsequent call won't produce results after
            // a_Cntxt.f_RunBitCount was zero on return.

            // Reset return vars.
            a_Cntxt.f_RunBitIndex = 0;
            a_Cntxt.f_RunBitCount = 0;

            if (0 == f_BitCount)
            {
                return;
            }

            uint l_BitsLeft = (a_Cntxt.f_BitIndex % c_BitsPerUInt);
            uint l_Mask;

            if (0 == l_BitsLeft)
            {
                l_BitsLeft = c_BitsPerUInt;
                l_Mask = 1;
            }
            else
            {
                l_Mask = (1u << (int)l_BitsLeft);
                l_BitsLeft = (c_BitsPerUInt - l_BitsLeft);
            }

            for (uint i = (a_Cntxt.f_BitIndex / c_BitsPerUInt), lim = (a_Cntxt.f_FinalBitIndex / c_BitsPerUInt); i <= lim; i++, l_BitsLeft = c_BitsPerUInt, l_Mask = 1)
            {
                uint l_UInt;

                if (i < lim)
                {
                    l_UInt = f_UIntArray[i];

                    if (0 == l_UInt)
                    {
                        if (0 != a_Cntxt.f_RunBitCount)
                        {
                            // We are within a run, so exit.

                            return;
                        }

                        a_Cntxt.f_BitIndex += l_BitsLeft;

                        continue;
                    }
                }
                else
                {
                    // The last uint is reached. l_BitsLeft might be too
                    // high and needs to be recalculated.

                    if (a_Cntxt.f_FinalBitIndex <= a_Cntxt.f_BitIndex)
                    {
                        return;
                    }

                    l_UInt = f_UIntArray[i];
                    l_BitsLeft = (a_Cntxt.f_FinalBitIndex - a_Cntxt.f_BitIndex);
                }

                for (; 0 != l_BitsLeft; l_BitsLeft--, l_Mask <<= 1)
                {
                    bool l_Set = (0 != (l_UInt & l_Mask));
                    bool l_WithinRun = (0 != a_Cntxt.f_RunBitCount);

                    a_Cntxt.f_BitIndex++;

                    if (false == l_Set)
                    {
                        if (true == l_WithinRun)
                        {
                            return;
                        }
                    }
                    else
                    {
                        if (false == l_WithinRun)
                        {
                            a_Cntxt.f_RunBitIndex = (a_Cntxt.f_BitIndex - 1);
                        }

                        a_Cntxt.f_RunBitCount++;

                        if (a_Cntxt.f_MaxBitsPerCall <= a_Cntxt.f_RunBitCount)
                        {
                            return;
                        }
                    }
                }
            }
        }

        public bool CollectSetBitIndices(ref EnumerateRunsContext a_Cntxt,
                                         ref FlexArray<ChunkAvail> a_Indices)
        {
            // The collected indices will be placed into a_Indices.
            // The count of a_Indices can get higher than
            // a_Cntxt.f_MaxBitsPerCall. Returns false if another call
            // won't produce results.

            while (true)
            {
                EnumerateSetRuns(ref a_Cntxt);

                if (0 == a_Cntxt.f_RunBitCount)
                {
                    // End reached.

                    return false;
                }

                int l_Count = a_Indices.p_Count;
                a_Indices.ReserveRange(l_Count, (int)a_Cntxt.f_RunBitCount);

                ChunkAvail[] l_Arr = a_Indices.p_Arr;
                for (uint i = 0; i < a_Cntxt.f_RunBitCount; i++)
                {
                    l_Arr[l_Count + i] = new ChunkAvail(a_Cntxt.f_RunBitIndex + i);
                }

                if (a_Cntxt.f_MaxBitsPerCall <= (uint)a_Indices.p_Count)
                {
                    // a_Indices's count can be (a_Cntxt.f_MaxBitsPerCall +
                    // c_BitsPerUInt - 1) at most.

                    return true;
                }
            }
        }

        public bool IsRunSet(uint a_Index,
                             uint a_Length)
        {
            return CheckRun(a_Index, a_Length, true);
        }

        public bool IsRunClear(uint a_Index,
                               uint a_Length)
        {
            return CheckRun(a_Index, a_Length, false);
        }

        private bool CheckRun(uint a_Index,
                              uint a_Length,
                              bool a_CheckForSet)
        {
            // Returns true for zero-length tests.

            uint l_Mask;
            int l_UIntIdx = GetUIntIndexAndMask(a_Index, out l_Mask);

            for (uint i = a_Index, lim = (a_Index + a_Length); i < lim; i++)
            {
                bool l_Value = (0 != (f_UIntArray[l_UIntIdx] & l_Mask));

                if (l_Value != a_CheckForSet)
                {
                    return false;
                }

                l_Mask <<= 1;

                if (0 == l_Mask)
                {
                    l_Mask = 1;
                    l_UIntIdx++;
                }
            }

            return true;
        }

        public bool Get(uint a_Index)
        {
            uint l_Mask;
            int l_UIntIndex = GetUIntIndexAndMask(a_Index, out l_Mask);

            return (0 != (f_UIntArray[l_UIntIndex] & l_Mask));
        }

        public byte GetByte(uint a_BitIndex,
                            out uint a_ByteIndex)
        {
            a_ByteIndex = (a_BitIndex / c_BitsPerByte);

            return Buffer.GetByte(f_UIntArray, (int)a_ByteIndex);
        }

        public bool Set(uint a_Index,
                        bool a_Value)
        {
            // Returns the previous value of the manipulated bit.

            uint l_Mask;
            int l_UIntIndex = GetUIntIndexAndMask(a_Index, out l_Mask);

            bool l_OldValue = (0 != (f_UIntArray[l_UIntIndex] & l_Mask));

            if (l_OldValue != a_Value)
            {
                if (false == a_Value)
                {
                    f_UIntArray[l_UIntIndex] &= ~l_Mask;
                    f_SetBitCount--;
                }
                else
                {
                    f_UIntArray[l_UIntIndex] |= l_Mask;
                    f_SetBitCount++;
                }
            }

            return l_OldValue;
        }

        public uint FilePosToIndex(ulong a_FilePos,
                                   ulong a_FinalFilePos,
                                   out uint a_Length)
        {
            return FilePosToIndex(false, a_FilePos, a_FinalFilePos, out a_Length);
        }

        private uint FilePosToIndex(bool a_ClipEnding,
                                    ulong a_FilePos,
                                    ulong a_FinalFilePos,
                                    out uint a_Length)
        {
            // The returned value and a_Length are in bit units (not
            // uint elements).

            uint l_FinalIndex = ((uint)(a_FinalFilePos / f_Granularity) +
                                 (0 == (a_FinalFilePos % f_Granularity) ? 0u : 1u));

            if (false == a_ClipEnding)
            {
            }
            else
            {
                // a_ClipEnding should be used if the caller isn't
                // sure that a_FinalFilePos is within the bounds of
                // (f_BitCount * f_Granularity).

                l_FinalIndex = Math.Min(f_BitCount, l_FinalIndex);
            }

            uint l_Index = (uint)(a_FilePos / f_Granularity);

            a_Length = (l_FinalIndex - l_Index);

            return l_Index;
        }

        public bool ToPieceBounds(uint a_ChunkIndex,
                                  uint a_ChunkSize,
                                  out PieceBounds a_Bounds)
        {
            a_Bounds = new PieceBounds();

            if (IncompleteFile.c_ChunkSize == f_Granularity)
            {
                if (true == Get(a_ChunkIndex))
                {
                    ChunkRange l_FullRange = new ChunkRange(0, a_ChunkSize);

                    return a_Bounds.Add(ref l_FullRange);
                }

                return true;
            }

            ulong l_Pos = (a_ChunkIndex * c_NativeGranularity);
            ulong l_FinalPos = (l_Pos + a_ChunkSize);
            uint l_BitLength;
            uint l_BitIndex = FilePosToIndex(l_Pos, l_FinalPos, out l_BitLength);

            if (0 == l_BitLength)
            {
                return true;
            }

            if (true == IsRunSet(l_BitIndex, l_BitLength))
            {
                ChunkRange l_FullRange = new ChunkRange(0, a_ChunkSize);

                return a_Bounds.Add(ref l_FullRange);
            }

            if (1 == l_BitLength)
            {
                return true;
            }

            // Set up the correct size for the first item.
            uint l_Rem = (uint)(l_Pos % f_Granularity);

            ChunkRange l_Range = new ChunkRange(0, f_Granularity - l_Rem);
            bool l_InitialSizeAdjustment = false;

            for (uint i = l_BitIndex, lim = (l_BitIndex + l_BitLength), l_LastIdx = (lim - 1);;)
            {
                if (true == Get(i) &&
                    false == a_Bounds.Add(ref l_Range))
                {
                    return false;
                }

                if (++i == lim)
                {
                    break;
                }

                l_Range.f_Offset += l_Range.f_Size;

                if (i == l_LastIdx)
                {
                    // We are about to enter the last round, so make sure
                    // that the size for the last item is within bounds.

                    l_Range.f_Size = (a_ChunkSize - l_Range.f_Offset);
                }
                else if (false == l_InitialSizeAdjustment)
                {
                    l_Range.f_Size = (uint)f_Granularity;
                    l_InitialSizeAdjustment = true;
                }
            }

            return true;
        }

        public void ReadBodyFromBuffer(uint a_ByteIndex,
                                       uint a_ByteCount,
                                       BufferIO a_BufferIO)
        {
            // The indices form a range of the bitmap, not of the
            // input buffer.

            ByteArray l_Data;
            a_BufferIO.ExposeRange(a_BufferIO.p_Position, a_ByteCount, out l_Data);
            FromBinary(ref l_Data, (int)a_ByteIndex);
        }

        public void FromBinary(ref ByteArray a_Src,
                               int a_DestIndex)
        {
            // a_DestIndex is a byte offset.

            int l_FinalByteOffset = (a_DestIndex + a_Src.f_Len);

            int l_UIntIndex = (a_DestIndex / sizeof(uint));
            int l_UIntFinal = ((l_FinalByteOffset / sizeof(uint)) +
                               (0 == (l_FinalByteOffset % sizeof(uint)) ? 0 : 1));
            uint l_OldSum;

            if (0 == l_UIntIndex &&
                f_UIntArray.Length == l_UIntFinal)
            {
                l_OldSum = f_SetBitCount;
            }
            else
            {
                if ((f_UIntArray.Length / 2) >= (l_UIntFinal - l_UIntIndex))
                {
                    // Positive calculation.

                    l_OldSum = CountSetBits(f_UIntArray, l_UIntIndex, l_UIntFinal);
                }
                else
                {
                    // Negative calculation (it's faster in this case).

                    l_OldSum = CountSetBits(f_UIntArray, 0, l_UIntIndex);
                    l_OldSum += CountSetBits(f_UIntArray, l_UIntFinal, f_UIntArray.Length);
                    l_OldSum = (f_SetBitCount - l_OldSum);
                }
            }

            Buffer.BlockCopy(a_Src.f_Arr, a_Src.f_Idx, f_UIntArray, a_DestIndex, a_Src.f_Len);
            ZeroSurplusBits(f_BitCount, f_UIntArray);

            uint l_NewSum = CountSetBits(f_UIntArray, l_UIntIndex, l_UIntFinal);

            f_SetBitCount = (f_SetBitCount + l_NewSum - l_OldSum);
        }

        public bool Has8MiBBitmapeD2kRelevance(ulong a_FileSize)
        {
            return Has8MiBBitmapeD2kRelevanceCore(a_FileSize);
        }

        private bool Has8MiBBitmapeD2kRelevanceCore(ulong a_FileSize)
        {
            if (0 == f_SetBitCount)
            {
                return false;
            }

            if (f_BitCount == f_SetBitCount)
            {
                return true;
            }

            EnumerateRunsContext l_EnumCntxt = new EnumerateRunsContext(3, 0, f_BitCount);
            uint l_LastIdx = (f_BitCount - 1);

            while (true)
            {
                EnumerateSetRuns(ref l_EnumCntxt);

                if (0 == l_EnumCntxt.f_RunBitCount)
                {
                    // No more runs.

                    return false;
                }

                if (1 == l_EnumCntxt.f_RunBitCount)
                {
                    if (l_LastIdx > l_EnumCntxt.f_RunBitIndex)
                    {
                        // One chunk is never enough, except maybe if it's
                        // the last chunk.

                        continue;
                    }
                }
                else if (3 <= l_EnumCntxt.f_RunBitCount)
                {
                    // Three adjacent chunks always form at least one
                    // part.

                    return true;
                }

                ulong l_FilePos = ((ulong)l_EnumCntxt.f_RunBitIndex * IncompleteFile.c_ChunkSize);
                ulong l_FinalFilePos = ((ulong)(l_EnumCntxt.f_RunBitIndex + l_EnumCntxt.f_RunBitCount) * IncompleteFile.c_ChunkSize);
                ulong l_eD2kPos;
                ulong l_eD2kFinalPos;

                if (true == IncompleteFile.CalceD2kPartBoundsNarrow(a_FileSize, l_FilePos, l_FinalFilePos, out l_eD2kPos, out l_eD2kFinalPos))
                {
                    return true;
                }
            }
        }

        private void TakeOverDirectly(BitMap a_Src,
                                      uint a_SrcIdx,
                                      uint a_SrcLen)
        {
            if (0 == a_SrcIdx &&
                a_Src.f_BitCount == a_SrcLen)
            {
                Buffer.BlockCopy(a_Src.f_UIntArray, 0, this.f_UIntArray, 0, this.f_UIntArray.Length * sizeof(uint));
                this.f_SetBitCount = a_Src.f_SetBitCount;
            }
            else
            {
                for (uint i = a_SrcIdx, lim = (a_SrcIdx + a_SrcLen); i < lim; i++)
                {
                    // In this case the set bit count will be updated
                    // automatically.
                    this.Set(i, a_Src.Get(i));
                }
            }
        }

        public void TakeOver(BitMap a_Src,
                             TakeOverMode a_Mode)
        {
            TakeOver(a_Src, 0, a_Src.f_BitCount, a_Mode);
        }

        public void TakeOver(BitMap a_Src,
                             uint a_SrcIdx,
                             uint a_SrcLen,
                             TakeOverMode a_Mode)
        {
            if (0 == a_SrcLen)
            {
                return;
            }

            if (this.f_Granularity == a_Src.f_Granularity)
            {
                this.TakeOverDirectly(a_Src, a_SrcIdx, a_SrcLen);

                return;
            }

            bool l_RoundDown = (TakeOverMode.RoundDown == a_Mode);

            ulong l_SrcPos = ((ulong)a_SrcIdx * a_Src.f_Granularity);
            ulong l_SrcFinalPos = (l_SrcPos + ((ulong)a_SrcLen * a_Src.f_Granularity));
            uint l_DestLength;
            uint l_DestFirst = this.FilePosToIndex(true, l_SrcPos, l_SrcFinalPos, out l_DestLength);

            if (this.f_BitCount <= l_DestFirst)
            {
                return;
            }

            ulong l_DestPos = ((ulong)l_DestFirst * this.f_Granularity);

            for (uint i = l_DestFirst, lim = (l_DestFirst + l_DestLength); i < lim; i++)
            {
                ulong l_DestFinalPos = (l_DestPos + this.f_Granularity);
                uint l_SrcLength;
                uint l_SrcFirst = a_Src.FilePosToIndex(true, l_DestPos, l_DestFinalPos, out l_SrcLength);

                bool l_Value = a_Src.CheckRun(l_SrcFirst, l_SrcLength, l_RoundDown);

                if (false == l_RoundDown)
                {
                    // We want to round up, so translate 'all bits are
                    // clear' into 'some bits are set'.
                    l_Value = !l_Value;
                }

                this.Set(i, l_Value);

                l_DestPos = l_DestFinalPos;
            }
        }

        private bool TryCountRelevantBits(BitMap a_Completed,
                                          out uint a_Count)
        {
            // See BuildRelevanceMap().

            a_Count = 0;

            if (true == this.p_AllClear)
            {
                return true;
            }

            if (true == a_Completed.p_AllClear)
            {
                a_Count = this.f_SetBitCount;

                return true;
            }

            if (true == this.p_AllSet &&
                true == a_Completed.p_AllSet)
            {
                return true;
            }

            // We cannot say anything without looking at the arrays.

            return false;
        }

        public uint CountRelevantBits(BitMap a_Completed)
        {
            // Same as BuildRelevanceMap(), but there are no changes
            // made and only a count is returned.

            uint l_Count;
            if (true == TryCountRelevantBits(a_Completed, out l_Count))
            {
                return l_Count;
            }

            l_Count = 0;

            for (int i = 0; i < this.f_UIntArray.Length; i++)
            {
                uint l_Res = (this.f_UIntArray[i] & (~a_Completed.f_UIntArray[i]));

                l_Count += CountSetBits(l_Res);
            }

            return l_Count;
        }

        public void BuildRelevanceMap(BitMap a_Completed)
        {
            // On exit, only those bits of this instance are still set,
            // that were set on entry and are not set in a_Completed.

            uint l_Count;

            if (true == TryCountRelevantBits(a_Completed, out l_Count))
            {
                if (0 == l_Count)
                {
                    // No relevance at all.
                    this.SetAll(false);
                }
                else if (this.f_BitCount == l_Count)
                {
                    // Total relevance.
                    this.SetAll(true);
                }
                else
                {
                    // This instace can remain as it is.
                }

                return;
            }

            uint[] l_ComplArr = a_Completed.f_UIntArray;

            for (int i = 0, lim = f_UIntArray.Length; i < lim; i++)
            {
                // Everything that isn't completed and is available is
                // relevant.
                this.f_UIntArray[i] &= ~l_ComplArr[i];
            }

            ZeroSurplusBits(this.f_BitCount, this.f_UIntArray);
            this.f_SetBitCount = CountSetBits(this.f_UIntArray);
        }

        public static void UpdateAvailabilityArray(NetworkAvailability a_Avail,
                                                   BitMap a_Old,
                                                   BitMap a_New)
        {
            // a_Old and a_New may be null, but only one of both
            // may be at most.

            if (null == a_Old)
            {
                if (true == a_New.p_AllClear)
                {
                    a_Avail.IncOne(false);

                    return;
                }

                if (true == a_New.p_AllSet)
                {
                    a_Avail.IncOne(true);

                    return;
                }
            }
            else
            {
                if (null == a_New)
                {
                    if (true == a_Old.p_AllClear)
                    {
                        a_Avail.DecOne(false);

                        return;
                    }

                    if (true == a_Old.p_AllSet)
                    {
                        a_Avail.DecOne(true);

                        return;
                    }
                }
                else
                {
                    if ((true == a_Old.p_AllClear &&
                         true == a_New.p_AllClear) ||
                        (true == a_Old.p_AllSet &&
                         true == a_New.p_AllSet))
                    {
                        // Nothing to do.

                        return;
                    }
                }
            }

            // We need to do things in depth.

            uint l_NonzeroItems = a_Avail.p_NonzeroItems;

            int l_Mask = 1;
            int l_UIntIdx = 0;
            for (int i = 0; a_Avail.f_Arr.Length > i; i++)
            {
                bool l_OldVal = (null == a_Old ? false : (0 != (a_Old.f_UIntArray[l_UIntIdx] & l_Mask)));
                bool l_NewVal = (null == a_New ? false : (0 != (a_New.f_UIntArray[l_UIntIdx] & l_Mask)));

                if (l_OldVal != l_NewVal)
                {
                    if (false == l_NewVal)
                    {
                        a_Avail.f_Arr[i]--;

                        if (0 == a_Avail.f_Arr[i])
                        {
                            l_NonzeroItems--;
                        }
                    }
                    else
                    {
                        if (0 == a_Avail.f_Arr[i])
                        {
                            l_NonzeroItems++;
                        }

                        a_Avail.f_Arr[i]++;
                    }
                }

                l_Mask <<= 1;

                if (0 == l_Mask)
                {
                    l_Mask = 1;
                    l_UIntIdx++;
                }
            }

            uint l_Contributors = a_Avail.p_Contributors;
            if (null != a_Old) l_Contributors--;
            if (null != a_New) l_Contributors++;

            a_Avail.Update(l_NonzeroItems, l_Contributors);

            return;
        }

        private static int GetUIntIndexAndMask(uint a_BitIndex,
                                               out uint a_Mask)
        {
            int l_UIntIndex = (int)(a_BitIndex / c_BitsPerUInt);
            a_Mask = (1u << (int)(a_BitIndex % c_BitsPerUInt));

            return l_UIntIndex;
        }

        public static ulong CalcBitCount(uint a_Granularity,
                                         ulong a_FileSize)
        {
            return ((a_FileSize / a_Granularity) +
                    (0 == (a_FileSize % a_Granularity) ? 0u : 1u));
        }

        public static uint CalcByteCount(uint a_BitCount)
        {
            return ((a_BitCount / c_BitsPerByte) + (0 == (a_BitCount % c_BitsPerByte) ? 0u : 1u));
        }

        private static uint CalcUIntCount(uint a_BitCount)
        {
            return ((a_BitCount / c_BitsPerUInt) +
                    (0 == (a_BitCount % c_BitsPerUInt) ? 0u : 1u));
        }

        private static uint CountSetBits(uint[] a_UIntArray)
        {
            return CountSetBits(a_UIntArray, 0, a_UIntArray.Length);
        }

        private static uint CountSetBits(uint[] a_UIntArray,
                                         int a_Index,
                                         int a_Final)
        {
            uint l_Count = 0;

            for (int i = a_Index; i < a_Final; i++)
            {
                l_Count += CountSetBits(a_UIntArray[i]);
            }

            return l_Count;
        }

        public static uint CountSetBits(uint a_Num)
        {
            if (0 == a_Num)
            {
                return 0;
            }

            if (uint.MaxValue == a_Num)
            {
                return c_BitsPerUInt;
            }

            uint l_ByteMask = byte.MaxValue;
            uint l_BitMask = 1;
            uint l_Count = 0;

            for (int i = sizeof(uint); 0 < i; i--, l_ByteMask <<= c_BitsPerByte)
            {
                uint l_Res = (l_ByteMask & a_Num);

                if (0 == l_Res)
                {
                    l_BitMask <<= c_BitsPerByte;

                    continue;
                }

                if (l_ByteMask == l_Res)
                {
                    l_Count += c_BitsPerByte;
                    l_BitMask <<= c_BitsPerByte;

                    continue;
                }

                for (int k = c_BitsPerByte; 0 < k; k--)
                {
                    if (0 != (l_BitMask & a_Num))
                    {
                        l_Count++;
                    }

                    l_BitMask <<= 1;
                }
            }

            return l_Count;
        }

        private static void ZeroSurplusBits(uint a_BitCount,
                                            uint[] a_UIntArray)
        {
            uint l_Rem = (a_BitCount % Util.c_BitsPerUInt);

            if (0 != l_Rem)
            {
                a_UIntArray[a_UIntArray.Length - 1] &= ((1u << (int)l_Rem) - 1);
            }
        }

        public enum TakeOverMode : byte
        {
            // If a bit it partially affected and partially not
            // affected...
            // - Set it.
            RoundUp = 0,
            // - Clear it.
            RoundDown = 1,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct EnumerateRunsContext
        {
            // Don't use the struct default constructor.

            public readonly uint f_MaxBitsPerCall;
            public uint f_BitIndex;
            public readonly uint f_FinalBitIndex;

            // Result of EnumerateSetRuns().
            public uint f_RunBitIndex;
            public uint f_RunBitCount;

            public EnumerateRunsContext(uint a_MaxBitsPerCall,
                                        uint a_BitIndex,
                                        uint a_FinalBitIndex)
            {
                f_MaxBitsPerCall = a_MaxBitsPerCall;
                f_BitIndex = a_BitIndex;
                f_FinalBitIndex = a_FinalBitIndex;

                f_RunBitIndex = 0;
                f_RunBitCount = 0;
            }
        }
    }

    class NetworkAvailability
    {
        public readonly byte[] f_Arr;
        private uint f_NonzeroItems;
        private uint f_Contributors;

        public uint p_NonzeroItems
        {
            get { return f_NonzeroItems; }
        }

        public uint p_Contributors
        {
            get { return f_Contributors; }
        }

        public NetworkAvailability(uint a_Items)
        {
            f_Arr = new byte[a_Items];
            f_NonzeroItems = 0;
            f_Contributors = 0;
        }

        public bool IsEmpty()
        {
            return (0 == f_NonzeroItems &&
                    0 == f_Contributors);
        }

        public uint CalcChunkAvailability(uint a_ChunkIndex)
        {
            ulong l_Pos = ((ulong)a_ChunkIndex * IncompleteFile.c_ChunkSize);
            ulong l_FinalPos = (l_Pos + IncompleteFile.c_ChunkSize);

            const uint l_ArrGranularity = eD2kHashSet.c_PartSize;
            int l_Idx = (int)(l_Pos / l_ArrGranularity);
            int l_FinalIdx = ((int)(l_FinalPos / l_ArrGranularity) +
                              (0 == (l_FinalPos % l_ArrGranularity) ? 0 : 1));
            l_FinalIdx = Math.Min(l_FinalIdx, f_Arr.Length);
            uint l_Sum = 0;

            for (int i = l_Idx; i < l_FinalIdx; i++)
            {
                l_Sum += f_Arr[i];
            }

            switch (l_FinalIdx - l_Idx)
            {
                case 2:
                    l_Sum >>= 2;
                    break;
            }

            return l_Sum;
        }

        public bool CheckFruitfulness(BitMap a_CompletionMap)
        {
            // Returns a nonzero value if all zero bits in
            // a_CompletionMap have a corresponding nonzero byte in
            // this instance. Returns zero if there are no zero
            // bits in a_CompletionMap.

            if (0 == f_NonzeroItems ||
                true == a_CompletionMap.p_AllSet)
            {
                return false;
            }

            uint l_ClearBytesLeft = ((uint)f_Arr.Length - f_NonzeroItems);

            if (0 != l_ClearBytesLeft)
            {
                if (true == a_CompletionMap.p_AllClear)
                {
                    return false;
                }

                uint l_BitIdx = 0;

                foreach (byte l_Avail in f_Arr)
                {
                    if (0 == l_Avail)
                    {
                        if (false == a_CompletionMap.Get(l_BitIdx))
                        {
                            return false;
                        }

                        if (0 == --l_ClearBytesLeft)
                        {
                            // End reached.

                            break;
                        }
                    }

                    l_BitIdx++;
                }
            }

            return true;
        }

        public void Update(uint a_NonzeroItems,
                           uint a_Contributors)
        {
            f_NonzeroItems = a_NonzeroItems;
            f_Contributors = a_Contributors;
        }

        public void IncOne(bool a_AllSet)
        {
            if (true == a_AllSet)
            {
                for (int i = (f_Arr.Length - 1); 0 <= i; i--)
                {
                    if (0 == f_Arr[i])
                    {
                        f_NonzeroItems++;
                    }

                    f_Arr[i]++;
                }
            }

            f_Contributors++;
        }

        public void DecOne(bool a_AllSet)
        {
            if (true == a_AllSet)
            {
                for (int i = (f_Arr.Length - 1); 0 <= i; i--)
                {
                    f_Arr[i]--;

                    if (0 == f_Arr[i])
                    {
                        f_NonzeroItems--;
                    }
                }
            }

            f_Contributors--;
        }
    }

    class IncomplFileProgress
    {
        public readonly ushort f_MapRevision;
        public readonly ushort f_HotIncomplRevision;
        // The actual bitfield of f_DoneMap is volatile, but
        // BitMap.Get() can be safely used.
        public readonly BitMap f_DoneMap;
        // f_HotIncomplChunks can be null.
        public readonly Dictionary<uint, ElmKey> f_HotIncomplChunks;
        public readonly ulong f_CompletedBytes;
        public readonly ulong f_CompletedBytesInclUnverified;

        public IncomplFileProgress(ushort a_MapRevision,
                                   ushort a_HotIncomplRevision,
                                   BitMap a_DoneMap,
                                   Dictionary<uint, ElmKey> a_HotIncomplChunks,
                                   ulong a_CompletedBytes,
                                   ulong a_CompletedBytesInclUnverified)
        {
            // a_HotIncomplChunks may be null.

            f_MapRevision = a_MapRevision;
            f_HotIncomplRevision = a_HotIncomplRevision;
            f_DoneMap = a_DoneMap;
            f_HotIncomplChunks = a_HotIncomplChunks;
            f_CompletedBytes = a_CompletedBytes;
            f_CompletedBytesInclUnverified = a_CompletedBytesInclUnverified;
        }

        public IncomplFileProgress RecreateWithClearedHotChunks(bool a_IncHotIncomplRevision)
        {
            ushort l_HotIncomplRevision = f_HotIncomplRevision;

            if (true == a_IncHotIncomplRevision)
            {
                l_HotIncomplRevision++;
            }

            return new IncomplFileProgress(f_MapRevision, l_HotIncomplRevision, f_DoneMap, null, f_CompletedBytes, f_CompletedBytesInclUnverified);
        }

        public bool CanBePreviewed()
        {
            return f_DoneMap.Get(0);
        }

        public void Write(BufferIO a_Buffer)
        {
            // We don't save f_HotIncomplChunks.

            f_DoneMap.Write(a_Buffer);
            a_Buffer.WriteUInt64LE(f_CompletedBytes);
            a_Buffer.WriteUInt64LE(f_CompletedBytesInclUnverified);
        }

        public static IncomplFileProgress FromBuffer(BufferIO a_Buffer,
                                                     ulong a_FileSize)
        {
            // Return will be null, on error.

            BitMap l_DoneMap = BitMap.FromBuffer(a_Buffer, a_FileSize);

            if (null == l_DoneMap ||
                IncompleteFile.c_ChunkSize != l_DoneMap.f_Granularity)
            {
                return null;
            }

            ulong l_CompletedBytes = a_Buffer.ReadUInt64LE();
            ulong l_CompletedBytesInclUnverified = a_Buffer.ReadUInt64LE();

            if (l_CompletedBytesInclUnverified > a_FileSize)
            {
                // Repair.

                Util.DebugBreak();

                l_CompletedBytesInclUnverified = a_FileSize;
            }

            if (l_CompletedBytes > l_CompletedBytesInclUnverified)
            {
                // Repair.

                Util.DebugBreak();

                l_CompletedBytes = l_CompletedBytesInclUnverified;
            }

            return new IncomplFileProgress(0, 0, l_DoneMap, null, l_CompletedBytes, l_CompletedBytesInclUnverified);
        }

        public static float CalcQuotient(ulong a_Dividend,
                                         ulong a_Divisor)
        {
            float l_Quot = ((float)a_Dividend / a_Divisor);

            if (a_Dividend < a_Divisor)
            {
                // Do not return 1f in this case.

                l_Quot = Math.Min(0.999f, l_Quot);
            }

            return l_Quot;
        }
    }
}

namespace NB.Content.Internal
{
    [Flags]
    enum IncomplFileFlags : ushort
    {
        None = 0,

        EnablePreview = 0x1,
        AlloweD2kVerification = 0x2,
        AllowTAICHVerification = 0x4,
        ExcessiveTAICHBlockReverification = 0x8,

        All = (EnablePreview | AlloweD2kVerification | AllowTAICHVerification | ExcessiveTAICHBlockReverification),
        Default =

        All,
    }

    class IncompleteFile : ListBaseClass<IncompleteFile>
    {
        // Eight MiB.
        public const uint c_ChunkSize = 0x800000;
        // Because 8 MiB / 512 KiB = 16 = 2^4.
        private const byte c_TAICHExponent = 4;
        private const uint c_eD2kAccBufferSize = (64 * 1024);
        private const uint c_TAICHAccBufferSize = (64 * 1024);

        private const uint c_MainMagic = 0x4DCECAC8;
        private const uint c_UsersMagic = 0xA3BD0A07;
        private const uint c_eD2kHashMagic = 0xE91C2267;
        private const uint c_TAICHMagic = 0xF9CA5CA4;
        private const uint c_ComplChunkMagic = 0xC0DC07FC;
        private const uint c_IncomplChunkMagic = 0x987564D7;

        // c_BufferSize should at least be able to hold as many bytes
        // as c_MaxChunkKeysPerBuffer.
        public const ushort c_BufferSize = IncomplChunkHeader.c_Size;
        public const ushort c_MaxChunkKeysPerBuffer = (c_BufferSize / sizeof(ulong));

        private const byte c_DefaultHeaderSize = 16;
        private const byte c_DefaultChunkHeaderSize = 24;
        private const ushort c_HeaderSize = 512;

        public const uint c_MaxPreviewChunks = 2;
        private const ushort c_MaxChunksPerAvailList = 512;

        private const ushort c_MaxUsers = 512;
        private const byte c_MaxHotIncomplChunks = 32;
        private const uint c_UpdateUsersInterval = (5 * Time.c_SecsPerMinute);

        // This limit is only of concern for TAICH block hashs.
        private const byte c_MaxHashReqs = 32;
        private const byte c_MaxUnmergedUsersPerIncomplChunk = 4;
        private const byte c_MaxAvoidedUsers = 32;
        private const ushort c_ChunkReverMaxTestsPerRun = 1024;
        private const byte c_ChunkReverMaxVerificationsPerRun = 5;

        private readonly Instance f_Inst;
        public readonly KnownFile f_Known;
        private readonly HardMemory f_IncomplMem;
        private readonly Resource f_Resource;
        private readonly HardMemory.Element f_MainElm;

        private readonly uint f_ChunkCount;
        private readonly uint f_ChunkKeyFieldOffset;
        private readonly BitMap f_DoneMap;
        private ushort f_MapRevision;

        // f_HotIncomplChunks maps chunk indices to element keys. Keys
        // in the dictionary can be zero.
        private Dictionary<uint, ElmKey> f_HotIncomplChunks;
        private ushort f_HotIncomplRevision;

        private bool f_NotifiedCompletion;

        private MainHeader f_Main;

        // f_eD2kHashSetElementCount will be zero for small files.
        private readonly uint f_eD2kHashSetElementCount;
        // p_Resource must be acquired before accessing
        // f_eD2kAccessor.
        private eD2kHashSetAccessor f_eD2kAccessor;

        private TAICHTreeLayout f_TAICHLayout;
        // f_TAICHGreatestUpperLevel will be zero for small files.
        private readonly int f_TAICHGreatestUpperLevel;
        // p_Resource must be acquired before accessing
        // f_TAICHAccessor.
        private TAICHTreeAccessor f_TAICHAccessor;

        // There may be several source cookies that have the same
        // user number (e.g. for up- and downloading).
        private readonly Dictionary<SourceCookie, SourceCookie> f_Sources;
        private ListHead<ObtainCookie> f_ObtainList;

        // All three list can be null.
        private List<HashDemand> f_HashDemand;
        private List<UserNum> f_AvoidedUsers;
        private List<uint> f_RecheckChunks;

        private bool f_ComplChunkReverInProgress;
        private uint f_ReverificationChunkIndex;

        private HashStorage f_HashStorage;
        private UserFileKnowledge f_OwnKnowledge;
        private UserFileKnowledge f_EstimatedKnowledge;
        private VerificationMode f_VerificationMode;
        private bool f_HadNonzeroVerificationMode;
        private FlexDictionary<UserNum, UserFileKnowledge> f_UserKnowledge;
        // f_UsersAreDirty indicates that the Close() routine should
        // write the content of f_UserKnowledge to disk.
        private bool f_UsersAreDirty;
        private UserManager.QueryUserInfosCookie f_QueryUserInfosCookie;
        private Time f_NextQueryUserInfoTime;
        // f_UsersUpdated indicates that the user manager was reached
        // at least once to update the content of f_UserKnowledge.
        private bool f_UsersUpdated;

        private TempBuffers f_Cntxt;
        private IncomplFileFlags f_IncomplFlags;

        private bool f_TryToLoadHashSetsOnce;

        private byte[] p_Buffer
        {
            get { return f_Cntxt.f_Buffer; }
        }

#region ConstructionAndDestruction

        private IncompleteFile(Instance a_Inst,
                               KnownFile a_Known,
                               HardMemory a_IncomplMem,
                               HardMemory.Element a_MainElm)
        {
            f_Inst = a_Inst;
            f_Known = a_Known;
            f_IncomplMem = a_IncomplMem;
            f_Resource = f_IncomplMem.p_Resource;
            f_MainElm = a_MainElm;

            uint l_Junk;
            CalcMainOffsets(f_Known.f_FileSize, out f_ChunkCount, out f_ChunkKeyFieldOffset, out l_Junk);
            f_DoneMap = new BitMap(c_ChunkSize, f_ChunkCount);

            f_TAICHLayout = new TAICHTreeLayout(f_Known.f_FileSize);
            f_TAICHGreatestUpperLevel = f_TAICHLayout.GetUpperTreeBoundaries(c_TAICHExponent);
            f_eD2kHashSetElementCount = eD2kHashSet.GeteD2kHashSetElementCount(f_Known.f_FileSize);

            f_Sources = new Dictionary<SourceCookie, SourceCookie>();
            f_ObtainList = new ListHead<ObtainCookie>();

            f_HotIncomplChunks = new Dictionary<uint, ElmKey>();

            HashStorage.Create(out f_HashStorage);
            f_OwnKnowledge = new UserFileKnowledge(UserNum.s_Self);
            f_EstimatedKnowledge = new UserFileKnowledge(UserNum.s_Self);
            FlexDictionary<UserNum, UserFileKnowledge>.Create(out f_UserKnowledge);

            f_IncomplFlags = IncomplFileFlags.Default;

            f_TryToLoadHashSetsOnce = true;
        }

        private static void CalcMainOffsets(ulong a_FileSize,
                                            out uint a_ChunkCount,
                                            out uint a_ChunkKeyFieldOffset,
                                            out uint a_MainElmSize)
        {
            a_ChunkCount = ((uint)(a_FileSize / c_ChunkSize) +
                            (0 == (a_FileSize % c_ChunkSize) ? 0u : 1u));
            a_ChunkKeyFieldOffset = (MainHeader.c_BitmapStartingOffset + a_ChunkCount);
            uint l_Remainder = (a_ChunkKeyFieldOffset % MainHeader.c_ChunkKeyFieldAlignment);
            a_ChunkKeyFieldOffset += (0 == l_Remainder ? 0 : (MainHeader.c_ChunkKeyFieldAlignment - l_Remainder));
            a_MainElmSize = (a_ChunkKeyFieldOffset + (a_ChunkCount * sizeof(ulong)));
        }

        private static OpStatus GetLoadErrorStatus(bool a_AfterReading)
        {
            return (false == a_AfterReading ? OpStatus.IoError : OpStatus.Corruption);
        }

        private OpStatus Load()
        {
            // Call this after instantiation.

            if ((f_ChunkKeyFieldOffset + (f_ChunkCount * sizeof(ulong))) != f_MainElm.f_MaxSize)
            {
                return OpStatus.Corruption;
            }

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Data = new ByteArray(p_Buffer, 0, MainHeader.c_Size);
                bool l_Read = f_IncomplMem.Read(f_MainElm, 0, ref l_Data);

                if (false == l_Read ||
                    false == CheckDefaultHeader(c_MainMagic) ||
                    MainHeader.ReadFileSizeFromHeaderBuffer(p_Buffer, 0) != f_Known.f_FileSize)
                {
                    f_Cntxt.DebugReleaseBuffers();

                    return GetLoadErrorStatus(l_Read);
                }

                f_Main = new MainHeader(p_Buffer, 0);
            }
            f_Cntxt.DebugReleaseBuffers();

            if (false == ReadWholeBitmap())
            {
                return OpStatus.IoError;
            }

            if (true == f_Main.f_UsersKey.IsNonzero())
            {
                OpStatus l_UserStatus = ReadUserKnowledge(f_Main.f_UsersKey);

                if (OpStatus.Done == l_UserStatus)
                {
                    UpdateVerificationMode();
                }
                else
                {
                    f_UserKnowledge.Clear();

                    if (OpStatus.Corruption == l_UserStatus)
                    {
                        // NOTE: We deliberately do not delete a corrupt
                        // element.

                        f_Main.f_UsersKey = new ElmKey();
                    }
                }
            }

            return OpStatus.Done;
        }

        public static OpStatus Initialize(Instance a_Inst,
                                          KnownFile a_Known,
                                          HardMemory a_IncomplMem,
                                          ref ElmKey a_MainElmKey,
                                          TempBuffers a_IncomplContext,
                                          out IncompleteFile a_Incompl)
        {
            // If a_MemKey is zero, a key will be generated. a_MemKey
            // and a_Incompl will be set to meaningful values
            // regardless of success.

            a_Incompl = null;

            OpStatus l_Status;
            Resource l_Resource = a_IncomplMem.p_Resource;
            ResourceAccessType l_CurAccessType;
            bool l_Created = false;
            bool l_Retrying = false;
            HardMemory.Element l_MainElm;

            uint l_Junk1;
            uint l_Junk2;
            uint l_ElmSize;
            CalcMainOffsets(a_Known.f_FileSize, out l_Junk1, out l_Junk2, out l_ElmSize);

            // Begin of critical section.
            {
                if (true == a_MainElmKey.IsZero())
                {
                    l_MainElm = null;

                    l_CurAccessType = ResourceAccessType.Exclusive;
                    l_Resource.AcquireExclusive();
                }
                else
                {
                    l_CurAccessType = ResourceAccessType.Shared;
                    l_Resource.AcquireShared();

                    l_MainElm = a_IncomplMem.GetElement(a_MainElmKey);
                }

                RETRY:

                if (null == l_MainElm)
                {
                    // Try to create a new element.

                    if (ComplStatus.Incomplete != a_Known.p_StatusVolatile)
                    {
                        // This isn't even an incomplete file, so
                        // refuse to create an element.

                        a_MainElmKey = new ElmKey();
                        l_Status = OpStatus.WontCreateElmSinceFileIsNotIncompl;

                        goto RELEASE_AND_EXIT;
                    }

                    l_Resource.SwitchToIfNecessaryAdv(ref l_CurAccessType, ResourceAccessType.Exclusive);
                    {
                        a_MainElmKey = FindUnoccupiedKey(a_IncomplMem);
                        l_MainElm = a_IncomplMem.Allocate(a_MainElmKey, l_ElmSize);

                        if (null == l_MainElm)
                        {
                            a_MainElmKey = new ElmKey();
                            l_Status = OpStatus.InsuffDiskSpace;

                            goto RELEASE_AND_EXIT;
                        }

                        bool l_HeaderWritten;

                        a_IncomplContext.DebugAcquireBuffers();
                        {
                            MainHeader l_Header = new MainHeader();
                            l_Header.Write(a_Known.f_FileNum, a_Known.f_FileSize, a_IncomplContext.f_Buffer, 0);
                            ByteArray l_Data = new ByteArray(a_IncomplContext.f_Buffer, 0, MainHeader.c_Size);

                            l_HeaderWritten = (true == a_IncomplMem.Write(l_MainElm, 0, ref l_Data) &&
                                               true == a_IncomplMem.Flush());
                        }
                        a_IncomplContext.DebugReleaseBuffers();

                        if (false == l_HeaderWritten)
                        {
                            // Remove the element just created.
                            bool l_Deleted = a_IncomplMem.Delete(l_MainElm);

                            a_MainElmKey = new ElmKey();
                            l_Status = OpStatus.IoError;

                            goto RELEASE_AND_EXIT;
                        }
                    }
                    l_Resource.Downgrade();
                    l_CurAccessType = ResourceAccessType.Shared;

                    l_Created = true;
                }

                IncompleteFile l_Incompl = new IncompleteFile(a_Inst, a_Known, a_IncomplMem, l_MainElm);
                l_Incompl.AttachThread(a_IncomplContext);

                l_Status = l_Incompl.Load();

                if (OpStatus.Done == l_Status)
                {
                    a_MainElmKey = l_MainElm.f_ElmKey;
                    a_Incompl = l_Incompl;
                }
                else
                {
                    Util.DebugBreak();

                    // If just an io error occurred, let a_MemKey as it
                    // is and return the error code.

                    // But if there is data corruption, try to create
                    // a new element and leave the old element alone.
                    if (OpStatus.Corruption == l_Status)
                    {
                        // We don't assume to land here after we've just
                        // created the header.

                        a_MainElmKey = new ElmKey();
                        l_MainElm = null;

                        if (false == l_Created &&
                            false == l_Retrying)
                        {
                            l_Retrying = true;

                            goto RETRY;
                        }
                    }
                }

                RELEASE_AND_EXIT:;
            }
            l_Resource.SwitchToIfNecessary(l_CurAccessType, ResourceAccessType.None);

            // End of critical section.

            return l_Status;
        }

        public void SetIncomplFileFlags(IncomplFileFlags a_Flags)
        {
            bool l_Change = (f_IncomplFlags != a_Flags);

            f_IncomplFlags = a_Flags;

            if (true == l_Change)
            {
                UpdateVerificationMode();
            }
        }

        public void Service(Time a_Time)
        {
            // This routine should be called from outside this class
            // once a while (especially after other operations).

            // First update the users.

            bool l_UsersJustUpdated;

            if (null == f_QueryUserInfosCookie)
            {
                if (false == f_NextQueryUserInfoTime.IsEarlierThan(a_Time))
                {
                    l_UsersJustUpdated = false;
                }
                else
                {
                    l_UsersJustUpdated = BeginUpdateUserInfos(a_Time);
                }
            }
            else
            {
                l_UsersJustUpdated = FinishUpdateUserInfos(a_Time);
            }

            if (true == l_UsersJustUpdated)
            {
                UpdateEstimatedHashes();
            }

            // Now take care of the hash sets.

            if (true == f_TryToLoadHashSetsOnce)
            {
                f_TryToLoadHashSetsOnce = false;

                LoadAllHashSets();
            }

            OpStatus l_HashStatus = SendHashRequests();

            // Now do chunk rechecks.

            if (VerificationMode.None != f_VerificationMode)
            {
                if (true == f_ComplChunkReverInProgress)
                {
                    DoComplChunkReverification();
                }

                if (null != f_RecheckChunks)
                {
                    DoChunkRechecks();
                }
            }
        }

        private void StartReverification()
        {
            f_ComplChunkReverInProgress = true;
            f_ReverificationChunkIndex = 0;
        }

        private void DoComplChunkReverification()
        {
            int l_InsertCount = 0;
            uint l_ChunkIndex = f_ReverificationChunkIndex;
            uint l_Limit = (uint)Math.Min((ulong)l_ChunkIndex + c_ChunkReverMaxTestsPerRun, f_ChunkCount);
            for (; l_ChunkIndex < l_Limit; l_ChunkIndex++)
            {
                if (false == f_DoneMap.Get(l_ChunkIndex) ||
                    (null != f_RecheckChunks &&
                     true == f_RecheckChunks.Contains(l_ChunkIndex)))
                {
                    continue;
                }

                bool l_Inserted = false;

                f_Resource.AcquireShared();
                {
                    ElmKey l_ElmKey;
                    OpStatus l_Status = ReadSingleChunkKey(l_ChunkIndex, out l_ElmKey);

                    if (OpStatus.Done == l_Status)
                    {
                        if (false == l_ElmKey.IsNonzero())
                        {
                            l_Status = OpStatus.Corruption;
                        }
                        else
                        {
                            ComplChunkHeader l_Header;
                            l_Status = ReadComplChunkHeader(l_ChunkIndex, l_ElmKey, out l_Header);

                            if (OpStatus.NeedReverification == l_Status)
                            {
                                l_Inserted = true;
                            }
                        }

                        if (OpStatus.Corruption == l_Status)
                        {
                            OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(l_ChunkIndex, l_ElmKey);
                        }
                    }
                }
                f_Resource.ReleaseShared();

                if (true == l_Inserted &&
                    c_ChunkReverMaxVerificationsPerRun == ++l_InsertCount)
                {
                    // Abort, or we'll hijack the thread.

                    l_ChunkIndex++;

                    break;
                }
            }

            if (f_ChunkCount > l_ChunkIndex)
            {
                f_ReverificationChunkIndex = l_ChunkIndex;
            }
            else
            {
                f_ComplChunkReverInProgress = false;
            }
        }

        private void DoChunkRechecks()
        {
            // NOTE: We can't enumerate on f_RecheckChunks, since it
            // may be modified in the code path.

            int l_Count = f_RecheckChunks.Count;

            for (int i = 0; i < l_Count; i++)
            {
                uint l_ChunkIndex = f_RecheckChunks[i];

                if (true == f_DoneMap.Get(l_ChunkIndex))
                {
                    // This chunk was verified in the meantime.

                    continue;
                }

                IncomplChunkHeader l_Header;
                OpStatus l_Status;

                f_Resource.AcquireShared();
                {
                    l_Status = ReadIncomplChunkHeaderWithKey(l_ChunkIndex, new ElmKey(), true, out l_Header);
                }
                f_Resource.ReleaseShared();

                if (OpStatus.Done == l_Status)
                {
                    OpStatus l_VerStatus = VerifyChunk(ref l_Header);
                }
            }

            f_RecheckChunks.RemoveRange(0, l_Count);
        }

        private ElmKey FindUnoccupiedKey()
        {
            return FindUnoccupiedKey(f_IncomplMem);
        }

        private static ElmKey FindUnoccupiedKey(HardMemory a_HardMem)
        {
            // Ensures that the zero key is never allocated.

            return a_HardMem.FindUnoccupiedKey(1);
        }

        public OpStatus Close()
        {
            UnregisterAllSourcesDueToUnload();

            if (true == f_UsersAreDirty)
            {
                OpStatus l_UserStatus = WriteUserKnowledge(ref f_Main.f_UsersKey);
            }

            bool l_Succ;

            f_Resource.AcquireShared();
            {
                l_Succ = WriteMainHeader(true);
            }
            f_Resource.ReleaseShared();

            return (false == l_Succ ? OpStatus.IoError : OpStatus.Done);
        }

        public static bool TouchElements(HardMemory a_HardMem,
                                         KnownFile a_Known,
                                         ref HardMemory.LeakFinder a_LeakFinder,
                                         TempBuffers a_Buffers)
        {
            // Returns false only if an io error occurred. Otherwise
            // touches all elements that are reachable through the
            // incomplete file of a_Known.

            HardMemory.Element l_MainElm = a_HardMem.GetElement(a_Known.f_IncomplKey);

            if (null == l_MainElm)
            {
                goto CLEAR_KEY_AND_EXIT;
            }

            uint l_ChunkCount;
            uint l_ChunkKeyFieldOffset;
            uint l_MainElmSize;
            CalcMainOffsets(a_Known.f_FileSize, out l_ChunkCount, out l_ChunkKeyFieldOffset, out l_MainElmSize);

            if (l_MainElmSize != l_MainElm.f_MaxSize)
            {
                goto CLEAR_KEY_AND_EXIT;
            }

            a_Buffers.DebugAcquireBuffers();

            try
            {
                ByteArray l_Data = new ByteArray(a_Buffers.f_Buffer, 0, MainHeader.c_Size);

                if (false == a_HardMem.Read(l_MainElm, 0, ref l_Data))
                {
                    return false;
                }

                MainHeader l_Header = new MainHeader(l_Data.f_Arr, 0);

                if (false == CheckDefaultHeader(l_Data.f_Arr, c_MainMagic, a_Known.f_FileNum) ||
                    MainHeader.ReadFileSizeFromHeaderBuffer(l_Data.f_Arr, 0) != a_Known.f_FileSize)
                {
                    goto CLEAR_KEY_AND_EXIT;
                }

                ulong[] l_KeyBuf = a_Buffers.f_KeyBuffer;
                uint l_ElmPos = l_ChunkKeyFieldOffset;
                for (uint l_KeysToGo = l_ChunkCount; 0 < l_KeysToGo; )
                {
                    int l_Cur = (int)Math.Min(IncompleteFile.c_MaxChunkKeysPerBuffer, l_KeysToGo);
                    l_Data = new ByteArray(a_Buffers.f_Buffer, 0, l_Cur * sizeof(ulong));

                    if (false == a_HardMem.Read(l_MainElm, l_ElmPos, ref l_Data))
                    {
                        return false;
                    }

                    Buffer.BlockCopy(l_Data.f_Arr, 0, l_KeyBuf, 0, l_Data.f_Len);
                    bool l_Dirty = false;

                    for (int k = 0; k < l_Cur; k++)
                    {
                        if (0 != l_KeyBuf[k])
                        {
                            HardMemory.LeakStates l_State = a_LeakFinder.Touch(new ElmKey(l_KeyBuf[k]));

                            if (HardMemory.LeakStates.NoError != l_State)
                            {
                                l_KeyBuf[k] = 0;
                                l_Dirty = true;
                            }
                        }
                    }

                    if (true == l_Dirty)
                    {
                        Buffer.BlockCopy(l_KeyBuf, 0, l_Data.f_Arr, 0, l_Data.f_Len);
                        bool l_WrittenBack = a_HardMem.Write(l_MainElm, l_ElmPos, ref l_Data);
                    }

                    l_ElmPos += (uint)l_Data.f_Len;
                    l_KeysToGo -= (uint)l_Cur;
                }

                HardMemory.LeakStates l_State1 = a_LeakFinder.TouchIfNotZero(l_Header.f_UsersKey);
                HardMemory.LeakStates l_State2 = a_LeakFinder.TouchIfNotZero(l_Header.f_eD2kHashSetKey);
                HardMemory.LeakStates l_State3 = a_LeakFinder.TouchIfNotZero(l_Header.f_TAICHUpperTreeKey);
                HardMemory.LeakStates l_State4 = a_LeakFinder.Touch(a_Known.f_IncomplKey);
            }
            finally
            {
                a_Buffers.DebugReleaseBuffers();
            }

            return true;

            CLEAR_KEY_AND_EXIT:

            Util.DebugBreak();

            a_Known.f_IncomplKey = new ElmKey();

            // NOTE: We land here when no io error happened, so return
            // true.
            return true;
        }

        public OpStatus DestroyForGood(ref ElmKey a_MainElmKey)
        {
            // Updates a_MainElmKey in any case.

            UnregisterAllSourcesDueToUnload();

            // We can't use f_KeyBuffer in this routine, since our
            // codepath uses the buffers too.
            ulong[] l_LocalKeyBuffer = f_Cntxt.CreateKeyBuffer(f_ChunkCount);

            f_Resource.AcquireExclusive();
            {
                // We won't check return values of DeleteElementChecked(),
                // since there isn't much we can do here if an element
                // should ever get stuck.

                // Delete file content chunks.

                for (uint l_ChunkIndex = 0; true; )
                {
                    uint l_Count = (f_ChunkCount - l_ChunkIndex);

                    if (0 == l_Count)
                    {
                        break;
                    }

                    l_Count = Math.Min(l_Count, c_MaxChunkKeysPerBuffer);
                    OpStatus l_Status = ReadChunkKeys(0, l_Count, l_LocalKeyBuffer);

                    if (OpStatus.Done != l_Status)
                    {
                        break;
                    }

                    for (int k = 0, lim = (int)l_Count; k < lim; k++)
                    {
                        if (0 != l_LocalKeyBuffer[k])
                        {
                            DeleteElementChecked(c_IncomplChunkMagic, new ElmKey(l_LocalKeyBuffer[k]));
                        }
                    }

                    l_ChunkIndex += l_Count;
                }

                // Delete special elements.

                bool l_Deleted1 = DeleteElementChecked(c_UsersMagic, f_Main.f_UsersKey);
                bool l_Deleted2 = DeleteElementChecked(c_eD2kHashMagic, f_Main.f_eD2kHashSetKey);
                bool l_Deleted3 = DeleteElementChecked(c_TAICHMagic, f_Main.f_TAICHUpperTreeKey);

                // The main element is now all that's left.

                bool l_Deleted4 = DeleteElementChecked(c_MainMagic, f_MainElm.f_ElmKey);
            }
            f_Resource.ReleaseExclusive();

            a_MainElmKey = new ElmKey();

            return OpStatus.Done;
        }

        private bool DeleteElementChecked(uint a_ExpectedMagic,
                                          ElmKey a_ElmKey)
        {
            if (true == a_ElmKey.IsZero())
            {
                return false;
            }

            HardMemory.Element l_Elm = f_IncomplMem.GetElement(a_ElmKey);

            return (null != l_Elm &&
                    true == DeleteElementChecked(a_ExpectedMagic, l_Elm));
        }

        private bool DeleteElementChecked(uint a_ExpectedMagic,
                                          HardMemory.Element a_Elm)
        {
            if (c_HeaderSize > a_Elm.f_MaxSize)
            {
                return false;
            }

            bool l_ReadAndVerified;

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Data = new ByteArray(p_Buffer, 0, c_DefaultHeaderSize);
                l_ReadAndVerified = f_IncomplMem.Read(a_Elm, 0, ref l_Data);

                if (true == l_ReadAndVerified &&
                    false == CheckDefaultHeader(a_ExpectedMagic))
                {
                    // If we've checked for an incomplete chunk and failed,
                    // we'll recheck for a complete chunk.

                    if (c_IncomplChunkMagic != a_ExpectedMagic ||
                        false == CheckDefaultHeader(c_ComplChunkMagic))
                    {
                        l_ReadAndVerified = false;
                    }
                }
            }
            f_Cntxt.DebugReleaseBuffers();

            if (false == l_ReadAndVerified)
            {
                // We couldn't read or we are unsure what we have, so
                // don't delete.

                return false;
            }

            return f_IncomplMem.Delete(a_Elm);
        }

        private OpStatus UpdateUserOrHashSetElement(uint a_ExpectedMagic,
                                                    bool a_PreserveOldOnFail,
                                                    ref ByteArray a_Header,
                                                    ref ByteArray a_Body,
                                                    FHandle a_File,
                                                    ref ElmKey a_ElmKeyInMain)
        {
            // a_Header may be empty. a_File can be used instead of
            // a_Body, in which case a_Body.f_Arr must be null and
            // a_Body.f_Len must still indicate the body size.

            OpStatus l_Status;
            uint l_HeaderSize = (uint)a_Header.f_Len;
            uint l_TotalSize = (l_HeaderSize + (uint)a_Body.f_Len);
            bool l_UpdateInPlace = (0 == a_Header.f_Len);
            HardMemory.Element l_Elm = null;

            f_Resource.AcquireExclusive();
            {
                // Load or create element.

                if (true == l_UpdateInPlace &&
                    true == a_ElmKeyInMain.IsNonzero())
                {
                    l_Elm = f_IncomplMem.GetElement(a_ElmKeyInMain);

                    if (null != l_Elm)
                    {
                        if (c_DefaultHeaderSize > l_Elm.f_MaxSize)
                        {
                            l_Elm = null;
                        }
                        else
                        {
                            f_Cntxt.DebugAcquireBuffers();
                            {
                                ByteArray l_Data = new ByteArray(p_Buffer, 0, c_DefaultHeaderSize);

                                if (false == f_IncomplMem.Read(l_Elm, 0, ref l_Data) ||
                                    false == CheckDefaultHeader(l_Data.f_Arr, a_ExpectedMagic, f_Known.f_FileNum))
                                {
                                    l_Elm = null;
                                }
                            }
                            f_Cntxt.DebugReleaseBuffers();
                        }
                    }
                }

                l_Status = OpStatus.Done;

                if (null == l_Elm)
                {
                    l_UpdateInPlace = false;

                    l_Elm = f_IncomplMem.Allocate(FindUnoccupiedKey(), l_TotalSize);

                    if (null == l_Elm)
                    {
                        l_Status = OpStatus.InsuffDiskSpace;
                    }
                }

                if (OpStatus.Done == l_Status)
                {
                    // Fill element.

                    bool l_Succ;
                    bool l_UseBodyBuffer = (null != a_Body.f_Arr);

                    if (false == l_UpdateInPlace)
                    {
                        if (0 == l_HeaderSize)
                        {
                            l_Succ = true;
                        }
                        else
                        {
                            l_Succ = f_IncomplMem.Write(l_Elm, 0, ref a_Header);
                        }

                        if (true == l_Succ)
                        {
                            if (false == l_UseBodyBuffer)
                            {
                                l_Succ = f_IncomplMem.Write(l_Elm, l_HeaderSize, a_File);
                            }
                            else
                            {
                                l_Succ = f_IncomplMem.Write(l_Elm, l_HeaderSize, ref a_Body);
                            }
                        }
                    }
                    else
                    {
                        if (false == l_UseBodyBuffer)
                        {
                            l_Succ = f_IncomplMem.SetAtomic(l_Elm.f_ElmKey, a_File);
                        }
                        else
                        {
                            l_Succ = f_IncomplMem.SetAtomic(l_Elm.f_ElmKey, ref a_Body);
                        }
                    }

                    if (false == l_Succ ||
                        false == f_IncomplMem.Flush())
                    {
                        l_Status = OpStatus.IoError;
                    }
                }

                // Handle main element references.

                const bool l_Flush = true;
                ElmKey l_OldElmKey = a_ElmKeyInMain;
                bool l_DelOldElm = false;
                bool l_DelElm = false;

                if (OpStatus.Done == l_Status &&
                    false == l_UpdateInPlace)
                {
                    a_ElmKeyInMain = l_Elm.f_ElmKey;

                    if (false == WriteMainHeader(l_Flush))
                    {
                        l_Status = OpStatus.IoError;

                        if (true == a_PreserveOldOnFail)
                        {
                            a_ElmKeyInMain = l_OldElmKey;
                        }

                        // Further handling below.
                    }
                    else
                    {
                        l_DelOldElm = l_OldElmKey.IsNonzero();
                    }
                }

                if (OpStatus.Done != l_Status)
                {
                    if (false == a_PreserveOldOnFail &&
                        true == l_OldElmKey.IsNonzero())
                    {
                        a_ElmKeyInMain = new ElmKey();
                        bool l_Written = WriteMainHeader(l_Flush);

                        l_DelOldElm = true;
                    }

                    l_DelElm = (null != l_Elm);
                }

                if (true == l_DelOldElm)
                {
                    bool l_Deleted = DeleteElementChecked(a_ExpectedMagic, l_OldElmKey);
                }

                if (true == l_DelElm)
                {
                    bool l_Deleted = DeleteElementChecked(a_ExpectedMagic, l_Elm);
                }
            }
            f_Resource.ReleaseExclusive();

            return l_Status;
        }

        public void AttachThread(TempBuffers a_Context)
        {
            f_Cntxt = a_Context;
        }

        public void DetachThread()
        {
            if (false == f_Resource.NoLockHeld())
            {
                // This should of course never happen, but we can heal
                // it if it happens (instead of just crashing).

                Util.DebugBreak();

                f_Resource.ReleaseAny();
            }

            f_Cntxt = null;
        }

#endregion

        private uint GetChunkSize(uint a_ChunkIndex)
        {
            if ((f_ChunkCount - 1) == a_ChunkIndex)
            {
                // The last chunk.

                uint l_Remainder = (uint)(f_Known.f_FileSize % c_ChunkSize);

                if (0 != l_Remainder)
                {
                    return l_Remainder;
                }
            }

            return c_ChunkSize;
        }

        private int GetVerifiedBufferSize(uint a_ChunkIndex)
        {
            return (int)(ComplChunkHeader.c_BodyStartingOffset + GetChunkSize(a_ChunkIndex));
        }

        private bool WriteMainHeader(bool a_FlushAfterwards)
        {
            bool l_Succ;

            f_Cntxt.DebugAcquireBuffers();
            {
                f_Main.Write(f_Known.f_FileNum, f_Known.f_FileSize, p_Buffer, 0);
                ByteArray l_Data = new ByteArray(p_Buffer, 0, MainHeader.c_Size);

                l_Succ = (true == f_IncomplMem.Write(f_MainElm, 0, ref l_Data) &&
                          (false == a_FlushAfterwards ||
                           true == f_IncomplMem.Flush()));
            }
            f_Cntxt.DebugReleaseBuffers();

            return l_Succ;
        }

        private bool WriteBytesInIncomplChunks()
        {
            bool l_Succ;

            f_Cntxt.DebugAcquireBuffers();
            {
                l_Succ = f_Main.WriteBytesInIncomplChunks(f_IncomplMem, f_MainElm, p_Buffer);
            }
            f_Cntxt.DebugReleaseBuffers();

            return l_Succ;
        }

#region Misc

        private OpStatus LockAndResize(ref IncomplChunkHeader a_Header,
                                       uint a_NewMaxSize)
        {
            // Resizes an incomplete element.

            OpStatus l_Status;

            if (a_Header.p_Element.f_MaxSize == a_NewMaxSize)
            {
                l_Status = OpStatus.Done;
            }
            else
            {
                f_Resource.AcquireExclusive();
                {
                    l_Status = a_Header.ResizeElement(f_IncomplMem, a_NewMaxSize);
                }
                f_Resource.ReleaseExclusive();
            }

            return l_Status;
        }

        private bool CheckDefaultHeader(uint a_ExpectedMagic)
        {
            // This routine uses p_Buffer (read only).

            return CheckDefaultHeader(p_Buffer, a_ExpectedMagic, f_Known.f_FileNum);
        }

        private static bool CheckDefaultHeader(byte[] a_Buffer,
                                               uint a_ExpectedMagic,
                                               FileNum a_FileNum)
        {
            return (Util.ReadUInt32LE(a_Buffer, 0) == a_ExpectedMagic &&
                    Util.ReadUInt32LE(a_Buffer, 4) == 0 &&
                    Util.ReadUInt64LE(a_Buffer, 8) == a_FileNum.f_Val);
        }

        private bool CheckDefaultChunkHeader(uint a_ExpectedMagic,
                                             uint a_ExpectedChunkIndex)
        {
            // This code path uses p_Buffer (read only).

            return (true == CheckDefaultHeader(a_ExpectedMagic) &&
                    Util.ReadUInt64LE(p_Buffer, 16) == a_ExpectedChunkIndex);
        }

        private static void WriteDefaultHeader(uint a_Magic,
                                               FileNum a_FileNum,
                                               byte[] a_Dest,
                                               int a_DestIndex)
        {
            Util.WriteUInt32LE(a_Magic, a_Dest, a_DestIndex);
            Util.WriteUInt32LE(0, a_Dest, a_DestIndex + 4);
            Util.WriteUInt64LE(a_FileNum.f_Val, a_Dest, a_DestIndex + 8);
        }

        private static void WriteDefaultChunkHeader(uint a_Magic,
                                                    FileNum a_FileNum,
                                                    uint a_ChunkIndex,
                                                    byte[] a_Dest,
                                                    int a_DestIndex)
        {
            WriteDefaultHeader(a_Magic, a_FileNum, a_Dest, a_DestIndex);
            Util.WriteUInt64LE(a_ChunkIndex, a_Dest, a_DestIndex + 16);
        }

        private OpStatus HandleChunkCorruptionFromShrdOrExcl(uint a_ChunkIndex,
                                                             ElmKey a_ElmKey)
        {
            OpStatus l_Status;

            ResourceAccessType l_OldAccessType = f_Resource.SwitchToExclusive();
            {
                l_Status = HandleChunkCorruptionFromExcl(a_ChunkIndex, a_ElmKey);
            }
            f_Resource.SwitchToIfNecessary(ResourceAccessType.Exclusive, l_OldAccessType);

            return l_Status;
        }

        private OpStatus HandleChunkCorruptionFromExcl(uint a_ChunkIndex,
                                                       ElmKey a_ElmKey)
        {
            // a_ElmKey may be zero.

            bool l_BitCleared = UpdateBitAndWrite(a_ChunkIndex, false);

            if (true == a_ElmKey.IsNonzero())
            {
                // NOTE: We don't delete the element, but leave it up
                // to the garbage collector.

                bool l_KeyCleared = WriteChunkKey(a_ChunkIndex, new ElmKey());
            }

            UpdateHotIncomplChunksIfEntryExists(a_ChunkIndex, new ElmKey());

            return OpStatus.Done;
        }

#endregion

#region UserHandling

        public void SetOwnFileKnowledge(ulong a_FileSize,
                                        ref MasterHashes a_MasterHashes)
        {
            UserFileKnowledge l_NewKnowledge = new UserFileKnowledge(UserNum.s_Self);
            l_NewKnowledge.Change(f_Inst.GetTime(), a_FileSize, ref f_HashStorage, ref a_MasterHashes);

            if (false == f_OwnKnowledge.HasSameKnowledge(ref l_NewKnowledge))
            {
                f_OwnKnowledge.Deref(ref f_HashStorage);
                f_HashStorage.TrimExcess();

                f_OwnKnowledge = l_NewKnowledge;
                UpdateEstimatedHashes();
            }
        }

        public bool HasActiveSources()
        {
            return (0 < f_Sources.Count);
        }

        public OpStatus RegisterSource(SourceCookie a_Cookie)
        {
            if (CookieStates.Pending != a_Cookie.p_PeerStateVolatile ||
                CookieStates.Pending != a_Cookie.p_ContentMngrState)
            {
                a_Cookie.RejectSource();

                return OpStatus.InputRejected;
            }

            if (true == a_Cookie.f_UserNum.IsNonGeneric() &&
                null != f_AvoidedUsers &&
                true == f_AvoidedUsers.Contains(a_Cookie.f_UserNum))
            {
                a_Cookie.RejectSource();

                return OpStatus.InputRejected;
            }

            f_Sources.Add(a_Cookie, a_Cookie);
            a_Cookie.AcceptSource();

            return OpStatus.Done;
        }

        private bool IsSourceCookieRegistered(SourceCookie a_Cookie)
        {
            return f_Sources.ContainsKey(a_Cookie);
        }

        public OpStatus UnregisterSource(SourceCookie a_Cookie)
        {
            CookieStates l_State = a_Cookie.p_ContentMngrState;

            if (CookieStates.WaitingForHashes == l_State)
            {
                // Due to timing issues, the network thread missed
                // that it is currently supposed to deliver hashs,
                // so abandon the query first.

                DeliverHashContext l_Cntxt = new DeliverHashContext(a_Cookie, null, null);
                OpStatus l_DeliverStatus = DeliverHashes(ref l_Cntxt);
            }
            else if (CookieStates.Success != l_State)
            {
                // It can happen that a cookie lands here without
                // being registered correctly, since the network
                // thread doesn't always know that a cookie was
                // rejected in the meantime.

                return OpStatus.InputRejected;
            }

            StopDownloadsThroughCookie(a_Cookie);

            bool l_Removed = f_Sources.Remove(a_Cookie);

            a_Cookie.UnacceptSource();

            return OpStatus.Done;
        }

        private void UnregisterAllSourcesDueToUnload()
        {
            // Remove obtain cookies first.

            for (ObtainCookie l_Cookie = f_ObtainList.p_First; null != l_Cookie; l_Cookie = l_Cookie.p_Next)
            {
                l_Cookie.Rejected(ObtainCookie.RejectionReasons.IncomplFileUnloads);
            }

            f_ObtainList.Clear();

            // Now remove source cookies.

            foreach (SourceCookie l_Cookie in f_Sources.Values)
            {
                CookieStates l_State = l_Cookie.p_ContentMngrState;

                if (CookieStates.WaitingForHashes == l_State)
                {
                    // Simulate the end of the remaining query.
                    l_Cookie.ResetAfterHashDelivery();
                }
                else
                {
                }

                l_Cookie.UnacceptSource();
            }

            f_Sources.Clear();
        }

        public OpStatus UpdateSource(ref UpdateSourceContext a_Cntxt)
        {
            if (true == a_Cntxt.f_Cookie.f_UserNum.IsGeneric())
            {
                // Should never happen.

                Util.DebugBreak();

                return OpStatus.Error;
            }

            a_Cntxt.f_Cookie.SetProvidableHashSets(a_Cntxt.f_ProvidableHashSets);

            if (true == UpdateUserKnowledge(ref a_Cntxt))
            {
                if (true == a_Cntxt.f_Cookie.IsRetrievingHashes())
                {
                    InvalidateHashQuery(a_Cntxt.f_Cookie);
                }

                UpdateEstimatedHashes();
            }

            return OpStatus.Done;
        }

        public OpStatus QueryUsers(QueryUsersCookie a_Cookie)
        {
            UserNum[] l_Users = f_UserKnowledge.ToKeyArray(true);

            if (null == l_Users)
            {
                a_Cookie.Fail();
            }
            else
            {
                a_Cookie.Succeed(l_Users);
            }

            return OpStatus.Done;
        }

        private bool UpdateUserKnowledge(ref UpdateSourceContext a_Cntxt)
        {
            // Returns true if new and old data differ.

            int l_Idx;
            FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr;
            bool l_Change;

            if (false == f_UserKnowledge.TryGet(a_Cntxt.f_Cookie.f_UserNum, out l_Idx))
            {
                // A user we don't know.

                if (c_MaxUsers <= f_UserKnowledge.p_Count)
                {
                    // Find the least important user.

                    l_Idx = (-1);
                    FlexDicEnum<UserNum, UserFileKnowledge> l_Enum;
                    l_Arr = f_UserKnowledge.ExposeItemArray(out l_Enum);
                    for (int i; true == l_Enum.GetNextIndex(l_Arr, out i);)
                    {
                        if (0 > l_Idx ||
                            true == l_Arr[l_Idx].f_Val.IsLessImportant(ref l_Arr[i].f_Val))
                        {
                            l_Idx = i;
                        }
                    }

                    bool l_Removed = f_UserKnowledge.Remove(l_Arr[l_Idx].p_Key);
                }

                bool l_Added = f_UserKnowledge.PrepareAdd(a_Cntxt.f_Cookie.f_UserNum, out l_Idx, out l_Arr);

                l_Change = true;
                l_Arr[l_Idx].f_Val = new UserFileKnowledge(a_Cntxt.f_Cookie.f_UserNum);
            }
            else
            {
                l_Change = false;
                l_Arr = f_UserKnowledge.ExposeItemArray();
            }

            if (true == l_Arr[l_Idx].f_Val.Change(f_Inst.GetTime(), a_Cntxt.f_FileSize, ref f_HashStorage, ref a_Cntxt.f_MasterHashes))
            {
                l_Change = true;

                f_HashStorage.TrimExcess();
            }

            if (true == l_Change)
            {
                f_UsersAreDirty = true;
            }

            return l_Change;
        }

        private void UpdateEstimatedHashes()
        {
            if (false == f_UsersUpdated)
            {
                // We're not ready yet.

                return;
            }

            UserFileKnowledge l_NewEstimatedKnowledge;
            UserFileKnowledge.EstimateKnowledge(ref f_OwnKnowledge, ref f_UserKnowledge, ref f_HashStorage, out l_NewEstimatedKnowledge);

            if (true == f_EstimatedKnowledge.HasSameKnowledge(ref l_NewEstimatedKnowledge))
            {
                l_NewEstimatedKnowledge.Deref(ref f_HashStorage);

                return;
            }

            P2PHashSets l_InvalidHashSets = P2PHashSets.None;
            int l_OldeD2kIdx = f_EstimatedKnowledge.p_eD2kHashIdx;
            int l_NeweD2kIdx = l_NewEstimatedKnowledge.p_eD2kHashIdx;

            if (0 <= l_OldeD2kIdx &&
                (0 > l_NeweD2kIdx ||
                 l_OldeD2kIdx != l_NeweD2kIdx))
            {
                if (0 != f_eD2kHashSetElementCount)
                {
                    f_eD2kAccessor = new eD2kHashSetAccessor();
                }

                l_InvalidHashSets = P2PHashSets.eD2kHashSet;
            }

            int l_OldTAICHIdx = f_EstimatedKnowledge.p_TAICHHashIdx;
            int l_NewTAICHIdx = l_NewEstimatedKnowledge.p_TAICHHashIdx;

            if (0 <= l_OldTAICHIdx &&
                (0 > l_NewTAICHIdx ||
                 l_OldTAICHIdx != l_NewTAICHIdx))
            {
                if (0 < f_TAICHGreatestUpperLevel)
                {
                    f_TAICHAccessor = new TAICHTreeAccessor();
                }

                l_InvalidHashSets |= (P2PHashSets.TAICHTree | P2PHashSets.Upper8MiBTAICHTree);
            }

            f_EstimatedKnowledge.Deref(ref f_HashStorage);
            f_EstimatedKnowledge = l_NewEstimatedKnowledge;

            if (P2PHashSets.None != l_InvalidHashSets)
            {
                InvalidateHashQueries(l_InvalidHashSets);
                f_TryToLoadHashSetsOnce = true;
            }

            // Do this even if no hash sets were invalidated.
            UpdateVerificationMode();
        }

        private void UpdateVerificationMode()
        {
            VerificationMode l_NewMode;

            // We prefer TAICH over other ways.

            if (0 != (IncomplFileFlags.AllowTAICHVerification & f_IncomplFlags) &&
                true == f_EstimatedKnowledge.p_HasTAICHHash &&
                true == HasUpperTAICHTree())
            {
                l_NewMode = VerificationMode.TAICH;
            }
            else if (0 != (IncomplFileFlags.AlloweD2kVerification & f_IncomplFlags) &&
                     true == f_EstimatedKnowledge.p_HaseD2kHash &&
                     true == HaseD2kHashSet())
            {
                l_NewMode = VerificationMode.eD2k;
            }
            else
            {
                l_NewMode = VerificationMode.None;
            }

            if (l_NewMode != f_VerificationMode)
            {
                f_VerificationMode = l_NewMode;

                if (VerificationMode.None != f_VerificationMode)
                {
                    if (false == f_HadNonzeroVerificationMode)
                    {
                        // In this case we do nothing, otherwise we
                        // would reverify after each restart.

                        f_HadNonzeroVerificationMode = true;
                    }
                    else
                    {
                        StartReverification();
                    }
                }
            }
        }

        public byte[] GetExpectedMasterHashForVerification(out P2PHashSets a_ExpectedHashSet)
        {
            // Return can be null.

            if (VerificationMode.eD2k == f_VerificationMode)
            {
                a_ExpectedHashSet = P2PHashSets.eD2kHashSet;

                return f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_eD2kHashIdx);
            }
            else if (VerificationMode.TAICH == f_VerificationMode)
            {
                a_ExpectedHashSet = P2PHashSets.TAICHTree;

                return f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx);
            }
            else
            {
                a_ExpectedHashSet = P2PHashSets.None;

                return null;
            }
        }

        private bool BeginUpdateUserInfos(Time a_Time)
        {
            // Returns true if an update happened.

            if (null != f_QueryUserInfosCookie)
            {
                return false;
            }

            int l_Count = f_UserKnowledge.p_Count;

            if (0 >= l_Count)
            {
                f_NextQueryUserInfoTime = a_Time.AddSeconds(c_UpdateUsersInterval);
                f_UsersUpdated = true;

                return true;
            }

            UserManager.UserInfo[] l_Infos = new UserManager.UserInfo[l_Count];
            FlexDicEnum<UserNum, UserFileKnowledge> l_Enum;
            FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = f_UserKnowledge.ExposeItemArray(out l_Enum);
            for (int i, k = 0; true == l_Enum.GetNextIndex(l_Arr, out i); k++)
            {
                l_Infos[k] = new UserManager.UserInfo(l_Arr[i].p_Key);
            }

            f_QueryUserInfosCookie = new UserManager.QueryUserInfosCookie(f_Inst.f_UserMngr, l_Infos);
            f_Inst.f_NetworkMngr.EnqueueAPC(f_QueryUserInfosCookie);

            return false;
        }

        private bool FinishUpdateUserInfos(Time a_Time)
        {
            // Returns true if an update happened.

            if (null == f_QueryUserInfosCookie)
            {
                return false;
            }

            CookieStates l_State = f_QueryUserInfosCookie.p_StateVolatile;

            if (CookieStates.Pending == l_State)
            {
                return false;
            }

            if (CookieStates.Success == l_State)
            {
                UserManager.UserInfo[] l_Infos = f_QueryUserInfosCookie.f_Infos;
                FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = null;
                for (int i = (l_Infos.Length - 1); 0 <= i; i--)
                {
                    int l_Idx;

                    if (false == f_UserKnowledge.TryGet(l_Infos[i].f_UserNum, out l_Idx))
                    {
                        // We've dropped this user in the meantime.

                        continue;
                    }

                    if (false == l_Infos[i].f_KnownUser)
                    {
                        // The user manager doesn't know this guy.

                        bool l_Removed = f_UserKnowledge.Remove(l_Infos[i].f_UserNum);

                        l_Arr = null;
                    }
                    else
                    {
                        if (null == l_Arr)
                        {
                            l_Arr = f_UserKnowledge.ExposeItemArray();
                        }

                        l_Arr[l_Idx].f_Val.SetReputation(l_Infos[i].f_Reputation);
                    }
                }

                f_NextQueryUserInfoTime = f_QueryUserInfosCookie.f_QueryTime.AddSeconds(c_UpdateUsersInterval);
                f_UsersUpdated = true;
                f_QueryUserInfosCookie = null;

                return true;
            }

            Util.DebugBreak();

            f_NextQueryUserInfoTime = a_Time;

            return false;
        }

        private void BewareOfUser(UserNum a_UserNum,
                                  CorruptDataObserver a_Observer)
        {
            // p_Resource doesn't matter here.

            if (true == a_UserNum.IsGeneric())
            {
                Util.DebugBreak();

                return;
            }

            if (CorruptDataObserver.HashDelivery == a_Observer)
            {
                // We are quite certain that this is not by accident,
                // so ban this user globally (for this session).

                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.ContentMngrBanUser, new object[] { a_UserNum, -(10 * 1024) });
                f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);
            }

            if (null == f_AvoidedUsers)
            {
                f_AvoidedUsers = new List<UserNum>(1);
            }
            else
            {
                if (true == f_AvoidedUsers.Contains(a_UserNum))
                {
                    return;
                }

                if (c_MaxAvoidedUsers < f_AvoidedUsers.Count)
                {
                    f_AvoidedUsers.RemoveAt(0);
                }
            }

            f_AvoidedUsers.Add(a_UserNum);
            StopDownloadsFromBadUser(a_UserNum);
        }

        private OpStatus ReadUserKnowledge(ElmKey a_Key)
        {
            HardMemory.Element l_Elm = f_IncomplMem.GetElement(a_Key);

            if (null == l_Elm ||
                UserKnowledgeHeader.c_LargeSize > l_Elm.f_MaxSize)
            {
                return OpStatus.Corruption;
            }

            // Check the header.

            UserKnowledgeHeader l_Header;

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Data = new ByteArray(p_Buffer, 0, UserKnowledgeHeader.c_ShortSize);
                bool l_Read = f_IncomplMem.Read(l_Elm, 0, ref l_Data);

                if (false == l_Read ||
                    false == CheckDefaultHeader(c_UsersMagic) ||
                    false == UserKnowledgeHeader.Load(l_Data.f_Arr, l_Data.f_Idx, out l_Header) ||
                    c_MaxUsers < l_Header.f_UsersCount)
                {
                    f_Cntxt.DebugReleaseBuffers();

                    return GetLoadErrorStatus(l_Read);
                }
            }
            f_Cntxt.DebugReleaseBuffers();

            // Read the whole element into a buffer.

            ByteArray l_FullData = new ByteArray((int)l_Elm.f_MaxSize);

            if (false == f_IncomplMem.Read(l_Elm, 0, ref l_FullData))
            {
                return OpStatus.IoError;
            }

            MemoryIO l_Buffer = MemoryIO.CreateReadable(ref l_FullData);

            // Read the header.

            l_Buffer.SetPosition(UserKnowledgeHeader.c_HashStorageOffset);

            try
            {
                if (false == HashStorage.FromBuffer(l_Buffer, out f_HashStorage))
                {
                    return OpStatus.Corruption;
                }
            }
            catch (BufferIO.InterpretBufferException)
            {
                return OpStatus.Corruption;
            }

            l_Buffer.Align(UserFileKnowledge.c_SuggestedStartingAlignment, false);
            uint l_UsersPos = l_Buffer.p_Position;

            l_Buffer.SetPosition(UserKnowledgeHeader.c_OwnKnowledgeOffset);
            if (false == UserFileKnowledge.FromBuffer(l_Buffer, ref f_HashStorage, out f_OwnKnowledge))
            {
                return OpStatus.Corruption;
            }

            l_Buffer.SetPosition(UserKnowledgeHeader.c_EstimatedKnowledgeOffset);
            if (false == UserFileKnowledge.FromBuffer(l_Buffer, ref f_HashStorage, out f_EstimatedKnowledge))
            {
                return OpStatus.Corruption;
            }

            // Read each users knowledge.

            l_Buffer.SetPosition(l_UsersPos);

            if ((l_Header.f_UsersCount * UserFileKnowledge.c_Size) != l_Buffer.GetRemaining())
            {
                return OpStatus.Corruption;
            }

            f_UserKnowledge.SetCapacity((int)l_Header.f_UsersCount);

            for (int i = ((int)l_Header.f_UsersCount - 1); 0 <= i; i--)
            {
                UserNum l_UserNum = new UserNum(l_Buffer.ReadUInt64LE());
                int l_Idx;
                FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr;

                if (false == l_UserNum.IsNonGeneric() ||
                    false == f_UserKnowledge.PrepareAdd(l_UserNum, out l_Idx, out l_Arr))
                {
                    return OpStatus.Corruption;
                }

                l_Buffer.JumpBack(sizeof(ulong));

                if (false == UserFileKnowledge.FromBuffer(l_Buffer, ref f_HashStorage, out l_Arr[l_Idx].f_Val))
                {
                    return OpStatus.Corruption;
                }
            }

            return OpStatus.Done;
        }

        private OpStatus WriteUserKnowledge(ref ElmKey a_ElmKeyInMain)
        {
            uint l_Count = (uint)f_UserKnowledge.p_Count;
            MemoryIO l_Buffer = MemoryIO.CreateWritable();
            // We don't know the real size of HashStorage at this
            // point (just assume 1024 bytes work).
            l_Buffer.SetCapacity(UserKnowledgeHeader.c_LargeSize + 1024 + (l_Count * UserFileKnowledge.c_Size));

            // Write the header.

            UserKnowledgeHeader l_Header = new UserKnowledgeHeader(l_Count);

            f_Cntxt.DebugAcquireBuffers();
            {
                l_Header.Write(f_Known.f_FileNum, p_Buffer, 0, true);
                l_Buffer.WriteBytes(p_Buffer, 0, UserKnowledgeHeader.c_ShortSize);
            }
            f_Cntxt.DebugReleaseBuffers();

            l_Buffer.SetPosition(UserKnowledgeHeader.c_OwnKnowledgeOffset);
            f_OwnKnowledge.Write(l_Buffer);
            l_Buffer.SetPosition(UserKnowledgeHeader.c_EstimatedKnowledgeOffset);
            f_EstimatedKnowledge.Write(l_Buffer);

            l_Buffer.SetPosition(UserKnowledgeHeader.c_HashStorageOffset);
            f_HashStorage.Write(l_Buffer);

            // Write each users knowledge.

            l_Buffer.Align(UserFileKnowledge.c_SuggestedStartingAlignment, true);

            FlexDicEnum<UserNum, UserFileKnowledge> l_Enum;
            FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = f_UserKnowledge.ExposeItemArray(out l_Enum);
            for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
            {
                l_Arr[i].f_Val.Write(l_Buffer);
            }

            // Write to disk.

            ByteArray l_HeaderData = new ByteArray();
            ByteArray l_BodyData;
            l_Buffer.ExposeRange(0, l_Buffer.p_Length, out l_BodyData);

            return UpdateUserOrHashSetElement(c_UsersMagic, true, ref l_HeaderData, ref l_BodyData, null, ref a_ElmKeyInMain);
        }

#endregion

#region HashQuerying

        private P2PHashSets GetAvailHashSets(out MasterHashes a_MasterHashes)
        {
            // Returns the available hash sets.

            P2PHashSets l_AvailHashSets = P2PHashSets.None;
            a_MasterHashes = new MasterHashes();

            int l_eD2kIdx = f_EstimatedKnowledge.p_eD2kHashIdx;
            int l_TAICHIdx = f_EstimatedKnowledge.p_TAICHHashIdx;

            if (0 <= l_eD2kIdx)
            {
                a_MasterHashes.SetHash(P2PHashSets.eD2kHashSet, f_HashStorage.ExposeHash(l_eD2kIdx), 0);

                if (true == f_eD2kAccessor.IsValid())
                {
                    l_AvailHashSets |= P2PHashSets.eD2kHashSet;
                }
            }

            if (0 <= l_TAICHIdx)
            {
                a_MasterHashes.SetHash(P2PHashSets.TAICHTree, f_HashStorage.ExposeHash(l_TAICHIdx), 0);

                if (true == f_TAICHAccessor.IsValid())
                {
                    l_AvailHashSets |= P2PHashSets.Upper8MiBTAICHTree;
                }
            }

            return l_AvailHashSets;
        }

        private bool HaseD2kHashSet()
        {
            // Returns true if the hash set is at least implicitly
            // present (in case of small files the master hash is
            // sufficient).

            return (0 == f_eD2kHashSetElementCount ||
                    true == f_eD2kAccessor.IsValid());
        }

        private bool HasUpperTAICHTree()
        {
            // Returns true if the upper tree is at least implicitly
            // present (in case of small files the top hash is
            // sufficient).

            return (0 == f_TAICHGreatestUpperLevel ||
                    true == f_TAICHAccessor.IsValid());
        }

        private void RequestTAICHBlocks(uint a_ChunkIndex)
        {
            if (null == f_HashDemand ||
                c_MaxHashReqs > f_HashDemand.Count)
            {
                HashDemand l_Demand = new HashDemand(HashDemandTypes.NeedTAICHBlockHashes, a_ChunkIndex);

                AddHashDemand(ref l_Demand);
            }
        }

        private void AddHashDemand(ref HashDemand a_Demand)
        {
            if (null == f_HashDemand)
            {
                f_HashDemand = new List<HashDemand>(1);
            }
            else if (true == f_HashDemand.Contains(a_Demand))
            {
                return;
            }

            f_HashDemand.Add(a_Demand);
        }

        private void RemoveTAICHBlockRequest(uint a_ChunkIndex)
        {
            // Undoes the effect of RequestTAICHBlocks() if the item
            // is still there, otherwise does nothing.

            if (null != f_HashDemand)
            {
                HashDemand l_Demand = new HashDemand(HashDemandTypes.NeedTAICHBlockHashes, a_ChunkIndex);

                f_HashDemand.Remove(l_Demand);
            }
        }

        private void InvalidateHashQueries(P2PHashSets a_HashSets)
        {
            // Stops running queries concerning a certain type of hash
            // set (in case we now anticipate a different master hash
            // than before). a_HashSets can be a combination of various
            // hash sets.

            foreach (SourceCookie l_Cookie in f_Sources.Values)
            {
                if (true == l_Cookie.IsRetrievingHashes(a_HashSets))
                {
                    InvalidateHashQuery(l_Cookie);
                }
            }
        }

        private void InvalidateHashQuery(SourceCookie a_Cookie)
        {
            // Abort the running query and requeue it.

            HashDemand l_Demand;
            a_Cookie.GetHashDemand(out l_Demand);

            a_Cookie.SetInternallyAborted();

            AddHashDemand(ref l_Demand);
        }

        private bool IsRetrievingHashes(ref HashDemand a_Demand)
        {
            // Meant to be called by SendHashRequests() only.

            foreach (SourceCookie l_Cookie in f_Sources.Values)
            {
                if (true == l_Cookie.IsRetrievingHashes(ref a_Demand))
                {
                    return true;
                }
            }

            return false;
        }

        private OpStatus SendHashRequests()
        {
            if (null == f_HashDemand)
            {
                return OpStatus.Done;
            }

            List<SourceCookie> l_eD2kSrcs = null;
            List<SourceCookie> l_TAICHSrcs = null;

            for (int i = 0; i < f_HashDemand.Count;)
            {
                HashDemand l_Demand = f_HashDemand[i];
                SourceCookie l_SuitableSrc = null;
                byte[] l_MasterHash = null;
                bool l_Remove = false;

                switch (l_Demand.f_Type)
                {
                    case HashDemandTypes.NeedTAICHBlockHashes:
                        {
                            if (0 == (IncomplFileFlags.AllowTAICHVerification & f_IncomplFlags) ||
                                false == f_EstimatedKnowledge.p_HasTAICHHash ||
                                false == HasUpperTAICHTree())
                            {
                                l_Remove = true;
                            }
                            else
                            {
                                // Do not use P2PHashSets.Upper8MiBTAICHTree here.
                                l_SuitableSrc = FindSuitableHashSource(P2PHashSets.TAICHTree, ref l_TAICHSrcs);
                                l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx);
                            }
                        }
                        break;

                    case HashDemandTypes.NeedeD2kHashSet:
                        {
                            if (0 == (IncomplFileFlags.AlloweD2kVerification & f_IncomplFlags) ||
                                false == f_EstimatedKnowledge.p_HaseD2kHash ||
                                true == HaseD2kHashSet())
                            {
                                l_Remove = true;
                            }
                            else
                            {
                                l_SuitableSrc = FindSuitableHashSource(P2PHashSets.eD2kHashSet, ref l_eD2kSrcs);
                                l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_eD2kHashIdx);
                            }
                        }
                        break;

                    case HashDemandTypes.NeedTAICHUpperTree:
                        {
                            if (0 == (IncomplFileFlags.AllowTAICHVerification & f_IncomplFlags) ||
                                false == f_EstimatedKnowledge.p_HasTAICHHash ||
                                true == HasUpperTAICHTree())
                            {
                                l_Remove = true;
                            }
                            else
                            {
                                // Do not use P2PHashSets.TAICHTree here.
                                l_SuitableSrc = FindSuitableHashSource(P2PHashSets.Upper8MiBTAICHTree, ref l_TAICHSrcs);
                                l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx);
                            }
                        }
                        break;

                    default: throw (new ApplicationException());
                }

                if (false == l_Remove)
                {
                    // Call IsRetrievingHashes() as late as now and only
                    // if l_SuitableSrc is nonzero, due to performance
                    // reasons (it can be quite expensive).
                    l_Remove = (null != l_SuitableSrc &&
                                true == IsRetrievingHashes(ref l_Demand));
                }

                if (false == l_Remove &&
                    null != l_SuitableSrc)
                {
                    SourceCookie.HashQueryParams l_Params;

                    switch (l_Demand.f_Type)
                    {
                        case HashDemandTypes.NeedTAICHBlockHashes:
                            {
                                List<uint> l_Indices = TAICHTree.CalcSubTreeIndices(ref f_TAICHLayout, c_TAICHExponent, l_Demand.f_ChunkIndex, f_TAICHGreatestUpperLevel);

                                l_Params = new SourceCookie.HashQueryParams(P2PHashSets.TAICHTree, l_MasterHash, l_Indices.ToArray(), (uint)l_Indices.Count * Tiger.c_HashByteSize);
                            }
                            break;

                        case HashDemandTypes.NeedeD2kHashSet:
                            {
                                l_Params = new SourceCookie.HashQueryParams(P2PHashSets.eD2kHashSet, l_MasterHash, 0, f_eD2kHashSetElementCount, f_eD2kHashSetElementCount * MD_4.c_HashByteSize);
                            }
                            break;

                        case HashDemandTypes.NeedTAICHUpperTree:
                            {
                                uint l_HashStartingIdx;
                                uint l_HashCnt;
                                int l_Lvl = f_TAICHLayout.GetUpperTreeBoundaries(c_TAICHExponent, out l_HashStartingIdx, out l_HashCnt);

                                l_Params = new SourceCookie.HashQueryParams(P2PHashSets.TAICHTree, l_MasterHash, l_HashStartingIdx, l_HashCnt, l_HashCnt * Tiger.c_HashByteSize);
                            }
                            break;

                        default: throw (new ApplicationException());
                    }

                    l_SuitableSrc.RequestHashes(ref l_Demand, ref l_Params);

                    l_Remove = true;
                }

                if (true == l_Remove)
                {
                    f_HashDemand.RemoveAt(i);
                }
                else
                {
                    i++;
                }
            }

            return OpStatus.Done;
        }

        private SourceCookie FindSuitableHashSource(P2PHashSets a_TargetHashSet,
                                                    ref List<SourceCookie> a_List)
        {
            // a_TargetHashSet may not be a combination.

            if (null == a_List)
            {
                int l_ExpectedIndex = f_EstimatedKnowledge.GetIndex(a_TargetHashSet);

                if (0 == f_Sources.Count ||
                    0 > l_ExpectedIndex)
                {
                    return null;
                }

                a_List = new List<SourceCookie>();

                FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = f_UserKnowledge.ExposeItemArray();
                foreach (SourceCookie l_Cookie in f_Sources.Values)
                {
                    // NOTE: We'll access p_PeerState without volatile
                    // semantics, due to performance reasons.
                    if (CookieStates.Success == l_Cookie.p_PeerState &&
                        CookieStates.Success == l_Cookie.p_ContentMngrState &&
                        0 != (l_Cookie.p_ProvidableHashSets & a_TargetHashSet))
                    {
                        // Check if this source has the hash we expect.

                        int l_Idx;
                        if (true == f_UserKnowledge.TryGet(l_Cookie.f_UserNum, out l_Idx))
                        {
                            if (l_Arr[l_Idx].f_Val.GetIndex(a_TargetHashSet) == l_ExpectedIndex)
                            {
                                // This user may have several source
                                // contexts, but we may only use the
                                // ones with a p_ProvidableHashSets
                                // that fits.

                                a_List.Add(l_Cookie);
                            }
                        }
                    }
                }
            }

            int l_Count = a_List.Count;

            if (0 < l_Count)
            {
                int l_RandomIndex = (1 >= l_Count ? 0 : (int)(Util.GetRandomUIntFast() % (uint)l_Count));
                SourceCookie l_Src = a_List[l_RandomIndex];

                a_List.RemoveAt(l_RandomIndex);

                return l_Src;
            }

            return null;
        }

        public OpStatus DeliverHashes(ref DeliverHashContext a_Cntxt)
        {
            // This routine will take ownership of a_Cntxt.f_File in
            // any case.

            if (CookieStates.WaitingForHashes != a_Cntxt.f_Cookie.p_ContentMngrState)
            {
                if (null != a_Cntxt.f_File)
                {
                    a_Cntxt.f_File.Close();
                }

                return OpStatus.InputRejected;
            }

            UserNum l_UserNum = a_Cntxt.f_Cookie.f_UserNum;
            HashDemand l_Demand;
            a_Cntxt.f_Cookie.GetHashDemand(out l_Demand);
            bool l_NeedRetry = false;
            bool l_AccessorCreated = false;
            Sanctions l_Sanctions;
            Sanctions.Create(out l_Sanctions);

            // NOTE: We only handle positive sanctions right here,
            // negative sanctions that stem from hashs are handled
            // by BewareOfUser().

            try
            {
                if (true == a_Cntxt.f_Cookie.WasInternallyAborted())
                {
                    // We already skipped and requeued this query.

                    return OpStatus.Done;
                }

                l_NeedRetry = true;

                if (null == a_Cntxt.f_Buffer &&
                    null == a_Cntxt.f_File)
                {
                    // The query failed in the network parts of the
                    // application, so we'll try again.

                    return OpStatus.Done;
                }

                FHandle l_File;
                uint l_DeliveredSize;

                {
                    ulong l_RawDeliveredSize;

                    if (null != a_Cntxt.f_Buffer)
                    {
                        l_File = null;
                        l_RawDeliveredSize = (uint)a_Cntxt.f_Buffer.Length;
                    }
                    else
                    {
                        l_File = a_Cntxt.f_File.ExposeFHandle();

                        if (false == l_File.GetSize(out l_RawDeliveredSize))
                        {
                            return OpStatus.IoError;
                        }
                    }

                    if (a_Cntxt.f_Cookie.f_HashQueryParams.f_ExpectedSize != l_RawDeliveredSize)
                    {
                        BewareOfUser(l_UserNum, CorruptDataObserver.HashDelivery);

                        return OpStatus.InputRejected;
                    }

                    l_DeliveredSize = (uint)l_RawDeliveredSize;
                }

                switch (l_Demand.f_Type)
                {
                    case HashDemandTypes.NeedTAICHBlockHashes:
                        {
                            uint l_ChunkIndex = l_Demand.f_ChunkIndex;

                            if (false == f_EstimatedKnowledge.p_HasTAICHHash ||
                                false == f_TAICHAccessor.IsValid() ||
                                true == f_DoneMap.Get(l_ChunkIndex))
                            {
                                l_NeedRetry = false;

                                return OpStatus.InputRejected;
                            }

                            byte[] l_Buffer;

                            if (null != a_Cntxt.f_Buffer)
                            {
                                l_Buffer = a_Cntxt.f_Buffer;
                            }
                            else
                            {
                                ByteArray l_Data = new ByteArray((int)l_DeliveredSize);

                                if (false == l_File.Read(0, ref l_Data))
                                {
                                    return OpStatus.IoError;
                                }

                                l_Buffer = l_Data.f_Arr;
                            }

                            TigerHashHive l_BlockHive;
                            TigerHashHive.CreateReadOnly(a_Cntxt.f_Cookie.f_HashQueryParams.f_HashIndices, l_Buffer, out l_BlockHive);
                            OpStatus l_Status;

                            f_Resource.AcquireShared();
                            {
                                // First verify the hashs.

                                TAICHVerifyResult l_VerRes = TAICHTree.VerifySubTree(ref f_TAICHLayout, c_TAICHExponent, l_ChunkIndex, ref f_TAICHAccessor, ref l_BlockHive, f_Cntxt.f_Tiger);

                                if (TAICHVerifyResult.Valid != l_VerRes)
                                {
                                    if (TAICHVerifyResult.AccessFailed != l_VerRes)
                                    {
                                        BewareOfUser(l_UserNum, CorruptDataObserver.HashDelivery);
                                    }

                                    l_Status = OpStatus.InputRejected;

                                    goto RELEASE_AND_EXIT;
                                }

                                l_Sanctions.AddSanction(l_UserNum, l_DeliveredSize, 0);
                                l_NeedRetry = false;

                                // Now load the chunk header and store the hashs.

                                IncomplChunkHeader l_Header;
                                l_Status = ReadIncomplChunkHeaderWithKey(l_ChunkIndex, new ElmKey(), true, out l_Header);

                                if (OpStatus.Done != l_Status)
                                {
                                    if (OpStatus.ChunkHasNoKey == l_Status)
                                    {
                                        // There seems to be no such chunk.

                                        l_Status = OpStatus.InputRejected;
                                    }

                                    goto RELEASE_AND_EXIT;
                                }

                                if (false == l_Header.HasTAICHBlockHashes(false))
                                {
                                    l_Header.SetTAICHBlockHashes(ref l_BlockHive);

                                    bool l_Succ;

                                    f_Cntxt.DebugAcquireBuffers();
                                    {
                                        l_Succ = l_Header.Update(f_Known.f_FileNum, ref f_HashStorage, ref f_EstimatedKnowledge, f_IncomplMem, p_Buffer);
                                    }
                                    f_Cntxt.DebugReleaseBuffers();

                                    if (false == l_Succ)
                                    {
                                        l_Status = OpStatus.IoError;

                                        goto RELEASE_AND_EXIT;
                                    }
                                }

                                f_Resource.ReleaseShared();

                                // Try to verify the chunk right away.

                                if (true == l_Header.IsMarkedDisputed())
                                {
                                    OpStatus l_VerStatus = VerifyChunk(ref l_Header);
                                }

                                return OpStatus.Done;
                            }

                            RELEASE_AND_EXIT:

                            f_Resource.ReleaseShared();

                            return l_Status;
                        }

                    case HashDemandTypes.NeedeD2kHashSet:
                        {
                            if (true == f_EstimatedKnowledge.p_HaseD2kHash &&
                                true == HaseD2kHashSet())
                            {
                                l_NeedRetry = false;

                                return OpStatus.Done;
                            }

                            // Validate input.

                            if (null != a_Cntxt.f_Buffer)
                            {
                                f_Cntxt.f_MD4.Reset();
                                f_Cntxt.f_MD4.Update(a_Cntxt.f_Buffer);
                                f_Cntxt.f_MD4.Finish();
                            }
                            else
                            {
                                const uint l_MaxHashBufferSize = (256 * 1024);
                                ulong l_Pos = 0;
                                byte[] l_TempBuf = new byte[(int)Math.Min(l_DeliveredSize, l_MaxHashBufferSize)];

                                while (true)
                                {
                                    uint l_MaxBytes = uint.MaxValue;
                                    HashingStatus l_Stat = HashSet.CalcFileHash(l_File, l_DeliveredSize, f_Cntxt.f_MD4, ref l_MaxBytes, ref l_Pos, l_TempBuf);

                                    if (HashingStatus.Done == l_Stat)
                                    {
                                        break;
                                    }

                                    if (HashingStatus.NeedFurtherCalls != l_Stat)
                                    {
                                        return OpStatus.IoError;
                                    }
                                }
                            }

                            bool l_HashFits;

                            f_Cntxt.DebugAcquireBuffers();
                            {
                                f_Cntxt.f_MD4.GetDigest(p_Buffer, 0);
                                l_HashFits = f_HashStorage.HasThisHash(f_EstimatedKnowledge.p_eD2kHashIdx, p_Buffer, 0);
                            }
                            f_Cntxt.DebugReleaseBuffers();

                            if (false == l_HashFits)
                            {
                                BewareOfUser(l_UserNum, CorruptDataObserver.HashDelivery);

                                return OpStatus.InputRejected;
                            }

                            l_Sanctions.AddSanction(l_UserNum, l_DeliveredSize, 0);
                            OpStatus l_Status = StoreeD2kHashSet(l_UserNum, l_DeliveredSize, a_Cntxt.f_Buffer, l_File);

                            if (OpStatus.Done == l_Status)
                            {
                                l_NeedRetry = false;

                                if (OpStatus.Done == LoadeD2kHashSet())
                                {
                                    l_AccessorCreated = true;
                                }
                            }

                            return l_Status;
                        }

                    case HashDemandTypes.NeedTAICHUpperTree:
                        {
                            if (true == f_EstimatedKnowledge.p_HasTAICHHash &&
                                true == HasUpperTAICHTree())
                            {
                                l_NeedRetry = false;

                                return OpStatus.Done;
                            }

                            // Validate input.

                            TAICHTreeAccessor l_TAICHAcc;

                            if (null != a_Cntxt.f_Buffer)
                            {
                                ByteArray l_Data = new ByteArray(a_Cntxt.f_Buffer);
                                l_TAICHAcc = new TAICHTreeAccessor(ref f_TAICHLayout, f_TAICHGreatestUpperLevel, ref l_Data);
                            }
                            else
                            {
                                ulong l_FileSize;

                                if (false == l_File.GetSize(out l_FileSize))
                                {
                                    return OpStatus.InputRejected;
                                }

                                l_TAICHAcc = new TAICHTreeAccessor(ref f_TAICHLayout, f_TAICHGreatestUpperLevel, c_TAICHAccBufferSize, l_File, 0, l_FileSize);
                            }

                            byte[] l_ExpectedTopHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx);
                            TAICHVerifyResult l_VerRes = l_TAICHAcc.VerifyTopHash(l_ExpectedTopHash);

                            if (TAICHVerifyResult.Valid != l_VerRes)
                            {
                                if (TAICHVerifyResult.AccessFailed != l_VerRes)
                                {
                                    BewareOfUser(l_UserNum, CorruptDataObserver.HashDelivery);
                                }

                                return OpStatus.InputRejected;
                            }

                            l_VerRes = TAICHTree.VerifyUpperTree(ref f_TAICHLayout, ref l_TAICHAcc, f_Cntxt.f_Tiger);

                            if (TAICHVerifyResult.Valid != l_VerRes)
                            {
                                if (TAICHVerifyResult.AccessFailed != l_VerRes)
                                {
                                    BewareOfUser(l_UserNum, CorruptDataObserver.HashDelivery);
                                }

                                return OpStatus.InputRejected;
                            }

                            l_Sanctions.AddSanction(l_UserNum, l_DeliveredSize, 0);
                            OpStatus l_Status = StoreTAICHTree(l_UserNum, l_DeliveredSize, a_Cntxt.f_Buffer, l_File);

                            if (OpStatus.Done == l_Status)
                            {
                                l_NeedRetry = false;

                                if (OpStatus.Done == LoadTAICHUpperTree())
                                {
                                    l_AccessorCreated = true;
                                }
                            }

                            return l_Status;
                        }

                    default: throw (new ApplicationException());
                }
            }
            finally
            {
                if (true == l_NeedRetry)
                {
                    AddHashDemand(ref l_Demand);
                }

                a_Cntxt.f_Cookie.ResetAfterHashDelivery();

                if (null != a_Cntxt.f_File)
                {
                    a_Cntxt.f_File.Close();
                }

                if (true == l_AccessorCreated)
                {
                    UpdateVerificationMode();
                }

                l_Sanctions.SendToUserManager(f_Inst);
            }
        }

        private OpStatus StoreeD2kHashSet(UserNum a_OriginUserNum,
                                          uint a_Size,
                                          byte[] a_Buffer,
                                          FHandle a_File)
        {
            HashHeader l_Header = new HashHeader(a_OriginUserNum, P2PHashSets.eD2kHashSet, a_Size);

            return StoreHashSet(ref l_Header, ref f_Main.f_eD2kHashSetKey, a_Buffer, a_File);
        }

        private OpStatus StoreTAICHTree(UserNum a_OriginUserNum,
                                        uint a_Size,
                                        byte[] a_Buffer,
                                        FHandle a_File)
        {
            HashHeader l_Header = new HashHeader(a_OriginUserNum, P2PHashSets.TAICHTree, a_Size);

            return StoreHashSet(ref l_Header, ref f_Main.f_TAICHUpperTreeKey, a_Buffer, a_File);
        }

        private OpStatus StoreHashSet(ref HashHeader a_Header,
                                      ref ElmKey a_ElmKeyInMain,
                                      byte[] a_Buffer,
                                      FHandle a_File)
        {
            // On failure any old key will be lost.

            // We can't use p_Buffer here, since the code path needs it.
            ByteArray l_HeaderData = new ByteArray(HashHeader.c_Size);
            a_Header.Write(f_Known.f_FileNum, ref f_HashStorage, ref f_EstimatedKnowledge, l_HeaderData.f_Arr, 0, true);

            ByteArray l_BodyData;

            if (null == a_Buffer)
            {
                // Though a_File contains the body, this structure
                // must indicate the size.
                l_BodyData = new ByteArray(null, 0, (int)a_Header.f_BodySize);
            }
            else
            {
                l_BodyData = new ByteArray(a_Buffer);
            }

            return UpdateUserOrHashSetElement(a_Header.GetMagic(), false, ref l_HeaderData, ref l_BodyData, a_File, ref a_ElmKeyInMain);
        }

        private void LoadAllHashSets()
        {
            bool l_Change = false;

            // Try to get the TAICH hash set first.

            if (false == HasUpperTAICHTree())
            {
                if (true == f_Main.f_TAICHUpperTreeKey.IsNonzero())
                {
                    // Don't stop on errors.

                    if (OpStatus.Done == LoadTAICHUpperTree())
                    {
                        l_Change = true;
                    }
                }

                if (false == HasUpperTAICHTree())
                {
                    HashDemand l_Demand = new HashDemand(HashDemandTypes.NeedTAICHUpperTree);

                    AddHashDemand(ref l_Demand);
                }
            }

            if (false == HaseD2kHashSet())
            {
                if (true == f_Main.f_eD2kHashSetKey.IsNonzero())
                {
                    // Don't stop on errors.

                    if (OpStatus.Done == LoadeD2kHashSet())
                    {
                        l_Change = true;
                    }
                }

                if (false == HaseD2kHashSet())
                {
                    HashDemand l_Demand = new HashDemand(HashDemandTypes.NeedeD2kHashSet);

                    AddHashDemand(ref l_Demand);
                }
            }

            if (true == l_Change)
            {
                // NOTE: We may not do the following update before
                // we tried to read both hash sets, otherwise we
                // might trigger an useless reverification.

                UpdateVerificationMode();
            }
        }

        private OpStatus LoadeD2kHashSet()
        {
            // This code path uses p_Buffer.

            OpStatus l_Status = LoadeD2kHashSetCore();

            if (OpStatus.Corruption == l_Status)
            {
                f_Main.f_eD2kHashSetKey = new ElmKey();
            }

            return l_Status;
        }

        private OpStatus LoadeD2kHashSetCore()
        {
            // This routine should only get called by LoadeD2kHashSet().

            HardMemory.Element l_Elm;
            HashHeader l_HashHeader;
            OpStatus l_Status = LoadHashSet(f_Main.f_eD2kHashSetKey, out l_Elm, out l_HashHeader);

            if (OpStatus.Done != l_Status)
            {
                return l_Status;
            }

            if (((ulong)f_eD2kHashSetElementCount * MD_4.c_HashByteSize) != l_HashHeader.f_BodySize)
            {
                return OpStatus.Corruption;
            }

            if (c_eD2kAccBufferSize >= l_HashHeader.f_BodySize)
            {
                bool l_Succ;
                ByteArray l_Data = new ByteArray((int)l_HashHeader.f_BodySize);

                f_Resource.AcquireShared();
                {
                    l_Succ = f_IncomplMem.Read(l_Elm, HashHeader.c_BodyStartingOffset, ref l_Data);
                }
                f_Resource.ReleaseShared();

                if (false == l_Succ)
                {
                    return OpStatus.IoError;
                }

                f_eD2kAccessor = new eD2kHashSetAccessor(f_eD2kHashSetElementCount, ref l_Data);
            }
            else
            {
                f_eD2kAccessor = new eD2kHashSetAccessor(f_eD2kHashSetElementCount, f_IncomplMem, l_Elm, HashHeader.c_BodyStartingOffset);
            }

            return OpStatus.Done;
        }

        private OpStatus LoadTAICHUpperTree()
        {
            // This code path uses p_Buffer.

            OpStatus l_Status = LoadTAICHUpperTreeCore();

            if (OpStatus.Corruption == l_Status)
            {
                f_Main.f_TAICHUpperTreeKey = new ElmKey();
            }

            return l_Status;
        }

        private OpStatus LoadTAICHUpperTreeCore()
        {
            // This routine should only get called by LoadTAICHUpperTree().

            HardMemory.Element l_Elm;
            HashHeader l_HashHeader;
            OpStatus l_Status = LoadHashSet(f_Main.f_TAICHUpperTreeKey, out l_Elm, out l_HashHeader);

            if (OpStatus.Done != l_Status)
            {
                return l_Status;
            }

            uint l_Junk0;
            uint l_HashCount;
            int l_Junk1 = f_TAICHLayout.GetUpperTreeBoundaries(c_TAICHExponent, out l_Junk0, out l_HashCount);

            if (((ulong)l_HashCount * Tiger.c_HashByteSize) != l_HashHeader.f_BodySize)
            {
                return OpStatus.Corruption;
            }

            f_TAICHAccessor = new TAICHTreeAccessor(ref f_TAICHLayout, f_TAICHGreatestUpperLevel, f_IncomplMem, l_Elm, HashHeader.c_BodyStartingOffset);

            return OpStatus.Done;
        }

        private OpStatus LoadHashSet(ElmKey a_ElmKey,
                                     out HardMemory.Element a_Elm,
                                     out HashHeader a_Header)
        {
            OpStatus l_Status;

            f_Resource.AcquireShared();
            {
                l_Status = LoadHashSetCore(a_ElmKey, out a_Elm, out a_Header);
            }
            f_Resource.ReleaseShared();

            return l_Status;
        }

        private OpStatus LoadHashSetCore(ElmKey a_ElmKey,
                                         out HardMemory.Element a_Elm,
                                         out HashHeader a_Header)
        {
            // This routine may only be called by LoadHashSet().

            OpStatus l_Status;

            a_Elm = f_IncomplMem.GetElement(a_ElmKey);

            if (null == a_Elm ||
                HashHeader.c_Size > a_Elm.f_MaxSize)
            {
                l_Status = OpStatus.Corruption;

                goto CLEAR_AND_EXIT;
            }

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Data = new ByteArray(p_Buffer, 0, HashHeader.c_Size);
                bool l_Read = f_IncomplMem.Read(a_Elm, 0, ref l_Data);

                if (false == l_Read ||
                    false == HashHeader.Load(p_Buffer, 0, out a_Header) ||
                    false == CheckDefaultHeader(a_Header.GetMagic()) ||
                    (HashHeader.c_Size + a_Header.f_BodySize) != a_Elm.f_MaxSize)
                {
                    f_Cntxt.DebugReleaseBuffers();

                    l_Status = GetLoadErrorStatus(l_Read);

                    goto CLEAR_AND_EXIT;
                }
            }
            f_Cntxt.DebugReleaseBuffers();

            return OpStatus.Done;

            CLEAR_AND_EXIT:

            // NOTE: We deliberately do not delete a corrupt element.

            a_Header = new HashHeader();

            return l_Status;
        }

        public OpStatus BuildFileDescriptor(FileDescriptorCookie a_Cookie)
        {
            MasterHashes l_MasterHashes;
            P2PHashSets l_AvailHashSets = GetAvailHashSets(out l_MasterHashes);

            if (false == l_MasterHashes.p_HaseD2kHash)
            {
                a_Cookie.Fail();

                return OpStatus.Done;
            }

            MemoryIO l_Buffer = MemoryIO.CreateWritable();
            l_Buffer.WriteBytes(f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_eD2kHashIdx));
            l_Buffer.WriteUInt64LE(f_Known.f_FileSize);
            l_Buffer.WriteUInt16LE((ushort)l_AvailHashSets);
            l_MasterHashes.ToBuffer(l_Buffer);

            a_Cookie.Succeed(l_Buffer.ToArray(true));

            return OpStatus.Done;
        }

        public OpStatus HandleHashQuery(HashQueryCookie a_Cookie)
        {
            byte[] l_Bytes = null;
            bool l_MalformedInput = false;

            if (HashQueryKinds.Range == a_Cookie.f_Kind)
            {
                byte[] l_MasterHash = null;
                bool l_OnlyMasterHash = false;

                if (P2PHashSets.eD2kHashSet == a_Cookie.f_HashType)
                {
                    if (true == f_EstimatedKnowledge.p_HaseD2kHash)
                    {
                        l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_eD2kHashIdx);
                        l_OnlyMasterHash = (0 == f_eD2kHashSetElementCount);
                    }
                }
                else if (P2PHashSets.TAICHTree == a_Cookie.f_HashType)
                {
                    if (true == f_EstimatedKnowledge.p_HasTAICHHash &&
                        0 < f_TAICHGreatestUpperLevel)
                    {
                        l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx);
                    }
                }

                if (null != l_MasterHash &&
                    true == Buffer2.ValueEquals(l_MasterHash, a_Cookie.f_MasterHash))
                {
                    if (false == l_OnlyMasterHash)
                    {
                        if (P2PHashSets.eD2kHashSet == a_Cookie.f_HashType)
                        {
                            f_eD2kAccessor.AcquireSharedIfDiskBased();
                            {
                                l_Bytes = f_eD2kAccessor.CheckInputAndBuildHashArray(a_Cookie.f_StartingIndex, a_Cookie.f_Count, out l_MalformedInput);
                            }
                            f_eD2kAccessor.ReleaseSharedIfDiskBased();
                        }
                        else
                        {
                            f_TAICHAccessor.AcquireSharedIfDiskBased();
                            {
                                l_Bytes = f_TAICHAccessor.CheckInputAndBuildHashArray(a_Cookie.f_StartingIndex, a_Cookie.f_Count, out l_MalformedInput);
                            }
                            f_TAICHAccessor.ReleaseSharedIfDiskBased();
                        }
                    }
                    else
                    {
                        if (0 == a_Cookie.f_StartingIndex &&
                            1 == a_Cookie.f_Count)
                        {
                            l_Bytes = Util.CloneArray(l_MasterHash);
                        }
                        else
                        {
                            l_MalformedInput = true;
                        }
                    }
                }
            }

            if (null != l_Bytes)
            {
                a_Cookie.Succeed(l_Bytes);

                return OpStatus.Done;
            }

            // We are not ready to deliver the requested data.

            a_Cookie.Fail(l_MalformedInput);

            return OpStatus.NoMatch;
        }

#endregion

#region BitmapHandling

        private bool ReadWholeBitmap()
        {
            return ReadBitmap(0, f_DoneMap.CalcByteCount());
        }

        private bool ReadBitmap(uint a_ByteIndex,
                                uint a_ByteCount)
        {
            bool l_Succ;

            f_Cntxt.DebugAcquireBuffers();
            {
                byte[] l_Buf = (c_BufferSize >= a_ByteCount ? p_Buffer : new byte[a_ByteCount]);
                ByteArray l_Data = new ByteArray(l_Buf, 0, (int)a_ByteCount);
                l_Succ = f_IncomplMem.Read(f_MainElm, MainHeader.c_BitmapStartingOffset + a_ByteIndex, ref l_Data);

                if (true == l_Succ)
                {
                    f_DoneMap.FromBinary(ref l_Data, (int)a_ByteIndex);
                }
            }
            f_Cntxt.DebugReleaseBuffers();

            return l_Succ;
        }

        private bool WriteBitmap(uint a_ByteIndex,
                                 uint a_ByteCount)
        {
            bool l_Succ;

            f_Cntxt.DebugAcquireBuffers();
            {
                byte[] l_Buf = (c_BufferSize >= a_ByteCount ? p_Buffer : new byte[a_ByteCount]);
                ByteArray l_Data = new ByteArray(l_Buf, 0, (int)a_ByteCount);
                f_DoneMap.CopyToBuffer(a_ByteIndex, a_ByteCount, ref l_Buf, 0);

                l_Succ = f_IncomplMem.Write(f_MainElm, MainHeader.c_BitmapStartingOffset + a_ByteIndex, ref l_Data);
            }
            f_Cntxt.DebugReleaseBuffers();

            return l_Succ;
        }

        public bool UpdateBitAndWrite(uint a_ChunkIndex,
                                      bool a_State)
        {
            // In case of failure, nothing will have changed.

            bool l_OldState = f_DoneMap.Set(a_ChunkIndex, a_State);

            if (l_OldState != a_State)
            {
                uint l_AffectedByte = (a_ChunkIndex / Util.c_BitsPerByte);

                if (false == WriteBitmap(l_AffectedByte, 1))
                {
                    f_DoneMap.Set(a_ChunkIndex, l_OldState);

                    return false;
                }

                f_MapRevision++;
            }

            return true;
        }

        private bool WriteChunkKey(uint a_ChunkIndex,
                                   ElmKey a_Key)
        {
            bool l_Succ;

            f_Cntxt.DebugAcquireBuffers();
            {
                Util.WriteUInt64LE(a_Key.f_Val, p_Buffer, 0);
                uint l_Offset = (f_ChunkKeyFieldOffset + (a_ChunkIndex * sizeof(ulong)));
                ByteArray l_Src = new ByteArray(p_Buffer, 0, sizeof(ulong));

                l_Succ = f_IncomplMem.Write(f_MainElm, l_Offset, ref l_Src);
            }
            f_Cntxt.DebugReleaseBuffers();

            return l_Succ;
        }

        private bool ReadChunkKeysIntoByteBuffer(uint a_StartingChunkIndex,
                                                 ref ByteArray a_Dest)
        {
            uint l_Offset = (f_ChunkKeyFieldOffset + (a_StartingChunkIndex * sizeof(ulong)));

            return f_IncomplMem.Read(f_MainElm, l_Offset, ref a_Dest);
        }

        private OpStatus ReadChunkKeys(uint a_StartingChunkIndex,
                                       uint a_Count,
                                       ulong[] a_DestKeyBuffer)
        {
            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Dest = new ByteArray(p_Buffer, 0, (int)(a_Count * sizeof(ulong)));

                if (false == ReadChunkKeysIntoByteBuffer(a_StartingChunkIndex, ref l_Dest))
                {
                    f_Cntxt.DebugReleaseBuffers();

                    return OpStatus.IoError;
                }

                Buffer.BlockCopy(p_Buffer, 0, a_DestKeyBuffer, 0, l_Dest.f_Len);
            }
            f_Cntxt.DebugReleaseBuffers();

            if (1 == a_Count)
            {
                UpdateHotIncomplChunksIfEntryExists(a_StartingChunkIndex, new ElmKey(a_DestKeyBuffer[0]));
            }

            return OpStatus.Done;
        }

        private OpStatus ReadSingleChunkKey(uint a_ChunkIndex,
                                            out ElmKey a_ElmKey)
        {
            // Returns OpStatus.Done also if the read key is zero.

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Dest = new ByteArray(p_Buffer, 0, sizeof(ulong));

                if (false == ReadChunkKeysIntoByteBuffer(a_ChunkIndex, ref l_Dest))
                {
                    f_Cntxt.DebugReleaseBuffers();

                    a_ElmKey = new ElmKey();

                    return OpStatus.IoError;
                }

                a_ElmKey = new ElmKey(Util.ReadUInt64LE(l_Dest.f_Arr, 0));
            }
            f_Cntxt.DebugReleaseBuffers();

            UpdateHotIncomplChunksIfEntryExists(a_ChunkIndex, a_ElmKey);

            return OpStatus.Done;
        }

        private ElmKey GetHotIncomplChunksEntry(uint a_ChunkIndex)
        {
            // Returns a zero key if there is no such entry.

            ElmKey l_ElmKey;
            if (true == f_HotIncomplChunks.TryGetValue(a_ChunkIndex, out l_ElmKey))
            {
                return l_ElmKey;
            }

            return new ElmKey();
        }

        private void CreateNewHotIncomplChunksEntryIfNoneExists(uint a_ChunkIndex)
        {
            if (false == f_HotIncomplChunks.ContainsKey(a_ChunkIndex))
            {
                f_HotIncomplChunks.Add(a_ChunkIndex, new ElmKey());
                f_HotIncomplRevision++;
            }
        }

        private void UpdateHotIncomplChunksIfEntryExists(uint a_ChunkIndex,
                                                         ElmKey a_NewElmKey)
        {
            // f_HotIncomplChunks[a_ChunkIndex] if it exists already.

            ElmKey l_OldElmKey;
            if (true == f_HotIncomplChunks.TryGetValue(a_ChunkIndex, out l_OldElmKey))
            {
                if (l_OldElmKey != a_NewElmKey)
                {
                    f_HotIncomplChunks[a_ChunkIndex] = a_NewElmKey;
                    f_HotIncomplRevision++;
                }
            }
        }

#endregion

#region ChunkIO

        public IncomplFileProgress GetProgress()
        {
            if (true == f_DoneMap.p_AllSet &&
                false == f_NotifiedCompletion)
            {
                GeneralAPCCookie l_Cookie = f_Inst.f_ContentMngr.IncomplNotifyFinishedAsync(f_Known);
                f_NotifiedCompletion = true;
            }

            ulong l_Completed = 0;
            uint l_SetBits = f_DoneMap.p_SetBitCount;

            if (true == f_DoneMap.Get(f_DoneMap.f_BitCount - 1))
            {
                // The last chunk is special.

                uint l_Remainder = (uint)(f_Known.f_FileSize % c_ChunkSize);

                if (0 != l_Remainder)
                {
                    l_SetBits--;
                    l_Completed += l_Remainder;
                }
            }

            l_Completed += ((ulong)l_SetBits * c_ChunkSize);
            ulong l_MaxHalfCompleted = (f_Known.f_FileSize - l_Completed);
            ulong l_HalfCompleted = f_Main.f_BytesInIncomplChunks;

            if (l_MaxHalfCompleted < l_HalfCompleted)
            {
                // There is something wrong with our count. This can
                // happen if the application wasn't shut down properly
                // in the past.

                f_Main.f_BytesInIncomplChunks = l_MaxHalfCompleted;
                l_HalfCompleted = l_MaxHalfCompleted;
            }

            Dictionary<uint, ElmKey> l_HotChunksClone = (0 == f_HotIncomplChunks.Count ? null : new Dictionary<uint, ElmKey>(f_HotIncomplChunks));

            return new IncomplFileProgress(f_MapRevision, f_HotIncomplRevision, f_DoneMap, l_HotChunksClone, l_Completed, l_Completed + l_HalfCompleted);
        }

        private void AddIncomplChunkBytes(ref IncomplChunkHeader a_Header)
        {
            // Call this routine if the header was just written to
            // disk and after the covered range of the header might
            // have changed.

            uint l_OldVal = a_Header.GetCoveredBytes();
            a_Header.UpdateCoveredBytes();
            uint l_NewVal = a_Header.GetCoveredBytes();

            AddIncomplChunkBytes((int)l_NewVal - (int)l_OldVal);
        }

        private void AddIncomplChunkBytes(int a_Diff)
        {
            if (0 < a_Diff)
            {
                f_Main.f_BytesInIncomplChunks += (uint)a_Diff;
            }
            else if (0 > a_Diff)
            {
                f_Main.f_BytesInIncomplChunks -= Math.Min(f_Main.f_BytesInIncomplChunks, (uint)-a_Diff);
            }
            else
            {
                return;
            }

            bool l_Written = WriteBytesInIncomplChunks();
        }

        private OpStatus ReadComplChunkHeader(uint a_ChunkIndex,
                                              ElmKey a_ElmKey,
                                              out ComplChunkHeader a_Header)
        {
            OpStatus l_Status;

            if (VerificationMode.None == f_VerificationMode)
            {
                l_Status = OpStatus.NeedReverification;

                goto CLEAR_AND_EXIT;
            }

            HardMemory.Element l_Elm = f_IncomplMem.GetElement(a_ElmKey);

            if (null == l_Elm ||
                (ComplChunkHeader.c_BodyStartingOffset + GetChunkSize(a_ChunkIndex)) != l_Elm.f_MaxSize)
            {
                l_Status = OpStatus.Corruption;

                goto CLEAR_AND_EXIT;
            }

            uint l_ChunkSize = GetChunkSize(a_ChunkIndex);

            f_Cntxt.DebugAcquireBuffers();
            {
                ByteArray l_Data = new ByteArray(p_Buffer, 0, ComplChunkHeader.c_Size);
                bool l_Read = f_IncomplMem.Read(l_Elm, 0, ref l_Data);

                if (false == l_Read ||
                    false == CheckDefaultChunkHeader(c_ComplChunkMagic, a_ChunkIndex))
                {
                    f_Cntxt.DebugReleaseBuffers();

                    l_Status = GetLoadErrorStatus(l_Read);

                    goto CLEAR_AND_EXIT;
                }

                P2PHashSets l_ExpectedHashSet;
                byte[] l_ExpectedMasterHash = GetExpectedMasterHashForVerification(out l_ExpectedHashSet);

                l_Status = ComplChunkHeader.Load(a_ChunkIndex, l_ChunkSize, l_ExpectedHashSet, l_ExpectedMasterHash, p_Buffer, 0, out a_Header);
            }
            f_Cntxt.DebugReleaseBuffers();

            if (OpStatus.Done == l_Status)
            {
                a_Header.SetElement(l_Elm);

                return OpStatus.Done;
            }

            if (OpStatus.NeedReverification == l_Status)
            {
                // Turn the complete chunk into an incomplete chunk again.

                if (false == f_ComplChunkReverInProgress)
                {
                    StartReverification();
                }

                IncomplChunkHeader l_IncomplHeader = new IncomplChunkHeader(a_ChunkIndex, l_ChunkSize);
                l_IncomplHeader.SetElement(l_Elm);

                l_IncomplHeader.f_Users.ReserveAt(0);
                UserHeader[] l_Users = l_IncomplHeader.f_Users.p_Arr;
                l_Users[0] = new UserHeader(UserNum.s_Anonymous);
                l_Users[0].SetSlot(0);
                l_Users[0].f_LastUpdateTime = f_Inst.GetTime();
                ChunkRange l_Range = new ChunkRange(0, l_ChunkSize);
                PieceBounds.CreateFromChunkRange(ref l_Range, out l_Users[0].f_Bounds);
                l_Users[0].f_DirtyMap.TakeOver(l_IncomplHeader.f_BlockMask);

                f_Main.f_BytesInIncomplChunks += l_ChunkSize;

                bool l_Written;

                f_Cntxt.DebugAcquireBuffers();
                {
                    l_Written = l_IncomplHeader.Update(f_Known.f_FileNum, ref f_HashStorage, ref f_EstimatedKnowledge, f_IncomplMem, p_Buffer);
                }
                f_Cntxt.DebugReleaseBuffers();

                if (false == l_Written ||
                    false == UpdateBitAndWrite(a_ChunkIndex, false) ||
                    false == WriteBytesInIncomplChunks() ||
                    false == f_IncomplMem.Flush())
                {
                    l_Status = OpStatus.IoError;
                }
                else
                {
                    if (null == f_RecheckChunks)
                    {
                        f_RecheckChunks = new List<uint>(1);
                    }

                    f_RecheckChunks.Add(a_ChunkIndex);
                }
            }

            return l_Status;

            CLEAR_AND_EXIT:

            a_Header = new ComplChunkHeader();

            return l_Status;
        }

        private OpStatus ReadComplChunk(ref ComplChunkHeader a_Header,
                                        uint a_Offset,
                                        ref ByteArray a_Dest)
        {
            if (false == f_IncomplMem.Read(a_Header.p_Element, ComplChunkHeader.c_BodyStartingOffset + a_Offset, ref a_Dest))
            {
                return OpStatus.IoError;
            }

            if (0 == a_Offset &&
                a_Header.f_ChunkSize == a_Dest.f_Len)
            {
                // Since we read the entire chunk, do an integrity
                // check.

                uint l_Expected = Adler32.ComputeHash(ref a_Dest);

                if (l_Expected != a_Header.f_ContentAdler32)
                {
                    return OpStatus.Corruption;
                }
            }

            return OpStatus.Done;
        }

        private OpStatus ReadIncomplChunkHeader(uint a_ChunkIndex,
                                                ElmKey a_ElmKey,
                                                bool a_Full,
                                                out IncomplChunkHeader a_Header)
        {
            OpStatus l_Status;
            HardMemory.Element l_Elm = f_IncomplMem.GetElement(a_ElmKey);

            if (null == l_Elm ||
                IncomplChunkHeader.c_Size > l_Elm.f_MaxSize)
            {
                l_Status = OpStatus.Corruption;

                goto CLEAR_AND_EXIT;
            }

            f_Cntxt.DebugAcquireBuffers();
            {
                int l_Size = (false == a_Full ? IncomplChunkHeader.c_ShortSize : IncomplChunkHeader.c_Size);
                ByteArray l_Data = new ByteArray(p_Buffer, 0, l_Size);
                bool l_Read = f_IncomplMem.Read(l_Elm, 0, ref l_Data);

                if (false == l_Read ||
                    false == CheckDefaultChunkHeader(c_IncomplChunkMagic, a_ChunkIndex) ||
                    false == IncomplChunkHeader.Load(a_ChunkIndex,
                                                     GetChunkSize(a_ChunkIndex),
                                                     l_Elm,
                                                     ref f_TAICHLayout,
                                                     ref f_HashStorage,
                                                     ref f_EstimatedKnowledge,
                                                     a_Full,
                                                     f_TAICHGreatestUpperLevel,
                                                     p_Buffer,
                                                     0,
                                                     out a_Header) ||
                    c_MaxUnmergedUsersPerIncomplChunk < a_Header.f_Users.p_Count)
                {
                    f_Cntxt.DebugReleaseBuffers();

                    l_Status = GetLoadErrorStatus(l_Read);

                    goto CLEAR_AND_EXIT;
                }
            }
            f_Cntxt.DebugReleaseBuffers();

            return OpStatus.Done;

            CLEAR_AND_EXIT:

            a_Header = new IncomplChunkHeader();

            return l_Status;
        }

        private OpStatus ReadIncomplChunkHeaderWithKey(uint a_ChunkIndex,
                                                       ElmKey a_ElmKeyHint,
                                                       bool a_Full,
                                                       out IncomplChunkHeader a_Header)
        {
            // a_ElmKeyHint can be zero, in which case the key will
            // be looked up.

            ElmKey l_ElmKey = a_ElmKeyHint;

            if (true == l_ElmKey.IsZero())
            {
                OpStatus l_KeyStatus = ReadSingleChunkKey(a_ChunkIndex, out l_ElmKey);

                if (OpStatus.Done != l_KeyStatus)
                {
                    a_Header = new IncomplChunkHeader();

                    return l_KeyStatus;
                }
            }

            if (true == l_ElmKey.IsZero())
            {
                a_Header = new IncomplChunkHeader();

                return OpStatus.ChunkHasNoKey;
            }

            OpStatus l_Status = ReadIncomplChunkHeader(a_ChunkIndex, l_ElmKey, a_Full, out a_Header);

            if (OpStatus.Corruption == l_Status)
            {
                OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(a_ChunkIndex, l_ElmKey);
            }

            return l_Status;
        }

        public OpStatus Read(ulong a_FilePos,
                             ref ByteArray a_Dest)
        {
            if (0 == a_Dest.f_Len)
            {
                return OpStatus.Error;
            }

            ulong l_FinalPos = (a_FilePos + (uint)a_Dest.f_Len);
            uint l_ChunkLen;
            uint l_StartingChunkIdx = f_DoneMap.FilePosToIndex(a_FilePos, l_FinalPos, out l_ChunkLen);

            if (false == f_DoneMap.IsRunSet(l_StartingChunkIdx, l_ChunkLen))
            {
                return OpStatus.NotEnoughData;
            }

            if (c_MaxChunkKeysPerBuffer < l_ChunkLen)
            {
                return OpStatus.RequestTooLarge;
            }

            OpStatus l_Status;
            uint l_Offset = (uint)(a_FilePos % c_ChunkSize);
            int l_DestIdx = a_Dest.f_Idx;
            int l_LeftLen = a_Dest.f_Len;

            // We can't use f_KeyBuffer in this routine, since our
            // codepath uses the buffers too.
            ulong[] l_LocalKeyBuffer = f_Cntxt.CreateKeyBuffer(l_ChunkLen);

            f_Resource.AcquireShared();
            {
                l_Status = ReadChunkKeys(l_StartingChunkIdx, l_ChunkLen, l_LocalKeyBuffer);

                if (OpStatus.Done == l_Status)
                {
                    for (uint i = 0, l_ChunkIndex = l_StartingChunkIdx; i < l_ChunkLen; i++, l_ChunkIndex++)
                    {
                        ElmKey l_ElmKey = new ElmKey(l_LocalKeyBuffer[i]);

                        if (false == l_ElmKey.IsNonzero())
                        {
                            l_Status = OpStatus.Corruption;
                        }
                        else
                        {
                            ComplChunkHeader l_Header;
                            l_Status = ReadComplChunkHeader(l_ChunkIndex, l_ElmKey, out l_Header);

                            if (OpStatus.Done == l_Status)
                            {
                                int l_CurLen = Math.Min(l_LeftLen, (int)(l_Header.f_ChunkSize - l_Offset));
                                ByteArray l_Data = new ByteArray(a_Dest.f_Arr, l_DestIdx, l_CurLen);

                                l_Status = ReadComplChunk(ref l_Header, l_Offset, ref l_Data);

                                l_Offset = 0;
                                l_DestIdx += l_CurLen;
                                l_LeftLen -= l_CurLen;
                            }
                        }

                        if (OpStatus.Corruption == l_Status)
                        {
                            OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(l_ChunkIndex, l_ElmKey);
                        }

                        if (OpStatus.Done != l_Status)
                        {
                            break;
                        }
                    }
                }
            }
            f_Resource.ReleaseShared();

            return l_Status;
        }

        public OpStatus Write(ref WriteContext a_Cntxt)
        {
            ObtainCookie l_Cookie = a_Cntxt.f_Cookie;

            if (CookieStates.RangeSet != l_Cookie.p_ContentMngrState)
            {
                // It can always happen that the obtain cookie got
                // invalid while the data was on its way too us.

                return OpStatus.InputRejected;
            }

            uint l_InChunkOffset;
            ByteArray l_ReadyMadeData;
            OpStatus l_Status = AlignInput(l_Cookie, ref a_Cntxt.f_Data, out l_InChunkOffset, out l_ReadyMadeData);

            if (OpStatus.Done != l_Status)
            {
                if (OpStatus.NotEnoughData == l_Status)
                {
                    // We've buffered the input.

                    return OpStatus.Done;
                }

                Util.DebugBreak();

                RejectAndRemoveFromObtainList(l_Cookie);

                return l_Status;
            }

            IncomplChunkHeader l_Header;
            ElmKey l_ElmKey = f_HotIncomplChunks[l_Cookie.p_ChunkIndex];
            UpdateFileChunkInputContext l_InputCntxt;
            UpdateFileChunkInputContext.CreateDefault(l_Cookie.p_UserNum, l_InChunkOffset, ref l_ReadyMadeData, out l_InputCntxt);

            f_Resource.AcquireExclusive();
            {
                l_Status = UpdateFileChunk(l_Cookie.p_ChunkIndex,
                                           l_ElmKey,
                                           ref l_InputCntxt,
                                           out l_Header);
            }
            f_Resource.ReleaseExclusive();

            if (OpStatus.Done != l_Status)
            {
                RejectAndRemoveFromObtainList(l_Cookie);

                return l_Status;
            }

            l_Cookie.f_BytesToGo -= (uint)l_ReadyMadeData.f_Len;

            if (0 == l_Cookie.f_BytesToGo)
            {
                FinishAndRemoveFromObtainList(l_Cookie);
            }

            OpStatus l_VerStatus = VerifyChunk(ref l_Header);

            return OpStatus.Done;
        }

        private OpStatus AlignInput(ObtainCookie a_Cookie,
                                    ref ByteArray a_InData,
                                    out uint a_InChunkOffset,
                                    out ByteArray a_OutData)
        {
            // Makes sure input is aligned to the ranges specified in
            // a_Cookie (this can avoid loss of existing data in some
            // cases, due to the nature of the PieceBounds structure).

            if (0 >= a_InData.f_Len)
            {
                goto FAIL_NOT_ENOUGH;
            }

            int l_LeftoverBytes = (null == a_Cookie.f_Buffer ? 0 : a_Cookie.f_Buffer.Length);
            int l_ReqBytes = ((int)a_Cookie.f_Ranges.p_Arr[a_Cookie.f_Index].f_Size - l_LeftoverBytes);

            bool l_CopyToOwnBuffer;
            int l_EffInLen = Math.Min(a_InData.f_Len, l_ReqBytes);

            if (a_InData.f_Len == l_ReqBytes)
            {
                l_CopyToOwnBuffer = (0 != l_LeftoverBytes);
            }
            else
            {
                l_CopyToOwnBuffer = true;
            }

            if (true == l_CopyToOwnBuffer)
            {
                Array.Resize(ref a_Cookie.f_Buffer, l_LeftoverBytes + l_EffInLen);

                Buffer.BlockCopy(a_InData.f_Arr, a_InData.f_Idx, a_Cookie.f_Buffer, l_LeftoverBytes, l_EffInLen);
            }

            if (l_EffInLen < l_ReqBytes)
            {
                goto FAIL_NOT_ENOUGH;
            }

            a_InChunkOffset = a_Cookie.f_Ranges.p_Arr[a_Cookie.f_Index].f_Offset;

            if (false == l_CopyToOwnBuffer)
            {
                a_OutData = a_InData;
            }
            else
            {
                a_OutData = new ByteArray(a_Cookie.f_Buffer);
                a_Cookie.f_Buffer = null;
            }

            a_Cookie.f_Index++;

            return OpStatus.Done;

            FAIL_NOT_ENOUGH:

            a_InChunkOffset = 0;
            a_OutData = new ByteArray();

            return OpStatus.NotEnoughData;
        }

        private OpStatus UpdateFileChunk(uint a_ChunkIndex,
                                         ElmKey a_ElmKey,
                                         ref UpdateFileChunkInputContext a_Cntxt,
                                         out IncomplChunkHeader a_Header)
        {
            // a_ElmKey may be zero.

            OpStatus l_Status = ReadIncomplChunkHeaderWithKey(a_ChunkIndex, a_ElmKey, true, out a_Header);
            ElmKey l_CreatedElmKey;

            if (OpStatus.Done == l_Status)
            {
                l_CreatedElmKey = new ElmKey();
            }
            else
            {
                if (OpStatus.ChunkHasNoKey != l_Status &&
                    OpStatus.Corruption != l_Status)
                {
                    return l_Status;
                }

                a_Header = new IncomplChunkHeader(a_ChunkIndex, GetChunkSize(a_ChunkIndex));

                l_CreatedElmKey = FindUnoccupiedKey();
                uint l_InitialSize = a_Header.SuggestElementSize(1);
                HardMemory.Element l_Elm = f_IncomplMem.Allocate(l_CreatedElmKey, l_InitialSize);

                if (null == l_Elm)
                {
                    return OpStatus.InsuffDiskSpace;
                }

                a_Header.SetElement(l_Elm);
            }

            l_Status = UpdateFileChunkCore(ref a_Cntxt, ref a_Header);

            if (OpStatus.Done == l_Status &&
                true == l_CreatedElmKey.IsNonzero() &&
                false == WriteChunkKey(a_ChunkIndex, l_CreatedElmKey))
            {
                l_Status = OpStatus.IoError;
            }

            if (OpStatus.Done == l_Status)
            {
                AddIncomplChunkBytes(ref a_Header);
            }
            else if (true == l_CreatedElmKey.IsNonzero())
            {
                bool l_Deleted = f_IncomplMem.Delete(l_CreatedElmKey);
            }

            return l_Status;
        }

        private OpStatus UpdateFileChunkCore(ref UpdateFileChunkInputContext a_Cntxt,
                                             ref IncomplChunkHeader a_Header)
        {
            UserHeader[] l_Users;
            int l_HeaderIdx;
            int l_Slot;
            bool l_NeedToRelocateAfterResizing = false;
            uint l_NewElmSize;

            if (false == a_Header.GetHeaderIndex(a_Cntxt.f_SrcUserNum, out l_HeaderIdx))
            {
                // We don't know this user yet.

                if (c_MaxUnmergedUsersPerIncomplChunk <= a_Header.f_Users.p_Count)
                {
                    l_HeaderIdx = a_Header.FindBestUserHeaderIndexForSubstitution();
                }
                else
                {
                    l_HeaderIdx = a_Header.f_Users.ReserveAtEnd();
                }

                l_Users = a_Header.f_Users.p_Arr;
                l_Users[l_HeaderIdx] = new UserHeader(a_Cntxt.f_SrcUserNum);
                a_Cntxt.RetrieveBounds(false, ref l_Users[l_HeaderIdx].f_Bounds);

                l_Slot = a_Header.SuggestSlot(ref l_Users[l_HeaderIdx].f_Bounds, l_HeaderIdx);
                l_Users[l_HeaderIdx].SetSlot(l_Slot);

                l_NewElmSize = a_Header.SuggestElementSize();
            }
            else
            {
                // This user already provided data.

                l_Users = a_Header.f_Users.p_Arr;
                a_Cntxt.RetrieveBounds(true, ref l_Users[l_HeaderIdx].f_Bounds);

                l_Slot = a_Header.SuggestSlot(ref l_Users[l_HeaderIdx].f_Bounds, l_HeaderIdx);
                int l_OldSlot = l_Users[l_HeaderIdx].p_Slot;

                if (l_OldSlot < l_Slot)
                {
                    l_NeedToRelocateAfterResizing = true;

                    l_NewElmSize = a_Header.SuggestElementSize(l_Slot + 1);
                }
                else
                {
                    if (l_OldSlot > l_Slot)
                    {
                        byte[] l_Buf = new byte[a_Header.f_ChunkSize];

                        if (false == a_Header.MoveBody(f_IncomplMem, ref l_Users[l_HeaderIdx], l_Slot, l_Buf))
                        {
                            return OpStatus.IoError;
                        }
                    }

                    l_NewElmSize = a_Header.SuggestElementSize();
                }
            }

            OpStatus l_ResizeStatus = a_Header.ResizeElement(f_IncomplMem, l_NewElmSize);

            if (OpStatus.Done != l_ResizeStatus)
            {
                return l_ResizeStatus;
            }

            if (true == l_NeedToRelocateAfterResizing)
            {
                byte[] l_Buf = new byte[a_Header.f_ChunkSize];

                if (false == a_Header.MoveBody(f_IncomplMem, ref l_Users[l_HeaderIdx], l_Slot, l_Buf))
                {
                    return OpStatus.IoError;
                }
            }

            bool l_BodyWritten;
            uint l_Mask;

            if (false == a_Cntxt.p_Import)
            {
                l_BodyWritten = a_Header.WriteBody(f_IncomplMem, l_Slot, a_Cntxt.f_Range.f_Offset, ref a_Cntxt.f_Data);

                uint l_BlockIdx;
                uint l_Count = BlockBitmap.CalcBlockBoundsWide(a_Cntxt.f_Range.f_Offset, a_Cntxt.f_Range.p_Final, out l_BlockIdx);
                l_Mask = BlockBitmap.BuildBlockMask(l_BlockIdx, l_Count);
            }
            else
            {
                l_BodyWritten = a_Header.WriteBody(f_IncomplMem, l_Slot, ref a_Cntxt.f_Bounds, a_Cntxt.f_Range.f_Offset, ref a_Cntxt.f_Data);

                l_Mask = a_Cntxt.f_Bounds.ToBitmap(false).p_Value;
            }

            if (false == l_BodyWritten ||
                false == f_IncomplMem.Flush())
            {
                return OpStatus.IoError;
            }

            l_Users[l_HeaderIdx].MarkAsDirty(l_Mask);
            l_Users[l_HeaderIdx].UpdateTime(f_Inst);

            bool l_Updated;

            f_Cntxt.DebugAcquireBuffers();
            {
                l_Updated = a_Header.Update(f_Known.f_FileNum, ref f_HashStorage, ref f_EstimatedKnowledge, f_IncomplMem, p_Buffer);
            }
            f_Cntxt.DebugReleaseBuffers();

            return (false == l_Updated ? OpStatus.IoError : OpStatus.Done);
        }

        private OpStatus CompactIncomplChunk(ref IncomplChunkHeader a_Header)
        {
            bool l_Change = false;
            byte[] l_Buf = null;
            uint l_SuggestedSize;

            // Begin of potential critical section.

            UserHeader[] l_Arr = a_Header.f_Users.p_Arr;
            for (int i = (a_Header.f_Users.p_Count - 1); 0 <= i; i--)
            {
                int l_SuggSlot = a_Header.SuggestSlot(ref l_Arr[i].f_Bounds, i);

                if (l_Arr[i].p_Slot > l_SuggSlot)
                {
                    if (false == l_Change)
                    {
                        l_Change = true;
                        l_Buf = new byte[a_Header.f_ChunkSize];

                        f_Resource.AcquireShared();
                    }

                    if (false == a_Header.MoveBody(f_IncomplMem, ref l_Arr[i], l_SuggSlot, l_Buf))
                    {
                        goto HANDLE_IO_ERROR;
                    }
                }
            }

            if (true == a_Header.f_MergedUser.IsSet())
            {
                int l_Slot = a_Header.f_MergedUser.p_Slot;
                int l_SuggSlot = a_Header.FindLowestUnoccupiedSlot(false);

                if (l_Slot > l_SuggSlot)
                {
                    if (false == l_Change)
                    {
                        l_Change = true;
                        l_Buf = new byte[a_Header.f_ChunkSize];

                        f_Resource.AcquireShared();
                    }

                    uint l_Blocks = a_Header.f_MergedUser.f_VerifiedMap.p_Value;
                    if (0 != l_Blocks &&
                        false == a_Header.CopyBlocks(f_IncomplMem, l_Slot, l_SuggSlot, l_Blocks, l_Buf))
                    {
                        goto HANDLE_IO_ERROR;
                    }

                    a_Header.f_MergedUser.SetSlot(l_SuggSlot);
                }
            }

            l_SuggestedSize = a_Header.SuggestElementSize();

            if (true == l_Change)
            {
                // We need to flush the changes we've made to the
                // body, then change the header.

                bool l_FlushAfterwards = (a_Header.p_Element.f_MaxSize != l_SuggestedSize);
                bool l_FlushedAndUpdated;

                f_Cntxt.DebugAcquireBuffers();
                {
                    l_FlushedAndUpdated = (true == f_IncomplMem.Flush() &&
                                           true == a_Header.UpdateUserHeaders(f_IncomplMem, p_Buffer, l_FlushAfterwards));
                }
                f_Cntxt.DebugReleaseBuffers();

                if (false == l_FlushedAndUpdated)
                {
                    goto HANDLE_IO_ERROR;
                }

                f_Resource.ReleaseShared();
            }

            // End of potential critical section. Now adjust (reduce)
            // the size.

            return LockAndResize(ref a_Header, l_SuggestedSize);

            HANDLE_IO_ERROR:

            f_Resource.ReleaseShared();

            return OpStatus.IoError;
        }

#endregion

#region VerificationHandling

        private OpStatus VerifyChunk(ref IncomplChunkHeader a_Header)
        {
            // Each of the following three ways to check an incomplete
            // chunk must (besides checking) do this:
            // - Set the header's disputed flag if suitable and request
            // TAICH block hashs.
            // - Fill in sanctions whenever there is some certainity
            // about right or wrong.
            // - Use BewareOfUser() if a malicious user is detected.

            OpStatus l_Status;
            // l_VerifiedChunkInPlace will be set if the verified data
            // is already stored at slot zero (so that we don't have
            // to move it).
            bool l_VerifiedChunkInPlace;
            byte[] l_VerifiedChunk;

            if (VerificationMode.TAICH == f_VerificationMode)
            {
                Sanctions l_Sanctions;
                Sanctions.Create(out l_Sanctions);

                if (false == a_Header.HasTAICHBlockHashes(true))
                {
                    l_Status = DoTAICHIncomplChunkCheck(ref a_Header,
                                                        ref l_Sanctions,
                                                        out l_VerifiedChunkInPlace,
                                                        out l_VerifiedChunk);
                }
                else
                {
                    l_Status = DoTAICHBlockCheck(ref a_Header,
                                                 ref l_Sanctions,
                                                 out l_VerifiedChunkInPlace,
                                                 out l_VerifiedChunk);
                }

                l_Status = HandleChunkCheckResult(l_Status, ref a_Header, ref l_Sanctions, l_VerifiedChunkInPlace, l_VerifiedChunk);

                return l_Status;
            }
            else if (VerificationMode.eD2k == f_VerificationMode)
            {
                // Chunk bounds.
                ulong l_Pos;
                ulong l_FinalPos;

                // Affected parts bounds.
                ulong l_eD2kPos;
                ulong l_eD2kFinalPos;

                CalceD2kPartBounds(ref a_Header, out l_Pos, out l_FinalPos, out l_eD2kPos, out l_eD2kFinalPos);

                // Now calculate the chunks that may profit from any
                // changes made to the chunk passed by the caller.
                uint l_ChunkLen;
                uint l_ChunkIndex = f_DoneMap.FilePosToIndex(l_eD2kPos, l_eD2kFinalPos, out l_ChunkLen);

                // Set up a simple cache to prevent unnecessary
                // reloading of headers.

                eD2kCheckHeaderCache l_HeaderCache = new eD2kCheckHeaderCache(8);

                {
                    eD2kCheckHeaderCacheItem l_FirstCacheItem = new eD2kCheckHeaderCacheItem(a_Header.f_ChunkIndex);
                    l_FirstCacheItem.SetIncomplete(ref a_Header);

                    int l_FirstIdx = l_HeaderCache.f_Items.ReserveAtEnd();
                    l_HeaderCache.f_Items.p_Arr[l_FirstIdx] = l_FirstCacheItem;
                }

                // Start checking each affected chunk.

                l_Status = OpStatus.ChunkNotVerified;

                Sanctions l_Sanctions;
                Sanctions.Create(out l_Sanctions);

                for (uint i = l_ChunkIndex, lim = (l_ChunkIndex + l_ChunkLen); i < lim; i++)
                {
                    if (true == f_DoneMap.Get(i))
                    {
                        continue;
                    }

                    OpStatus l_CurStatus = DoeD2kIncomplChunkCheck(i,
                                                                   ref l_HeaderCache,
                                                                   ref l_Sanctions,
                                                                   out l_VerifiedChunkInPlace,
                                                                   out l_VerifiedChunk);

                    int l_CacheIdx;
                    if (true == l_HeaderCache.Has(i, out l_CacheIdx))
                    {
                        l_CurStatus = HandleChunkCheckResult(l_CurStatus, ref l_HeaderCache.f_Items.p_Arr[l_CacheIdx].f_UnionHeader.f_Incompl, ref l_Sanctions, l_VerifiedChunkInPlace, l_VerifiedChunk);

                        if (OpStatus.ChunkNotVerified != l_CurStatus)
                        {
                            l_HeaderCache.f_Items.DeleteAt(l_CacheIdx);
                        }
                    }

                    if (a_Header.f_ChunkIndex == i)
                    {
                        // Return the status for the chunk the caller
                        // is interested in.

                        l_Status = l_CurStatus;
                    }
                }

                return l_Status;
            }

            return OpStatus.ChunkNotVerified;
        }

        private OpStatus HandleChunkCheckResult(OpStatus a_CheckResult,
                                                ref IncomplChunkHeader a_Header,
                                                ref Sanctions a_Sanctions,
                                                bool a_VerifiedChunkInPlace,
                                                byte[] a_VerifiedChunk)
        {
            if (OpStatus.ChunkNotVerified != a_CheckResult &&
                OpStatus.ChunkVerified != a_CheckResult)
            {
                // We don't handle that here at all.

                return a_CheckResult;
            }

            a_Sanctions.SendToUserManager(f_Inst);

            if (OpStatus.ChunkNotVerified == a_CheckResult)
            {
                OpStatus l_CompactStatus = CompactIncomplChunk(ref a_Header);

                if (OpStatus.Done != l_CompactStatus)
                {
                    return l_CompactStatus;
                }
            }
            else
            {
                // Create complete chunk.

                P2PHashSets l_VerificationHashSet;
                byte[] l_VerificationMasterHash = GetExpectedMasterHashForVerification(out l_VerificationHashSet);

                ByteArray l_HashableData = new ByteArray(a_VerifiedChunk, ComplChunkHeader.c_BodyStartingOffset);
                uint l_Checksum = Adler32.ComputeHash(ref l_HashableData);

                ComplChunkHeader l_ComplHeader = new ComplChunkHeader(f_Inst, a_Header.f_ChunkIndex, a_Header.f_ChunkSize, l_Checksum);
                l_ComplHeader.Write(f_Known.f_FileNum, a_VerifiedChunk, 0, l_VerificationHashSet, l_VerificationMasterHash);

                bool l_Succ;

                // Use exclusive access here, even though we don't need
                // it for some of the operations.
                f_Resource.AcquireExclusive();
                {
                    if (false == a_VerifiedChunkInPlace)
                    {
                        // Overwrite what's there.

                        ByteArray l_Data = new ByteArray(a_VerifiedChunk);

                        l_Succ = f_IncomplMem.SetAtomic(a_Header.p_Element.f_ElmKey, ref l_Data);
                    }
                    else
                    {
                        // We only need to adjust the header and the
                        // element size.

                        ByteArray l_Data = new ByteArray(a_VerifiedChunk, 0, ComplChunkHeader.c_Size);

                        l_Succ = f_IncomplMem.Write(a_Header.p_Element, 0, ref l_Data);
                    }

                    if (true == l_Succ)
                    {
                        // a_Header is now invalid, so there is no
                        // going back.

                        HardMemory.Element l_Elm = a_Header.p_Element;
                        a_Header.SetElement(null);
                        AddIncomplChunkBytes(-(int)a_Header.GetCoveredBytes());

                        if (true == a_VerifiedChunkInPlace)
                        {
                            // If this fails, it is probably not due to
                            // insufficient disk space, since we're
                            // trying to shrink.

                            l_Succ = f_IncomplMem.Resize(ref l_Elm, (uint)a_VerifiedChunk.Length, true);
                        }

                        l_Succ = (true == l_Succ &&
                                  true == f_IncomplMem.Flush() &&
                                  true == UpdateBitAndWrite(a_Header.f_ChunkIndex, true));

                        if (false == l_Succ)
                        {
                            OpStatus l_CorrStatus = HandleChunkCorruptionFromExcl(a_Header.f_ChunkIndex, l_Elm.f_ElmKey);
                        }
                        else
                        {
                            l_ComplHeader.SetElement(l_Elm);
                        }
                    }
                }
                f_Resource.ReleaseExclusive();

                if (false == l_Succ)
                {
                    return OpStatus.IoError;
                }

                StopDownloadsOfChunk(a_Header.f_ChunkIndex);
            }

            return a_CheckResult;
        }

        private OpStatus DoTAICHBlockCheck(ref IncomplChunkHeader a_Header,
                                           ref Sanctions a_Sanctions,
                                           out bool a_VerifiedChunkInPlace,
                                           out byte[] a_VerifiedChunk)
        {
            a_VerifiedChunkInPlace = false;
            a_VerifiedChunk = null;

            bool l_MergedSlotPresent = a_Header.f_MergedUser.IsSet();
            int l_MergedSlot;
            uint l_Verified;

            if (false == l_MergedSlotPresent)
            {
                l_MergedSlot = a_Header.FindLowestUnoccupiedSlot(false);
                a_Header.f_MergedUser.SetSlot(l_MergedSlot);
                l_Verified = 0;

                // We don't mark the header as dirty yet, since we'll
                // only save changes if the new slot is really filled.
            }
            else
            {
                l_MergedSlot = a_Header.f_MergedUser.p_Slot;
                l_Verified = a_Header.f_MergedUser.f_VerifiedMap.p_Value;
            }

            bool l_HeaderDirty = false;
            bool l_BodyDirty = false;

            uint l_StoredVerified = l_Verified;
            byte[] l_BlockBuf = null;

            UserHeader[] l_Arr = a_Header.f_Users.p_Arr;
            for (int i = (a_Header.f_Users.p_Count - 1); 0 <= i; i--)
            {
                uint l_Dirty = l_Arr[i].f_DirtyMap.p_Value;

                if (0 == l_Dirty)
                {
                    continue;
                }

                // We will check dirty blocks even if we have the
                // corresponding data already, to make sure each
                // user gets sanctionized.

                uint l_AcceptedBytes = 0;
                uint l_RejectedBytes = 0;

                for (uint k = 0; a_Header.f_BlockCount > k; k++)
                {
                    if (false == BlockBitmap.TestBit(k, l_Dirty))
                    {
                        continue;
                    }

                    uint l_Offset;
                    ByteArray l_Data;
                    OpStatus l_BlockStat = ReadAndVerifyBlock(ref a_Header,
                                                              ref l_Arr[i],
                                                              k,
                                                              ref l_BlockBuf,
                                                              out l_Offset,
                                                              out l_Data);

                    BlockBitmap.ClearBit(k, ref l_Dirty);

                    if (OpStatus.NotEnoughData == l_BlockStat)
                    {
                        if (0 == l_Dirty)
                        {
                            break;
                        }

                        continue;
                    }
                    else if (OpStatus.BlockVerified == l_BlockStat)
                    {
                        l_Arr[i].f_AcceptedMap.Set(k);
                        l_AcceptedBytes += (uint)l_Data.f_Len;

                        if (true == BlockBitmap.TestBit(k, l_Verified))
                        {
                            // This block was already added to the
                            // merged user.

                            continue;
                        }

                        if (false == l_MergedSlotPresent)
                        {
                            uint l_NewElmSize = a_Header.SuggestElementSize(l_MergedSlot + 1);
                            OpStatus l_ResizeStatus = LockAndResize(ref a_Header, l_NewElmSize);

                            if (OpStatus.Done != l_ResizeStatus)
                            {
                                return l_ResizeStatus;
                            }

                            l_MergedSlotPresent = true;
                        }

                        bool l_Succ;

                        f_Resource.AcquireShared();
                        {
                            l_Succ = a_Header.WriteBody(f_IncomplMem, l_MergedSlot, l_Offset, ref l_Data);
                        }
                        f_Resource.ReleaseShared();

                        if (false == l_Succ)
                        {
                            return OpStatus.IoError;
                        }

                        BlockBitmap.SetBit(k, ref l_Verified);
                        l_BodyDirty = true;
                    }
                    else if (OpStatus.BlockNotVerified == l_BlockStat)
                    {
                        l_RejectedBytes += (uint)l_Data.f_Len;
                    }
                    else
                    {
                        return l_BlockStat;
                    }
                }

                l_Arr[i].f_DirtyMap.ClearAll();
                l_HeaderDirty = true;

                a_Sanctions.AddSanction(l_Arr[i].f_UserNum, l_AcceptedBytes, l_RejectedBytes);

                if (0 != l_RejectedBytes)
                {
                    a_Header.MarkDisputed();
                    BewareOfUser(l_Arr[i].f_UserNum, CorruptDataObserver.TAICHVerification);

                    if (0 == l_AcceptedBytes)
                    {
                        a_Header.f_Users.DeleteAt(i);
                    }
                }
            }

            if (l_StoredVerified != l_Verified)
            {
                a_Header.f_MergedUser.f_VerifiedMap.TakeOver(l_Verified);
            }

            // Write out any changes to disk (even if we will soon
            // turn the incomplete chunk into a complete chunk).

            if (true == l_HeaderDirty ||
                true == l_BodyDirty)
            {
                // If both are dirty, flush the body before updating
                // the header.

                if (true == l_BodyDirty ||
                    true == l_HeaderDirty)
                {
                    bool l_Succ;

                    f_Cntxt.DebugAcquireBuffers();
                    f_Resource.AcquireShared();
                    {
                        l_Succ = ((false == l_BodyDirty ||
                                   true == f_IncomplMem.Flush()) &&
                                  (false == l_HeaderDirty ||
                                   true == a_Header.UpdateUserHeaders(f_IncomplMem, p_Buffer, false)));
                    }
                    f_Resource.ReleaseShared();
                    f_Cntxt.DebugReleaseBuffers();

                    if (false == l_Succ)
                    {
                        return OpStatus.IoError;
                    }
                }

                // No need to call AddIncomplChunksBytes() here.
            }

            if (a_Header.f_BlockMask == l_Verified)
            {
                // All bits are set, so this should be the complete
                // chunk.

                int l_VerifiedSlot = (a_Header.f_BlockMask == a_Header.f_Users.p_Arr[0].f_AcceptedMap.p_Value ? 0 : l_MergedSlot);
                ByteArray l_Data;

                if (false == ReadIncomplChunk(ref a_Header, l_VerifiedSlot, null, out l_Data))
                {
                    return OpStatus.IoError;
                }

                if (0 != (IncomplFileFlags.ExcessiveTAICHBlockReverification & f_IncomplFlags))
                {
                    // We will reverify to make sure that l_VerifiedSlot
                    // and l_Data really contain verified data only
                    // (protection against e.g. hardware errors).

                    OpStatus l_ReverStatus = VerifyTAICHChunk(ref a_Header, ref l_Data);

                    if (OpStatus.ChunkNotVerified == l_ReverStatus)
                    {
                        // Should never happen.

                        Util.DebugBreak();

                        a_Header.ReverseTAICHBlockVerification();

                        bool l_Succ;

                        f_Cntxt.DebugAcquireBuffers();
                        f_Resource.AcquireShared();
                        {
                            l_Succ = a_Header.UpdateUserHeaders(f_IncomplMem, p_Buffer, false);
                        }
                        f_Resource.ReleaseShared();
                        f_Cntxt.DebugReleaseBuffers();

                        if (false == l_Succ)
                        {
                            return OpStatus.IoError;
                        }

                        return OpStatus.ChunkNotVerified;
                    }
                    else if (OpStatus.ChunkVerified != l_ReverStatus)
                    {
                        return l_ReverStatus;
                    }
                }

                a_VerifiedChunkInPlace = (0 == l_VerifiedSlot);
                a_VerifiedChunk = l_Data.f_Arr;

                return OpStatus.ChunkVerified;
            }

            return OpStatus.ChunkNotVerified;
        }

        private bool ReadIncomplChunk(ref IncomplChunkHeader a_Header,
                                      int a_SrcSlot,
                                      byte[] a_Buffer,
                                      out ByteArray a_Data)
        {
            // a_Buffer can be null.

            if (null == a_Buffer)
            {
                a_Buffer = new byte[IncomplChunkHeader.c_BodyStartingOffset + a_Header.f_ChunkSize];
            }
            else
            {
            }

            a_Data = new ByteArray(a_Buffer, IncomplChunkHeader.c_BodyStartingOffset, (int)a_Header.f_ChunkSize);

            bool l_Succ;

            f_Resource.AcquireShared();
            {
                l_Succ = a_Header.ReadBody(f_IncomplMem, a_SrcSlot, 0, ref a_Data);
            }
            f_Resource.ReleaseShared();

            if (false == l_Succ)
            {
                a_Data = new ByteArray();
            }

            return l_Succ;
        }

        private OpStatus ReadAndVerifyBlock(ref IncomplChunkHeader a_Header,
                                            ref UserHeader a_UserHeader,
                                            uint a_BlockIndex,
                                            ref byte[] a_TempBuffer,
                                            out uint a_Offset,
                                            out ByteArray a_Data)
        {
            // If a_TempBuffer is null, a buffer will be allocated.

            uint l_BlockSize;
            a_Offset = a_Header.GetBlockOffsetAndSize(a_BlockIndex, out l_BlockSize);

            if (false == a_UserHeader.f_Bounds.IsRangeAvailable(a_Offset, a_Offset + l_BlockSize))
            {
                a_Data = new ByteArray();

                return OpStatus.NotEnoughData;
            }

            if (null == a_TempBuffer)
            {
                a_TempBuffer = new byte[TAICHTree.c_BlockSize];
            }

            a_Data = new ByteArray(a_TempBuffer, 0, (int)l_BlockSize);
            bool l_Succ;

            f_Resource.AcquireShared();
            {
                l_Succ = a_Header.ReadBody(f_IncomplMem, a_UserHeader.p_Slot, a_Offset, ref a_Data);
            }
            f_Resource.ReleaseShared();

            if (false == l_Succ)
            {
                return OpStatus.IoError;
            }

            f_Cntxt.f_Tiger.Reset();
            f_Cntxt.f_Tiger.Update(ref a_Data);
            f_Cntxt.f_Tiger.Finish();

            if (false == a_Header.VerifyTAICHBlockHashByBlockIndex(a_BlockIndex, f_Cntxt.f_Tiger))
            {
                return OpStatus.BlockNotVerified;
            }

            return OpStatus.BlockVerified;
        }

        private OpStatus DoTAICHIncomplChunkCheck(ref IncomplChunkHeader a_Header,
                                                  ref Sanctions a_Sanctions,
                                                  out bool a_VerifiedChunkInPlace,
                                                  out byte[] a_VerifiedChunk)
        {
            a_VerifiedChunkInPlace = false;
            a_VerifiedChunk = null;

            if (false == a_Header.HasTAICHBlockHashes(false) &&
                true == a_Header.NeedExtraHashesForTAICHChunkCheck(ref f_TAICHLayout, f_TAICHGreatestUpperLevel))
            {
                RequestTAICHBlocks(a_Header.f_ChunkIndex);

                return OpStatus.ChunkNotVerified;
            }

            bool l_HeaderDirty = false;
            bool l_Disputes = false;
            bool l_Verified = false;
            byte[] l_Buffer = null;

            AvailabilityAnalyzer l_AA = new AvailabilityAnalyzer(0, a_Header.f_ChunkSize, ref a_Header.f_Users);

            if (false == l_AA.IsRangeAvailable() ||
                false == l_AA.PrepareCombinations())
            {
                return OpStatus.ChunkNotVerified;
            }

            UserHeader[] l_Users = a_Header.f_Users.p_Arr;
            FlexArray<UserRange> l_CombLs;
            FlexArray<UserRange>.Create(a_Header.f_Users.p_Count, out l_CombLs);
            Lookup3 l_Lookup3 = new Lookup3();
            Dictionary<ulong, bool> l_Tried = new Dictionary<ulong, bool>();

            while (true)
            {
                bool l_MultipleContributors;

                if (false == l_AA.GetNextCombination(ref l_CombLs, out l_MultipleContributors))
                {
                    break;
                }

                int l_SingleUserIdx = (-1);

                if (false == l_MultipleContributors)
                {
                    bool l_OK = a_Header.GetHeaderIndex(l_CombLs.p_Arr[0].f_UserNum, out l_SingleUserIdx);

                    if (true == l_Users[l_SingleUserIdx].f_ChunkTestedAgainstUpperTree)
                    {
                        continue;
                    }
                }

                l_Lookup3.Reset();
                f_Cntxt.DebugAcquireBuffers();
                {
                    AvailabilityAnalyzer.AddToHashRoutine(l_Lookup3, p_Buffer, ref l_CombLs);
                }
                f_Cntxt.DebugReleaseBuffers();

                ulong l_CombDigest = l_Lookup3.GetDigestAsULong();

                if (true == l_Tried.ContainsKey(l_CombDigest))
                {
                    // We've already tried this combination.

                    continue;
                }

                if (null == l_Buffer)
                {
                    l_Buffer = new byte[GetVerifiedBufferSize(a_Header.f_ChunkIndex)];
                }

                bool l_Succ;
                ByteArray l_Data = new ByteArray(l_Buffer, (int)ComplChunkHeader.c_BodyStartingOffset, (int)a_Header.f_ChunkSize);
                int l_CommonSlot;

                f_Resource.AcquireShared();
                {
                    l_Succ = ReadIncomplChunkMixed(ref a_Header, 0, ref l_CombLs, ref l_Data, out l_CommonSlot);
                }
                f_Resource.ReleaseShared();

                if (false == l_Succ)
                {
                    return OpStatus.IoError;
                }

                OpStatus l_Status = VerifyTAICHChunk(ref a_Header, ref l_Data);

                if (OpStatus.ChunkVerified != l_Status &&
                    OpStatus.ChunkNotVerified != l_Status)
                {
                    return l_Status;
                }

                if (false == l_MultipleContributors)
                {
                    l_Users[l_SingleUserIdx].SetChunkTestedAgainstUpperTree();
                    l_HeaderDirty = true;
                }

                if (OpStatus.ChunkVerified == l_Status)
                {
                    a_Sanctions.AddSanctions(ref l_CombLs, true);

                    a_VerifiedChunkInPlace = (0 == l_CommonSlot);
                    l_Verified = true;

                    break;
                }

                if (false == l_MultipleContributors)
                {
                    a_Sanctions.AddSanctions(ref l_CombLs, false);
                    BewareOfUser(l_Users[l_SingleUserIdx].f_UserNum, CorruptDataObserver.TAICHVerification);
                }

                l_Disputes = true;
                l_Tried.Add(l_CombDigest, true);
            }

            if (true == l_Disputes)
            {
                if (false == a_Header.IsMarkedDisputed())
                {
                    a_Header.MarkDisputed();
                    l_HeaderDirty = true;
                }

                if (false == l_Verified &&
                    false == a_Header.HasTAICHBlockHashes(false))
                {
                    RequestTAICHBlocks(a_Header.f_ChunkIndex);
                }
            }

            if (true == l_HeaderDirty)
            {
                bool l_Succ;

                f_Cntxt.DebugAcquireBuffers();
                f_Resource.AcquireShared();
                {
                    l_Succ = a_Header.UpdateUserHeaders(f_IncomplMem, p_Buffer, false);
                }
                f_Resource.ReleaseShared();
                f_Cntxt.DebugReleaseBuffers();

                if (false == l_Succ)
                {
                    return OpStatus.IoError;
                }
            }

            if (false == l_Verified)
            {
                return OpStatus.ChunkNotVerified;
            }

            a_VerifiedChunk = l_Buffer;

            return OpStatus.ChunkVerified;
        }

        private bool ReadIncomplChunkMixed(ref IncomplChunkHeader a_Header,
                                           uint a_OffsetCorrection,
                                           ref FlexArray<UserRange> a_Ranges,
                                           ref ByteArray a_Dest,
                                           out int a_CommonSlot)
        {
            // Reads data according to the source instructions provided
            // through a_Ranges. a_CommonSlot is negative on return if
            // there are multiple slots involved.

            a_CommonSlot = (-1);

            int l_Count = a_Ranges.p_Count;
            UserRange[] l_Ranges = a_Ranges.p_Arr;

            // First find out if there is a common slot index.

            for (int i = 0; i < l_Count; i++)
            {
                if (0 == i)
                {
                    a_CommonSlot = l_Ranges[0].f_Slot;
                }
                else if (l_Ranges[i].f_Slot != a_CommonSlot)
                {
                    a_CommonSlot = (-1);

                    break;
                }
            }

            for (int k = 0; k < l_Count; k++)
            {
                ByteArray l_CurDest = new ByteArray(a_Dest.f_Arr,
                                                    a_Dest.f_Idx + (int)(l_Ranges[k].f_Rng.f_Offset - a_OffsetCorrection),
                                                    (int)l_Ranges[k].f_Rng.f_Size);

                if (false == a_Header.ReadBody(f_IncomplMem, l_Ranges[k].f_Slot, l_Ranges[k].f_Rng.f_Offset, ref l_CurDest))
                {
                    return false;
                }
            }

            return true;
        }

        private OpStatus VerifyTAICHChunk(ref IncomplChunkHeader a_Header,
                                          ref ByteArray a_Data)
        {
            if (1 == f_TAICHLayout.f_Total)
            {
                // The downloaded file's size is TAICHTree.c_BlockSize
                // at most. There is no tree in this case, the TAICH
                // hash is at the same time the Tiger hash of the file.

                f_Cntxt.f_Tiger.Reset();
                f_Cntxt.f_Tiger.Update(ref a_Data);
                f_Cntxt.f_Tiger.Finish();

                if (false == f_HashStorage.HasThisHash(f_EstimatedKnowledge.p_TAICHHashIdx, f_Cntxt.f_Tiger))
                {
                    return OpStatus.ChunkNotVerified;
                }

                return OpStatus.ChunkVerified;
            }

            bool l_BlockVerification = !a_Header.HasTAICHBlockHashes(true);
            TigerHashHive l_Hive;

            if (false == l_BlockVerification)
            {
                l_Hive = new TigerHashHive();
            }
            else
            {
                if (true == a_Header.NeedExtraHashesForTAICHChunkCheck(ref f_TAICHLayout, f_TAICHGreatestUpperLevel))
                {
                    // Should never happen (this routine may not be
                    // called in this case).

                    throw (new ApplicationException());
                }

                TigerHashHive.CreateExpandable((int)a_Header.f_BlockCount, out l_Hive);
            }

            // Hash each block and compare the resulting hash directly
            // OR add it to the hive.

            for (uint i = 0, l_HashIndex = a_Header.f_FirstHashIndex; a_Header.f_BlockCount > i; i++, l_HashIndex++)
            {
                uint l_BlockSize;
                uint l_Offset = a_Header.GetBlockOffsetAndSize(i, out l_BlockSize);

                f_Cntxt.f_Tiger.Reset();
                f_Cntxt.f_Tiger.Update(a_Data.f_Arr, a_Data.f_Idx + (int)l_Offset, (int)l_BlockSize);
                f_Cntxt.f_Tiger.Finish();

                if (false == l_BlockVerification)
                {
                    if (false == a_Header.VerifyTAICHBlockHashByHashIndex(l_HashIndex, f_Cntxt.f_Tiger))
                    {
                        return OpStatus.ChunkNotVerified;
                    }
                }
                else
                {
                    l_Hive.Add(l_HashIndex, f_Cntxt.f_Tiger);
                }
            }

            if (true == l_BlockVerification)
            {
                bool l_NeedTempAccessor = (0 == f_TAICHGreatestUpperLevel);

                if (true == l_NeedTempAccessor)
                {
                    // If we land here, the file we are downloading is
                    // larger than TAICHTree.c_BlockSize but no larger
                    // than c_ChunkSize. We'll create a temp accessor.
                    // We may not return without destroying it again.

                    byte[] l_TopHash = Util.CloneArray(f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_TAICHHashIdx));
                    ByteArray l_TopHashData = new ByteArray(l_TopHash);
                    f_TAICHAccessor = new TAICHTreeAccessor(ref f_TAICHLayout, 0, ref l_TopHashData);
                }

                TAICHVerifyResult l_VerRes;

                f_TAICHAccessor.AcquireSharedIfDiskBased();
                {
                    l_VerRes = TAICHTree.VerifySubTree(ref f_TAICHLayout,
                                                       c_TAICHExponent,
                                                       a_Header.f_ChunkIndex,
                                                       ref f_TAICHAccessor,
                                                       ref l_Hive,
                                                       f_Cntxt.f_Tiger);
                }
                f_TAICHAccessor.ReleaseSharedIfDiskBased();

                if (true == l_NeedTempAccessor)
                {
                    // Remove the temp accessor again.
                    f_TAICHAccessor = new TAICHTreeAccessor();
                }

                if (TAICHVerifyResult.Valid != l_VerRes)
                {
                    if (TAICHVerifyResult.HashMismatch == l_VerRes)
                    {
                        return OpStatus.ChunkNotVerified;
                    }

                    if (TAICHVerifyResult.AccessFailed == l_VerRes)
                    {
                        return OpStatus.IoError;
                    }

                    // Should never happen.

                    throw (new ApplicationException());
                }
            }

            return OpStatus.ChunkVerified;
        }

        private OpStatus DoeD2kIncomplChunkCheck(uint a_ChunkIndex,
                                                 ref eD2kCheckHeaderCache a_HeaderCache,
                                                 ref Sanctions a_Sanctions,
                                                 out bool a_VerifiedChunkInPlace,
                                                 out byte[] a_VerifiedChunk)
        {
            a_VerifiedChunkInPlace = false;
            a_VerifiedChunk = null;

            if (true == a_HeaderCache.DidLoadFail(a_ChunkIndex))
            {
                return OpStatus.ChunkNotVerified;
            }

            int l_CacheIdx;
            if (false == a_HeaderCache.Has(a_ChunkIndex, out l_CacheIdx))
            {
                OpStatus l_ChunkStatus;

                f_Resource.AcquireShared();
                {
                    l_ChunkStatus = LoadHeaderIntoeD2kCache(ref a_HeaderCache, a_ChunkIndex, out l_CacheIdx);
                }
                f_Resource.ReleaseShared();

                if (OpStatus.Done != l_ChunkStatus)
                {
                    return OpStatus.ChunkNotVerified;
                }
            }

            IncomplChunkHeader l_Header = a_HeaderCache.f_Items.p_Arr[l_CacheIdx].f_UnionHeader.f_Incompl;

            if (false == l_Header.IsRangeAvailable(0, l_Header.f_ChunkSize, true))
            {
                return OpStatus.ChunkNotVerified;
            }

            // Chunk bounds.
            ulong l_Pos;
            ulong l_FinalPos;

            // Affected parts bounds.
            ulong l_eD2kPos;
            ulong l_eD2kFinalPos;

            CalceD2kPartBounds(ref l_Header, out l_Pos, out l_FinalPos, out l_eD2kPos, out l_eD2kFinalPos);

            uint l_eD2kSize = (uint)(l_eD2kFinalPos - l_eD2kPos);

            int l_AffectedChunks;
            FlexArray<eD2kVerifyContext> l_Cntxts;

            {
                uint l_LeftChunkIdx = (uint)(l_eD2kPos / c_ChunkSize);

                uint l_InitialSpacing = (uint)(l_eD2kPos % c_ChunkSize);
                uint l_LeftSize = 0;

                if (0 != l_InitialSpacing)
                {
                    l_LeftSize = (GetChunkSize(l_LeftChunkIdx) - l_InitialSpacing);
                }

                int l_MidChunks = (int)((l_eD2kSize - l_LeftSize) / c_ChunkSize);
                uint l_RightSize = (l_eD2kSize - ((uint)l_MidChunks * c_ChunkSize) - l_LeftSize);

                l_AffectedChunks = (l_MidChunks + (0 == l_LeftSize ? 0 : 1) + (0 == l_RightSize ? 0 : 1));
                FlexArray<eD2kVerifyContext>.Create(out l_Cntxts);
                l_Cntxts.ReserveRange(0, l_AffectedChunks);

                int l_Idx = 0;
                uint l_CumOffset = 0;

                if (0 != l_LeftSize)
                {
                    l_Cntxts.p_Arr[l_Idx] = new eD2kVerifyContext(l_LeftChunkIdx, 0, l_InitialSpacing, l_LeftSize);
                    l_Idx++;
                    l_CumOffset = l_LeftSize;
                }

                for (int i = 0; i < l_MidChunks; i++)
                {
                    l_Cntxts.p_Arr[l_Idx] = new eD2kVerifyContext(l_LeftChunkIdx + (uint)l_Idx, l_CumOffset, 0, c_ChunkSize);
                    l_Idx++;
                    l_CumOffset += c_ChunkSize;
                }

                if (0 != l_RightSize)
                {
                    l_Cntxts.p_Arr[l_Idx] = new eD2kVerifyContext(l_LeftChunkIdx + (uint)l_Idx, l_CumOffset, 0, l_RightSize);
                    l_Idx++;
                    l_CumOffset += l_RightSize;
                }
            }

            OpStatus l_Status = OpStatus.Done;
            byte[] l_Buffer = null;
            ResourceAccessType l_CurAccessType = ResourceAccessType.None;

            // Begin of potential critical section.

            {
                for (int i = 0; l_AffectedChunks > i; i++)
                {
                    uint l_CurChunkIndex = l_Cntxts.p_Arr[i].f_ChunkIndex;
                    int l_CurCacheIdx;

                    if (true == a_HeaderCache.DidLoadFail(l_CurChunkIndex))
                    {
                        l_Status = OpStatus.ChunkNotVerified;

                        goto RELEASE_AND_EXIT;
                    }

                    if (false == a_HeaderCache.Has(l_CurChunkIndex, out l_CurCacheIdx))
                    {
                        f_Resource.SwitchToIfNecessaryAdv(ref l_CurAccessType, ResourceAccessType.Shared);

                        l_Status = LoadHeaderIntoeD2kCache(ref a_HeaderCache, l_CurChunkIndex, out l_CurCacheIdx);

                        if (OpStatus.Done != l_Status)
                        {
                            l_Status = OpStatus.ChunkNotVerified;

                            goto RELEASE_AND_EXIT;
                        }
                    }

                    bool l_Avail = l_Cntxts.p_Arr[i].SetCacheItem(ref a_HeaderCache.f_Items.p_Arr[l_CurCacheIdx]);

                    if (false == l_Avail)
                    {
                        l_Status = OpStatus.ChunkNotVerified;

                        goto RELEASE_AND_EXIT;
                    }
                }

                // Now that we know it makes sense, load any data from
                // completed chunks.

                l_Buffer = new byte[(int)l_eD2kSize];

                eD2kVerifyContext[] l_Arr = l_Cntxts.p_Arr;
                for (int i = (l_Cntxts.p_Count - 1); 0 <= i; i--)
                {
                    if (UnionChunkHeader.State.CompleteChunkSet == l_Arr[i].p_State)
                    {
                        f_Resource.SwitchToIfNecessaryAdv(ref l_CurAccessType, ResourceAccessType.Shared);

                        ByteArray l_Data;
                        l_Arr[i].BuildByteArray(l_Buffer, out l_Data);

                        l_Status = ReadComplChunk(ref l_Arr[i].f_Item.f_UnionHeader.f_Compl, l_Arr[i].f_Offset, ref l_Data);

                        if (OpStatus.Done != l_Status)
                        {
                            if (OpStatus.Corruption == l_Status)
                            {
                                OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(l_Arr[i].f_Item.f_UnionHeader.f_Compl.f_ChunkIndex, l_Arr[i].f_Item.f_UnionHeader.f_Compl.p_Element.f_ElmKey);
                            }

                            goto RELEASE_AND_EXIT;
                        }

                        l_Cntxts.DeleteAt(i);
                    }
                }
            }
            f_Resource.SwitchToIfNecessaryAdv(ref l_CurAccessType, ResourceAccessType.None);

            // End of potential critical section.

            uint l_PartIndex = (uint)(l_eD2kPos / eD2kHashSet.c_PartSize);
            uint l_AffectedParts = ((l_eD2kSize / eD2kHashSet.c_PartSize) +
                                    (0 == (l_eD2kSize % eD2kHashSet.c_PartSize) ? 0u : 1u));

            l_Status = DoeD2kIncomplChunkCheckCore(l_Cntxts.ToArray(true), ref a_Sanctions, l_PartIndex, l_AffectedParts, l_Buffer);

            if (OpStatus.ChunkVerified == l_Status)
            {
                a_VerifiedChunkInPlace = (1 == l_Header.f_Users.p_Count &&
                                          0 == l_Header.f_Users.p_Arr[0].p_Slot);
                a_VerifiedChunk = new byte[ComplChunkHeader.c_BodyStartingOffset + l_Header.f_ChunkSize];

                Buffer.BlockCopy(l_Buffer, (int)(l_Pos - l_eD2kPos), a_VerifiedChunk, ComplChunkHeader.c_BodyStartingOffset, (int)l_Header.f_ChunkSize);
            }

            return l_Status;

            RELEASE_AND_EXIT:

            f_Resource.SwitchToIfNecessary(l_CurAccessType, ResourceAccessType.None);

            return l_Status;
        }

        private void CalceD2kPartBounds(ref IncomplChunkHeader a_Header,
                                        out ulong a_ChunkPos,
                                        out ulong a_ChunkFinalPos,
                                        out ulong a_eD2kPos,
                                        out ulong a_eD2kFinalPos)
        {
            // Chunk bounds.
            a_ChunkPos = ((ulong)a_Header.f_ChunkIndex * c_ChunkSize);
            a_ChunkFinalPos = (a_ChunkPos + a_Header.f_ChunkSize);

            // Expand chunk bounds to part bounds.
            CalceD2kPartBoundsWide(f_Known.f_FileSize, a_ChunkPos, a_ChunkFinalPos, out a_eD2kPos, out a_eD2kFinalPos);
        }

        public static void CalceD2kPartBoundsWide(ulong a_FileSize,
                                                  ulong a_FilePos,
                                                  ulong a_FinalFilePos,
                                                  out ulong a_eD2kPos,
                                                  out ulong a_eD2kFinalPos)
        {
            a_eD2kPos = (a_FilePos - (a_FilePos % eD2kHashSet.c_PartSize));
            uint l_Remainder = (uint)(a_FinalFilePos % eD2kHashSet.c_PartSize);
            a_eD2kFinalPos = (a_FinalFilePos + (0 == l_Remainder ? 0u : (eD2kHashSet.c_PartSize - l_Remainder)));
            a_eD2kFinalPos = Math.Min(a_FileSize, a_eD2kFinalPos);
        }

        public static bool CalceD2kPartBoundsNarrow(ulong a_FileSize,
                                                    ulong a_FilePos,
                                                    ulong a_FinalFilePos,
                                                    out ulong a_eD2kPos,
                                                    out ulong a_eD2kFinalPos)
        {
            // Returns true if a nonzero range was calculated.

            uint l_Remainder = (uint)(a_FilePos % eD2kHashSet.c_PartSize);
            a_eD2kPos = (0 == l_Remainder ? a_FilePos : (a_FilePos + (eD2kHashSet.c_PartSize - l_Remainder)));

            if (a_FileSize > a_FinalFilePos)
            {
                a_eD2kFinalPos = (a_FinalFilePos - (a_FinalFilePos % eD2kHashSet.c_PartSize));
            }
            else
            {
                a_eD2kFinalPos = a_FileSize;
            }

            return (a_eD2kPos < a_eD2kFinalPos);
        }

        private OpStatus LoadHeaderIntoeD2kCache(ref eD2kCheckHeaderCache a_Cache,
                                                 uint a_ChunkIndex,
                                                 out int a_CacheIndex)
        {
            a_CacheIndex = (-1);

            OpStatus l_Status;
            ElmKey l_ElmKey = GetHotIncomplChunksEntry(a_ChunkIndex);

            if (true == l_ElmKey.IsZero())
            {
                l_Status = ReadSingleChunkKey(a_ChunkIndex, out l_ElmKey);

                if (OpStatus.Done != l_Status)
                {
                    return l_Status;
                }

                if (true == l_ElmKey.IsZero())
                {
                    // We cannot go on, since there isn't enough data
                    // yet.

                    a_Cache.LoadFailed(a_ChunkIndex);

                    return OpStatus.ChunkNotVerified;
                }
            }

            bool l_Done = f_DoneMap.Get(a_ChunkIndex);
            UnionChunkHeader l_UnionHeader = new UnionChunkHeader();

            if (false == l_Done)
            {
                l_Status = ReadIncomplChunkHeader(a_ChunkIndex, l_ElmKey, false, out l_UnionHeader.f_Incompl);
            }
            else
            {
                l_Status = ReadComplChunkHeader(a_ChunkIndex, l_ElmKey, out l_UnionHeader.f_Compl);
            }

            if (OpStatus.Done != l_Status)
            {
                if (OpStatus.Corruption == l_Status)
                {
                    OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(a_ChunkIndex, l_ElmKey);
                }

                return l_Status;
            }

            a_CacheIndex = a_Cache.f_Items.ReserveAtEnd();
            eD2kCheckHeaderCacheItem[] l_Arr = a_Cache.f_Items.p_Arr;

            l_Arr[a_CacheIndex] = new eD2kCheckHeaderCacheItem(a_ChunkIndex);

            if (false == l_Done)
            {
                l_Arr[a_CacheIndex].SetIncomplete(ref l_UnionHeader.f_Incompl);
            }
            else
            {
                l_Arr[a_CacheIndex].SetComplete(ref l_UnionHeader.f_Compl);
            }

            return OpStatus.Done;
        }

        private OpStatus DoeD2kIncomplChunkCheckCore(eD2kVerifyContext[] a_Cntxts,
                                                     ref Sanctions a_Sanctions,
                                                     uint a_PartIndex,
                                                     uint a_AffectedParts,
                                                     byte[] a_PartBuffer)
        {
            int l_ContextCount = a_Cntxts.Length;
            Lookup3 l_Lookup3 = new Lookup3();
            Dictionary<ulong, bool> l_Tried = new Dictionary<ulong, bool>();
            bool l_FirstRound = true;

            byte[] l_PartHash = null;
            bool l_IsDisputed = false;
            OpStatus l_ErrStatus;

            while (true)
            {
                if (true == l_FirstRound)
                {
                    // Initialize.

                    for (int i = 0; i < l_ContextCount; i++)
                    {
                        a_Cntxts[i].LoadNextCombination();
                    }

                    l_FirstRound = false;
                }
                else
                {
                    bool l_EndReached = false;

                    // Make the latter elements change more frequently
                    // than the starting elements.
                    for (int i = (l_ContextCount - 1); 0 <= i; i--)
                    {
                        bool l_StageEndReached = a_Cntxts[i].LoadNextCombination();

                        if (false == l_StageEndReached)
                        {
                            break;
                        }

                        if (0 == i)
                        {
                            l_EndReached = true;

                            break;
                        }
                    }

                    if (true == l_EndReached)
                    {
                        break;
                    }
                }

                l_Lookup3.Reset();

                f_Cntxt.DebugAcquireBuffers();
                {
                    for (int i = 0; i < l_ContextCount; i++)
                    {
                        a_Cntxts[i].AddToHashRoutine(l_Lookup3, p_Buffer);
                    }
                }
                f_Cntxt.DebugReleaseBuffers();

                ulong l_CombDigest = l_Lookup3.GetDigestAsULong();

                if (true == l_Tried.ContainsKey(l_CombDigest))
                {
                    // We've already tried this combination.

                    continue;
                }

                f_Resource.AcquireShared();
                {
                    for (int i = 0; i < l_ContextCount; i++)
                    {
                        if (true == a_Cntxts[i].f_Dirty)
                        {
                            ByteArray l_Data;
                            a_Cntxts[i].BuildByteArray(a_PartBuffer, out l_Data);
                            int l_Junk;

                            if (false == ReadIncomplChunkMixed(ref a_Cntxts[i].f_Item.f_UnionHeader.f_Incompl, a_Cntxts[i].f_Offset, ref a_Cntxts[i].f_CombRanges, ref l_Data, out l_Junk))
                            {
                                l_ErrStatus = OpStatus.IoError;

                                goto RELEASE_AND_EXIT;
                            }

                            a_Cntxts[i].f_Dirty = false;
                        }
                    }

                    if (null == l_PartHash)
                    {
                        l_PartHash = new byte[(int)a_AffectedParts * MD_4.c_HashByteSize];

                        l_ErrStatus = GeteD2kPartHashes(a_PartIndex, a_AffectedParts, l_PartHash, 0);

                        if (OpStatus.Done != l_ErrStatus)
                        {
                            goto RELEASE_AND_EXIT;
                        }
                    }
                }
                f_Resource.ReleaseShared();

                bool l_Verified = true;

                for (int i = 0, l_BufPos = 0; i < a_AffectedParts; i++, l_BufPos += (int)eD2kHashSet.c_PartSize)
                {
                    f_Cntxt.f_MD4.Reset();
                    f_Cntxt.f_MD4.Update(a_PartBuffer, l_BufPos, Math.Min((int)eD2kHashSet.c_PartSize, a_PartBuffer.Length - l_BufPos));
                    f_Cntxt.f_MD4.Finish();

                    if (false == f_Cntxt.f_MD4.VerifyBE(l_PartHash, i * MD_4.c_HashByteSize))
                    {
                        l_Verified = false;

                        break;
                    }
                }

                if (true == l_Verified)
                {
                    // Exit and ignore l_IsDisputed, since we've made
                    // it finally.

                    eD2kVerifyContext.AddSanctions(a_Cntxts, ref a_Sanctions);

                    return OpStatus.ChunkVerified;
                }

                l_Tried.Add(l_CombDigest, true);
                l_IsDisputed = true;

                UserNum l_TrgtUserNum;
                if (true == eD2kVerifyContext.HaveOneOrigin(a_Cntxts, out l_TrgtUserNum))
                {
                    BewareOfUser(l_TrgtUserNum, CorruptDataObserver.eD2kVerification);
                }
            }

            if (true == l_IsDisputed)
            {
                // We can't figure out from which chunk the dispute
                // originates, so we'll assume all.

                ResourceAccessType l_CurAccessType = ResourceAccessType.None;

                for (int i = 0; i < l_ContextCount; i++)
                {
                    if (false == a_Cntxts[i].f_Item.f_UnionHeader.f_Incompl.IsMarkedDisputed())
                    {
                        if (0 != (IncomplFileFlags.AllowTAICHVerification & f_IncomplFlags) &&
                            false == a_Cntxts[i].f_Item.f_UnionHeader.f_Incompl.HasTAICHBlockHashes(false))
                        {
                            RequestTAICHBlocks(a_Cntxts[i].f_ChunkIndex);
                        }

                        a_Cntxts[i].f_Item.f_UnionHeader.f_Incompl.MarkDisputed();

                        f_Resource.SwitchToIfNecessaryAdv(ref l_CurAccessType, ResourceAccessType.Shared);

                        f_Cntxt.DebugAcquireBuffers();
                        {
                            if (false == a_Cntxts[i].f_Item.f_UnionHeader.f_Incompl.UpdateUserHeaders(f_IncomplMem, p_Buffer, false))
                            {
                                // Don't bother if it didn't work.

                                Util.DebugBreak();
                            }
                        }
                        f_Cntxt.DebugReleaseBuffers();

                        // We don't need to call AddIncomplChunksBytes()
                        // here, since our update was neutral.
                    }
                }

                f_Resource.SwitchToIfNecessary(l_CurAccessType, ResourceAccessType.None);
            }

            return OpStatus.ChunkNotVerified;

            RELEASE_AND_EXIT:

            f_Resource.ReleaseShared();

            return l_ErrStatus;
        }

        private OpStatus GeteD2kPartHashes(uint a_PartIndex,
                                           uint a_PartCount,
                                           byte[] a_Dest,
                                           int a_DestIndex)
        {
            if (0 == f_eD2kHashSetElementCount)
            {
                // In this case the master hash is the hash of the
                // first part.

                byte[] l_MasterHash = f_HashStorage.ExposeHash(f_EstimatedKnowledge.p_eD2kHashIdx);

                if (null == l_MasterHash)
                {
                    return OpStatus.NotEnoughData;
                }

                Buffer.BlockCopy(l_MasterHash, 0, a_Dest, a_DestIndex, MD_4.c_HashByteSize);
            }
            else
            {
                // We don't need to call f_eD2kAccessor.AcquireSharedIfDiskBased()
                // here, since we have shared access already.
                bool l_Succ = f_eD2kAccessor.GetPartHashes(a_PartIndex, a_PartCount, a_Dest, a_DestIndex);

                if (false == l_Succ)
                {
                    return OpStatus.IoError;
                }
            }

            return OpStatus.Done;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct eD2kCheckHeaderCache
        {
            private FlexArray<uint> f_FailedLoads;
            public FlexArray<eD2kCheckHeaderCacheItem> f_Items;

            public eD2kCheckHeaderCache(int a_InitCapacity)
            {
                FlexArray<uint>.Create(a_InitCapacity, out f_FailedLoads);
                FlexArray<eD2kCheckHeaderCacheItem>.Create(a_InitCapacity, out f_Items);
            }

            public void LoadFailed(uint a_ChunkIndex)
            {
                int l_Idx = f_FailedLoads.ReserveAtEnd();
                f_FailedLoads.p_Arr[l_Idx] = a_ChunkIndex;
            }

            public bool DidLoadFail(uint a_ChunkIndex)
            {
                for (int i = 0; f_FailedLoads.p_Count > i; i++)
                {
                    if (a_ChunkIndex == f_FailedLoads.p_Arr[i])
                    {
                        return true;
                    }
                }

                return false;
            }

            public bool Has(uint a_ChunkIndex,
                            out int a_ListIndex)
            {
                for (int i = 0; f_Items.p_Count > i; i++)
                {
                    if (a_ChunkIndex == f_Items.p_Arr[i].f_ChunkIndex)
                    {
                        a_ListIndex = i;

                        return true;
                    }
                }

                a_ListIndex = (-1);

                return false;
            }
        }

        //[StructLayout(LayoutKind.Explicit)]
#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct UnionChunkHeader
        {
            //[FieldOffset(0)]
            public State f_State;

            // Union (this only works safely without any JIT drama
            // if none of the involved structs contains a pointer -
            // otherwise pray that the JIT compiler has a good day).
            // Release decision: Deactivate this until it's confirmed
            // that the runtime accepts it on x86 (confirmed) AND x86-64.
            //[FieldOffset(8)]
            public IncomplChunkHeader f_Incompl;
            //[FieldOffset(8)]
            public ComplChunkHeader f_Compl;

            public enum State : byte
            {
                Invalid = 0,
                IncompleteChunkSet = 1,
                CompleteChunkSet = 2,
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct eD2kCheckHeaderCacheItem
        {
            public readonly uint f_ChunkIndex;

            public UnionChunkHeader f_UnionHeader;

            public UnionChunkHeader.State p_State
            {
                get { return f_UnionHeader.f_State; }
            }

            public eD2kCheckHeaderCacheItem(uint a_ChunkIndex)
            {
                f_ChunkIndex = a_ChunkIndex;

                f_UnionHeader = new UnionChunkHeader();
            }

            public void SetIncomplete(ref IncomplChunkHeader a_Header)
            {
                // Returns true if the specified range is available.

                f_UnionHeader.f_State = UnionChunkHeader.State.IncompleteChunkSet;
                f_UnionHeader.f_Incompl = a_Header;
            }

            public void SetComplete(ref ComplChunkHeader a_Header)
            {
                f_UnionHeader.f_State = UnionChunkHeader.State.CompleteChunkSet;
                f_UnionHeader.f_Compl = a_Header;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct eD2kVerifyContext
        {
            // This should be a very large value (to make it less
            // likely to appear in the wild), so we start with 0xF.
            private const uint c_ChunkIndicationMagic = 0xF9181636;

            public readonly uint f_ChunkIndex;
            public readonly uint f_CumOffset;
            public readonly uint f_Offset;
            public readonly uint f_Size;

            public eD2kCheckHeaderCacheItem f_Item;

            // Only valid for incomplete chunks.
            private AvailabilityAnalyzer f_AA;
            public bool f_Dirty;
            public FlexArray<UserRange> f_CombRanges;

            public UnionChunkHeader.State p_State
            {
                get { return f_Item.p_State; }
            }

            public eD2kVerifyContext(uint a_ChunkIndex,
                                     uint a_CumOffset,
                                     uint a_Offset,
                                     uint a_Size)
            {
                f_ChunkIndex = a_ChunkIndex;
                f_CumOffset = a_CumOffset;
                f_Offset = a_Offset;
                f_Size = a_Size;

                f_Item = new eD2kCheckHeaderCacheItem();

                f_AA = new AvailabilityAnalyzer();
                f_Dirty = false;
                FlexArray<UserRange>.Create(out f_CombRanges);
            }

            public void BuildByteArray(byte[] a_PartBuffer,
                                       out ByteArray a_Data)
            {
                a_Data = new ByteArray(a_PartBuffer, (int)f_CumOffset, (int)f_Size);
            }

            public bool SetCacheItem(ref eD2kCheckHeaderCacheItem a_CacheItem)
            {
                // For complete chunks this routine always returns
                // true. For incomplete chunks this routine returns
                // true if the specified range is available.

                f_Item = a_CacheItem;

                if (UnionChunkHeader.State.IncompleteChunkSet == f_Item.p_State)
                {
                    f_AA = new AvailabilityAnalyzer(f_Offset, f_Size, ref f_Item.f_UnionHeader.f_Incompl.f_Users);

                    return f_AA.IsRangeAvailable();
                }

                return true;
            }

            public bool LoadNextCombination()
            {
                // Loads the next combination, updates the dirty flag
                // and returns true if the end of the iteration was
                // reached, in which case an automatic reset happens
                // before returning to the caller.

                if (false == f_AA.IsPrepared())
                {
                    // Initialize.

                    if (false == f_AA.PrepareCombinations())
                    {
                        // Don't go on if SetIncomplete() returns false
                        // or if there are too many users.

                        throw (new ApplicationException());
                    }

                    f_CombRanges.SetCapacity(f_Item.f_UnionHeader.f_Incompl.f_Users.p_Count);
                }

                f_Dirty = true;

                if (false == f_AA.GetNextCombination(ref f_CombRanges))
                {
                    // Do a reset.

                    bool l_OK = f_AA.PrepareCombinations();

                    l_OK = f_AA.GetNextCombination(ref f_CombRanges);

                    return true;
                }

                return false;
            }

            public void AddToHashRoutine(IHashFunction a_HashAlgo,
                                         byte[] a_TempBuffer)
            {
                // Write the chunk identifier.

                Util.WriteUInt32LE(c_ChunkIndicationMagic, a_TempBuffer, 0);
                Util.WriteUInt32LE(f_ChunkIndex, a_TempBuffer, sizeof(uint));

                a_HashAlgo.Update(a_TempBuffer, 0, 2 * sizeof(uint));

                // Write the ranges belonging to this chunk.

                AvailabilityAnalyzer.AddToHashRoutine(a_HashAlgo, a_TempBuffer, ref f_CombRanges);
            }

            public static void AddSanctions(eD2kVerifyContext[] a_Cntxts,
                                            ref Sanctions a_Sanctions)
            {
                for (int i = 0; a_Cntxts.Length > i; i++)
                {
                    a_Sanctions.AddSanctions(ref a_Cntxts[i].f_CombRanges, true);
                }
            }

            public static bool HaveOneOrigin(eD2kVerifyContext[] a_Cntxts,
                                             out UserNum a_TrgtUserNum)
            {
                // Returns true if only one user is forming the current
                // combination.

                bool l_GotOne = false;
                a_TrgtUserNum = new UserNum();

                for (int i = 0; a_Cntxts.Length > i; i++)
                {
                    for (int k = 0; a_Cntxts[i].f_CombRanges.p_Count > k; k++)
                    {
                        if (false == l_GotOne)
                        {
                            l_GotOne = true;
                            a_TrgtUserNum = a_Cntxts[i].f_CombRanges.p_Arr[k].f_UserNum;
                        }
                        else if (a_Cntxts[i].f_CombRanges.p_Arr[k].f_UserNum != a_TrgtUserNum)
                        {
                            return false;
                        }
                    }
                }

                return l_GotOne;
            }
        }

#endregion

        public OpStatus Extraction(ExtractionCookie a_Cookie)
        {
            // The requested range doesn't have to be aligned in any
            // way.

            if (0 == a_Cookie.f_Size)
            {
                a_Cookie.Fail(ExtractionCookie.Result.InputCorrupt);

                return OpStatus.Error;
            }

            if (VerificationMode.None == f_VerificationMode)
            {
                a_Cookie.Fail(ExtractionCookie.Result.IOError);

                return OpStatus.Error;
            }

            ulong l_FilePos = a_Cookie.f_FilePos;
            ulong l_Size = a_Cookie.f_Size;
            ulong l_FinalPos = (l_FilePos + l_Size);
            uint l_ChunkIndex = (uint)(l_FilePos / c_ChunkSize);
            uint l_LastChunkIndex = (uint)(l_FinalPos / c_ChunkSize);
            uint l_StartingOffset = (uint)(l_FilePos % c_ChunkSize);
            uint l_EndingOffset = (uint)(l_FinalPos % c_ChunkSize);

            if (0 == l_EndingOffset)
            {
                l_LastChunkIndex--;
                l_EndingOffset = c_ChunkSize;
            }

            if (false == f_DoneMap.IsRunSet(l_ChunkIndex, l_LastChunkIndex - l_ChunkIndex + 1))
            {
                a_Cookie.Fail(ExtractionCookie.Result.NotAvail);

                return OpStatus.NotEnoughData;
            }

            byte[] l_Temp = new byte[(int)Math.Min(c_ChunkSize, l_Size)];
            FHandle l_DestFile = a_Cookie.f_TempFile.ExposeFHandle();

            f_Resource.AcquireShared();

            for (uint i = l_ChunkIndex; i <= l_LastChunkIndex; i++)
            {
                ElmKey l_ElmKey;
                OpStatus l_Status = ReadSingleChunkKey(i, out l_ElmKey);

                if (OpStatus.Done != l_Status)
                {
                    goto CHUNKERR_RELEASE_AND_EXIT;
                }

                if (true == l_ElmKey.IsZero())
                {
                    // The corruption handling will make sure the
                    // chunk's bit is cleared, as it belongs.

                    l_Status = OpStatus.Corruption;

                    goto CHUNKERR_RELEASE_AND_EXIT;
                }

                ComplChunkHeader l_Header;
                l_Status = ReadComplChunkHeader(i, l_ElmKey, out l_Header);

                if (OpStatus.Done != l_Status)
                {
                    goto CHUNKERR_RELEASE_AND_EXIT;
                }

                bool l_LastLoopRound = (i == l_LastChunkIndex);
                ByteArray l_Data = new ByteArray(l_Temp, 0, (int)((false == l_LastLoopRound ? c_ChunkSize : l_EndingOffset) - l_StartingOffset));
                l_Status = ReadComplChunk(ref l_Header, l_StartingOffset, ref l_Data);

                if (OpStatus.Done == l_Status)
                {
                    // Write to destination file.

                    if (true == l_LastLoopRound)
                    {
                        f_Resource.ReleaseShared();
                    }

                    int l_StartBufferBytesToGo = (null == a_Cookie.f_StartBuffer ? 0 : (a_Cookie.f_StartBuffer.Length - a_Cookie.f_StartBufferIndex));
                    ulong l_WriteOffset = l_FilePos;
                    ByteArray l_WriteData = l_Data;

                    if (0 < l_StartBufferBytesToGo)
                    {
                        int l_CurStartBytes = Math.Min(l_Data.f_Len, l_StartBufferBytesToGo);
                        Buffer.BlockCopy(l_Data.f_Arr, l_Data.f_Idx, a_Cookie.f_StartBuffer, a_Cookie.f_StartBufferIndex, l_CurStartBytes);

                        a_Cookie.f_StartBufferIndex += l_CurStartBytes;
                        l_WriteOffset += (uint)l_CurStartBytes;
                        l_WriteData = new ByteArray(l_Data.f_Arr, l_Data.f_Idx + l_CurStartBytes, l_Data.f_Len - l_CurStartBytes);
                    }

                    if (0 != l_WriteData.f_Len &&
                        false == l_DestFile.Write(l_WriteOffset, ref l_WriteData))
                    {
                        a_Cookie.Fail(ExtractionCookie.Result.DestFileAccessErr);
                        l_Status = OpStatus.IoError;

                        if (true == l_LastLoopRound)
                        {
                            // We've already released.

                            return l_Status;
                        }

                        goto RELEASE_AND_EXIT;
                    }

                    l_StartingOffset = 0;
                    l_FilePos += (uint)l_Data.f_Len;

                    continue;
                }

                CHUNKERR_RELEASE_AND_EXIT:

                if (OpStatus.Corruption == l_Status)
                {
                    OpStatus l_CorrStatus = HandleChunkCorruptionFromShrdOrExcl(i, l_ElmKey);
                }

                a_Cookie.Fail(ExtractionCookie.Result.IOError);

                RELEASE_AND_EXIT:

                f_Resource.ReleaseShared();

                return l_Status;
            }

            // We should have released already, so calling ReleaseAny()
            // here is not really necessary (just to be on the safe side).
            f_Resource.ReleaseAny();

            if (true == a_Cookie.f_FlushOnSuccess)
            {
                l_DestFile.Flush();
            }

            a_Cookie.Succeed(ExtractionCookie.Result.Success);

            return OpStatus.Done;
        }

        public OpStatus ContinueImport(ImportCookie a_Cookie)
        {
            uint l_ChunkIndex;
            int l_GapIndexHint;
            a_Cookie.GetPosition(out l_ChunkIndex, out l_GapIndexHint);

            // Take over eD2k hash set.

            if (0 == l_ChunkIndex &&
                true == a_Cookie.f_ImportFile.p_HasPartHashes &&
                false == HaseD2kHashSet() &&
                true == f_EstimatedKnowledge.p_HaseD2kHash &&
                true == f_HashStorage.HasThisHash(f_EstimatedKnowledge.p_eD2kHashIdx, a_Cookie.f_ImportFile.f_FileID.f_Val.ToByteArrayBE(), 0))
            {
                byte[] l_Hashs = a_Cookie.f_ImportFile.f_PartHashes;
                OpStatus l_HashSetStatus = StoreeD2kHashSet(UserNum.s_Import, (uint)l_Hashs.Length, l_Hashs, null);
            }

            // Import file body.

            FlexArray<FileRange> l_GapRanges;
            a_Cookie.f_ImportFile.ExposeGapRanges(out l_GapRanges);
            OpStatus l_Status = OpStatus.Done;

            for (uint l_ChunksImported = 0;;)
            {
                if (f_ChunkCount <= l_ChunkIndex)
                {
                    a_Cookie.Succeed(ImportCookie.Result.SuccessFinished);

                    break;
                }

                if (true == f_DoneMap.Get(l_ChunkIndex))
                {
                    l_ChunkIndex++;

                    continue;
                }

                // See if there is data we can use.

                uint l_ChunkSize = GetChunkSize(l_ChunkIndex);
                PieceBounds.FromGapRangesCreationOutputContext l_TranslCntxt;
                bool l_CreatedBounds = PieceBounds.CreateFromGapRanges(l_ChunkIndex, l_ChunkSize, ref l_GapIndexHint, ref l_GapRanges, out l_TranslCntxt);

                a_Cookie.NotifyElapsed(l_TranslCntxt.f_ElapsedByteCount);

                if (false == l_CreatedBounds)
                {
                    if (l_TranslCntxt.f_NextPossibleChunkIndex <= l_ChunkIndex)
                    {
                        // Should never happen (this is a safeguard).

                        Util.DebugBreak();

                        l_ChunkIndex++;
                    }
                    else
                    {
                       l_ChunkIndex = l_TranslCntxt.f_NextPossibleChunkIndex;

                        // l_ChunkIndex can now be uint.MaxValue, which
                        // will make us exit (above).
                    }

                    continue;
                }

                // Read the data we are interested in from the import
                // source file.

                ChunkRange l_Outline;
                bool l_GotOutline = l_TranslCntxt.f_Bounds.OutlineToChunkRange(out l_Outline);

                bool l_SrcFileAccessFailed = false;
                ByteArray l_Data = new ByteArray((int)l_Outline.f_Size);

                PieceBounds.Enumerator l_Enum;
                l_TranslCntxt.f_Bounds.BuildEnumerator(out l_Enum);
                ChunkRange l_Range;
                while (true == l_Enum.GetNextRange(out l_Range))
                {
                    ByteArray l_PartialData = new ByteArray(l_Data.f_Arr, (int)(l_Range.f_Offset - l_Outline.f_Offset), (int)l_Range.f_Size);
                    if (false == a_Cookie.f_SrcFile.Read(l_TranslCntxt.f_ChunkPos + l_Range.f_Offset, ref l_PartialData))
                    {
                        l_SrcFileAccessFailed = true;

                        break;
                    }
                }

                if (true == l_SrcFileAccessFailed)
                {
                    a_Cookie.Fail(ImportCookie.Result.FailedToAccessSrcFile);

                    break;
                }

                // Update the incomplete chunk.

                IncomplChunkHeader l_Header;
                UpdateFileChunkInputContext l_UpdateCntxt;
                UpdateFileChunkInputContext.CreateForImport(ref l_Outline, ref l_TranslCntxt.f_Bounds, ref l_Data, out l_UpdateCntxt);

                f_Resource.AcquireExclusive();
                {
                    l_Status = UpdateFileChunk(l_ChunkIndex, new ElmKey(), ref l_UpdateCntxt, out l_Header);
                }
                f_Resource.ReleaseExclusive();

                if (l_Status != OpStatus.Done)
                {
                    a_Cookie.Fail(ImportCookie.Result.FailedToUpdateChunk);

                    break;
                }

                OpStatus l_VerStatus = VerifyChunk(ref l_Header);

                // The chunk was updated.

                l_ChunkIndex++;
                a_Cookie.NotifyTranslation(l_TranslCntxt.f_TranslatedByteCount);

                if (a_Cookie.f_MaxChunksPerCall == ++l_ChunksImported)
                {
                    a_Cookie.Succeed(ImportCookie.Result.SuccessCallAgain);

                    break;
                }
            }

            a_Cookie.SetPosition(l_ChunkIndex, l_GapIndexHint);

            return l_Status;
        }

#region RangeSelection

        public OpStatus RegisterObtainCookie(ref ObtainContext a_Cntxt)
        {
            // Successfully registered cookies must be processed in
            // the same order they were registered. a_MaxBytes should
            // be well above PieceBounds.c_PieceSize.

            CookieStates l_ContentMngrState = a_Cntxt.f_Cookie.f_SourceCookie.p_ContentMngrState;

            if (CookieStates.Success != l_ContentMngrState &&
                CookieStates.WaitingForHashes != l_ContentMngrState)
            {
                // The source cookie is no longer valid.

                a_Cntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.OtherReason);

                return OpStatus.InputRejected;
            }

            if (VerificationMode.eD2k != f_VerificationMode &&
                VerificationMode.TAICH != f_VerificationMode)
            {
                a_Cntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.NeedHashesFirst);

                return OpStatus.NotReady;
            }

            int l_Idx;

            if (true == f_UserKnowledge.TryGet(a_Cntxt.f_Cookie.f_SourceCookie.f_UserNum, out l_Idx))
            {
                FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = f_UserKnowledge.ExposeItemArray();
                P2PHashSets l_HashSet = (VerificationMode.eD2k == f_VerificationMode ? P2PHashSets.eD2kHashSet : P2PHashSets.TAICHTree);
                int l_PeerHashIndex = l_Arr[l_Idx].f_Val.GetIndex(l_HashSet);

                if (0 <= l_PeerHashIndex &&
                    f_EstimatedKnowledge.GetIndex(l_HashSet) != l_PeerHashIndex)
                {
                    // This peer is known to have a different hash
                    // knowledge than we do, so don't dwell there.

                    a_Cntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.NoRelevantRange);

                    return OpStatus.NoMatch;
                }
            }

            uint l_ChunkIndex;
            PieceBounds l_Bounds;
            OpStatus l_Status = SuggestDownloadBounds(a_Cntxt.f_Cookie.p_UserNum, a_Cntxt.f_CompletedByUser, a_Cntxt.f_VolatileAvail, out l_ChunkIndex, out l_Bounds);

            if (OpStatus.Done != l_Status)
            {
                if (OpStatus.NoMatch == l_Status)
                {
                    a_Cntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.NoRelevantRange);
                }
                else
                {
                    a_Cntxt.f_Cookie.Rejected(ObtainCookie.RejectionReasons.OtherReason);
                }

                return l_Status;
            }

            l_Bounds.ReduceSize(a_Cntxt.f_MaxBytes);
            a_Cntxt.f_Cookie.SetBounds(this, l_ChunkIndex, ref l_Bounds);
            InsertIntoObtainList(a_Cntxt.f_Cookie);

            return OpStatus.Done;
        }

        private OpStatus SuggestDownloadBounds(UserNum a_UserNum,
                                               BitMap a_CompletedByUser,
                                               NetworkAvailability a_VolatileAvail,
                                               out uint a_ChunkIndex,
                                               out PieceBounds a_Bounds)
        {
            Dictionary<uint, bool> l_CheckedChunks = new Dictionary<uint, bool>();

            RemoveDeadCookiesFromObtainList();

            if (0 != (IncomplFileFlags.EnablePreview & f_IncomplFlags))
            {
                // Take care of the starting chunks of the file in order
                // to enable previewing.

                for (uint p = 0, lim = Math.Min(f_DoneMap.f_BitCount, c_MaxPreviewChunks); p < lim; p++)
                {
                    if (true == f_DoneMap.Get(p))
                    {
                        continue;
                    }

                    ElmKey l_ElmKey = GetHotIncomplChunksEntry(p);
                    OpStatus l_Stat = SuggestDownloadBoundsForChunk(a_UserNum, p, l_ElmKey, a_CompletedByUser, out a_Bounds);

                    if (OpStatus.NoMatch != l_Stat)
                    {
                        a_ChunkIndex = p;

                        return l_Stat;
                    }

                    l_CheckedChunks.Add(p, true);
                }
            }

            // Now try to continue dowloading one of the chunks that we
            // download currently.

            ObtainCookieEnumerator l_Enum = new ObtainCookieEnumerator(false, f_ObtainList.p_First);

            for (ObtainCookie l_Cookie; true == l_Enum.GetNext(out l_Cookie); )
            {
                if (l_Cookie.p_UserNum == a_UserNum &&
                    false == l_CheckedChunks.ContainsKey(l_Cookie.p_ChunkIndex))
                {
                    uint l_ChunkIndex = l_Cookie.p_ChunkIndex;

                    ElmKey l_ElmKey = GetHotIncomplChunksEntry(l_ChunkIndex);
                    OpStatus l_Stat = SuggestDownloadBoundsForChunk(a_UserNum, l_ChunkIndex, l_ElmKey, a_CompletedByUser, out a_Bounds);

                    if (OpStatus.NoMatch != l_Stat)
                    {
                        a_ChunkIndex = l_ChunkIndex;

                        return l_Stat;
                    }

                    l_CheckedChunks.Add(l_ChunkIndex, true);
                }
            }

            // Now try to make this user download some chunk we are
            // already trying to get from anyone.

            if (0 < f_HotIncomplChunks.Count)
            {
                // We need a clone, since it is possible that we write
                // on f_HotIncomplChunks while enumerating.

                Dictionary<uint, ElmKey> l_HotChunksClone = new Dictionary<uint, ElmKey>(f_HotIncomplChunks);

                foreach (KeyValuePair<uint, ElmKey> l_Pair in l_HotChunksClone)
                {
                    if (true == l_CheckedChunks.ContainsKey(l_Pair.Key))
                    {
                        continue;
                    }

                    OpStatus l_Stat = SuggestDownloadBoundsForChunk(a_UserNum, l_Pair.Key, l_Pair.Value, a_CompletedByUser, out a_Bounds);

                    if (OpStatus.NoMatch != l_Stat)
                    {
                        a_ChunkIndex = l_Pair.Key;

                        return l_Stat;
                    }

                    l_CheckedChunks.Add(l_Pair.Key, true);
                }
            }

            // Since nothing worked so far, we need to find a whole new
            // place. We'd prefer a weak spot in the network (where
            // availability is quite low), so that we can strengthen
            // availability.

            BitMap l_SrcMap;

            if (c_ChunkSize == a_CompletedByUser.f_Granularity)
            {
                // The input is in our native format.

                l_SrcMap = a_CompletedByUser.Clone();
            }
            else
            {
                // Transform the user's bitmap format into our native
                // format.

                l_SrcMap = f_DoneMap.RecreateEmpty();
                l_SrcMap.TakeOver(a_CompletedByUser, BitMap.TakeOverMode.RoundUp);
            }

            l_SrcMap.BuildRelevanceMap(f_DoneMap);

            // Now l_SrcMap contains set bits for every chunk we need
            // AND the user might be able to provide (this is not
            // exact since we needed to 'round' bits, and that way
            // we might have implemented some rounding errors).

            if (false == l_SrcMap.p_AllClear)
            {
                // Start scanning at a random position, to prevent that
                // some chunks get picked more often than others.
                uint l_BitBarrier = (Util.GetRandomUInt() % f_ChunkCount);
                FlexArray<ChunkAvail> l_Chunks;
                FlexArray<ChunkAvail>.Create((int)Math.Min(l_SrcMap.p_SetBitCount, c_MaxChunksPerAvailList), out l_Chunks);
                bool l_FirstRound = true;
                BitMap.EnumerateRunsContext l_EnumCntxt = new BitMap.EnumerateRunsContext(c_MaxChunksPerAvailList, 0, l_BitBarrier);

                while (true)
                {
                    bool l_CallAgain = l_SrcMap.CollectSetBitIndices(ref l_EnumCntxt, ref l_Chunks);
                    int l_ChunkCount = l_Chunks.p_Count;

                    if (0 < l_ChunkCount)
                    {
                        ChunkAvail[] l_Arr = l_Chunks.p_Arr;
                        for (int k = 0; k < l_ChunkCount; k++)
                        {
                            if (true == l_CheckedChunks.ContainsKey(l_Arr[k].f_ChunkIndex))
                            {
                                // We've already checked this chunk
                                // without success.

                                l_Arr[k].SetNetworkAvail(uint.MaxValue);
                                l_Arr[k].Invalidate();

                                continue;
                            }

                            uint l_Avail = a_VolatileAvail.CalcChunkAvailability(l_Arr[k].f_ChunkIndex);
                            l_Arr[k].SetNetworkAvail(l_Avail);
                        }

                        // We'll inspect those chunks with the worst
                        // availability first.
                        l_Chunks.Sort(CompareChunkAvails);

                        l_Arr = l_Chunks.p_Arr;
                        int l_RangeStart = 0;
                        for (int k = 0, lim = (l_ChunkCount - 1); k <= lim; k++)
                        {
                            if (false == l_Arr[k].p_Invalidated)
                            {
                                // Try to form a cluster of completed
                                // chunks. So look if the current chunk
                                // lies somewhere where we operated
                                // recently.

                                a_ChunkIndex = l_Arr[k].f_ChunkIndex;

                                if (true == IsNearHotIncomplChunk(a_ChunkIndex))
                                {
                                    OpStatus l_Stat = SuggestDownloadBoundsForChunk(a_UserNum, a_ChunkIndex, new ElmKey(), a_CompletedByUser, out a_Bounds);

                                    if (OpStatus.NoMatch != l_Stat)
                                    {
                                        return l_Stat;
                                    }

                                    // We may not change p_NetworkAvail
                                    // when invalidating, or the array's
                                    // order will be ruined.
                                    l_Arr[k].Invalidate();
                                }
                            }

                            bool l_EndReached = (k == lim);

                            if (false == l_EndReached &&
                                l_Arr[k].p_NetworkAvail == l_Arr[k + 1].p_NetworkAvail)
                            {
                                // The next chunk has the same
                                // availability, so take a look
                                // at that one.

                                continue;
                            }

                            // Now that we are about to leave a cluster
                            // of chunks with the same availability,
                            // try to pick any one.

                            for (; l_RangeStart <= k; l_RangeStart++)
                            {
                                if (true == l_Arr[l_RangeStart].p_Invalidated)
                                {
                                    continue;
                                }

                                a_ChunkIndex = l_Arr[l_RangeStart].f_ChunkIndex;

                                OpStatus l_Stat = SuggestDownloadBoundsForChunk(a_UserNum, a_ChunkIndex, new ElmKey(), a_CompletedByUser, out a_Bounds);

                                if (OpStatus.NoMatch != l_Stat)
                                {
                                    return l_Stat;
                                }
                            }
                        }
                    }

                    if (false == l_CallAgain)
                    {
                        if (false == l_FirstRound)
                        {
                            break;
                        }

                        // We just ran: [0; l_BitBarrier[
                        // Now run: [l_BitBarrier; f_ChunkCount[

                        l_FirstRound = false;
                        l_EnumCntxt = new BitMap.EnumerateRunsContext(c_MaxChunksPerAvailList, l_BitBarrier, f_ChunkCount);
                    }

                    l_Chunks.Clear();
                }
            }

            a_ChunkIndex = 0;
            a_Bounds = new PieceBounds();

            return OpStatus.NoMatch;
        }

        private static int CompareChunkAvails(ref ChunkAvail a_A,
                                              ref ChunkAvail a_B)
        {
            return a_A.CompareTo(ref a_B);
        }

        private OpStatus SuggestDownloadBoundsForChunk(UserNum a_UserNum,
                                                       uint a_ChunkIndex,
                                                       ElmKey a_HintedKey,
                                                       BitMap a_CompletedByUser,
                                                       out PieceBounds a_BoundsPossible)
        {
            // a_HintedKey may be zero. This routine should return an
            // error code other than OpStatus.NoMatch only on io
            // failure.

            if (false == a_CompletedByUser.ToPieceBounds(a_ChunkIndex, GetChunkSize(a_ChunkIndex), out a_BoundsPossible) ||
                false == a_BoundsPossible.p_IsSet)
            {
                return OpStatus.NoMatch;
            }

            ObtainCookie l_FirstCookie = FindFirstCookieOfChunk(a_ChunkIndex);

            if (null != l_FirstCookie)
            {
                // Prevent this chunk from becoming too crowded.

                ObtainCookieEnumerator l_UsersEnum = new ObtainCookieEnumerator(true, l_FirstCookie);
                ObtainCookie l_Cookie;
                List<UserNum> l_ActiveUsers = new List<UserNum>(c_MaxUnmergedUsersPerIncomplChunk - 1);
                while (true == l_UsersEnum.GetNext(out l_Cookie))
                {
                    UserNum l_UserNum = l_Cookie.p_UserNum;

                    if (false == l_ActiveUsers.Contains(l_UserNum))
                    {
                        if (c_MaxUnmergedUsersPerIncomplChunk <= l_ActiveUsers.Count)
                        {
                            // Limit reached.

                            return OpStatus.NoMatch;
                        }

                        l_ActiveUsers.Add(l_UserNum);
                    }
                }
            }

            OpStatus l_Status;
            IncomplChunkHeader l_Header;

            f_Resource.AcquireShared();
            {
                l_Status = ReadIncomplChunkHeaderWithKey(a_ChunkIndex, a_HintedKey, false, out l_Header);
            }
            f_Resource.ReleaseShared();

            if (OpStatus.Done != l_Status &&
                OpStatus.ChunkHasNoKey != l_Status)
            {
                return (OpStatus.IoError == l_Status ? OpStatus.IoError : OpStatus.NoMatch);
            }

            bool l_CalledVerifyChunk = false;

            BOUNDS_ANALYSIS:

            if (false == l_Header.IsValid())
            {
                if (null == l_FirstCookie)
                {
                    // This chunk doesn't really exist yet.

                    return OpStatus.Done;
                }
            }
            else
            {
                if (VerificationMode.TAICH == f_VerificationMode &&
                    true == l_Header.IsMarkedDisputed() &&
                    false == l_Header.HasTAICHBlockHashes(false))
                {
                    RequestTAICHBlocks(l_Header.f_ChunkIndex);

                    return OpStatus.NoMatch;
                }
            }

            PieceBounds l_BoundsDone;
            bool l_ChunkIsUndisputedAndFilled;

            ObtainCookieEnumerator l_BoundsEnum = new ObtainCookieEnumerator(true, l_FirstCookie);
            CalcCoveredBounds(a_UserNum, ref l_BoundsEnum, ref l_Header, out l_BoundsDone, out l_ChunkIsUndisputedAndFilled);

            a_BoundsPossible.Sub(ref l_BoundsDone, PieceBounds.SubtractionMode.MinimalRemoval);

            if (true == a_BoundsPossible.p_IsSet)
            {
                return OpStatus.Done;
            }

            if (true == l_ChunkIsUndisputedAndFilled &&
                false == l_CalledVerifyChunk)
            {
                // This chunk is filled but not verified. This usually
                // happens after we switch from eD2k to TAICH hash set.

                OpStatus l_VerStatus = VerifyChunk(ref l_Header);

                if (OpStatus.ChunkNotVerified == l_VerStatus)
                {
                    // Take a second look.

                    l_CalledVerifyChunk = true;

                    goto BOUNDS_ANALYSIS;
                }
            }

            return OpStatus.NoMatch;
        }

        private void CalcCoveredBounds(UserNum a_UserNum,
                                       ref ObtainCookieEnumerator a_Enum,
                                       ref IncomplChunkHeader a_Header,
                                       out PieceBounds a_BoundsDone,
                                       out bool a_ChunkIsUndisputedAndFilled)
        {
            // a_Header may be invalid. On return, a_ChunkIsUndisputedAndFilled
            // does not account for requested ranges (contrary to
            // a_BoundsDone).

            a_BoundsDone = new PieceBounds();
            a_ChunkIsUndisputedAndFilled = false;
            bool l_Disputed = false;

            if (true == a_Header.IsValid())
            {
                l_Disputed = a_Header.IsMarkedDisputed();

                if (true == a_Header.HasTAICHBlockHashes(false) &&
                    true == a_Header.f_MergedUser.IsSet())
                {
                    a_BoundsDone.Add(a_Header.f_MergedUser.f_VerifiedMap);
                }

                if (false == l_Disputed)
                {
                    UserHeader[] l_Users = a_Header.f_Users.p_Arr;
                    for (int i = (a_Header.f_Users.p_Count - 1); 0 <= i; i--)
                    {
                        a_BoundsDone.Add(ref l_Users[i].f_Bounds);
                    }

                    if (a_BoundsDone.GetCoveredBytes() == a_Header.f_ChunkSize)
                    {
                        a_ChunkIsUndisputedAndFilled = true;
                    }
                }
                else
                {
                    int l_HeaderIdx;
                    if (true == a_Header.GetHeaderIndex(a_UserNum, out l_HeaderIdx))
                    {
                        a_BoundsDone.Add(ref a_Header.f_Users.p_Arr[l_HeaderIdx].f_Bounds);
                    }
                }
            }

            a_Enum.Reset();

            ObtainCookie l_Cookie;
            while (true == a_Enum.GetNext(out l_Cookie))
            {
                if (false == l_Disputed ||
                    l_Cookie.p_UserNum == a_UserNum)
                {
                    a_BoundsDone.Add(ref l_Cookie.f_Bounds);
                }
            }
        }

#endregion

#region ObtainCookieHandling

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ObtainCookieEnumerator
        {
            // Don't use the struct default constructor.

            private readonly bool f_RestrictToFirstChunksUser;
            private readonly ObtainCookie f_First;
            private ObtainCookie f_Current;

            public ObtainCookieEnumerator(bool a_RestrictToFirstChunksUser,
                                          ObtainCookie a_First)
            {
                // a_First may be null.

                f_RestrictToFirstChunksUser = a_RestrictToFirstChunksUser;
                f_First = a_First;
                f_Current = f_First;
            }

            public void Reset()
            {
                f_Current = f_First;
            }

            public bool GetNext(out ObtainCookie a_Cookie)
            {
                // After successful return, the node referenced by the
                // returned a_Cookie can be safely removed from the
                // list even while enumerating (but don't reset the
                // enumeration after any node removal, create a new
                // one instead).

                if (null == f_Current)
                {
                    a_Cookie = null;

                    return false;
                }

                if (true == f_RestrictToFirstChunksUser &&
                    f_First.p_ChunkIndex != f_Current.p_ChunkIndex)
                {
                    a_Cookie = null;
                    f_Current = null;

                    return false;
                }

                a_Cookie = f_Current;
                f_Current = f_Current.p_Next;

                return true;
            }
        }

        private void InsertIntoObtainList(ObtainCookie a_Cookie)
        {
            uint l_ChunkIndex = a_Cookie.p_ChunkIndex;
            bool l_Handled = false;

            for (ObtainCookie l_Cookie = f_ObtainList.p_First; null != l_Cookie; l_Cookie = l_Cookie.p_Next)
            {
                if (l_Cookie.p_ChunkIndex >= l_ChunkIndex)
                {
                    l_Handled = true;

                    f_ObtainList.AddBefore(l_Cookie, a_Cookie);

                    break;
                }
            }

            if (false == l_Handled)
            {
                f_ObtainList.AddLast(a_Cookie);
            }

            CreateNewHotIncomplChunksEntryIfNoneExists(l_ChunkIndex);
        }

        private void RemoveFromObtainListCore(ObtainCookie a_Cookie)
        {
            // Should only be called by RejectAndRemoveFromObtainList()
            // and FinishAndRemoveFromObtainList().

            uint l_ChunkIndex = a_Cookie.p_ChunkIndex;

            if ((null == a_Cookie.p_Previous ||
                 a_Cookie.p_Previous.p_ChunkIndex != l_ChunkIndex) &&
                (null == a_Cookie.p_Next ||
                 a_Cookie.p_Next.p_ChunkIndex != l_ChunkIndex))
            {
                // This chunk is inactive now.

                bool l_Removed = f_HotIncomplChunks.Remove(l_ChunkIndex);
                f_HotIncomplRevision++;
            }

            f_ObtainList.Remove(a_Cookie);
        }

        private void RejectAndRemoveFromObtainList(ObtainCookie a_Cookie,
                                                   ObtainCookie.RejectionReasons a_Reason)
        {
            a_Cookie.Rejected(a_Reason);
            RemoveFromObtainListCore(a_Cookie);
        }

        private void RejectAndRemoveFromObtainList(ObtainCookie a_Cookie)
        {
            RejectAndRemoveFromObtainList(a_Cookie, ObtainCookie.RejectionReasons.OtherReason);
        }

        private void FinishAndRemoveFromObtainList(ObtainCookie a_Cookie)
        {
            a_Cookie.Finished();

            RemoveFromObtainListCore(a_Cookie);
        }

        private void RemoveDeadCookiesFromObtainList()
        {
            ObtainCookie l_Cookie = f_ObtainList.p_First;
            ObtainCookieEnumerator l_Enum = new ObtainCookieEnumerator(false, l_Cookie);

            while (l_Enum.GetNext(out l_Cookie))
            {
                if (CookieStates.Failure == l_Cookie.p_PeerState)
                {
                    RejectAndRemoveFromObtainList(l_Cookie);
                }
            }
        }

        private ObtainCookie FindFirstCookieOfChunk(uint a_ChunkIndex)
        {
            // Returns the first obtain cookie concerning a certain
            // chunk or null if this chunk isn't active.

            for (ObtainCookie l_Cookie = f_ObtainList.p_First; null != l_Cookie; l_Cookie = l_Cookie.p_Next)
            {
                if (l_Cookie.p_ChunkIndex >= a_ChunkIndex)
                {
                    if (l_Cookie.p_ChunkIndex == a_ChunkIndex)
                    {
                        return l_Cookie;
                    }

                    break;
                }
            }

            return null;
        }

        public bool IsNearHotIncomplChunk(uint a_ChunkIndex)
        {
            return ((0 != a_ChunkIndex &&
                     true == f_HotIncomplChunks.ContainsKey(a_ChunkIndex - 1)) ||
                    (f_ChunkCount > (a_ChunkIndex + 1) &&
                     true == f_HotIncomplChunks.ContainsKey(a_ChunkIndex + 1)));
        }

        private void StopDownloadsOfChunk(uint a_ChunkIndex)
        {
            // Stops all downloads concerning a certain chunk.

            ObtainCookie l_Cookie = FindFirstCookieOfChunk(a_ChunkIndex);
            ObtainCookieEnumerator l_Enum = new ObtainCookieEnumerator(true, l_Cookie);

            while (true == l_Enum.GetNext(out l_Cookie))
            {
                RejectAndRemoveFromObtainList(l_Cookie);
            }
        }

        private void StopDownloadsFromBadUser(UserNum a_TargetUserNum)
        {
            ObtainCookie l_Cookie = f_ObtainList.p_First;
            ObtainCookieEnumerator l_Enum = new ObtainCookieEnumerator(false, l_Cookie);

            while (true == l_Enum.GetNext(out l_Cookie))
            {
                if (l_Cookie.p_UserNum == a_TargetUserNum)
                {
                    RejectAndRemoveFromObtainList(l_Cookie, ObtainCookie.RejectionReasons.BadSource);
                }
            }
        }

        private void StopDownloadsThroughCookie(SourceCookie a_TargetCookie)
        {
            // Stops all downloads that stem from a given cookie.

            ObtainCookie l_Cookie = f_ObtainList.p_First;
            ObtainCookieEnumerator l_Enum = new ObtainCookieEnumerator(false, l_Cookie);

            while (true == l_Enum.GetNext(out l_Cookie))
            {
                if (l_Cookie.f_SourceCookie == a_TargetCookie)
                {
                    RejectAndRemoveFromObtainList(l_Cookie);
                }
            }
        }

#endregion

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct UpdateFileChunkInputContext
        {
            // Don't use the struct default constructor.

            public readonly Mode f_Mode;
            public readonly UserNum f_SrcUserNum;
            public readonly ChunkRange f_Range;
            // f_Bounds is only valid if f_Mode is ImportWithPrecomputedBounds.
            public PieceBounds f_Bounds;
            public ByteArray f_Data;

            public bool p_Import
            {
                get { return (Mode.ImportWithPrecomputedBounds == f_Mode); }
            }

            private UpdateFileChunkInputContext(UserNum a_SrcUserNum,
                                                uint a_InChunkOffset,
                                                ref ByteArray a_Data)
            {
                f_Mode = Mode.NormalDownloadInput;
                f_SrcUserNum = a_SrcUserNum;
                f_Range = new ChunkRange(a_InChunkOffset, (uint)a_Data.f_Len);
                f_Bounds = new PieceBounds();
                f_Data = a_Data;
            }

            private UpdateFileChunkInputContext(ref ChunkRange a_Range,
                                                ref PieceBounds a_Bounds,
                                                ref ByteArray a_Data)
            {
                f_Mode = Mode.ImportWithPrecomputedBounds;
                f_SrcUserNum = UserNum.s_Import;
                f_Range = a_Range;
                f_Bounds = a_Bounds;
                f_Data = a_Data;
            }

            public static void CreateDefault(UserNum a_SrcUserNum,
                                             uint a_InChunkOffset,
                                             ref ByteArray a_Data,
                                             out UpdateFileChunkInputContext a_Dest)
            {
                a_Dest = new UpdateFileChunkInputContext(a_SrcUserNum, a_InChunkOffset, ref a_Data);
            }

            public static void CreateForImport(ref ChunkRange a_Range,
                                               ref PieceBounds a_Bounds,
                                               ref ByteArray a_Data,
                                               out UpdateFileChunkInputContext a_Dest)
            {
                a_Dest = new UpdateFileChunkInputContext(ref a_Range, ref a_Bounds, ref a_Data);
            }

            public void RetrieveBounds(bool a_Add,
                                       ref PieceBounds a_DestBounds)
            {
                if (false == p_Import)
                {
                    ChunkRange l_Range = f_Range;
                    a_DestBounds.Add(ref l_Range);
                }
                else
                {
                    if (false == a_Add)
                    {
                        a_DestBounds = f_Bounds;
                    }
                    else
                    {
                        a_DestBounds.Add(ref f_Bounds);
                    }
                }
            }

            public enum Mode : byte
            {
                NormalDownloadInput = 0,
                ImportWithPrecomputedBounds = 1,
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct MainHeader
        {
            public const ushort c_Size = c_HeaderSize;
            public const ushort c_BitmapStartingOffset = c_Size;
            public const byte c_ChunkKeyFieldAlignment = 16;
            private const byte c_BytesInIncomplChunksOffset = 56;

            // [c_DefaultHeaderSize]
            // File size, ulong at 24.

            public ElmKey f_UsersKey; // 32
            public ElmKey f_eD2kHashSetKey; // 40
            public ElmKey f_TAICHUpperTreeKey; // 48

            public ulong f_BytesInIncomplChunks; // 56

            public MainHeader(byte[] a_Src,
                              int a_SrcIdx)
            {
                f_UsersKey = new ElmKey(Util.ReadUInt64LE(a_Src, a_SrcIdx + 32));
                f_eD2kHashSetKey = new ElmKey(Util.ReadUInt64LE(a_Src, a_SrcIdx + 40));
                f_TAICHUpperTreeKey = new ElmKey(Util.ReadUInt64LE(a_Src, a_SrcIdx + 48));
                f_BytesInIncomplChunks = Util.ReadUInt64LE(a_Src, a_SrcIdx + c_BytesInIncomplChunksOffset);
            }

            public void Write(FileNum a_FileNum,
                              ulong a_FileSize,
                              byte[] a_Dest,
                              int a_DestIdx)
            {
                IncompleteFile.WriteDefaultHeader(c_MainMagic, a_FileNum, a_Dest, a_DestIdx);
                Util.WriteUInt64LE(a_FileSize, a_Dest, a_DestIdx + 24);
                Util.WriteUInt64LE(f_UsersKey.f_Val, a_Dest, a_DestIdx + 32);
                Util.WriteUInt64LE(f_eD2kHashSetKey.f_Val, a_Dest, a_DestIdx + 40);
                Util.WriteUInt64LE(f_TAICHUpperTreeKey.f_Val, a_Dest, a_DestIdx + 48);
                Util.WriteUInt64LE(f_BytesInIncomplChunks, a_Dest, a_DestIdx + c_BytesInIncomplChunksOffset);

                // Zero unused bytes.
                Buffer2.BlockZero(a_Dest, a_DestIdx + 64, c_Size - 64);
            }

            public bool WriteBytesInIncomplChunks(HardMemory a_HardMem,
                                                  HardMemory.Element a_Elm,
                                                  byte[] a_Dest)
            {
                Util.WriteUInt64LE(f_BytesInIncomplChunks, a_Dest, 0);
                ByteArray l_Data = new ByteArray(a_Dest, 0, sizeof(ulong));

                return a_HardMem.Write(a_Elm, c_BytesInIncomplChunksOffset, ref l_Data);
            }

            public static ulong ReadFileSizeFromHeaderBuffer(byte[] a_Src,
                                                             int a_SrcIdx)
            {
                return Util.ReadUInt64LE(a_Src, a_SrcIdx + 24);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct UserKnowledgeHeader
        {
            public const ushort c_ShortSize = 20;
            public const ushort c_LargeSize = c_HeaderSize;
            public const ushort c_OwnKnowledgeOffset = 32;
            public const ushort c_EstimatedKnowledgeOffset = 128;
            public const ushort c_HashStorageOffset = c_LargeSize;

            // [c_DefaultHeaderSize]
            public readonly uint f_UsersCount; // 16

            public UserKnowledgeHeader(uint a_UsersCount)
            {
                f_UsersCount = a_UsersCount;
            }

            public static bool Load(byte[] a_Src,
                                    int a_SrcIdx,
                                    out UserKnowledgeHeader a_Dest)
            {
                uint l_UsersCount = Util.ReadUInt32LE(a_Src, a_SrcIdx + 16);
                a_Dest = new UserKnowledgeHeader(l_UsersCount);

                return true;
            }

            public void Write(FileNum a_FileNum,
                              byte[] a_Dest,
                              int a_DestIdx,
                              bool a_ZeroUnusedBytes)
            {
                if (true == a_ZeroUnusedBytes)
                {
                    Buffer2.BlockZero(a_Dest, a_DestIdx, UserKnowledgeHeader.c_ShortSize);
                }

                WriteDefaultHeader(c_UsersMagic, a_FileNum, a_Dest, a_DestIdx);

                Util.WriteUInt32LE(f_UsersCount, a_Dest, a_DestIdx + 16);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct HashHeader
        {
            public const ushort c_Size = c_HeaderSize;
            public const ushort c_BodyStartingOffset = c_HeaderSize;
            public const ushort c_DigestOffset = 64;

            // [c_DefaultHeaderSize]
            public readonly UserNum f_FromUserNum; // 16
            public readonly P2PHashSets f_Type; // 24
            public readonly ulong f_BodySize; // 32

            public HashHeader(UserNum a_FromUserNum,
                              P2PHashSets a_Type,
                              uint a_BodySize)
            {
                f_FromUserNum = a_FromUserNum;
                f_Type = a_Type;
                f_BodySize = a_BodySize;
            }

            public static bool Load(byte[] a_Src,
                                    int a_SrcIdx,
                                    out HashHeader a_Dest)
            {
                uint l_TypeAsUInt = Util.ReadUInt32LE(a_Src, a_SrcIdx + 24);
                ulong l_BodySize = Util.ReadUInt64LE(a_Src, a_SrcIdx + 32);

                if (((uint)P2PHashSets.eD2kHashSet != l_TypeAsUInt &&
                     (uint)P2PHashSets.TAICHTree != l_TypeAsUInt) ||
                    uint.MaxValue < l_BodySize)
                {
                    a_Dest = new HashHeader();

                    return false;
                }

                UserNum l_FromUserNum = new UserNum(Util.ReadUInt64LE(a_Src, a_SrcIdx + 16));
                P2PHashSets l_Type = (P2PHashSets)l_TypeAsUInt;

                a_Dest = new HashHeader(l_FromUserNum, l_Type, (uint)l_BodySize);

                return true;
            }

            public void Write(FileNum a_FileNum,
                              ref HashStorage a_HashStorage,
                              ref UserFileKnowledge a_Knowledge,
                              byte[] a_Dest,
                              int a_DestIdx,
                              bool a_ZeroUnusedBytes)
            {
                WriteDefaultHeader(GetMagic(), a_FileNum, a_Dest, a_DestIdx);

                Util.WriteUInt64LE(f_FromUserNum.f_Val, a_Dest, a_DestIdx + 16);
                Util.WriteUInt32LE((uint)f_Type, a_Dest, a_DestIdx + 24);
                Util.WriteUInt64LE(f_BodySize, a_Dest, a_DestIdx + 32);

                // Finally copy the digest.

                byte[] l_SrcBuf;
                int l_SrcLen;

                switch (f_Type)
                {
                    case P2PHashSets.eD2kHashSet:
                        l_SrcBuf = a_HashStorage.ExposeHash(a_Knowledge.p_eD2kHashIdx);
                        l_SrcLen = MD_4.c_HashByteSize;
                        break;

                    case P2PHashSets.TAICHTree:
                        l_SrcBuf = a_HashStorage.ExposeHash(a_Knowledge.p_TAICHHashIdx);
                        l_SrcLen = Tiger.c_HashByteSize;
                        break;

                    default:
                        throw (new ApplicationException());
                }

                Buffer.BlockCopy(l_SrcBuf, 0, a_Dest, a_DestIdx + c_DigestOffset, l_SrcLen);

                if (true == a_ZeroUnusedBytes)
                {
                    int l_Filled = (c_DigestOffset + l_SrcLen);
                    Buffer2.BlockZero(a_Dest, a_DestIdx + l_Filled, HashHeader.c_Size - l_Filled);
                }
            }

            public uint GetMagic()
            {
                uint l_Magic;
                if (true == GetMagic(f_Type, out l_Magic))
                {
                    return l_Magic;
                }

                throw (new ApplicationException());
            }

            public static bool GetMagic(P2PHashSets a_Type,
                                        out uint a_Magic)
            {
                switch (a_Type)
                {
                    case P2PHashSets.eD2kHashSet:
                        a_Magic = c_eD2kHashMagic;
                        return true;

                    case P2PHashSets.TAICHTree:
                        a_Magic = c_TAICHMagic;
                        return true;

                    default:
                        a_Magic = 0;
                        return false;
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ComplChunkHeader
        {
            public const ushort c_Size = c_HeaderSize;
            public const ushort c_BodyStartingOffset = IncomplChunkHeader.c_BodyStartingOffset;
            private const ushort c_HashSetOffset = 36;
            private const ushort c_MasterHashOffset = 40;
            private const ushort c_MaxMasterHashSize = (c_HeaderSize - c_MasterHashOffset);

            // [c_DefaultChunkHeaderSize]
            public readonly uint f_ChunkIndex;
            public readonly uint f_ChunkSize;

            public Time f_CreationTime; // 24
            public uint f_ContentAdler32; // 32

            // The attributes used for verification.
            // P2PHashSets at 36.
            // Master hash at 40.

            // f_Element can be null.
            private HardMemory.Element f_Element;

            public HardMemory.Element p_Element
            {
                get { return f_Element; }
            }

            public ComplChunkHeader(Instance a_Inst,
                                    uint a_ChunkIndex,
                                    uint a_ChunkSize,
                                    uint a_ContentAdler32)
            {
                f_ChunkIndex = a_ChunkIndex;
                f_ChunkSize = a_ChunkSize;

                f_CreationTime = a_Inst.GetTime();
                f_ContentAdler32 = a_ContentAdler32;

                f_Element = null;
            }

            private ComplChunkHeader(uint a_ChunkIndex,
                                     uint a_ChunkSize)
            {
                // Constructor for loading.

                f_ChunkIndex = a_ChunkIndex;
                f_ChunkSize = a_ChunkSize;

                f_CreationTime = new Time();
                f_ContentAdler32 = 0;

                f_Element = null;
            }

            public static OpStatus Load(uint a_ChunkIndex,
                                        uint a_ChunkSize,
                                        P2PHashSets a_ExpectedHashSet,
                                        byte[] a_ExpectedMasterHash,
                                        byte[] a_Src,
                                        int a_SrcIdx,
                                        out ComplChunkHeader a_Dest)
            {
                if (a_ExpectedHashSet != (P2PHashSets)Util.ReadUInt16LE(a_Src, a_SrcIdx + c_HashSetOffset) ||
                    false == Buffer2.ValueEquals(a_Src, a_SrcIdx + c_MasterHashOffset, a_ExpectedMasterHash, 0, a_ExpectedMasterHash.Length))
                {
                    a_Dest = new ComplChunkHeader();

                    return OpStatus.NeedReverification;
                }

                a_Dest = new ComplChunkHeader(a_ChunkIndex, a_ChunkSize);

                a_Dest.f_CreationTime = Time.FromBinary(Util.ReadUInt64LE(a_Src, a_SrcIdx + 24));
                a_Dest.f_ContentAdler32 = Util.ReadUInt32LE(a_Src, a_SrcIdx + 32);

                return OpStatus.Done;
            }

            public bool IsValid()
            {
                return (null != p_Element);
            }

            public void SetElement(HardMemory.Element a_Element)
            {
                f_Element = a_Element;
            }

            public void Write(FileNum a_FileNum,
                              byte[] a_Dest,
                              int a_DestIdx,
                              P2PHashSets a_VerificationHashSet,
                              byte[] a_VerificationMasterHash)
            {
                // We assume that there is no need to zero anything.

                IncompleteFile.WriteDefaultChunkHeader(c_ComplChunkMagic, a_FileNum, f_ChunkIndex, a_Dest, a_DestIdx);

                Util.WriteUInt64LE(f_CreationTime.ToBinary(), a_Dest, a_DestIdx + 24);
                Util.WriteUInt32LE(f_ContentAdler32, a_Dest, a_DestIdx + 32);

                Util.WriteUInt16LE((ushort)a_VerificationHashSet, a_Dest, a_DestIdx + c_HashSetOffset);
                Buffer.BlockCopy(a_VerificationMasterHash, 0, a_Dest, a_DestIdx + c_MasterHashOffset, a_VerificationMasterHash.Length);
            }
        }

        [Flags]
        private enum ChunkFlags : uint
        {
            None = 0,
            TAICHBlockHashesPresent = 0x1,
            // Means that validation of received content failed
            // already.
            Disputed = 0x2,
        }

        private enum VerificationMode : byte
        {
            // Items are not combinable.

            None = 0,

            eD2k = 1,
            TAICH = 2,
        }

        private enum CorruptDataObserver : byte
        {
            // Items are not combinable.

            HashDelivery = 0,
            eD2kVerification = 1,
            TAICHVerification = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct IncomplChunkHeader
        {
            // Don't use the struct default constructor.

            public const ushort c_ShortSize = c_HeaderSize;
            public const ushort c_Size = (2 * c_HeaderSize);
            public const ushort c_BodyStartingOffset = (4 * 1024);
            private const ushort c_TAICHTopHashOffset = 32;
            private const ushort c_Challenge1Offset = 56;
            private const ushort c_Challenge2Offset = 512;
            private const ushort c_TAICHBlockHashesOffset = 520;
            private const ushort c_IncomplFlagsOffset = 116;
            private const ushort c_MergedUserHeadersOffset = 120;
            private const ushort c_UserIncomplHeadersOffset = 128;
            public const byte c_MaxUserHeaders = ((c_HeaderSize - c_UserIncomplHeadersOffset) / UserHeader.c_Size);

            public readonly uint f_ChunkIndex;
            public readonly uint f_ChunkSize;
            public readonly uint f_BlockCount;
            public readonly uint f_BlockMask;
            public readonly uint f_FirstHashIndex;
            private uint f_CoveredBytes;

            // [c_DefaultChunkHeaderSize]
            // Tiger hash at 32.
            // Block challenge at 56.

            public ChunkFlags f_ChunkFlags; // 116
            public MergedUserHeader f_MergedUser; // 120
            public FlexArray<UserHeader> f_Users; // 128

            private TigerHashHive f_TAICHBlockHashes; // 520
            // f_TAICHChunkCheckReq exists in memory only.
            private TAICHChunkCheckReqs f_TAICHChunkCheckReq;

            // f_Element can be null.
            private HardMemory.Element f_Element;

            public HardMemory.Element p_Element
            {
                get {return f_Element; }
            }

            public IncomplChunkHeader(uint a_ChunkIndex,
                                      uint a_ChunkSize)
            {
                f_ChunkIndex = a_ChunkIndex;
                f_ChunkSize = a_ChunkSize;
                DoDefaultInit(f_ChunkIndex, f_ChunkSize, out f_BlockCount, out f_BlockMask, out f_FirstHashIndex);

                f_CoveredBytes = 0;
                f_ChunkFlags = ChunkFlags.None;
                f_MergedUser = new MergedUserHeader();
                FlexArray<UserHeader>.Create(out f_Users);
                f_TAICHBlockHashes = new TigerHashHive();
                f_TAICHChunkCheckReq = TAICHChunkCheckReqs.Unknown;
                f_Element = null;
            }

            public static bool Load(uint a_ChunkIndex,
                                    uint a_ChunkSize,
                                    HardMemory.Element a_Element,
                                    ref TAICHTreeLayout a_Layout,
                                    ref HashStorage a_HashStorage,
                                    ref UserFileKnowledge a_Knowledge,
                                    bool a_LoadTAICHBlockData,
                                    int a_GreatestUpperLevel,
                                    byte[] a_Src,
                                    int a_SrcIdx,
                                    out IncomplChunkHeader a_Dest)
            {
                // If a_LoadTAICHBlockData is true, a_Src needs to have
                // an accessable range of c_Size bytes, otherwise only
                // c_ShortSize bytes are required.

                a_Dest = new IncomplChunkHeader(a_ChunkIndex, a_ChunkSize);
                a_Dest.f_ChunkFlags = (ChunkFlags)Util.ReadUInt32LE(a_Src, a_SrcIdx + c_IncomplFlagsOffset);

                bool l_AcceptBlockData = false;

                if (0 <= a_Knowledge.p_TAICHHashIdx)
                {
                    ByteArray l_Data = new ByteArray(a_Src, a_SrcIdx + c_TAICHTopHashOffset, Tiger.c_HashByteSize);
                    bool l_HasTAICH = a_HashStorage.IsHash(a_Knowledge.p_TAICHHashIdx, ref l_Data);

                    if (true == l_HasTAICH &&
                        0 < a_GreatestUpperLevel &&
                        0 != (ChunkFlags.TAICHBlockHashesPresent & a_Dest.f_ChunkFlags))
                    {
                        if (false == a_LoadTAICHBlockData)
                        {
                            l_AcceptBlockData = true;
                        }
                        else if (true == Buffer2.ValueEquals(a_Src, c_Challenge1Offset, a_Src, c_Challenge2Offset, sizeof(ulong)))
                        {
                            List<uint> l_Indices = TAICHTree.CalcSubTreeIndices(ref a_Layout, c_TAICHExponent, a_Dest.f_ChunkIndex, a_GreatestUpperLevel);
                            byte[] a_Bytes = Util.CopyBytes(a_Src, c_TAICHBlockHashesOffset, l_Indices.Count * Tiger.c_HashByteSize);

                            l_AcceptBlockData = true;
                            TigerHashHive.CreateReadOnly(l_Indices, a_Bytes, out a_Dest.f_TAICHBlockHashes);
                            a_Dest.f_TAICHChunkCheckReq = (a_Dest.f_BlockCount == l_Indices.Count ?
                                                           TAICHChunkCheckReqs.NoExtraHashesNeeded : TAICHChunkCheckReqs.ExtraHashesNeeded);
                        }
                    }
                }

                if (false == l_AcceptBlockData)
                {
                    a_Dest.f_ChunkFlags &= ~ChunkFlags.TAICHBlockHashesPresent;
                }

                a_Dest.f_MergedUser = new MergedUserHeader(a_Src, a_SrcIdx + c_MergedUserHeadersOffset);
                int l_HeaderCount = a_Src[a_SrcIdx + c_MergedUserHeadersOffset + 4];

                int l_MergedSlot = a_Dest.GetMergedSlot();
                int l_MaxSlots = CalcSlotsFromElementSize(a_Element.f_MaxSize, a_Dest.f_ChunkSize);

                if (l_MergedSlot >= l_MaxSlots ||
                    c_MaxUserHeaders < l_HeaderCount)
                {
                    a_Dest = new IncomplChunkHeader();

                    return false;
                }

                a_Dest.f_Users.ReserveRange(0, l_HeaderCount);

                int l_Pos = (a_SrcIdx + c_UserIncomplHeadersOffset);
                UserHeader[] l_Arr = a_Dest.f_Users.p_Arr;
                for (int i = 0; i < l_HeaderCount; i++)
                {
                    bool l_OK = UserHeader.Load(a_Src, l_Pos, l_AcceptBlockData, out l_Arr[i]);

                    int l_Slot = l_Arr[i].p_Slot;

                    l_OK = (true == l_OK &&
                            l_MaxSlots > l_Slot &&
                            l_MergedSlot != l_Slot);

                    if (true == l_OK)
                    {
                        for (int k = 0; k < i; k++)
                        {
                            if (l_Arr[k].f_UserNum == l_Arr[i].f_UserNum)
                            {
                                l_OK = false;

                                break;
                            }

                            if (l_Arr[k].p_Slot == l_Slot &&
                                true == l_Arr[k].f_Bounds.DoesOverlap(ref l_Arr[i].f_Bounds))
                            {
                                l_OK = false;

                                break;
                            }
                        }
                    }

                    if (false == l_OK)
                    {
                        a_Dest = new IncomplChunkHeader();

                        return false;
                    }

                    if (false == l_AcceptBlockData)
                    {
                        l_Arr[i].f_DirtyMap.TakeOver(a_Dest.f_BlockMask);
                    }
                    else
                    {
                        l_Arr[i].f_DirtyMap.RestrictTo(a_Dest.f_BlockMask);
                        l_Arr[i].f_AcceptedMap.RestrictTo((~l_Arr[i].f_DirtyMap.p_Value) & a_Dest.f_BlockMask);
                    }

                    l_Pos += UserHeader.c_Size;
                }

                a_Dest.f_Element = a_Element;
                a_Dest.UpdateCoveredBytes();

                return true;
            }

            private static void DoDefaultInit(uint a_ChunkIndex,
                                              uint a_ChunkSize,
                                              out uint a_BlockCount,
                                              out uint a_BlockMask,
                                              out uint a_FirstHashIndex)
            {
                uint l_Junk;
                a_BlockCount = BlockBitmap.CalcBlockBoundsWide(0, a_ChunkSize, out l_Junk);
                a_BlockMask = BlockBitmap.BuildBlockMask(0, a_BlockCount);
                a_FirstHashIndex = (a_ChunkIndex << c_TAICHExponent);
            }

            public bool IsValid()
            {
                return (null != p_Element);
            }

            public void SetElement(HardMemory.Element a_Element)
            {
                f_Element = a_Element;
            }

            public int Write(FileNum a_FileNum,
                             ref HashStorage a_HashStorage,
                             ref UserFileKnowledge a_Knowledge,
                             byte[] a_Dest,
                             int a_DestIdx)
            {
                // Returns the size of the space used up.

                IncompleteFile.WriteDefaultChunkHeader(c_IncomplChunkMagic, a_FileNum, f_ChunkIndex, a_Dest, a_DestIdx);

                bool l_StoredTAICHBlockData = (true == a_Knowledge.p_HasTAICHHash &&
                                               true == f_TAICHBlockHashes.IsValid());

                if (false == l_StoredTAICHBlockData)
                {
                    f_ChunkFlags &= ~ChunkFlags.TAICHBlockHashesPresent;
                    Buffer2.BlockZero(a_Dest, c_TAICHTopHashOffset, (c_Challenge1Offset + sizeof(ulong)) - c_TAICHTopHashOffset);
                }
                else
                {
                    // Page 1.
                    ulong l_Challenge = Util.GetRandomULong();
                    Buffer.BlockCopy(a_HashStorage.ExposeHash(a_Knowledge.p_TAICHHashIdx), 0, a_Dest, a_DestIdx + c_TAICHTopHashOffset, Tiger.c_HashByteSize);
                    Util.WriteUInt64LE(l_Challenge, a_Dest, a_DestIdx + c_Challenge1Offset);

                    // Page 2.
                    Util.WriteUInt64LE(l_Challenge, a_Dest, a_DestIdx + c_Challenge2Offset);
                    int l_BytesWritten = f_TAICHBlockHashes.WriteAll(a_Dest, a_DestIdx + c_TAICHBlockHashesOffset);

                    // Zero unused bytes.
                    l_BytesWritten += c_TAICHBlockHashesOffset;
                    Buffer2.BlockZero(a_Dest, a_DestIdx + l_BytesWritten, c_Size - l_BytesWritten);
                }

                // Zero unused bytes.
                Buffer2.BlockZero(a_Dest, a_DestIdx + c_Challenge1Offset + sizeof(ulong), c_IncomplFlagsOffset - (c_Challenge1Offset + sizeof(ulong)));

                ByteArray l_Data;
                WriteHeaders(a_Dest, a_DestIdx, out l_Data);

                return (false == l_StoredTAICHBlockData ? c_ShortSize : c_Size);
            }

            private void WriteHeaders(byte[] a_Dest,
                                      int a_DestIdx,
                                      out ByteArray a_Data)
            {
                // f_ChunkFlags will also be written.

                int l_Count = f_Users.p_Count;
                a_Data = new ByteArray(a_Dest, a_DestIdx + c_IncomplFlagsOffset, 12 + (l_Count * UserHeader.c_Size));

                Util.WriteUInt32LE((uint)f_ChunkFlags, a_Dest, a_DestIdx + c_IncomplFlagsOffset);
                f_MergedUser.Write(a_Dest, a_DestIdx + c_MergedUserHeadersOffset);
                Util.WriteUInt32LE((uint)l_Count, a_Dest, a_DestIdx + c_MergedUserHeadersOffset + 4);

                int l_Pos = (a_DestIdx + c_UserIncomplHeadersOffset);
                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int i = 0; i < l_Count; i++)
                {
                    l_Arr[i].Write(a_Dest, l_Pos);

                    l_Pos += UserHeader.c_Size;
                }
            }

            public bool Update(FileNum a_FileNum,
                               ref HashStorage a_HashStorage,
                               ref UserFileKnowledge a_Knowledge,
                               HardMemory a_HardMem,
                               byte[] a_TempBuffer)
            {
                int l_Size = Write(a_FileNum, ref a_HashStorage, ref a_Knowledge, a_TempBuffer, 0);
                ByteArray l_Data = new ByteArray(a_TempBuffer, 0, l_Size);

                return a_HardMem.Write(f_Element, 0, ref l_Data);
            }

            public bool UpdateUserHeaders(HardMemory a_HardMem,
                                          byte[] a_TempBuffer,
                                          bool a_FlushAfterwards)
            {
                ByteArray l_Data;
                WriteHeaders(a_TempBuffer, 0, out l_Data);

                return (true == a_HardMem.Write(f_Element, c_IncomplFlagsOffset, ref l_Data) &&
                        (false == a_FlushAfterwards ||
                         true == a_HardMem.Flush()));
            }

            public bool NeedExtraHashesForTAICHChunkCheck(ref TAICHTreeLayout a_Layout,
                                                          int a_GreatestUpperLevel)
            {
                if (TAICHChunkCheckReqs.Unknown == f_TAICHChunkCheckReq)
                {
                    if (0 == a_GreatestUpperLevel)
                    {
                        // This is a tiny file (no bigger than
                        // c_ChunkSize).

                        f_TAICHChunkCheckReq = TAICHChunkCheckReqs.NoExtraHashesNeeded;
                    }
                    else
                    {
                        List<uint> l_Indices = TAICHTree.CalcSubTreeIndices(ref a_Layout, c_TAICHExponent, (uint)f_ChunkIndex, a_GreatestUpperLevel);

                        if (f_BlockCount == l_Indices.Count)
                        {
                            f_TAICHChunkCheckReq = TAICHChunkCheckReqs.NoExtraHashesNeeded;
                        }
                        else
                        {
                            f_TAICHChunkCheckReq = TAICHChunkCheckReqs.ExtraHashesNeeded;
                        }
                    }
                }

                return (TAICHChunkCheckReqs.ExtraHashesNeeded == f_TAICHChunkCheckReq);
            }

            public void MarkDisputed()
            {
                f_ChunkFlags |= ChunkFlags.Disputed;
            }

            public bool IsMarkedDisputed()
            {
                return (0 != (ChunkFlags.Disputed & f_ChunkFlags));
            }

            public int GetMergedSlot()
            {
                // Returns a negative value if there is no such slot.

                return (false == f_MergedUser.IsSet() ? (-1) : f_MergedUser.p_Slot);
            }

            public int FindLowestUnoccupiedSlot(bool a_IncludeMerged)
            {
                int l_MergedSlot = GetMergedSlot();
                int l_Count = f_Users.p_Count;
                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int l_Slot = 0; true; l_Slot++)
                {
                    if (true == a_IncludeMerged &&
                        l_MergedSlot == l_Slot)
                    {
                        continue;
                    }

                    bool l_Conflict = false;

                    // i is the header index.
                    for (int i = 0; i < l_Count; i++)
                    {
                        if (l_Arr[i].p_Slot == l_Slot)
                        {
                            l_Conflict = true;

                            break;
                        }
                    }

                    if (false == l_Conflict)
                    {
                        return l_Slot;
                    }
                }
            }

            public int SuggestSlot(ref PieceBounds a_Bounds,
                                   int a_IgnoreConflictsHeaderIndex)
            {
                int l_MergedSlot = GetMergedSlot();
                int l_Count = f_Users.p_Count;
                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int l_Slot = 0; true; l_Slot++)
                {
                    if (l_MergedSlot == l_Slot)
                    {
                        continue;
                    }

                    bool l_Conflict = false;

                    // i is the header index.
                    for (int i = 0; i < l_Count; i++)
                    {
                        if (i == a_IgnoreConflictsHeaderIndex)
                        {
                            continue;
                        }

                        if (l_Arr[i].p_Slot == l_Slot &&
                            true == a_Bounds.DoesOverlap(ref l_Arr[i].f_Bounds))
                        {
                            l_Conflict = true;

                            break;
                        }
                    }

                    if (false == l_Conflict)
                    {
                        return l_Slot;
                    }
                }
            }

            public uint SuggestElementSize()
            {
                return SuggestElementSize(0);
            }

            public uint SuggestElementSize(int a_MinSlotCount)
            {
                int l_Slots = Math.Max(a_MinSlotCount, GetOccupiedSlotCount());

                if (0 == l_Slots)
                {
                    return c_Size;
                }

                return (c_BodyStartingOffset + ((uint)l_Slots * f_ChunkSize));
            }

            public int GetOccupiedSlotCount()
            {
                int l_HighestSlot = (-1);

                if (true == f_MergedUser.IsSet())
                {
                    l_HighestSlot = Math.Max(l_HighestSlot, f_MergedUser.p_Slot);
                }

                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int i = (f_Users.p_Count - 1); 0 <= i; i--)
                {
                    l_HighestSlot = Math.Max(l_HighestSlot, l_Arr[i].p_Slot);
                }

                return (l_HighestSlot + 1);
            }

            public uint GetCoveredBytes()
            {
                // No recalculation takes place.

                return f_CoveredBytes;
            }

            public void UpdateCoveredBytes()
            {
                int l_Count = f_Users.p_Count;

                if (false == f_MergedUser.IsSet() &&
                    1 >= l_Count)
                {
                    if (0 == l_Count)
                    {
                        f_CoveredBytes = 0;
                    }
                    else
                    {
                        f_CoveredBytes = f_Users.p_Arr[0].f_Bounds.GetCoveredBytes();
                    }

                    return;
                }

                PieceBounds l_Bounds;
                CalcCoveredBounds(out l_Bounds);

                f_CoveredBytes = l_Bounds.GetCoveredBytes();
            }

            public void CalcCoveredBounds(out PieceBounds a_Bounds)
            {
                a_Bounds = new PieceBounds();

                if (true == f_MergedUser.IsSet())
                {
                    a_Bounds.Add(f_MergedUser.f_VerifiedMap);
                }

                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int i = (f_Users.p_Count - 1); 0 <= i; i--)
                {
                    a_Bounds.Add(ref l_Arr[i].f_Bounds);
                }
            }

            private uint GetSlotPos(int a_UserSlot)
            {
                return (c_BodyStartingOffset + ((uint)a_UserSlot * f_ChunkSize));
            }

            public uint GetBlockOffsetAndSize(uint a_BlockIndex,
                                              out uint a_Size)
            {
                // The returned offset is slot agnostic, so add
                // GetUserSlotPos() to fix this.

                uint l_Offset = (a_BlockIndex * TAICHTree.c_BlockSize);
                a_Size = Math.Min(f_ChunkSize - l_Offset, TAICHTree.c_BlockSize);

                return l_Offset;
            }

            public bool MoveBody(HardMemory a_HardMem,
                                 ref UserHeader a_Header,
                                 int a_DestSlot,
                                 byte[] a_TempBuffer)
            {
                int l_SrcSlot = a_Header.p_Slot;

                if (l_SrcSlot == a_DestSlot ||
                    false == a_Header.f_Bounds.p_IsSet)
                {
                    return true;
                }

                uint l_SrcSlotPos = GetSlotPos(l_SrcSlot);
                uint l_DestSlotPos = GetSlotPos(a_DestSlot);

                PieceBounds.Enumerator l_Enum;
                a_Header.f_Bounds.BuildEnumerator(out l_Enum);

                ChunkRange l_Range;
                while (true == l_Enum.GetNextRange(out l_Range))
                {
                    ByteArray l_Data = new ByteArray(a_TempBuffer, 0, (int)l_Range.f_Size);

                    if (false == a_HardMem.Read(f_Element, l_SrcSlotPos + l_Range.f_Offset, ref l_Data) ||
                        false == a_HardMem.Write(f_Element, l_DestSlotPos + l_Range.f_Offset, ref l_Data))
                    {
                        return false;
                    }
                }

                a_Header.SetSlot(a_DestSlot);

                return true;
            }

            public bool CopyBlocks(HardMemory a_HardMem,
                                   int a_SrcSlot,
                                   int a_DestSlot,
                                   uint a_BlockBitmap,
                                   byte[] a_TempBuffer)
            {
                // Try not to call when there's nothing to do.

                uint l_SrcChunkOffset = GetSlotPos(a_SrcSlot);
                uint l_DestChunkOffset = GetSlotPos(a_DestSlot);

                for (uint i = 0; f_BlockCount > i; i++)
                {
                    if (true == BlockBitmap.TestBit(i, a_BlockBitmap))
                    {
                        uint l_Count = 1;

                        for (uint k = (i + 1); f_BlockCount > k; k++)
                        {
                            if (false == BlockBitmap.TestBit(k, a_BlockBitmap))
                            {
                                break;
                            }

                            l_Count++;
                        }

                        uint a_SrcOffset = (l_SrcChunkOffset + (i * TAICHTree.c_BlockSize));
                        uint a_DestOffset = (l_DestChunkOffset + (i * TAICHTree.c_BlockSize));
                        ByteArray l_BlockData = new ByteArray(a_TempBuffer, 0, (int)(l_Count * TAICHTree.c_BlockSize));

                        if (false == a_HardMem.Read(f_Element, a_SrcOffset, ref l_BlockData) ||
                            false == a_HardMem.Write(f_Element, a_DestOffset, ref l_BlockData))
                        {
                            return false;
                        }

                        // Add the count and after this loop round we
                        // will add one, since we've already checked
                        // things so far.
                        i += l_Count;
                    }
                }

                return true;
            }

            public bool ReadBody(HardMemory a_HardMem,
                                 int a_UserSlot,
                                 uint a_Offset,
                                 ref ByteArray a_Data)
            {
                uint a_SrcOffset = (GetSlotPos(a_UserSlot) + a_Offset);
                bool l_OK = a_HardMem.Read(f_Element, a_SrcOffset, ref a_Data);

                return l_OK;
            }

            public bool WriteBody(HardMemory a_HardMem,
                                  int a_UserSlot,
                                  uint a_Offset,
                                  ref ByteArray a_Data)
            {
                uint a_DestOffset = (GetSlotPos(a_UserSlot) + a_Offset);
                return a_HardMem.Write(f_Element, a_DestOffset, ref a_Data);
            }

            public bool WriteBody(HardMemory a_HardMem,
                                  int a_UserSlot,
                                  ref PieceBounds a_Bounds,
                                  uint a_BufferBaseOffset,
                                  ref ByteArray a_Data)
            {
                PieceBounds.Enumerator l_Enum;
                a_Bounds.BuildEnumerator(out l_Enum);

                ChunkRange l_Range;
                while (true == l_Enum.GetNextRange(out l_Range))
                {
                    ByteArray l_SrcData = new ByteArray(a_Data.f_Arr, (int)(l_Range.f_Offset - a_BufferBaseOffset), (int)l_Range.f_Size);
                    uint a_DestOffset = (GetSlotPos(a_UserSlot) + l_Range.f_Offset);

                    if (false == a_HardMem.Write(f_Element, a_DestOffset, ref l_SrcData))
                    {
                        return false;
                    }
                }

                return true;
            }

            public void ReverseTAICHBlockVerification()
            {
                // Removes the merged user and resets f_GeneralMap and
                // f_AcceptedMap of normal users.

                f_MergedUser.Invalidate();

                for (int i = (f_Users.p_Count - 1); 0 <= i; i--)
                {
                    f_Users.p_Arr[i].f_DirtyMap.TakeOver(f_BlockMask);
                    f_Users.p_Arr[i].f_AcceptedMap.ClearAll();
                }
            }

            public bool GetHeaderIndex(UserNum a_UserNum,
                                       out int a_HeaderIndex)
            {
                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int i = (f_Users.p_Count - 1); 0 <= i; i--)
                {
                    if (l_Arr[i].f_UserNum == a_UserNum)
                    {
                        a_HeaderIndex = i;

                        return true;
                    }
                }

                a_HeaderIndex = (-1);

                return false;
            }

            public int FindBestUserHeaderIndexForSubstitution()
            {
                int l_TrgtIdx = 0;

                UserHeader[] l_Arr = f_Users.p_Arr;
                for (int i = (f_Users.p_Count - 1); 1 <= i; i--)
                {
                    // Pick the oldest header.

                    if (true == l_Arr[i].f_LastUpdateTime.IsEarlierThan(l_Arr[l_TrgtIdx].f_LastUpdateTime))
                    {
                        l_TrgtIdx = i;
                    }
                }

                return l_TrgtIdx;
            }

            public bool IsRangeAvailable(uint a_Offset,
                                         uint a_Size,
                                         bool a_Join)
            {
                // NOTE: The merged user is never checked.

                int l_Count = f_Users.p_Count;

                if (0 == l_Count)
                {
                    return false;
                }

                if (1 == l_Count ||
                    false == a_Join)
                {
                    uint l_FinalOffset = (a_Offset + a_Size);
                    UserHeader[] l_Arr = f_Users.p_Arr;
                    for (int i = 0; i < l_Count; i++)
                    {
                        if (true == l_Arr[i].f_Bounds.IsRangeAvailable(a_Offset, l_FinalOffset))
                        {
                            return true;
                        }
                    }

                    return false;
                }

                AvailabilityAnalyzer l_AA = new AvailabilityAnalyzer(a_Offset, a_Size, ref f_Users);

                return l_AA.IsRangeAvailable();
            }

            public OpStatus ResizeElement(HardMemory a_HardMem,
                                          uint a_NewSize)
            {
                // Resizes and updates f_Element.

                if (false == a_HardMem.Resize(ref f_Element, a_NewSize, true))
                {
                    return (f_Element.f_MaxSize < a_NewSize ? OpStatus.InsuffDiskSpace : OpStatus.IoError);
                }

                return OpStatus.Done;
            }

            public bool HasTAICHBlockHashes(bool a_ActualUsePlanned)
            {
                if (false == a_ActualUsePlanned)
                {
                    // In this case f_TAICHBlockHashes.IsValid() can be
                    // false even if we return true (if the header
                    // wasn't loaded completely, using the Load()
                    // argument a_LoadTAICHBlockData).

                    return (0 != (ChunkFlags.TAICHBlockHashesPresent & f_ChunkFlags));
                }
                else
                {
                    return f_TAICHBlockHashes.IsValid();
                }
            }

            public void SetTAICHBlockHashes(ref TigerHashHive a_TAICHBlockHashes)
            {
                f_TAICHBlockHashes = a_TAICHBlockHashes;
                f_ChunkFlags |= ChunkFlags.TAICHBlockHashesPresent;
            }

            public bool HasTAICHBlockHash(uint a_HashIndex)
            {
                return f_TAICHBlockHashes.HasHash(a_HashIndex);
            }

            public bool VerifyTAICHBlockHashByBlockIndex(uint a_BlockIndex,
                                                         Tiger a_Tiger)
            {
                return VerifyTAICHBlockHashByHashIndex(f_FirstHashIndex + a_BlockIndex, a_Tiger);
            }

            public bool VerifyTAICHBlockHashByHashIndex(uint a_HashIndex,
                                                        Tiger a_Tiger)
            {
                return f_TAICHBlockHashes.Verify(a_HashIndex, a_Tiger);
            }

            public static int CalcSlotsFromElementSize(uint a_ElementSize,
                                                       uint a_ChunkSize)
            {
                if (c_BodyStartingOffset >= a_ElementSize)
                {
                    return 0;
                }

                return (int)((a_ElementSize - c_BodyStartingOffset) / a_ChunkSize);
            }

            private enum TAICHChunkCheckReqs : byte
            {
                // That is TAICHChunkCheckRequirements.

                Unknown = 0,
                ExtraHashesNeeded = 1,
                NoExtraHashesNeeded = 2,
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct MergedUserHeader
        {
            // For incomplete chunks.

            public const ushort c_Size = 4;

            private byte f_Slot; // 2
            public BlockBitmap f_VerifiedMap; // 0

            public int p_Slot
            {
                get { return f_Slot; }
            }

            public MergedUserHeader(byte[] a_Src,
                                    int a_SrcIdx)
            {
                f_Slot = a_Src[a_SrcIdx + 2];
                f_VerifiedMap = new BlockBitmap(a_Src, a_SrcIdx);
            }

            public void Write(byte[] a_Dest,
                              int a_DestIdx)
            {
                a_Dest[a_DestIdx + 2] = (false == f_VerifiedMap.IsSet() ? byte.MaxValue : f_Slot);
                f_VerifiedMap.Write(a_Dest, a_DestIdx);

                // Zero unused byte.
                a_Dest[a_DestIdx + 3] = 0;
            }

            public void SetSlot(int a_NewSlot)
            {
                f_Slot = (byte)a_NewSlot;
            }

            public bool IsSet()
            {
                return f_VerifiedMap.IsSet();
            }

            public void Invalidate()
            {
                f_VerifiedMap.ClearAll();
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct UserHeader
        {
            // For incomplete chunks.

            public const ushort c_Size = 48;

            public readonly UserNum f_UserNum; // 0
            private byte f_Slot; // 42
            public Time f_LastUpdateTime; // 8
            public PieceBounds f_Bounds; // 16

            // TAICH block checking.
            public BlockBitmap f_DirtyMap; // 38
            public BlockBitmap f_AcceptedMap; // 40

            // TAICH chunk checking.
            public bool f_ChunkTestedAgainstUpperTree; // 43

            public int p_Slot
            {
                get { return f_Slot; }
            }

            public UserHeader(UserNum a_UserNum)
            {
                f_UserNum = a_UserNum;
                f_Slot = byte.MaxValue;
                f_LastUpdateTime = new Time();
                f_Bounds = new PieceBounds();
                f_DirtyMap = new BlockBitmap();
                f_AcceptedMap = new BlockBitmap();
                f_ChunkTestedAgainstUpperTree = false;
            }

            public static bool Load(byte[] a_Src,
                                    int a_SrcIdx,
                                    bool a_AcceptTAICHBlockData,
                                    out UserHeader a_Dest)
            {
                UserNum l_UserNum = new UserNum(Util.ReadUInt64LE(a_Src, a_SrcIdx));
                a_Dest = new UserHeader(l_UserNum);

                if (true == l_UserNum.IsGeneric() &&
                    UserNum.s_Anonymous != l_UserNum &&
                    UserNum.s_Import != l_UserNum)
                {
                    // Should never happen.

                    return false;
                }

                if (false == PieceBounds.Load(a_Src, a_SrcIdx + 16, out a_Dest.f_Bounds))
                {
                    // Corruption.

                    return false;
                }

                a_Dest.f_Slot = a_Src[a_SrcIdx + 42];
                a_Dest.f_LastUpdateTime = Time.FromBinary(Util.ReadUInt64LE(a_Src, a_SrcIdx + 8));

                if (false == a_AcceptTAICHBlockData)
                {
                    a_Dest.f_DirtyMap = new BlockBitmap();
                    a_Dest.f_AcceptedMap = new BlockBitmap();
                }
                else
                {
                    a_Dest.f_DirtyMap = new BlockBitmap(a_Src, a_SrcIdx + 38);
                    a_Dest.f_AcceptedMap = new BlockBitmap(a_Src, a_SrcIdx + 40);
                }

                a_Dest.f_ChunkTestedAgainstUpperTree = (0 != a_Src[a_SrcIdx + 43]);

                return true;
            }

            public void Write(byte[] a_Dest,
                              int a_DestIdx)
            {
                Util.WriteUInt64LE(f_UserNum.f_Val, a_Dest, a_DestIdx);
                Util.WriteUInt64LE(f_LastUpdateTime.ToBinary(), a_Dest, a_DestIdx + 8);
                f_Bounds.Write(a_Dest, a_DestIdx + 16);
                f_DirtyMap.Write(a_Dest, a_DestIdx + 38);
                f_AcceptedMap.Write(a_Dest, a_DestIdx + 40);
                a_Dest[a_DestIdx + 42] = f_Slot;
                a_Dest[a_DestIdx + 43] = (byte)(false == f_ChunkTestedAgainstUpperTree ? 0 : 1);
            }

            public void SetSlot(int a_NewSlot)
            {
                f_Slot = (byte)a_NewSlot;
            }

            public void UpdateTime(Instance a_Inst)
            {
                f_LastUpdateTime = a_Inst.GetTime();
            }

            public void MarkAsDirty(uint a_Mask)
            {
                // Sets the bits in the dirty bitmap and clears the
                // bits in the accepted bitmap.

                f_DirtyMap.AddMask(a_Mask);
                f_AcceptedMap.DeleteMask(a_Mask);
            }

            private uint GetBlockMask(uint a_ChunkSize)
            {
                return f_Bounds.ToBitmap(true).p_Value;
            }

            public void GetPosAndNegVerifiedBlocks(uint a_ChunkSize,
                                                   out uint a_PosMask,
                                                   out uint a_NegMask)
            {
                a_PosMask = f_AcceptedMap.p_Value;
                a_NegMask = ~a_PosMask;

                uint l_ClipMask = GetBlockMask(a_ChunkSize);

                a_PosMask &= l_ClipMask;
                a_NegMask &= l_ClipMask;
            }

            public bool HasRejectedBlocks(uint a_ChunkIndex,
                                          uint a_Verified)
            {
                uint l_Mask = (a_Verified & GetBlockMask(a_ChunkIndex));

                return (f_AcceptedMap.p_Value != l_Mask);
            }

            public void SetChunkTestedAgainstUpperTree()
            {
                f_ChunkTestedAgainstUpperTree = true;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct AvailabilityAnalyzer
        {
            // Don't use the struct default constructor.

            public readonly uint f_Offset;
            public readonly uint f_Size;
            public readonly uint f_FinalOffset;
            private readonly FlexArray<UserHeader> f_UserHeaders;

            // This block will be initialized by PrepareCombinations()
            // on demand.
            private PermutationGenerator f_Gen;
            private bool f_Virgin;

            public AvailabilityAnalyzer(uint a_Offset,
                                        uint a_Size,
                                        ref FlexArray<UserHeader> a_UserHeaders)
            {
                f_Offset = a_Offset;
                f_Size = a_Size;
                f_FinalOffset = (f_Offset + f_Size);
                f_UserHeaders = a_UserHeaders;

                // Clear extras.
                f_Gen = new PermutationGenerator();
                f_Virgin = false;
            }

            public bool IsValid()
            {
                return (0 != f_Size);
            }

            public bool IsRangeAvailable()
            {
                // Returns true if the range described by f_Offset and
                // f_Size is available in f_UserHeaders.

                return (f_Size == GetCoveredBytes());
            }

            public uint GetCoveredBytes()
            {
                uint l_BytesCovered = 0;
                FlexArray<UserRange> l_Ls;
                FlexArray<UserRange>.Create(2 * f_UserHeaders.p_Count, out l_Ls);

                for (int i = 0; f_UserHeaders.p_Count > i; i++)
                {
                    l_BytesCovered += AddUserToList(ref l_Ls, ref f_UserHeaders.p_Arr[i]);

                    if (f_Size == l_BytesCovered)
                    {
                        break;
                    }
                }

                return l_BytesCovered;
            }

            private uint AddUserToList(ref FlexArray<UserRange> a_DestList,
                                       ref UserHeader a_UserHeader)
            {
                uint l_BytesAdded = 0;

                PieceBounds.Enumerator l_Enum;
                a_UserHeader.f_Bounds.BuildEnumerator(out l_Enum);

                ChunkRange l_Range;
                while (true == l_Enum.GetNextRange(out l_Range))
                {
                    l_BytesAdded += AddRangeToList(ref a_DestList, ref a_UserHeader, ref l_Range);
                }

                return l_BytesAdded;
            }

            private uint AddRangeToList(ref FlexArray<UserRange> a_DestList,
                                        ref UserHeader a_UserHeader,
                                        ref ChunkRange a_Range)
            {
                // Returns the number of bytes added.

                if (f_FinalOffset <= a_Range.f_Offset)
                {
                    return 0;
                }

                uint l_Final = a_Range.p_Final;

                if (f_Offset >= l_Final)
                {
                    return 0;
                }

                int l_Count = a_DestList.p_Count;
                uint l_BytesAdded = 0;

                // Try to find a gap to insert a_Range.

                UserRange[] l_Ranges = a_DestList.p_Arr;
                for (int k = 0; k <= l_Count; k++)
                {
                    // Find the ending border first, due to performance
                    // reasons.

                    uint l_BorderAft;

                    if (k < l_Count)
                    {
                        l_BorderAft = l_Ranges[k].f_Rng.f_Offset;

                        if (a_Range.f_Offset >= l_BorderAft)
                        {
                            continue;
                        }
                    }
                    else
                    {
                        l_BorderAft = f_FinalOffset;
                    }

                    uint l_BorderBef;

                    if (0 == k)
                    {
                        l_BorderBef = f_Offset;
                    }
                    else
                    {
                        l_BorderBef = l_Ranges[k - 1].f_Rng.p_Final;

                        if (l_Final <= l_BorderBef)
                        {
                            break;
                        }
                    }

                    uint l_NewOffset = Math.Max(l_BorderBef, a_Range.f_Offset);
                    uint l_NewFinal = Math.Min(l_BorderAft, l_Final);

                    // Check, since it can happen that l_BorderBef is
                    // l_BorderAft.
                    if (l_NewOffset < l_NewFinal)
                    {
                        a_DestList.ReserveAt(k);
                        l_Ranges = a_DestList.p_Arr;

                        ChunkRange l_NewRange;
                        ChunkRange.CreateFromOffsets(l_NewOffset, l_NewFinal, out l_NewRange);
                        l_Ranges[k] = new UserRange(a_UserHeader.f_UserNum, a_UserHeader.p_Slot, ref l_NewRange);
                        l_Count++;
                        l_BytesAdded += l_NewRange.f_Size;

                        // We need to make k go up two steps now,
                        // otherwise we will do a useless recheck.
                        k++;
                    }
                }

                return l_BytesAdded;
            }

            public bool IsPrepared()
            {
                return f_Gen.IsValid();
            }

            public bool PrepareCombinations()
            {
                if (false == f_Gen.IsValid())
                {
                    int l_Count = f_UserHeaders.p_Count;

                    if (PermutationGenerator.c_MinElements > l_Count ||
                        PermutationGenerator.c_MaxElements < l_Count)
                    {
                        return false;
                    }

                    f_Gen = new PermutationGenerator(l_Count);
                }
                else
                {
                    f_Gen.BuildFirst();
                }

                f_Virgin = true;

                return true;
            }

            public bool GetNextCombination(ref FlexArray<UserRange> a_DestList)
            {
                bool l_Junk;
                return GetNextCombination(ref a_DestList, out l_Junk);
            }

            public bool GetNextCombination(ref FlexArray<UserRange> a_DestList,
                                           out bool a_MultipleContributors)
            {
                // Call PrepareCombinations() before the first call to
                // this routine. Returns false when there are no more
                // combinations.

                a_DestList.Clear();

                while (true)
                {
                    if (false == f_Virgin)
                    {
                        if (false == f_Gen.HasNext())
                        {
                            a_MultipleContributors = false;

                            return false;
                        }

                        f_Gen.BuildNext();
                    }
                    else
                    {
                        f_Virgin = false;
                    }

                    if (true == FillListFromGenerator(ref a_DestList, out a_MultipleContributors))
                    {
                        return true;
                    }
                }
            }

            private bool FillListFromGenerator(ref FlexArray<UserRange> a_DestList,
                                               out bool a_MultipleContributors)
            {
                // Returns true if the whole range is available.

                uint l_Contribs = 0;
                uint l_BytesAdded = 0;
                int[] l_Priorities = f_Gen.ExposeArray();

                for (int i = 0, lim = l_Priorities.Length; i < lim; i++)
                {
                    int l_HeaderIdx = l_Priorities[i];
                    uint l_NowAdded = AddUserToList(ref a_DestList, ref f_UserHeaders.p_Arr[l_HeaderIdx]);

                    if (0 != l_NowAdded)
                    {
                        l_Contribs++;
                    }

                    l_BytesAdded += l_NowAdded;

                    if (f_Size == l_BytesAdded)
                    {
                        a_MultipleContributors = (2 <= l_Contribs);

                        return true;
                    }
                }

                a_DestList.Clear();
                a_MultipleContributors = false;

                return false;
            }

            public static void AddToHashRoutine(IHashFunction a_HashAlgo,
                                                byte[] a_TempBuffer,
                                                ref FlexArray<UserRange> a_CombRanges)
            {
                // The caller should reset a_HashAlgo before calling.
                // We assume a_TempBuffer is large enough to hold all
                // ranges.

                int l_Idx = 0;

                for (int k = 0; a_CombRanges.p_Count > k; k++)
                {
                    a_CombRanges.p_Arr[k].Write(a_TempBuffer, l_Idx);

                    l_Idx += UserRange.c_Size;
                }

                a_HashAlgo.Update(a_TempBuffer, 0, l_Idx);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct HashStorage
        {
            // 'Index' in the context of this container refers not to
            // the 'hash index' used by the HashSet class.

            private readonly Dictionary<int, Item> f_Hashes;

            private HashStorage(Dictionary<int, Item> a_Hashes)
            {
                f_Hashes = a_Hashes;
            }

            public static void Create(out HashStorage a_Dest)
            {
                a_Dest = new HashStorage(new Dictionary<int, Item>());
            }

            public void TrimExcess()
            {
                List<Item> l_RemLs = null;

                foreach (Item l_Item in f_Hashes.Values)
                {
                    if (0 == l_Item.p_RefCount)
                    {
                        if (null == l_RemLs)
                        {
                            l_RemLs = new List<Item>(1);
                        }

                        l_RemLs.Add(l_Item);
                    }
                }

                if (null != l_RemLs)
                {
                    foreach (Item l_Item in l_RemLs)
                    {
                        bool l_Removed = f_Hashes.Remove(l_Item.f_Index);
                    }
                }
            }

            public static bool FromBuffer(BufferIO a_Buffer,
                                          out HashStorage a_HashStorage)
            {
                uint l_Count = a_Buffer.ReadUInt32LE();

                if (ushort.MaxValue < l_Count)
                {
                    a_HashStorage = new HashStorage();

                    return false;
                }

                Dictionary<int, Item> l_Hashes = new Dictionary<int, Item>((int)l_Count);

                for (uint i = 0; i < l_Count; i++)
                {
                    int l_Index = (int)a_Buffer.ReadUInt32LE();
                    int l_Len = a_Buffer.ReadUInt16LE();
                    byte[] l_Bytes = a_Buffer.ReadBytes((uint)l_Len);

                    try
                    {
                        l_Hashes.Add(l_Index, new Item(l_Index, l_Bytes));
                    }
                    catch (ArgumentException)
                    {
                        // Duplicate key.

                        a_HashStorage = new HashStorage();

                        return false;
                    }
                }

                a_HashStorage = new HashStorage(l_Hashes);

                return true;
            }

            public void Write(BufferIO a_Buffer)
            {
                a_Buffer.WriteUInt32LE((uint)f_Hashes.Count);

                foreach (Item l_Item in f_Hashes.Values)
                {
                    a_Buffer.WriteUInt32LE((uint)l_Item.f_Index);
                    a_Buffer.WriteUInt16LE((ushort)l_Item.f_Hash.Length);
                    a_Buffer.WriteBytes(l_Item.f_Hash);
                }
            }

            public byte[] ExposeHash(int a_Index)
            {
                return f_Hashes[a_Index].f_Hash;
            }

            public bool HasThisHash(int a_Index,
                                    byte[] a_Buf,
                                    int a_BufIdx)
            {
                // a_Index may be subzero.

                if (0 > a_Index)
                {
                    return false;
                }

                byte[] l_Hash = f_Hashes[a_Index].f_Hash;

                return Buffer2.ValueEquals(a_Buf, a_BufIdx, l_Hash, 0, l_Hash.Length);
            }

            public bool HasThisHash(int a_Index,
                                    Tiger a_Tiger)
            {
                // a_Index may be subzero.

                if (0 > a_Index)
                {
                    return false;
                }

                return a_Tiger.VerifyBE(f_Hashes[a_Index].f_Hash, 0);
            }

            private int FindUnoccupiedIndex()
            {
                for (int i = 0; true; i++)
                {
                    if (false == f_Hashes.ContainsKey(i))
                    {
                        return i;
                    }
                }
            }

            public bool IsLegalHash(int a_Index,
                                    int a_ExpectedSize)
            {
                // a_Index may be subzero.

                if (0 > a_Index)
                {
                    return true;
                }

                Item l_Item;
                if (true == f_Hashes.TryGetValue(a_Index, out l_Item))
                {
                    return (l_Item.f_Hash.Length == a_ExpectedSize);
                }

                return false;
            }

            public bool IsHash(int a_Index,
                               ref ByteArray a_Data)
            {
                return f_Hashes[a_Index].IsHash(ref a_Data);
            }

            private Item FindHash(ref ByteArray a_Data)
            {
                foreach (Item l_Item in f_Hashes.Values)
                {
                    if (true == l_Item.IsHash(ref a_Data))
                    {
                        return l_Item;
                    }
                }

                return null;
            }

            public int RefHash(ref ByteArray a_Data)
            {
                Item l_Item = FindHash(ref a_Data);

                if (null == l_Item)
                {
                    l_Item = new Item(FindUnoccupiedIndex(), a_Data.ToArray(true));

                    f_Hashes.Add(l_Item.f_Index, l_Item);
                }

                l_Item.IncRefCount();

                return l_Item.f_Index;
            }

            public int RefHash(P2PHashSets a_HashSet,
                               ref MasterHashes a_MasterHashes)
            {
                // It's fine if the specified type of hash set is not
                // contained in a_MasterHashes, in this case a negative
                // value is returned.

                ByteArray l_Data;
                if (true == a_MasterHashes.ExposeHash(a_HashSet, out l_Data))
                {
                    return RefHash(ref l_Data);
                }

                return (-1);
            }

            public void RefHash(int a_Index)
            {
                // It's fine if the specified index is subzero,
                // otherwise the index must be valid.

                if (0 <= a_Index)
                {
                    f_Hashes[a_Index].IncRefCount();
                }
            }

            public void DerefHash(ref int a_Idx)
            {
                // a_Idx may be subzero, in which case nothing happens.

                if (0 <= a_Idx)
                {
                    f_Hashes[a_Idx].DecRefCount();

                    a_Idx = (-1);
                }
            }

            private class Item
            {
                public readonly int f_Index;
                // Hashes are stored in big endian notation.
                public readonly byte[] f_Hash;

                private uint f_RefCount;

                public uint p_RefCount
                {
                    get { return f_RefCount; }
                }

                public Item(int a_Index,
                            byte[] a_Hash)
                {
                    f_Index = a_Index;
                    f_Hash = a_Hash;
                }

                public void IncRefCount()
                {
                    f_RefCount++;
                }

                public void DecRefCount()
                {
                    f_RefCount--;
                }

                public bool IsHash(ref ByteArray a_Data)
                {
                    return (f_Hash.Length == a_Data.f_Len &&
                            true == Buffer2.ValueEquals(f_Hash, 0, a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len));
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct UserFileKnowledge
        {
            public const byte c_SuggestedStartingAlignment = 16;
            public const byte c_Size = 40;

            private const int c_BelieveHashReputation = 128;
            private const int c_BelieveHashUserCount = 5;

            public readonly UserNum f_UserNum; // 0

            private int f_Reputation; // not on disk
            private Time f_LastUpdateTime; // 8
            private ulong f_FileSize; // 16
            private int f_eD2kHashIdx; // 24
            private int f_AICHHashIdx; // 28
            private int f_TAICHHashIdx; // 32

            public int p_Reputation
            {
                get { return f_Reputation; }
            }

            public ulong p_FileSize
            {
                get { return f_FileSize; }
            }

            public int p_eD2kHashIdx
            {
                get { return f_eD2kHashIdx; }
            }

            public int p_AICHHashIdx
            {
                get { return f_AICHHashIdx; }
            }

            public int p_TAICHHashIdx
            {
                get { return f_TAICHHashIdx; }
            }

            public bool p_HaseD2kHash
            {
                get { return (0 <= f_eD2kHashIdx); }
            }

            public bool p_HasAICHHash
            {
                get { return (0 <= f_AICHHashIdx); }
            }

            public bool p_HasTAICHHash
            {
                get { return (0 <= f_TAICHHashIdx); }
            }
            public UserFileKnowledge(UserNum a_SrcUserNum)
            {
                // This instace takes ownership of a_MasterHashes.

                f_UserNum = a_SrcUserNum;
                f_Reputation = 0;
                f_LastUpdateTime = new Time();
                f_FileSize = 0;
                f_eD2kHashIdx = (-1);
                f_AICHHashIdx = (-1);
                f_TAICHHashIdx = (-1);
            }

            private void Ref(ulong a_FileSize,
                             ref HashStorage a_HashStorage,
                             ref MasterHashes a_MasterHashes)
            {
                f_FileSize = a_FileSize;
                f_eD2kHashIdx = a_HashStorage.RefHash(P2PHashSets.eD2kHashSet, ref a_MasterHashes);
                f_AICHHashIdx = a_HashStorage.RefHash(P2PHashSets.AICHTree, ref a_MasterHashes);
                f_TAICHHashIdx = a_HashStorage.RefHash(P2PHashSets.TAICHTree, ref a_MasterHashes);
            }

            public void Deref(ref HashStorage a_HashStorage)
            {
                f_FileSize = 0;
                a_HashStorage.DerefHash(ref f_eD2kHashIdx);
                a_HashStorage.DerefHash(ref f_AICHHashIdx);
                a_HashStorage.DerefHash(ref f_TAICHHashIdx);
            }

            public bool Change(Time a_Time,
                               ulong a_FileSize,
                               ref HashStorage a_HashStorage,
                               ref MasterHashes a_MasterHashes)
            {
                // Returns true if new and old data differ.

                ulong l_FileSize = this.f_FileSize;
                int l_eD2kHashIdx = this.f_eD2kHashIdx;
                int l_AICHHashIdx = this.f_AICHHashIdx;
                int l_TAICHHashIdx = this.f_TAICHHashIdx;

                Deref(ref a_HashStorage);

                // Don't use HashStorage.TrimExcess() here, or we won't
                // be able to compare the indices later.

                Ref(a_FileSize, ref a_HashStorage, ref a_MasterHashes);
                this.f_LastUpdateTime = a_Time;

                return (l_FileSize != this.f_FileSize ||
                        l_eD2kHashIdx != this.f_eD2kHashIdx ||
                        l_AICHHashIdx != this.f_AICHHashIdx ||
                        l_TAICHHashIdx != this.f_TAICHHashIdx);
            }

            public static bool FromBuffer(BufferIO a_Buffer,
                                          ref HashStorage a_HashStorage,
                                          out UserFileKnowledge a_Dest)
            {
                UserNum l_UserNum = new UserNum(a_Buffer.ReadUInt64LE());
                Time l_LastUpdateTime = a_Buffer.ReadTime();
                ulong l_FileSize = a_Buffer.ReadUInt64LE();
                int l_eD2kHashIdx = (int)a_Buffer.ReadUInt32LE();
                int l_AICHHashIdx = (int)a_Buffer.ReadUInt32LE();
                int l_TAICHHashIdx = (int)a_Buffer.ReadUInt32LE();
                uint l_Reserved = a_Buffer.ReadUInt32LE();

                if (false == a_HashStorage.IsLegalHash(l_eD2kHashIdx, MD_4.c_HashByteSize) ||
                    false == a_HashStorage.IsLegalHash(l_AICHHashIdx, SHA_1.c_HashByteSize) ||
                    false == a_HashStorage.IsLegalHash(l_TAICHHashIdx, Tiger.c_HashByteSize) ||
                    0 != l_Reserved)
                {
                    // At least one invalid index.

                    a_Dest = new UserFileKnowledge();

                    return false;
                }

                a_HashStorage.RefHash(l_eD2kHashIdx);
                a_HashStorage.RefHash(l_AICHHashIdx);
                a_HashStorage.RefHash(l_TAICHHashIdx);

                a_Dest = new UserFileKnowledge(l_UserNum);
                a_Dest.f_LastUpdateTime = l_LastUpdateTime;
                a_Dest.f_FileSize = l_FileSize;
                a_Dest.f_eD2kHashIdx = l_eD2kHashIdx;
                a_Dest.f_AICHHashIdx = l_AICHHashIdx;
                a_Dest.f_TAICHHashIdx = l_TAICHHashIdx;

                return true;
            }

            public void Write(BufferIO a_Buffer)
            {
                a_Buffer.WriteUInt64LE(f_UserNum.f_Val);
                a_Buffer.WriteTime(f_LastUpdateTime);
                a_Buffer.WriteUInt64LE(f_FileSize);
                a_Buffer.WriteUInt32LE((uint)f_eD2kHashIdx);
                a_Buffer.WriteUInt32LE((uint)f_AICHHashIdx);
                a_Buffer.WriteUInt32LE((uint)f_TAICHHashIdx);
                a_Buffer.ZeroBytes(sizeof(uint));
            }

            public void SetReputation(int a_Reputation)
            {
                f_Reputation = a_Reputation;
            }

            public bool IsLessImportant(ref UserFileKnowledge a_Other)
            {
                // Returns true if this is less important (to keep)
                // than a_Other.

                if (this.f_Reputation < a_Other.f_Reputation)
                {
                    return true;
                }

                return (this.f_Reputation == a_Other.f_Reputation &&
                        true == this.f_LastUpdateTime.IsEarlierThan(a_Other.f_LastUpdateTime));
            }

            public int GetIndex(P2PHashSets a_HashSet)
            {
                switch (a_HashSet)
                {
                    case P2PHashSets.eD2kHashSet:
                        return f_eD2kHashIdx;

                    case P2PHashSets.AICHTree:
                        return f_AICHHashIdx;

                    case P2PHashSets.TAICHTree:
                    case P2PHashSets.Upper8MiBTAICHTree:
                        return f_TAICHHashIdx;

                    default: throw (new ApplicationException());
                }
            }

            public bool HasSameKnowledge(ref UserFileKnowledge a_Other)
            {
                return (this.f_FileSize == a_Other.f_FileSize &&
                        this.f_eD2kHashIdx == a_Other.f_eD2kHashIdx &&
                        this.f_AICHHashIdx == a_Other.f_AICHHashIdx &&
                        this.f_TAICHHashIdx == a_Other.f_TAICHHashIdx);
            }

            public bool HasSameKnowledgeSelective(P2PHashSets a_HashSet,
                                                  ref UserFileKnowledge a_Other)
            {
                // a_HashSet may have only one flag set.

                if (this.f_FileSize != a_Other.f_FileSize)
                {
                    return false;
                }

                switch (a_HashSet)
                {
                    case P2PHashSets.eD2kHashSet:
                        return (this.f_eD2kHashIdx == a_Other.f_eD2kHashIdx);

                    case P2PHashSets.AICHTree:
                        return (this.f_AICHHashIdx == a_Other.f_AICHHashIdx);

                    case P2PHashSets.TAICHTree:
                    case P2PHashSets.Upper8MiBTAICHTree:
                        return (this.f_TAICHHashIdx == a_Other.f_TAICHHashIdx);

                    default: throw (new ApplicationException());
                }
            }

            public static void EstimateKnowledge(ref UserFileKnowledge a_Own,
                                                 ref FlexDictionary<UserNum, UserFileKnowledge> a_Others,
                                                 ref HashStorage a_HashStorage,
                                                 out UserFileKnowledge a_Dest)
            {
                int l_eD2kHashIdx = FindPlausibleHash(P2PHashSets.eD2kHashSet, ref a_Own, ref a_Others);
                int l_AICHHashIdx = FindPlausibleHash(P2PHashSets.AICHTree, ref a_Own, ref a_Others);
                int l_TAICHHashIdx = FindPlausibleHash(P2PHashSets.TAICHTree, ref a_Own, ref a_Others);

                a_HashStorage.RefHash(l_eD2kHashIdx);
                a_HashStorage.RefHash(l_AICHHashIdx);
                a_HashStorage.RefHash(l_TAICHHashIdx);

                a_Dest = new UserFileKnowledge(UserNum.s_Self);
                a_Dest.f_FileSize = a_Own.f_FileSize;
                a_Dest.f_eD2kHashIdx = l_eD2kHashIdx;
                a_Dest.f_AICHHashIdx = l_AICHHashIdx;
                a_Dest.f_TAICHHashIdx = l_TAICHHashIdx;
            }

            private static int FindPlausibleHash(P2PHashSets a_HashSet,
                                                 ref UserFileKnowledge a_Own,
                                                 ref FlexDictionary<UserNum, UserFileKnowledge> a_Others)
            {
                int l_OwnIndex = a_Own.GetIndex(a_HashSet);

                if (0 <= l_OwnIndex)
                {
                    // Always trust yourself.

                    return l_OwnIndex;
                }

                FlexArray<Cluster> l_Clusters;
                FlexArray<Cluster>.Create(out l_Clusters);

                FlexDicEnum<UserNum, UserFileKnowledge> l_Enum;
                FlexDicItem<UserNum, UserFileKnowledge>[] l_Arr = a_Others.ExposeItemArray(out l_Enum);
                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i);)
                {
                    int l_Index = l_Arr[i].f_Val.GetIndex(a_HashSet);

                    if (0 > l_Index)
                    {
                        continue;
                    }

                    AddToClusterArray(l_Index, l_Arr[i].f_Val.f_Reputation, ref l_Clusters);
                }

                return PickStrongestCluster(ref l_Clusters);
            }

            private static void AddToClusterArray(int a_Index,
                                                  int a_Reputation,
                                                  ref FlexArray<Cluster> a_Clusters)
            {
                Cluster[] l_Arr = a_Clusters.p_Arr;
                for (int i = (a_Clusters.p_Count - 1); 0 <= i; i--)
                {
                    if (l_Arr[i].f_Index == a_Index)
                    {
                        l_Arr[i].f_Count++;
                        l_Arr[i].f_CumReputation += a_Reputation;

                        return;
                    }
                }

                int l_Idx = a_Clusters.ReserveAtEnd();
                a_Clusters.p_Arr[l_Idx] = new Cluster(a_Index, a_Reputation);
            }

            private static int PickStrongestCluster(ref FlexArray<Cluster> a_Clusters)
            {
                // Returns the f_Index field of the strongest cluster,
                // or subzero if all clusters are too weak.

                int l_Count = a_Clusters.p_Count;

                if (0 < l_Count)
                {
                    int l_BestIndex = 0;

                    Cluster[] l_Arr = a_Clusters.p_Arr;
                    for (int i = 1; i < l_Count; i++)
                    {
                        if (true == l_Arr[i].IsBetter(ref l_Arr[l_BestIndex]))
                        {
                            l_BestIndex = i;
                        }
                    }

                    if (c_BelieveHashUserCount <= l_Arr[l_BestIndex].f_Count &&
                        c_BelieveHashReputation <= l_Arr[l_BestIndex].f_CumReputation)
                    {
                        return l_Arr[l_BestIndex].f_Index;
                    }
                }

                return (-1);
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct Cluster
            {
                public readonly int f_Index;

                public int f_Count;
                public long f_CumReputation;

                public Cluster(int a_Index,
                               int a_Reputation)
                {
                    f_Index = a_Index;

                    f_Count = 1;
                    f_CumReputation = a_Reputation;
                }

                public bool IsBetter(ref Cluster a_Other)
                {
                    // Returns true if this is better than a_Other.

                    if (this.f_CumReputation > a_Other.f_CumReputation)
                    {
                        return true;
                    }

                    return (this.f_CumReputation == a_Other.f_CumReputation &&
                            this.f_Count > a_Other.f_Count);
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Sanctions
        {
            private const byte c_PosPointsPer8KiB = 1;
            private const byte c_NegPointsPer8KiB = 2;

            private FlexArray<UserManager.UserRepChange> f_Users;

            private Sanctions(int a_InitialCapacity)
            {
                FlexArray<UserManager.UserRepChange>.Create(a_InitialCapacity, out f_Users);
            }

            public static void Create(out Sanctions a_Dest)
            {
                a_Dest = new Sanctions(0);
            }

            public void Clear()
            {
                f_Users.Clear();
            }

            public void AddSanction(UserNum a_UserNum,
                                    uint a_PosSize,
                                    uint a_NegSize)
            {
                int l_Rep = Mul(a_PosSize, c_PosPointsPer8KiB);
                l_Rep -= Mul(a_NegSize, c_NegPointsPer8KiB);

                if (0 == l_Rep &&
                    0 != a_PosSize &&
                    0 == a_NegSize)
                {
                    l_Rep = 1;
                }

                AddSanction(a_UserNum, l_Rep);
            }

            public void AddSanctions(ref FlexArray<UserRange> a_Ranges,
                                     bool a_Positive)
            {
                UserRange[] l_Range = a_Ranges.p_Arr;
                for (int i = (a_Ranges.p_Count - 1); 0 <= i; i--)
                {
                    uint l_Size = l_Range[i].f_Rng.f_Size;
                    int l_Rep = (false == a_Positive ? -Mul(l_Size, c_NegPointsPer8KiB) : Mul(l_Size, c_PosPointsPer8KiB));

                    AddSanction(l_Range[i].f_UserNum, l_Rep);
                }
            }

            private void AddSanction(UserNum a_UserNum,
                                     int a_Points)
            {
                if (0 == a_Points)
                {
                    return;
                }

                UserManager.UserRepChange[] l_Arr = f_Users.p_Arr;
                for (int i = (f_Users.p_Count - 1); 0 <= i; i--)
                {
                    if (l_Arr[i].f_UserNum == a_UserNum)
                    {
                        l_Arr[i].f_ReputationDelta += a_Points;

                        return;
                    }
                }

                int l_Idx = f_Users.ReserveAtEnd();
                f_Users.p_Arr[l_Idx] = new UserManager.UserRepChange(a_UserNum, a_Points);
            }

            public void SendToUserManager(Instance a_Inst)
            {
                // After sending all content will be cleared.

                if (0 < f_Users.p_Count)
                {
                    UserManager.UserRepChange[] l_Arr = f_Users.ToArray(true);
                    f_Users.Clear();

                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(a_Inst.f_UserMngr.ContentMngrChangeReputation, l_Arr);
                    a_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);
                }

                // We don't wait.
            }

            private static int Mul(uint a_Size,
                                   uint a_PointsPer8KiB)
            {
                return (int)((a_PointsPer8KiB * a_Size) / (8 * 1024));
            }
        }

        public class TempBuffers
        {
            public readonly byte[] f_Buffer;
            public readonly ulong[] f_KeyBuffer;

            public readonly Tiger f_Tiger;
            public readonly MD_4 f_MD4;

            public TempBuffers()
            {
                f_Buffer = new byte[IncompleteFile.c_BufferSize];
                f_KeyBuffer = new ulong[IncompleteFile.c_MaxChunkKeysPerBuffer];

                f_Tiger = new Tiger();
                f_MD4 = new MD_4();
            }

            public ulong[] CreateKeyBuffer(uint a_ChunkCount)
            {
                return new ulong[(ushort)Math.Min(a_ChunkCount, c_MaxChunkKeysPerBuffer)];
            }

            // The following routines can be used for detecting
            // conflicting uses of the buffers along the codepath.

            [Conditional("DEBUG")]
            public void DebugAcquireBuffers()
            {
            }

            public bool DebugAreBuffersAcquired()
            {
                return false;
            }

            [Conditional("DEBUG")]
            public void DebugReleaseBuffers()
            {
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct PieceBounds
    {
        // This structure is used to describe portions of a chunk
        // (8 MiB). For this, the chunk is divided into 128 pieces
        // (64 KiB each). A bitmap is used to hold the state of each
        // piece.
        // But there's the problem of eD2k compatibility. eD2k uses
        // 9,500 KiB parts, that we cannot describe with our 64 KiB
        // bitmap. For this reason we've got two 'extras' that bridge
        // the gap. We choose this design since we desperately want
        // to get away from the non-2^n design eD2k suffers from.

        private const byte c_Size = 22;

        private const byte c_PieceCount = 128;
        // 64 KiB.
        public const uint c_PieceSize = (IncompleteFile.c_ChunkSize / c_PieceCount);

        private const uint c_BitsPerBitmap = c_PieceCount;
        private const uint c_BitsPerULong = Util.c_BitsPerULong; // c_PieceCount / 2

        // c_BlocksPerBitmap is 8.
        private const byte c_BlocksPerULong = (byte)(BlockBitmap.c_Blocks / 2);
        private const ulong c_BlockMask = byte.MaxValue;

        // The following three fields are not safed to disk.
        private StateFlags f_Flags;
        // If the bitmap flag is set, there may not be set bits
        // outside the range described by the following two indices.
        private byte f_BitmapLowIdx;
        private byte f_BitmapHighIdx;

        private ulong f_Bitmap0;
        private ulong f_Bitmap1;

        private byte f_ExtraAheadIdx;
        private byte f_ExtraBehindIdx;

        // a_ExtraAhead is the byte extent of the extra and works
        // from the end of the piece f_ExtraAheadIdx indicates
        // towards the start.
        private ushort f_ExtraAhead;
        // f_ExtraBehind is the byte extent of the extra and works
        // from the start of the piece f_ExtraBehindIdx indicates
        // towards the end.
        private ushort f_ExtraBehind;

        // Rules concerning extras:
        // - Extras are only valid if the corresponding flag is set
        // and if they have a non-zero extent.
        // - Extras may not lie on a piece that is set in the bitmap.
        // - Both extra ranges may lie on the same piece, but they must
        // not together cover the full piece (otherwise the bitmap must
        // be used).

        public bool p_IsSet
        {
            get { return (StateFlags.None != f_Flags); }
        }

        public bool p_HasBitmap
        {
            get { return (0 != (StateFlags.BitmapValid & f_Flags)); }
        }

        public bool p_HasExtra
        {
            get { return (0 != (StateFlags.ExtraPresent & f_Flags)); }
        }

        public static bool Load(byte[] a_Src,
                                int a_SrcIdx,
                                out PieceBounds a_Dest)
        {
            a_Dest = new PieceBounds();

            a_Dest.f_Bitmap0 = Util.ReadUInt64LE(a_Src, a_SrcIdx);
            a_Dest.f_Bitmap1 = Util.ReadUInt64LE(a_Src, a_SrcIdx + 8);

            a_Dest.f_ExtraAheadIdx = a_Src[a_SrcIdx + 16];
            a_Dest.f_ExtraBehindIdx = a_Src[a_SrcIdx + 17];

            a_Dest.f_ExtraAhead = Util.ReadUInt16LE(a_Src, a_SrcIdx + 18);
            a_Dest.f_ExtraBehind = Util.ReadUInt16LE(a_Src, a_SrcIdx + 20);

            // Pre-init to make the struct methods useable.
            a_Dest.f_Flags = StateFlags.None;
            a_Dest.f_BitmapLowIdx = 0;
            a_Dest.f_BitmapHighIdx = 0;

            if (0 != a_Dest.f_Bitmap0 ||
                0 != a_Dest.f_Bitmap1)
            {
                a_Dest.f_Flags = StateFlags.BitmapValid;
                a_Dest.CalcLowestAndHighestSetBitIndices();

                // Validate extras.

                if (0 != a_Dest.f_ExtraAhead &&
                    (c_PieceCount <= a_Dest.f_ExtraAheadIdx ||
                     true == a_Dest.GetPieceState(a_Dest.f_ExtraAheadIdx)))
                {
                    return false;
                }

                if (0 != a_Dest.f_ExtraBehind)
                {
                    if (c_PieceCount <= a_Dest.f_ExtraBehindIdx ||
                        true == a_Dest.GetPieceState(a_Dest.f_ExtraBehindIdx))
                    {
                        return false;
                    }

                    // Do an extra on extra overlap check.

                    if (0 != a_Dest.f_ExtraAhead &&
                        a_Dest.f_ExtraAheadIdx == a_Dest.f_ExtraBehindIdx &&
                        c_PieceSize <= (a_Dest.f_ExtraAhead + a_Dest.f_ExtraBehind))
                    {
                        return false;
                    }
                }
            }

            if (0 != a_Dest.f_ExtraAhead ||
                0 != a_Dest.f_ExtraBehind)
            {
                a_Dest.f_Flags |= StateFlags.ExtraPresent;
            }

            a_Dest.DebugNotify();

            return true;
        }

        public void Write(byte[] a_Dest,
                          int a_DestIdx)
        {
            // Since we don't write the flags, we need to make sure
            // the bitmap and the extras are zeroed if the flag isn't
            // set.

            DebugNotify();

            if (false == this.p_HasBitmap)
            {
                f_Bitmap0 = 0;
                f_Bitmap1 = 0;
            }

            if (false == this.p_HasExtra)
            {
                f_ExtraAhead = 0;
                f_ExtraBehind = 0;
            }

            Util.WriteUInt64LE(f_Bitmap0, a_Dest, a_DestIdx);
            Util.WriteUInt64LE(f_Bitmap1, a_Dest, a_DestIdx + 8);

            a_Dest[a_DestIdx + 16] = f_ExtraAheadIdx;
            a_Dest[a_DestIdx + 17] = f_ExtraBehindIdx;

            Util.WriteUInt16LE(f_ExtraAhead, a_Dest, a_DestIdx + 18);
            Util.WriteUInt16LE(f_ExtraBehind, a_Dest, a_DestIdx + 20);
        }

        public void Clear()
        {
            f_Flags = StateFlags.None;
        }

        public void InvalidateBitmap()
        {
            f_Flags &= ~StateFlags.BitmapValid;
        }

        public void InvalidateExtras()
        {
            f_Flags &= ~StateFlags.ExtraPresent;
        }

        public int EstimateMaxPieces()
        {
            DebugNotify();

            int l_PiecesAffected = 0;

            if (true == this.p_HasBitmap)
            {
                l_PiecesAffected = (f_BitmapHighIdx - f_BitmapLowIdx + 1);
            }

            if (true == this.p_HasExtra)
            {
                if (0 != f_ExtraAhead)
                {
                    l_PiecesAffected++;
                }

                if (0 != f_ExtraBehind)
                {
                    l_PiecesAffected++;
                }
            }

            return l_PiecesAffected;
        }

        public bool OutlineToChunkRange(out ChunkRange a_Range)
        {
            DebugNotify();

            uint l_StartPos = uint.MaxValue;
            uint l_FinalPos = 0;
            bool l_Valid = false;

            if (true == this.p_HasBitmap)
            {
                l_StartPos = CalcPieceOffset(f_BitmapLowIdx);
                l_FinalPos = CalcPieceOffset(f_BitmapHighIdx + 1u);
                l_Valid = true;
            }

            if (true == this.p_HasExtra)
            {
                if (0 != f_ExtraAhead)
                {
                    uint l_Base = CalcPieceOffset(f_ExtraAheadIdx + 1u);
                    l_StartPos = Math.Min(l_StartPos, l_Base - f_ExtraAhead);
                    l_FinalPos = Math.Max(l_FinalPos, l_Base);
                    l_Valid = true;
                }

                if (0 != f_ExtraBehind)
                {
                    uint l_Base = CalcPieceOffset(f_ExtraBehindIdx);
                    l_StartPos = Math.Min(l_StartPos, l_Base);
                    l_FinalPos = Math.Max(l_FinalPos, l_Base + f_ExtraBehind);
                    l_Valid = true;
                }
            }

            if (false == l_Valid)
            {
                a_Range = new ChunkRange();

                return false;
            }

            ChunkRange.CreateFromOffsets(l_StartPos, l_FinalPos, out a_Range);

            return true;
        }

        public bool Equals(ref PieceBounds a_Other)
        {
            this.DebugNotify();
            a_Other.DebugNotify();

            if (this.f_Flags != a_Other.f_Flags)
            {
                return false;
            }

            if (true == this.p_HasBitmap)
            {
                if (this.f_Bitmap0 != a_Other.f_Bitmap0 ||
                    this.f_Bitmap1 != a_Other.f_Bitmap1)
                {
                    return false;
                }
            }

            if (true == this.p_HasExtra)
            {
                return (this.f_ExtraAheadIdx == a_Other.f_ExtraAheadIdx &&
                        this.f_ExtraBehindIdx == a_Other.f_ExtraBehindIdx &&
                        this.f_ExtraAhead == a_Other.f_ExtraAhead &&
                        this.f_ExtraBehind == a_Other.f_ExtraBehind);
            }

            return true;
        }

        public void ReduceSize(uint a_MaxSize)
        {
            DebugNotify();

            bool l_HasBitmap = this.p_HasBitmap;

            if (true == p_HasExtra)
            {
                if (false == l_HasBitmap &&
                    ((uint)f_ExtraAhead + f_ExtraBehind) <= a_MaxSize)
                {
                    return;
                }

                InvalidateExtras();
            }

            uint l_KeepPieces = (a_MaxSize / c_PieceSize);

            if (true == l_HasBitmap)
            {
                bool l_ClearBitmap = true;

                if (0 != l_KeepPieces)
                {
                    for (uint i = f_BitmapLowIdx, lim = f_BitmapHighIdx; i <= lim; i++)
                    {
                        if (true == GetPieceState(i))
                        {
                            l_ClearBitmap = false;

                            if (0 == --l_KeepPieces)
                            {
                                f_BitmapHighIdx = (byte)i;

                                if ((c_BitsPerBitmap - 1) > i)
                                {
                                    SetPieceStates(i + 1, c_BitsPerBitmap - 1, false);
                                }

                                break;
                            }
                        }
                    }
                }

                if (true == l_ClearBitmap)
                {
                    InvalidateBitmap();
                }
            }

            DebugNotify();
        }

        public static bool CreateFromChunkRange(ref ChunkRange a_ChunkRange,
                                                out PieceBounds a_Bounds)
        {
            a_Bounds = new PieceBounds();

            if (0 == a_ChunkRange.f_Size)
            {
                return true;
            }

            uint l_ChunkRangeFinal = a_ChunkRange.p_Final;

            // NOTE: l_ExtraAhead is inverted (compared with how we
            // use it normally).
            uint l_ExtraAhead = (a_ChunkRange.f_Offset % c_PieceSize);
            uint l_ExtraBehind = (l_ChunkRangeFinal % c_PieceSize);

            uint l_BitIdx = ((a_ChunkRange.f_Offset / c_PieceSize) +
                             (0 == l_ExtraAhead ? 0u : 1u));
            uint l_FinalBitIdx = (l_ChunkRangeFinal / c_PieceSize);

            if (l_BitIdx > l_FinalBitIdx)
            {
                // We cannot represent that information.

                return false;
            }

            if (l_BitIdx < l_FinalBitIdx)
            {
                a_Bounds.f_Flags |= StateFlags.BitmapValid;
                a_Bounds.f_BitmapLowIdx = (byte)l_BitIdx;
                a_Bounds.f_BitmapHighIdx = (byte)(l_FinalBitIdx - 1);
                a_Bounds.f_Bitmap0 = CreateLowMask(a_Bounds.f_BitmapLowIdx, a_Bounds.f_BitmapHighIdx);
                a_Bounds.f_Bitmap1 = CreateHighMask(a_Bounds.f_BitmapLowIdx, a_Bounds.f_BitmapHighIdx);
            }

            if (0 != l_ExtraAhead)
            {
                a_Bounds.f_Flags |= StateFlags.ExtraPresent;
                a_Bounds.f_ExtraAhead = (ushort)(c_PieceSize - l_ExtraAhead);
                a_Bounds.f_ExtraAheadIdx = (byte)(l_BitIdx - 1);
            }

            if (0 != l_ExtraBehind)
            {
                a_Bounds.f_Flags |= StateFlags.ExtraPresent;
                a_Bounds.f_ExtraBehind = (ushort)l_ExtraBehind;
                a_Bounds.f_ExtraBehindIdx = (byte)l_FinalBitIdx;
            }

            a_Bounds.DebugNotify();

            return true;
        }

        public static bool CreateFromGapRanges(uint a_ChunkIndex,
                                               uint a_ChunkSize,
                                               ref int a_GapIndex,
                                               ref FlexArray<FileRange> a_GapRanges,
                                               out FromGapRangesCreationOutputContext a_Cntxt)
        {
            // f_GapRanges must be sorted and may not contain zero
            // length gaps or seamless gaps or gaps beyond the file's
            // size. f_GapRanges's count may be zero.

            uint l_ElapsedByteCount = 0;
            PieceBounds l_Bounds = new PieceBounds();

            ulong l_ChunkPos = ((ulong)a_ChunkIndex * IncompleteFile.c_ChunkSize);
            ulong l_ChunkFinalPos = (l_ChunkPos + a_ChunkSize);
            bool l_InitializedSeamInfo = false;
            eD2kSeamInfo l_SeamInfo = new eD2kSeamInfo();

            int l_GapCount = a_GapRanges.p_Count;
            FileRange[] l_Arr = a_GapRanges.p_Arr;
            for (; a_GapIndex <= l_GapCount; a_GapIndex++)
            {
                ulong l_MinPos = (0 == a_GapIndex ? 0 : l_Arr[a_GapIndex - 1].p_FinalOffset);
                ulong l_MaxPos;

                if (l_GapCount == a_GapIndex)
                {
                    if (l_MinPos >= l_ChunkFinalPos)
                    {
                        break;
                    }

                    l_MaxPos = l_ChunkFinalPos;
                }
                else
                {
                    l_MaxPos = l_Arr[a_GapIndex].f_StartingOffset;

                    if (l_MaxPos <= l_ChunkPos)
                    {
                        continue;
                    }
                }

                ChunkRange l_Range;
                ChunkRange.CreateClipped(l_ChunkPos, l_ChunkFinalPos, l_MinPos, l_MaxPos, out l_Range);

                if (0 != l_Range.f_Size)
                {
                    l_ElapsedByteCount += l_Range.f_Size;

                    if (false == l_InitializedSeamInfo)
                    {
                        l_InitializedSeamInfo = true;
                        l_SeamInfo = new eD2kSeamInfo(l_ChunkPos, l_ChunkFinalPos, true);
                    }

                    ClampRange(ref l_Range, ref l_SeamInfo);

                    if (0 != l_Range.f_Size)
                    {
                        l_Bounds.Add(ref l_Range);
                    }
                }

                if (l_MaxPos >= l_ChunkFinalPos)
                {
                    break;
                }
            }

            // Provide a hint to the caller, about which chunk index
            // might by tried next.

            if (true == l_Bounds.p_IsSet)
            {
                a_Cntxt = new FromGapRangesCreationOutputContext(l_ChunkPos, l_ElapsedByteCount, a_ChunkIndex + 1, ref l_Bounds);

                return true;
            }

            uint l_NextPossChunkIdx;

            if (l_GapCount <= a_GapIndex)
            {
                // End reached.

                l_NextPossChunkIdx = uint.MaxValue;
            }
            else if (0 == a_GapIndex)
            {
                // This should never happen.

                l_NextPossChunkIdx = (a_ChunkIndex + 1);
            }
            else
            {
                l_NextPossChunkIdx = (uint)(l_Arr[a_GapIndex - 1].p_FinalOffset / IncompleteFile.c_ChunkSize);
            }

            a_Cntxt = new FromGapRangesCreationOutputContext(l_ChunkPos, l_ElapsedByteCount, l_NextPossChunkIdx);

            return false;
        }

        private static void ClampRange(ref ChunkRange a_Range,
                                       ref eD2kSeamInfo a_SeamInfo)
        {
            // Removes extras from a ChunkRange, except if they
            // are compatible within the eD2k seam.

            uint l_StartingPos = a_Range.f_Offset;
            uint l_EndPos = a_Range.p_Final;

            bool l_ClampStart = true;
            bool l_ClampEnd = true;

            if (true == a_SeamInfo.f_Valid)
            {
                if (a_SeamInfo.f_PiecePos <= l_StartingPos &&
                    a_SeamInfo.f_PieceFinalPos >= l_StartingPos &&
                    a_SeamInfo.f_PieceFinalPos <= l_EndPos)
                {
                    l_ClampStart = false;
                }

                if (a_SeamInfo.f_PiecePos <= l_EndPos &&
                    a_SeamInfo.f_PieceFinalPos >= l_EndPos &&
                    a_SeamInfo.f_PiecePos >= l_StartingPos)
                {
                    if (false == l_ClampStart)
                    {
                        return;
                    }

                    l_ClampEnd = false;
                }
            }

            if (true == l_ClampStart)
            {
                uint l_Remainder = (l_StartingPos % c_PieceSize);

                if (0 != l_Remainder)
                {
                    l_StartingPos += (c_PieceSize - l_Remainder);
                }
            }

            if (true == l_ClampEnd)
            {
                l_EndPos -= (l_EndPos % c_PieceSize);
            }

            if (l_StartingPos >= l_EndPos)
            {
                a_Range = new ChunkRange();

                return;
            }

            ChunkRange.CreateFromOffsets(l_StartingPos, l_EndPos, out a_Range);
        }

        public bool GetPieceState(uint a_BitIndex)
        {
            // a_BitIndex is zero based. This routine does not bother
            // about f_BitmapLowIdx and f_BitmapHighIdx.

            ulong l_Mask;

            if (c_BitsPerULong > a_BitIndex)
            {
                l_Mask = (1UL << (int)a_BitIndex);

                return (0 != (f_Bitmap0 & l_Mask));
            }

            l_Mask = (1UL << (int)(a_BitIndex - c_BitsPerULong));

            return (0 != (f_Bitmap1 & l_Mask));
        }

        private bool SetPieceState(uint a_BitIndex,
                                   bool a_NewState)
        {
            // a_BitIndex is zero based. Returns the previous state of
            // the affected bit. This routine only updates the bitmap
            // itself.

            if (c_BitsPerULong > a_BitIndex)
            {
                return ModifyULong(ref f_Bitmap0, a_BitIndex, a_NewState);
            }

            return ModifyULong(ref f_Bitmap1, a_BitIndex - c_BitsPerULong, a_NewState);
        }

        private void SetPieceStates(uint a_FromBitIndex,
                                    uint a_ToBitIndex,
                                    bool a_NewState)
        {
            // a_BitIndex is zero based. This routine only updates the
            // bitmap itself.

            ulong l_Mask = CreateLowMask(a_FromBitIndex, a_ToBitIndex);

            if (0 != l_Mask)
            {
                ApplyMask(ref f_Bitmap0, l_Mask, a_NewState);
            }

            l_Mask = CreateHighMask(a_FromBitIndex, a_ToBitIndex);

            if (0 != l_Mask)
            {
                ApplyMask(ref f_Bitmap1, l_Mask, a_NewState);
            }
        }

        public bool DoesOverlap(ref PieceBounds a_Other)
        {
            this.DebugNotify();
            a_Other.DebugNotify();

            if (true == a_Other.p_HasBitmap)
            {
                if (true == this.p_HasBitmap)
                {
                    if (0 != (this.f_Bitmap0 & a_Other.f_Bitmap0) ||
                        0 != (this.f_Bitmap1 & a_Other.f_Bitmap1))
                    {
                        return true;
                    }
                }

                if (true == this.p_HasExtra)
                {
                    if (0 != this.f_ExtraAhead &&
                        true == a_Other.GetPieceState(this.f_ExtraAheadIdx))
                    {
                        return true;
                    }

                    if (0 != this.f_ExtraBehind &&
                        true == a_Other.GetPieceState(this.f_ExtraBehindIdx))
                    {
                        return true;
                    }
                }
            }

            if (true == a_Other.p_HasExtra)
            {
                if (true == this.p_HasExtra)
                {
                    if (0 != this.f_ExtraAhead &&
                        0 != a_Other.f_ExtraAhead &&
                        this.f_ExtraAheadIdx == a_Other.f_ExtraAheadIdx)
                    {
                        return true;
                    }

                    if (0 != this.f_ExtraBehind &&
                        0 != a_Other.f_ExtraBehind &&
                        this.f_ExtraBehindIdx == a_Other.f_ExtraBehindIdx)
                    {
                        return true;
                    }
                }

                if (true == this.p_HasBitmap)
                {
                    if (0 != a_Other.f_ExtraAhead &&
                        true == this.GetPieceState(a_Other.f_ExtraAheadIdx))
                    {
                        return true;
                    }

                    if (0 != a_Other.f_ExtraBehind &&
                        true == this.GetPieceState(a_Other.f_ExtraBehindIdx))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public bool IsRangeAvailable(ref ChunkRange a_Range)
        {
            return IsRangeAvailable(a_Range.f_Offset, a_Range.f_Offset + a_Range.f_Size);
        }

        public bool IsRangeAvailable(uint a_Offset,
                                     uint a_FinalOffset)
        {
            DebugNotify();

            // NOTE: l_ExtraAhead is inverted (compared with how we use
            // it normally).
            uint l_ExtraAhead = (a_Offset % c_PieceSize);
            uint l_ExtraBehind = (a_FinalOffset % c_PieceSize);

            uint l_BitIdx = ((a_Offset / c_PieceSize) +
                             (0 == l_ExtraAhead ? 0u : 1u));
            uint l_FinalBitIdx = (a_FinalOffset / c_PieceSize);

            bool l_HasBitmap = this.p_HasBitmap;

            if (l_BitIdx < l_FinalBitIdx)
            {
                uint l_ToBitIdx = (l_FinalBitIdx - 1);

                if (false == l_HasBitmap ||
                    this.f_BitmapLowIdx > l_BitIdx ||
                    this.f_BitmapHighIdx < l_ToBitIdx)
                {
                    return false;
                }

                ulong l_Mask = CreateLowMask(l_BitIdx, l_ToBitIdx);

                if (l_Mask != (this.f_Bitmap0 & l_Mask))
                {
                    return false;
                }

                l_Mask = CreateHighMask(l_BitIdx, l_ToBitIdx);

                if (l_Mask != (this.f_Bitmap1 & l_Mask))
                {
                    return false;
                }
            }
            else if (l_BitIdx > l_FinalBitIdx)
            {
                // This is a special case where the range is completely
                // in-between a piece.

                if (false == l_HasBitmap ||
                    false == this.GetPieceState(l_FinalBitIdx))
                {
                    return false;
                }
            }

            bool l_HasExtras = this.p_HasExtra;

            if (0 != l_ExtraAhead)
            {
                uint l_ExtraAheadIdx = (l_BitIdx - 1);

                if ((false == l_HasBitmap ||
                     false == this.GetPieceState(l_ExtraAheadIdx)) &&
                    (false == l_HasExtras ||
                     this.f_ExtraAheadIdx != l_ExtraAheadIdx ||
                     this.f_ExtraAhead < (c_PieceSize - l_ExtraAhead)))
                {
                    return false;
                }
            }

            if (0 != l_ExtraBehind)
            {
                if ((false == l_HasBitmap ||
                     false == this.GetPieceState(l_FinalBitIdx)) &&
                    (false == l_HasExtras ||
                     this.f_ExtraBehindIdx != l_FinalBitIdx ||
                     this.f_ExtraBehind < l_ExtraBehind))
                {
                    return false;
                }
            }

            return true;
        }

        public uint GetCoveredBytes()
        {
            DebugNotify();

            uint l_Sum = 0;

            if (true == p_HasBitmap)
            {
                for (ulong l_Mask = 1; 0 != l_Mask; l_Mask <<= 1)
                {
                    if (0 != (f_Bitmap0 & l_Mask))
                    {
                        l_Sum += c_PieceSize;
                    }

                    if (0 != (f_Bitmap1 & l_Mask))
                    {
                        l_Sum += c_PieceSize;
                    }
                }
            }

            if (true == p_HasExtra)
            {
                l_Sum += f_ExtraAhead;
                l_Sum += f_ExtraBehind;
            }

            return l_Sum;
        }

        public BlockBitmap ToBitmap(bool a_Narrow)
        {
            DebugNotify();

            BlockBitmap l_Map = new BlockBitmap();

            if (true == p_HasBitmap)
            {
                ulong l_Mask = c_BlockMask;
                for (uint i = 0; c_BlocksPerULong > i; i++, l_Mask <<= c_BlocksPerULong)
                {
                    bool l_Set0;
                    bool l_Set1;

                    if (false == a_Narrow)
                    {
                        l_Set0 = (0 != (f_Bitmap0 & l_Mask));
                        l_Set1 = (0 != (f_Bitmap1 & l_Mask));
                    }
                    else
                    {
                        l_Set0 = (l_Mask == (f_Bitmap0 & l_Mask));
                        l_Set1 = (l_Mask == (f_Bitmap1 & l_Mask));
                    }

                    if (true == l_Set0)
                    {
                        l_Map.Set(i);
                    }

                    if (true == l_Set1)
                    {
                        l_Map.Set(c_BlocksPerULong + i);
                    }
                }
            }

            if (false == a_Narrow &&
                true == p_HasExtra)
            {
                if (0 != f_ExtraAhead)
                {
                    l_Map.Set((uint)f_ExtraAheadIdx / c_BlocksPerULong);
                }

                if (0 != f_ExtraBehind)
                {
                    l_Map.Set((uint)f_ExtraBehindIdx / c_BlocksPerULong);
                }
            }

            return l_Map;
        }

        public void Add(BlockBitmap a_Map)
        {
            DebugNotify();

            uint l_Val = a_Map.p_Value;

            if (0 == l_Val)
            {
                return;
            }

            if (false == p_HasBitmap)
            {
                // Prepare (clear) bitmap fields.

                f_Bitmap0 = 0;
                f_Bitmap1 = 0;
            }

            ulong l_Mask = c_BlockMask;
            for (uint i = 0; c_BlocksPerULong > i; i++, l_Mask <<= c_BlocksPerULong)
            {
                if (true == a_Map.IsSet(i))
                {
                    f_Bitmap0 |= l_Mask;
                }

                if (true == a_Map.IsSet(c_BlocksPerULong + i))
                {
                    f_Bitmap1 |= l_Mask;
                }
            }

            f_Flags |= StateFlags.BitmapValid;
            CalcLowestAndHighestSetBitIndices();
            UpdateExtrasAfterBitmapChange();

            DebugNotify();
        }

        public bool Add(ref ChunkRange a_Range)
        {
            PieceBounds l_Temp;
            if (true == PieceBounds.CreateFromChunkRange(ref a_Range, out l_Temp))
            {
                this.Add(ref l_Temp);

                return true;
            }

            return false;
        }

        public void Add(ref PieceBounds a_Other)
        {
            Add(ref a_Other, AdditionMode.DontOverwriteExistingExtra);
        }

        public void Add(ref PieceBounds a_Other,
                        AdditionMode a_Mode)
        {
            this.DebugNotify();
            a_Other.DebugNotify();

            if (true == a_Other.p_HasBitmap)
            {
                // Take over bits from a_Other.

                if (false == this.p_HasBitmap)
                {
                    this.f_Flags |= StateFlags.BitmapValid;
                    this.f_BitmapLowIdx = a_Other.f_BitmapLowIdx;
                    this.f_BitmapHighIdx = a_Other.f_BitmapHighIdx;
                    this.f_Bitmap0 = a_Other.f_Bitmap0;
                    this.f_Bitmap1 = a_Other.f_Bitmap1;
                }
                else
                {
                    this.f_BitmapLowIdx = Math.Min(this.f_BitmapLowIdx, a_Other.f_BitmapLowIdx);
                    this.f_BitmapHighIdx = Math.Max(this.f_BitmapHighIdx, a_Other.f_BitmapHighIdx);
                    this.f_Bitmap0 |= a_Other.f_Bitmap0;
                    this.f_Bitmap1 |= a_Other.f_Bitmap1;
                }

                UpdateExtrasAfterBitmapChange();
            }

            if (true == a_Other.p_HasExtra)
            {
                // Take over extras from a_Other.

                bool l_HasBitmap = this.p_HasBitmap;
                bool l_HasExtras = this.p_HasExtra;

                if (0 != a_Other.f_ExtraAhead &&
                    (false == l_HasBitmap ||
                     false == this.GetPieceState(a_Other.f_ExtraAheadIdx)))
                {
                    bool l_TakeOver;

                    if (false == l_HasExtras)
                    {
                        l_HasExtras = true;
                        l_TakeOver = true;

                        // When we set the flag, we need to make sure
                        // that both extras are initialized properly.
                        this.f_Flags |= StateFlags.ExtraPresent;
                        this.f_ExtraBehind = 0;
                    }
                    else
                    {
                        if (0 == this.f_ExtraAhead)
                        {
                            l_TakeOver = true;
                        }
                        else
                        {
                            if (this.f_ExtraAheadIdx == a_Other.f_ExtraAheadIdx)
                            {
                                l_TakeOver = false;
                                this.f_ExtraAhead = Math.Max(this.f_ExtraAhead, a_Other.f_ExtraAhead);
                            }
                            else
                            {
                                l_TakeOver = (AdditionMode.AlwaysTakeOverExtra == a_Mode);
                            }
                        }
                    }

                    if (true == l_TakeOver)
                    {
                        this.f_ExtraAheadIdx = a_Other.f_ExtraAheadIdx;
                        this.f_ExtraAhead = a_Other.f_ExtraAhead;
                    }
                }

                if (0 != a_Other.f_ExtraBehind &&
                    (false == l_HasBitmap ||
                     false == this.GetPieceState(a_Other.f_ExtraBehindIdx)))
                {
                    bool l_TakeOver;

                    if (false == l_HasExtras)
                    {
                        l_HasExtras = true;
                        l_TakeOver = true;

                        // When we set the flag, we need to make sure
                        // that both extras are initialized properly.
                        this.f_Flags |= StateFlags.ExtraPresent;
                        this.f_ExtraAhead = 0;
                    }
                    else
                    {
                        if (0 == this.f_ExtraBehind)
                        {
                            l_TakeOver = true;
                        }
                        else
                        {
                            if (this.f_ExtraBehindIdx == a_Other.f_ExtraBehindIdx)
                            {
                                l_TakeOver = false;
                                this.f_ExtraBehind = Math.Max(this.f_ExtraBehind, a_Other.f_ExtraBehind);
                            }
                            else
                            {
                                l_TakeOver = (AdditionMode.AlwaysTakeOverExtra == a_Mode);
                            }
                        }
                    }

                    if (true == l_TakeOver)
                    {
                        this.f_ExtraBehindIdx = a_Other.f_ExtraBehindIdx;
                        this.f_ExtraBehind = a_Other.f_ExtraBehind;
                    }
                }

                if (true == l_HasExtras &&
                    0 != this.f_ExtraAhead &&
                    0 != this.f_ExtraBehind &&
                    this.f_ExtraAheadIdx == this.f_ExtraBehindIdx &&
                    c_PieceSize <= (this.f_ExtraAhead + this.f_ExtraBehind))
                {
                    // The extras are so big that they can be removed.

                    this.InvalidateExtras();

                    if (false == l_HasBitmap)
                    {
                        this.f_Flags |= StateFlags.BitmapValid;
                        this.f_BitmapLowIdx = this.f_ExtraAheadIdx;
                        this.f_BitmapHighIdx = this.f_BitmapLowIdx;
                        this.f_Bitmap0 = 0;
                        this.f_Bitmap1 = 0;
                    }
                    else
                    {
                        this.f_BitmapLowIdx = Math.Min(this.f_BitmapLowIdx, this.f_ExtraAheadIdx);
                        this.f_BitmapHighIdx = Math.Max(this.f_BitmapHighIdx, this.f_ExtraAheadIdx);
                    }

                    this.SetPieceState(this.f_ExtraAheadIdx, true);
                }
            }

            DebugNotify();
        }

        private void UpdateExtrasAfterBitmapChange()
        {
            if (false == p_HasExtra)
            {
                return;
            }

            if (0 != f_ExtraAhead &&
                true == GetPieceState(f_ExtraAheadIdx))
            {
                f_ExtraAhead = 0;
            }

            if (0 != f_ExtraBehind &&
                true == GetPieceState(f_ExtraBehindIdx))
            {
                f_ExtraBehind = 0;
            }

            if (0 == f_ExtraAhead &&
                0 == f_ExtraBehind)
            {
                // Clear the extra flag.

                InvalidateExtras();
            }
        }

        public void Sub(ref PieceBounds a_Subtrahend,
                        SubtractionMode a_Mode)
        {
            // This is the minuend.

            this.DebugNotify();
            a_Subtrahend.DebugNotify();

            if (false == this.p_IsSet ||
                false == a_Subtrahend.p_IsSet)
            {
                return;
            }

            bool l_HasBitmap = this.p_HasBitmap;
            bool l_HasExtras = this.p_HasExtra;
            bool l_RecheckBitmap = false;
            bool l_RecheckExtras = false;

            if (true == a_Subtrahend.p_HasBitmap)
            {
                // Take over bits from a_Other.

                if (true == l_HasBitmap)
                {
                    if (this.f_BitmapLowIdx <= a_Subtrahend.f_BitmapHighIdx &&
                        this.f_BitmapHighIdx >= a_Subtrahend.f_BitmapLowIdx)
                    {
                        this.f_Bitmap0 &= ~a_Subtrahend.f_Bitmap0;
                        this.f_Bitmap1 &= ~a_Subtrahend.f_Bitmap1;
                        l_RecheckBitmap = true;
                    }
                }

                if (true == l_HasExtras)
                {
                    if (0 != this.f_ExtraAhead &&
                        true == a_Subtrahend.GetPieceState(this.f_ExtraAheadIdx))
                    {
                        this.f_ExtraAhead = 0;
                        l_RecheckExtras = true;
                    }

                    if (0 != this.f_ExtraBehind &&
                        true == a_Subtrahend.GetPieceState(this.f_ExtraBehindIdx))
                    {
                        this.f_ExtraBehind = 0;
                        l_RecheckExtras = true;
                    }
                }
            }

            if (true == a_Subtrahend.p_HasExtra)
            {
                if (0 != a_Subtrahend.f_ExtraAhead)
                {
                    uint l_EffIdx = a_Subtrahend.f_ExtraAheadIdx;

                    if (true == l_HasExtras)
                    {
                        if (0 != this.f_ExtraAhead &&
                            l_EffIdx == this.f_ExtraAheadIdx &&
                            (SubtractionMode.MaximalRemoval == a_Mode ||
                             this.f_ExtraAhead <= a_Subtrahend.f_ExtraAhead))
                        {
                            this.f_ExtraAhead = 0;
                            l_RecheckExtras = true;
                        }

                        if (0 != this.f_ExtraBehind &&
                            l_EffIdx == this.f_ExtraBehindIdx)
                        {
                            this.f_ExtraBehind = Math.Min(this.f_ExtraBehind, (byte)(c_PieceSize - a_Subtrahend.f_ExtraAhead));
                            l_RecheckExtras = true;
                        }
                    }

                    if (true == l_HasBitmap &&
                        SubtractionMode.MaximalRemoval == a_Mode)
                    {
                        this.SetPieceState(l_EffIdx, false);
                        l_RecheckBitmap = true;
                    }
                }

                if (0 != a_Subtrahend.f_ExtraBehind)
                {
                    uint l_EffIdx = a_Subtrahend.f_ExtraBehindIdx;

                    if (true == l_HasExtras)
                    {
                        if (0 != this.f_ExtraAhead &&
                            l_EffIdx == this.f_ExtraAheadIdx)
                        {
                            this.f_ExtraAhead = Math.Min(this.f_ExtraAhead, (byte)(c_PieceSize - a_Subtrahend.f_ExtraBehind));
                            l_RecheckExtras = true;
                        }

                        if (0 != this.f_ExtraBehind &&
                            l_EffIdx == this.f_ExtraBehindIdx &&
                            (SubtractionMode.MaximalRemoval == a_Mode ||
                             this.f_ExtraBehind <= a_Subtrahend.f_ExtraBehind))
                        {
                            this.f_ExtraBehind = 0;
                            l_RecheckExtras = true;
                        }
                    }

                    if (true == l_HasBitmap &&
                        SubtractionMode.MaximalRemoval == a_Mode)
                    {
                        this.SetPieceState(l_EffIdx, false);
                        l_RecheckBitmap = true;
                    }
                }
            }

            // Update flags.

            if (true == l_RecheckBitmap &&
                false == CalcLowestAndHighestSetBitIndices())
            {
                this.InvalidateBitmap();
            }

            if (true == l_RecheckExtras &&
                0 == this.f_ExtraAhead &&
                0 == this.f_ExtraBehind)
            {
                this.InvalidateExtras();
            }

            DebugNotify();
        }

        private bool CalcLowestAndHighestSetBitIndices()
        {
            // Returns false if there is no set bit.

            if (0 == f_Bitmap0)
            {
                if (0 == f_Bitmap1)
                {
                    return false;
                }

                f_BitmapLowIdx = (byte)(c_BitsPerULong + Util.GetClearLowBits64(f_Bitmap1));
                f_BitmapHighIdx = (byte)(c_BitsPerBitmap - 1 - Util.GetClearHighBits64(f_Bitmap1));
            }
            else
            {
                f_BitmapLowIdx = (byte)Util.GetClearLowBits64(f_Bitmap0);

                if (0 == f_Bitmap1)
                {
                    f_BitmapHighIdx = (byte)(c_BitsPerULong - 1 - Util.GetClearHighBits64(f_Bitmap0));
                }
                else
                {
                    f_BitmapHighIdx = (byte)(c_BitsPerBitmap - 1 - Util.GetClearHighBits64(f_Bitmap1));
                }
            }

            return true;
        }

        public void BuildEnumerator(out Enumerator a_Enum)
        {
            a_Enum = new Enumerator(ref this, int.MaxValue);
        }

        public void BuildEnumerator(int a_MaxPiecesPerGroup,
                                    out Enumerator a_Enum)
        {
            DebugNotify();

            a_Enum = new Enumerator(ref this, a_MaxPiecesPerGroup);
        }

        public static uint CalcPieceOffset(uint a_PieceIndex)
        {
            return (a_PieceIndex * c_PieceSize);
        }

        private static ulong CreateLowMask(uint a_FromBitIdx,
                                           uint a_ToBitIdx)
        {
            // Return can be null.

            if (c_BitsPerULong > a_FromBitIdx)
            {
                return CreateMask(a_FromBitIdx, Math.Min(c_BitsPerULong - 1, a_ToBitIdx));
            }

            return 0;
        }

        private static ulong CreateHighMask(uint a_FromBitIdx,
                                            uint a_ToBitIdx)
        {
            // Return can be null.

            if (c_BitsPerULong <= a_ToBitIdx)
            {
                uint l_EffFromBitIdx = (c_BitsPerULong >= a_FromBitIdx ? 0 : (a_FromBitIdx - c_BitsPerULong));
                return CreateMask(l_EffFromBitIdx, a_ToBitIdx - c_BitsPerULong);
            }

            return 0;
        }

        private static ulong CreateMask(uint a_FromBitIdx,
                                        uint a_ToBitIdx)
        {
            // a_FromLowBit and a_ToHighBit form an inclusive
            // range.

            int l_Count = (int)((a_ToBitIdx - a_FromBitIdx) + 1u);

            if (c_BitsPerULong == l_Count)
            {
                return (ulong.MaxValue << (int)a_FromBitIdx);
            }

            return (((1UL << l_Count) - 1) << (int)a_FromBitIdx);
        }

        private static bool ModifyULong(ref ulong a_Target,
                                        uint a_BitIndex,
                                        bool a_NewState)
        {
            ulong l_Mask = (1UL << (int)(a_BitIndex - c_BitsPerULong));
            bool l_OldState = (0 != (a_Target & l_Mask));

            ApplyMask(ref a_Target, l_Mask, a_NewState);

            return l_OldState;
        }

        private static void ApplyMask(ref ulong a_Target,
                                      ulong a_Mask,
                                      bool a_NewState)
        {
            if (false == a_NewState)
            {
                a_Target &= ~a_Mask;
            }
            else
            {
                a_Target |= a_Mask;
            }
        }

        [Conditional("DEBUG")]
        private void DebugNotify()
        {
        }

        public enum AdditionMode : byte
        {
            DontOverwriteExistingExtra = 0,
            AlwaysTakeOverExtra = 1,
        }

        public enum SubtractionMode : byte
        {
            // This enum controls the handling of extras that don't
            // really overlap, but would need to be removed in order
            // to get rid of anything that overlaps when subtracting.

            MinimalRemoval = 0,
            MaximalRemoval = 1,
        }

        [Flags]
        private enum StateFlags : byte
        {
            None = 0,

            BitmapValid = 0x1,
            ExtraPresent = 0x2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct FromGapRangesCreationOutputContext
        {
            // Don't use the struct default constructor.

            public readonly ulong f_ChunkPos;
            public readonly uint f_ElapsedByteCount;
            public readonly uint f_TranslatedByteCount;
            public readonly uint f_NextPossibleChunkIndex;

            public PieceBounds f_Bounds;

            public FromGapRangesCreationOutputContext(ulong a_ChunkPos,
                                                      uint a_ElapsedByteCount,
                                                      uint a_NextPossibleChunkIndex,
                                                      ref PieceBounds a_Bounds)
            {
                // Constructor for success.

                f_ChunkPos = a_ChunkPos;
                f_ElapsedByteCount = a_ElapsedByteCount;
                f_TranslatedByteCount = a_Bounds.GetCoveredBytes();
                f_NextPossibleChunkIndex = a_NextPossibleChunkIndex;
                f_Bounds = a_Bounds;
            }

            public FromGapRangesCreationOutputContext(ulong a_ChunkPos,
                                                      uint a_ElapsedByteCount,
                                                      uint a_NextPossibleChunkIndex)
            {
                // Constructor for failure.

                f_ChunkPos = a_ChunkPos;
                f_ElapsedByteCount = a_ElapsedByteCount;
                f_TranslatedByteCount = 0;
                f_NextPossibleChunkIndex = a_NextPossibleChunkIndex;
                f_Bounds = new PieceBounds();
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct eD2kSeamInfo
        {
            // Don't use the struct default constructor.

            public readonly bool f_Valid;
            public readonly uint f_PiecePos;
            public readonly uint f_PieceFinalPos;

            public eD2kSeamInfo(ulong a_ChunkPos,
                                ulong a_ChunkFinalPos,
                                bool a_IfThereIsNoSeamTreatFileEndAsSeam)
            {
                uint l_ChunkSize = (uint)(a_ChunkFinalPos - a_ChunkPos);

                uint l_Remainder = (uint)(a_ChunkFinalPos % eD2kHashSet.c_PartSize);
                uint l_SeamPos;

                if (0 != l_Remainder &&
                    l_Remainder < l_ChunkSize)
                {
                    // There is a eD2k seam right within this chunk.
                    // l_SeamPos can be a multiple of c_PieceSize.

                    ulong l_InnerSeamPos = (a_ChunkFinalPos - l_Remainder);

                    l_SeamPos = (uint)(l_InnerSeamPos - a_ChunkPos);
                }
                else
                {
                    // There is no eD2k seam within this chunk (but
                    // maybe right at the start or at the end).

                    if (false == a_IfThereIsNoSeamTreatFileEndAsSeam ||
                        IncompleteFile.c_ChunkSize == l_ChunkSize ||
                        0 == (l_ChunkSize % c_PieceSize))
                    {
                        f_Valid = false;
                        f_PiecePos = 0;
                        f_PieceFinalPos = 0;

                        return;
                    }

                    // We are at the end of the file and treat the end
                    // and as a seam.

                    l_SeamPos = l_ChunkSize;
                }

                f_Valid = true;
                f_PiecePos = (l_SeamPos - (l_SeamPos % c_PieceSize));
                f_PieceFinalPos = Math.Min(l_ChunkSize, f_PiecePos + c_PieceSize);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Enumerator
        {
            private const uint c_Start = uint.MaxValue;
            private const uint c_EndReached = (uint.MaxValue - 1);

            public readonly PieceBounds f_Bounds;
            public readonly int f_MaxPiecesPerGroup;
            // The estimated maximum times GetNextRange() will return
            // true in a row.
            public readonly int f_MaxGroups;

            private uint f_Index;
            private uint f_MaxIndex;
            private bool f_NeedToHandleExtraAhead;
            private bool f_NeedToHandleExtraBehind;

            public Enumerator(ref PieceBounds a_Bounds,
                              int a_MaxPiecesPerGroup)
            {
                f_Bounds = a_Bounds;
                f_MaxPiecesPerGroup = a_MaxPiecesPerGroup;

                int l_MaxPieces = f_Bounds.EstimateMaxPieces();
                f_MaxGroups = ((l_MaxPieces / a_MaxPiecesPerGroup) +
                               (0 == (l_MaxPieces % a_MaxPiecesPerGroup) ? 0 : 1));

                // Reset.
                f_Index = c_Start;
                f_MaxIndex = 0;
                f_NeedToHandleExtraAhead = false;
                f_NeedToHandleExtraBehind = false;
            }

            public void Reset()
            {
                f_Index = c_Start;
            }

            public bool GetNextRange(out ChunkRange a_Range)
            {
                a_Range = new ChunkRange();

                if (c_Start == f_Index)
                {
                    if (false == f_Bounds.p_IsSet)
                    {
                        return false;
                    }

                    if (false == f_Bounds.p_HasBitmap)
                    {
                        f_MaxIndex = 0;
                    }
                    else
                    {
                        f_Index = f_Bounds.f_BitmapLowIdx;
                        f_MaxIndex = f_Bounds.f_BitmapHighIdx;
                    }

                    f_NeedToHandleExtraAhead = false;
                    f_NeedToHandleExtraBehind = false;

                    if (true == f_Bounds.p_HasExtra)
                    {
                        if (0 != f_Bounds.f_ExtraAhead)
                        {
                            f_Index = Math.Min(f_Index, f_Bounds.f_ExtraAheadIdx);
                            f_MaxIndex = Math.Max(f_MaxIndex, f_Bounds.f_ExtraAheadIdx);
                            f_NeedToHandleExtraAhead = true;
                        }

                        if (0 != f_Bounds.f_ExtraBehind)
                        {
                            f_Index = Math.Min(f_Index, f_Bounds.f_ExtraBehindIdx);
                            f_MaxIndex = Math.Max(f_MaxIndex, f_Bounds.f_ExtraBehindIdx);
                            f_NeedToHandleExtraBehind = true;
                        }
                    }
                }

                int l_Count = 0;

                for (; f_Index <= f_MaxIndex; )
                {
                    bool l_Set = (true == f_Bounds.p_HasBitmap &&
                                  f_Index >= f_Bounds.f_BitmapLowIdx &&
                                  f_Index <= f_Bounds.f_BitmapHighIdx &&
                                  f_Bounds.GetPieceState(f_Index));

                    if (false == l_Set)
                    {
                        if (true == f_NeedToHandleExtraBehind &&
                            f_Index == f_Bounds.f_ExtraBehindIdx)
                        {
                            if (0 == l_Count)
                            {
                                a_Range.f_Offset = CalcPieceOffset(f_Index);
                            }
                            else
                            {
                            }

                            f_NeedToHandleExtraBehind = false;
                            a_Range.f_Size += f_Bounds.f_ExtraBehind;

                            l_Count++;

                            // Don't inc f_Index here, we might be able
                            // to add the extra ahead also.
                        }

                        if (true == f_NeedToHandleExtraAhead &&
                            f_Index == f_Bounds.f_ExtraAheadIdx)
                        {
                            if (0 == l_Count)
                            {
                                a_Range.f_Offset = (CalcPieceOffset(f_Index + 1) - f_Bounds.f_ExtraAhead);
                                a_Range.f_Size = f_Bounds.f_ExtraAhead;

                                l_Count++;
                            }
                            else
                            {
                                if (f_Index == f_Bounds.f_ExtraBehindIdx &&
                                    c_PieceSize == (f_Bounds.f_ExtraBehind + f_Bounds.f_ExtraAhead))
                                {
                                    // Continue from extra behind.

                                    a_Range.f_Size += f_Bounds.f_ExtraAhead;
                                }
                                else
                                {
                                    return true;
                                }
                            }

                            f_NeedToHandleExtraAhead = false;

                            f_Index++;

                            if (f_MaxPiecesPerGroup == l_Count)
                            {
                                return true;
                            }

                            continue;
                        }

                        // We land here if no extra ahead handling was done.

                        f_Index++;

                        if (0 != l_Count)
                        {
                            return true;
                        }
                    }
                    else
                    {
                        if (0 == l_Count)
                        {
                            a_Range.f_Offset = CalcPieceOffset(f_Index);
                        }
                        else
                        {
                        }

                        a_Range.f_Size += c_PieceSize;

                        f_Index++;
                        l_Count++;

                        if (f_MaxPiecesPerGroup == l_Count)
                        {
                            return true;
                        }
                    }
                }

                return (0 != l_Count);
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct BlockBitmap
    {
        public const byte c_Size = sizeof(ushort);
        // c_Blocks is 16.
        public const byte c_Blocks = (byte)(IncompleteFile.c_ChunkSize / TAICHTree.c_BlockSize);

        // Only the low 16 bits of f_Value are used.
        private uint f_Value;

        public uint p_Value
        {
            get { return f_Value; }
        }

        public BlockBitmap(uint a_Value)
        {
            f_Value = a_Value;
        }

        public BlockBitmap(byte[] a_Src,
                           int a_SrcIdx)
        {
            f_Value = Util.ReadUInt16LE(a_Src, a_SrcIdx);
        }

        public void Write(byte[] a_Dest,
                          int a_DestIdx)
        {
            Util.WriteUInt16LE((ushort)f_Value, a_Dest, a_DestIdx);
        }

        public bool IsSet()
        {
            return (0 != f_Value);
        }

        public void RestrictTo(uint a_Mask)
        {
            f_Value &= a_Mask;
        }

        public void TakeOver(uint a_Value)
        {
            f_Value = a_Value;
        }

        public void ClearAll()
        {
            f_Value = 0;
        }

        public uint CountSetBits()
        {
            return BitMap.CountSetBits(f_Value);
        }

        public bool IsSet(uint a_Index)
        {
            return TestBit(a_Index, f_Value);
        }

        public void Clear(uint a_Index)
        {
            ClearBit(a_Index, ref f_Value);
        }

        public void Set(uint a_Index)
        {
            SetBit(a_Index, ref f_Value);
        }

        public void DeleteMask(uint a_Mask)
        {
            f_Value &= ~a_Mask;
        }

        public void AddMask(uint a_Mask)
        {
            f_Value |= a_Mask;
        }

        public static void ClearBit(uint a_BitIndex,
                                    ref uint a_Bitmap)
        {
            a_Bitmap &= ~(1u << (int)a_BitIndex);
        }

        public static void SetBit(uint a_BitIndex,
                                  ref uint a_Bitmap)
        {
            a_Bitmap |= (1u << (int)a_BitIndex);
        }

        public static bool TestBit(uint a_BitIndex,
                                   uint a_Bitmap)
        {
            return (0 != (a_Bitmap & (1u << (int)a_BitIndex)));
        }

        public static uint BuildBlockMask(uint a_BlockIndex,
                                          uint a_BlockCount)
        {
            return (((1u << (int)a_BlockCount) - 1) << (int)a_BlockIndex);
        }

        public static uint CalcBlockBoundsWide(uint a_Offset,
                                               uint a_FinalOffset,
                                               out uint a_BlockIndex)
        {
            // Returns the number of blocks affected.

            a_BlockIndex = (a_Offset / TAICHTree.c_BlockSize);
            uint l_FinalBlockIdx = ((a_FinalOffset / TAICHTree.c_BlockSize) +
                                    (0 == (a_FinalOffset % TAICHTree.c_BlockSize) ? 0u : 1u));

            return (l_FinalBlockIdx - a_BlockIndex);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct ChunkAvail
    {
        public readonly uint f_ChunkIndex;
        private readonly uint f_Rand;

        private bool f_Invalidated;
        private uint f_NetworkAvail;

        public bool p_Invalidated
        {
            get { return f_Invalidated; }
        }

        public uint p_NetworkAvail
        {
            get { return f_NetworkAvail; }
        }

        public ChunkAvail(uint a_ChunkIndex)
        {
            f_ChunkIndex = a_ChunkIndex;
            f_Rand = Util.GetRandomUIntFast();
            f_Invalidated = false;
            f_NetworkAvail = 0;
        }

        public void SetNetworkAvail(uint a_NetworkAvail)
        {
            f_NetworkAvail = a_NetworkAvail;
        }

        public void Invalidate()
        {
            f_Invalidated = true;
        }

        public int CompareTo(ref ChunkAvail a_Other)
        {
            int l_Res = this.f_NetworkAvail.CompareTo(a_Other.f_NetworkAvail);

            if (0 == l_Res)
            {
                // If two chunks have the same availability, flip a
                // coin to determine which one to pick first.
                l_Res = this.f_Rand.CompareTo(a_Other.f_Rand);
            }

            return l_Res;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct ChunkRange
    {
        public uint f_Offset;
        public uint f_Size;

        public uint p_Final
        {
            get { return (f_Offset + f_Size); }
        }

        public ChunkRange(uint a_Offset,
                          uint a_Size)
        {
            f_Offset = a_Offset;
            f_Size = a_Size;
        }

        public bool IsAvailable(uint a_Offset,
                                uint a_FinalOffset)
        {
            return (f_Offset <= a_Offset &&
                    p_Final >= a_FinalOffset);
        }

        public static void CreateFromOffsets(uint a_Offset,
                                             uint a_FinalOffset,
                                             out ChunkRange a_Range)
        {
            a_Range = new ChunkRange(a_Offset, a_FinalOffset - a_Offset);
        }

        public static void CreateClipped(ulong a_BasePos,
                                         ulong a_BaseFinalPos,
                                         ulong a_MinPos,
                                         ulong a_MaxPos,
                                         out ChunkRange a_ChunkRange)
        {
            if (a_MinPos >= a_BaseFinalPos ||
                a_MaxPos <= a_BasePos)
            {
                // Ranges don't overlap.

                a_ChunkRange = new ChunkRange();

                return;
            }

            ChunkRange.CreateFromOffsets(a_MinPos <= a_BasePos ? 0u : (uint)(a_MinPos - a_BasePos),
                                         (uint)(Math.Min(a_MaxPos, a_BaseFinalPos) - a_BasePos),
                                         out a_ChunkRange);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct UserRange
    {
        public const byte c_Size = 16;

        public readonly UserNum f_UserNum; // 0
        public readonly int f_Slot; // not written
        public readonly ChunkRange f_Rng; // 8

        public UserRange(UserNum a_UserNum,
                         int a_Slot,
                         ref ChunkRange a_ChunkRange)
        {
            f_UserNum = a_UserNum;
            f_Slot = a_Slot;
            f_Rng = a_ChunkRange;
        }

        public void Write(byte[] a_Dest,
                          int a_DestIdx)
        {
            Util.WriteUInt64LE(f_UserNum.f_Val, a_Dest, a_DestIdx);
            Util.WriteUInt32LE(f_Rng.f_Offset, a_Dest, a_DestIdx + 8);
            Util.WriteUInt32LE(f_Rng.f_Size, a_Dest, a_DestIdx + 12);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct HashDemand
    {
        public readonly HashDemandTypes f_Type;
        public readonly uint f_ChunkIndex;

        public HashDemand(HashDemandTypes a_Type) :
                          this(a_Type, 0)
        {
        }

        public HashDemand(HashDemandTypes a_Type,
                               uint a_ChunkIndex)
        {
            f_Type = a_Type;
            f_ChunkIndex = a_ChunkIndex;
        }

        public bool ValueEquals(ref HashDemand a_Other)
        {
            return (this.f_Type == a_Other.f_Type &&
                    this.f_ChunkIndex == a_Other.f_ChunkIndex);
        }

        public override int GetHashCode()
        {
            return ((int)f_Type ^ (int)f_ChunkIndex);
        }
    }

    enum HashDemandTypes : byte
    {
        NeedeD2kHashSet = 0,
        NeedTAICHUpperTree = 1,
        NeedTAICHBlockHashes = 2,
    }
}

namespace NB.Content.Internal.IncomplInput
{
    enum OpStatus : byte
    {
        Done = 0,
        WontCreateElmSinceFileIsNotIncompl,
        NotReady,
        IoError,
        InsuffDiskSpace,
        NeedFurtherCalls,
        NotEnoughData,
        Corruption,
        NoMatch,
        InputRejected,
        RequestTooLarge,
        NothingToDo,
        NeedReverification,
        ChunkHasNoKey,
        Error,

        // Verification.
        ChunkVerified,
        BlockVerified,
        ChunkNotVerified,
        BlockNotVerified,
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct UpdateSourceContext
    {
        public readonly SourceCookie f_Cookie;
        public readonly P2PHashSets f_ProvidableHashSets;
        public readonly ulong f_FileSize;
        public MasterHashes f_MasterHashes;

        public UpdateSourceContext(SourceCookie a_Cookie,
                                   P2PHashSets a_ProvidableHashSets,
                                   ulong a_FileSize,
                                   ref MasterHashes a_MasterHashes)
        {
            f_Cookie = a_Cookie;
            f_ProvidableHashSets = a_ProvidableHashSets;
            f_FileSize = a_FileSize;
            f_MasterHashes = a_MasterHashes;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct DeliverHashContext
    {
        public readonly SourceCookie f_Cookie;
        public readonly byte[] f_Buffer;
        public readonly TemporaryFile f_File;

        public DeliverHashContext(SourceCookie a_Cookie,
                                  byte[] a_Buffer,
                                  TemporaryFile a_File)
        {
            f_Cookie = a_Cookie;
            f_Buffer = a_Buffer;
            f_File = a_File;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct ObtainContext
    {
        public readonly ObtainCookie f_Cookie;
        public readonly uint f_MaxBytes;
        public readonly BitMap f_CompletedByUser;
        public readonly NetworkAvailability f_VolatileAvail;

        public ObtainContext(ObtainCookie a_Cookie,
                             uint a_MaxBytes,
                             BitMap a_CompletedByUser,
                             NetworkAvailability a_VolatileAvail)
        {
            f_Cookie = a_Cookie;
            f_MaxBytes = a_MaxBytes;
            f_CompletedByUser = a_CompletedByUser;
            f_VolatileAvail = a_VolatileAvail;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct WriteContext
    {
        public readonly ObtainCookie f_Cookie;
        public ByteArray f_Data;

        public WriteContext(ObtainCookie a_Cookie,
                            ref ByteArray a_Data)
        {
            f_Cookie = a_Cookie;
            f_Data = a_Data;
        }
    }
}

namespace NB.Content
{
#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct MasterHashes
    {
        // The default constructor may be used.

        private const byte c_HashInfoCount = 3;

        private static readonly HashInfo[] s_Infos;

        private P2PHashSets f_HashSets;
        // Hashes are stored in big endian notation.
        private byte[] f_Buffer;

        public bool p_HaseD2kHash
        {
            get { return (0 != (P2PHashSets.eD2kHashSet & f_HashSets)); }
        }

        public bool p_HasAICHHash
        {
            get { return (0 != (P2PHashSets.AICHTree & f_HashSets)); }
        }

        public bool p_HasTAICHHash
        {
            get { return (0 != (P2PHashSets.TAICHTree & f_HashSets)); }
        }

        static MasterHashes()
        {
            s_Infos = new HashInfo[3];
            s_Infos[0] = new HashInfo(P2PHashSets.eD2kHashSet, MD_4.c_HashByteSize);
            s_Infos[1] = new HashInfo(P2PHashSets.AICHTree, SHA_1.c_HashByteSize);
            s_Infos[2] = new HashInfo(P2PHashSets.TAICHTree, Tiger.c_HashByteSize);
        }

        private MasterHashes(P2PHashSets a_HashSets,
                             byte[] a_Buffer)
        {
            f_HashSets = a_HashSets;
            f_Buffer = a_Buffer;
        }

        public MasterHashes(FileID a_FileID)
        {
            f_HashSets = P2PHashSets.eD2kHashSet;
            f_Buffer = a_FileID.f_Val.ToByteArrayBE();
        }

        public void Clone(out MasterHashes a_MasterHashes)
        {
            byte[] l_NewBuffer = (null == f_Buffer ? null : Util.CloneArray(f_Buffer));
            a_MasterHashes = new MasterHashes(f_HashSets, l_NewBuffer);
        }

        public static void CloneVolatile(ref MasterHashes a_VolatileSrc,
                                         out MasterHashes a_Dest)
        {
            while (true)
            {
                P2PHashSets l_HashSets = a_VolatileSrc.f_HashSets;
                byte[] l_Buffer = a_VolatileSrc.f_Buffer;
                Util.MemoryBarrier();

                if (P2PHashSets.None == l_HashSets)
                {
                    a_Dest = new MasterHashes();

                    return;
                }

                if (null != l_Buffer &&
                    CalcBufferSize(l_HashSets) == l_Buffer.Length)
                {
                    // We cannot be 100% sure that there was no concurrent
                    // write, but at least no access error will occur.

                    a_Dest = new MasterHashes(l_HashSets, Util.CloneArray(l_Buffer));

                    return;
                }
            }
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt16LE((ushort)f_HashSets);

            if (P2PHashSets.None != f_HashSets)
            {
                a_Buffer.WriteUInt16LE((ushort)f_Buffer.Length);
                a_Buffer.WriteBytes(f_Buffer);
            }
        }

        public static bool FromBuffer(bool a_UpwardCompatible,
                                      BufferIO a_Buffer,
                                      out MasterHashes a_MasterHashes)
        {
            P2PHashSets l_HashSets = (P2PHashSets)a_Buffer.ReadUInt16LE();

            if (0 == l_HashSets)
            {
                a_MasterHashes = new MasterHashes();

                return true;
            }

            if (false == a_UpwardCompatible)
            {
                if (l_HashSets != (P2PHashSets._ALL_HASHSETS_ & l_HashSets))
                {
                    // Unsupported flags.

                    a_MasterHashes = new MasterHashes();

                    return false;
                }
            }
            else
            {
                l_HashSets &= P2PHashSets._ALL_HASHSETS_;
            }

            uint l_Size = a_Buffer.ReadUInt16LE();

            if (CalcBufferSize(l_HashSets) != l_Size)
            {
                // Corruption.

                a_MasterHashes = new MasterHashes();

                return false;
            }

            a_MasterHashes = new MasterHashes(l_HashSets, a_Buffer.ReadBytes(l_Size));

            return true;
        }

        public bool ValueEquals(ref MasterHashes a_Other)
        {
            return (this.f_HashSets == a_Other.f_HashSets &&
                    true == Buffer2.ValueEquals(this.f_Buffer, a_Other.f_Buffer));
        }

        public uint CountHashes()
        {
            uint l_Count = 0;

            if (0 != (P2PHashSets.eD2kHashSet & f_HashSets)) l_Count++;
            if (0 != (P2PHashSets.AICHTree & f_HashSets)) l_Count++;
            if (0 != (P2PHashSets.TAICHTree & f_HashSets)) l_Count++;

            return l_Count;
        }

        public bool HasAllHashes(P2PHashSets a_HashSets)
        {
            return (a_HashSets == (f_HashSets & a_HashSets));
        }

        public void SeteD2kHash(byte[] a_Src,
                                int a_Idx)
        {
            SetHash(P2PHashSets.eD2kHashSet, a_Src, a_Idx);
        }

        public void SetAICHHash(byte[] a_Src,
                                int a_Idx)
        {
            SetHash(P2PHashSets.AICHTree, a_Src, a_Idx);
        }

        public void SetTAICHHash(byte[] a_Src,
                                 int a_Idx)
        {
            SetHash(P2PHashSets.TAICHTree, a_Src, a_Idx);
        }

        public void SetHash(P2PHashSets a_HashSet,
                             byte[] a_Src,
                               int a_Idx)
        {
            int l_Size;
            int l_DestIdx = GetLocation(a_HashSet, true, out l_Size);

            Buffer.BlockCopy(a_Src, a_Idx, f_Buffer, l_DestIdx, l_Size);
        }

        public bool ExposeHash(P2PHashSets a_HashSet,
                               out ByteArray a_Data)
        {
            int l_Size;
            int l_Idx = GetLocation(a_HashSet, false, out l_Size);

            if (0 <= l_Idx)
            {
                a_Data = new ByteArray(f_Buffer, l_Idx, l_Size);

                return true;
            }

            a_Data = new ByteArray();

            return false;
        }

        public bool IsHashEqual(P2PHashSets a_HashSet,
                                byte[] a_Buf,
                                int a_Idx)
        {
            int l_Size;
            int l_Idx = GetLocation(a_HashSet, false, out l_Size);

            if (0 <= l_Idx)
            {
                return Buffer2.ValueEquals(f_Buffer, l_Idx, a_Buf, a_Idx, l_Size);
            }

            return false;
        }

        public bool IsHashEqual(P2PHashSets a_HashSet,
                                ref MasterHashes a_Other)
        {
            int l_Size1;
            int l_Idx1 = this.GetLocation(a_HashSet, false, out l_Size1);

            if (0 <= l_Idx1)
            {
                int l_Size2;
                int l_Idx2 = a_Other.GetLocation(a_HashSet, false, out l_Size2);

                if (0 <= l_Idx2)
                {
                    return Buffer2.ValueEquals(this.f_Buffer, l_Idx1, a_Other.f_Buffer, l_Idx2, l_Size1);
                }
            }

            return false;
        }

        private int GetLocation(P2PHashSets a_HashSet,
                                bool a_CreateIfNotThere,
                                out int a_Size)
        {
            // If a_CreateIfNotThere is false, return can be subzero to
            // indicate that the hash is missing.

            int l_Idx = 0;

            // Iteration must be done from start to end (direction
            // matters).
            for (int i = 0; i < c_HashInfoCount; i++)
            {
                bool l_Present = (0 != (f_HashSets & s_Infos[i].f_HashSet));

                if (a_HashSet == s_Infos[i].f_HashSet)
                {
                    if (false == l_Present)
                    {
                        if (false == a_CreateIfNotThere)
                        {
                            break;
                        }

                        // Create.

                        int l_OldSize = (null == f_Buffer ? 0 : f_Buffer.Length);
                        a_Size = s_Infos[i].f_Size;

                        byte[] l_NewBuf = new byte[l_OldSize + a_Size];

                        if (null != f_Buffer)
                        {
                            Array.Copy(f_Buffer, l_NewBuf, l_OldSize);
                        }

                        l_OldSize -= l_Idx;

                        if (0 < l_OldSize)
                        {
                            // Preserve old data.

                            Buffer.BlockCopy(l_NewBuf, l_Idx, l_NewBuf, l_Idx + a_Size, l_OldSize);
                        }

                        f_HashSets |= s_Infos[i].f_HashSet;
                        Util.VolatileWrite(out f_Buffer, l_NewBuf);
                    }
                    else
                    {
                        a_Size = s_Infos[i].f_Size;
                    }

                    return l_Idx;
                }
                else
                {
                    if (true == l_Present)
                    {
                        l_Idx += s_Infos[i].f_Size;
                    }
                }
            }

            a_Size = 0;

            return (-1);
        }

        public FileID CreateFileID()
        {
            UInt128 l_UInt128 = UInt128.FromByteArrayBE(f_Buffer, 0);

            return new FileID(l_UInt128);
        }

        public string eD2kHashAsStr()
        {
            return HashAsStr(P2PHashSets.eD2kHashSet, null);
        }

        public string AICHHashAsStr()
        {
            return HashAsStr(P2PHashSets.AICHTree, null);
        }

        public string TAICHHashAsStr()
        {
            return HashAsStr(P2PHashSets.TAICHTree, null);
        }

        public string eD2kHashAsStr(string a_Replacement)
        {
            return HashAsStr(P2PHashSets.eD2kHashSet, a_Replacement);
        }

        public string AICHHashAsStr(string a_Replacement)
        {
            return HashAsStr(P2PHashSets.AICHTree, a_Replacement);
        }

        public string TAICHHashAsStr(string a_Replacement)
        {
            return HashAsStr(P2PHashSets.TAICHTree, a_Replacement);
        }

        public string HashAsStr(P2PHashSets a_HashSet,
                                string a_Replacement)
        {
            ByteArray l_Data;
            if (false == ExposeHash(a_HashSet, out l_Data))
            {
                return a_Replacement;
            }

            switch (a_HashSet)
            {
                case P2PHashSets.eD2kHashSet:
                    return HexEncoding.ToHex(l_Data.f_Arr, l_Data.f_Idx, l_Data.f_Len);

                case P2PHashSets.AICHTree:
                    return Base32.ToBase32(l_Data.f_Arr, l_Data.f_Idx, l_Data.f_Len);

                case P2PHashSets.TAICHTree:
                    return Base64.ToBase64(l_Data.f_Arr, l_Data.f_Idx, l_Data.f_Len);

                default: throw (new ApplicationException());
            }
        }

        private static uint CalcBufferSize(P2PHashSets a_HashSets)
        {
            uint l_Size = 0;

            if (0 != (a_HashSets & P2PHashSets.eD2kHashSet)) l_Size += MD_4.c_HashByteSize;
            if (0 != (a_HashSets & P2PHashSets.AICHTree)) l_Size += SHA_1.c_HashByteSize;
            if (0 != (a_HashSets & P2PHashSets.TAICHTree)) l_Size += Tiger.c_HashByteSize;

            return l_Size;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct HashInfo
        {
            public readonly P2PHashSets f_HashSet;
            public readonly byte f_Size;

            public HashInfo(P2PHashSets a_HashSet,
                            uint a_Size)
            {
                f_HashSet = a_HashSet;
                f_Size = (byte)a_Size;
            }
        }
    }

    class SourceCookie
    {
        // Used for interaction between a download peer context and
        // the content manager.

        public readonly UserNum f_UserNum;
        private P2PHashSets f_ProvidableHashSets;
        // Volatile.
        private CookieStates f_PeerState;

        private HashDemand f_HashDemand;
        private bool f_QueryInternallyAborted;
        private uint f_HashQueryNumber;
        public HashQueryParams f_HashQueryParams;
        // Volatile.
        private CookieStates f_ContentMngrState;

        public uint p_HashQueryNumber
        {
            get { return f_HashQueryNumber; }
        }

        public CookieStates p_PeerState
        {
            get { return f_PeerState; }
        }

        public CookieStates p_PeerStateVolatile
        {
            get { return Util.VolatileRead(ref f_PeerState); }
        }

        public P2PHashSets p_ProvidableHashSets
        {
            get { return f_ProvidableHashSets; }
        }

        public CookieStates p_ContentMngrState
        {
            get { return f_ContentMngrState; }
        }

        public CookieStates p_ContentMngrStateVolatile
        {
            get { return Util.VolatileRead(ref f_ContentMngrState); }
        }

        public SourceCookie(UserNum a_UserNum)
        {
            f_UserNum = a_UserNum;
        }

        public bool ShouldUnregister()
        {
            return (CookieStates.Failure != p_ContentMngrStateVolatile);
        }

        public void SetPeerAlive()
        {
            // Used by the PeerContext class only.

            Util.VolatileWrite(out f_PeerState, CookieStates.Success);
        }

        public void SetPeerGone()
        {
            // Used by the PeerContext class only.

            Util.VolatileWrite(out f_PeerState, CookieStates.Failure);
        }

        public void SetProvidableHashSets(P2PHashSets a_ProvidableHashSets)
        {
            // Used by the IncompleteFile class.

            f_ProvidableHashSets = a_ProvidableHashSets;
        }

        public void AcceptSource()
        {
            // Used by the IncompleteFile class.

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Success);
        }

        public void UnacceptSource()
        {
            // Used by the IncompleteFile class.

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Failure);
        }

        public void RejectSource()
        {
            // Used by the content manager.

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Failure);
        }

        public void RequestHashes(ref HashDemand a_HashDemand,
                                  ref HashQueryParams a_HashQueryParams)
        {
            // Used by the IncompleteFile class.

            f_HashDemand = a_HashDemand;
            f_QueryInternallyAborted = false;
            f_HashQueryNumber++;
            f_HashQueryParams = a_HashQueryParams;

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.WaitingForHashes);
        }

        public bool IsRetrievingHashes()
        {
            // Used by the IncompleteFile class.

            return (CookieStates.WaitingForHashes == f_ContentMngrState &&
                    false == f_QueryInternallyAborted);
        }

        public bool IsRetrievingHashes(ref HashDemand a_HashDemand)
        {
            // Used by the IncompleteFile class.

            return (true == IsRetrievingHashes() &&
                    true == f_HashDemand.ValueEquals(ref a_HashDemand));
        }

        public bool IsRetrievingHashes(P2PHashSets a_HashSets)
        {
            // Used by the IncompleteFile class. a_HashSets can be a
            // combination of hash types.

            return (true == IsRetrievingHashes() &&
                    0 != (f_HashQueryParams.f_HashSet & a_HashSets));
        }

        public void GetHashDemand(out HashDemand a_HashDemand)
        {
            // Used by the IncompleteFile class.

            a_HashDemand = f_HashDemand;
        }

        public void SetInternallyAborted()
        {
            // Used by the IncompleteFile class.

            f_QueryInternallyAborted = true;
        }

        public bool WasInternallyAborted()
        {
            // Used by the IncompleteFile class.

            return f_QueryInternallyAborted;
        }

        public void ResetAfterHashDelivery()
        {
            // Used by the IncompleteFile class.

            // Release some memory.
            f_HashQueryParams = new HashQueryParams();

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Success);
        }

        public bool GuiWaitForCompletion()
        {
            return APCCookie.GuiWaitForCompletion(ref f_ContentMngrState);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct HashQueryParams
        {
            public readonly P2PHashSets f_HashSet;
            // f_ExpectedMasterHash has big endian notation.
            public readonly byte[] f_ExpectedMasterHash;
            public readonly uint[] f_HashIndices;
            public readonly uint f_HashOffset;
            public readonly uint f_HashCount;
            public readonly uint f_ExpectedSize;

            public HashQueryParams(P2PHashSets a_HashSet,
                                   byte[] a_ExpectedMasterHash,
                                   uint[] a_HashIndices,
                                   uint a_ExpectedSize)
            {
                f_HashSet = a_HashSet;
                f_ExpectedMasterHash = a_ExpectedMasterHash;
                f_HashIndices = a_HashIndices;
                f_HashOffset = 0;
                f_HashCount = 0;
                f_ExpectedSize = a_ExpectedSize;
            }

            public HashQueryParams(P2PHashSets a_HashSet,
                                   byte[] a_ExpectedMasterHash,
                                   uint a_HashOffset,
                                   uint a_HashCount,
                                   uint a_ExpectedSize)
            {
                f_HashSet = a_HashSet;
                f_ExpectedMasterHash = a_ExpectedMasterHash;
                f_HashIndices = null;
                f_HashOffset = a_HashOffset;
                f_HashCount = a_HashCount;
                f_ExpectedSize = a_ExpectedSize;
            }
        }
    }

    class ObtainCookie : ListBaseClass<ObtainCookie>
    {
        // Used for interaction between a download peer context and
        // the content manager (primarily to download file content).

        public readonly SourceCookie f_SourceCookie;
        public readonly int f_MaxPiecesPerGroup;

        // Volatile.
        private CookieStates f_PeerState;

        private IncompleteFile f_IncomplFile;
        private uint f_ChunkIndex;
        public PieceBounds f_Bounds;
        // f_Size indicates the total size defined in f_Bounds.
        private uint f_Size;
        public uint f_BytesToGo;
        public FlexArray<ChunkRange> f_Ranges;
        public int f_Index;
        public byte[] f_Buffer;
        private RejectionReasons f_RejectionReason;
        // Volatile.
        private CookieStates f_ContentMngrState;

        public CookieStates p_PeerState
        {
            get { return f_PeerState; }
        }

        public uint p_ChunkIndex
        {
            get { return f_ChunkIndex; }
        }

        public uint p_Size
        {
            get { return f_Size; }
        }

        public RejectionReasons p_RejectionReason
        {
            get { return f_RejectionReason; }
        }

        public CookieStates p_ContentMngrState
        {
            get { return f_ContentMngrState; }
        }

        public CookieStates p_ContentMngrStateVolatile
        {
            get { return Util.VolatileRead(ref f_ContentMngrState); }
        }

        public ulong p_ChunkBaseOffset
        {
            get { return ((ulong)f_ChunkIndex * IncompleteFile.c_ChunkSize); }
        }

        public UserNum p_UserNum
        {
            get { return f_SourceCookie.f_UserNum; }
        }

        public ObtainCookie(SourceCookie a_SourceCookie,
                            int a_MaxPiecesPerGroup)
        {
            f_SourceCookie = a_SourceCookie;
            f_MaxPiecesPerGroup = a_MaxPiecesPerGroup;
        }

        public void SetPeerState(CookieStates a_State)
        {
            // Used by the IncompleteFile class.

            Util.VolatileWrite(out f_PeerState, a_State);
        }

        public void Rejected(RejectionReasons a_RejectionReason)
        {
            // Used by the content manager.

            f_Buffer = null;

            f_RejectionReason = a_RejectionReason;
            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Failure);
        }

        public void Finished()
        {
            // Used by the IncompleteFile class.

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.Success);
        }

        public void SetBounds(IncompleteFile a_IncomplFile,
                              uint a_ChunkIndex,
                              ref PieceBounds a_Bounds)
        {
            // Used by the IncompleteFile class.

            f_IncomplFile = a_IncomplFile;
            f_ChunkIndex = a_ChunkIndex;
            f_Bounds = a_Bounds;
            f_Size = f_Bounds.GetCoveredBytes();
            f_BytesToGo = f_Size;

            PieceBounds.Enumerator l_Enum;
            f_Bounds.BuildEnumerator(f_MaxPiecesPerGroup, out l_Enum);

            FlexArray<ChunkRange>.Create(f_Bounds.EstimateMaxPieces(), out f_Ranges);

            ChunkRange l_Range;
            while (true == l_Enum.GetNextRange(out l_Range))
            {
                int l_Idx = f_Ranges.ReserveAtEnd();
                f_Ranges.p_Arr[l_Idx] = l_Range;
            }

            f_Ranges.TrimExcess();

            Util.VolatileWrite(out f_ContentMngrState, CookieStates.RangeSet);
        }

        public bool BelongsToIncomplFile(IncompleteFile a_IncomplFile)
        {
            return (f_IncomplFile == a_IncomplFile);
        }

        public enum RejectionReasons : byte
        {
            NotRejected = 0,

            CriticalDiskSpaceLevel = 1,
            IncomplFileNotLoaded = 2,
            IncomplFileUnloads = 3,
            NeedHashesFirst = 4,
            NoRelevantRange = 5,
            BadSource = 6,
            OtherReason = 7,
        }
    }

    class ExtractionCookie : APCCookie
    {
        private const uint c_PreviewMinAvailDisk = (64 * 1024 * 1024);

        public readonly ulong f_FilePos;
        public readonly ulong f_Size;
        public readonly TemporaryFile f_TempFile;
        public readonly bool f_FlushOnSuccess;

        public byte[] f_StartBuffer;
        public int f_StartBufferIndex;

        private Result f_Result;

        public Result p_Result
        {
            get { return f_Result; }
        }

        public ExtractionCookie(ulong a_FilePos,
                                uint a_Size,
                                TemporaryFile a_TempFile,
                                bool a_FlushOnSuccess)
        {
            f_FilePos = a_FilePos;
            f_Size = a_Size;
            f_TempFile = a_TempFile;
            f_FlushOnSuccess = a_FlushOnSuccess;
        }

        public void Fail(Result a_Result)
        {
            f_Result = a_Result;

            SetState(CookieStates.Failure);
        }

        public void Succeed(Result a_Result)
        {
            f_Result = a_Result;

            SetState(CookieStates.Success);
        }

        public static Result CreateForPreviewing(Instance a_Inst,
                                                 string a_FileName,
                                                 uint a_Size,
                                                 out ExtractionCookie a_Cookie)
        {
            string l_Path = a_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_PreviewFolder;

            if (false == Directory.Exists(l_Path))
            {
                a_Cookie = null;

                return Result.PreviewFolderDoesntExist;
            }

            string l_Ext = LongPath.GetExtension(a_FileName);

            // Do not preview files that are executable.

            if (null == l_Ext)
            {
                // No extension at all.

                a_Cookie = null;

                return Result.UnsuitableExtension;
            }
            else
            {
                FileType l_FileType = FileTypeDetector.GuessFileTypeFromExtension(l_Ext);

                if (FileType.FT_AUDIO != l_FileType &&
                    FileType.FT_VIDEO != l_FileType &&
                    FileType.FT_IMAGE != l_FileType &&
                    FileType.FT_DOCUMENT != l_FileType)
                {
                    a_Cookie = null;

                    return Result.UnsuitableExtension;
                }
            }

            string l_FileName = Path.GetFileNameWithoutExtension(a_FileName);

            if (64 < l_FileName.Length)
            {
                // Truncate this long file name to avoid troubling the
                // OS.

                l_FileName = l_FileName.Substring(0, 64);
            }

            string l_SuggestedFileName = a_Inst.f_TextSet.Format(NB.Language.TextKeys.MsgPreviewFileName, l_FileName, l_Ext);
            string l_FullFileName = LongPath.Combine(l_Path, l_SuggestedFileName);
            TemporaryFile l_TempFile;
            Result l_Result = CreateTempFile(true, l_FullFileName, a_Size, out l_TempFile);

            if (Result.Success == l_Result)
            {
                a_Cookie = new ExtractionCookie(0, a_Size, l_TempFile, true);

                return Result.Success;
            }

            a_Cookie = null;

            return l_Result;
        }

        public static Result CreateTempFileForCompletion(string a_FullFileName,
                                                         ulong a_Size,
                                                         out TemporaryFile a_File)
        {
            return CreateTempFile(false, a_FullFileName, a_Size, out a_File);
        }

        private static Result CreateTempFile(bool a_ForPreviewing,
                                             string a_FullFileName,
                                             ulong a_Size,
                                             out TemporaryFile a_File)
        {
            ulong l_AvailBytes;
            if (true == FHandle.GetAvailableVolumeSpace(a_FullFileName, out l_AvailBytes))
            {
                if (a_Size > l_AvailBytes)
                {
                    a_File = null;

                    return Result.NotEnoughDiskSpace;
                }

                if ((false == a_ForPreviewing ? ContentManager.c_MinAvailDiskSpace : c_PreviewMinAvailDisk) > (l_AvailBytes - a_Size))
                {
                    // The size would suffice, but we prefer not to use it
                    // up since it's quite small.

                    a_File = null;

                    return Result.LowDiskSpace;
                }
            }

            a_File = TemporaryFile.Create(a_FullFileName,
                                          false == a_ForPreviewing ? FileMode.CreateNew : FileMode.OpenOrCreate,
                                          TemporaryFile.DeletionKind.DeleteOnFinalize);

            if (null == a_File)
            {
                return Result.FailedToCreateFile;
            }

            if (false == a_ForPreviewing)
            {
                // We're extracting, so allocate the whole file at once.

                FHandle l_Handle = a_File.ExposeFHandle();

                if (false == l_Handle.SetSize(a_Size))
                {
                    a_File.Close();

                    return Result.NotEnoughDiskSpace;
                }

                byte[] l_WarnMsg = BuildWarnMsgBytes(a_Size);
                uint l_MsgLen = (uint)l_WarnMsg.Length;
                ByteArray l_Data = new ByteArray(l_WarnMsg, 0, (int)Math.Min(l_MsgLen, a_Size));
                bool l_WriteSucc = l_Handle.Write(0, ref l_Data);

                if (true == l_WriteSucc &&
                    (2 * l_MsgLen) <= a_Size &&
                    (false == l_Handle.Write(a_Size - l_MsgLen, ref l_Data) ||
                     false == l_Handle.Flush()))
                {
                    l_WriteSucc = false;
                }

                if (false == l_WriteSucc)
                {
                    a_File.Close();

                    return Result.IOError;
                }
            }

            return Result.Success;
        }

        private static byte[] BuildWarnMsgBytes(ulong a_FileSize)
        {
            string l_WarnMsg = string.Format(Util.s_InvariantCulture, "THIS FILE WAS NOT COMPLETELY FILLED WITH MEANINGFUL DATA YET AND IS PROBABLY NOT USEFUL IN ITS CURRENT FORM. EXPECTED FILE SIZE IS {0:N0} BYTES.", a_FileSize);

            return Encoding.ASCII.GetBytes(l_WarnMsg);
        }

        public enum Result : byte
        {
            None = 0,

            Success = 1,
            UnsuitableExtension = 2,
            FileNotActive = 3,
            IOError = 4,
            NotEnoughDiskSpace = 5,
            LowDiskSpace = 6,
            FailedToCreateFile = 7,
            InputCorrupt = 8,
            NotAvail = 9,
            DestFileAccessErr = 10,
            PreviewFolderDoesntExist = 11,
        }
    }

    class ImportCookie : APCCookie
    {
        private const byte c_DefaultMaxChunksPerCall = 2;

        public readonly ImportableFile f_ImportFile;
        public readonly FHandle f_SrcFile;
        public readonly uint f_MaxChunksPerCall;

        private uint f_ChunkIndex;
        private int f_GapIndexHint;

        private Result f_Result;
        private uint f_ElapsedByteCount;
        private uint f_TranslatedByteCount;

        public Result p_Result
        {
            get { return f_Result; }
        }

        public ImportCookie(ImportableFile a_ImportFile,
                            FHandle a_SrcFile)
        {
            f_ImportFile = a_ImportFile;
            f_SrcFile = a_SrcFile;
            f_MaxChunksPerCall = c_DefaultMaxChunksPerCall;
        }

        public void GetPosition(out uint a_ChunkIndex,
                                out int a_GapIndexHint)
        {
            a_ChunkIndex = f_ChunkIndex;
            a_GapIndexHint = f_GapIndexHint;
        }

        public void SetPosition(uint a_ChunkIndex,
                                int a_GapIndexHint)
        {
            f_ChunkIndex = a_ChunkIndex;
            f_GapIndexHint = a_GapIndexHint;
        }

        public void NotifyElapsed(uint a_ElapsedByteCount)
        {
            // Used by the IncompleteFile class.

            f_ElapsedByteCount += a_ElapsedByteCount;
        }

        public void NotifyTranslation(uint a_TranslatedByteCount)
        {
            // Used by the IncompleteFile class.

            f_TranslatedByteCount += a_TranslatedByteCount;
        }

        public void QueryCounts(ref ulong a_ElapsedSize,
                                ref ulong a_TranslatedSize)
        {
            a_ElapsedSize += f_ElapsedByteCount;
            a_TranslatedSize += f_TranslatedByteCount;
        }

        public void Recycle()
        {
            // Preserves f_ChunkIndex and f_GapIndexHint.

            f_Result = Result.None;
            f_ElapsedByteCount = 0;
            f_TranslatedByteCount = 0;

            SetState(CookieStates.Pending);
        }

        public void Fail(Result a_Result)
        {
            // Used by the IncompleteFile class.

            f_Result = a_Result;

            SetState(CookieStates.Failure);
        }

        public void Succeed(Result a_Result)
        {
            // Used by the IncompleteFile class.

            f_Result = a_Result;

            SetState(CookieStates.Success);
        }

        public enum Result : byte
        {
            None = 0,

            SuccessFinished = 1,
            SuccessCallAgain = 2,

            FileNotActive = 3,
            FailedToAccessSrcFile = 4,
            FailedToUpdateChunk = 5,
        }
    }
}

// End of file.
