/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\GUIPrimitives.cs
 * PURPOSE: Basic types of the graphical user interface
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using NB.I2P;
using NB.Language;
using NB.Network;
using NB.Network.eD2k;
using Support;

namespace NB.Gui
{
    sealed class GuiFilterTextBox : LabeledControlBase<TextBox>
    {
        private static int s_ButtonWidth = (Lyt.s_Space + Lyt.s_SmallSpace2);

        private readonly GuiButton f_Button;

        // f_FullText can be null.
        private string f_FullText;

        public event EventHandler e_ButtonClick;

        public string p_FilterText
        {
            get { return f_FullText; }
        }

        public GuiFilterTextBox(Control a_Owner,
                                int a_Width) :
                         base(a_Owner, new TextBox(), Math.Max(0, a_Width - s_ButtonWidth - Lyt.s_SmallSpace), int.MaxValue)
        {
            const string l_ButtonText = "\x00B1";

            f_Button = new GuiButton(a_Owner);

            f_Box.ReadOnly = true;
            f_Box.MaxLength = int.MaxValue;
            f_Box.TabStop = false;
            f_Button.Size = new Size(s_ButtonWidth, f_Box.Height);

            f_Button.Click += Button_Click;

            f_Button.Text = l_ButtonText;
        }

        public override void Dispose()
        {
            f_Button.Dispose();
        }

        public void FocusButton()
        {
            f_Button.Focus();
        }

        public override void SetLocation(int a_x,
                                         int a_y)
        {
            base.SetLocation(a_x, a_y);

            UpdateLocation();
        }

        public override void SetWidth(int a_Width)
        {
            base.SetWidth(a_Width);

            UpdateLocation();
            UpdateDisplayText();
        }

        private void UpdateLocation()
        {
            // This routine should get called after the base class
            // was updated.

            f_Button.Location = new Point(f_Box.Right + Lyt.s_SmallSpace, f_Box.Top);
        }

        public void SetFilterText(string a_Text)
        {
            // a_Text may be null.

            bool l_Change = (f_FullText != a_Text);

            f_FullText = a_Text;

            if (true == l_Change)
            {
                UpdateDisplayText();
            }
        }

        private void UpdateDisplayText()
        {
            if (null == f_FullText ||
                0 >= f_FullText.Length)
            {
                f_Box.Clear();

                return;
            }

            const string l_Ellipsis = "...";
            int l_MaxPixelWidth = (f_Box.Width - Lyt.s_TinySpace2);
            Font l_Font = f_Box.Font;
            string l_EffText = l_Ellipsis;

            Bitmap l_Map = new Bitmap(1, 1);
            Graphics l_G = Graphics.FromImage(l_Map);
            {
                if ((int)l_G.MeasureString(f_FullText, l_Font).Width <= l_MaxPixelWidth)
                {
                    l_EffText = f_FullText;

                    goto EXIT;
                }

                int l_Idx = 0;

                while (true)
                {
                    int l_NextIdx = f_FullText.IndexOf(Util.c_Space, l_Idx);

                    if (0 > l_NextIdx)
                    {
                        goto EXIT;
                    }

                    l_NextIdx++;

                    string l_NewText = (f_FullText.Substring(0, l_NextIdx) + l_Ellipsis);

                    if ((int)l_G.MeasureString(l_NewText, l_Font).Width > l_MaxPixelWidth)
                    {
                        goto EXIT;
                    }

                    l_Idx = l_NextIdx;
                    l_EffText = l_NewText;
                }

                EXIT:;
            }
            l_G.Dispose();
            l_Map.Dispose();

            f_Box.Text = l_EffText;
        }

        private void Button_Click(object a_Sender,
                                  EventArgs e)
        {
            // Call the custom handler if the owner registered one.
            // Use this 'control' as sender.

            if (null != e_ButtonClick)
            {
                e_ButtonClick(this, e);
            }
        }
    }

    sealed class TextItemsDialog : ButtonDialog
    {
        private const uint c_NoItemID = 0;

        public delegate bool AcceptTextHandler(TextItemsDialog a_Sender,
                                               ref string a_ProposedText,
                                               out string a_ErrMsg);

        private readonly Instance f_Inst;
        private readonly uint f_MaxItemCount;
        // Limit is in UTF-16 code units.
        private readonly uint f_MaxCharsPerItem;
        private readonly uint f_MaxTotalChars;
        private readonly char f_SeparatorChar;
        private readonly AcceptTextHandler f_AcceptText;
        private readonly Comparison<string> f_StringComparer;

        private readonly GuiGroupBox f_GroupBox;
        private readonly GuiLabel f_TopLabel;
        private readonly LabeledTextBox f_TextBox;
        private readonly GridView<uint, string> f_GridView;
        private readonly GuiButton f_AddButton;
        private readonly GuiButton f_ModifySelButton;
        private readonly GuiButton f_RemoveSelButton;
        private readonly GuiButton f_RemoveAllButton;
        private readonly GuiButton f_AddFromClipboardButton;
        private readonly GuiButton f_CopySelToClipboardButton;

        private uint f_TotalRawChars;
        private Dictionary<uint, DummyStruct> f_DeleteItemDic;

        private bool f_SelHandlerMustNotModifyTextbox;

        private string f_FinalText;

        private TextItemsDialog(Instance a_Inst,
                                uint a_MaxItemCount,
                                uint a_MaxCharsPerItem,
                                uint a_MaxTotalChars,
                                char a_SeparatorChar,
                                AcceptTextHandler a_AcceptText) :
                           base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            f_Inst = a_Inst;
            f_MaxItemCount = a_MaxItemCount;
            f_MaxCharsPerItem = a_MaxCharsPerItem;
            f_MaxTotalChars = a_MaxTotalChars;
            f_SeparatorChar = a_SeparatorChar;
            f_AcceptText = a_AcceptText;
            f_StringComparer = Util.CompareStringsOrdinal;

            const int lc_NativeDlgWidth = 550;
            const int lc_NativeMinGridViewHeight = 450;

            int l_DlgWidth = Lyt.CalcX(lc_NativeDlgWidth);
            int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
            f_GroupBox = new GuiGroupBox(this);
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            int l_RightControlsWidth = Lyt.s_GenerousButtonWidth;
            int l_LeftControlsWidth = (l_InnerWidth - Lyt.s_Space - l_RightControlsWidth);
            f_TopLabel = new GuiLabel(f_GroupBox);
            f_TextBox = new LabeledTextBox(f_GroupBox, l_LeftControlsWidth);
            f_GridView = new GridView<uint, string>(a_Inst, f_GroupBox, a_Inst.f_MainForm.f_VisualCust.f_TextItemsDlgColSet);
            f_AddButton = new GuiButton(f_GroupBox, true);
            f_ModifySelButton = new GuiButton(f_GroupBox, true);
            f_RemoveSelButton = new GuiButton(f_GroupBox, true);
            f_RemoveAllButton = new GuiButton(f_GroupBox, true);
            f_AddFromClipboardButton = new GuiButton(f_GroupBox, true);
            f_CopySelToClipboardButton = new GuiButton(f_GroupBox, true);

            int l_LeftPadding = f_GroupBox.Padding.Left;
            int l_RightButtonsX = (l_LeftPadding + l_LeftControlsWidth + Lyt.s_Space);
            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_TopLabel.Location = new Point(l_LeftPadding, f_GroupBox.Padding.Top);
            f_TopLabel.Text = "Use this dialog to specify relevant keyword patterns for the filter. Wildcard characters like ? and * may be used. Note that these keyword patterns are not matched against an entire file name, but against each individual keyword that a file name is comprised of.\n\nThe filter is case-insensitive and all input is converted into its lowercase representation.";
            f_TopLabel.Width = l_InnerWidth;
            Lyt.SetFontAndHeight(f_TopLabel, f_TopLabel.Font);
            f_TextBox.SetLocation(l_LeftPadding, f_TopLabel.GetBottom() + Lyt.s_SmallSpace);
            f_TextBox.SetMaxChars(f_MaxCharsPerItem);
            f_GridView.Dock = DockStyle.None;
            f_GridView.Location = new Point(l_LeftPadding, f_TextBox.GetBottom());
            f_GridView.Width = (l_RightButtonsX - Lyt.s_Space - l_LeftPadding);
            f_GridView.LoadColumnAdjustments();
            f_AddButton.Location = new Point(l_RightButtonsX, f_TextBox.f_Box.Top);
            f_ModifySelButton.Location = new Point(l_RightButtonsX, f_AddButton.GetBottom());
            f_RemoveSelButton.Location = new Point(l_RightButtonsX, f_ModifySelButton.GetBottom());
            f_RemoveAllButton.Location = new Point(l_RightButtonsX, f_RemoveSelButton.GetBottom());
            int l_BottomButtonsHeight = ((2 * f_AddFromClipboardButton.Height) + Lyt.s_SmallSpace);
            f_GridView.Height = Math.Max((f_RemoveAllButton.Bottom + Lyt.s_LargeSpace + l_BottomButtonsHeight) - f_GridView.Top, lc_NativeMinGridViewHeight);
            f_AddFromClipboardButton.Location = new Point(l_RightButtonsX, f_GridView.Bottom - l_BottomButtonsHeight);
            f_CopySelToClipboardButton.Location = new Point(l_RightButtonsX, f_AddFromClipboardButton.GetBottom());
            f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_GridView.Bottom);

            base.AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
            base.SetSelectOnShowControl(f_TextBox.f_Box);
            this.AcceptButton = f_AddButton;
            base.CancelButton = null;

            f_AddButton.Click += AddButton_Click;
            f_ModifySelButton.Click += ModifyButton_Click;
            f_RemoveSelButton.Click += RemoveSelButton_Click;
            f_RemoveAllButton.Click += RemoveAllButton_Click;
            f_AddFromClipboardButton.Click += AddFromClipboardButton_Click;
            f_CopySelToClipboardButton.Click += CopySelToClipboardButton_Click;
            f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
            f_GridView.e_UpdateItem += GridView_UpdateItem;
            f_GridView.SelectionChanged += GridView_SelectionChanged;
            f_OKButton.Click += OKButton_Click;

            this.Text = "Modify Text Set";
            f_GroupBox.Text = "Keywords";
            f_TextBox.SetCaption("Keyword pattern");
            f_AddButton.Text = "Add";
            f_ModifySelButton.Text = "Modify Selection";
            f_RemoveSelButton.Text = "Remove Selection";
            f_RemoveAllButton.Text = "Remove All";
            f_AddFromClipboardButton.Text = "Add from Clipboard";
            f_CopySelToClipboardButton.Text = "Copy Selection to Clipboard";
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                f_GridView.StoreColumnAdjustments();

                MainForm.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        private string TryGetText(string a_SubstitutedText,
                                  bool a_NeedSeparator,
                                  uint a_IgnoreRuleIDDuringDuplicateCheck)
        {
            // a_SubstitutedText may be null. a_IgnoreRuleIDDuringDuplicateCheck may
            // be c_NoItemID. Return can be null if the current text
            // is invalid. In this case a message box will appear.

            string l_ErrMsg;
            string l_Text = f_TextBox.GetBoxTextTrimmedAndNormalized();

            if (0 >= l_Text.Length)
            {
                l_ErrMsg = "The specified text must not be an empty string.";

                goto FAIL;
            }

            if (false == f_AcceptText(this, ref l_Text, out l_ErrMsg))
            {
                goto FAIL;
            }

            uint l_TextLen = Util.GetStringLength(l_Text);

            if (null == a_SubstitutedText ||
                false == Util.StringsEqual(l_Text, a_SubstitutedText))
            {
                foreach (GridViewItem<uint, string> l_Item in f_GridView.Rows)
                {
                    if (c_NoItemID != a_IgnoreRuleIDDuringDuplicateCheck &&
                        l_Item.f_Ref == a_IgnoreRuleIDDuringDuplicateCheck)
                    {
                        continue;
                    }

                    if (true == Util.StringsEqual(l_Item.f_Tag, l_Text))
                    {
                        l_ErrMsg = "A text equivalent to the specified text is already in the list.";

                        goto FAIL;
                    }
                }

                uint l_SubstitutedLength = Util.GetStringLengthNullSafe(a_SubstitutedText);
                ulong l_NewTotalLen = ((ulong)(f_TotalRawChars - l_SubstitutedLength) + (Util.BoolToByte(a_NeedSeparator) + l_TextLen));

                if (f_MaxTotalChars < l_NewTotalLen)
                {
                    l_ErrMsg = "The specified text cannot be used, as the maximum number of characters has been reached for this text set.";

                    goto FAIL;
                }

                f_TotalRawChars = (uint)l_NewTotalLen;
            }

            return l_Text;

            FAIL:

            f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);

            return null;
        }

        private void AddButton_Click(object a_Sender,
                                     EventArgs e)
        {
            uint l_Count = (uint)f_GridView.p_ItemCount;

            if (l_Count >= f_MaxItemCount)
            {
                f_Inst.f_MainForm.ErrMsgBox("The maximum number of items is reached.");

                return;
            }

            string l_Text = TryGetText(null, 0 != l_Count, c_NoItemID);

            if (null == l_Text)
            {
                return;
            }

            f_SelHandlerMustNotModifyTextbox = true;
            {
                GridViewItem<uint, string> l_Item = f_GridView.CreateSingleItem(FindUnoccupiedItemID(), l_Text);
                f_GridView.SelectAndDisplayItem(l_Item);
            }
            f_SelHandlerMustNotModifyTextbox = false;
        }

        private void ModifyButton_Click(object a_Sender,
                                        EventArgs e)
        {
            if (1 != f_GridView.p_SelectedCount)
            {
                return;
            }

            GridViewItem<uint, string> l_Item = f_GridView.GetSelectedItem(0);
            uint l_ItemID = l_Item.f_Ref;
            string l_OldText = l_Item.f_Tag;
            string l_Text = TryGetText(l_OldText, false, l_ItemID);

            if (null == l_Text ||
                true == Util.StringsEqual(l_Text, l_OldText))
            {
                return;
            }

            l_Item.f_Tag = l_Text;

            f_SelHandlerMustNotModifyTextbox = true;
            {
                int l_RowIdx = l_Item.Index;
                f_GridView.NotifyValuesChanged(false);
                f_GridView.UpdateCellValue(0, l_RowIdx);

                f_GridView.SelectAndDisplayItem(l_Item);
            }
            f_SelHandlerMustNotModifyTextbox = false;

            f_TextBox.SetFocus();
        }

        private void RemoveSelButton_Click(object a_Sender,
                                           EventArgs e)
        {
            uint l_SelItem = (uint)f_GridView.p_SelectedCount;

            if (0 == l_SelItem)
            {
                return;
            }

            if ((uint)f_GridView.p_ItemCount == l_SelItem)
            {
                RemoveAllItemsCore(true);

                return;
            }

            f_TotalRawChars -= l_SelItem;
            f_DeleteItemDic = new Dictionary<uint, DummyStruct>((int)l_SelItem);

            for (uint i = l_SelItem; 0 != i--; )
            {
                GridViewItem<uint, string> l_Item = f_GridView.GetSelectedItem((int)i);
                f_TotalRawChars -= Util.GetStringLength(l_Item.f_Tag);
                f_DeleteItemDic.Add(l_Item.f_Ref, new DummyStruct());
            }

            UpdateItems();
        }

        private void RemoveAllButton_Click(object a_Sender,
                                           EventArgs e)
        {
            if (0 >= f_GridView.p_ItemCount)
            {
                return;
            }

            if (true == f_Inst.f_MainForm.YesNoMsgBox("Really remove all items?"))
            {
                RemoveAllItemsCore(true);
            }
        }

        private void RemoveAllItemsCore(bool a_AlsoUpdate)
        {
            f_GridView.ClearItems();
            f_TotalRawChars = 0;

            if (true == a_AlsoUpdate)
            {
                UpdateItems();
            }
        }

        private void UpdateButtons()
        {
            uint l_ItemCount = (uint)f_GridView.p_ItemCount;
            uint l_SelCount = (uint)f_GridView.p_SelectedCount;

            f_ModifySelButton.Enabled = (1 == l_SelCount);
            f_RemoveSelButton.Enabled = (0 != l_SelCount);
            f_RemoveAllButton.Enabled = (0 != l_ItemCount);
            f_CopySelToClipboardButton.Enabled = (0 != l_SelCount);
        }

        private uint FindUnoccupiedItemID()
        {
            while (true)
            {
                uint l_ItemID = Util.GetRandomUIntFast();

                if (c_NoItemID != l_ItemID &&
                    null == f_GridView.GetItemByKey(l_ItemID))
                {
                    return l_ItemID;
                }
            }
        }

        private void UpdateItems()
        {
            uint[] l_Arr = f_GridView.CreateItemKeyArr();

            if (null == l_Arr)
            {
                f_GridView.ClearItems();

                return;
            }

            Dictionary<uint, DummyStruct> l_Dic = f_DeleteItemDic;
            f_DeleteItemDic = null;

            f_GridView.StartChanging();

            for (uint i = Util.GetArrayLength(l_Arr); 0 != i--; )
            {
                uint l_ItemID = l_Arr[i];

                if (null == l_Dic ||
                    false == l_Dic.ContainsKey(l_ItemID))
                {
                    GridViewItem<uint, string> l_Item = f_GridView.CreateItem(l_ItemID);
                }
            }

            f_GridView.EndChanging();
        }

        private void GridView_SelectionChanged(object a_Sender,
                                               EventArgs e)
        {
            if (false == f_SelHandlerMustNotModifyTextbox)
            {
                uint l_SelCount = (uint)f_GridView.p_SelectedCount;

                if (1 == l_SelCount)
                {
                    string l_Text = f_GridView.GetSelectedItem(0).f_Tag;
                    f_TextBox.f_Box.Text = l_Text;
                }
                else
                {
                    f_TextBox.f_Box.Clear();
                }
            }

            UpdateButtons();
        }

        private void GridView_ItemDoubleClicked(GridViewItem<uint, string> a_Item)
        {
            f_TextBox.SetFocus();
        }

        private void GridView_UpdateItem(Time a_Time,
                                         GridViewItem<uint, string> a_Item)
        {
            string l_ItemText = a_Item.f_Tag;
            a_Item.SetTextWithCompObj(0, l_ItemText, l_ItemText);
        }

        private void AddFromClipboardButton_Click(object a_Sender,
                                                  EventArgs e)
        {
            string l_Input;

            try
            {
                l_Input = Clipboard.GetText(TextDataFormat.UnicodeText);
            }
            catch (System.Runtime.InteropServices.ExternalException)
            {
                l_Input = string.Empty;
            }

            if (0 >= l_Input.Length)
            {
                f_Inst.f_MainForm.ErrMsgBox("The clipboard seems not to contain usable text.");

                return;
            }

            l_Input = KademliaFilter.NormalizeAndLowercaseString(l_Input);

            uint l_StringsProcessed = 0;
            FlexArray<string> l_NewStrsLs;
            FlexArray<string>.Create(out l_NewStrsLs);

            using (System.IO.StringReader l_Rdr = new System.IO.StringReader(l_Input))
            {
                bool l_NeedSeparator = (0 != f_GridView.p_ItemCount);

                for (bool l_Recheck = true; ; )
                {
                    if (true == l_Recheck)
                    {
                        l_Recheck = false;

                        if (f_MaxItemCount <= (uint)f_GridView.p_ItemCount ||
                            f_MaxTotalChars <= f_TotalRawChars)
                        {
                            break;
                        }
                    }

                    string l_Str = l_Rdr.ReadLine();

                    if (null == l_Str)
                    {
                        break;
                    }

                    l_StringsProcessed++;

                    if (true == TryAddStringFromClipboard(l_Str, l_NeedSeparator, ref l_NewStrsLs))
                    {
                        l_Recheck = true;
                        l_NeedSeparator = true;
                    }
                }
            }

            l_Input = null;

            uint l_NewItemCount = (uint)l_NewStrsLs.p_Count;

            if (0 != l_NewItemCount)
            {
                uint[] l_ExistingKeyArr = f_GridView.CreateItemKeyArr();

                f_GridView.StartChanging();
                {
                    if (null != l_ExistingKeyArr)
                    {
                        foreach (uint l_Key in l_ExistingKeyArr)
                        {
                            f_GridView.CreateItem(l_Key);
                        }
                    }

                    string[] l_NewStrsArr = l_NewStrsLs.p_Arr;

                    for (uint i = 0; i != l_NewItemCount; i++)
                    {
                        f_GridView.CreateItem(FindUnoccupiedItemID()).f_Tag = l_NewStrsArr[i];
                    }
                }
                f_GridView.EndChanging();
            }

            f_Inst.f_MainForm.InfoMsgBox(f_Inst.f_TextSet.Format("The clipboard content was processed. In total {0} text items were added.", l_NewItemCount));

            UpdateItems();
        }

        private bool TryAddStringFromClipboard(string a_Str,
                                               bool a_NeedSeparator,
                                               ref FlexArray<string> a_NewStrsLs)
        {
            // A helper for AddFromClipboardButton_Click().

            string l_Text = a_Str;
            string l_Junk;

            if (0 >= l_Text.Length ||
                false == f_AcceptText(this, ref l_Text, out l_Junk))
            {
                return false;
            }

            uint l_TextLen = Util.GetStringLength(l_Text);

            ulong l_NewTotalLen = ((ulong)f_TotalRawChars + (Util.BoolToByte(a_NeedSeparator) + l_TextLen));

            if (f_MaxTotalChars <= l_NewTotalLen)
            {
                return false;
            }

            foreach (GridViewItem<uint, string> l_Item in f_GridView.Rows)
            {
                if (true == Util.StringsEqual(l_Item.f_Tag, l_Text))
                {
                    return false;
                }
            }

            string[] l_NewStrArr = a_NewStrsLs.p_Arr;

            for (uint i = (uint)a_NewStrsLs.p_Count; 0 != i--; )
            {
                if (true == Util.StringsEqual(l_NewStrArr[i], l_Text))
                {
                    return false;
                }
            }

            int l_ArrIdx = a_NewStrsLs.ReserveAtEnd();
            a_NewStrsLs.p_Arr[l_ArrIdx] = l_Text;

            f_TotalRawChars = (uint)l_NewTotalLen;

            return true;
        }

        private void CopySelToClipboardButton_Click(object a_Sender,
                                                    EventArgs e)
        {
            string l_Str = BuildSortedString(true, Environment.NewLine);

            if (null == l_Str)
            {
                return;
            }

            Util.SetClipboard(l_Str);

            f_Inst.f_MainForm.InfoMsgBox("The selected texts were copied into the clipboard.");
        }

        private void OKButton_Click(object a_Sender,
                                    EventArgs e)
        {
            f_FinalText = BuildSortedString(false, f_SeparatorChar.ToString());

            base.DefaultOKCloseHandler(a_Sender, e);
        }

        private string BuildSortedString(bool a_SelectedOnly,
                                         string a_Separator)
        {
            // Return can be null if there are no entries.

            uint l_Count = (uint)(false == a_SelectedOnly ? f_GridView.p_ItemCount : f_GridView.p_SelectedCount);

            if (0 == l_Count)
            {
                return null;
            }

            string[] l_Arr = new string[l_Count];
            uint l_Idx = 0;

            if (false == a_SelectedOnly)
            {
                foreach (GridViewItem<uint, string> l_Item in f_GridView.Rows)
                {
                    l_Arr[l_Idx++] = l_Item.f_Tag;
                }
            }
            else
            {
                foreach (GridViewItem<uint, string> l_SelItem in f_GridView.SelectedRows)
                {
                    l_Arr[l_Idx++] = l_SelItem.f_Tag;
                }
            }

            List<string> l_Ls = new List<string>(l_Arr);
            l_Ls.Sort(f_StringComparer);
            l_Ls.CopyTo(l_Arr);

            string l_OutStr = string.Join(a_Separator, l_Arr);

            return l_OutStr;
        }

        private void InitializeList(string a_Input)
        {
            // a_Input may be null.

            RemoveAllItemsCore(false);

            if (null == a_Input)
            {
                return;
            }

            string[] l_Arr = Util.Split(a_Input, 0, f_SeparatorChar);
            uint l_ItemCount = Util.GetArrayLength(l_Arr);

            if (0 == l_ItemCount)
            {
                return;
            }

            uint l_CoreLen = 0;

            f_GridView.StartChanging();
            {
                foreach (string l_Str in l_Arr)
                {
                    uint l_StrLen = Util.GetStringLength(l_Str);

                    uint l_ItemID = FindUnoccupiedItemID();
                    f_GridView.CreateItem(l_ItemID).f_Tag = l_Str;
                    l_CoreLen += l_StrLen;
                }
            }
            f_GridView.EndChanging();

            f_TotalRawChars = (l_CoreLen + (l_ItemCount - 1));
        }

        public static bool ShowDialog(Instance a_Inst,
                                      uint a_MaxItemCount,
                                      uint a_MaxCharsPerItem,
                                      uint a_MaxTotalChars,
                                      char a_SeparatorChar,
                                      AcceptTextHandler a_AcceptText,
                                      IWin32Window a_Owner,
                                      string a_InitialText,
                                      out string a_FinalText)
        {
            // a_InitialText may be null. a_FinalText can be null on
            // successful return.

            using (TextItemsDialog l_Dlg = new TextItemsDialog(a_Inst, a_MaxItemCount, a_MaxCharsPerItem, a_MaxTotalChars, a_SeparatorChar, a_AcceptText))
            {
                l_Dlg.InitializeList(a_InitialText);

                if (DialogResult.OK == l_Dlg.ShowDialog(a_Owner))
                {
                    a_FinalText = l_Dlg.f_FinalText;

                    return true;
                }
            }

            a_FinalText = null;

            return false;
        }
    }

    sealed class TextBoxDialog : ButtonDialog
    {
        // A dialog with an optional label and a text box.

        private readonly TextSet f_TextSet;

        private readonly GuiGroupBox f_GroupBox;
        // f_TopLabel can be null.
        private readonly GuiLabel f_TopLabel;
        private readonly LabeledTextBox f_TextBox;

        public event EventHandler e_OKClick;

        public string p_TextBoxText
        {
            get { return f_TextBox.f_Box.Text; }
            set { f_TextBox.f_Box.Text = value; }
        }

        public TextBoxDialog(TextSet a_TextSet,
                             TextKeys a_DlgKey,
                             TextKeys a_GroupBoxKey,
                             TextKeys a_LabelKey,
                             TextKeys a_TextBoxKey,
                             uint a_MaxTextBoxChars) :
                        base(a_TextSet, DialogButtons.OKCancel)
        {
            // a_LabelKey may be TextKeys.Empty.

            f_TextSet = a_TextSet;

            const int lc_NativeDlgWidth = 500;
            int l_DlgWidth = Lyt.CalcX(lc_NativeDlgWidth);
            int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
            f_GroupBox = new GuiGroupBox(this);
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            bool l_ShowLabel = (TextKeys.Empty != a_LabelKey);

            if (true == l_ShowLabel)
            {
                f_TopLabel = new GuiLabel(f_GroupBox);
                f_TopLabel.Text = a_TextSet[a_LabelKey];
            }

            f_TextBox = new LabeledTextBox(f_GroupBox, l_InnerWidth);

            int l_LeftPadding = f_GroupBox.Padding.Left;
            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);

            if (true == l_ShowLabel)
            {
                f_TopLabel.Location = new Point(l_LeftPadding, f_GroupBox.Padding.Top);
                f_TopLabel.Width = l_InnerWidth;
                Lyt.SetFontAndHeight(f_TopLabel, f_TopLabel.Font);
            }

            f_TextBox.SetLocation(l_LeftPadding, f_TopLabel.GetBottom() + Lyt.s_SmallSpace);
            f_TextBox.SetMaxChars(a_MaxTextBoxChars);
            f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_TextBox.f_Box.Bottom);

            base.AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
            base.SetSelectOnShowControl(f_TextBox.f_Box);

            f_OKButton.Click += OKButton_Click;

            this.Text = a_TextSet[a_DlgKey];
            f_GroupBox.Text = a_TextSet[a_GroupBoxKey];
            f_TextBox.SetCaption(a_TextSet, a_TextBoxKey);
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                MainForm.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        public void AdjustButtonTexts(TextKeys a_OKKey,
                                      TextKeys a_CancelKey)
        {
            f_OKButton.Text = f_TextSet[a_OKKey];
            f_CancelButton.Text = f_TextSet[a_CancelKey];
        }

        private void OKButton_Click(object a_Sender,
                                    EventArgs e)
        {
            // Call the custom handler if the owner registered one.
            // Use this dialog as sender.

            if (null != e_OKClick)
            {
                e_OKClick(this, e);
            }
        }
    }

    class Invoker
    {
        // This class automatically calls EndInvoke() for each
        // BeginInvoke() so that the caller doesn't have to care
        // about handle leaks.

        private readonly Control f_Cntrl;
        private readonly List<IAsyncResult> f_Ls;

        public Invoker(Control a_Cntrl)
        {
            f_Cntrl = a_Cntrl;
            f_Ls = new List<IAsyncResult>();
        }

        public void TidyUp()
        {
            if (true == f_Cntrl.IsDisposed)
            {
                // This should prevent an ObjectDisposedException,
                // at least if the GUI thread is calling.

                return;
            }

            for (int i = (f_Ls.Count - 1); 0 <= i; i--)
            {
                IAsyncResult l_AsyncRes = f_Ls[i];

                if (true == l_AsyncRes.IsCompleted)
                {
                    object l_Junk = f_Cntrl.EndInvoke(l_AsyncRes);

                    f_Ls.RemoveAt(i);
                }
            }
        }

        public bool InvokeAsync(Delegate a_Delegate,
                                params object[] a_Args)
        {
            // a_Args may be null.

            TidyUp();

            IAsyncResult l_AsyncRes;

            try
            {
                l_AsyncRes = f_Cntrl.BeginInvoke(a_Delegate, a_Args);
            }
            catch (ObjectDisposedException)
            {
                return false;
            }

            f_Ls.Add(l_AsyncRes);

            return true;
        }
    }

#region GridViewHandling

    class ColumnSet
    {
        private const ushort c_Magic = 0x319A;

        private static string s_ReorderColsStr;
        private static string s_ResetColsStr;

        private readonly Column[] f_Cols;

        public int p_Count
        {
            get { return f_Cols.Length; }
        }

        public ColumnSet(Column[] a_Cols)
        {
            f_Cols = a_Cols;
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_ReorderColsStr = a_TextSet[TextKeys.ColumnsReorder];
            s_ResetColsStr = a_TextSet[TextKeys.ColumnsReset];
        }

        public Column[] ExposeArray()
        {
            return f_Cols;
        }

        public ColumnKind GetColumnKind(int a_Idx)
        {
            return f_Cols[a_Idx].f_Style;
        }

        public void MoveToTopOfSortOrder(DataGridViewColumn a_GridViewColumn)
        {
            // We store indices one-based so that we can flip positive
            // into negative values and vice versa.

            int l_ColIdx = (a_GridViewColumn.Index + 1);
            int l_InvColIdx = -l_ColIdx;

            for (int i = 0, lim = f_Cols.Length; i < lim; i++)
            {
                int l_CurColIdx = f_Cols[i].f_Adjustment.f_SortOrder;

                if (l_CurColIdx == l_ColIdx ||
                    l_CurColIdx == l_InvColIdx)
                {
                    for (int k = i; 0 < k; k--)
                    {
                        f_Cols[k].f_Adjustment.f_SortOrder = f_Cols[k - 1].f_Adjustment.f_SortOrder;
                    }

                    if (0 == i)
                    {
                        // Invert only if the column was already at the
                        // start of the sort order.

                        l_CurColIdx = -l_CurColIdx;
                    }

                    f_Cols[0].f_Adjustment.f_SortOrder = (sbyte)l_CurColIdx;
                    a_GridViewColumn.HeaderCell.SortGlyphDirection = f_Cols[0].f_Adjustment.SortOrderAsEnum();

                    return;
                }
            }

            throw (new ApplicationException());
        }

        public void CreateMenu(ContextMenuStrip a_Menu,
                               EventHandler a_ReorderClick,
                               EventHandler a_ResetClick,
                               EventHandler a_ToggleVisibilityClick)
        {
            for (int i = 0, lim = f_Cols.Length; i < lim; i++)
            {
                GenericMenuItem<int>.CreateItem(a_Menu, f_Cols[i].f_Name, a_ToggleVisibilityClick, i).Checked = f_Cols[i].f_Adjustment.f_Visible;
            }

            MainForm.AddMenuSeparator(a_Menu);
            MainForm.AddMenuItem(a_Menu, s_ReorderColsStr, a_ReorderClick, IconTypes.File);
            MainForm.AddMenuItem(a_Menu, s_ResetColsStr, a_ResetClick, IconTypes.Recycle);
        }

        public void ResetAdjustments()
        {
            for (int i = (f_Cols.Length - 1); 0 <= i; i--)
            {
                f_Cols[i].ResetAdjustments();
            }
        }

        public void AdjustmentsToBuffer(BufferIO a_Buffer)
        {
            int l_Count = f_Cols.Length;

            a_Buffer.WriteUInt16LE(c_Magic);
            a_Buffer.WriteByte((byte)l_Count);

            for (int i = 0; i < l_Count; i++)
            {
                f_Cols[i].f_Adjustment.ToBuffer(a_Buffer);
            }
        }

        public bool AdjustmentsFromBuffer(BufferIO a_Buffer)
        {
            // This routine is backward compatible and forward
            // tolerant.

            if (c_Magic != a_Buffer.ReadUInt16LE())
            {
                return false;
            }

            uint l_UseableCount = (uint)p_Count;
            uint l_ProvidedCount = a_Buffer.ReadByte();

            Column[] l_Cols = f_Cols;
            uint l_CommonCount = Math.Min(l_UseableCount, l_ProvidedCount);

            for (uint i = 0; i != l_CommonCount; i++)
            {
                ColumnAdjustment.FromBuffer(a_Buffer, out l_Cols[i].f_Adjustment);
            }

            ColumnAdjustment.CheckAndCorrectArray(l_Cols, l_CommonCount);

            if (l_UseableCount == l_ProvidedCount)
            {
                return true;
            }

            if (l_UseableCount > l_ProvidedCount)
            {
                for (uint i = l_ProvidedCount; i != l_UseableCount; i++)
                {
                    l_Cols[i].ResetAdjustments();
                }
            }
            else
            {
                if (false == ColumnAdjustment.DummyReadFromBuffer(a_Buffer, l_ProvidedCount - l_UseableCount))
                {
                    return false;
                }
            }

            return true;
        }

        public static bool DummyReadFromBuffer(BufferIO a_Buffer,
                                               uint a_DummyCount)
        {
            for (uint i = a_DummyCount; 0 != i--; )
            {
                if (c_Magic != a_Buffer.ReadUInt16LE())
                {
                    return false;
                }

                uint l_ColCount = a_Buffer.ReadByte();

                if (false == ColumnAdjustment.DummyReadFromBuffer(a_Buffer, l_ColCount))
                {
                    return false;
                }
            }

            return true;
        }

        public void InitializeGridViewColumns(DataGridViewColumnCollection a_Coll,
                                              int a_MinWidth)
        {
            int l_Count = f_Cols.Length;
            DataGridViewColumn[] l_Arr = new DataGridViewColumn[l_Count];

            for (int i = 0; i < l_Count; i++)
            {
                if (ColumnKind.Bool == f_Cols[i].f_Style)
                {
                    l_Arr[i] = new DataGridViewCheckBoxColumn(false);
                }
                else
                {
                    l_Arr[i] = new DataGridViewTextBoxColumn();
                }

                l_Arr[i].Name = f_Cols[i].f_Name;
                l_Arr[i].MinimumWidth = a_MinWidth;
                l_Arr[i].ReadOnly = true;
                l_Arr[i].SortMode = DataGridViewColumnSortMode.Programmatic;
            }

            a_Coll.AddRange(l_Arr);
        }

        public void LoadAdjustments(DataGridViewColumnCollection a_Coll)
        {
            for (int i = (f_Cols.Length - 1); 0 <= i; i--)
            {
                DataGridViewColumn l_Col = a_Coll[i];

                l_Col.HeaderCell.SortGlyphDirection = f_Cols[i].f_Adjustment.SortOrderAsEnum();
                l_Col.Visible = f_Cols[i].f_Adjustment.f_Visible;
                l_Col.Width = Lyt.CalcX(f_Cols[i].f_Adjustment.f_Width);
                l_Col.DisplayIndex = f_Cols[i].f_Adjustment.f_DisplayIndex;
            }
        }

        public void StoreAdjustments(DataGridViewColumnCollection a_Coll)
        {
            // Since the visibilty and the width are saved using a
            // different mechanism, this routine only saves the
            // display index.

            for (int i = (f_Cols.Length - 1); 0 <= i; i--)
            {
                f_Cols[i].f_Adjustment.f_DisplayIndex = (byte)a_Coll[i].DisplayIndex;
            }
        }

        public bool FlipVisibilty(int a_ColIndex)
        {
            bool l_NewState = !f_Cols[a_ColIndex].f_Adjustment.f_Visible;

            if (false == l_NewState)
            {
                bool l_AtLeastOneLeft = false;

                for (int i = p_Count; 0 < i--; )
                {
                    if (i != a_ColIndex &&
                        true == f_Cols[i].f_Adjustment.f_Visible)
                    {
                        l_AtLeastOneLeft = true;

                        break;
                    }
                }

                if (false == l_AtLeastOneLeft)
                {
                    return !l_NewState;
                }
            }

            f_Cols[a_ColIndex].f_Adjustment.f_Visible = l_NewState;

            return l_NewState;
        }

        public void ChangeColumnWidth(int a_ColIndex,
                                      int a_Width)
        {
            int l_ConvWidth = Math.Min(ushort.MaxValue, Lyt.InvCalcX(a_Width));
            f_Cols[a_ColIndex].f_Adjustment.f_Width = (ushort)l_ConvWidth;
        }

        public ReorderItemsDialog<int>.Item[] CreateItemsArrayForReordering(DataGridViewColumnCollection a_Coll)
        {
            int l_Count = f_Cols.Length;
            ReorderItemsDialog<int>.Item[] l_Items = new ReorderItemsDialog<int>.Item[l_Count];

            int l_ColIdx = 0;
            foreach (DataGridViewColumn l_Col in a_Coll)
            {
                // Take only the name from f_Cols, since e.g. the
                // display indices of f_Cols might not be up to date.
                int l_DisplIdx = l_Col.DisplayIndex;
                l_Items[l_DisplIdx] = new ReorderItemsDialog<int>.Item(f_Cols[l_ColIdx].f_Name, l_ColIdx, l_Col.Visible, l_DisplIdx);

                l_ColIdx++;
            }

            return l_Items;
        }

        public void ApplyItemsArrayAfterReordering(DataGridViewColumnCollection a_Coll,
                                                   ReorderItemsDialog<int>.Item[] a_Items,
                                                   bool a_OrderChanged)
        {
            // At least one item in a_Items must be visible. The order
            // of a_Items does not matter.

            foreach (ReorderItemsDialog<int>.Item l_Item in a_Items)
            {
                int l_ColIdx = l_Item.f_Tag;
                DataGridViewColumn l_GridViewCol = a_Coll[l_ColIdx];

                bool l_Visible = l_Item.p_Visible;
                f_Cols[l_ColIdx].f_Adjustment.f_Visible = l_Visible;
                l_GridViewCol.Visible = l_Visible;

                if (true == a_OrderChanged)
                {
                    int l_DisplOrderIdx = l_Item.p_DisplayOrderIndex;
                    f_Cols[l_ColIdx].f_Adjustment.f_DisplayIndex = (byte)l_DisplOrderIdx;
                    l_GridViewCol.DisplayIndex = l_DisplOrderIdx;
                }
            }
        }

        public enum ColumnKind : byte
        {
            Text = 0,
            Bool = 1,
            OwnerDrawn = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Creator
        {
            private readonly TextSet f_TextSet;
            private readonly TextKeys f_InitialTextKey;
            private readonly Column[] f_Cols;
            private int f_Index;

            public Creator(TextSet a_TextSet,
                           TextKeys a_InitialTextKey,
                           int a_ColumnCount)
            {
                f_TextSet = a_TextSet;
                f_InitialTextKey = a_InitialTextKey;
                f_Cols = new Column[a_ColumnCount];
                f_Index = 0;
            }

            public void AddText(Enum a_ColIdxAsEnum,
                                bool a_DefVisible,
                                int a_DefWidth)
            {
                Add(a_ColIdxAsEnum, a_DefVisible, a_DefWidth, ColumnKind.Text);
            }

            public void Add(Enum a_ColIdxAsEnum,
                            bool a_DefVisible,
                            int a_DefWidth,
                            ColumnKind a_Style)
            {
                string l_Name = f_TextSet[(TextKeys)((int)f_InitialTextKey + f_Index)];
                f_Cols[f_Index] = new Column(f_Index, l_Name, a_DefVisible, a_DefWidth, a_Style);
                f_Index++;
            }

            public ColumnSet Finish()
            {
                return new ColumnSet(f_Cols);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct Column
        {
            public readonly byte f_Index;
            public readonly string f_Name;
            public readonly bool f_DefVisible;
            public readonly ushort f_DefWidth;
            public readonly ColumnKind f_Style;

            public ColumnAdjustment f_Adjustment;

            public Column(int a_Index,
                          string a_Name,
                          bool a_DefVisible,
                          int a_DefWidth,
                          ColumnKind a_Style)
            {
                f_Index = (byte)a_Index;
                f_Name = a_Name;
                f_DefVisible = a_DefVisible;
                f_DefWidth = (ushort)a_DefWidth;
                f_Style = a_Style;

                f_Adjustment = new ColumnAdjustment(f_Index + 1, f_DefVisible, f_DefWidth, f_Index);
            }

            public void ResetAdjustments()
            {
                f_Adjustment = new ColumnAdjustment(f_Index + 1, f_DefVisible, f_DefWidth, f_Index);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ColumnMiniInfo
        {
            public readonly byte f_Index;
            public readonly string f_Name;
            public readonly ColumnKind f_Style;

            public ColumnMiniInfo(ref Column a_SrcCol)
            {
                f_Index = a_SrcCol.f_Index;
                f_Name = a_SrcCol.f_Name;
                f_Style = a_SrcCol.f_Style;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ColumnAdjustment
        {
            private const byte c_WrittenByteSize = 5;

            // f_SortOrder stores indices one-based, since zero
            // can't be signed.
            public sbyte f_SortOrder;
            public bool f_Visible;
            public ushort f_Width;
            public byte f_DisplayIndex;

            public byte p_SortOrderIndex
            {
                get { return (byte)((0 > f_SortOrder ? -(int)f_SortOrder : Math.Max(1, (int)f_SortOrder)) - 1); }
            }

            public bool p_SortDescending
            {
                get { return (0 > f_SortOrder); }
            }

            public ColumnAdjustment(int a_SortOrder,
                                    bool a_Visible,
                                    int a_Width,
                                    int a_DisplayIndex)
            {
                f_SortOrder = (sbyte)a_SortOrder;
                f_Visible = a_Visible;
                f_Width = (ushort)a_Width;
                f_DisplayIndex = (byte)a_DisplayIndex;
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteByte((byte)f_SortOrder);
                a_Buffer.WriteBool(f_Visible);
                a_Buffer.WriteUInt16LE(f_Width);
                a_Buffer.WriteByte(f_DisplayIndex);
            }

            public static void FromBuffer(BufferIO a_Buffer,
                                          out ColumnAdjustment a_Dest)
            {
                a_Dest = new ColumnAdjustment((sbyte)a_Buffer.ReadByte(),
                                              a_Buffer.ReadBool(),
                                              a_Buffer.ReadUInt16LE(),
                                              a_Buffer.ReadByte());
            }

            public static bool DummyReadFromBuffer(BufferIO a_Buffer,
                                                   uint a_DummyCount)
            {
                a_Buffer.JumpAhead(a_DummyCount * c_WrittenByteSize);

                return true;
            }

            public static void CheckAndCorrectArray(Column[] a_Cols,
                                                    uint a_Len)
            {
                // Checks and corrects the indices of the first a_Len
                // column adjustments of the input array.

                bool l_NeededCorrection = false;

                for (uint i = 0; i != a_Len; i++)
                {
                    CheckAndCorrectContext l_SortOrderIdxCntxt = new CheckAndCorrectContext(false);
                    CheckAndCorrectContext l_DisplayIdxCntxt = new CheckAndCorrectContext(false);

                    for (uint k = 0; k != a_Len; k++)
                    {
                        uint l_CurSortOrderIdx;
                        bool l_CurSortOrderDescending;
                        a_Cols[k].f_Adjustment.GetSortOrder(out l_CurSortOrderIdx, out l_CurSortOrderDescending);

                        if (true == l_SortOrderIdxCntxt.ProcessItem(k, i, l_CurSortOrderIdx, l_CurSortOrderDescending))
                        {
                            a_Cols[k].f_Adjustment.SetSortOrder(i + 1, l_CurSortOrderDescending);
                            l_NeededCorrection = true;
                        }

                        uint l_CurDisplIdx = a_Cols[k].f_Adjustment.f_DisplayIndex;

                        if (true == l_DisplayIdxCntxt.ProcessItem(k, i, l_CurDisplIdx, false))
                        {
                            a_Cols[k].f_Adjustment.f_DisplayIndex = (byte)(i + 1);
                            l_NeededCorrection = true;
                        }
                    }

                    uint l_ItemIdx;
                    bool l_ValueBool;

                    if (true == l_SortOrderIdxCntxt.GetResult(out l_ItemIdx, out l_ValueBool))
                    {
                        a_Cols[l_ItemIdx].f_Adjustment.SetSortOrder(i, l_ValueBool);
                        l_NeededCorrection = true;
                    }

                    if (true == l_DisplayIdxCntxt.GetResult(out l_ItemIdx, out l_ValueBool))
                    {
                        a_Cols[l_ItemIdx].f_Adjustment.f_DisplayIndex = (byte)i;
                        l_NeededCorrection = true;
                    }
                }

                bool l_HasVisibleCol = false;

                for (uint i = 0; i != a_Len; i++)
                {
                    if (true == a_Cols[i].f_Adjustment.f_Visible)
                    {
                        l_HasVisibleCol = true;

                        break;
                    }
                }

                if (false == l_HasVisibleCol &&
                    0 != a_Len)
                {
                    a_Cols[0].f_Adjustment.f_Visible = true;

                    if (10 > a_Cols[0].f_Adjustment.f_Width)
                    {
                        a_Cols[0].f_Adjustment.f_Width = 10;
                    }

                    l_NeededCorrection = true;
                }

                if (true == l_NeededCorrection)
                {
                    Util.DebugBreak();
                }
            }

            public void GetSortOrder(out uint a_SortOrderIndex,
                                     out bool a_SortDescending)
            {
                a_SortOrderIndex = p_SortOrderIndex;
                a_SortDescending = p_SortDescending;
            }

            public void SetSortOrder(uint a_SortOrderIndex,
                                     bool a_SortDescending)
            {
                a_SortOrderIndex++;
                f_SortOrder = (sbyte)(false == a_SortDescending ? -(int)a_SortOrderIndex : (int)a_SortOrderIndex);
            }

            public SortOrder SortOrderAsEnum()
            {
                if (0 < f_SortOrder)
                {
                    return SortOrder.Ascending;
                }

                if (0 > f_SortOrder)
                {
                    return SortOrder.Descending;
                }

                return SortOrder.None;
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct CheckAndCorrectContext
            {
                // This is a helper structure for CheckAndCorrectArray().

                // Don't use the struct default constructor.

                private bool f_GotIdealItem;
                private byte f_TargetItemIndex;
                private byte f_SmallestValueInteger;
                private bool f_SmallestValueBool;

                public CheckAndCorrectContext(bool a_Dummy)
                {
                    f_GotIdealItem = false;
                    f_TargetItemIndex = byte.MaxValue;
                    f_SmallestValueInteger = byte.MaxValue;
                    f_SmallestValueBool = false;
                }

                public bool ProcessItem(uint a_ItemIndex,
                                        uint a_AimedValue,
                                        uint a_ValueInteger,
                                        bool a_ValueBool)
                {
                    // a_ValueInteger and a_ValueBool may be any value.
                    // Returns a nonzero value if the caller should
                    // increment the value of the current item.

                    if (a_AimedValue > a_ValueInteger)
                    {
                        return false;
                    }

                    if (true == f_GotIdealItem)
                    {
                        if (a_AimedValue == a_ValueInteger)
                        {
                            return true;
                        }
                    }
                    else if (a_AimedValue == a_ValueInteger)
                    {
                        f_GotIdealItem = true;
                        f_TargetItemIndex = (byte)a_ItemIndex;
                    }
                    else if (a_ValueInteger < f_SmallestValueInteger ||
                             byte.MaxValue == f_TargetItemIndex)
                    {
                        f_TargetItemIndex = (byte)a_ItemIndex;
                        f_SmallestValueInteger = (byte)a_ValueInteger;
                        f_SmallestValueBool = a_ValueBool;
                    }

                    return false;
                }

                public bool GetResult(out uint a_TargetItemIndex,
                                      out bool a_SmallestValueBool)
                {
                    // Returns a nonzero value if the caller should
                    // update the current item with the returned values.

                    if (false == f_GotIdealItem)
                    {
                        a_TargetItemIndex = f_TargetItemIndex;
                        a_SmallestValueBool = f_SmallestValueBool;

                        return true;
                    }

                    a_TargetItemIndex = Util.c_NoIndex;
                    a_SmallestValueBool = false;

                    return false;
                }
            }
        }
    }

    enum GridViewBorderStyles : byte
    {
        NoBorder = 0,
        Horizontal = 1,
        Vertical = 2,
        Both = 3,
    }

    interface IGridView
    {
        void TakeOverFontAndBorderStyle();
    }

    sealed class GridView<TKey, TVal> : DataGridView, System.Collections.IComparer, IGridView
    {
        public delegate void UpdateItemHandler(Time a_Time,
                                               GridViewItem<TKey, TVal> a_Item);

        public delegate bool FillContextMenuHandler(GridView<TKey, TVal> a_Sender,
                                                    ContextMenuStripEx a_Menu,
                                                    bool a_ItemHit);

        public delegate void ItemDoubleClickedHandler(GridViewItem<TKey, TVal> a_Item);

        private static readonly int s_MinHeaderWidth;

        private readonly Instance f_Inst;
        private readonly ColumnSet f_ColSet;

        private readonly ContextMenuStripEx f_Menu;

        private readonly Dictionary<TKey, GridViewItem<TKey, TVal>> f_Items;
        private readonly List<DataGridViewRow> f_NewlyCreatedItems;

        private uint f_UpdateGen;
        private int f_RemoveCount;
        private GridViewItem<TKey, TVal> f_ProcessedItem;

        private bool f_DontUpdateNow;

        // e_UpdateItem is a mandatory event, a handler must be
        // present if items are inserted into the grid view.
        public event UpdateItemHandler e_UpdateItem;
        public event FillContextMenuHandler e_FillContextMenu;
        public event ToolStripDropDownClosedEventHandler e_ClosedContextMenu;
        public event ItemDoubleClickedHandler e_ItemDoubleClicked;

        public uint p_UpdateGeneration
        {
            get { return f_UpdateGen; }
        }

        public int p_ItemCount
        {
            get { return f_Items.Count; }
        }

        public int p_SelectedCount
        {
            get { return base.SelectedRows.Count; }
        }

        static GridView()
        {
            s_MinHeaderWidth = Lyt.CalcX(50);
        }

        public GridView(Instance a_Inst,
                        Control a_Owner,
                        ColumnSet a_ColSet)
        {
            f_Inst = a_Inst;
            f_ColSet = a_ColSet;

            f_ColSet.InitializeGridViewColumns(this.Columns, s_MinHeaderWidth);
            f_Menu = new ContextMenuStripEx(this, true);

            f_Items = new Dictionary<TKey, GridViewItem<TKey, TVal>>();
            f_NewlyCreatedItems = new List<DataGridViewRow>();

            this.VirtualMode = true;
            this.EditMode = DataGridViewEditMode.EditProgrammatically;
            this.AllowUserToOrderColumns = true;
            this.AllowUserToAddRows = false;
            this.AllowUserToDeleteRows = false;
            this.AllowUserToResizeRows = false;
            this.AutoSizeRowsMode = System.Windows.Forms.DataGridViewAutoSizeRowsMode.AllCells;
            this.ColumnHeadersHeightSizeMode = DataGridViewColumnHeadersHeightSizeMode.DisableResizing;
            this.RowHeadersVisible = false;
            this.SelectionMode = DataGridViewSelectionMode.FullRowSelect;
            this.DoubleBuffered = true;
            this.ColumnHeadersHeight = ((int)this.Font.GetHeight() + Lyt.s_Space);
            this.ContextMenuStrip = f_Menu;

            a_Owner.Controls.Add(this);
            this.Dock = DockStyle.Fill;

            f_Menu.Opening += Menu_Opening;
            f_Menu.Closed += Menu_Closed;

            TakeOverFontAndBorderStyle();
            f_Inst.f_MainForm.RegisterGridView(this);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_Inst.f_MainForm.UnregisterGridView(this);
                f_DontUpdateNow = true;

                f_Menu.Dispose();
                DisposeColumnHeaders();
                DisposeItems();
            }

            base.Dispose(disposing);

#if MONO
            if (true == disposing &&
                null != this.Parent)
            {
                // For some reason Mono doesn't automatically remove the
                // disposed control (as would be normal).

                this.Parent.Controls.Remove(this);
            }
#endif
        }

        private void DisposeColumnHeaders()
        {
            int l_Count = ColumnCount;
            DataGridViewColumn[] l_Cols = new DataGridViewColumn[l_Count];

            for (int i = 0; i < l_Count; i++)
            {
                l_Cols[i] = this.Columns[i];
            }

            this.Columns.Clear();

            foreach (DataGridViewColumn l_Col in l_Cols)
            {
                l_Col.Dispose();
            }
        }

        private void DisposeItems()
        {
            // The row count can be smaller than the element count of
            // f_Items (in this case the grid view is disposing
            // already). Still we need to dispose any items in f_Items.

            if (0 == f_Items.Count)
            {
                return;
            }

            base.Rows.Clear();

            foreach (GridViewItem<TKey, TVal> l_Item in f_Items.Values)
            {
                l_Item.Dispose();
            }

            f_Items.Clear();
        }

        public GridViewItem<TKey, TVal> GetItemByKey(TKey a_Key)
        {
            // Should not be used during an update. Return can be null.

            GridViewItem<TKey, TVal> l_Item;
            if (true == f_Items.TryGetValue(a_Key, out l_Item))
            {
                return l_Item;
            }

            return null;
        }

        public GridViewItem<TKey, TVal> GetItem(int a_RowIndex)
        {
            return (GridViewItem<TKey, TVal>)base.Rows[a_RowIndex];
        }

        public GridViewItem<TKey, TVal> GetSelectedItem(int a_SelIndex)
        {
            return (GridViewItem<TKey, TVal>)base.SelectedRows[a_SelIndex];
        }

        public TKey[] CreateItemKeyArr()
        {
            // Return value is null if no items exist.

            int l_Count = f_Items.Count;

            if (0 == l_Count)
            {
                return null;
            }

            TKey[] l_Arr = new TKey[l_Count];
            f_Items.Keys.CopyTo(l_Arr, 0);

            return l_Arr;
        }

        public TKey[] SelectedItemsToKeyArr()
        {
            // Return value is null if no items are selected.

            int l_Count = p_SelectedCount;

            if (0 == l_Count)
            {
                return null;
            }

            TKey[] l_Arr = new TKey[l_Count];

            int l_Idx = 0;
            foreach (GridViewItem<TKey, TVal> l_Item in base.SelectedRows)
            {
                l_Arr[l_Idx++] = l_Item.f_Ref;
            }

            return l_Arr;
        }

        public void SelectIfTabPageIsSelected(AdvancedTabPage a_TabPage)
        {
            if (true == a_TabPage.IsPageSelected())
            {
                this.Select();
            }
        }

        public void SelectAndDisplayItem(GridViewItem<TKey, TVal> a_Item)
        {
            this.ClearSelection();
            a_Item.Selected = true;
            EnsureItemIsDisplayed(a_Item);
        }

        public void EnsureItemIsDisplayed(GridViewItem<TKey, TVal> a_Item)
        {
            int l_RowIndex = a_Item.Index;

            Rectangle l_DisplRect = this.GetRowDisplayRectangle(l_RowIndex, true);

            if (l_DisplRect.Height < a_Item.Height)
            {
                this.FirstDisplayedScrollingRowIndex = l_RowIndex;
            }
        }

        protected override void OnCellDoubleClick(DataGridViewCellEventArgs e)
        {
            // Need to call base class implementation.

            // We also land here if the column header row was clicked.

            if (null != e_ItemDoubleClicked &&
                0 <= e.RowIndex)
            {
                e_ItemDoubleClicked(GetItem(e.RowIndex));
            }

            base.OnCellDoubleClick(e);
        }

        private void UpdateAllItems()
        {
            Time l_Time = f_Inst.GetTime();

            foreach (GridViewItem<TKey, TVal> l_Item in f_Items.Values)
            {
                UpdateSingleItem(l_Time, l_Item);
            }
        }

        private void UpdateSingleItem(Time a_Time,
                                      GridViewItem<TKey, TVal> a_Item)
        {
            if (a_Item.p_UpdateGeneration != f_UpdateGen)
            {
                e_UpdateItem(a_Time, a_Item);
                a_Item.SetUpdateGeneration(f_UpdateGen);
            }
        }

        private void PrepareItem(int a_RowIndex)
        {
            f_ProcessedItem = GetItem(a_RowIndex);

            if (f_ProcessedItem.p_UpdateGeneration != f_UpdateGen)
            {
                e_UpdateItem(f_Inst.GetTime(), f_ProcessedItem);
                f_ProcessedItem.SetUpdateGeneration(f_UpdateGen);
            }
        }

        public string CopyToString()
        {
            ColumnSet.ColumnMiniInfo[] l_Arr = GetViewedColsInDisplayOrder();
            GridViewItem<TKey, TVal>.Writer l_Wtr = new GridViewItem<TKey, TVal>.Writer();

            // Write the header.

            for (int i = 0, lim = l_Arr.Length; i < lim; i++)
            {
                l_Wtr.AddCol(l_Arr[i].f_Name);
            }

            l_Wtr.NextRow();

            if (0 < SelectedRows.Count)
            {
                // Write the body. Copy only selected items, but in
                // the order of the overall list (not of the
                // selection).

                Time l_Time = f_Inst.GetTime();

                foreach (GridViewItem<TKey, TVal> l_Item in f_Items.Values)
                {
                    if (true == l_Item.Selected)
                    {
                        UpdateSingleItem(l_Time, l_Item);
                        l_Item.AsString(l_Arr, l_Wtr);
                    }
                }
            }

            return l_Wtr.AsString();
        }

        private ColumnSet.ColumnMiniInfo[] GetViewedColsInDisplayOrder()
        {
            // Returns ColumnMiniInfos in display order, leaving out
            // owner drawn and invisible columns.

            ColumnSet.Column[] l_ColArr = f_ColSet.ExposeArray();
            SortedDictionary<int, int> l_Map = new SortedDictionary<int, int>();

            foreach (DataGridViewColumn l_Col in this.Columns)
            {
                int l_ColIdx = l_Col.Index;

                if (ColumnSet.ColumnKind.OwnerDrawn != l_ColArr[l_ColIdx].f_Style &&
                    true == l_Col.Visible)
                {
                    l_Map.Add(l_Col.DisplayIndex, l_ColIdx);
                }
            }

            int l_Count = l_Map.Count;
            ColumnSet.ColumnMiniInfo[] l_NewColArr = new ColumnSet.ColumnMiniInfo[l_Count];

            int l_VirtIdx = 0;
            foreach (int l_ColIdx in l_Map.Values)
            {
                l_NewColArr[l_VirtIdx++] = new ColumnSet.ColumnMiniInfo(ref l_ColArr[l_ColIdx]);
            }

            return l_NewColArr;
        }

        protected override void OnCellValueNeeded(DataGridViewCellValueEventArgs e)
        {
            // Need to call base class implementation.

            if (false == f_DontUpdateNow)
            {
                PrepareItem(e.RowIndex);

                int l_ColIdx = e.ColumnIndex;
                switch (f_ColSet.GetColumnKind(l_ColIdx))
                {
                    case ColumnSet.ColumnKind.Text:
                        e.Value = f_ProcessedItem.GetText(l_ColIdx);
                        break;

                    case ColumnSet.ColumnKind.Bool:
                        e.Value = f_ProcessedItem.GetBool(l_ColIdx);
                        break;
                }
            }

            base.OnCellValueNeeded(e);
        }

        protected override void OnCellPainting(DataGridViewCellPaintingEventArgs e)
        {
            // Need to call base class implementation.

            if (false == f_DontUpdateNow &&
                0 <= e.RowIndex &&
                0 <= e.ColumnIndex)
            {
                GridView_CellPainting(e);
            }

            base.OnCellPainting(e);
        }

        private void GridView_CellPainting(DataGridViewCellPaintingEventArgs e)
        {
            PrepareItem(e.RowIndex);

            OwnerDrawContext l_DrawCntxt = f_ProcessedItem.GetOwnerDrawContext(e.ColumnIndex);

            if (null == l_DrawCntxt)
            {
                return;
            }

            Rectangle l_DestRect = e.CellBounds;
            bool l_Sel = f_ProcessedItem.Selected;
            e.PaintBackground(l_DestRect, l_Sel);
            Graphics l_G = e.Graphics;
            GridViewBorderStyles l_Style = f_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_GridViewBorderStyle;

            switch (l_Style)
            {
                case GridViewBorderStyles.NoBorder:
                {
                    l_DestRect.Y++;
                    l_DestRect.Height -= 2;
                    BarDrawer.DrawFrame(l_G, ref l_DestRect, Pens.Black);
                    break;
                }

                case GridViewBorderStyles.Horizontal:
                    l_DestRect.Height--;
                    goto case GridViewBorderStyles.NoBorder;

                case GridViewBorderStyles.Vertical:
                {
                    l_DestRect.Y++;
                    l_DestRect.Width--;
                    l_DestRect.Height -= 2;
                    break;
                }

                case GridViewBorderStyles.Both:
                    l_DestRect.Width--;
                    l_DestRect.Height--;
                    break;

                default: throw (new ApplicationException());
            }

            if (1 > l_DestRect.Width ||
                1 > l_DestRect.Height)
            {
                return;
            }

            if (false == l_DrawCntxt.f_DrawProgressBar)
            {
                Bitmap l_SrcMap = l_DrawCntxt.f_Map;
                Rectangle l_SrcRect = new Rectangle(Point.Empty, l_SrcMap.Size);
                l_G.DrawImage(l_SrcMap, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
            }
            else
            {
                f_Inst.f_MainForm.f_BarDrawer.DrawProgressBar(f_Inst.f_TextSet, l_G, l_DestRect, l_DrawCntxt.f_ProgressBarStyle, l_DrawCntxt.f_Progress);
            }

            e.Handled = true;
        }

        protected override void OnColumnHeaderMouseClick(DataGridViewCellMouseEventArgs e)
        {
            // Need to call base class implementation.

            if (MouseButtons.Left == e.Button)
            {
                int l_Idx = e.ColumnIndex;
                f_ColSet.MoveToTopOfSortOrder(Columns[l_Idx]);

                Sort(true);
            }

            base.OnColumnHeaderMouseClick(e);
        }

        protected override void OnColumnWidthChanged(DataGridViewColumnEventArgs e)
        {
            // Need to call base class implementation.

            f_ColSet.ChangeColumnWidth(e.Column.Index, e.Column.Width);

            base.OnColumnWidthChanged(e);
        }

        private void Menu_Opening(object sender, System.ComponentModel.CancelEventArgs e)
        {
            f_Menu.DisposeItems();

            Point l_MousePos = this.PointToClient(Form.MousePosition);
            HitTestInfo l_Info = this.HitTest(l_MousePos.X, l_MousePos.Y);

            if (DataGridViewHitTestType.ColumnHeader == l_Info.Type)
            {
                f_ColSet.CreateMenu(f_Menu, MenuReorder_Click, MenuReset_Click, MenuToggleColumnVisibility_Click);
            }
            else
            {
                if (null == e_FillContextMenu)
                {
                    e.Cancel = true;
                }
                else
                {
                    bool l_ItemHit = (DataGridViewHitTestType.Cell == l_Info.Type);

                    e.Cancel = !e_FillContextMenu(this, f_Menu, l_ItemHit);
                }
            }
        }

        private void Menu_Closed(object sender, ToolStripDropDownClosedEventArgs e)
        {
            if (null != e_ClosedContextMenu)
            {
                e_ClosedContextMenu(sender, e);
            }
        }

        public void TakeOverFontAndBorderStyle()
        {
            MainForm l_Form = f_Inst.f_MainForm;
            Font l_Font = l_Form.f_GridViewFont;
            GridViewBorderStyles l_Style = l_Form.p_Prefs.f_GeneralPrefs.f_GridViewBorderStyle;

            this.Font = l_Font;
            this.ColumnHeadersHeight = (l_Font.Height + Lyt.s_Space);
            this.CellBorderStyle = TranslateBorderStyle(l_Style);
#if MONO
            this.DefaultCellStyle.Font = l_Font;
            this.ColumnHeadersDefaultCellStyle.Font = l_Font;
#endif
        }

        public void LoadColumnAdjustments()
        {
            f_ColSet.LoadAdjustments(this.Columns);
        }

        public void StoreColumnAdjustments()
        {
            f_ColSet.StoreAdjustments(this.Columns);
        }

        private void MenuReorder_Click(object sender, EventArgs e)
        {
            ReorderItemsDialog<int>.Item[] l_Items = f_ColSet.CreateItemsArrayForReordering(this.Columns);
            bool l_OrderChanged;

            using (ReorderItemsDialog<int> l_Dlg = new ReorderItemsDialog<int>(f_Inst, TextKeys.ColumnsReorder))
            {
                if (false == l_Dlg.ShowDialog(f_Inst.f_MainForm, 0, ModifyItemVisibilityMode.OneItemMustStayVisible, l_Items, out l_OrderChanged))
                {
                    return;
                }
            }

            f_ColSet.ApplyItemsArrayAfterReordering(this.Columns, l_Items, l_OrderChanged);
        }

        private void MenuReset_Click(object sender, EventArgs e)
        {
            if (false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.ColumnsMsgReallyReset]))
            {
                return;
            }

            f_ColSet.ResetAdjustments();
            LoadColumnAdjustments();
        }

        private void MenuToggleColumnVisibility_Click(object sender, EventArgs e)
        {
            int l_ColIndex = ((GenericMenuItem<int>)sender).f_GenericTag;

            Columns[l_ColIndex].Visible = f_ColSet.FlipVisibilty(l_ColIndex);
        }

        public void ClearItems()
        {
            if (0 != f_Items.Count)
            {
                // Make sure that Clear() doesn't update (it might).
                f_DontUpdateNow = true;
                DisposeItems();
                f_DontUpdateNow = false;

                f_UpdateGen++;
            }
        }

        public void NotifyValuesChanged(bool a_ForceValueRequery)
        {
            // Call this routine if the underlying database changed,
            // but StartChanging() wasn't called.

            if (0 != f_Items.Count)
            {
                f_UpdateGen++;

                if (true == a_ForceValueRequery)
                {
                    this.Invalidate();
                }
            }
        }

#if MONO

        private DataGridViewSelectedRowCollection f_SavedSelColl;

        private void MonoStoreBeforeChange()
        {
            f_SavedSelColl = this.SelectedRows;
        }

        private void DeselectAll()
        {
            foreach (DataGridViewRow l_Row in this.SelectedRows)
            {
                l_Row.Selected = false;
            }
        }

        private void MonoRestoreAfterChange()
        {
            // Usually Mono selects the first row after changing,
            // regardless of the previous situation.
            DeselectAll();

            if (0 < f_SavedSelColl.Count)
            {
                foreach (DataGridViewRow l_Row in f_SavedSelColl)
                {
                    // Select any rows that are still alive.

                    if (this == l_Row.DataGridView)
                    {
                        l_Row.Selected = true;
                    }
                }
            }

            f_SavedSelColl = null;

            // As of Mono v2.10.8 neither the FirstDisplayedCell
            // nor the HorizontalScrollingOffset property could be
            // used to restore the scroll offset.
        }

#endif

        public void StartChanging()
        {
            f_RemoveCount = f_Items.Count;

#if MONO
            MonoStoreBeforeChange();
#endif
        }

        public GridViewItem<TKey, TVal> CreateItem(TKey a_Key)
        {
            bool l_Junk;
            return CreateItem(a_Key, out l_Junk);
        }

        public GridViewItem<TKey, TVal> CreateItem(TKey a_Key,
                                                   out bool a_AlreadyChanged)
        {
            GridViewItem<TKey, TVal> l_Item;

            if (false == f_Items.TryGetValue(a_Key, out l_Item))
            {
                a_AlreadyChanged = false;

                l_Item = new GridViewItem<TKey, TVal>(a_Key, f_ColSet.p_Count);

                // For performance reasons we don't add one by one, but
                // use AddRange() later.

                f_Items.Add(a_Key, l_Item);
                f_NewlyCreatedItems.Add(l_Item);
            }
            else
            {
                a_AlreadyChanged = l_Item.p_Touched;

                if (false == a_AlreadyChanged)
                {
                    f_RemoveCount--;
                }
            }

            l_Item.Touch();

            return l_Item;
        }

        public GridViewItem<TKey, TVal> CreateSingleItem(TKey a_Key,
                                                         TVal a_Val)
        {
            GridViewItem<TKey, TVal> l_NewItem;

            StartChanging();
            {
                foreach (GridViewItem<TKey, TVal> l_Item in f_Items.Values)
                {
                    l_Item.Touch();
                }

                f_RemoveCount = 0;

                l_NewItem = CreateItem(a_Key);
                l_NewItem.f_Tag = a_Val;
            }
            EndChanging();

            return l_NewItem;
        }

        public void EndChanging()
        {
            // Removes unmarked items and unmarks the rest.

            TKey[] l_RemArr = new TKey[f_RemoveCount];

            // Do not allow updates now, since the base class has a
            // tedency to call OnCellValueNeeded() while Remove()
            // is invoked, which our child classes don't expect.
            f_DontUpdateNow = true;
            {
                foreach (KeyValuePair<TKey, GridViewItem<TKey, TVal>> l_Pair in f_Items)
                {
                    if (false == l_Pair.Value.p_Touched)
                    {
                        // This one by one removal is a performance issue,
                        // but there seems to be no better way.

                        base.Rows.Remove(l_Pair.Value);
                        l_Pair.Value.Dispose();
                        l_RemArr[--f_RemoveCount] = l_Pair.Key;
                    }
                    else
                    {
                        l_Pair.Value.Untouch();
                    }
                }

                foreach (TKey l_Key in l_RemArr)
                {
                    bool l_Removed = f_Items.Remove(l_Key);
                }

                base.Rows.AddRange(f_NewlyCreatedItems.ToArray());
                f_NewlyCreatedItems.Clear();
            }
            f_DontUpdateNow = false;

            f_UpdateGen++;
            Sort(false);

#if MONO
            MonoRestoreAfterChange();
#endif

            // Call Redraw() in the hope that it will speed up things.
            Redraw();
        }

        public void Redraw()
        {
            base.Invalidate();
        }

        private void Sort(bool a_Force)
        {
            if (false == a_Force &&
                f_Inst.f_MainForm.p_GridViewAutoSortItemThreshold < f_Items.Count)
            {
                return;
            }

            // Updating is expensive.

            UpdateAllItems();

            // Sorting is expensive as well.

            this.VirtualMode = false;
            base.Sort(this);
            this.VirtualMode = true;
        }

        public int Compare(object a_A,
                           object a_B)
        {
            return GridViewItem<TKey, TVal>.Compare(f_ColSet, (GridViewItem<TKey, TVal>)a_A, (GridViewItem<TKey, TVal>)a_B);
        }

        private DataGridViewCellBorderStyle TranslateBorderStyle(GridViewBorderStyles a_Style)
        {
            switch (a_Style)
            {
                case GridViewBorderStyles.NoBorder:
                    return DataGridViewCellBorderStyle.None;

                case GridViewBorderStyles.Horizontal:
                    return DataGridViewCellBorderStyle.SingleHorizontal;

                case GridViewBorderStyles.Vertical:
                    return DataGridViewCellBorderStyle.SingleVertical;

                case GridViewBorderStyles.Both:
                    return DataGridViewCellBorderStyle.Single;

                default:
                    throw(new ApplicationException());
            }
        }
    }

    sealed class GridViewItem<TKey, TVal> : DataGridViewRow
    {
        public readonly TKey f_Ref;
        private readonly Extra[] f_Extras;

        private bool f_Touched;
        private uint f_UpdateGen;

        public TVal f_Tag;

        public bool p_Touched
        {
            get { return f_Touched; }
        }

        public uint p_UpdateGeneration
        {
            get { return f_UpdateGen; }
        }

        public GridViewItem(TKey a_Ref,
                            int a_ColumnCount)
        {
            f_Ref = a_Ref;
            f_Extras = new Extra[a_ColumnCount];
        }

        public void Touch()
        {
            f_Touched = true;
        }

        public void Untouch()
        {
            f_Touched = false;
        }

        public void SetUpdateGeneration(uint a_Gen)
        {
            f_UpdateGen = a_Gen;
        }

        public OwnerDrawContext GetOwnerDrawContext(int a_ColSetColIdx)
        {
            // Return can be null (for a text column).

            return f_Extras[a_ColSetColIdx].f_OwnerContext;
        }

        public void SetTextWithCompObj(int a_ColSetColIdx,
                                       IComparable a_CompareObj,
                                       string a_Text)
        {
            f_Extras[a_ColSetColIdx] = new Extra(a_CompareObj, a_Text);
        }

        public void SetText(int a_ColSetColIdx,
                            float a_CompareWeight,
                            string a_Text)
        {
            f_Extras[a_ColSetColIdx] = new Extra((ulong)(a_CompareWeight * 1000), a_Text);
        }

        public void SetText(int a_ColSetColIdx,
                            ulong a_CompareWeight,
                            string a_Text)
        {
            f_Extras[a_ColSetColIdx] = new Extra(a_CompareWeight, a_Text);
        }

        public void SetBool(int a_ColSetColIdx,
                            bool a_BooleanValue)
        {
            f_Extras[a_ColSetColIdx] = new Extra(a_BooleanValue);
        }

        public void SetTime(int a_ColSetColIdx,
                            TextSet a_TextSet,
                            Time a_RefTime,
                            Time a_Time)
        {
            SetTime(a_ColSetColIdx, a_TextSet, a_RefTime, a_Time, null);
        }

        public void SetTime(int a_ColSetColIdx,
                            TextSet a_TextSet,
                            Time a_RefTime,
                            Time a_Time,
                            string a_TextIfNotSet)
        {
            // a_TextIfNotSet may be null.

            string l_Str;

            if (a_RefTime <= a_Time)
            {
                l_Str = TimeStrHelper.s_NowStr;
            }
            else if (false == a_Time.IsSet())
            {
                l_Str = a_TextSet.TimeStr(a_Time, a_TextIfNotSet);
            }
            else
            {
                Time l_Limit = a_RefTime.SubSeconds(10 * Time.c_SecsPerHour);

                if (l_Limit < a_Time)
                {
                    TimeStrHelper l_Hlpr;
                    TimeStrHelper.Create(a_Time, out l_Hlpr);
                    l_Str = l_Hlpr.GetRelativeTimeString(a_RefTime, null == a_TextIfNotSet ? string.Empty : a_TextIfNotSet);
                }
                else
                {
                    l_Str = a_TextSet.TimeStr(a_Time, a_TextIfNotSet);
                }
            }

            SetText(a_ColSetColIdx, a_Time.ToBinary(), l_Str);
        }

        public void SetDrawContext(int a_ColSetColIdx,
                                   OwnerDrawContext a_DrawCntxt)
        {
            // a_DrawCntxt can be null.

            if (null == a_DrawCntxt)
            {
                f_Extras[a_ColSetColIdx] = new Extra();
            }
            else
            {
                f_Extras[a_ColSetColIdx] = new Extra(a_DrawCntxt);
            }
        }

        public string GetText(int a_ColumnIndex)
        {
            return f_Extras[a_ColumnIndex].f_Text;
        }

        public bool GetBool(int a_ColumnIndex)
        {
            return (0 != f_Extras[a_ColumnIndex].f_CompareWeight);
        }

        public void AsString(ColumnSet.ColumnMiniInfo[] a_Arr,
                             Writer a_Wtr)
        {
            for (int i = 0, lim = a_Arr.Length; i < lim; i++)
            {
                int l_ColIdx = a_Arr[i].f_Index;

                switch (a_Arr[i].f_Style)
                {
                    case ColumnSet.ColumnKind.Text:
                        a_Wtr.AddCol(f_Extras[l_ColIdx].f_Text);
                        break;

                    case ColumnSet.ColumnKind.Bool:
                        if (0 != f_Extras[l_ColIdx].f_CompareWeight)
                        {
                            a_Wtr.AddCol('x');
                        }
                        break;
                }
            }

            a_Wtr.NextRow();
        }

        public static int Compare(ColumnSet a_ColSet,
                                  GridViewItem<TKey, TVal> a_A,
                                  GridViewItem<TKey, TVal> a_B)
        {
            ColumnSet.Column[] l_Arr = a_ColSet.ExposeArray();
            for (int i = 0, lim = a_ColSet.p_Count; i < lim; i++)
            {
                int l_Idx = l_Arr[i].f_Adjustment.f_SortOrder;
                bool l_Inv = (0 > l_Idx);

                if (true == l_Inv)
                {
                    l_Idx = -l_Idx;
                }

                // Convert one-based index into zero-based index.
                l_Idx--;

                int l_Res;

                if (null == a_A.f_Extras[l_Idx].f_CompareObj)
                {
                    l_Res = a_A.f_Extras[l_Idx].f_CompareWeight.CompareTo(a_B.f_Extras[l_Idx].f_CompareWeight);
                }
                else
                {
                    l_Res = a_A.f_Extras[l_Idx].f_CompareObj.CompareTo(a_B.f_Extras[l_Idx].f_CompareObj);
                }

                if (0 < l_Res)
                {
                    return (false == l_Inv ? 1 : (-1));
                }

                if (0 > l_Res)
                {
                    return (false == l_Inv ? (-1) : 1);
                }
            }

            return 0;
        }

        public static uint BuildCompareWeight(uint a_Val0,
                                              uint a_Val1)
        {
            uint l_Res = (((a_Val0 + a_Val1) << Util.c_BitsPerUShort) |
                          (byte.MaxValue + a_Val0 - a_Val1));

            return l_Res;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Extra
        {
            public readonly IComparable f_CompareObj;
            // f_CompareWeight will be used if f_CompareObj is null.
            public readonly ulong f_CompareWeight;

            public readonly string f_Text;
            public readonly OwnerDrawContext f_OwnerContext;

            public Extra(IComparable a_CompareObj,
                         string a_Text)
            {
                // For text columns.

                f_CompareObj = a_CompareObj;
                f_CompareWeight = 0;
                f_Text = a_Text;
                f_OwnerContext = null;
            }

            public Extra(ulong a_CompareWeight,
                         string a_Text)
            {
                // For text columns.

                f_CompareObj = null;
                f_CompareWeight = a_CompareWeight;
                f_Text = a_Text;
                f_OwnerContext = null;
            }

            public Extra(bool a_Flag)
            {
                // For boolean columns.

                f_CompareObj = null;
                f_CompareWeight = (false == a_Flag ? 0u : 1u);
                f_Text = null;
                f_OwnerContext = null;
            }

            public Extra(OwnerDrawContext a_OwnerContext)
            {
                // For owner drawn columns.

                f_CompareObj = null;
                f_CompareWeight = a_OwnerContext.f_CompareWeight;
                f_Text = null;
                f_OwnerContext = a_OwnerContext;
            }
        }

        public class Writer
        {
            private const char c_ColSepChar = '\t';
            private const char c_ColSepSubstChar = ' ';

            private readonly StringBuilder f_Bld;
            private bool f_FreshCol;

            public Writer()
            {
                f_Bld = new StringBuilder();
                f_FreshCol = true;
            }

            public void AddCol(string a_Str)
            {
                BeginCol();

                int l_Idx = f_Bld.Length;
                f_Bld.Append(a_Str);
                f_Bld.Replace(c_ColSepChar, c_ColSepSubstChar, l_Idx, a_Str.Length);
            }

            public void AddCol(char a_Char)
            {
                BeginCol();

                f_Bld.Append(c_ColSepChar == a_Char ? c_ColSepSubstChar : a_Char);
            }

            private void BeginCol()
            {
                if (false == f_FreshCol)
                {
                    f_Bld.Append(c_ColSepChar);
                }
                else
                {
                    f_FreshCol = false;
                }
            }

            public void NextRow()
            {
                f_Bld.AppendLine();
                f_FreshCol = true;
            }

            public string AsString()
            {
                return f_Bld.ToString();
            }
        }
    }

#endregion

    class GuiLabel : Label
    {
        public GuiLabel(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            Lyt.SetFontAndHeight(this, this.Font);

            this.BackColor = Color.Transparent;
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_SmallSpace);
        }
    }

    class GuiLinkLabel : LinkLabel
    {
        public GuiLinkLabel(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            Lyt.SetFontAndHeight(this, this.Font);

            this.BackColor = Color.Transparent;
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_SmallSpace);
        }
    }

    class TextGroup : IDisposable
    {
        private const ushort c_DefLeftWidth = 100;
        private const ushort c_LargeLeftWidth = 115;

        public readonly static int s_DefLeftWidth;
        private readonly static int s_LargeLeftWidth;

        private readonly Label f_Left;
        private readonly Label f_Right;

        static TextGroup()
        {
            s_DefLeftWidth = Lyt.CalcX(c_DefLeftWidth);
            s_LargeLeftWidth = Lyt.CalcX(c_LargeLeftWidth);
        }

        private TextGroup(Control a_Owner,
                          int a_Left,
                          int a_Top,
                          int a_LeftWidth)
        {
            f_Left = new Label();
            f_Right = new Label();

            a_Owner.Controls.Add(f_Left);
            a_Owner.Controls.Add(f_Right);

            f_Left.TextAlign = ContentAlignment.TopRight;
            Lyt.SetFontAndHeight(f_Left, f_Left.Font);
            f_Left.Left = a_Left;
            f_Left.Top = a_Top;
            f_Left.Width = a_LeftWidth;
            f_Left.BackColor = Color.Transparent;
            f_Right.Left = (a_Left + a_LeftWidth + Lyt.s_SmallSpace);
            f_Right.Top = a_Top;
            f_Right.Height = f_Left.Height;
            f_Right.BackColor = Color.Transparent;
        }

        public static TextGroup Create(Control a_Owner,
                                       int a_Left,
                                       int a_Top)
        {
            return new TextGroup(a_Owner, a_Left, a_Top, s_DefLeftWidth);
        }

        public static TextGroup CreateLargeRight(Control a_Owner,
                                                 int a_Left,
                                                 int a_Top)
        {
            TextGroup l_TextGroup = new TextGroup(a_Owner, a_Left, a_Top, s_DefLeftWidth);
            l_TextGroup.f_Right.Width = ushort.MaxValue;

            return l_TextGroup;
        }

        public static TextGroup CreateLargeLeft(Control a_Owner,
                                                int a_Left,
                                                int a_Top)
        {
            return new TextGroup(a_Owner, a_Left, a_Top, s_LargeLeftWidth);
        }

        public void Dispose()
        {
            f_Left.Dispose();
            f_Right.Dispose();
        }

        public void SetWidth(int a_Width)
        {
            f_Right.Width = Math.Max(0, a_Width - (f_Right.Left - f_Left.Left));
        }

        public void SetWidthAndAlignLeft(int a_Width)
        {
            f_Left.TextAlign = ContentAlignment.TopLeft;
            SetWidth(a_Width);
        }

        public void SetLeftText(TextSet a_TextSet,
                                TextKeys a_Key)
        {
            SetLeftText(a_TextSet.GetAsKeyPortion(a_Key));
        }

        public void SetLeftText(string a_Str)
        {
            f_Left.Text = a_Str;
        }

        public void ClearRightText()
        {
            SetRightText(string.Empty);
        }

        public void SetRightText(string a_Str)
        {
            f_Right.Text = a_Str;
        }

        public int GetBottom()
        {
            return (f_Left.Bottom + Lyt.s_SmallSpace);
        }

        public void SetVisibility(bool a_Visible)
        {
            f_Left.Visible = a_Visible;
            f_Right.Visible = a_Visible;
        }
    }

    class GuiCheckBox : CheckBox
    {
        private readonly int f_MaxWidth;

        public GuiCheckBox(Control a_Owner,
                           int a_MaxWidth)
        {
            f_MaxWidth = a_MaxWidth;

            a_Owner.Controls.Add(this);

            Lyt.SetFontAndHeight(this, this.Font);

            this.BackColor = Color.Transparent;
        }

        public void SetTextAndUpdateSize(string a_Text)
        {
            this.Text = a_Text;
            this.Width = GetSuggestedWidth(this, f_MaxWidth);
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_Space);
        }

        public static int GetSuggestedWidth(Control a_Cntrl,
                                            int a_MaxWidth)
        {
#if MONO
            return a_MaxWidth;
#else
            return Math.Min(a_MaxWidth, a_Cntrl.PreferredSize.Width);
#endif
        }
    }

    class GuiRadioButton : RadioButton
    {
        private readonly int f_MaxWidth;

        public GuiRadioButton(Control a_Owner,
                              int a_MaxWidth)
        {
            f_MaxWidth = a_MaxWidth;

            a_Owner.Controls.Add(this);

            Lyt.SetFontAndHeight(this, this.Font);

            this.BackColor = Color.Transparent;
        }

        public void SetTextAndUpdateSize(string a_Text)
        {
            this.Text = a_Text;
            this.Width = GuiCheckBox.GetSuggestedWidth(this, f_MaxWidth);
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_SmallSpace);
        }
    }

    class GuiButton : Button
    {
        public GuiButton(Control a_Owner) :
                    this(a_Owner, false)
        {
        }

        public GuiButton(Control a_Owner,
                         bool a_GenerousWidth)
        {
            a_Owner.Controls.Add(this);

            Lyt.SetFontAndHeight(this, this.Font);

            this.Width = (false == a_GenerousWidth ? Lyt.s_DefaultButtonWidth : Lyt.s_GenerousButtonWidth);
#if MONO
            // Higher buttons look slightly better on Mono.
            this.Height += (Lyt.s_SmallSpace / 2);
#endif
            this.BackColor = Color.Transparent;
        }

        public int SuggestOptimalWidth()
        {
            Size l_Size;
            MainForm.MeasureText(this.Text, this.Font, out l_Size);

            return (Lyt.s_Space2 + l_Size.Width);
        }

        public int GetRight()
        {
            return (this.Right + Lyt.s_SmallSpace);
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_SmallSpace);
        }
    }

    class PriorityPickerControl : LabeledComboBox
    {
        private static readonly int s_ComboBoxWidth;

        public static string[] s_TitleCasedTexts;
        public static string[] s_TextsPlusEmpty;
        public static string s_NoPriorityErrMsg;

        static PriorityPickerControl()
        {
            s_ComboBoxWidth = Lyt.CalcX(100);
        }

        public PriorityPickerControl(Control a_Owner,
                                     int a_Width) :
                                base(a_Owner, a_Width, s_ComboBoxWidth, false)
        {
            base.SetTexts(s_TextsPlusEmpty);
            Clear();
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            string[] l_RawTexts = a_TextSet.ReadArray(TextKeys.PriorityVeryLow, Priority.c_GuiPriorityStages);
            s_TitleCasedTexts = new string[Priority.c_GuiPriorityStages];
            s_TextsPlusEmpty = new string[Priority.c_GuiPriorityStages + 1];

            s_TextsPlusEmpty[0] = string.Empty;
            for (int i = 0; i < Priority.c_GuiPriorityStages; i++)
            {
                s_TitleCasedTexts[i] = a_TextSet.MakeTitlecasedIfEn(l_RawTexts[i]);
                s_TextsPlusEmpty[i + 1] = l_RawTexts[i];
            }

            s_NoPriorityErrMsg = a_TextSet[TextKeys.PriorityNoSuchErrMsg];
        }

        public void Clear()
        {
            f_Box.SelectedIndex = 0;
        }

        public bool GetPriority(out Priority a_Priority,
                                out string a_ErrMsg)
        {
            // If this routine fails, a_Priority is guaranteed to be
            // the default value of the Priority type.

            int l_Idx = f_Box.SelectedIndex;

            if (0 == l_Idx)
            {
                a_Priority = new Priority();
                a_ErrMsg = s_NoPriorityErrMsg;

                return false;
            }

            a_Priority = Priority.DecodeFromIndex(--l_Idx);
            a_ErrMsg = null;

            return true;
        }

        public void SetPriority(Priority a_Priority)
        {
            int l_Idx = a_Priority.EncodeIntoIndex();

            if (0 > l_Idx)
            {
                Clear();
            }
            else
            {
                f_Box.SelectedIndex = (l_Idx + 1);
            }
        }

        public static string PriorityToString(Priority a_Priority,
                                              string a_FallbackValue)
        {
            int l_Idx = a_Priority.EncodeIntoIndex();

            if (0 > l_Idx)
            {
                return a_FallbackValue;
            }

            return s_TextsPlusEmpty[l_Idx + 1];
        }
    }

    class SamControl : GuiGroupBox
    {
        private const string c_Localhost = "localhost";

        private readonly TextSet f_TextSet;

        private readonly GuiLabel f_Label;
        private readonly LabeledTextBox f_SamHostBox;
        private readonly LabeledNumberBox f_SamPortBox;

        public SamControl(TextSet a_TextSet,
                          Control a_Owner,
                          int a_Width) :
                      base(a_Owner)
        {
            f_TextSet = a_TextSet;

            int l_NumWidth = Lyt.CalcX(75);
            int l_HostTextWidth = (a_Width - Lyt.s_Space - l_NumWidth - this.Padding.Horizontal);
            f_Label = new GuiLabel(this);
            f_SamHostBox = new LabeledTextBox(this, l_HostTextWidth);
            f_SamPortBox = new LabeledNumberBox(this, l_NumWidth);

            f_Label.Location = new Point(this.Padding.Left, this.Padding.Top);
            f_Label.Width = (a_Width - this.Padding.Horizontal);
            f_Label.Text = f_TextSet[TextKeys.OptionsDlgAdvancedSamWarnText];
            Lyt.SetFontAndHeight(f_Label, f_Label.Font);
            int l_LabelBottom = (f_Label.Bottom + Lyt.s_Space);
            f_SamHostBox.SetLocation(this.Padding.Left, l_LabelBottom);
            f_SamPortBox.SetLocation(f_SamHostBox.GetRight(), l_LabelBottom);
            this.SetSizePlusBottomPadding(a_Width, f_SamHostBox.f_Box.Bottom);

            f_SamPortBox.SetIPPortLimits();

            this.Text = f_TextSet[TextKeys.OptionsDlgAdvancedSamCaption];
            f_SamHostBox.SetCaption(f_TextSet, TextKeys.OptionsDlgAdvancedSamTcpHost);
            f_SamPortBox.SetCaption(f_TextSet, TextKeys.OptionsDlgAdvancedSamTcpPort);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        public void SetFocus()
        {
            f_SamHostBox.SetFocus();
        }

        public void Set(System.Net.IPEndPoint a_EndPoint)
        {
            bool l_IsLoopback = System.Net.IPAddress.Loopback.Equals(a_EndPoint.Address);
            f_SamHostBox.f_Box.Text = (false == l_IsLoopback ? a_EndPoint.Address.ToString() : c_Localhost);
            f_SamPortBox.SetValueClipped(a_EndPoint.Port);
        }

        public System.Net.IPEndPoint TryParse(out string a_ErrMsg)
        {
            string l_Text = f_SamHostBox.f_Box.Text.Trim();
            System.Net.IPAddress l_Addr;

            if (c_Localhost == l_Text)
            {
                l_Addr = System.Net.IPAddress.Loopback;
            }
            else
            {
                if (false == System.Net.IPAddress.TryParse(l_Text, out l_Addr))
                {
                    a_ErrMsg = f_TextSet.Format(TextKeys.ParseIPAddrErrMsgFormat, l_Text);

                    return null;
                }
            }

            a_ErrMsg = null;

            return new System.Net.IPEndPoint(l_Addr, f_SamPortBox.p_ValueAsInt);
        }
    }

    class ButtonDialog : Dialog
    {
         // All three can be null.
        protected readonly GuiButton f_OKButton;
        protected readonly GuiButton f_CancelButton;
        protected readonly GuiButton f_ApplyButton;

        private readonly int f_ButtonCount;
        private readonly int f_ButtonHeight;

        // f_SelectOnShowCntrl can be null.
        private Control f_SelectOnShowCntrl;

        protected Rectangle f_InnerRect;

        protected int p_ButtonHeightPlusSpace
        {
            get { return (f_ButtonHeight + Lyt.s_Space); }
        }

        public ButtonDialog(TextSet a_TextSet,
                            DialogButtons a_Buttons)
        {
            this.SuspendLayout();

            if (0 != (DialogButtons.OK & a_Buttons))
            {
                f_OKButton = new GuiButton(this);
                f_OKButton.Text = a_TextSet[TextKeys.ButtonTextOK];
                this.AcceptButton = f_OKButton;
                f_ButtonCount++;
                f_ButtonHeight = f_OKButton.Height;
            }

            if (0 != (DialogButtons.Cancel & a_Buttons))
            {
                f_CancelButton = new GuiButton(this);
                f_CancelButton.Text = a_TextSet[TextKeys.ButtonTextCancel];
                this.CancelButton = f_CancelButton;
                f_ButtonCount++;
                f_ButtonHeight = f_CancelButton.Height;
            }

            if (0 != (DialogButtons.Apply & a_Buttons))
            {
                f_ApplyButton = new GuiButton(this);
                f_ApplyButton.Text = a_TextSet[TextKeys.ButtonTextApply];
                f_ButtonCount++;
                f_ButtonHeight = f_ApplyButton.Height;
            }

            UpdateSizes();

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing &&
                0 < f_ButtonCount)
            {
                if (null != f_OKButton)
                {
                    f_OKButton.Dispose();
                }

                if (null != f_CancelButton)
                {
                    f_CancelButton.Dispose();
                }

                if (null != f_ApplyButton)
                {
                    f_ApplyButton.Dispose();
                }
            }

            base.Dispose(disposing);
        }

        public void AsFixedButtonDialog(int a_Width,
                                        GuiGroupBox a_GroupBox)
        {
            AsFixedButtonDialog(a_Width, a_GroupBox.GetBottom());
        }

        public void AsFixedButtonDialog(int a_Width,
                                        int a_Height)
        {
            // Adds the height of the button (plus space) to the
            // passed height.

            AsFixedDialog(new Size(a_Width, a_Height + p_ButtonHeightPlusSpace));
        }

        protected void SetSelectOnShowControl(Control a_Cntrl)
        {
            // a_Cntrl may be null.

            f_SelectOnShowCntrl = a_Cntrl;
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            if (null != f_SelectOnShowCntrl)
            {
                f_SelectOnShowCntrl.Select();
            }

            base.OnShown(e);
        }

        protected override void OnResize(EventArgs e)
        {
            // Need to call base class implementation.

            UpdateSizes();

            base.OnResize(e);
        }

        private void UpdateSizes()
        {
            f_InnerRect = ClientRectangle;
            f_InnerRect.Height -= (f_ButtonHeight + Lyt.s_Space);
            f_InnerRect.Inflate(-Lyt.s_Space, -Lyt.s_Space);

            if (0 == f_ButtonCount)
            {
                return;
            }

            int l_Left = f_InnerRect.Right;
            int l_Top = (f_InnerRect.Bottom + Lyt.s_Space);

            this.SuspendLayout();

            if (null != f_ApplyButton)
            {
                l_Left -= f_ApplyButton.Width;
                f_ApplyButton.Location = new Point(l_Left, l_Top);
                l_Left -= Lyt.s_SmallSpace;
            }

            if (null != f_CancelButton)
            {
                l_Left -= f_CancelButton.Width;
                f_CancelButton.Location = new Point(l_Left, l_Top);
                l_Left -= Lyt.s_SmallSpace;
            }

            if (null != f_OKButton)
            {
                l_Left -= f_OKButton.Width;
                f_OKButton.Location = new Point(l_Left, l_Top);
                l_Left -= Lyt.s_SmallSpace;
            }

            this.ResumeLayout();
        }

        [Flags]
        public enum DialogButtons : byte
        {
            None = 0,
            OK = 0x1,
            Cancel = 0x2,
            Apply = 0x4,

            OKCancel = (OK | Cancel),
            All = (OK | Cancel | Apply),
        }
    }

    class Dialog : Form
    {
        private const float c_CoreBoundsRatio = .8f;

        public static readonly Color s_DialogBackgroundColor = Color.FromArgb(243, 242, 231);

        public Dialog()
        {
            this.AutoScaleMode = AutoScaleMode.None;
            this.BackColor = s_DialogBackgroundColor;
        }

        // NOTE: The below routines will invoke any OnResize handler.

        protected void AsSizeableDialog(Size a_Size)
        {
            this.FormBorderStyle = FormBorderStyle.Sizable;
            this.ShowIcon = false;
            this.ShowInTaskbar = false;
            this.SizeGripStyle = SizeGripStyle.Show;

            Size l_DlgSize = new Size(a_Size.Width, SuggestWindowHeight(a_Size.Height));
            SetSizes(l_DlgSize, false);
            base.MinimumSize = new Size(this.Width, Math.Min(this.Height, a_Size.Height));
            this.StartPosition = FormStartPosition.CenterScreen;
        }

        protected void AsFixedDialog(Size a_ClientSize)
        {
            this.FormBorderStyle = FormBorderStyle.FixedDialog;
            SetNonSizeableDialogDefaults();

            SetSizes(a_ClientSize, true);
            this.StartPosition = FormStartPosition.CenterScreen;
        }

        protected void AsToolDialog(Size a_ClientSize)
        {
            this.FormBorderStyle = FormBorderStyle.FixedToolWindow;
            SetNonSizeableDialogDefaults();

            SetSizes(a_ClientSize, true);
            this.StartPosition = FormStartPosition.Manual;
        }

        private void SetNonSizeableDialogDefaults()
        {
            this.MinimizeBox = false;
            this.MaximizeBox = false;
            this.ShowIcon = false;
            this.ShowInTaskbar = false;
            this.SizeGripStyle = SizeGripStyle.Hide;
        }

        private void SetSizes(Size a_ClientSize,
                              bool a_MinMaxSizeAlso)
        {
            if (false == a_MinMaxSizeAlso)
            {
                this.ClientSize = a_ClientSize;
            }
            else
            {
                // This will update the relation between Size and
                // ClientSize.
                this.Size = a_ClientSize;
                // Now we can retrieve the correct spacing.
                Size l_Diff = (this.Size - this.ClientSize);

                Size l_MinMax = (a_ClientSize + l_Diff);
                this.MinimumSize = l_MinMax;
                this.MaximumSize = l_MinMax;
                this.Size = l_MinMax;
            }
        }

        public void SetChildWindowPosition(Rectangle a_Origin)
        {
            int l_x = (a_Origin.Right - Lyt.s_SmallSpace);
            int l_y = (a_Origin.Bottom - Lyt.s_SmallSpace);
            Rectangle l_DesktopRect = Screen.FromControl(this).WorkingArea;

            if ((l_x + this.Width) > l_DesktopRect.Right)
            {
                l_x = Math.Max(a_Origin.Left - this.Width + Lyt.s_SmallSpace, l_DesktopRect.Left);
            }

            if ((l_y + this.Height) > l_DesktopRect.Bottom)
            {
                l_y = Math.Max(a_Origin.Top - this.Height + Lyt.s_SmallSpace, l_DesktopRect.Top);
            }

            this.Location = new Point(l_x, l_y);
        }

        protected Size GetWorkingArea()
        {
            return Screen.GetWorkingArea(this).Size;
        }

        public Size SuggestWindowSize(Size a_MinSize)
        {
            // Suggest a window size in a specified aspect ratio and
            // with respect to the desktop bounds.

            SizeF l_MaxSize = GetWorkingArea();
            l_MaxSize.Width *= c_CoreBoundsRatio;
            l_MaxSize.Height *= c_CoreBoundsRatio;

            // NOTE: A division by zero is virtually impossible here,
            // but wouldn't harm.
            float l_Ratio = ((float)a_MinSize.Height / a_MinSize.Width);
            float l_MaxRation = ((float)l_MaxSize.Height / l_MaxSize.Width);

            if (l_Ratio <= l_MaxRation)
            {
                // Excess of height.

                if (a_MinSize.Width >= l_MaxSize.Width)
                {
                    return a_MinSize;
                }

                l_MaxSize.Height = (l_MaxSize.Width * l_Ratio);
            }
            else
            {
                // Excess of width.

                if (a_MinSize.Height >= l_MaxSize.Height)
                {
                    return a_MinSize;
                }

                l_MaxSize.Width = (l_MaxSize.Height / l_Ratio);
            }

            return l_MaxSize.ToSize();
        }

        public int SuggestWindowHeight(int a_MinHeight)
        {
            float l_MaxHeight = GetWorkingArea().Height;
            l_MaxHeight *= c_CoreBoundsRatio;

            if (l_MaxHeight <= a_MinHeight)
            {
                return a_MinHeight;
            }

            return (int)l_MaxHeight;
        }

        protected void DefaultCancelCloseHandler(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.Cancel;
            this.Close();
        }

        protected void DefaultOKCloseHandler(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        public static void DisposeAllChildren(Control a_Cntrl)
        {
            int l_Count = a_Cntrl.Controls.Count;

            if (0 < l_Count)
            {
                Control[] l_Arr = new Control[l_Count];
                a_Cntrl.Controls.CopyTo(l_Arr, 0);

                foreach (Control l_Cntrl in l_Arr)
                {
                    l_Cntrl.Dispose();
                }
            }
        }

#if MONO

        private bool f_Shown;

        protected override void OnShown(EventArgs e)
        {
            f_Shown = true;

            base.OnShown(e);

            // Mono doesn't focus the form.
            this.Focus();
        }

        protected override void OnFormClosed(FormClosedEventArgs e)
        {
            if (true == f_Shown)
            {
                // Mono might call the closed handler more than once
                // when the form is closed. This prevents it.

                f_Shown = false;

                base.OnFormClosed(e);
            }
        }

#endif

        public static bool HasNonShiftModifiers(Keys a_Modifiers)
        {
            // In some cases we only allow the shift modifier this
            // routine aids in checking.

            return (0 != (a_Modifiers & (Keys.Alt | Keys.Control)));
        }
    }

    class DirectoryControl : Panel
    {
        private const ushort c_ProblematicPathChars = 248;
        private const ushort c_ButtonWidth = 25;

        private static readonly ushort s_ButtonWidth;

        private readonly Instance f_Inst;

        private readonly LabeledTextBox f_TextBox;
        private readonly GuiButton f_NavigateButton;
        private readonly GuiButton f_BrowseButton;

        private readonly string f_FolderBrowserMsg;
        private NB.Content.PathNormalizer f_PathNormalizer;

        public event EventHandler e_TextChanged;

        static DirectoryControl()
        {
            s_ButtonWidth = (ushort)Lyt.CalcX(c_ButtonWidth);
        }

        public DirectoryControl(Instance a_Inst,
                                Control a_Cntrl,
                                int a_Width,
                                TextKeys a_FolderBrowserMsgKey)
        {
            f_Inst = a_Inst;

            a_Cntrl.Controls.Add(this);

            f_TextBox = new LabeledTextBox(this, a_Width - (2 * (s_ButtonWidth + Lyt.s_SmallSpace)));
            f_NavigateButton = new GuiButton(this);
            f_BrowseButton = new GuiButton(this);

            this.BackColor = Color.Transparent;
            f_TextBox.SetLocation(Point.Empty);
            int l_Top = f_TextBox.f_Box.Top;
            int l_Left = (f_TextBox.GetRight() - Lyt.s_SmallSpace);
            f_NavigateButton.Location = new Point(l_Left, l_Top);
            f_BrowseButton.Location = new Point(l_Left + s_ButtonWidth + Lyt.s_SmallSpace, l_Top);
            f_NavigateButton.Size = new Size(s_ButtonWidth, f_TextBox.f_Box.Height);
            f_BrowseButton.Size = f_NavigateButton.Size;
            this.Size = new Size(a_Width, f_TextBox.f_Box.Bottom);

            f_TextBox.f_Box.TextChanged += Box_TextChanged;
            f_NavigateButton.Click += NavigateButton_Click;
            f_BrowseButton.Click += BrowseButton_Click;

            TextSet l_TextSet = f_Inst.f_TextSet;
            f_NavigateButton.Text = l_TextSet[TextKeys.CntrlsOpenDirectoryButton];
            f_BrowseButton.Text = l_TextSet[TextKeys.CntrlsBrowseFileSystemButton];
            f_FolderBrowserMsg = l_TextSet[a_FolderBrowserMsgKey];

            NB.Content.PathNormalizer.Create(out f_PathNormalizer);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        protected override void OnGotFocus(EventArgs e)
        {
            // Need to call base class implementation.

            f_TextBox.f_Box.Select();

            base.OnGotFocus(e);
        }

        public void SetCaption(TextKeys a_LabelTextKey)
        {
            f_TextBox.SetCaption(f_Inst.f_TextSet, a_LabelTextKey);
        }

        public void SetFocus()
        {
            f_TextBox.SetFocus();
        }

        private void Box_TextChanged(object sender, EventArgs e)
        {
            if (null != e_TextChanged)
            {
                e_TextChanged(this, EventArgs.Empty);
            }
        }

        private void NavigateButton_Click(object sender, EventArgs e)
        {
            bool l_Succ = f_Inst.f_MainForm.NavigateToFolder(GetPath(true));
        }

        private void BrowseButton_Click(object sender, EventArgs e)
        {
            using (FolderBrowserDialog l_Dlg = new FolderBrowserDialog())
            {
                l_Dlg.Description = f_FolderBrowserMsg;
                l_Dlg.SelectedPath = f_TextBox.f_Box.Text;
                l_Dlg.ShowNewFolderButton = true;

                if (DialogResult.OK == l_Dlg.ShowDialog(this))
                {
                    string l_Path = l_Dlg.SelectedPath;
                    bool l_Junk;

                    if (true == LongPath.IsPathRooted(l_Path, out l_Junk))
                    {
                        f_TextBox.f_Box.Text = l_Path;
                    }
                }
            }
        }

        public void SetMaxPathChars(int a_MaxChars)
        {
            f_TextBox.f_Box.MaxLength = a_MaxChars;
        }

        public string GetPath(bool a_Trim)
        {
            string l_Path = f_TextBox.f_Box.Text;

            if (true == a_Trim)
            {
                l_Path = l_Path.Trim();
            }

            return f_PathNormalizer.Normalize(l_Path);
        }

        public bool GetPathWithCheck(bool a_MaySuggestDirCreation,
                                     out string a_Path,
                                     out string a_ErrMsg)
        {
            a_Path = GetPath(true);

            return CheckPath(f_Inst, a_Path, a_MaySuggestDirCreation, out a_ErrMsg);
        }

        public void SetPath(string a_Path)
        {
            f_TextBox.f_Box.Text = a_Path;
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_Space);
        }

        public static bool CheckPath(Instance a_Inst,
                                     string a_Path,
                                     bool a_MaySuggestDirCreation,
                                     out string a_ErrMsg)
        {
            // NOTE: We are using some text keys from the shared dirs
            // options page here.

            TextSet l_TextSet = a_Inst.f_TextSet;

            if (0 == a_Path.Length)
            {
                a_ErrMsg = l_TextSet[TextKeys.OptionsDlgSharedDirsNoDir];

                return false;
            }

            bool l_IsValidPath;

            if (false == LongPath.IsPathRooted(a_Path, out l_IsValidPath) &&
                true == l_IsValidPath)
            {
                a_ErrMsg = "The path must be rooted.";

                return false;
            }

            if (true == l_IsValidPath &&
                true == LongPath.HasWindowsExtendedLengthPrefix(a_Path))
            {
                l_IsValidPath = false;
            }

            if (false == l_IsValidPath)
            {
                a_ErrMsg = "The path must not contain invalid characters.";

                return false;
            }

            if (false == System.IO.Directory.Exists(a_Path))
            {
                if (false == a_MaySuggestDirCreation)
                {
                    // Sadly .NET is unable to process paths longer
                    // than 247 chars.

                    string l_Extra = (c_ProblematicPathChars > a_Path.Length ? null : " " + l_TextSet[TextKeys.OptionsDlgSharedDirsWarnNoLongPathSupport]);
                    a_ErrMsg = (l_TextSet.Format(TextKeys.OptionsDlgSharedDirsDirNotFoundFormat, a_Path) + l_Extra);

                    return false;
                }

                if (false == a_Inst.f_MainForm.YesNoMsgBox(l_TextSet.Format("The directory \"{0}\" seems not to exist.\n\nShould {1} attempt to create this directory now?", a_Path, l_TextSet.f_AppName)))
                {
                    a_ErrMsg = "Please specify an existing directory.";

                    return false;
                }

                if (false == Util.EnsureDirectoryExists(a_Path))
                {
                    a_ErrMsg = l_TextSet.Format("Failed to create the directory \"{0}\".", a_Path);

                    return false;
                }
            }

            a_ErrMsg = null;

            return true;
        }
    }

    class ListAttributesControl : GuiGroupBox
    {
        public readonly LabeledNumberBoxWithUnit f_TextSizeRatio;
        public readonly LabeledComboBox f_BorderStyleBox;

        public ListAttributesControl(Control a_Owner,
                                     int a_Width) :
                                base(a_Owner)
        {
            int l_Width = (a_Width - this.Padding.Horizontal);
            int l_LeftWidth = (LabeledNumberBox.s_DefMaxTextBoxWidth + Lyt.CalcX(25));
            int l_RightWidth = Math.Min(Lyt.CalcX(140), l_Width - l_LeftWidth - Lyt.s_Space);
            f_TextSizeRatio = new LabeledNumberBoxWithUnit(this, l_LeftWidth);
            f_BorderStyleBox = new LabeledComboBox(this, l_RightWidth, false);

            int l_Top = this.Padding.Top;
            f_TextSizeRatio.SetLocation(this.Padding.Left, l_Top);
            f_TextSizeRatio.SetLimits(GeneralPrefs.c_MinGridViewTextSizeRatio, GeneralPrefs.c_MaxGridViewTextSizeRatio);
            f_BorderStyleBox.SetLocation(f_TextSizeRatio.GetRight(), l_Top);
            this.SetSizePlusBottomPadding(a_Width, Math.Max(f_TextSizeRatio.f_Box.Bottom, f_BorderStyleBox.f_Box.Bottom));
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_TextSizeRatio.Dispose();
                f_BorderStyleBox.Dispose();
            }

            base.Dispose(disposing);
        }

        public void SetTexts(TextSet a_TextSet)
        {
            this.Text = a_TextSet[TextKeys.OptionsDlgDisplListViewCaption];
            f_TextSizeRatio.SetCaption(a_TextSet, TextKeys.OptionsDlgDisplListViewTextSize);
            f_TextSizeRatio.SetUnitText(a_TextSet, TextKeys.OptionsDlgDisplListViewTextSizeUnit);
            string[] l_ItemTexts = a_TextSet.ReadArray(TextKeys.BorderStylesNoBorder, 4);
            f_BorderStyleBox.SetTexts(a_TextSet, TextKeys.OptionsDlgDisplListViewBorderStyle, l_ItemTexts);

            f_TextSizeRatio.f_Box.Value = GeneralPrefs.c_DefaultGridViewTextSizeRatio;
            f_BorderStyleBox.f_Box.SelectedIndex = 0;
        }

        public byte GetTextSizeRatio()
        {
            return (byte)f_TextSizeRatio.f_Box.Value;
        }

        public GridViewBorderStyles GetBorderStyle()
        {
            return (GridViewBorderStyles)f_BorderStyleBox.f_Box.SelectedIndex;
        }

        public void SetValues(byte a_TextSizeRatio,
                              GridViewBorderStyles a_BorderStyle)
        {
            f_TextSizeRatio.f_Box.Value = a_TextSizeRatio;
            f_BorderStyleBox.f_Box.SelectedIndex = (int)a_BorderStyle;
        }
    }

    class KadPublishControl : GuiGroupBox
    {
        private const byte c_CheckBoxCount = 3;

        private readonly GuiCheckBox[] f_CheckBoxes;

        public KadPublishControl(Control a_Owner,
                                 int a_Width) :
                            base(a_Owner)
        {
            f_CheckBoxes = new GuiCheckBox[c_CheckBoxCount];

            int l_InnerWidth = (a_Width - this.Padding.Horizontal);
            int l_Left = this.Padding.Left;
            int l_Top = this.Padding.Top;

            for (int i = 0; i < c_CheckBoxCount; i++)
            {
                f_CheckBoxes[i] = new GuiCheckBox(this, l_InnerWidth);
                f_CheckBoxes[i].Location = new Point(l_Left, l_Top);
                l_Top = f_CheckBoxes[i].GetBottom();
            }

            this.SetSizePlusBottomPadding(a_Width, f_CheckBoxes[c_CheckBoxCount - 1].Bottom);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        public void SetTexts(TextSet a_TextSet)
        {
            this.Text = a_TextSet[TextKeys.OptionsDlgKademliaPublCaption];

            string[] l_SubTexts = a_TextSet.ReadArray(TextKeys.OptionsDlgKademliaPublSource, c_CheckBoxCount);
            for (int i = 0; i < c_CheckBoxCount; i++)
            {
                f_CheckBoxes[i].SetTextAndUpdateSize(l_SubTexts[i]);
            }
        }

        public PublishFlags GetFlags()
        {
            PublishFlags l_Flags = PublishFlags.None;

            for (int i = 0; i < c_CheckBoxCount; i++)
            {
                if (true == f_CheckBoxes[i].Checked)
                {
                    l_Flags |= (PublishFlags)(1u << i);
                }
            }

            return l_Flags;
        }

        public void SetFlags(PublishFlags a_Flags)
        {
            uint l_Val = (uint)a_Flags;
            for (int i = 0; i < c_CheckBoxCount; i++)
            {
                f_CheckBoxes[i].Checked = (0 != ((1u << i) & l_Val));
            }
        }
    }

    class OwnPersonaControl : GuiGroupBox
    {
        private readonly GuiLabel f_UserIDLabel;
        private readonly GuiButton f_ResetButton;

        private OwnUserPersona f_OwnPersona;

        public event ArglessHandler e_Reset;

        public OwnUserPersona p_OwnPersona
        {
            get { return f_OwnPersona; }
        }

        public OwnPersonaControl(TextSet a_TextSet,
                                 Control a_Owner,
                                 int a_Width) :
                            base(a_Owner)
        {
            int l_TextWidth = (a_Width - this.Padding.Horizontal);
            f_UserIDLabel = new GuiLabel(this);
            f_ResetButton = new GuiButton(this);

            f_UserIDLabel.Location = new Point(this.Padding.Left, this.Padding.Top);
            f_UserIDLabel.Width = l_TextWidth;
            f_ResetButton.Location = new Point(this.Padding.Left, f_UserIDLabel.Bottom + Lyt.s_Space);
            f_ResetButton.Width = Lyt.s_GenerousButtonWidth;
            this.SetSizePlusBottomPadding(a_Width, f_ResetButton.Bottom);

            f_ResetButton.Click += ResetButton_Click;

            this.Text = a_TextSet[TextKeys.OptionsDlgeD2kUserCaption];
            f_ResetButton.Text = a_TextSet[TextKeys.OptionsDlgeD2kUserGenerateNew];
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_UserIDLabel.Dispose();
                f_ResetButton.Dispose();
            }

            base.Dispose(disposing);
        }

        public void Set(TextSet a_TextSet,
                        OwnUserPersona a_OwnPersona)
        {
            f_OwnPersona = a_OwnPersona;
            f_UserIDLabel.Text = a_TextSet.BuildKeyValuePair(TextKeys.OptionsDlgeD2kUserHash, f_OwnPersona.f_UserID.f_Val.ToHexString());
        }

        private void ResetButton_Click(object sender, EventArgs e)
        {
            if (null != e_Reset)
            {
                e_Reset();
            }
        }
    }

    class I2PAddressControl : GuiGroupBox
    {
        public delegate void ClearClickHandler(I2PAddressControl a_Cntrl);

        private readonly Instance f_Inst;

        private readonly LabeledTextBox f_AddressBox;
        private readonly GuiButton f_ClearButton;
        private readonly GuiRadioButton f_KeepButton;
        private readonly GuiRadioButton f_AutoManageButton;

        // f_KeyPair can be null.
        private I2PKeyPair f_KeyPair;

        public event ClearClickHandler e_ClearClick;

        public I2PKeyPair p_KeyPair
        {
            get { return f_KeyPair; }
        }

        public bool p_AutoManage
        {
            get { return f_AutoManageButton.Checked; }
        }

        public I2PAddressControl(Instance a_Inst,
                                 Control a_Owner,
                                 int a_Width,
                                 int a_SuggestedAgeInDays) :
                            base(a_Owner)
        {
            f_Inst = a_Inst;

            int l_ChildWidth = (a_Width - this.Padding.Horizontal);
            f_ClearButton = new GuiButton(this);
            f_AddressBox = new LabeledTextBox(this, l_ChildWidth - Lyt.s_Space - f_ClearButton.Width);
            f_KeepButton = new GuiRadioButton(this, l_ChildWidth);
            f_AutoManageButton = new GuiRadioButton(this, l_ChildWidth);

            int l_Left = this.Padding.Left;
            f_AddressBox.SetLocation(l_Left, this.Padding.Top);
            f_AddressBox.f_Box.ReadOnly = true;
            f_ClearButton.Location = new Point(f_AddressBox.GetRight(), f_AddressBox.f_Box.Top);
            f_KeepButton.Location = new Point(l_Left, f_AddressBox.GetBottom());
            f_AutoManageButton.Location = new Point(l_Left, f_KeepButton.GetBottom());
            this.SetSizePlusBottomPadding(a_Width, f_AutoManageButton.Bottom);

            f_ClearButton.Click += ClearButton_Click;

            f_KeepButton.SetTextAndUpdateSize(f_Inst[TextKeys.OptionsDlgAddressNeverChange]);
            f_AutoManageButton.SetTextAndUpdateSize(f_Inst.f_TextSet.Format(TextKeys.OptionsDlgAddressChangeAfterNDays, a_SuggestedAgeInDays));
            f_ClearButton.Text = f_Inst[TextKeys.OptionsDlgAddressDeleteAddr];

            SetKeyPair(null);
            AdjustAutoManagement(true);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        public void ClearKeyPair()
        {
            SetKeyPair(null);
        }

        public void SetKeyPair(I2PKeyPair a_KeyPair)
        {
            // a_KeyPair may be null.

            f_KeyPair = a_KeyPair;
            string l_Caption;
            string l_AddrText;

            if (null == f_KeyPair)
            {
                l_Caption = f_Inst[TextKeys.OptionsDlgAddressNewAddrWhenConnecting];
                l_AddrText = string.Empty;
            }
            else
            {
                long l_AgeInDays = f_KeyPair.CalcAgeInDays(f_Inst.GetTime());
                l_Caption = f_Inst.f_TextSet.Format(TextKeys.OptionsDlgAddressExistingAddrAge, l_AgeInDays);
                l_AddrText = f_KeyPair.f_PublicKey.ToFullString();
            }

            f_AddressBox.SetCaption(l_Caption);
            f_AddressBox.f_Box.Text = l_AddrText;
            f_AddressBox.f_Box.Enabled = (null != f_KeyPair);
            f_ClearButton.Enabled = (null != f_KeyPair);
        }

        private void ClearButton_Click(object sender, EventArgs e)
        {
            if (null != e_ClearClick)
            {
                e_ClearClick(this);
            }
        }

        public void AdjustAutoManagement(bool a_Set)
        {
            if (false == a_Set)
            {
                f_KeepButton.Checked = true;
            }
            else
            {
                f_AutoManageButton.Checked = true;
            }
        }
    }

    class TunnelControl : GuiGroupBox
    {
        private const byte c_MinQuantity = 1;
        private const byte c_MaxQuantity = 6;
        private const byte c_MaxRandExtraHops = 3;

        private static string[] s_QuantityStrs;
        private static string[] s_RandExtraHopsStrs;

        // f_UseInboundSettingsBox can be null.
        private readonly GuiCheckBox f_UseInboundSettingsBox;
        private readonly LabeledComboBox f_DatagramQuantity;
        private readonly LabeledComboBox f_StreamQuantity;
        private readonly LabeledComboBox f_MinHops;
        private readonly LabeledComboBox f_RandomExtraHops;

        public bool p_UseInboundSettings
        {
            get { return f_UseInboundSettingsBox.Checked; }
        }

        public TunnelControl(TextSet a_TextSet,
                             Control a_Owner,
                             int a_Width,
                             bool a_WithCheckBox) :
                        base(a_Owner)
        {
            if (true == a_WithCheckBox)
            {
                f_UseInboundSettingsBox = new GuiCheckBox(this, a_Width - this.Padding.Horizontal);
            }

            int l_ChildWidth = ((a_Width - Lyt.s_Space - this.Padding.Horizontal) / 2);
            int l_ComboWidth = Math.Min(l_ChildWidth, Lyt.CalcX(75));
            f_DatagramQuantity = new LabeledComboBox(this, l_ChildWidth, l_ComboWidth, false);
            f_StreamQuantity = new LabeledComboBox(this, l_ChildWidth, l_ComboWidth, false);
            f_MinHops = new LabeledComboBox(this, l_ChildWidth, l_ComboWidth, false);
            f_RandomExtraHops = new LabeledComboBox(this, l_ChildWidth, l_ComboWidth, false);

            int l_Left = this.Padding.Left;
            int l_Left2 = (l_Left + l_ChildWidth + Lyt.s_Space);
            int l_Top = this.Padding.Top;

            if (true == a_WithCheckBox)
            {
                f_UseInboundSettingsBox.Location = new Point(this.Padding.Left, l_Top);
                l_Top = f_UseInboundSettingsBox.GetBottom();
                f_UseInboundSettingsBox.Click += SameAsInboundBox_Click;
            }

            f_DatagramQuantity.SetLocation(l_Left, l_Top);
            f_StreamQuantity.SetLocation(l_Left2, l_Top);
            l_Top = f_DatagramQuantity.GetBottom();
            f_MinHops.SetLocation(l_Left, l_Top);
            f_RandomExtraHops.SetLocation(l_Left2, l_Top);
            this.SetSizePlusBottomPadding(a_Width, f_MinHops.f_Box.Bottom);

            if (true == a_WithCheckBox)
            {
                f_UseInboundSettingsBox.SetTextAndUpdateSize(a_TextSet[TextKeys.OptionsDlgTunnelsUseInbound]);
            }

            f_DatagramQuantity.SetTexts(a_TextSet, TextKeys.OptionsDlgTunnelsDatagramQuant, s_QuantityStrs);
            f_StreamQuantity.SetTexts(a_TextSet, TextKeys.OptionsDlgTunnelsStreamQuant, s_QuantityStrs);
            f_MinHops.SetTexts(a_TextSet, TextKeys.OptionsDlgTunnelsMinHops, s_QuantityStrs);
            f_RandomExtraHops.SetTexts(a_TextSet, TextKeys.OptionsDlgTunnelsRandExtraHops, s_RandExtraHopsStrs);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_QuantityStrs = new string[c_MaxQuantity - c_MinQuantity + 1];

            for (int i = (s_QuantityStrs.Length - 1); 0 <= i; i--)
            {
                s_QuantityStrs[i] = a_TextSet.NumStr(c_MinQuantity + i);
            }

            s_RandExtraHopsStrs = new string[c_MaxRandExtraHops + 1];
            s_RandExtraHopsStrs[0] = a_TextSet[TextKeys.OptionsDlgTunnelsNoHops];

            string l_PosNumFormat = a_TextSet[TextKeys.OptionsDlgTunnelsPosNumFormat];
            for (int k = 1; k <= c_MaxRandExtraHops; k++)
            {
                s_RandExtraHopsStrs[k] = string.Format(l_PosNumFormat, a_TextSet.NumStr(k));
            }
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                if (null != f_UseInboundSettingsBox)
                {
                    f_UseInboundSettingsBox.Dispose();
                }

                f_DatagramQuantity.Dispose();
                f_StreamQuantity.Dispose();
                f_MinHops.Dispose();
                f_RandomExtraHops.Dispose();
            }

            base.Dispose(disposing);
        }

        public void Set(bool a_SameAsInbound,
                        TunnelSettings a_Settings)
        {
            if (null == f_UseInboundSettingsBox)
            {
            }
            else
            {
                f_UseInboundSettingsBox.Checked = a_SameAsInbound;
                SameAsInboundBox_Click(null, null);
            }

            f_DatagramQuantity.f_Box.SelectedIndex = (a_Settings.f_DatagramQuantity - c_MinQuantity);
            f_StreamQuantity.f_Box.SelectedIndex = (a_Settings.f_StreamQuantity - c_MinQuantity);
            f_MinHops.f_Box.SelectedIndex = (a_Settings.f_MinHops - c_MinQuantity);
            f_RandomExtraHops.f_Box.SelectedIndex = a_Settings.f_RandomExtraHops;
        }

        public TunnelSettings Get()
        {
            return new TunnelSettings((uint)f_DatagramQuantity.f_Box.SelectedIndex + c_MinQuantity,
                                      (uint)f_StreamQuantity.f_Box.SelectedIndex + c_MinQuantity,
                                      (uint)f_MinHops.f_Box.SelectedIndex + c_MinQuantity,
                                      (uint)f_RandomExtraHops.f_Box.SelectedIndex);
        }

        private void SameAsInboundBox_Click(object sender, EventArgs e)
        {
            bool l_Enable = !f_UseInboundSettingsBox.Checked;
            f_DatagramQuantity.f_Box.Enabled = l_Enable;
            f_StreamQuantity.f_Box.Enabled = l_Enable;
            f_MinHops.f_Box.Enabled = l_Enable;
            f_RandomExtraHops.f_Box.Enabled = l_Enable;
        }
    }

    class CulturePickerControl : GuiGroupBox
    {
        private readonly BootConfig f_BootConfig;
        private readonly Instance f_Inst;

        private readonly LabeledComboBox f_LanguageBox;
        private readonly GuiButton f_VerifyLangButton;
        private readonly GuiButton f_CopyLangButton;
        private readonly LabeledComboBox f_FormatBox;

        public CulturePickerControl(Instance a_Inst,
                                    Control a_Owner,
                                    int a_ComboBoxMaxWidth,
                                    int a_Width) :
                               base(a_Owner)
        {
            f_BootConfig = Software.s_BootConfig;
            f_Inst = a_Inst;

            int l_ChildWidth = (a_Width - this.Padding.Horizontal);
            f_LanguageBox = new LabeledComboBox(this, l_ChildWidth, a_ComboBoxMaxWidth, false);
            f_VerifyLangButton = new GuiButton(this);
            f_CopyLangButton = new GuiButton(this);
            f_FormatBox = new LabeledComboBox(this, l_ChildWidth, a_ComboBoxMaxWidth, false);

            int l_Left = this.Padding.Left;
            int l_Top = this.Padding.Top;
            f_LanguageBox.SetLocation(l_Left, l_Top);
            f_CopyLangButton.Width = f_VerifyLangButton.Width = ((l_ChildWidth - Lyt.s_SmallSpace) / 2);
            f_VerifyLangButton.Location = new Point(l_Left, f_LanguageBox.GetBottom());
            f_CopyLangButton.Location = new Point(f_VerifyLangButton.GetRight(), f_VerifyLangButton.Top);
            f_FormatBox.SetLocation(l_Left, f_VerifyLangButton.GetBottom() + Lyt.s_SmallSpace);
            this.SetSizePlusBottomPadding(a_Width, f_FormatBox.f_Box.Bottom);

            f_LanguageBox.f_Box.SelectedIndexChanged += LanguageBox_SelectedIndexChanged;
            f_VerifyLangButton.Click += VerifyLangButton_Click;
            f_CopyLangButton.Click += CopyLangButton_Click;

            TextSet l_TextSet = f_Inst.f_TextSet;
            this.Text = l_TextSet[TextKeys.OptionsDlgLangLocalization];
            f_LanguageBox.SetCaption(l_TextSet, TextKeys.OptionsDlgLangTextLang);
            f_LanguageBox.SetTexts(BootConfig.ScanForLanguageCultures(l_TextSet[TextKeys.OptionsDlgLangBuiltInEng]));
            f_VerifyLangButton.Text = l_TextSet[TextKeys.OptionsDlgLangVerifyTextSet];
            f_CopyLangButton.Text = l_TextSet[TextKeys.OptionsDlgLangCopyTextsToClipboard];
            f_FormatBox.SetCaption(l_TextSet, TextKeys.OptionsDlgLangFormatCulture);
            f_FormatBox.SetTexts(BootConfig.BuildFormatCulturesArray(l_TextSet[TextKeys.OptionsDlgLangOSSetting]));

            f_LanguageBox.f_Box.SelectedIndex = 0;
            f_FormatBox.f_Box.SelectedIndex = 0;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        private void LanguageBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            f_VerifyLangButton.Enabled = (0 != f_LanguageBox.f_Box.SelectedIndex);
        }

        private void VerifyLangButton_Click(object sender, EventArgs e)
        {
            int l_LangIdx = f_LanguageBox.f_Box.SelectedIndex;

            if (0 == l_LangIdx)
            {
                Util.DebugBreak();

                return;
            }

            System.Globalization.CultureInfo l_LangCulture = GetCulture(f_LanguageBox, l_LangIdx);
            string l_Summary;
            TextSet l_NewTextSet = TextSet.LoadCustom(Software.s_DefaultEngTextSet, l_LangCulture, f_Inst.f_TextSet.f_FormatCulture, false, out l_Summary);

            if (null != l_NewTextSet &&
                0 == l_Summary.Length)
            {
                // This file seems to be in perfect shape.

                f_Inst.f_MainForm.InfoMsgBox(f_Inst[TextKeys.OptionsDlgLangMsgNoErrorsInTextSet]);
            }
            else
            {
                l_Summary = string.Format("BEGIN OF LANGUAGE FILE SUMMARY\nLanguage: {0}\n{1}\nEND OF LANGUAGE FILE SUMMARY", l_LangCulture.EnglishName, l_Summary);
                f_Inst.f_MainForm.AddToLog(l_Summary);

                f_Inst.f_MainForm.WarnMsgBox(f_Inst[TextKeys.OptionsDlgLangMsgErrorsDetectedSeeLog]);
            }
        }

        private void CopyLangButton_Click(object sender, EventArgs e)
        {
            int l_LangIdx = f_LanguageBox.f_Box.SelectedIndex;
            string l_Text;

            if (0 == l_LangIdx)
            {
                l_Text = Software.s_DefaultEngTextSet.JoinAllStringsForSpellChecking(null);
            }
            else
            {
                System.Globalization.CultureInfo l_LangCulture = GetCulture(f_LanguageBox, l_LangIdx);
                string l_Summary;
                TextSet l_NewTextSet = TextSet.LoadCustom(Software.s_DefaultEngTextSet, l_LangCulture, f_Inst.f_TextSet.f_FormatCulture, false, out l_Summary);

                if (null == l_NewTextSet)
                {
                    f_Inst.f_MainForm.ShowOperationFailedMsg();

                    return;
                }

                l_Text = l_NewTextSet.JoinAllStringsForSpellChecking(Software.s_DefaultEngTextSet);
            }

            Util.SetClipboard(l_Text);
        }

        public void ResetToDefaults()
        {
            f_BootConfig.SetLanguageCultureToDefault();
            f_BootConfig.SetFormatCultureToOSSetting();

            Load();
        }

        public void Load()
        {
            int l_LangIdx;

            if (true == f_BootConfig.HasDefaultLanguageCulture() ||
                false == FindCultureIndex(f_BootConfig.GetLanguageCulture(), f_LanguageBox, out l_LangIdx))
            {
                l_LangIdx = 0;
            }

            f_LanguageBox.f_Box.SelectedIndex = l_LangIdx;

            int l_FormatIdx;

            if (true == f_BootConfig.ShouldUseOSFormatCulture() ||
                false == FindCultureIndex(f_BootConfig.GetFormatCulture(), f_FormatBox, out l_FormatIdx))
            {
                l_FormatIdx = 0;
            }

            f_FormatBox.f_Box.SelectedIndex = l_FormatIdx;
        }

        public void Store()
        {
            int l_LangIdx = f_LanguageBox.f_Box.SelectedIndex;

            if (0 == l_LangIdx)
            {
                f_BootConfig.SetLanguageCultureToDefault();
            }
            else
            {
                f_BootConfig.SetLanguageCulture(GetCulture(f_LanguageBox, l_LangIdx));
            }

            int l_FormatIdx = f_FormatBox.f_Box.SelectedIndex;

            if (0 == l_FormatIdx)
            {
                f_BootConfig.SetFormatCultureToOSSetting();
            }
            else
            {
                f_BootConfig.SetFormatCulture(GetCulture(f_FormatBox, l_FormatIdx));
            }
        }

        private static System.Globalization.CultureInfo GetCulture(LabeledComboBox a_ComboBox,
                                                                   int a_Idx)
        {
            return ((BootConfig.CultureItem)a_ComboBox.f_Box.Items[a_Idx]).f_Culture;
        }

        private static bool FindCultureIndex(System.Globalization.CultureInfo a_Culture,
                                             LabeledComboBox a_ComboBox,
                                             out int a_Index)
        {
            a_Index = 0;
            foreach (BootConfig.CultureItem l_Item in a_ComboBox.f_Box.Items)
            {
                if (true == l_Item.HasCulture(a_Culture))
                {
                    return true;
                }

                a_Index++;
            }

            a_Index = (-1);

            return false;
        }
    }

    class DataSizePicker : LabeledNumberBoxWithUnit
    {
        private const byte c_DecimalPlaces = 2;

        private readonly decimal f_Granularity;
        private uint f_MinBytes;
        private uint f_MaxBytes;

        public DataSizePicker(Control a_Owner,
                              int a_Width,
                              bool a_SI) :
                         base(a_Owner, a_Width)
        {
            uint l_Granul = SizeFormatter.GetGranularity(a_SI);

            for (int i = 1; i < c_DecimalPlaces; i++)
            {
                l_Granul *= l_Granul;
            }

            f_Granularity = l_Granul;
            f_MaxBytes = uint.MaxValue;

            f_Box.DecimalPlaces = c_DecimalPlaces;
            f_UnitLabel.Text = SizeFormatter.GetUnitStr(a_SI, 2);
        }

        public void SetDataSizeLimits(uint a_MinBytes,
                                      uint a_MaxBytes)
        {
            f_MinBytes = a_MinBytes;
            f_MaxBytes = a_MaxBytes;

            decimal l_Min = Math.Round(f_MinBytes / f_Granularity, c_DecimalPlaces);
            decimal l_Max = Math.Round(f_MaxBytes / f_Granularity, c_DecimalPlaces);
            l_Min = Math.Min(l_Min, l_Max);

            f_Box.Minimum = l_Min;
            f_Box.Maximum = l_Max;
        }

        public uint GetDataSize()
        {
            // Return unit is bytes.

            uint l_Val = (uint)(f_Box.Value * f_Granularity);

            // Make sure we don't round outside the limit.

            return Math.Max(f_MinBytes, Math.Min(f_MaxBytes, l_Val));
        }

        public void SetDataSize(uint a_Bytes)
        {
            base.SetValueClipped(a_Bytes / f_Granularity);
        }
    }

    class BandwidthLimitControls : IDisposable
    {
        private readonly TextSet f_TextSet;

        private readonly GuiCheckBox f_CheckBox;
        private readonly TextBox f_TextBox;
        private readonly GuiLabel f_SideLbl;

        private ushort f_Granularity;
        private float f_Maximum;

        public BandwidthLimitControls(TextSet a_TextSet,
                                      Control a_Owner,
                                      int a_Width,
                                      bool a_SI)
        {
            f_TextSet = a_TextSet;

            f_CheckBox = new GuiCheckBox(a_Owner, a_Width);
            f_TextBox = new TextBox();
            f_SideLbl = new GuiLabel(a_Owner);

            a_Owner.Controls.Add(f_TextBox);

            f_TextBox.Width = LabeledNumberBox.s_DefMaxTextBoxWidth;
            f_SideLbl.Size = new Size(a_Width - LabeledNumberBox.s_DefMaxTextBoxWidth - Lyt.s_SmallSpace, f_TextBox.Height);
            f_SideLbl.TextAlign = ContentAlignment.MiddleLeft;

            f_TextBox.MaxLength = 15;
            f_CheckBox.Click += CheckBox_Click;

            AdjustDisplayStyle(a_SI);
            SetLimit(Speed.s_UnknownSpeed);
        }

        public void Dispose()
        {
            f_CheckBox.Dispose();
            f_TextBox.Dispose();
            f_SideLbl.Dispose();
        }

        private void CheckBox_Click(object sender, EventArgs e)
        {
            bool l_Checked = f_CheckBox.Checked;
            SetLimit(false == l_Checked ? Speed.s_MaxSpeed : Speed.s_UnknownSpeed);

            if (true == l_Checked)
            {
                f_TextBox.Select();
            }
        }

        public void AdjustDisplayStyle(bool a_SI)
        {
            f_Granularity = SizeFormatter.GetGranularity(a_SI);
            // NOTE: Single precision is not sufficient for this.
            f_Maximum = (float)Math.Floor(Speed.s_MaxSpeed.p_BytesPerSec / (double)f_Granularity);
            f_SideLbl.Text = SizeFormatter.GetUnitStr(a_SI, true, 1);
        }

        public void SetFocus()
        {
            f_TextBox.Select();
            f_TextBox.SelectAll();
        }

        public bool LoadSpeed(out Speed a_Speed,
                              out string a_ErrMsg)
        {
            // If this routine fails, a_Speed is guaranteed to be
            // Speed.s_UnknownSpeed.

            if (false == f_CheckBox.Checked)
            {
                a_Speed = Speed.s_MaxSpeed;
                a_ErrMsg = null;

                return true;
            }

            float l_Float;
            if (false == f_TextSet.TryParseFloat(f_TextBox.Text, 0f, f_Maximum, out l_Float, out a_ErrMsg))
            {
                a_ErrMsg = f_TextSet.Format(TextKeys.SpeedLimitFailedToParse, a_ErrMsg);

                a_Speed = Speed.s_UnknownSpeed;

                return false;
            }

            uint l_BytesPerSec = (uint)Math.Round((double)l_Float * f_Granularity, MidpointRounding.ToEven);
            l_BytesPerSec = Math.Min(l_BytesPerSec, Speed.c_MaxSpeedValue);

            a_Speed = new Speed(l_BytesPerSec);
            a_ErrMsg = null;

            return true;
        }

        public void SetLimit(Speed a_Speed)
        {
            bool l_Enable;
            string l_Text;

            if (true == Speed.s_UnknownSpeed.Equals(a_Speed))
            {
                l_Enable = true;
                l_Text = string.Empty;
            }
            else if (true == Speed.s_MaxSpeed.Equals(a_Speed))
            {
                l_Enable = false;
                l_Text = f_TextSet[TextKeys.SpeedUnlimited];
            }
            else
            {
                l_Enable = true;
                float l_EffLimit = Math.Min(a_Speed.p_BytesPerSec / (float)f_Granularity, f_Maximum);
                l_Text = f_TextSet.ShortFloatStr(l_EffLimit);
            }

            f_CheckBox.Checked = l_Enable;
            f_TextBox.Enabled = l_Enable;
            f_TextBox.Text = l_Text;
        }

        public void SetLocation(int a_x,
                                int a_y)
        {
            SetLocation(new Point(a_x, a_y));
        }

        public void SetLocation(Point a_Pnt)
        {
            f_CheckBox.Location = a_Pnt;
            f_TextBox.Location = new Point(a_Pnt.X, f_CheckBox.Bottom + Lyt.s_SmallSpace);
            f_SideLbl.Location = new Point(f_TextBox.Right + Lyt.s_SmallSpace, f_TextBox.Top);
        }

        public void SetText(string a_CheckBoxText)
        {
            f_CheckBox.SetTextAndUpdateSize(a_CheckBoxText);
        }

        public int GetRight()
        {
            return (f_SideLbl.Right + Lyt.s_Space);
        }

        public int GetBottom()
        {
            return (f_SideLbl.Bottom + Lyt.s_Space);
        }
    }

    class SearchSizePicker : IDisposable
    {
        private const ushort c_BoxWidth = 70;
        private const byte c_DefaultSelectedIndex = 2;

        private static readonly int s_BoxWidth;

        private readonly GuiCheckBox f_TopCheckBox;
        public readonly NumericUpDown f_NumBox;
        public readonly ComboBox f_ComboBox;

        private bool f_SI;

        static SearchSizePicker()
        {
            s_BoxWidth = Lyt.CalcX(c_BoxWidth);
        }

        public SearchSizePicker(Control a_Owner,
                                int a_Width)
        {
            f_TopCheckBox = new GuiCheckBox(a_Owner, a_Width);
            f_NumBox = new NumericUpDown();
            f_ComboBox = new ComboBox();

            a_Owner.Controls.Add(f_NumBox);
            a_Owner.Controls.Add(f_ComboBox);

            f_NumBox.Width = s_BoxWidth;
            f_ComboBox.DropDownStyle = ComboBoxStyle.DropDownList;
            f_ComboBox.Width = f_NumBox.Width;

            f_TopCheckBox.Click += TopCheckBox_Click;

            // Finish initialization.

            SetSizeFormatCore(false);
            Reset();
        }

        public void Dispose()
        {
            f_TopCheckBox.Dispose();
            f_NumBox.Dispose();
            f_ComboBox.Dispose();
        }

        public void SetLocation(int a_x,
                                int a_y)
        {
            SetLeft(a_x);
            SetTop(a_y);
        }

        public void SetLocation(Point a_Pnt)
        {
            SetLeft(a_Pnt.X);
            SetTop(a_Pnt.Y);
        }

        public void SetLeft(int a_x)
        {
            f_TopCheckBox.Left = a_x;
            f_NumBox.Left = a_x;
            f_ComboBox.Left = (f_NumBox.Right + Lyt.s_Space);
        }

        public void SetTop(int a_y)
        {
            f_TopCheckBox.Top = a_y;
            f_NumBox.Top = (f_TopCheckBox.Bottom + Lyt.s_SmallSpace);
            f_ComboBox.Top = f_NumBox.Top;
        }

        public void SetWidth(int a_Width)
        {
            f_TopCheckBox.Width = Math.Min(a_Width, f_TopCheckBox.PreferredSize.Width);
            f_ComboBox.Left = (f_NumBox.Right + Lyt.s_Space);
        }

        public void SetCaption(TextSet a_TextSet,
                               TextKeys a_Key)
        {
            f_TopCheckBox.SetTextAndUpdateSize(a_TextSet.GetAsInputControlCaption(a_Key));
        }

        public int GetBottom()
        {
            return (f_NumBox.Bottom + Lyt.s_Space);
        }

        private void TopCheckBox_Click(object sender, EventArgs e)
        {
            bool l_Enabled = f_TopCheckBox.Checked;

            f_NumBox.Enabled = l_Enabled;
            f_ComboBox.Enabled = l_Enabled;

            if (true == l_Enabled)
            {
                SetFocus();
            }
        }

        public void SetSizeFormat(bool a_SI)
        {
            if (a_SI != f_SI)
            {
                SetSizeFormatCore(a_SI);
            }
        }

        public void SetSizeFormatCore(bool a_SI)
        {
            ulong l_OldVal = GetSizeCore();

            f_SI = a_SI;
            f_NumBox.Maximum = (SizeFormatter.GetGranularity(f_SI) - 1);
            SizeFormatter.FillComboBox(f_SI, f_ComboBox);

            SetSizeCore(l_OldVal);
        }

        public void SetVisibilty(bool a_Set)
        {
            f_TopCheckBox.Visible = a_Set;
            f_NumBox.Visible = a_Set;
            f_ComboBox.Visible = a_Set;
        }

        public void Reset()
        {
            f_TopCheckBox.Checked = false;
            TopCheckBox_Click(null, null);
            f_NumBox.Value = 0;
            f_ComboBox.SelectedIndex = c_DefaultSelectedIndex;
        }

        public bool GetSize(out ulong a_Bytes)
        {
            if (false == f_TopCheckBox.Checked)
            {
                a_Bytes = 0;

                return false;
            }

            a_Bytes = GetSizeCore();

            return true;
        }

        public void SetSize(ulong a_Bytes)
        {
            f_TopCheckBox.Checked = true;
            TopCheckBox_Click(null, null);

            SetSizeCore(a_Bytes);
        }

        private ulong GetSizeCore()
        {
            ulong a_Bytes = (ulong)f_NumBox.Value;
            int l_Magnitude = f_ComboBox.SelectedIndex;

            if (0 < l_Magnitude)
            {
                a_Bytes *= SizeFormatter.GetSize(f_SI, l_Magnitude - 1);
            }

            return a_Bytes;
        }

        private void SetSizeCore(ulong a_Bytes)
        {
            int l_Magnitude = SizeFormatter.FindMagnitude(f_SI, a_Bytes);

            if (f_ComboBox.Items.Count <= l_Magnitude)
            {
                // We are not able to hold this large value.

                f_ComboBox.SelectedIndex = (f_ComboBox.Items.Count - 1);
                f_NumBox.Value = f_NumBox.Maximum;

                return;
            }

            f_ComboBox.SelectedIndex = l_Magnitude;

            if (0 == l_Magnitude)
            {
                f_NumBox.Value = a_Bytes;
            }
            else
            {
                f_NumBox.Value = (a_Bytes / SizeFormatter.GetSize(f_SI, l_Magnitude - 1));
            }
        }

        public void SetFocus()
        {
            if (false == f_TopCheckBox.Checked)
            {
                f_TopCheckBox.Select();
            }
            else
            {
                f_NumBox.Select();
                f_NumBox.Select(0, int.MaxValue);
            }
        }
    }

    class ReadOnlyTextBox : RichTextBox
    {
        private static bool s_HadScrollToCaretException;

        public ReadOnlyTextBox(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            base.DoubleBuffered = true;
            this.ScrollBars = RichTextBoxScrollBars.Vertical;
            this.ReadOnly = true;
            this.Multiline = true;
            this.DetectUrls = false;
            this.LanguageOption = RichTextBoxLanguageOptions.UIFonts;
        }

        public int GetViewedLine()
        {
            // Tests suggest not to use Point.Empty for that.
            int l_CharIdx = GetCharIndexFromPosition(new Point(1, 1));

            return GetLineFromCharIndex(l_CharIdx);
        }

        public void SetViewedLine(int a_LineIdx)
        {
            int l_CharIdx = GetFirstCharIndexFromLine(a_LineIdx);

            if (0 > l_CharIdx)
            {
                ScrollToEnd();
            }
            else
            {
                Select(l_CharIdx, 0);
                this.TryScrollToCaret();
            }
        }

        public void SetTextAndHoldPosition(string a_Text)
        {
            int l_LineIdx = GetViewedLine();
            Text = a_Text;
            SetViewedLine(l_LineIdx);
        }

        public void AppendTextAndHoldPosition(string a_Text)
        {
            int l_LineIdx = GetViewedLine();
            AppendText(a_Text);
            SetViewedLine(l_LineIdx);
        }

        public string GetTextWithPlatfromLineBreak()
        {
            StringBuilder l_Bld = new StringBuilder(this.Text);
            Util.TransformNewLineIntoPlatformLineBreak(l_Bld, 0, l_Bld.Length);

            return l_Bld.ToString();
        }

        public void ScrollToEnd()
        {
            int l_Pos = Text.Length;

            if (0 < l_Pos)
            {
                this.Select(l_Pos - 1, 0);
                this.TryScrollToCaret();
            }
        }

        private void TryScrollToCaret()
        {
            // ScrollToCaret() throws an exception on Wine 1.6.2.

            if (true == s_HadScrollToCaretException)
            {
                return;
            }

            try
            {
                this.ScrollToCaret();
            }
            catch (NotImplementedException)
            {
                s_HadScrollToCaretException = true;
            }
        }
    }

    class GuiPictureBox : PictureBox
    {
        private readonly Color f_ClearColor;
        // Both bitmaps can be null.
        private Bitmap f_Bitmap;
        private Bitmap f_RotatedBitmap;
        private float f_Rotation;

        public GuiPictureBox(Control a_Owner,
                             Color a_ClearColor)
        {
            f_ClearColor = a_ClearColor;

            a_Owner.Controls.Add(this);

            this.BackColor = f_ClearColor;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Clear();
            }

            base.Dispose(disposing);
        }

        public void ResetRotation()
        {
            SetRotation(0);
        }

        public void SetRotation(float a_Degrees)
        {
            if (a_Degrees != f_Rotation)
            {
                f_Rotation = a_Degrees;

                UpdatePicture(f_ClearColor);
            }
        }

        public void DisplayColor(Color a_Color)
        {
            if (null != f_Bitmap)
            {
                f_Bitmap.Dispose();
                f_Bitmap = null;
            }

            UpdatePicture(a_Color);
        }

        public void Clear()
        {
            SetPicture(null);
        }

        public void SetPictureAndSize(Bitmap a_Bitmap)
        {
            SetPicture(a_Bitmap);
            this.Size = a_Bitmap.Size;
        }

        public void SetPicture(Bitmap a_Bitmap)
        {
            // a_Bitmap may be null. If it's not, this instance takes
            // ownership of a_Bitmap.

            if (null != f_Bitmap)
            {
                f_Bitmap.Dispose();
            }

            f_Bitmap = a_Bitmap;

            UpdatePicture(f_ClearColor);
        }

        private void UpdatePicture(Color a_ClearColor)
        {
            // f_Bitmap may be null on entry.

            this.Image = null;

            if (null != f_RotatedBitmap)
            {
                f_RotatedBitmap.Dispose();
                f_RotatedBitmap = null;
            }

            if (null == f_Bitmap)
            {
                this.BackColor = a_ClearColor;

                return;
            }

            if (0 == f_Rotation)
            {
                SetPictureCore(f_Bitmap);

                return;
            }

            f_RotatedBitmap = new Bitmap(f_Bitmap.Width, f_Bitmap.Height);
            Point l_Pos = new Point(f_Bitmap.Width / 2, f_Bitmap.Height / 2);

            Graphics G = Graphics.FromImage(f_RotatedBitmap);
            {
                G.TranslateTransform(l_Pos.X, l_Pos.Y);
                G.RotateTransform(f_Rotation);
                G.DrawImageUnscaled(f_Bitmap, new Point(-l_Pos.X, -l_Pos.Y));
            }
            G.Dispose();

            SetPictureCore(f_RotatedBitmap);
        }

        private void SetPictureCore(Bitmap a_Bitmap)
        {
            this.Image = a_Bitmap;

            // By setting the back color to the color of the image's
            // lower right corner, we avoid a graphical error that
            // can occur when using PictureBoxSizeMode.StretchImage.
            this.BackColor = a_Bitmap.GetPixel(a_Bitmap.Width - 1, a_Bitmap.Height - 1);
        }
    }

    class GuiPanel : Panel
    {
        // Very similar to GuiGroupBox.

        public GuiPanel(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            this.BackColor = Color.Transparent;
        }

        public int GetBottom(bool a_PlusSpace)
        {
            return (this.Bottom + (false == a_PlusSpace ? 0 : (int)Lyt.s_Space));
        }
    }

    class GuiGroupBox : GroupBox
    {
        public static readonly MouseTimer s_MouseTimer;

        private static readonly Point[] s_SoftRectPoints;
        private static readonly SolidBrush s_GroupBoxBackgroundBrushNormal;
        private static readonly SolidBrush s_GroupBoxBackgroundBrushMoused;
        public static readonly Color s_SuggestedTrackBarBgColor;

        private bool f_EnableMouseOverEffect;
        private bool f_Moused;

        public bool p_EnableMouseOverEffect
        {
            get { return f_EnableMouseOverEffect; }
        }

        static GuiGroupBox()
        {
            s_MouseTimer = new MouseTimer();
            s_SoftRectPoints = new Point[10];
            s_GroupBoxBackgroundBrushNormal = new SolidBrush(Color.FromArgb(20, SystemColors.ControlDark));
            s_GroupBoxBackgroundBrushMoused = new SolidBrush(Color.FromArgb(50, SystemColors.ControlDark));

            using (Bitmap l_Bitmap = new Bitmap(1, 1))
            {
                using (Graphics l_G = Graphics.FromImage(l_Bitmap))
                {
                    l_G.Clear(Dialog.s_DialogBackgroundColor);
                    l_G.FillRectangle(s_GroupBoxBackgroundBrushNormal, 0, 0, 1, 1);
                }

                s_SuggestedTrackBarBgColor = l_Bitmap.GetPixel(0, 0);
            }
        }

        public GuiGroupBox(Control a_Owner) :
                      this(a_Owner, true)
        {
        }

        public GuiGroupBox(Control a_Owner,
                           bool a_EnableMouseOverEffect)
        {
            a_Owner.Controls.Add(this);

            const byte lc_FrameSize = 2;
            int l_NonTopPadding = (Lyt.s_Space + lc_FrameSize);
            this.Padding = new Padding(l_NonTopPadding, this.Font.Height + Lyt.s_Space, l_NonTopPadding, l_NonTopPadding);
            this.DoubleBuffered = true;

            f_EnableMouseOverEffect = a_EnableMouseOverEffect;
        }

        protected override void OnMouseEnter(EventArgs e)
        {
            // Need to call base class implementation.

            MouseTimer l_Timer = s_MouseTimer;

            if (true == l_Timer.p_Enabled &&
                true == f_EnableMouseOverEffect)
            {
                Point l_Pos = this.PointToClient(Form.MousePosition);

                if (true == BoxHitTest(l_Pos))
                {
                    l_Timer.AddManually(this);
                }
            }

            base.OnMouseEnter(e);
        }

        private bool BoxHitTest(Point a_ClientPoint)
        {
            // Returns a nonzero value if the mouse is inside the
            // group box frame.

            return (true == this.ClientRectangle.Contains(a_ClientPoint) &&
                    (this.Font.Height / 2) < a_ClientPoint.Y);
        }

        private void AdjustMousedFlag(bool a_Moused)
        {
            // Should only be used by the nested MouseTimer class. This
            // routine may get called after the control was disposed.

            if (a_Moused == f_Moused)
            {
                return;
            }

            f_Moused = a_Moused;

            if (false == this.IsDisposed)
            {
                this.Invalidate();
            }
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            Rectangle l_Rect = this.ClientRectangle;
            DrawGroupBox(e.Graphics, this, f_Moused, ref l_Rect);

            // Do not call the base class.
        }

        private static void DrawGroupBox(Graphics a_G,
                                         GroupBox a_GroupBox,
                                         bool a_Moused,
                                         ref Rectangle a_Rect)
        {
            Font l_Font = a_GroupBox.Font;
            int l_HalfFontHeight = (l_Font.Height / 2);
            Rectangle l_FrameRect = new Rectangle(a_Rect.X, a_Rect.Y + l_HalfFontHeight, a_Rect.Width, a_Rect.Height - l_HalfFontHeight);

            if (0 >= a_Rect.Width ||
                0 >= a_Rect.Height)
            {
                return;
            }

            int l_TopGapX = 0;
            int l_TopGapWidth = 0;
            string l_Caption = a_GroupBox.Text;

            if (0 < l_Caption.Length)
            {
                bool l_Enabled = a_GroupBox.Enabled;
                Brush l_TextBrush;
                bool l_DisposeTextBrush = false;

                if (false == l_Enabled)
                {
                    l_TextBrush = SystemBrushes.ButtonShadow;
                }
                else
                {
                    Color l_TextColor = a_GroupBox.ForeColor;

                    if (0xFF000000 == (uint)l_TextColor.ToArgb())
                    {
                        l_TextBrush = Brushes.Black;
                    }
                    else
                    {
                        l_TextBrush = new SolidBrush(l_TextColor);
                        l_DisposeTextBrush = true;
                    }
                }

                System.Drawing.Text.TextRenderingHint l_StoredHint = a_G.TextRenderingHint;
                {
                    ushort l_GapMargin = Lyt.s_SmallSpace;
                    ushort l_GapPadding = Lyt.s_TinySpace;
                    bool l_FromRight = (RightToLeft.Yes == a_GroupBox.RightToLeft);
                    SizeF l_SizeF = a_G.MeasureString(l_Caption, l_Font, a_Rect.Width);
                    Size l_Size = new Size((int)l_SizeF.Width + 1, (int)l_SizeF.Height + 1);
                    l_TopGapWidth = (l_Size.Width + (2 * l_GapPadding));
                    l_TopGapX = (false == l_FromRight ? l_GapMargin : Math.Max(l_GapMargin, a_Rect.Width - l_GapMargin - l_TopGapWidth));

                    Rectangle l_TextRect = new Rectangle(a_Rect.X + l_TopGapX + l_GapPadding, a_Rect.Y, l_Size.Width, l_Size.Height);
                    a_G.DrawString(l_Caption, l_Font, l_TextBrush, l_TextRect);
                }
                a_G.TextRenderingHint = l_StoredHint;

                if (true == l_DisposeTextBrush)
                {
                    l_TextBrush.Dispose();
                }
            }

            DrawSoftRectangle(a_G, ref l_FrameRect, l_TopGapX, l_TopGapWidth, SystemPens.ControlDark, SystemPens.ControlLightLight, false == a_Moused ? s_GroupBoxBackgroundBrushNormal : s_GroupBoxBackgroundBrushMoused);
        }

        private static void DrawSoftRectangle(Graphics a_G,
                                              ref Rectangle a_DestRect,
                                              int a_TopGapX,
                                              int a_TopGapWidth,
                                              Pen a_MainPen,
                                              Pen a_SecondPen,
                                              Brush a_BackgroundBrush)
        {
            // a_TopGapX and a_TopGapWidth may be greater than the
            // the horizontal bounds of a_DestRect.

            const byte lc_MinSize = 3;

            if (lc_MinSize > a_DestRect.Width ||
                lc_MinSize > a_DestRect.Height)
            {
                return;
            }

            Point[] l_Points = s_SoftRectPoints;

            int l_X = (a_DestRect.X + 1);
            int l_Y = (a_DestRect.Y + 1);
            int l_Right = (a_DestRect.Right - 2);
            int l_Bottom = (a_DestRect.Bottom - 1);
            int l_SeamX = Util.MinMaxSigned(l_X, l_Right, a_DestRect.X + a_TopGapX);
            int l_SeamRight = Math.Min(l_SeamX + a_TopGapWidth, l_Right);

            {
                // Draw the background of the frame.

                Rectangle l_BgRect = new Rectangle(l_X + 1, l_Y, a_DestRect.Width - 4, a_DestRect.Height - 3);

                if (l_SeamX == l_SeamRight)
                {
                    // No gap, we can reduce the rect.

                    l_BgRect.Y++;
                    l_BgRect.Height--;
                }

                a_G.FillRectangle(a_BackgroundBrush, l_BgRect);
            }

            // Draw the (mostly) inner frame.

            l_Points[0] = new Point(l_SeamX, l_Y);
            l_Points[1] = new Point(l_X, l_Y);
            l_Points[2] = new Point(l_X, l_Bottom - 3);
            l_Points[3] = new Point(l_X - 1, l_Bottom - 2);
            l_Points[4] = new Point(l_X + 1, l_Bottom);
            l_Points[5] = new Point(l_Right - 1, l_Bottom);
            l_Points[6] = new Point(l_Right + 1, l_Bottom - 2);
            l_Points[7] = new Point(l_Right, l_Bottom - 3);
            l_Points[8] = new Point(l_Right, l_Y);
            l_Points[9] = new Point(l_SeamRight, l_Y);
            a_G.DrawLines(a_SecondPen, l_Points);

            // Draw the (mostly) outer frame.

            l_Points[0].Y--;
            l_Points[1].Offset(+1, -1);
            l_Points[2] = new Point(l_X - 1, l_Y + 1);
            l_Points[3].Y--;
            l_Points[4].Y--;
            l_Points[5].Y--;
            l_Points[6].Y--;
            l_Points[7] = new Point(l_Right + 1, l_Y + 1);
            l_Points[8].Offset(-1, -1);
            l_Points[9].Y--;
            a_G.DrawLines(a_MainPen, l_Points);
        }

        public void SetSize(int a_Width,
                            int a_Height)
        {
            this.Size = new Size(a_Width, a_Height);
        }

        public void SetSizePlusBottomPadding(int a_Width,
                                             int a_Height)
        {
            SetSize(a_Width, a_Height + this.Padding.Bottom);
        }

        public int GetRight()
        {
            return (this.Right + Lyt.s_Space);
        }

        public int GetBottom()
        {
            return (this.Bottom + Lyt.s_Space);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public class MouseTimer : IDisposable
        {
            private bool f_Enabled;
            private readonly GuiTimer f_Timer;
            private FlexArray<GuiGroupBox> f_MousedBoxes;

            public bool p_Enabled
            {
                get { return f_Enabled; }
            }

            public MouseTimer()
            {
                f_Timer = new GuiTimer();
                FlexArray<GuiGroupBox>.Create(out f_MousedBoxes);

                const byte lc_IntervalMs = 50;
                f_Timer.Interval = lc_IntervalMs;
                f_Timer.e_AdvancedTick += Timer_AdvancedTick;
            }

            public void Dispose()
            {
                Stop();

                f_Timer.Dispose();
            }

            public void Start()
            {
                if (true == f_Enabled)
                {
                    return;
                }

                f_Enabled = true;
                f_Timer.Start();
            }

            public void Stop()
            {
                if (false == f_Enabled)
                {
                    return;
                }

                f_Enabled = false;
                f_Timer.Stop();

                if (0 < f_MousedBoxes.p_Count)
                {
                    FlexArray<GuiGroupBox> l_MousedBoxes;
                    FlexArray<GuiGroupBox>.Create(0, out l_MousedBoxes);
                    HandleMousedBoxes(ref l_MousedBoxes);
                }

                f_MousedBoxes.Clear(false);
            }

            public void AddManually(GuiGroupBox a_GroupBox)
            {
                // May only get called from within the GuiGroupBox
                // class.

                bool l_Added = TryAddToMousedBoxes(a_GroupBox);
            }

            private void Timer_AdvancedTick(GuiTimer a_Sender,
                                            uint a_Times)
            {
                // Find the topmost (modal) dialog.

                FormCollection l_Coll = Application.OpenForms;
                int l_Forms = l_Coll.Count;

                if (0 >= l_Forms)
                {
                    return;
                }

                Control l_Cntrl = l_Coll[l_Forms - 1];
                Type l_CntrlType = l_Cntrl.GetType();

                if (false == l_CntrlType.IsSubclassOf(typeof(Dialog)))
                {
                    return;
                }

                // Find all group boxes that the mouse is on.

                // This is an arbitrary limit.
                const uint lc_RecursionLimit = 48;
                Point l_ScreenPoint = Form.MousePosition;
                Point l_ClientPoint = l_Cntrl.PointToClient(l_ScreenPoint);

                FlexArray<GuiGroupBox> l_MousedBoxes;
                FlexArray<GuiGroupBox>.Create(out l_MousedBoxes);

                for (uint i = 0; lc_RecursionLimit != i; i++)
                {
                    const GetChildAtPointSkip l_SkipFlags = (GetChildAtPointSkip.Invisible | GetChildAtPointSkip.Disabled);
                    Control l_ChildCntrl = l_Cntrl.GetChildAtPoint(l_ClientPoint, l_SkipFlags);

                    if (null == l_ChildCntrl)
                    {
                        break;
                    }

                    l_Cntrl = l_ChildCntrl;
                    l_CntrlType = l_Cntrl.GetType();
                    l_ClientPoint = l_Cntrl.PointToClient(l_ScreenPoint);

                    if (typeof(GuiGroupBox) != l_CntrlType &&
                        false == l_CntrlType.IsSubclassOf(typeof(GuiGroupBox)))
                    {
                        // This control is not relevant.

                        continue;
                    }

                    GuiGroupBox l_GroupBox = (GuiGroupBox)l_Cntrl;

                    if (false == l_GroupBox.p_EnableMouseOverEffect ||
                        false == l_GroupBox.BoxHitTest(l_ClientPoint))
                    {
                        // The group box has the feature disabled or
                        // the mouse is not inside the group box frame.

                        continue;
                    }

                    int l_Idx = l_MousedBoxes.ReserveAtEnd();
                    l_MousedBoxes.p_Arr[l_Idx] = l_GroupBox;
                }

                HandleMousedBoxes(ref l_MousedBoxes);
            }

            private void HandleMousedBoxes(ref FlexArray<GuiGroupBox> a_MousedBoxes)
            {
                // a_MousedBoxes may contain duplicates.

                int l_NewCount = a_MousedBoxes.p_Count;
                GuiGroupBox[] l_OldArr = f_MousedBoxes.p_Arr;
                GuiGroupBox[] l_NewArr = a_MousedBoxes.p_Arr;

                // Disable the moused state.

                for (int i = f_MousedBoxes.p_Count; 0 != i--; )
                {
                    GuiGroupBox l_CurBox = l_OldArr[i];

                    for (int k = l_NewCount; 0 != k--; )
                    {
                        if (l_CurBox == l_NewArr[k])
                        {
                            goto CONTINUE;
                        }
                    }

                    l_CurBox.AdjustMousedFlag(false);
                    f_MousedBoxes.DeleteAt(i);

                    CONTINUE:;
                }

                // Enable the moused state.

                for (int k = l_NewCount; 0 != k--; )
                {
                    bool l_Added = TryAddToMousedBoxes(l_NewArr[k]);
                }
            }

            private bool TryAddToMousedBoxes(GuiGroupBox a_GroupBox)
            {
                GroupBox[] l_Arr = f_MousedBoxes.p_Arr;

                for (int i = f_MousedBoxes.p_Count; 0 != i--; )
                {
                    if (a_GroupBox == l_Arr[i])
                    {
                        // Item already exists.

                        return false;
                    }
                }

                a_GroupBox.AdjustMousedFlag(true);
                int l_Idx = f_MousedBoxes.ReserveAtEnd();
                f_MousedBoxes.p_Arr[l_Idx] = a_GroupBox;

                return true;
            }
        }
    }

    class RadioButtonBattery : GuiGroupBox
    {
        private readonly GuiRadioButton[] f_Buttons;

        public RadioButtonBattery(Control a_Owner,
                                  int a_Width,
                                  int a_Count) :
                             base(a_Owner)
        {
            f_Buttons = new GuiRadioButton[a_Count];

            int l_Left = this.Padding.Left;
            int l_Top = this.Padding.Top;
            int l_MaxWidth = (a_Width - this.Padding.Horizontal);
            for (int i = 0; i < a_Count; i++)
            {
                f_Buttons[i] = new GuiRadioButton(this, l_MaxWidth);
                f_Buttons[i].Location = new Point(l_Left, l_Top);

                l_Top = f_Buttons[i].GetBottom();
            }

            if (0 < a_Count)
            {
                l_Top -= Lyt.s_SmallSpace;
            }

            this.Size = new Size(a_Width, l_Top + this.Padding.Bottom);

            f_Buttons[0].Checked = true;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                foreach (GuiRadioButton l_Button in f_Buttons)
                {
                    l_Button.Dispose();
                }
            }

            base.Dispose(disposing);
        }

        public void SetTexts(TextSet a_TextSet,
                             TextKeys a_CaptionKey,
                             params TextKeys[] a_ButtonTextKeys)
        {
            this.Text = a_TextSet[a_CaptionKey];

            for (int i = (f_Buttons.Length - 1); 0 <= i; i--)
            {
                f_Buttons[i].SetTextAndUpdateSize(a_TextSet[a_ButtonTextKeys[i]]);
            }
        }

        public void SetTexts(string a_CaptionText,
                             params string[] a_ButtonText)
        {
            this.Text = a_CaptionText;

            for (int i = (f_Buttons.Length - 1); 0 <= i; i--)
            {
                f_Buttons[i].SetTextAndUpdateSize(a_ButtonText[i]);
            }
        }

        public int GetSelectedIndex()
        {
            for (int i = (f_Buttons.Length - 1); 0 <= i; i--)
            {
                if (true == f_Buttons[i].Checked)
                {
                    return i;
                }
            }

            return (-1);
        }

        public void SetSelectedIndex(int a_Index)
        {
            f_Buttons[a_Index].Checked = true;
        }

        public void SetSelectedIndexFromFlag(bool a_Flag)
        {
            f_Buttons[false == a_Flag ? 0 : 1].Checked = true;
        }
    }

    abstract class LabeledControlBase<T> : IDisposable
                                where T : Control
    {
        private static readonly int s_LabelExtraWidth;

        private readonly int f_MaxBoxWidth;
        protected readonly GuiLabel f_Label;
        public readonly T f_Box;

        static LabeledControlBase()
        {
            s_LabelExtraWidth = Lyt.CalcX(2);
        }

        protected LabeledControlBase(Control a_Owner,
                                     T a_Box,
                                     int a_Width,
                                     int a_MaxBoxWidth)
        {
            f_MaxBoxWidth = a_MaxBoxWidth;
            f_Label = new GuiLabel(a_Owner);
            f_Box = a_Box;

            a_Owner.Controls.Add(f_Box);

            SetWidthBase(a_Width);
        }

        public virtual void Dispose()
        {
            f_Label.Dispose();
            f_Box.Dispose();
        }

        public void SetLocation(Point a_Pnt)
        {
            SetLocation(a_Pnt.X, a_Pnt.Y);
        }

        public virtual void SetLocation(int a_x,
                                        int a_y)
        {
            SetLeft(a_x);
            SetTop(a_y);
        }

        public void SetLeft(int a_x)
        {
            // Display the label somewhat more to the left.
            f_Label.Left = (a_x - s_LabelExtraWidth);
            f_Box.Left = a_x;
        }

        public void SetTop(int a_y)
        {
            f_Label.Top = a_y;
            f_Box.Top = (f_Label.Bottom + Lyt.s_SmallSpace);
        }

        public virtual void SetWidth(int a_Width)
        {
            SetWidthBase(a_Width);
        }

        private void SetWidthBase(int a_Width)
        {
            f_Label.Width = (a_Width + s_LabelExtraWidth);
            f_Box.Width = Math.Min(a_Width, f_MaxBoxWidth);
        }

        public int GetRight()
        {
            return (f_Label.Right + Lyt.s_Space);
        }

        public int GetBottom()
        {
            return (f_Box.Bottom + Lyt.s_Space);
        }

        public void SetCaption(TextSet a_TextSet,
                               TextKeys a_LabelTextKey)
        {
            SetCaption(a_TextSet.GetAsInputControlCaption(a_LabelTextKey));
        }

        public void SetCaption(string a_Text)
        {
            // This routine doesn't add ':'.

            f_Label.Text = a_Text;
        }
    }

    class LabeledComboBox : LabeledControlBase<ComboBox>
    {
        // NOTE: ComboBox items don't need disposal.

        public LabeledComboBox(Control a_Owner,
                               int a_TotalWidth,
                               int a_MaxComboWidth,
                               bool a_CustomText) :
                          base(a_Owner, new ComboBox(), a_TotalWidth, a_MaxComboWidth)
        {
            if (false == a_CustomText)
            {
                f_Box.DropDownStyle = ComboBoxStyle.DropDownList;
            }

            f_Box.MaxLength = LabeledTextBox.c_DefaultMaxChars;
        }

        public LabeledComboBox(Control a_Owner,
                               int a_Width,
                               bool a_CustomText) :
                          this(a_Owner, a_Width, int.MaxValue, a_CustomText)
        {
        }

        public void ClearText()
        {
            f_Box.Text = string.Empty;
        }

        public void SetTexts(TextSet a_TextSet,
                             TextKeys a_LabelTextKey,
                             string[] a_ComboItemTexts)
        {
            SetCaption(a_TextSet, a_LabelTextKey);
            SetTexts(a_ComboItemTexts);
        }

        public void SetTexts<T>(T[] a_ComboItems)
        {
            f_Box.BeginUpdate();
            {
                f_Box.Items.Clear();

                foreach (T l_Obj in a_ComboItems)
                {
                    // Do not insert strings (ToString()), but insert
                    // the actual object.

                    f_Box.Items.Add(l_Obj);
                }
            }
            f_Box.EndUpdate();
        }

        public void SetFocus()
        {
            SetFocus(true);
        }

        public void SetFocus(bool a_DropDown)
        {
            f_Box.Select();

            if (false == a_DropDown)
            {
                f_Box.Select(f_Box.Text.Length, 0);
            }
            else
            {
                f_Box.DroppedDown = true;
            }
        }
    }

    sealed class LabeledIconComboBox<TTag> : LabeledComboBox
    {
        public LabeledIconComboBox(Control a_Owner,
                                   int a_TotalWidth,
                                   int a_MaxComboWidth) :
                              base(a_Owner, a_TotalWidth, a_MaxComboWidth, false)
        {
            f_Box.DrawMode = DrawMode.OwnerDrawFixed;
            f_Box.DrawItem += Box_DrawItem;
        }

        private void Box_DrawItem(object sender, DrawItemEventArgs e)
        {
            e.DrawBackground();

            // The index can be negative.
            int l_ItemIndex = e.Index;

            if (0 <= l_ItemIndex)
            {
                Item l_Item = GetItem(l_ItemIndex);

                // Draw icon.

                if (null != l_Item.f_Image)
                {
                    Rectangle l_SrcRect = new Rectangle(Point.Empty, l_Item.f_Image.Size);
                    Rectangle l_DestRect = e.Bounds;
                    l_DestRect.Width = Math.Min(l_DestRect.Width, l_DestRect.Height);
                    e.Graphics.DrawImage(l_Item.f_Image, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
                }

                // Draw text.

                Point l_TextPos = e.Bounds.Location;
                l_TextPos.X += (e.Bounds.Height + Lyt.s_SmallSpace);
                l_TextPos.Y += ((e.Bounds.Height - e.Font.Height) / 2);

                Brush l_Brush = new SolidBrush(e.ForeColor);
                e.Graphics.DrawString(l_Item.f_Text, e.Font, l_Brush, l_TextPos);
                l_Brush.Dispose();
            }

            e.DrawFocusRectangle();
        }

        public Item GetItem(int a_Idx)
        {
            return (Item)f_Box.Items[a_Idx];
        }

        public sealed class Item
        {
            // This class does not own f_Image and won't dispose it.

            public readonly string f_Text;
            // f_Image can be null.
            public readonly Image f_Image;

            public TTag f_Tag;

            public Item(string a_Text,
                        Image a_Image,
                        TTag a_Tag)
            {
                // a_Image and f_Tag may be null.

                f_Text = a_Text;
                f_Image = a_Image;
                f_Tag = a_Tag;
            }

            public override string ToString()
            {
                return f_Text;
            }
        }
    }

    sealed class LabeledTextBox : LabeledControlBase<TextBox>
    {
        public const ushort c_DefaultMaxChars = 1024;

        public LabeledTextBox(Control a_Owner,
                              int a_Width) :
                         this(a_Owner, a_Width, false)
        {
        }

        public LabeledTextBox(Control a_Owner,
                              int a_Width,
                              bool a_MultiLine) :
                         base(a_Owner, new TextBox(), a_Width, int.MaxValue)
        {
            f_Box.Width = a_Width;
            f_Box.MaxLength = c_DefaultMaxChars;
            f_Box.Multiline = a_MultiLine;
        }

        public void SetFocus()
        {
            f_Box.Select();
            f_Box.SelectAll();
        }

        public void SetMaxChars(uint a_MaxChars)
        {
            f_Box.MaxLength = (int)Math.Min(int.MaxValue, a_MaxChars);
        }

        public string GetBoxTextTrimmedAndNormalized()
        {
            return Util.UnicodeNormalizeString(f_Box.Text.Trim());
        }
    }

    class LabeledNumberBoxWithUnit : LabeledNumberBox
    {
        protected readonly GuiLabel f_UnitLabel;

        public LabeledNumberBoxWithUnit(Control a_Owner,
                                        int a_Width) :
                                   this(a_Owner, a_Width, LabeledNumberBox.s_DefMaxTextBoxWidth)
        {
        }

        public LabeledNumberBoxWithUnit(Control a_Owner,
                                        int a_Width,
                                        int a_MaxBoxWidth) :
                                   base(a_Owner, a_Width, a_MaxBoxWidth)
        {
            f_UnitLabel = new GuiLabel(a_Owner);
            f_UnitLabel.Height = f_Box.Height;
            f_UnitLabel.TextAlign = ContentAlignment.MiddleLeft;
        }

        public override void Dispose()
        {
            f_UnitLabel.Dispose();

            base.Dispose();
        }

        public override void SetLocation(int a_x,
                                         int a_y)
        {
            base.SetLocation(a_x, a_y);

            f_UnitLabel.Location = new Point(f_Box.Right + Lyt.s_SmallSpace, f_Box.Top);
            f_UnitLabel.Width = (f_Label.Right - f_UnitLabel.Left);
        }

        public void SetUnitText(TextSet a_TextSet,
                                TextKeys a_TextKey)
        {
            f_UnitLabel.Text = a_TextSet[a_TextKey];
        }
    }

    class LabeledNumberBox : LabeledControlBase<NumericUpDown>
    {
        public static readonly ushort s_DefMaxTextBoxWidth;

        public int p_ValueAsInt
        {
            get { return (int)f_Box.Value; }
        }

        public uint p_ValueAsUInt
        {
            get { return (uint)f_Box.Value; }
        }

        static LabeledNumberBox()
        {
            s_DefMaxTextBoxWidth = (ushort)Lyt.CalcX(75);
        }

        public LabeledNumberBox(Control a_Owner,
                                int a_Width) :
                           this(a_Owner, a_Width, s_DefMaxTextBoxWidth)
        {
        }

        public LabeledNumberBox(Control a_Owner,
                                int a_Width,
                                int a_MaxBoxWidth) :
                           base(a_Owner, new NumericUpDown(), a_Width, a_MaxBoxWidth)
        {
        }

        public void SetIPPortLimits()
        {
            SetLimits(System.Net.IPEndPoint.MinPort, System.Net.IPEndPoint.MaxPort);
        }

        public void SetLimits(int a_Min,
                              int a_Max)
        {
            f_Box.Minimum = a_Min;
            f_Box.Maximum = a_Max;
        }

        public void SetValueClipped(decimal a_Value)
        {
            if (f_Box.Minimum > a_Value)
            {
                f_Box.Value = f_Box.Minimum;
            }
            else if (f_Box.Maximum < a_Value)
            {
                f_Box.Value = f_Box.Maximum;
            }
            else
            {
                f_Box.Value = a_Value;
            }
        }
    }

    class GenericMenuItem<T> : ToolStripMenuItem
    {
        public T f_GenericTag;

        public GenericMenuItem(string a_Text,
                               EventHandler a_ClickHandler) :
                          base(a_Text, null, a_ClickHandler)
        {
        }

        public static GenericMenuItem<T> CreateItem(ContextMenuStrip a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    T a_GenericTag)
        {
            return CreateItem(a_Parent, a_Text, a_ClickHandler, null, a_GenericTag);
        }

        public static GenericMenuItem<T> CreateItem(ContextMenuStrip a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    IconTypes a_IconType,
                                                    T a_GenericTag)
        {
            return CreateItem(a_Parent, a_Text, a_ClickHandler, Software.s_IconStorage.GetDefaultIcon(a_IconType), a_GenericTag);
        }

        public static GenericMenuItem<T> CreateItem(ContextMenuStrip a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    Image a_Image,
                                                    T a_GenericTag)
        {
            GenericMenuItem<T> l_NewItem = new GenericMenuItem<T>(a_Text, a_ClickHandler);
            l_NewItem.f_GenericTag = a_GenericTag;
            l_NewItem.Image = a_Image;
            a_Parent.Items.Add(l_NewItem);

            return l_NewItem;
        }

        public static GenericMenuItem<T> CreateSubItem(ToolStripMenuItem a_Parent,
                                                       string a_Text,
                                                       EventHandler a_ClickHandler,
                                                       T a_GenericTag)
        {
            return CreateSubItem(a_Parent, a_Text, a_ClickHandler, null, a_GenericTag);
        }

        public static GenericMenuItem<T> CreateSubItem(ToolStripMenuItem a_Parent,
                                                       string a_Text,
                                                       EventHandler a_ClickHandler,
                                                       IconTypes a_IconType,
                                                       T a_GenericTag)
        {
            return CreateSubItem(a_Parent, a_Text, a_ClickHandler, Software.s_IconStorage.GetDefaultIcon(a_IconType), a_GenericTag);
        }

        public static GenericMenuItem<T> CreateSubItem(ToolStripMenuItem a_Parent,
                                                       string a_Text,
                                                       EventHandler a_ClickHandler,
                                                       Image a_Image,
                                                       T a_GenericTag)
        {
            GenericMenuItem<T> l_NewItem = new GenericMenuItem<T>(a_Text, a_ClickHandler);
            l_NewItem.f_GenericTag = a_GenericTag;
            l_NewItem.Image = a_Image;
            a_Parent.DropDownItems.Add(l_NewItem);

            return l_NewItem;
        }

        public static void AddLinkMenu(TextSet a_TextSet,
                                       ContextMenuStrip a_DestMenu,
                                       bool a_Enable,
                                       EventHandler a_Click)
        {
            ToolStripMenuItem l_Links = MainForm.AddMenuItem(a_DestMenu, a_TextSet[TextKeys.FilesTabListCntxtMenuLinks], null, IconTypes.Key);

            if (false == a_Enable)
            {
                l_Links.Enabled = false;
            }
            else
            {
                GenericMenuItem<FileLink.CopyLinkToClipboardOps>.CreateSubItem(l_Links, a_TextSet[TextKeys.FilesTabListCntxtMenuCopyeD2kLink], a_Click, FileLink.CopyLinkToClipboardOps.eD2kUri);
                GenericMenuItem<FileLink.CopyLinkToClipboardOps>.CreateSubItem(l_Links, a_TextSet[TextKeys.FilesTabListCntxtMenuCopyeD2kLinkPlusHtml], a_Click, FileLink.CopyLinkToClipboardOps.eD2kUriPlusHtml);
                GenericMenuItem<FileLink.CopyLinkToClipboardOps>.CreateSubItem(l_Links, a_TextSet[TextKeys.FilesTabListCntxtMenuCopyMagnetLink], a_Click, FileLink.CopyLinkToClipboardOps.MagnetUri);
                GenericMenuItem<FileLink.CopyLinkToClipboardOps>.CreateSubItem(l_Links, a_TextSet[TextKeys.FilesTabListCntxtMenuCopyMagnetLinkPlusHtml], a_Click, FileLink.CopyLinkToClipboardOps.MagnetUriPlusHtml);
            }
        }
    }

    sealed class ContextMenuStripEx : ContextMenuStrip
    {
        // f_TabControl can be null.
        private readonly AdvancedTabControl f_TabControl;
        private uint f_ClickedTabPageIndex;

        public int p_ClickedTabPageIndex
        {
            get { return (int)f_ClickedTabPageIndex; }
        }

        private ContextMenuStripEx(bool a_AddDummyItem)
        {
            f_ClickedTabPageIndex = uint.MaxValue;

            if (true == a_AddDummyItem)
            {
                // We need to insert a dummy item, otherwise the menu
                // won't be shown the first time it's opened.
                MainForm.AddMenuSeparator(this);
            }

            this.ImageScalingSize = Lyt.s_SmallIconSize;
        }

        public ContextMenuStripEx(AdvancedTabControl a_TabControl,
                                  bool a_AddDummyItem) :
            this(a_AddDummyItem)
        {
            f_TabControl = a_TabControl;
            f_TabControl.p_TabContextMenuStrip = this;
        }

        public ContextMenuStripEx(Control a_Cntrl,
                                  bool a_AddDummyItem) :
            this(a_AddDummyItem)
        {
            a_Cntrl.ContextMenuStrip = this;
        }

        public static ContextMenuStripEx CreateUnassociated(bool a_AddDummyItem)
        {
            return new ContextMenuStripEx(a_AddDummyItem);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                DisposeItems();
            }

            base.Dispose(disposing);
        }

        public void DisposeItems()
        {
            MainForm.DisposeMenuItems(this);
        }

        protected override void OnOpening(System.ComponentModel.CancelEventArgs e)
        {
            if (null != f_TabControl)
            {
                f_ClickedTabPageIndex = f_TabControl.p_SelectedPageIndex;
            }

            base.OnOpening(e);
        }
    }

    class GuiAutoPanel : Panel
    {
        // Since the original auto scroll feature of .NET doesn't
        // seem to work correctly (there is no way of hiding the
        // horizontal scroll bar as mentioned in the documentation
        // using HScroll/HorizontalScroll.Visible), this is a working
        // approach that introduces a vertical scroll bar alone.

        private readonly Panel f_ChildPnl;
        private readonly VScrollBar f_VScrollBar;
        public readonly int f_ScrollBarWidth;
        private readonly bool f_OverrideControlMember;

        private bool f_VertAutoScroll;
        private int f_PresetOccupiedHeight;
        private bool f_VertScrollBarIsVisible;

        protected int p_ScrollBarWidthInclSpace
        {
            get { return (f_ScrollBarWidth + Lyt.s_Space); }
        }

        public bool p_VertScrollBarIsVisible
        {
            get { return f_VertScrollBarIsVisible; }
        }

        public int p_ScrollPosition
        {
            get { return f_VScrollBar.Value; }
            set { ExternScrollTo(value); }
        }

        public GuiAutoPanel(bool a_InitiallyShowScrollBar)
        {
            f_ChildPnl = new Panel();
            f_VScrollBar = new VScrollBar();

            this.Controls.Add(f_ChildPnl);
            this.Controls.Add(f_VScrollBar);

            f_ScrollBarWidth = f_VScrollBar.Width;
            f_OverrideControlMember = true;
            f_VertAutoScroll = a_InitiallyShowScrollBar;

            f_VScrollBar.Scroll += new ScrollEventHandler(VScrollBar_Scroll);

            UpdateSize();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                DisposeAutoPanelChildren();

                f_ChildPnl.Dispose();
                f_VScrollBar.Dispose();
            }

            base.Dispose(disposing);
        }

        protected virtual void DisposeAutoPanelChildren()
        {
            // This is optional. Child classes might need it.
        }

        protected void DisposeSubClassCreatedChildren()
        {
            // Sub classes might use this from DisposeAutoPanelChildren().

            Dialog.DisposeAllChildren(f_ChildPnl);
        }

        public int GetActualControlCount()
        {
            return f_ChildPnl.Controls.Count;
        }

        public bool ExpectScrollBar(int a_OccupiedHeight)
        {
            return (true == f_VertAutoScroll &&
                    a_OccupiedHeight > this.Height);
        }

        private int GetOccupiedHeight()
        {
            return (0 >= f_PresetOccupiedHeight ?
                    CalcOccupiedHeight() : f_PresetOccupiedHeight);
        }

        private void UpdateSize()
        {
            this.SuspendLayout();

            bool l_ScrollBarRequired = false;

            if (true == f_VertAutoScroll)
            {
                int l_OccupiedHeight = GetOccupiedHeight();

                if (l_OccupiedHeight > this.Height)
                {
                    // In this case (since we'll have a scroll bar)
                    // we add some extra space.
                    l_OccupiedHeight += Lyt.s_SmallSpace;

                    l_ScrollBarRequired = true;

                    int l_ScrollViews = (l_OccupiedHeight - this.Height + 1);
                    int l_NewScrollPos = Math.Min(f_VScrollBar.Value, l_ScrollViews - 1);

                    f_VertScrollBarIsVisible = true;
                    f_VScrollBar.Visible = true;
                    int l_Left = Math.Max(0, this.Width - f_ScrollBarWidth);
                    f_VScrollBar.Location = new Point(l_Left, 0);
                    f_VScrollBar.Height = this.Height;

                    SetScrollBarViews(f_VScrollBar, l_ScrollViews);
                    f_VScrollBar.Value = l_NewScrollPos;
                    ScrollTo(l_NewScrollPos);
                }
            }

            if (false == l_ScrollBarRequired)
            {
                f_VertScrollBarIsVisible = false;
                f_VScrollBar.Visible = false;
                f_VScrollBar.Value = 0;

                f_ChildPnl.Location = Point.Empty;
                f_ChildPnl.Size = this.Size;
            }

            this.ResumeLayout();
        }

        public void ChildBoundsUpdated()
        {
            f_PresetOccupiedHeight = 0;

            UpdateSize();
        }

        public void ChildSizeUpdated(int a_PresetOccupiedHeight)
        {
            f_PresetOccupiedHeight = a_PresetOccupiedHeight;

            UpdateSize();
        }

        private int CalcOccupiedHeight()
        {
            int l_OccupiedHeight = 0;

            foreach (Control CurCntrl in f_ChildPnl.Controls)
            {
                if (DockStyle.Fill != CurCntrl.Dock)
                {
                    l_OccupiedHeight = Math.Max(l_OccupiedHeight, CurCntrl.Bottom);
                }
            }

            return l_OccupiedHeight;
        }

        public void ExternScrollTo(int a_Y)
        {
            // a_Y may exceed the maximum scroll position.

            int l_Y = Math.Min(a_Y, f_VScrollBar.Maximum);
            f_VScrollBar.Value = l_Y;
            ScrollTo(l_Y);
        }

        private void ScrollTo(int a_Y)
        {
            f_ChildPnl.Location = new Point(0, -a_Y);
            f_ChildPnl.Size = new Size(this.Width - f_VScrollBar.Width, this.Height + a_Y);
        }

        protected void ScrollToStart()
        {
            if (true == f_VertAutoScroll)
            {
                f_VScrollBar.Value = 0;
                ScrollTo(0);
            }
        }

        private void VScrollBar_Scroll(object sender, ScrollEventArgs e)
        {
            if (e.NewValue != e.OldValue)
            {
                ScrollTo(e.NewValue);
            }
        }

        protected override void OnResize(EventArgs eventargs)
        {
            UpdateSize();

            base.OnResize(eventargs);
        }

        protected override void OnControlAdded(ControlEventArgs e)
        {
            // Since the Controls member is not virtual, we can't
            // override it (and hiding it doesn't do the trick),
            // so we use an event to reroute the added control.

            if (false == f_OverrideControlMember)
            {
                base.OnControlAdded(e);
            }
            else
            {
                // NOTE: Adding a control to another control also
                // removes that child from the previous parent.
                f_ChildPnl.Controls.Add(e.Control);
            }
        }

        public override bool AutoScroll
        {
            // Don't use AutoScroll.
            set { throw (new ApplicationException()); }
        }

        public void AdjustVertAutoScroll(bool a_State)
        {
            if (a_State != f_VertAutoScroll)
            {
                f_VertAutoScroll = a_State;

                UpdateSize();
            }
        }

        public static int GetScrollBarViews(ScrollBar a_ScrollBar)
        {
            return checked(a_ScrollBar.Maximum - a_ScrollBar.LargeChange + 2);
        }

        public static void SetScrollBarViews(ScrollBar a_ScrollBar,
                                             int a_ViewCount)
        {
            // This routine tries to hide away several tricky issues
            // that may occur when dealing with scroll bars.

            int l_MaxScrollValue = checked(a_ViewCount - 2 + a_ScrollBar.LargeChange);

            if (l_MaxScrollValue != a_ScrollBar.Maximum)
            {
                a_ScrollBar.Maximum = l_MaxScrollValue;

                // For some high LargeChange values Maximum won't be
                // set correctly.
                if (l_MaxScrollValue != a_ScrollBar.Maximum)
                {
                    throw (new ApplicationException());
                }
            }
        }
    }

    class HorSplitContainer : Panel
    {
        private const byte c_DefaultSplitterHeight = 4;
        private const byte c_MinUnlockedSplitterHeight = 1;

        private static readonly ushort s_DefaultSplitterHeight;

        // NOTE: Both controls are not supposed to exist right after
        // construction.
        private Control f_UpperCntrl;
        private Control f_LowerCntrl;

        private int f_SplitterHeight;
        private bool f_UpdateWhileMoving;
        private int f_MinUpperHeight;
        private int f_MinLowerHeight;
        // f_FixedPanel may never be FixedPanel.None.
        private FixedPanel f_FixedPanel;

        private bool f_Locked;
        private bool f_Moving;
        private bool f_SplitterMoused;

        private int f_PreferredHeight;

        public int p_SplitterHeight
        {
            get { return f_SplitterHeight; }
            set { ChangeParam(ref f_SplitterHeight, false == f_Locked ? (int)c_MinUnlockedSplitterHeight : 0, value); }
        }

        public bool p_UpdateWhileMoving
        {
            get { return f_UpdateWhileMoving; }
            set { f_UpdateWhileMoving = value; }
        }

        public bool p_Locked
        {
            get { return f_Locked; }
            set
            {
                if (value != f_Locked)
                {
                    f_Locked = value;

                    if (false == f_Locked)
                    {
                        f_SplitterHeight = Math.Max(f_SplitterHeight, c_MinUnlockedSplitterHeight);
                    }
                    else
                    {
                        if (true == f_Moving)
                        {
                            EndMove();
                        }

                        ChangeSplitterMousedState(false);
                    }
                }
            }
        }

        public int p_MinUpperHeight
        {
            get { return f_MinUpperHeight; }
            set { ChangeParam(ref f_MinUpperHeight, 0, value); }
        }

        public int p_MinLowerHeight
        {
            get { return f_MinLowerHeight; }
            set { ChangeParam(ref f_MinLowerHeight, 0, value); }
        }

        public FixedPanel p_FixedPanel
        {
            get { return f_FixedPanel; }
            set
            {
                if (value != f_FixedPanel)
                {
                    f_FixedPanel = value;

                    ParamsUpdated();
                }
            }
        }

        static HorSplitContainer()
        {
            s_DefaultSplitterHeight = (ushort)Lyt.CalcY(c_DefaultSplitterHeight);
        }

        public HorSplitContainer(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            f_SplitterHeight = s_DefaultSplitterHeight;
            f_FixedPanel = FixedPanel.Panel1;

            this.BackColor = Color.Transparent;
        }

        public static HorSplitContainer CreateFixed(Control a_Owner)
        {
            HorSplitContainer l_Splitter = new HorSplitContainer(a_Owner);
            l_Splitter.p_Locked = true;
            l_Splitter.p_SplitterHeight = 0;
            l_Splitter.Dock = DockStyle.Fill;

            return l_Splitter;
        }

        public void SetControls(Control a_UpperControl,
                                Control a_LowerControl)
        {
            f_UpperCntrl = a_UpperControl;
            f_LowerCntrl = a_LowerControl;

            f_UpperCntrl.Location = new Point();
            f_LowerCntrl.Left = 0;

            UpdateSizes();
        }

        private void ChangeParam(ref int a_Value,
                                 int a_MinValue,
                                 int a_NewValue)
        {
            int l_Val = Math.Max(a_MinValue, a_NewValue);

            if (l_Val != a_Value)
            {
                a_Value = l_Val;

                ParamsUpdated();
            }
        }

        private void ParamsUpdated()
        {
            // The fixed panel, splitter height or a minimum size
            // changed.

            if (null != f_UpperCntrl)
            {
                // SetControls() was called already.

                UpdateHeights();
            }
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            // Need to call base class implementation.

            if (false == f_Locked &&
                false == f_Moving &&
                true == IsMouseOnSplitter(e.Location))
            {
                BeginMove();
            }

            base.OnMouseDown(e);
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            // Need to call base class implementation.

            if (true == f_Moving)
            {
                EndMove();

                SplitterMove(e.Y);
            }

            base.OnMouseUp(e);
        }

        private void BeginMove()
        {
            f_Moving = true;
            UpdateCursor();
        }

        private void EndMove()
        {
            f_Moving = false;
            UpdateCursor();
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            // Need to call base class implementation.

            if (false == f_Moving)
            {
                if (false == f_Locked)
                {
                    ChangeSplitterMousedState(IsMouseOnSplitter(e.Location));
                }
            }
            else if (true == f_UpdateWhileMoving)
            {
                SplitterMove(e.Y);
            }

            base.OnMouseMove(e);
        }

        protected override void OnMouseLeave(EventArgs e)
        {
            // Need to call base class implementation.

            ChangeSplitterMousedState(false);

            base.OnMouseLeave(e);
        }

        private bool IsMouseOnSplitter(Point a_Pos)
        {
            Rectangle l_Rect = new Rectangle(0, f_UpperCntrl.Height, this.Width, f_SplitterHeight);

            return l_Rect.Contains(a_Pos);
        }

        private void ChangeSplitterMousedState(bool a_Set)
        {
            if (a_Set != f_SplitterMoused)
            {
                f_SplitterMoused = a_Set;
                UpdateCursor();
            }
        }

        private void UpdateCursor()
        {
            bool l_Special = (true == f_Moving ||
                              true == f_SplitterMoused);
            this.Cursor = (false == l_Special ? Cursors.Default : Cursors.HSplit);
        }

        private void SplitterMove(int a_NewUpperHeight)
        {
            SetUpperHeight(a_NewUpperHeight);

            // Make the control redraw itself now.
            this.Update();
        }

        protected override void OnResize(EventArgs eventargs)
        {
            // Need to call base class implementation.

            UpdateSizes();

            base.OnResize(eventargs);
        }

        private int CalcDistributableHeight()
        {
            return (this.Height - f_SplitterHeight);
        }

        public void SetUpperHeight(int a_Height)
        {
            SetHeight(FixedPanel.Panel1, a_Height);
        }

        public void SetLowerHeight(int a_Height)
        {
            SetHeight(FixedPanel.Panel2, a_Height);
        }

        private void SetHeight(FixedPanel a_FixedPanel,
                               int a_Height)
        {
            if (a_FixedPanel == f_FixedPanel)
            {
                f_PreferredHeight = a_Height;
            }
            else
            {
                f_PreferredHeight = (CalcDistributableHeight() - a_Height);
            }

            UpdateHeights();
        }

        private void UpdateSizes()
        {
            UpdateHeights();

            int l_Width = this.Width;
            f_UpperCntrl.Width = l_Width;
            f_LowerCntrl.Width = l_Width;
        }

        private void UpdateHeights()
        {
            int l_UpperHeight;
            int l_LowerHeight;

            if (FixedPanel.Panel1 == f_FixedPanel)
            {
                l_UpperHeight = CalcFixedSize(f_MinUpperHeight, f_MinLowerHeight, out l_LowerHeight);
            }
            else
            {
                l_LowerHeight = CalcFixedSize(f_MinLowerHeight, f_MinUpperHeight, out l_UpperHeight);
            }

            f_UpperCntrl.Height = l_UpperHeight;
            f_LowerCntrl.Top = (l_UpperHeight + f_SplitterHeight);
            f_LowerCntrl.Height = l_LowerHeight;
        }

        private int CalcFixedSize(int a_PrimaryCntrlMin,
                                  int a_SecondaryCntrlMin,
                                  out int a_SeconderyCntrlHeight)
        {
            int l_HeightToDistribute = CalcDistributableHeight();
            int l_Remaining = (l_HeightToDistribute - a_PrimaryCntrlMin);

            if (0 > l_Remaining)
            {
                // Very low space.

                a_SeconderyCntrlHeight = 0;

                return l_HeightToDistribute;
            }

            if (l_Remaining <= a_SecondaryCntrlMin ||
                f_PreferredHeight < a_PrimaryCntrlMin)
            {
                a_SeconderyCntrlHeight = l_Remaining;

                return a_PrimaryCntrlMin;
            }

            // There's enough space available to honor all constraints.

            int l_PrimaryHeight = Math.Min(f_PreferredHeight, l_HeightToDistribute - a_SecondaryCntrlMin);
            a_SeconderyCntrlHeight = (l_HeightToDistribute - l_PrimaryHeight);

            return l_PrimaryHeight;
        }

        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            // Need to call base class implementation (if unhandled).

            if (false == f_Locked)
            {
                if (true == f_Moving)
                {
                    EndMove();
                }

                if (true == IsMouseOnSplitter(e.Location))
                {
                    if (FixedPanel.Panel1 == f_FixedPanel)
                    {
                        if (0 < f_MinUpperHeight)
                        {
                            SetUpperHeight(f_MinUpperHeight);

                            return;
                        }
                    }
                    else
                    {
                        if (0 < f_MinLowerHeight)
                        {
                            SetLowerHeight(f_MinLowerHeight);

                            return;
                        }
                    }
                }
            }

            base.OnMouseDoubleClick(e);
        }
    }

    interface IDialogController
    {
        bool IsAbortable();
        void Abort();
        // Returns true if the operation is still going on.
        bool GetProgress(out uint a_Percentage);
    }

    class UnabortableControllerContext : ControllerContext
    {
        public UnabortableControllerContext() :
                                       base(false)
        {
        }

        public void SetFinished()
        {
            base.SetFinishedBase();
        }
    }

    abstract class AbortableControllerContext : ControllerContext
    {
        private bool f_Aborted;

        public bool p_Aborted
        {
            get
            {
                return f_Aborted;
            }
        }

        public bool p_AbortedVolatile
        {
            get
            {
                return Util.VolatileRead(ref f_Aborted);
            }
        }

        public AbortableControllerContext() :
                                     base(true)
        {
        }

        public override void Abort()
        {
            Util.VolatileWrite(out f_Aborted, true);
        }
    }

    abstract class ControllerContext : IDialogController
    {
        protected readonly bool f_Abortable;

        private byte f_Percentage;
        private bool f_Finished;

        public bool p_Finished
        {
            get { return f_Finished; }
        }

        public bool p_FinishedVolatile
        {
            get { return Util.VolatileRead(ref f_Finished); }
        }

        public ControllerContext(bool a_Abortable)
        {
            f_Abortable = a_Abortable;
        }

        public bool IsAbortable()
        {
            return f_Abortable;
        }

        public virtual void Abort()
        {
        }

        public bool GetProgress(out uint a_Percentage)
        {
            a_Percentage = f_Percentage;
            bool l_NotFinished = !f_Finished;
            Util.MemoryBarrier();

            return l_NotFinished;
        }

        protected void SetPercentage(uint a_Percentage)
        {
            // It would be good to do this volatile, but we don't
            // due to performance reasons.
            f_Percentage = (byte)a_Percentage;
        }

        protected void SetPercentage(uint a_Done,
                                     uint a_Total)
        {
            // a_Total may be zero.

            uint l_Perc = (0 == a_Total ? 0 : ((a_Done * 100) / a_Total));
            SetPercentage(l_Perc);
        }

        protected void SetPercentage(ulong a_Done,
                                     ulong a_Total)
        {
            // a_Total may be zero. Same as above.

            uint l_Perc = (uint)(0 == a_Total ? 0 : ((a_Done * 100) / a_Total));
            SetPercentage(l_Perc);
        }

        protected void SetFinishedBase()
        {
            // May be called several times.

            Util.VolatileWrite(out f_Finished, true);
        }
    }

#region MessageBoxes

    class CustomMsgBox : Dialog
    {
        private const byte c_MaxWidthPerc = 60;
        private const float c_FontSizeIncrement = 1.5f;
        private const ushort c_TimerInterval = 50;
        private const DialogResult c_ControllerAbortDialogResult = DialogResult.Abort;

        private readonly static Size s_Spaces;
        private readonly static Bitmap s_MessageBoxIcons;

        private readonly string f_Message;
        // f_Icon can be null.
        private readonly Bitmap f_Icon;
        private readonly ButtonInfo[] f_Buttons;
        // f_AbortButtonIdx matters if the user presses escape or
        // if the control box is clicked. It can be negative.
        private readonly int f_AbortButtonIdx;
        private readonly GuiButton f_InitiallySelButton;
        private readonly System.Media.SystemSound f_Sound;
        private readonly Font f_MsgFont;

        private readonly Rectangle f_PicRect;
        private readonly Rectangle f_MsgRect;
        private readonly bool f_CenterMsgVertical;
        private readonly Panel f_ButtonPnl;

        private bool f_DialogResultIsTakenCareOf;

        // f_Controller and f_Timer can be null.
        private IDialogController f_Controller;
        private System.Windows.Forms.Timer f_Timer;

        static CustomMsgBox()
        {
            s_Spaces = new Size(Lyt.CalcX(25), Lyt.CalcY(25));

            System.IO.Stream l_Strm = Software.OpenResource("MsgBoxIcons.png", false);
            s_MessageBoxIcons = new Bitmap(l_Strm);
            l_Strm.Dispose();
        }

        private CustomMsgBox(string a_Caption,
                             string a_Message,
                             Bitmap a_Icon,
                             ButtonInfo[] a_Buttons,
                             int a_AbortButtonIdx,
                             DialogResult a_SelButton,
                             System.Media.SystemSound a_Sound)
        {
            // a_Icon may be null, otherwise this instance takes
            // ownership of it. a_AbortButtonIdx may be negative, if
            // the message is not abortable by the user. a_SelButton
            // may be an invalid value indicating the first button
            // should be selected.

            int l_ButtonCount = a_Buttons.Length;

            f_Message = a_Message;
            f_Icon = a_Icon;
            f_Buttons = a_Buttons;
            f_AbortButtonIdx = a_AbortButtonIdx;
            f_Sound = a_Sound;
            Font l_DlgFont = this.Font;
            f_MsgFont = MainForm.RecreateFont(l_DlgFont, l_DlgFont.SizeInPoints + c_FontSizeIncrement, l_DlgFont.Style);

            this.SuspendLayout();

            f_ButtonPnl = new Panel();

            this.Controls.Add(f_ButtonPnl);

            f_ButtonPnl.BackColor = this.BackColor;
            this.BackColor = SystemColors.Window;
            this.Text = a_Caption;

            Point l_LeftTop = new Point(s_Spaces.Width, s_Spaces.Height);

            if (null == a_Icon)
            {
                f_PicRect = new Rectangle(l_LeftTop, Size.Empty);
            }
            else
            {
                f_PicRect = new Rectangle(l_LeftTop, a_Icon.Size);
                l_LeftTop.X = (f_PicRect.Right + Lyt.s_SmallSpace);
            }

            Size l_WorkingAreaSz = base.GetWorkingArea();
            int l_MaxDlgWidth = ((l_WorkingAreaSz.Width / 100) * c_MaxWidthPerc);
            int l_WidthOfButtons = (((l_ButtonCount * Lyt.s_DefaultButtonWidth) + ((l_ButtonCount - 1) * Lyt.s_Space)) + Lyt.s_Space2);
            int l_MinTextWidth = (l_WidthOfButtons - l_LeftTop.X - s_Spaces.Width);
            int l_MaxTextWidth = (l_MaxDlgWidth - l_LeftTop.X - s_Spaces.Width);
            Size l_TextSize;

            Bitmap l_TempMap = new Bitmap(1, 1);
            Graphics l_G = Graphics.FromImage(l_TempMap);
            {
                SizeF l_SizeF = l_G.MeasureString(f_Message, f_MsgFont, l_MaxTextWidth);
                l_TextSize = new Size((int)l_SizeF.Width + 1, (int)l_SizeF.Height + 1);
            }
            l_G.Dispose();
            l_TempMap.Dispose();

            f_MsgRect = new Rectangle(l_LeftTop, Size.Empty);
            f_MsgRect.Width = Math.Max(l_MinTextWidth, Math.Min(l_MaxTextWidth, l_TextSize.Width));
            f_ButtonPnl.Width = (f_MsgRect.Right + s_Spaces.Width);

            int l_Left = (f_ButtonPnl.Width - l_WidthOfButtons + Lyt.s_Space);
            for (int i = 0; i < l_ButtonCount; i++)
            {
                f_Buttons[i].CreateButton(f_ButtonPnl, Button_Click);
                GuiButton l_Btn = f_Buttons[i].f_Button;
                l_Btn.Location = new Point(l_Left, Lyt.s_Space);
                l_Left += (Lyt.s_DefaultButtonWidth + Lyt.s_Space);

                if (i == f_AbortButtonIdx)
                {
                    this.CancelButton = l_Btn;
                }

                // Select the first button in case no other is
                // specified.
                if (0 == i ||
                    f_Buttons[i].f_Action == a_SelButton)
                {
                    f_InitiallySelButton = l_Btn;
                }
            }

            int l_MaxDlgHeight = (l_WorkingAreaSz.Height - Lyt.s_LargeSpace - (this.Height - this.ClientSize.Height));
            int l_HeightOfButtons = (f_Buttons[0].f_Button.Height + Lyt.s_Space2);
            int l_MinTextHeight = f_PicRect.Height;
            int l_MaxTextHeight = (l_MaxDlgHeight - l_LeftTop.Y - s_Spaces.Height - l_HeightOfButtons);
            f_MsgRect.Height = Math.Max(l_MinTextHeight, Math.Min(l_MaxTextHeight, l_TextSize.Height));

            if (l_MinTextHeight >= l_TextSize.Height)
            {
                f_CenterMsgVertical = true;
            }

            f_ButtonPnl.Top = (f_MsgRect.Bottom + s_Spaces.Height);
            f_ButtonPnl.Height = l_HeightOfButtons;

            this.AsFixedDialog(new Size(f_ButtonPnl.Right, f_ButtonPnl.Bottom));

            if (0 > f_AbortButtonIdx)
            {
                this.ControlBox = false;
            }

            this.KeyPreview = true;

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(f_ButtonPnl);
                Dialog.DisposeAllChildren(this);
                DisposeTimer();

                if (null != f_Icon)
                {
                    f_Icon.Dispose();
                }

                f_MsgFont.Dispose();
            }

            base.Dispose(disposing);
        }

        private void DisposeTimer()
        {
            if (null != f_Timer)
            {
                f_Timer.Dispose();
                f_Timer = null;
            }
        }

        private void SetDialogResultSafe(DialogResult a_DlgRes)
        {
            // f_DialogResultIsTakenCareOf must be set before the
            // property DialogResult is used, otherwise we might
            // run into problems on Mono.
            f_DialogResultIsTakenCareOf = true;
            this.DialogResult = a_DlgRes;
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            // Need to call base class implementation.

            if (true == e.Control &&
                (Keys.Insert == e.KeyCode ||
                 Keys.C == e.KeyCode))
            {
                CopyContentToClipboard();

                e.Handled = true;
            }

            base.OnKeyDown(e);
        }

        private void CopyContentToClipboard()
        {
            // This is the way the .NET framework does it.

            string l_Rule = new string('-', 27);
            string l_BtnSpacer = new string(' ', 3);

            StringBuilder l_Bld = new StringBuilder();

            l_Bld.AppendLine(l_Rule);
            l_Bld.AppendLine(this.Text);
            l_Bld.AppendLine(l_Rule);
            int l_Idx = l_Bld.Length;
            l_Bld.Append(f_Message);
            Util.TransformNewLineIntoPlatformLineBreak(l_Bld, l_Idx, f_Message.Length);
            l_Bld.AppendLine();
            l_Bld.AppendLine(l_Rule);

            for (int i = 0, lim = f_Buttons.Length; i < lim; i++)
            {
                if (0 < i)
                {
                    l_Bld.Append(l_BtnSpacer);
                }

                l_Bld.Append(f_Buttons[i].f_Text);
            }

            l_Bld.AppendLine();
            l_Bld.AppendLine(l_Rule);

            Util.SetClipboard(l_Bld.ToString());
        }

        protected override void OnClosed(EventArgs e)
        {
            // Need to call base class implementation.

            DisposeTimer();

            if (false == f_DialogResultIsTakenCareOf)
            {
                // We never expect to land here if f_AbortButtonIdx
                // is negative.

                SetDialogResultSafe(f_Buttons[f_AbortButtonIdx].f_Button.DialogResult);
            }

            base.OnClosed(e);
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            f_DialogResultIsTakenCareOf = false;

            if (null != f_Controller &&
                true == f_Controller.IsAbortable())
            {
                f_Timer = new System.Windows.Forms.Timer();
                f_Timer.Interval = c_TimerInterval;
                f_Timer.Tick += Timer_Tick;
                f_Timer.Start();
            }

            try
            {
                f_Sound.Play();
            }
            catch
            {
            }

            base.OnShown(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            // Need to call base class implementation.

            Graphics l_G = e.Graphics;
            System.Drawing.Text.TextRenderingHint l_Stored = l_G.TextRenderingHint;
            {
                if (null != f_Icon)
                {
                    l_G.DrawImage(f_Icon, f_PicRect);
                }

                StringFormat l_StrFrmt = (StringFormat)StringFormat.GenericDefault.Clone();
                {
                    if (true == f_CenterMsgVertical)
                    {
                        l_StrFrmt.LineAlignment = StringAlignment.Center;
                    }

                    l_G.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
                    l_G.DrawString(f_Message, f_MsgFont, SystemBrushes.ControlText, f_MsgRect, l_StrFrmt);
                }
                l_StrFrmt.Dispose();
            }
            l_G.TextRenderingHint = l_Stored;

            base.OnPaint(e);
        }

        private void Timer_Tick(object sender, EventArgs e)
        {
            // Returns true if the controller has aborted.

            uint l_Junk;
            if (true == f_Controller.GetProgress(out l_Junk))
            {
                return;
            }

            f_Timer.Stop();

            SetDialogResultSafe(c_ControllerAbortDialogResult);
            this.Close();
        }

        private void Button_Click(object sender, EventArgs e)
        {
            DialogResult l_DlgRes = (DialogResult)((Control)sender).Tag;
            SetDialogResultSafe(l_DlgRes);
            this.Close();
        }

        public DialogResult ShowMsgBox()
        {
            return this.ShowMsgBox(null, null);
        }

        public DialogResult ShowMsgBox(IWin32Window a_Owner)
        {
            return this.ShowMsgBox(a_Owner, null);
        }

        public DialogResult ShowMsgBox(IDialogController a_Controller)
        {
            return this.ShowMsgBox(null, a_Controller);
        }

        public DialogResult ShowMsgBox(IWin32Window a_Owner,
                                       IDialogController a_Controller)
        {
            // a_Owner and a_Controller may be null.

            f_InitiallySelButton.Select();

            DialogResult l_DlgRes;
            f_Controller = a_Controller;

            if (null == a_Owner)
            {
                l_DlgRes = this.ShowDialog();
            }
            else
            {
                l_DlgRes = this.ShowDialog(a_Owner);
            }

            f_Controller = null;

            return l_DlgRes;
        }

        public static DialogResult ShowMsgBox(TextSet a_TextSet,
                                              IWin32Window a_Owner,
                                              string a_Caption,
                                              string a_Message,
                                              MessageBoxIcon a_Icons,
                                              MessageBoxButtons a_Buttons,
                                              DialogResult a_SelButton)
        {
            // a_Owner may be null.

            using (CustomMsgBox l_MsgBox = Create(a_TextSet, a_Caption, a_Message, a_Icons, a_Buttons, a_SelButton))
            {
                return l_MsgBox.ShowMsgBox(a_Owner, null);
            }
        }

        public static DialogResult ShowAbortableMsgBox(TextSet a_TextSet,
                                                       IWin32Window a_Owner,
                                                       IDialogController a_Controller,
                                                       string a_Caption,
                                                       string a_Message,
                                                       MessageBoxIcon a_Icons,
                                                       MessageBoxButtons a_Buttons,
                                                       DialogResult a_SelButton)
        {
            // a_Owner may be null.

            uint l_Junk;
            if (false == a_Controller.GetProgress(out l_Junk))
            {
                return c_ControllerAbortDialogResult;
            }

            using (CustomMsgBox l_MsgBox = Create(a_TextSet, a_Caption, a_Message, a_Icons, a_Buttons, a_SelButton))
            {
                return l_MsgBox.ShowMsgBox(a_Owner, a_Controller);
            }
        }

        public static CustomMsgBox Create(TextSet a_TextSet,
                                          string a_Caption,
                                          string a_Message,
                                          MessageBoxIcon a_Icons,
                                          MessageBoxButtons a_Buttons,
                                          DialogResult a_SelButton)
        {
            // a_Owner may be null.

            ButtonInfo[] l_Buttons;
            int l_AbortButtonIdx;

            switch (a_Buttons)
            {
                case MessageBoxButtons.OK:
                    l_Buttons = new ButtonInfo[1] { new ButtonInfo(a_TextSet[TextKeys.ButtonTextOK], DialogResult.OK) };
                    l_AbortButtonIdx = 0;
                    break;

                case MessageBoxButtons.OKCancel:
                    l_Buttons = new ButtonInfo[2] { new ButtonInfo(a_TextSet[TextKeys.ButtonTextOK], DialogResult.OK), new ButtonInfo(a_TextSet[TextKeys.ButtonTextCancel], DialogResult.Cancel) };
                    l_AbortButtonIdx = 1;
                    break;

                case MessageBoxButtons.YesNo:
                    l_Buttons = new ButtonInfo[2] { new ButtonInfo(a_TextSet[TextKeys.ButtonTextYes], DialogResult.Yes), new ButtonInfo(a_TextSet[TextKeys.ButtonTextNo], DialogResult.No) };
                    // This message can't be skipped by the user.
                    l_AbortButtonIdx = (-1);
                    break;

                case MessageBoxButtons.YesNoCancel:
                    l_Buttons = new ButtonInfo[3] { new ButtonInfo(a_TextSet[TextKeys.ButtonTextYes], DialogResult.Yes), new ButtonInfo(a_TextSet[TextKeys.ButtonTextNo], DialogResult.No), new ButtonInfo(a_TextSet[TextKeys.ButtonTextCancel], DialogResult.Cancel) };
                    l_AbortButtonIdx = 2;
                    break;

                default: throw (new ApplicationException());
            }

            Size l_IconSize = new Size(Lyt.CalcX(43), Lyt.CalcY(43));
            System.Media.SystemSound l_Sound;
            Bitmap l_Icon = GetMessageBoxIcon(a_Icons, l_IconSize, out l_Sound);

            return new CustomMsgBox(a_Caption, a_Message, l_Icon, l_Buttons, l_AbortButtonIdx, a_SelButton, l_Sound);
        }

        private static Bitmap GetMessageBoxIcon(MessageBoxIcon a_Icon,
                                                Size a_Size,
                                                out System.Media.SystemSound a_Sound)
        {
            uint l_SrcIdx;

            switch (a_Icon)
            {
                case MessageBoxIcon.None:
                    a_Sound = System.Media.SystemSounds.Beep;
                    return null;

                case MessageBoxIcon.Error:
                    a_Sound = System.Media.SystemSounds.Hand;
                    l_SrcIdx = 0;
                    break;

                case MessageBoxIcon.Question:
                    a_Sound = System.Media.SystemSounds.Question;
                    l_SrcIdx = 1;
                    break;

                case MessageBoxIcon.Exclamation:
                    a_Sound = System.Media.SystemSounds.Exclamation;
                    l_SrcIdx = 2;
                    break;

                case MessageBoxIcon.Information:
                    a_Sound = System.Media.SystemSounds.Asterisk;
                    l_SrcIdx = 3;
                    break;

                default: throw (new ApplicationException());
            }

            const byte lc_OriginalIconSize = 64;
            Bitmap l_ResImg = new Bitmap(a_Size.Width, a_Size.Height);
            Graphics l_G = Graphics.FromImage(l_ResImg);
            {
                Rectangle l_SrcRect = new Rectangle((int)(l_SrcIdx * lc_OriginalIconSize), 0, lc_OriginalIconSize, lc_OriginalIconSize);
                Rectangle l_DestRect = new Rectangle(Point.Empty, a_Size);
                l_G.DrawImage(s_MessageBoxIcons, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
            }
            l_G.Dispose();

            return l_ResImg;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ButtonInfo
        {
            public readonly string f_Text;
            public readonly DialogResult f_Action;
            public GuiButton f_Button;

            public ButtonInfo(string a_Text,
                              DialogResult a_Action)
            {
                f_Text = a_Text;
                f_Action = a_Action;
                f_Button = null;
            }

            public void CreateButton(Control a_Owner,
                                     EventHandler a_ClickHandler)
            {
                f_Button = new GuiButton(a_Owner);
                // Do not use the Button.DialogResult property, since
                // Mono has issues (it executes handlers in a different
                // order than the .NET framework).
                f_Button.Tag = f_Action;
                f_Button.Click += a_ClickHandler;

                f_Button.Text = f_Text;
            }
        }
    }

#endregion

    class ReorderItemsDialog<TTag> : ButtonDialog
    {
        private const ushort c_DialogWidth = 400;
        private const ushort c_DialogHeight = 350;

        private readonly Instance f_Inst;

        private readonly GuiGroupBox f_GroupBox;
        private readonly ListView f_ListView;
        private readonly GuiButton f_ToTopButton;
        private readonly GuiButton f_UpButton;
        private readonly GuiButton f_DownButton;
        private readonly GuiButton f_ToBottomButton;

        private bool f_UseItemCheckHandler;

        // f_FixedItemCount indicates how many of the starting items
        // may not reorder.
        private int f_FixedItemCount;
        private ModifyItemVisibilityMode f_ModifyVisMode;
        private Item[] f_Arr;
        private bool f_VisibilityChanged;
        private int f_CheckedItemCount;
        private bool f_OrderChanged;

        public ReorderItemsDialog(Instance a_Inst, TextKeys a_TitleTextKey) :
            base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            f_Inst = a_Inst;

            this.SuspendLayout();
            base.AsFixedDialog(Lyt.TranformSize(new Size(c_DialogWidth, c_DialogHeight)));

            f_GroupBox = new GuiGroupBox(this);
            f_ListView = new ListView();
            f_ToTopButton = new GuiButton(f_GroupBox);
            f_UpButton = new GuiButton(f_GroupBox);
            f_DownButton = new GuiButton(f_GroupBox);
            f_ToBottomButton = new GuiButton(f_GroupBox);

            f_GroupBox.Controls.Add(f_ListView);

            TextSet l_TextSet = f_Inst.f_TextSet;
            f_GroupBox.Location = f_InnerRect.Location;
            f_GroupBox.Size = f_InnerRect.Size;
            int l_ListViewWidth = (f_InnerRect.Width - f_GroupBox.Padding.Horizontal - Lyt.s_Space - f_ToTopButton.Width);
            f_ListView.FullRowSelect = true;
            f_ListView.View = View.Details;
            f_ListView.HideSelection = false;
            f_ListView.Columns.Add(l_TextSet[TextKeys.ReorderItemsDlgListViewColCaption], l_ListViewWidth - Lyt.s_LargeSpace2);
            f_ListView.MultiSelect = true;
            f_ListView.Location = new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top);
            f_ListView.Size = new Size(l_ListViewWidth, f_InnerRect.Height - f_GroupBox.Padding.Vertical);
            int l_BtnLeft = (f_ListView.Right + Lyt.s_Space);
            f_ToTopButton.Location = new Point(l_BtnLeft, f_ListView.Top);
            f_UpButton.Location = new Point(l_BtnLeft, f_ToTopButton.GetBottom());
            f_DownButton.Location = new Point(l_BtnLeft, f_UpButton.GetBottom());
            f_ToBottomButton.Location = new Point(l_BtnLeft, f_DownButton.GetBottom());

            f_ListView.SelectedIndexChanged += ListView_SelectedIndexChanged;
            f_ListView.ItemCheck += ListView_ItemCheck;
            f_ToTopButton.Click += ToTopButton_Click;
            f_UpButton.Click += UpButton_Click;
            f_DownButton.Click += DownButton_Click;
            f_ToBottomButton.Click += ToBottomButton_Click;
            f_OKButton.Click += OKButton_Click;

            this.Text = l_TextSet[a_TitleTextKey];
            f_GroupBox.Text = l_TextSet[TextKeys.ReorderItemsDlgGroupBoxCaption];
            f_ToTopButton.Text = l_TextSet[TextKeys.ReorderItemsDlgTop];
            f_UpButton.Text = l_TextSet[TextKeys.ReorderItemsDlgUp];
            f_DownButton.Text = l_TextSet[TextKeys.ReorderItemsDlgDown];
            f_ToBottomButton.Text = l_TextSet[TextKeys.ReorderItemsDlgBottom];

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                // Calling dispose on a ListView will also dispose
                // the ColumnHeaders of that list view.
                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(disposing);
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            f_UseItemCheckHandler = true;

            base.OnShown(e);
        }

        private void ListView_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (false == f_UseItemCheckHandler)
            {
                // Block calls that happen when this.Visible is already
                // set, but before the OnShown handler ran (thus
                // blocking non user-triggered events).

                return;
            }

            if (ModifyItemVisibilityMode.FirstItemMustStayVisible == f_ModifyVisMode)
            {
                if (0 == e.Index)
                {
                    e.NewValue = CheckState.Checked;
                }
            }
            else if (ModifyItemVisibilityMode.OneItemMustStayVisible == f_ModifyVisMode)
            {
                if (1 >= f_CheckedItemCount)
                {
                    e.NewValue = CheckState.Checked;
                }
            }

            if (CheckState.Unchecked == e.NewValue)
            {
                if (CheckState.Checked == e.CurrentValue)
                {
                    f_CheckedItemCount--;
                }
            }
            else
            {
                if (CheckState.Unchecked == e.CurrentValue)
                {
                    f_CheckedItemCount++;
                }
            }
        }

        private void ListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateButtons();
        }

        private void UpdateButtons()
        {
            bool l_Enable = (0 < f_ListView.SelectedItems.Count);

            if (true == l_Enable &&
                0 < f_FixedItemCount)
            {
                foreach (ListViewItem l_Item in f_ListView.SelectedItems)
                {
                    if (l_Item.Index < f_FixedItemCount)
                    {
                        // An item is selected that may not be moved.

                        l_Enable = false;

                        break;
                    }
                }
            }

            if (l_Enable != f_ToTopButton.Enabled)
            {
                f_ToTopButton.Enabled = l_Enable;
                f_UpButton.Enabled = l_Enable;
                f_DownButton.Enabled = l_Enable;
                f_ToBottomButton.Enabled = l_Enable;
            }
        }

        private Item[] GetSelItemsSortedByListViewIndex()
        {
            // Returns null if no items are selected.

            int l_Count = f_ListView.SelectedItems.Count;

            if (0 >= l_Count)
            {
                return null;
            }

            Item[] l_Arr = new Item[l_Count];
            f_ListView.SelectedItems.CopyTo(l_Arr, 0);

            // Make sure the array is sorted by the Index property.
            Item.SortArrayByListViewIndex(l_Arr);

            return l_Arr;
        }

        private void MoveSelectedItems(int a_IndexDelta)
        {
            // a_IndexDelta make a_IndexDelta too extreme.

            if (0 == a_IndexDelta)
            {
                return;
            }

            Item[] l_Items = GetSelItemsSortedByListViewIndex();

            if (null == l_Items ||
                // Check for forbidden index.
                f_FixedItemCount > l_Items[0].Index)
            {
                return;
            }

            int l_TotalCount = f_ListView.Items.Count;
            int l_SelCount = l_Items.Length;

            if ((l_TotalCount - f_FixedItemCount) <= l_SelCount)
            {
                return;
            }

            if (0 > a_IndexDelta)
            {
                int l_MinIdx = f_FixedItemCount;
                for (int i = 0; i < l_SelCount; i++)
                {
                    int l_TrgtIdx = Math.Max(l_Items[i].Index + a_IndexDelta, l_MinIdx);

                    if (true == SetItemIndex(l_Items[i], l_TrgtIdx))
                    {
                        l_MinIdx = (l_TrgtIdx + 1);
                    }
                }
            }
            else
            {
                int l_MaxIdx = (l_TotalCount - 1);
                for (int i = (l_SelCount - 1); 0 <= i; i--)
                {
                    int l_TrgtIdx = Math.Min(l_Items[i].Index + a_IndexDelta, l_MaxIdx);

                    if (true == SetItemIndex(l_Items[i], l_TrgtIdx))
                    {
                        l_MaxIdx = Math.Max(l_TrgtIdx - 1, f_FixedItemCount);
                    }
                }
            }
        }

        private bool SetItemIndex(Item a_Item,
                                  int a_NewIndex)
        {
            int l_OldIdx = a_Item.Index;

            if (l_OldIdx < f_FixedItemCount)
            {
                return false;
            }

            if (l_OldIdx != a_NewIndex)
            {
                f_ListView.Items.RemoveAt(l_OldIdx);
                f_ListView.Items.Insert(a_NewIndex, a_Item);
            }

            return true;
        }

        private void ToTopButton_Click(object sender, EventArgs e)
        {
            MoveSelectedItems(short.MinValue);
        }

        private void UpButton_Click(object sender, EventArgs e)
        {
            MoveSelectedItems(-1);
        }

        private void DownButton_Click(object sender, EventArgs e)
        {
            MoveSelectedItems(+1);
        }

        private void ToBottomButton_Click(object sender, EventArgs e)
        {
            MoveSelectedItems(short.MaxValue);
        }

        private void OKButton_Click(object sender, EventArgs e)
        {
            for (int i = (f_Arr.Length - 1); 0 <= i; i--)
            {
                if (ModifyItemVisibilityMode.Disabled != f_ModifyVisMode)
                {
                    bool l_Visible = f_Arr[i].Checked;

                    if (f_Arr[i].p_Visible != l_Visible)
                    {
                        f_Arr[i].SetVisibility(l_Visible);
                        f_VisibilityChanged = true;
                    }
                }

                // Copy the indices from the list view.

                int l_NewIndex = f_Arr[i].Index;
                if (f_Arr[i].p_DisplayOrderIndex != l_NewIndex)
                {
                    f_Arr[i].SetNewIndex(l_NewIndex);
                    f_OrderChanged = true;
                }
            }

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        public bool ShowDialog(IWin32Window a_Owner,
                               int a_FixedCount,
                               ModifyItemVisibilityMode a_ModifyVisMode,
                               Item[] a_Arr,
                               out bool a_OrderChanged)
        {
            // Returns true if the user changed anything (visibility
            // or index-order). The element-order of a_Arr is always
            // undefined on return.

            if (0 == a_Arr.Length)
            {
                a_OrderChanged = false;

                return false;
            }

            f_FixedItemCount = a_FixedCount;
            f_ModifyVisMode = a_ModifyVisMode;
            f_Arr = a_Arr;
            f_VisibilityChanged = false;
            f_CheckedItemCount = 0;
            f_OrderChanged = false;

            // Prepare list view.

            f_ListView.BeginUpdate();
            {
                bool l_WithCheckBox = (ModifyItemVisibilityMode.Disabled != f_ModifyVisMode);
                f_ListView.Items.Clear();
                f_ListView.CheckBoxes = l_WithCheckBox;

                Item.SortArrayByDisplayOrderIndex(a_Arr);

                foreach (Item l_Item in a_Arr)
                {
                    if (true == l_WithCheckBox)
                    {
                        l_Item.Checked = l_Item.p_Visible;

                        if (true == l_Item.p_Visible)
                        {
                            f_CheckedItemCount++;
                        }
                    }

                    f_ListView.Items.Add(l_Item);
                }
            }
            f_ListView.EndUpdate();

            UpdateButtons();

            DialogResult l_Junk = this.ShowDialog(a_Owner);

            f_ListView.Items.Clear();
            f_UseItemCheckHandler = false;
            f_Arr = null;

            // Examine results.

            a_OrderChanged = f_OrderChanged;

            return (true == f_VisibilityChanged ||
                    true == f_OrderChanged);
        }

        public class Item : ListViewItem
        {
            public readonly string f_DisplName;
            public readonly TTag f_Tag;

            private bool f_Visible;
            private int f_DisplayOrderIndex;

            public bool p_Visible
            {
                get { return f_Visible; }
            }

            public int p_DisplayOrderIndex
            {
                get { return f_DisplayOrderIndex; }
            }

            public Item(string a_DisplName,
                        TTag a_Tag,
                        bool a_Visible,
                        int a_DisplayOrderIndex) :
                base(a_DisplName)
            {
                f_DisplName = a_DisplName;
                f_Tag = a_Tag;
                f_Visible = a_Visible;
                f_DisplayOrderIndex = a_DisplayOrderIndex;
            }

            public void SetVisibility(bool a_Visible)
            {
                f_Visible = a_Visible;
            }

            public void SetNewIndex(int a_DisplayOrderIndex)
            {
                f_DisplayOrderIndex = a_DisplayOrderIndex;
            }

            public static void SortArrayByListViewIndex(Item[] a_Arr)
            {
                // Only needed by the parent dialog class.

                Util.SortArrayIfNecessary(a_Arr, CompareByListViewIndex);
            }

            public static void SortArrayByDisplayOrderIndex(Item[] a_Arr)
            {
                Util.SortArrayIfNecessary(a_Arr, CompareByDisplayOrderIndex);
            }

            private static int CompareByListViewIndex(Item a_A,
                                                      Item a_B)
            {
                return a_A.Index.CompareTo(a_B.Index);
            }

            private static int CompareByDisplayOrderIndex(Item a_A,
                                                          Item a_B)
            {
                return a_A.f_DisplayOrderIndex.CompareTo(a_B.f_DisplayOrderIndex);
            }
        }
    }

    enum ModifyItemVisibilityMode : byte
    {
        // Belongs to ReorderItemsDialog.

        Disabled = 0,
        FirstItemMustStayVisible = 1,
        OneItemMustStayVisible = 2,
    }

    class GuiMenuStrip : MenuStrip
    {
        // A menu control that is able to display a message in the
        // unused portion.

        private readonly Font f_MsgFont;
        private readonly StringFormat f_StringFormat;

        private bool f_MayShowMessage;
        private string f_Message;

        public GuiMenuStrip()
        {
            f_MsgFont = new Font(this.Font, FontStyle.Bold);
            f_StringFormat = new StringFormat(StringFormatFlags.NoWrap);
            f_MayShowMessage = true;
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                f_MsgFont.Dispose();
                f_StringFormat.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        public void ClearMessage()
        {
            if (null != f_Message)
            {
                f_Message = null;

                if (true == f_MayShowMessage)
                {
                    RedrawMessage();
                }
            }
        }

        public void SetMessage(string a_Msg)
        {
            if (null == a_Msg ||
                0 >= a_Msg.Length)
            {
                ClearMessage();
            }
            else if (false == object.ReferenceEquals(f_Message, a_Msg))
            {
                f_Message = a_Msg;

                if (true == f_MayShowMessage)
                {
                    RedrawMessage();
                }
            }
        }

        public void AdjustVisibility(bool a_Visible)
        {
            if (a_Visible != f_MayShowMessage)
            {
                f_MayShowMessage = a_Visible;

                if (null != f_Message)
                {
                    RedrawMessage();
                }
            }
        }

        private void RedrawMessage()
        {
            this.Invalidate();
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            base.OnPaintBackground(e);

            if (false == f_MayShowMessage ||
                null == f_Message)
            {
                return;
            }

            int l_ControlWidth = this.Width;
            int l_PreferredSize = this.PreferredSize.Width;
            int l_AvailWidth = (l_ControlWidth - l_PreferredSize);

            if (0 >= l_AvailWidth)
            {
                return;
            }

            Graphics l_G = e.Graphics;
            System.Drawing.Text.TextRenderingHint l_StoredHint = l_G.TextRenderingHint;
            {
                l_G.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;
                int l_X;
                int l_TextWidth = ((int)l_G.MeasureString(f_Message, f_MsgFont, int.MaxValue, f_StringFormat).Width + 1);

                if (RightToLeft.Yes == this.RightToLeft)
                {
                    l_X = 0;
                    l_TextWidth = Math.Min(l_AvailWidth, l_TextWidth);
                }
                else
                {
                    l_X = Math.Max(l_PreferredSize, l_ControlWidth - l_TextWidth);
                }

                int l_ControlHeight = this.Height;
                int l_TextHeight = Math.Min(l_ControlHeight, f_MsgFont.Height + 1);
                l_G.DrawString(f_Message, f_MsgFont, SystemBrushes.ControlText, new RectangleF(l_X, (l_ControlHeight - l_TextHeight) / 2, l_TextWidth, l_TextHeight), f_StringFormat);
            }
            l_G.TextRenderingHint = l_StoredHint;
        }

        protected override void OnRightToLeftChanged(EventArgs e)
        {
            base.OnRightToLeftChanged(e);

            RedrawMessage();
        }
    }

    class ColorPicker : GuiPanel
    {
        private const int c_MaxAlpha = byte.MaxValue;

        public delegate void ColorChangedHandler(ColorPicker a_Sender,
                                                 bool a_Default,
                                                 Color a_Color);

        private readonly Instance f_Inst;

        private readonly GuiLabel f_Caption;
        private readonly GuiRadioButton f_UseDefaultButton;
        private readonly GuiRadioButton f_UseCustomButton;
        private readonly GuiPanel f_ColorPanel;

        private bool f_Default;
        private Color f_CustomColor;

        public event ColorChangedHandler e_ColorChanged;

        public ColorPicker(Instance a_Inst,
                           Control a_Owner,
                           int a_Width,
                           bool a_ExtraCaptionSpace) :
                      base(a_Owner)
        {
            f_Inst = a_Inst;

            int l_RadioButtonWidth = (a_Width - Lyt.s_LargeSpace);
            f_Caption = new GuiLabel(this);
            f_UseDefaultButton = new GuiRadioButton(this, l_RadioButtonWidth);
            f_UseCustomButton = new GuiRadioButton(this, l_RadioButtonWidth);
            f_ColorPanel = new GuiPanel(this);

            // Set the texts here already, since we need to know the
            // final sizes.
            TextSet l_TextSet = f_Inst.f_TextSet;
            f_Caption.Text = l_TextSet.GetAsInputControlCaption(TextKeys.IconColorPickerCaption);
            f_UseDefaultButton.SetTextAndUpdateSize(l_TextSet[TextKeys.IconColorPickerDefaultCol]);
            f_UseCustomButton.SetTextAndUpdateSize(l_TextSet[TextKeys.IconColorPickerCustomCol]);

            f_Caption.Width = a_Width;
            f_UseDefaultButton.Location = new Point(Lyt.s_LargeSpace, f_Caption.GetBottom() + (false == a_ExtraCaptionSpace ? 0 : (int)Lyt.s_SmallSpace));
            f_UseCustomButton.Location = new Point(Lyt.s_LargeSpace, f_UseDefaultButton.GetBottom());
            int l_ColPanelPixels = f_UseCustomButton.Height;
            int l_MaxCustomButtonWidth = (l_RadioButtonWidth - Lyt.s_SmallSpace - l_ColPanelPixels);

            if (l_MaxCustomButtonWidth < f_UseCustomButton.Width)
            {
                // The radio button must not span into the range
                // reserved for the color panel.

                f_UseCustomButton.Width = l_MaxCustomButtonWidth;
            }

#if !MONO
            // Don't do this for Mono, since the panel is already
            // quite small.
            f_ColorPanel.BorderStyle = BorderStyle.Fixed3D;
#endif
            f_ColorPanel.Location = new Point(f_UseCustomButton.Right + Lyt.s_SmallSpace, f_UseCustomButton.Top);
            f_ColorPanel.Size = new Size(l_ColPanelPixels, l_ColPanelPixels);
            this.Size = new Size(a_Width, f_UseCustomButton.Bottom);

            ResetAttributes();

            f_UseDefaultButton.CheckedChanged += RadioButton_Click;
            f_UseCustomButton.CheckedChanged += RadioButton_Click;
            f_ColorPanel.Click += ColorPanel_Click;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
            }

            base.Dispose(disposing);
        }

        public void ResetAttributes()
        {
            // Suggest a kind of orange as custom color.
            SetAttributes(true, Color.FromArgb(255, 128, 0), true);
        }

        public void SetAttributes(bool a_Default,
                                  Color a_CustomColor,
                                  bool a_ForceUpdate)
        {
            bool l_Change = false;

            if (a_Default != f_Default ||
                true == a_ForceUpdate)
            {
                l_Change = true;

                f_Default = a_Default;

                if (false == f_Default)
                {
                    f_UseCustomButton.Checked = true;
                }
                else
                {
                    f_UseDefaultButton.Checked = true;
                }
            }

            if (a_CustomColor != f_CustomColor ||
                true == a_ForceUpdate)
            {
                l_Change = true;

                f_CustomColor = a_CustomColor;
                f_ColorPanel.BackColor = f_CustomColor;
            }

            if (true == l_Change &&
                null != e_ColorChanged)
            {
                e_ColorChanged(this, f_Default, f_CustomColor);
            }
        }

        public void GetAttributes(out bool a_Default,
                                  out Color a_CustomColor)
        {
            a_Default = f_Default;
            a_CustomColor = f_CustomColor;
        }

        public void SetColorEasy(Color a_Color,
                                 bool a_ForceUpdate)
        {
            bool l_Default = !IsColorUsable(a_Color);
            SetAttributes(l_Default, Color.FromArgb(c_MaxAlpha, a_Color), a_ForceUpdate);
        }

        public Color GetColorEasy()
        {
            // Translates the default state into an alpha value of
            // zero, otherwise an maximum alpha value is enfored.

            bool l_Default;
            Color l_CustomColor;
            GetAttributes(out l_Default, out l_CustomColor);

            int l_Alpha = (false == l_Default ? c_MaxAlpha : 0);
            return Color.FromArgb(l_Alpha, l_CustomColor);
        }

        private void RadioButton_Click(object sender, EventArgs e)
        {
            SetAttributes(f_UseDefaultButton.Checked, f_CustomColor, false);
        }

        private void ColorPanel_Click(object sender, EventArgs e)
        {
            using (ColorDialog l_Dlg = new ColorDialog())
            {
                l_Dlg.FullOpen = true;
                l_Dlg.AnyColor = true;
                l_Dlg.Color = Color.FromArgb(f_CustomColor.ToArgb() | unchecked((int)0xFF000000));

                if (DialogResult.OK == l_Dlg.ShowDialog(f_Inst.f_MainForm))
                {
                    Color l_Col = Color.FromArgb(l_Dlg.Color.ToArgb() | unchecked((int)0xFF000000));

                    SetAttributes(false, l_Col, false);
                }
            }
        }

        public static bool IsColorUsable(Color a_Color)
        {
            return (c_MaxAlpha == a_Color.A);
        }

        public static Color ChooseColorEasy(Color a_PossColor,
                                            Color a_FallbackColor)
        {
            // a_PossColor can be the return value of GetColorEasy().

            return (false == IsColorUsable(a_PossColor) ? a_FallbackColor : a_PossColor);
        }
    }

    class BarBackgroundImages : IDisposable
    {
        private const uint c_MaxFileByteSize = (2 * 1024 * 1024);
        private const uint c_MaxPixelProduct = (16 * 1024 * 1024);
        public const byte c_MaxImages = 32;
        public const byte c_AspectDivisor = 7;

        private readonly static Color s_BackgroundColor = Color.White;

        private readonly Bitmap[] f_Bitmaps;

        public int p_Count
        {
            get { return f_Bitmaps.Length; }
        }

        private BarBackgroundImages(Bitmap[] a_Bitmaps)
        {
            f_Bitmaps = a_Bitmaps;
        }

        public void Dispose()
        {
            foreach (Bitmap l_Bitmap in f_Bitmaps)
            {
                l_Bitmap.Dispose();
            }
        }

        public Bitmap GetBitmap(int a_Idx)
        {
            // The return value remains owned by this class and may
            // not be disposed by the caller.

            return f_Bitmaps[a_Idx];
        }

        public int GetBitmapIndex(Content.KnownFile a_Known)
        {
            return (int)(a_Known.f_FileNum.f_Val % (uint)f_Bitmaps.Length);
        }

        public static BarBackgroundImages LoadFromDisk(string a_Folder,
                                                       int a_MaxWidth)
        {
            // Return can be null, if no suitable files were found.

            string[] l_FileNames;

            try
            {
                l_FileNames = System.IO.Directory.GetFiles(a_Folder);
            }
            catch
            {
                return null;
            }

            a_MaxWidth = Math.Max(a_MaxWidth, 1);
            Size l_AimedSize = new Size(a_MaxWidth, Math.Max(1, Util.RoundDoubleToInt((float)a_MaxWidth / c_AspectDivisor)));
            List<Bitmap> l_Bitmaps = new List<Bitmap>(Math.Min(l_FileNames.Length, c_MaxImages));

            foreach (string l_FileName in l_FileNames)
            {
                System.IO.FileStream l_Strm = null;
                Image l_Img = null;

                try
                {
                    l_Strm = System.IO.File.OpenRead(l_FileName);

                    if (c_MaxFileByteSize < l_Strm.Length)
                    {
                        continue;
                    }

                    l_Img = Image.FromStream(l_Strm);
                }
                catch
                {
                    if (null != l_Img)
                    {
                        l_Img.Dispose();
                    }

                    continue;
                }
                finally
                {
                    if (null != l_Strm)
                    {
                        l_Strm.Dispose();
                    }
                }

                if (c_MaxPixelProduct < (l_Img.Width * l_Img.Height))
                {
                    l_Img.Dispose();

                    continue;
                }

                if (l_Img.Height > l_Img.Width)
                {
                    l_Img.RotateFlip(RotateFlipType.Rotate90FlipNone);
                }

                Bitmap l_Bitmap = TranslateImage(l_Img, l_AimedSize);
                l_Img.Dispose();

                l_Bitmaps.Add(l_Bitmap);

                if (c_MaxImages == l_Bitmaps.Count)
                {
                    break;
                }
            }

            if (0 >= l_Bitmaps.Count)
            {
                return null;
            }

            return new BarBackgroundImages(l_Bitmaps.ToArray());
        }

        private static Bitmap TranslateImage(Image a_Img,
                                             Size a_NewSize)
        {
            // The caller gains ownership over the returned bitmap.

            int l_Width = a_Img.Width;
            int l_Height = a_Img.Height;
            int l_AimedHeight = Util.RoundDoubleToInt((float)l_Width / c_AspectDivisor);
            Rectangle l_SrcRect;

            if (l_Height == l_AimedHeight)
            {
                // Perfect fit.

                l_SrcRect = new Rectangle(Point.Empty, a_Img.Size);
            }
            else
            {
                // Input is too high.

                l_SrcRect = new Rectangle(0, (l_Height - l_AimedHeight) / 2, l_Width, l_AimedHeight);
            }

            Bitmap l_Bitmap = new Bitmap(a_NewSize.Width, a_NewSize.Height);

            Graphics l_G = Graphics.FromImage(l_Bitmap);
            {
                l_G.Clear(s_BackgroundColor);
                l_G.DrawImage(a_Img, new Rectangle(Point.Empty, a_NewSize), l_SrcRect, GraphicsUnit.Pixel);
            }
            l_G.Dispose();

            return l_Bitmap;
        }
    }

    class IconPicker : IDisposable
    {
        // NOTE: Make sure Dispose() gets called (this class is not a
        // control by itself).

        private const byte c_ButtonExtraPixels = 13;

        private readonly Instance f_Inst;

        private readonly GuiLabel f_Caption;
        private readonly GuiButton f_Button;

        private readonly Bitmap f_Icon;

        private IconTypes f_IconType;
        private Color f_BgColor;

        public IconTypes p_IconType
        {
            get { return f_IconType; }
        }

        public IconPicker(Instance a_Inst,
                          Control a_Owner,
                          int a_Width)
        {
            f_Inst = a_Inst;

            f_Caption = new GuiLabel(a_Owner);
            f_Button = new GuiButton(a_Owner);

            int l_BtnWidth = (IconStorage.c_PixelSize + c_ButtonExtraPixels);
            f_Caption.Width = a_Width;
            f_Button.Width = Math.Min(Lyt.CalcX(l_BtnWidth), a_Width);
            f_Button.Height = Lyt.CalcY(l_BtnWidth);
#if MONO
            // The BackColor property doesn't really work otherwise on
            // Mono.
            f_Button.FlatStyle = FlatStyle.Flat;
#endif

            f_Button.Click += IconPicker_Click;

            f_Caption.Text = f_Inst.f_TextSet.GetAsInputControlCaption(TextKeys.FilesTabCustomTabDlgIcon);

            f_Icon = IconStorage.BuildEmptyIcon();
        }

        public void Dispose()
        {
            f_Caption.Dispose();
            f_Button.Dispose();

            f_Icon.Dispose();
        }

        public void SetBackColor(Color a_BgColor)
        {
            f_BgColor = a_BgColor;

            UpdateBgColor();
        }

        private void UpdateBgColor()
        {
            f_Button.BackColor = (IconTypes.NoIcon == f_IconType ? SystemColors.Control : f_BgColor);
        }

        public void SetIconType(IconTypes a_IconType)
        {
            f_IconType = a_IconType;

            if (IconTypes.NoIcon == f_IconType)
            {
                f_Button.Image = null;
            }
            else
            {
                Software.s_IconStorage.FillIcon(f_IconType, Color.Transparent, f_Icon);
                f_Button.Image = f_Icon;
            }

            UpdateBgColor();
        }

        private void IconPicker_Click(object sender, EventArgs e)
        {
            using (PickerForm l_Dlg = new PickerForm(f_Inst))
            {
                Rectangle l_ButtonRect = f_Button.RectangleToScreen(f_Button.ClientRectangle);
                l_Dlg.SetChildWindowPosition(l_ButtonRect);

                IconTypes l_IconType;
                if (true == l_Dlg.ShowDialog(f_Button, out l_IconType))
                {
                    SetIconType(l_IconType);
                }
            }
        }

        public void SetLocation(int a_x,
                                int a_y)
        {
            f_Caption.Location = new Point(a_x, a_y);
            f_Button.Location = new Point(a_x, f_Caption.GetBottom());
        }

        public int GetBottom()
        {
            return (f_Button.Bottom + Lyt.s_Space);
        }

        private class PickerForm : Dialog
        {
            private const ushort c_Lines = (IconStorage.c_IconCount / IconStorage.c_PicsPerLine);

            private readonly Size f_IconSize;
            private readonly Rectangle f_NoIconRect;
            private readonly Rectangle f_IconsRect;
            private readonly Bitmap f_NeutralImage;
            private readonly Bitmap f_SelectionImage;

            private bool f_HasSelection;
            private IconTypes f_SelIconType;

            private IconTypes f_ClickedIconType;

            public PickerForm(Instance a_Inst)
            {
                TextSet l_TextSet = a_Inst.f_TextSet;
                Bitmap l_RawImage = Software.s_IconStorage.ExposeResourceImage();

                Color l_BgColor = Color.Black;
                Pen l_FramePen = Pens.Gray;
                string l_NoIconStr = l_TextSet[TextKeys.FilesTabCustomTabDlgPickIconDlgNoIcon];

                f_IconSize = Lyt.TranformSize(IconStorage.c_PixelSize, IconStorage.c_PixelSize);
                Size l_Sz = new Size(IconStorage.c_PicsPerLine * f_IconSize.Width, c_Lines * f_IconSize.Height);
                f_NoIconRect = new Rectangle(1, 1, l_Sz.Width, f_IconSize.Height);
                f_IconsRect = new Rectangle(1, f_NoIconRect.Bottom, l_Sz.Width, l_Sz.Height);
                f_NeutralImage = new Bitmap(f_IconsRect.Right + 1, f_IconsRect.Bottom + 1);

                Graphics l_G = Graphics.FromImage(f_NeutralImage);
                Font l_Fnt = new Font(this.Font, FontStyle.Bold);
                StringFormat l_StrFrmt = BarDrawer.CreateCenterStringFormat();
                {
                    l_G.Clear(l_BgColor);
                    l_G.DrawRectangle(l_FramePen, new Rectangle(0, 0, f_NeutralImage.Width - 1, f_NeutralImage.Height - 1));
                    l_G.DrawString(l_NoIconStr, l_Fnt, Brushes.White, f_NoIconRect, l_StrFrmt);
                    l_G.DrawImage(l_RawImage, f_IconsRect, new Rectangle(Point.Empty, l_RawImage.Size), GraphicsUnit.Pixel);
                }
                l_StrFrmt.Dispose();
                l_Fnt.Dispose();
                l_G.Dispose();

                f_SelectionImage = IconStorage.CloneBitmap(f_NeutralImage);

                this.AsToolDialog(f_NeutralImage.Size);
                this.StartPosition = FormStartPosition.Manual;
                this.DoubleBuffered = true;

                this.Text = l_TextSet[TextKeys.FilesTabCustomTabDlgPickIconDlgCaption];
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    f_NeutralImage.Dispose();
                    f_SelectionImage.Dispose();
                }

                base.Dispose(disposing);
            }

            protected override void OnPaint(PaintEventArgs e)
            {
                e.Graphics.DrawImageUnscaled(f_SelectionImage, Point.Empty);

                // Don't call the base class.
            }

            protected override void OnKeyDown(KeyEventArgs e)
            {
                // Need to call base class implementation.

                if (Keys.Escape == e.KeyCode)
                {
                    this.Close();

                    e.Handled = true;
                }

                base.OnKeyDown(e);
            }

            protected override void OnMouseLeave(EventArgs e)
            {
                MouseOver(new Point(int.MinValue, int.MinValue));

                base.OnMouseLeave(e);
            }

            protected override void OnMouseMove(MouseEventArgs e)
            {
                MouseOver(e.Location);

                base.OnMouseMove(e);
            }

            private void MouseOver(Point a_Pnt)
            {
                IconTypes l_SelIconType;
                bool l_HasSelection = GetIconTypeAtPos(a_Pnt, out l_SelIconType);

                if (l_HasSelection != f_HasSelection ||
                    l_SelIconType != f_SelIconType)
                {
                    f_HasSelection = l_HasSelection;
                    f_SelIconType = l_SelIconType;

                    FillSelectionImage(f_HasSelection, f_SelIconType);
                    this.Invalidate();
                }
            }

            protected override void OnMouseClick(MouseEventArgs e)
            {
                if (true == GetIconTypeAtPos(e.Location, out f_ClickedIconType))
                {
                    this.DialogResult = DialogResult.OK;
                    this.Close();
                }

                base.OnMouseClick(e);
            }

            private bool GetIconTypeAtPos(Point a_Pos,
                                          out IconTypes a_IconType)
            {
                // Returns true if the point is within a sensitive
                // region.

                if (true == f_IconsRect.Contains(a_Pos))
                {
                    int l_x = ((a_Pos.X - f_IconsRect.X) / f_IconSize.Width);
                    int l_y = ((a_Pos.Y - f_IconsRect.Y) / f_IconSize.Height);
                    a_IconType = IconStorage.IconIndexToIconType((l_y * IconStorage.c_PicsPerLine) + l_x);

                    return true;
                }

                a_IconType = IconTypes.NoIcon;

                return f_NoIconRect.Contains(a_Pos);
            }

            private void FillSelectionImage(bool a_WithSelection,
                                            IconTypes a_IconType)
            {
                Graphics l_G = Graphics.FromImage(f_SelectionImage);
                {
                    l_G.DrawImageUnscaled(f_NeutralImage, Point.Empty);

                    if (true == a_WithSelection)
                    {
                        if (IconTypes.NoIcon == a_IconType)
                        {
                            DrawSelection(l_G, f_NoIconRect);
                        }
                        else
                        {
                            Rectangle l_Rect;
                            IconStorage.CalcIconRect(a_IconType, f_IconSize, out l_Rect);
                            l_Rect.Offset(f_IconsRect.Location);

                            DrawSelection(l_G, l_Rect);
                        }
                    }
                }
                l_G.Dispose();
            }

            private static void DrawSelection(Graphics a_G,
                                              Rectangle a_Region)
            {
                Pen l_SelFramePen = Pens.White;

                a_G.DrawRectangle(l_SelFramePen, a_Region.X, a_Region.Y, a_Region.Width - 1, a_Region.Height - 1);
            }

            public bool ShowDialog(IWin32Window a_Parent,
                                   out IconTypes a_IconType)
            {
                // Returns true if the user did not cancel.

                f_HasSelection = false;
                f_SelIconType = IconTypes.NoIcon;

                if (DialogResult.OK == this.ShowDialog(a_Parent))
                {
                    a_IconType = f_ClickedIconType;

                    return true;
                }

                a_IconType = IconTypes.NoIcon;

                return false;
            }
        }
    }

    class IconStorage : IDisposable
    {
        public static readonly Color s_SuggestedBgColor = System.Drawing.Color.Black;

        // This value can be used to decrease the size of the merged
        // file.
        public const byte c_PicsPerLine = 7;
        public const byte c_PixelSize = 26;
        public const byte c_IconCount = 42;

        private readonly Bitmap f_SourceImage;
        private readonly Bitmap[] f_DefIcons;

        private Color f_BgColor;

        public Color p_BackgroundColor
        {
            get { return f_BgColor; }
        }

        public IconStorage(Color a_BgColor)
        {
            System.IO.Stream l_Strm = Software.OpenResource("Icons.png", false);
            f_SourceImage = new Bitmap(l_Strm);
            l_Strm.Dispose();

            f_DefIcons = CreateIconArray();
            FillIconArray(a_BgColor, f_DefIcons);

            f_BgColor = a_BgColor;
        }

        public void Dispose()
        {
            DisposeIconArray(f_DefIcons);
        }

        public static void DisposeIconArray(Bitmap[] a_Icons)
        {
            foreach (Bitmap l_Map in a_Icons)
            {
                l_Map.Dispose();
            }
        }

        public static Bitmap[] CreateIconArray()
        {
            Bitmap[] l_Icons = new Bitmap[c_IconCount];

            for (int i = 0; i < c_IconCount; i++)
            {
                l_Icons[i] = BuildEmptyIcon();
            }

            return l_Icons;
        }

        public static Bitmap BuildEmptyIcon()
        {
            return new Bitmap(c_PixelSize, c_PixelSize);
        }

        public Bitmap ExposeResourceImage()
        {
            // The caller does not own the returned image.

            return f_SourceImage;
        }

        public void FillIconArray(Color a_BgColor,
                                  Bitmap[] a_Dest)
        {
            // a_BgColor may be Color.Transparent.

            int l_Count = a_Dest.Length;

            if (0 >= l_Count)
            {
                return;
            }

            Size l_Size = a_Dest[0].Size;

            Rectangle l_DestRect = new Rectangle(Point.Empty, l_Size);
            Rectangle l_SrcRect = l_DestRect;
            for (int i = 0; i < l_Count; )
            {
                Graphics l_G = Graphics.FromImage(a_Dest[i]);
                l_G.Clear(a_BgColor);
                l_G.DrawImage(f_SourceImage, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
                l_G.Dispose();

                if (0 == (++i % c_PicsPerLine))
                {
                    l_SrcRect.X = 0;
                    l_SrcRect.Y += l_Size.Height;
                }
                else
                {
                    l_SrcRect.X += l_Size.Width;
                }
            }
        }

        public bool SetBackgroundColor(Color a_BgColor)
        {
            // Returns true if the color changed.

            if (a_BgColor == f_BgColor)
            {
                return false;
            }

            // We need to draw on the existing images (f_Icons).

            FillIconArray(a_BgColor, f_DefIcons);
            f_BgColor = a_BgColor;

            return true;
        }

        public Bitmap GetDefaultIcon(IconTypes a_IconType)
        {
            // a_IconType may be IconTypes.NoIcon, which leads to a
            // return value of null, otherwise an image will be
            // returned that the caller does not own (it may not be
            // disposed by the caller).

            if (IconTypes.NoIcon == a_IconType)
            {
                return null;
            }

            return f_DefIcons[IconTypeToIconIndex(a_IconType)];
        }

        public void FillIcon(IconTypes a_IconType,
                             Color a_BgColor,
                             Bitmap a_Dest)
        {
            // The caller takes ownership and is responsible for
            // disposing the returned image.

            Graphics l_G = Graphics.FromImage(a_Dest);

            if (a_BgColor.ToArgb() == f_BgColor.ToArgb())
            {
                l_G.DrawImageUnscaled(GetDefaultIcon(a_IconType), Point.Empty);
            }
            else
            {
                Rectangle l_DestRect = new Rectangle(0, 0, c_PixelSize, c_PixelSize);
                Rectangle l_SrcRect;
                CalcIconRect(a_IconType, new Size(c_PixelSize, c_PixelSize), out l_SrcRect);

                l_G.Clear(a_BgColor);
                l_G.DrawImage(f_SourceImage, l_DestRect, l_SrcRect, GraphicsUnit.Pixel);
            }

            l_G.Dispose();
        }

        public static void CalcIconRect(IconTypes a_IconType,
                                        Size a_Size,
                                        out Rectangle a_Rect)
        {
            int l_Idx = IconTypeToIconIndex(a_IconType);
            a_Rect = new Rectangle((l_Idx % c_PicsPerLine) * a_Size.Width,
                                   (l_Idx / c_PicsPerLine) * a_Size.Height,
                                   a_Size.Width,
                                   a_Size.Height);
        }

        public static int IconTypeToIconIndex(IconTypes a_IconType)
        {
            return ((int)a_IconType - 1);
        }

        public static IconTypes IconIndexToIconType(int a_IconIndex)
        {
            if (0 > a_IconIndex ||
                c_IconCount <= a_IconIndex)
            {
                return IconTypes.NoIcon;
            }

            return (IconTypes)(a_IconIndex + 1);
        }

        public static Bitmap CloneBitmap(Bitmap a_Src)
        {
            return (Bitmap)a_Src.Clone();
        }
    }

    enum IconTypes : byte
    {
        NoIcon = 0,

        AidShield = 1,
        Armchair,
        Audio,
        Balloon,
        Battery,
        Bell,
        Box,
        Bug,
        Camera,
        Clock,
        CloudUp,
        Cup,
        Cutlery,
        Disk,
        Download,
        File,
        FilmReel,
        Forbidden,
        Gear,
        Heart,
        Info,
        Key,
        LetterClosed,
        LetterOpen,
        Lightning,
        Magnifier,
        Pendant,
        Person,
        Picture,
        PieChart,
        Progress,
        Radio,
        RadioTower,
        Recycle,
        Scissors,
        Sheet,
        SheetPencil,
        Shirt,
        Speaker,
        Star,
        TrashBin,
        Umbrella,
    }

    sealed class GuiTimer : System.Windows.Forms.Timer
    {
        // This is a winforms timer that is aware of the actual
        // elapsed time and can more seriously be used for time based
        // event handling than the default timer (in some scenarios).

        public delegate void AdvancedTickHandler(GuiTimer a_Sender,
                                                 uint a_Times);

        public readonly bool f_InvokeOnStart;
        public Time f_ElapsedTime;

        public event AdvancedTickHandler e_AdvancedTick;

        public override bool Enabled
        {
            get { return base.Enabled; }
            set { AdjustEnabled(value); }
        }

        public GuiTimer() :
            this(false)
        {
        }

        public GuiTimer(bool a_InvokeOnStart) :
            base()
        {
            f_InvokeOnStart = a_InvokeOnStart;
        }

        private void AdjustEnabled(bool a_Enabled)
        {
            if (a_Enabled == base.Enabled)
            {
                return;
            }

            base.Enabled = a_Enabled;

            if (true == a_Enabled)
            {
                f_ElapsedTime = Time.Now();

                if (true == f_InvokeOnStart)
                {
                    OnAdvancedTick(1);
                }
            }
        }

        protected override void OnTick(EventArgs e)
        {
            // We deliberately do not call the base class routine.

            const ulong l_MinTicks = TimeSpan.TicksPerMillisecond;

            Time l_Time = Time.Now();
            ulong l_Ticks = (l_Time.ToBinary() - f_ElapsedTime.ToBinary());

            if (l_MinTicks > l_Ticks)
            {
                return;
            }

            ulong l_Interval = ((uint)this.Interval * l_MinTicks);
            ulong l_Times = (l_Ticks / l_Interval);

            if (0 == l_Times)
            {
                // No full interval elapsed yet.

                return;
            }

            f_ElapsedTime = f_ElapsedTime.AddTicks(l_Times * l_Interval);

            OnAdvancedTick((uint)l_Times);
        }

        private void OnAdvancedTick(uint a_Times)
        {
            if (null != e_AdvancedTick)
            {
                e_AdvancedTick(this, a_Times);
            }
        }
    }

    sealed class GuiToolTip : ToolTip
    {
        public GuiToolTip()
        {
        }

        public void ClearToolTip(Control a_Cntrl)
        {
            this.SetToolTip(a_Cntrl, null);
        }
    }

    sealed class GuiProgressBar : Control
    {
        public const byte c_PercentsPerBase = 100;
        // 1,000,000,000
        public const uint c_BaseValue = 0x3B9ACA00;
        // 10,000,000
        public const uint c_PercentValue = (c_BaseValue / c_PercentsPerBase);
        public const byte c_MaxDecimals = 5;

        private static BarBrushes s_Brushes;

        private uint f_Value;

        private bool f_DisplayText;
        private byte f_Decimals;
        private string f_CachedText;

        private bool f_BlackOuterFrame;
        private bool f_NoRoundingToCompleteness;

        static GuiProgressBar()
        {
            s_Brushes = new BarBrushes(Lyt.CalcY(25));
        }

        public GuiProgressBar(Control a_Owner)
        {
            a_Owner.Controls.Add(this);

            this.Size = new Size(8 * Lyt.s_Space, s_Brushes.f_SuggestedControlHeight);
            this.DoubleBuffered = true;

            const byte lc_DefaultDecimals = 1;
            f_DisplayText = true;
            f_Decimals = lc_DefaultDecimals;
            f_NoRoundingToCompleteness = true;
        }

        protected override void OnRightToLeftChanged(EventArgs e)
        {
            // Need to call base class implementation.

            InvalidateBar();

            base.OnRightToLeftChanged(e);
        }

        private void DrawFrame(Graphics a_G,
                               ref Rectangle a_Rect,
                               ref Rectangle a_ClipRect)
        {
            Pen l_Pen = (false == f_BlackOuterFrame ? s_Brushes.f_AltFramePen : Pens.Black);

            int l_X = (a_Rect.X + 1);
            int l_Y = (a_Rect.Y);
            int l_Right = (a_Rect.Right - 2);
            int l_Bottom = (a_Rect.Bottom - 1);

            // Draw horizontal lines.

            if (a_ClipRect.Y <= l_Y)
            {
                a_G.DrawLine(l_Pen, l_X, l_Y, l_Right, l_Y);
            }

            if (a_ClipRect.Bottom > l_Bottom)
            {
                a_G.DrawLine(l_Pen, l_X, l_Bottom, l_Right, l_Bottom);
            }

            l_X--;
            l_Y++;
            l_Right++;
            l_Bottom--;

            // Draw vertical lines.

            if (a_ClipRect.X <= l_X)
            {
                a_G.DrawLine(l_Pen, l_X, l_Y, l_X, l_Bottom);
            }

            if (a_ClipRect.Right > l_Right)
            {
                a_G.DrawLine(l_Pen, l_Right, l_Y, l_Right, l_Bottom);
            }
        }

        private void DrawBar(Graphics a_G,
                             ref Rectangle a_Rect,
                             ref Rectangle a_ClipRect)
        {
            bool l_FromRight = (RightToLeft.Yes == this.RightToLeft);
            int l_ComplWidth = GetCompletedWidth(a_Rect.Width);
            int l_ComplX, l_IncomplX;

            if (false == l_FromRight)
            {
                l_ComplX = a_Rect.X;
                l_IncomplX = (l_ComplX + l_ComplWidth);
            }
            else
            {
                l_ComplX = (a_Rect.Right - l_ComplWidth);
                l_IncomplX = a_Rect.X;
            }

            if (0 < l_ComplWidth &&
                (l_ComplX + l_ComplWidth) > a_ClipRect.X)
            {
                int l_LowerTop = (a_Rect.Y + s_Brushes.f_UpperHeight);
                a_G.FillRectangle(s_Brushes.f_ComplUpperGradient, l_ComplX, a_Rect.Y, l_ComplWidth, s_Brushes.f_UpperHeight);
                a_G.FillRectangle(s_Brushes.f_ComplLowerGradient, l_ComplX, l_LowerTop, l_ComplWidth, s_Brushes.f_LowerHeight);
            }

            if (a_Rect.Width > l_ComplWidth &&
                a_ClipRect.Right > l_IncomplX)
            {
                int l_RightWidth = (a_Rect.Width - l_ComplWidth);
                a_G.FillRectangle(s_Brushes.f_IncomplGradient, l_IncomplX, a_Rect.Y, l_RightWidth, a_Rect.Height);
            }
        }

        private void DrawText(Graphics a_G,
                              ref Rectangle a_Rect,
                              ref Rectangle a_ClipRect)
        {
            if (null == f_CachedText)
            {
                f_CachedText = BuildText();
            }

            System.Drawing.Text.TextRenderingHint l_StoredHint = a_G.TextRenderingHint;
            {
                a_G.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;
                a_G.DrawString(f_CachedText, s_Brushes.f_Font, s_Brushes.f_TextBrush, a_Rect, s_Brushes.f_StrFrmt);
            }
            a_G.TextRenderingHint = l_StoredHint;
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            // Need to call base class implementation.

            Rectangle l_Rect, l_InnerRect;

            if (true == GetRects(out l_Rect, out l_InnerRect))
            {
                Rectangle l_ClipRect = e.ClipRectangle;

                DrawFrame(e.Graphics, ref l_Rect, ref l_ClipRect);
                DrawBar(e.Graphics, ref l_InnerRect, ref l_ClipRect);

                if (true == f_DisplayText)
                {
                    DrawText(e.Graphics, ref l_InnerRect, ref l_ClipRect);
                }
            }

            base.OnPaint(e);
        }

        private bool GetRects(out Rectangle a_Rect,
                              out Rectangle a_InnerRect)
        {
            // Returns a nonzero value if the inner rect has a nonzero
            // size.

            int l_MaxHeight = Math.Min(this.Height, s_Brushes.f_SuggestedControlHeight);
            a_Rect = new Rectangle(0, 0, this.Width, l_MaxHeight);

            int l_InnerHeight = Math.Min(l_MaxHeight, s_Brushes.f_SuggestedControlHeight - 2);
            a_InnerRect = new Rectangle(1, 1, a_Rect.Width - 2, l_InnerHeight);

            return (0 < a_InnerRect.Width &&
                    0 < a_InnerRect.Height);
        }

        private void InvalidateBar()
        {
            if (null != f_CachedText)
            {
                f_CachedText = null;
            }

            Rectangle l_Rect, l_InnerRect;

            if (true == GetRects(out l_Rect, out l_InnerRect))
            {
                this.Invalidate(l_Rect);
            }
        }

        public void AdjustDisplayText(bool a_DisplayText)
        {
            if (a_DisplayText == f_DisplayText)
            {
                return;
            }

            f_DisplayText = a_DisplayText;

            InvalidateBar();
        }

        public void AdjustDecimalsDisplay(uint a_NewDecimals)
        {
            uint l_Decimals = Math.Min(a_NewDecimals, c_MaxDecimals);

            if (l_Decimals == f_Decimals)
            {
                return;
            }

            f_Decimals = (byte)l_Decimals;

            if (true == f_DisplayText)
            {
                InvalidateBar();
            }
        }

        public void AdjustOuterFrameColor(bool a_Black)
        {
            if (a_Black == f_BlackOuterFrame)
            {
                return;
            }

            f_BlackOuterFrame = a_Black;
            this.Invalidate();
        }

        public void AdjustNoRoundingToCompleteness(bool a_NoRounding)
        {
            if (a_NoRounding == f_NoRoundingToCompleteness)
            {
                return;
            }

            f_NoRoundingToCompleteness = a_NoRounding;

            if (true == f_DisplayText &&
                (c_BaseValue - c_PercentValue) < f_Value)
            {
                InvalidateBar();
            }
        }

        private string BuildText()
        {
            TextSet l_TextSet = Software.sp_TextSet;
            float l_Float = GetProgressFloat();

            if (true == f_NoRoundingToCompleteness &&
                c_BaseValue <= f_Value)
            {
                float l_Limit;

                switch (f_Decimals)
                {
                    case 0:
                        l_Limit = .99f;
                        break;

                    case 1:
                        l_Limit = .999f;
                        break;

                    case 2:
                        l_Limit = .9999f;
                        break;

                    case 3:
                        l_Limit = .99999f;
                        break;

                    case 4:
                        l_Limit = .999999f;
                        break;

                    default:
                        l_Limit = .9999999f;
                        break;
                }

                l_Float = Math.Min(l_Float, l_Limit);
            }

            if (1 == f_Decimals)
            {
                return l_TextSet.PercStr(l_Float);
            }

            return l_Float.ToString("P" + f_Decimals, l_TextSet.f_FormatCulture);
        }

        private int GetCompletedWidth(int a_TotalWidth)
        {
            if (0 == f_Value)
            {
                return 0;
            }

            if (c_BaseValue == f_Value)
            {
                return a_TotalWidth;
            }

            uint l_ComplPixels = (uint)(Util.Mul64((uint)a_TotalWidth, f_Value) / c_BaseValue);

            return (int)l_ComplPixels;
        }

        public float GetProgressFloat()
        {
            return ((float)f_Value / c_BaseValue);
        }

        public uint GetProgressPercent()
        {
            return (f_Value / c_PercentValue);
        }

        public void SetMinProgress()
        {
            SetProgressCore(0);
        }

        public void SetMaxProgress()
        {
            SetProgressCore(c_BaseValue);
        }

        public void SetProgressPercent(uint a_Percent)
        {
            uint l_Value = (Math.Min(a_Percent, c_PercentsPerBase) * c_PercentValue);
            SetProgressCore(l_Value);
        }

        public void SetProgress(uint a_NewValue)
        {
            // This is the primary progress value setter.

            SetProgressCore(Math.Min(a_NewValue, c_BaseValue));
        }

        private void SetProgressCore(uint a_NewValue)
        {
            // a_NewValue must be within range.

            if (a_NewValue == f_Value)
            {
                return;
            }

            f_Value = a_NewValue;

            InvalidateBar();
        }

        public void SetProgress(ulong a_Elapsed,
                                ulong a_Total)
        {
            uint l_Value = CalcProgressValue(a_Elapsed, a_Total);
            SetProgressCore(l_Value);
        }

        public static uint CalcProgressValue(ulong a_Elapsed,
                                             ulong a_Total)
        {
            if (a_Elapsed >= a_Total)
            {
                // This also protects against the case where
                // a_Total is zero.

                return c_BaseValue;
            }

            if (0 == a_Elapsed)
            {
                return 0;
            }

            if (uint.MaxValue >= a_Total)
            {
                // We can handle this with integral arithmetic.

                ulong l_ULongVal = ((a_Elapsed * c_BaseValue) / a_Total);

                return (uint)l_ULongVal;
            }

            // Use a double precision floating point
            // calculation for this.

            double l_DoubleVal = (((double)a_Elapsed * c_BaseValue) / a_Total);

            return Math.Min((uint)l_DoubleVal, c_BaseValue);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct BarBrushes : IDisposable
        {
            private static readonly Color s_TextColor;
            private static readonly Color s_AltFrameColor;

            private static readonly Color s_ComplUpperCol0;
            private static readonly Color s_ComplUpperCol1;
            private static readonly Color s_ComplLowerCol0;
            private static readonly Color s_ComplLowerCol1;
            private static readonly Color s_IncomplCol0;
            private static readonly Color s_IncomplCol1;

            public readonly int f_SuggestedControlHeight;
            public readonly int f_UpperHeight;
            public readonly int f_LowerHeight;

            public readonly Font f_Font;
            public readonly StringFormat f_StrFrmt;

            public readonly Brush f_TextBrush;
            public readonly Pen f_AltFramePen;
            public readonly LinearGradientBrush f_ComplUpperGradient;
            public readonly LinearGradientBrush f_ComplLowerGradient;
            public readonly LinearGradientBrush f_IncomplGradient;

            static BarBrushes()
            {
                s_TextColor = Color.FromArgb(255, 255, 255);
                s_AltFrameColor = Color.FromArgb(101, 152, 202);

                s_ComplUpperCol0 = Color.FromArgb(101, 162, 225);
                s_ComplUpperCol1 = Color.FromArgb(82, 134, 185);
                s_ComplLowerCol0 = Color.FromArgb(74, 120, 174);
                s_ComplLowerCol1 = Color.FromArgb(55, 92, 134);
                s_IncomplCol0 = Color.FromArgb(95, 95, 95);
                s_IncomplCol1 = Color.FromArgb(55, 55, 55);
            }

            public BarBrushes(int a_SuggestedControlHeight)
            {
                f_SuggestedControlHeight = Math.Max(4, a_SuggestedControlHeight);
                int l_InnerHeight = (f_SuggestedControlHeight - 2);
                f_UpperHeight = (l_InnerHeight / 2);
                f_LowerHeight = (l_InnerHeight - f_UpperHeight);

                f_Font = new Font(FontFamily.GenericSansSerif, Lyt.CalcFontPointSize(7.5f), FontStyle.Regular);
                f_StrFrmt = BarDrawer.CreateCenterStringFormat();

                f_TextBrush = new SolidBrush(s_TextColor);
                f_AltFramePen = new Pen(s_AltFrameColor);

                const LinearGradientMode lc_Mode = LinearGradientMode.Vertical;
                Rectangle l_Rect = new Rectangle(0, 0, int.MaxValue, f_UpperHeight);
                f_ComplUpperGradient = new LinearGradientBrush(l_Rect, s_ComplUpperCol0, s_ComplUpperCol1, lc_Mode);

                l_Rect.Height = f_LowerHeight;
                f_ComplLowerGradient = new LinearGradientBrush(l_Rect, s_ComplLowerCol0, s_ComplLowerCol1, lc_Mode);

                l_Rect.Height = l_InnerHeight;
                f_IncomplGradient = new LinearGradientBrush(l_Rect, s_IncomplCol0, s_IncomplCol1, lc_Mode);
            }

            public void Dispose()
            {
                f_Font.Dispose();
                f_StrFrmt.Dispose();
                f_TextBrush.Dispose();
                f_AltFramePen.Dispose();
                f_ComplUpperGradient.Dispose();
                f_ComplLowerGradient.Dispose();
                f_IncomplGradient.Dispose();
            }
        }
    }

    static class Lyt
    {
        public static readonly Size s_DisplayBounds;
        public static readonly SizeF s_DisplayPpi;
        public static readonly SizeF s_DesignPpi;
        public static readonly bool s_IsNative;

        private static readonly float s_FactorX;
        private static readonly float s_FactorY;

        public static readonly ushort s_TinySpace;
        public static readonly ushort s_TinySpace2;
        public static readonly ushort s_SmallSpace;
        public static readonly ushort s_SmallSpace2;
        public static readonly ushort s_Space;
        public static readonly ushort s_Space2;
        public static readonly ushort s_LargeSpace;
        public static readonly ushort s_LargeSpace2;

        public static readonly ushort s_DefaultButtonWidth;
        public static readonly ushort s_GenerousButtonWidth;

#if MONO
        private static readonly Graphics s_G;
        private static readonly int s_ButtonPlus;
#endif

        public static readonly Size s_SmallIconSize;

        public static ushort sp_ControlIndentation
        {
            get { return s_LargeSpace; }
        }

        static Lyt()
        {
            s_DisplayBounds = Screen.PrimaryScreen.Bounds.Size;

            Bitmap l_TempMap = new Bitmap(1, 1);
            {
                s_DisplayPpi = new SizeF(l_TempMap.HorizontalResolution, l_TempMap.VerticalResolution);
            }
            l_TempMap.Dispose();

            s_DesignPpi = new SizeF(96f, 96f);
            s_IsNative = (s_DesignPpi == s_DisplayPpi);
            CalcFactors(s_DisplayPpi, out s_FactorX, out s_FactorY);

            s_TinySpace = (ushort)Calc(2);
            s_SmallSpace = (ushort)Calc(5);
            s_Space = (ushort)Calc(10);
            s_LargeSpace = (ushort)Calc(15);

            s_TinySpace2 = (ushort)(2 * s_TinySpace);
            s_SmallSpace2 = (ushort)(2 * s_SmallSpace);
            s_Space2 = (ushort)(2 * s_Space);
            s_LargeSpace2 = (ushort)(2 * s_LargeSpace);

            s_DefaultButtonWidth = (ushort)CalcX(100);
            s_GenerousButtonWidth = (ushort)CalcX(160);

#if MONO
            s_G = Graphics.FromImage(new Bitmap(1, 1));
            s_ButtonPlus = Calc(7);
#endif

            s_SmallIconSize = Lyt.TranformSize(16, 16);
        }

        private static void CalcFactors(SizeF a_DestPpi,
                                        out float a_FactorX,
                                        out float a_FactorY)
        {
            a_FactorX = (a_DestPpi.Width / s_DesignPpi.Width);
            a_FactorY = (a_DestPpi.Height / s_DesignPpi.Height);
        }

        public static float CalcFontPointSize(float a_PointSize)
        {
            return (a_PointSize * s_FactorY);
        }

        public static int Calc(int a_Num)
        {
            return CalcX(a_Num);
        }

        public static int CalcX(int a_DesignWidth)
        {
            return Util.RoundDoubleToInt(a_DesignWidth * s_FactorX);
        }

        public static int CalcY(int a_DesignHeight)
        {
            return Util.RoundDoubleToInt(a_DesignHeight * s_FactorY);
        }

        public static int InvCalcX(int a_Width)
        {
            return Util.RoundDoubleToInt(a_Width / s_FactorX);
        }

        public static int InvCalcY(int a_Height)
        {
            return Util.RoundDoubleToInt(a_Height / s_FactorY);
        }

        public static Point TranformPoint(Point a_Pos)
        {
            return new Point(CalcX(a_Pos.X), CalcY(a_Pos.Y));
        }

        public static Size TranformSize(int a_DesignWidth,
                                        int a_DesignHeight)
        {
            return TranformSize(new Size(a_DesignWidth, a_DesignHeight));
        }

        public static Size TranformSize(Size a_Size)
        {
            return new Size(CalcX(a_Size.Width), CalcY(a_Size.Height));
        }

        public static void SetFontAndHeight(Control a_Cntrl,
                                            Font a_Font)
        {
            a_Cntrl.Font = a_Font;
            bool l_NoText = (0 == a_Cntrl.Text.Length);

            if (true == l_NoText)
            {
                a_Cntrl.Text = "Dummy";
            }

            int l_Height;

#if MONO
            // Mono v2.10 doesn't correctly implement GetPreferredSize(),
            // so we use a different way.

            l_Height = MonoGetPreferredTextSize(a_Cntrl, a_Cntrl.Width).Height;

            if (true == a_Cntrl.GetType().IsSubclassOf(typeof(Button)))
            {
                l_Height += s_ButtonPlus;
            }
#else
            l_Height = a_Cntrl.GetPreferredSize(new Size(a_Cntrl.Width, 0)).Height;
#endif

            a_Cntrl.Height = l_Height;

            if (true == l_NoText)
            {
                a_Cntrl.Text = string.Empty;
            }
        }

#if MONO
        public static Size MonoGetPreferredTextSize(Control a_Cntrl,
                                                    int a_MaxWidth)
        {
            SizeF l_SizeF = s_G.MeasureString(a_Cntrl.Text, a_Cntrl.Font, a_MaxWidth, StringFormat.GenericDefault);

            return new Size((int)l_SizeF.Width + 1, (int)l_SizeF.Height + 1);
        }
#endif
    }

    [Flags]
    enum SelectionChangeOrigins : byte
    {
        None = 0,

        ItemRemoval = 0x1,
        AutoSelect = 0x2,
        Programmatic = 0x4,
        UserInteraction = 0x8,
    }
}

// End of file.
