/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.25-src\Nachtblitz\MainWindow.cs
 * PURPOSE: Main application window logic
 * VERSION: 1.4.25
 * DATE: January 24, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Net;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using NB.Content;
using NB.Gui;
using NB.I2P;
using NB.Language;
using NB.Network;
using NB.Network.eD2k;
using Support;
using Support.TaskBar;

namespace NB.Gui
{
    class MainForm : Dialog
    {
        public const uint c_MaxLogChars = 0x20000;
        private const uint c_AutoSaveIntervalSecs = (3 * Time.c_SecsPerMinute);
        private const uint c_AutoConnectIntervalSecs = Time.c_SecsPerMinute;

        private const uint c_PrefsMagic = 0xD8127498;
        private const uint c_PrefsHeaderSize = 512;

        public const byte c_MainTabsCount = 6;

        private const string c_UpdatePrefsWorkerName = "UpdatePrefsWorker";
        private const string c_LoadAndStartWorkerName = "LAndSWorker";

        public delegate void ProcessShellInputHandler(string a_Stmt);
        public delegate void NotifyDownloadCompletionHandler(string a_FileName,
                                                             string a_FullFileName);
        private delegate void BalloonClickHandler(object a_Context);

        private readonly TextSet f_TextSet;
        public readonly Instance f_Inst;
        public readonly Invoker f_Invoker;
        public readonly BarDrawer f_BarDrawer;
        public TaskBarProgress f_TaskBar;
        public readonly FormInput f_Input;
        private readonly StringBuilder f_LogStrBld;
        public readonly VisualCustomization f_VisualCust;
        private readonly List<ExtractionCookie> f_ExtractionCookies;
        private readonly NotifyIcon f_NotifyIcon;

        public readonly int f_GridViewLineHeight;
        public Font f_GridViewFont;
        private Font f_BoldGridViewFont;
        // f_FileSearchCellStyles is compatible with the ComplStatus
        // enumeration.
        public readonly DataGridViewCellStyle[] f_FileSearchCellStyles;
        public readonly DataGridViewCellStyle f_FriendCellStyle;
        public readonly DataGridViewCellStyle f_BoldCellStyle;
        private readonly List<IGridView> f_InstancedGridViews;

        private readonly HorSplitContainer f_Splitter;
        private readonly GuiMenuStrip f_Menu;
        private readonly AdvancedTabControl f_TabCntrl;

        private readonly ToolStripMenuItem f_ConnMenuItem;
        private readonly ToolStripMenuItem f_ShowConnSecInfoMenuItem;
        private readonly ToolStripMenuItem f_CopyOwnStreamAddressMenuItem;
        private readonly ToolStripMenuItem f_RepublishMenuItem;
        private ConnectMenuFunction f_ConnMenuFunc;
        private readonly ToolStripMenuItem f_OpenOutputDirMenuItem;

        private readonly KademliaPanel f_KadPanel;
        private readonly FileSearchPanel f_SearchPanel;
        private readonly FilesPanel f_FilesPanel;
        private readonly GeneralPeersPanel f_PeersPanel;
        private readonly StatsPanel f_StatsPanel;
        private readonly ReadOnlyTextBox f_LogTextBox;
        private readonly Font f_StatsAndLogPanelFont;

        private readonly System.Windows.Forms.Timer f_Timer;

        private CultureInfo f_GuiCulture;
        // f_InitialShellCmd can be null.
        private string f_InitialShellCmd;

        // f_ActiveOptionsForm is null while the options form is
        // closed.
        private OptionsDialog f_ActiveOptionsForm;
        private int[] f_ScrollStates;
        private FormWindowState f_WndStateBeforeMinimization;

        // f_LoadedPrefs can be null.
        private Preferences f_LoadedPrefs;
        private FormWindowState f_PrevSessionWndState;
        // f_PrefForwardData can be null.
        private Preferences.ForwardData[] f_PrefForwardData;
        private Preferences f_Prefs;
        private BitField f_AcceptedLicenseEditions;
        private MainTabs f_LastSelectedMainTabOfPreviousSession;
        private bool f_NeedRestartToTakeOverChanges;
        public CategoryList f_Categories;
        private ConnData f_ConnData;

        private Time f_AutoSaveTime;
        private byte f_AutoSaveMode;

        private uint f_LastContactInfosGeneration;
        private uint f_LastFilesGeneration;
        private uint f_LastPeersGeneration;
        private uint f_LastStatsGeneration;
        private bool f_TitleUpdateFlag;
        private bool f_CopyStatsToClipboardFlag;

        private BalloonClickHandler f_BalloonClickHandler;
        private object f_BalloonClickContext;

        private ApplicationState f_AppState;
        private bool f_TryingToClose;

        private bool f_NewerVersionInfoIsDisplayed;

        public CultureInfo p_GuiCulture
        {
            // Should only be used by the GUI thread.
            get { return f_GuiCulture; }
        }

        public Preferences p_Prefs
        {
            get { return f_Prefs; }
        }

        public bool p_DisplaySISizes
        {
            get { return f_Prefs.f_GeneralPrefs.f_DisplaySISizes; }
        }

        public int p_GridViewAutoSortItemThreshold
        {
            get { return (int)f_Prefs.f_GeneralPrefs.f_GridViewAutoSortItemThreshold; }
        }

        public bool p_AllowComplFilesDeletion
        {
            get { return f_Prefs.f_GeneralPrefs.f_AllowComplFilesDeletion; }
        }

        public BridgeStatus p_BridgeStatus
        {
            get { return f_ConnData.f_BridgeStatus; }
        }

        public bool p_IsRunning
        {
            get { return (ApplicationState.Started == f_AppState); }
        }

        static MainForm()
        {
            s_TinyBitmap = new Bitmap(1, 1);
            s_MeasurementGraphics = Graphics.FromImage(s_TinyBitmap);
        }

        public MainForm(TextSet a_TextSet,
                        string a_InitialShellCmd,
                        HardMemory a_Mem,
                        HardMemory a_IncomplMem)
        {
            // a_InitialShellCmd may be null.

            f_TextSet = a_TextSet;

            Color l_IncomplFilesColor = Color.FromArgb(255, 153, 35);
            Color l_IncomplFilesColorDark = Color.FromArgb(185, 85, 5);
            Color l_MissingFilesColor = Color.Red;
            Color l_ComplFilesColor = Color.FromArgb(134, 196, 64);
            // l_ComplFilesColorDark is also used for friends in the
            // general peers panel.
            Color l_ComplFilesColorDark = Color.FromArgb(20, 82, 0);
            Color l_ActiveFilesColor = Color.FromArgb(51, 153, 255);
            Color l_FriendPeersColor = Color.FromArgb(0, 57, 159);

            Software.SetMainForm(this);
            f_InitialShellCmd = a_InitialShellCmd;
            f_ScrollStates = new int[OptionsDialog.c_PageCount];

            f_Invoker = new Invoker(this);
            f_BarDrawer = new BarDrawer(this.Font,
                                        Color.Black,
                                        l_IncomplFilesColor,
                                        l_MissingFilesColor,
                                        l_ComplFilesColor,
                                        l_ActiveFilesColor,
                                        Screen.FromControl(this).Bounds.Width);
            f_TaskBar = TaskBarProgress.TryCreate(this.Handle, true);
            f_Input = new FormInput();
            f_LogStrBld = new StringBuilder();
            f_VisualCust = new VisualCustomization(a_TextSet);
            f_ExtractionCookies = new List<ExtractionCookie>();
            f_NotifyIcon = new NotifyIcon();
            f_NotifyIcon.Icon = Software.s_AppIcon;
            f_NotifyIcon.MouseClick += NotifyIcon_MouseClick;
            f_NotifyIcon.BalloonTipClicked += NotifyIcon_BalloonTipClicked;
            this.Icon = Software.s_AppIcon;

            {
                // Create grid view fonts.
                DataGridView l_GridView = new DataGridView();
                DataGridViewCellStyle l_DefStyle = l_GridView.DefaultCellStyle;

                f_GridViewLineHeight = l_GridView.RowTemplate.Height;
                f_GridViewFont = new Font(this.Font, this.Font.Style);
                f_FileSearchCellStyles = new DataGridViewCellStyle[4];
                f_FileSearchCellStyles[0] = CreateCellStyle(l_DefStyle, f_GridViewFont, SystemColors.ControlText);
                f_FileSearchCellStyles[1] = CreateCellStyle(l_DefStyle, f_GridViewFont, l_IncomplFilesColorDark);
                f_FileSearchCellStyles[2] = CreateCellStyle(l_DefStyle, f_GridViewFont, l_MissingFilesColor);
                f_FileSearchCellStyles[3] = CreateCellStyle(l_DefStyle, f_GridViewFont, l_ComplFilesColorDark);
                f_FriendCellStyle = CreateCellStyle(l_DefStyle, f_GridViewFont, l_FriendPeersColor);
                f_BoldGridViewFont = new Font(f_GridViewFont, FontStyle.Bold);
                f_BoldCellStyle = CreateCellStyle(l_DefStyle, f_BoldGridViewFont, SystemColors.ControlText);

                l_GridView.Dispose();
                f_InstancedGridViews = new List<IGridView>();
            }

            // Don't use CultureInfo.CurrentUICulture.
            f_GuiCulture = CultureInfo.CurrentCulture;

            this.SuspendLayout();

            this.Size = base.SuggestWindowSize(Lyt.TranformSize(600, 475));
            this.MinimumSize = Lyt.TranformSize(400, 350);
            this.StartPosition = FormStartPosition.CenterScreen;

            // Startup.

            // Use a dummy persona during startup.
            f_Prefs = Preferences.CreateDefault(OwnUserPersona.CreateDummy());
            UpdateFormTitle(false);

            string l_NodesDatFileName = LongPath.Combine(Software.sp_ModulePath, Software.c_NodesDat);
            string l_NodeLoadErrMsg;
            byte[] l_InitialNodes = ReadNodesFile(l_NodesDatFileName, out l_NodeLoadErrMsg);

            f_Inst = new Instance(a_TextSet,
                                  this,
                                  f_Prefs,
                                  a_Mem,
                                  a_IncomplMem,
                                  l_InitialNodes,
                                  out f_Inst);

            AddToLog(f_TextSet.FormatWithAppName(TextKeys.MsgWelcomeToApp));
            string l_FrameworkStr;
            string l_AdditionalPlatformInfoStr;
            string l_RuntimeStr = Software.GetRuntimeString(out l_FrameworkStr, out l_AdditionalPlatformInfoStr);
            AddToLog(l_RuntimeStr);

            if (null != l_FrameworkStr)
            {
                AddToLog(l_FrameworkStr);
            }

            if (null != l_AdditionalPlatformInfoStr)
            {
                AddToLog(l_AdditionalPlatformInfoStr);
            }

            // Splitter.

            f_Splitter = HorSplitContainer.CreateFixed(this);

            // Menu.

            f_Menu = new GuiMenuStrip();
            f_Splitter.Controls.Add(f_Menu);
            f_Menu.Height = (f_Menu.Font.Height + Lyt.s_Space);
            f_Menu.ImageScalingSize = Lyt.s_SmallIconSize;

            ToolStripMenuItem l_AppMenu = new ToolStripMenuItem(a_TextSet[TextKeys.MainMenuApplication]);
            AddMenuItem(l_AppMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuOptions), MenuShowOptions_Click, Keys.Control | Keys.O, IconTypes.Gear);
            AddMenuSeparator(l_AppMenu);
            AddMenuItem(l_AppMenu, a_TextSet[TextKeys.MainMenuExit], DefaultCancelCloseHandler, Keys.Alt | Keys.F4, IconTypes.Armchair);

            ToolStripMenuItem l_NetworkMenu = new ToolStripMenuItem(a_TextSet[TextKeys.MainMenuNetwork]);
            f_ConnMenuItem = AddMenuItem(l_NetworkMenu, string.Empty, MenuConnect_Click);
            AddMenuItem(l_NetworkMenu, f_TextSet.GetAsDialogOpener(TextKeys.SamConnTestDlgTitle), MenuSamTest_Click);
            f_ShowConnSecInfoMenuItem = AddMenuItem(l_NetworkMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuShowConnSecInfo), MenuShowConnSecInfo_Click, IconTypes.Umbrella);
            f_CopyOwnStreamAddressMenuItem = AddMenuItem(l_NetworkMenu, a_TextSet[TextKeys.MainMenuCopyOwnStrmAddrToClipboard], MenuCopyOwnStreamAddressToClipboard_Click, IconTypes.Pendant);

            AddMenuSeparator(l_NetworkMenu);
            ToolStripMenuItem l_KademliaSubMenu = AddMenuItem(l_NetworkMenu, a_TextSet[TextKeys.MainMenuKademlia], null, IconTypes.RadioTower);
            f_RepublishMenuItem = AddMenuItem(l_KademliaSubMenu, a_TextSet[TextKeys.MainMenuRepublishOwnFiles], MenuRepublish_Click, IconTypes.Bell);
            AddMenuSeparator(l_KademliaSubMenu);
            // FIXME: Make translatable.
            AddMenuItem(l_KademliaSubMenu, f_TextSet.GetAsDialogOpener("Add Address for Bootstrapping"), MenuAddKadAddress_Click);
            AddMenuItem(l_KademliaSubMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuImportNodes), MenuImportNodes_Click);
            AddMenuItem(l_KademliaSubMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuExportNodes), MenuExportNodes_Click);
            AddMenuItem(l_NetworkMenu, a_TextSet[TextKeys.MainMenuResetAllBans], MenuResetBans_Click, IconTypes.Recycle);
            AdjustConnMenuItemsReceptiveness(false);

            AddMenuSeparator(l_NetworkMenu);
            AddMenuItem(l_NetworkMenu, f_TextSet.GetAsDialogOpener("View I2P Router Status"), MenuViewI2PRouterStatus_Click);

            ToolStripMenuItem l_FilesMenu = new ToolStripMenuItem(a_TextSet[TextKeys.MainMenuFiles]);
            AddMenuItem(l_FilesMenu, a_TextSet[TextKeys.MainMenuPasteLinkFromClipboard], MenuPasteLink_Click, Keys.Control | Keys.L, IconTypes.Key);
            AddMenuItem(l_FilesMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuOpenCollection), MenuOpenCollection_Click, IconTypes.LetterOpen);
            AddMenuItem(l_FilesMenu, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuImportPartialDlsFromiMule), MenuImportIncomplFilesFromiMule_Click, IconTypes.Download);
            AddMenuSeparator(l_FilesMenu);
            AddMenuItem(l_FilesMenu, a_TextSet[TextKeys.MainMenuRescanSharedDirs], MenuRefreshSharedFiles_Click, Keys.F9, IconTypes.Camera);
            f_OpenOutputDirMenuItem = AddMenuItem(l_FilesMenu, "Browse Output Directory for Downloaded Files", MenuOpenOutputDirectory_Click);
            f_OpenOutputDirMenuItem.Enabled = false;

            ToolStripMenuItem l_Info = new ToolStripMenuItem(a_TextSet[TextKeys.MainMenuInfo]);
            AddMenuItem(l_Info, a_TextSet[TextKeys.MainMenuCopyStatsToClipboard], MenuCopyStatsToClipboard_Click, IconTypes.PieChart);
            AddMenuItem(l_Info, a_TextSet[TextKeys.MainMenuCopyLogToClipboard], MenuCopyLogToClipboard_Click, IconTypes.SheetPencil);
            AddMenuSeparator(l_Info);
            AddMenuItem(l_Info, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuShowLicense), MenuViewLicenseDialog_Click);
            AddMenuItem(l_Info, a_TextSet.GetAsDialogOpener(TextKeys.MainMenuViewReleaseNotes), MenuViewReleaseNotes_Click);
            AddMenuItem(l_Info, a_TextSet.GetAsDialogOpener(a_TextSet.FormatWithAppName(TextKeys.MainMenuAbout)), MenuShowAboutDialog_Click, IconTypes.Info);

            f_Menu.Items.Add(l_AppMenu);
            f_Menu.Items.Add(l_NetworkMenu);
            f_Menu.Items.Add(l_FilesMenu);
            f_Menu.Items.Add(l_Info);
            this.MainMenuStrip = f_Menu;

            // Tab control.

            f_TabCntrl = new AdvancedTabControl(f_Splitter);
            f_TabCntrl.p_IncludeHiddenPagesInAutoPageListing = false;
            f_Splitter.SetControls(f_Menu, f_TabCntrl);
            f_Splitter.SetUpperHeight(f_Menu.Height);

            f_KadPanel = new KademliaPanel(f_Inst);
            f_SearchPanel = new FileSearchPanel(f_Inst);
            f_FilesPanel = new FilesPanel(f_Inst);
            f_PeersPanel = new GeneralPeersPanel(f_Inst, f_FileSearchCellStyles[0], f_FriendCellStyle);
            f_StatsPanel = new StatsPanel(f_Inst);
            AdvancedTabPage l_LogPanel = new AdvancedTabPage();

            l_LogPanel.p_PageVisible = f_PeersPanel.p_PageVisible = f_KadPanel.p_PageVisible = false;

            f_TabCntrl.AddPageWithoutSelecting(f_KadPanel, IconTypes.RadioTower);
            f_TabCntrl.AddPageWithoutSelecting(f_SearchPanel, IconTypes.Magnifier);
            f_TabCntrl.AddPageWithoutSelecting(f_FilesPanel, IconTypes.File);
            f_TabCntrl.AddPageWithoutSelecting(f_PeersPanel, IconTypes.Person);
            f_TabCntrl.AddPageWithoutSelecting(f_StatsPanel, IconTypes.PieChart);
            f_TabCntrl.AddPageWithoutSelecting(l_LogPanel, IconTypes.SheetPencil);
            f_LogTextBox = new ReadOnlyTextBox(l_LogPanel);
            f_LogTextBox.Dock = DockStyle.Fill;

            // Increase the stats and log font.

            Font l_FormFont = this.Font;
            float l_LogFontSize = (l_FormFont.SizeInPoints + 0.5f);

            if (false == CreateFont("Verdana", l_LogFontSize, l_FormFont.Style, out f_StatsAndLogPanelFont))
            {
                f_StatsAndLogPanelFont = RecreateFont(l_FormFont, l_LogFontSize, l_FormFont.Style);
            }

            f_StatsPanel.SetTextBoxFont(f_StatsAndLogPanelFont);
            f_LogTextBox.Font = f_StatsAndLogPanelFont;

            // Setup event handlers.

            this.KeyPreview = true;
            f_TabCntrl.e_SelectedPageChanged += TabCntrl_SelectedPageChanged;

            SelectMainTab(MainTabs.Files);

            this.ResumeLayout();

            f_Timer = new System.Windows.Forms.Timer();
            f_Timer.Tick += ProcessGui;
            f_Timer.Interval = 500;

            // Set Texts.
            f_KadPanel.p_Caption = a_TextSet[TextKeys.KademliaTabTitle];
            f_SearchPanel.p_Caption = a_TextSet[TextKeys.FileSearchTabTitle];
            f_FilesPanel.p_Caption = a_TextSet[TextKeys.FilesTabTitle];
            f_PeersPanel.p_Caption = a_TextSet[TextKeys.PeersTabTitle];
            f_StatsPanel.p_Caption = a_TextSet[TextKeys.StatsTabTitle];
            l_LogPanel.p_Caption = a_TextSet[TextKeys.LogTabTitle];
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing &&
                false == this.IsDisposed)
            {
                DisposeMenuItems(f_Menu);
                f_Menu.Dispose();
                f_LogTextBox.Dispose();
                f_TabCntrl.Dispose();
                f_Splitter.Dispose();

                f_BarDrawer.Dispose();
                f_Invoker.TidyUp();
                f_NotifyIcon.Dispose();

                f_GridViewFont.Dispose();
                f_BoldGridViewFont.Dispose();

                f_Categories.Dispose();

                f_StatsAndLogPanelFont.Dispose();

                if (null != f_TaskBar)
                {
                    f_TaskBar.Dispose();
                }
            }

            base.Dispose(disposing);
        }

        private bool ShouldAcceptExternalCall()
        {
            return (ApplicationState.Started == f_AppState);
        }

        protected override void OnLoad(EventArgs e)
        {
            // Need to call base class implementation.

            // Since the StartPosition property doesn't really seem to
            // work for forms that are launched through Application.Run(),
            // we center the form manually.
            Rectangle l_Rect = Screen.FromControl(this).WorkingArea;
            this.Location = new Point((l_Rect.Width - Width) / 2, (l_Rect.Height - Height) / 2);

            base.OnLoad(e);
        }

        protected override void OnResize(EventArgs e)
        {
            // Need to call base class implementation.

            if (ApplicationState.Started == f_AppState)
            {
                if (false == IsMinimized())
                {
                    FormWindowState l_CurWndState = this.WindowState;

                    if (l_CurWndState != f_WndStateBeforeMinimization)
                    {
                        f_WndStateBeforeMinimization = l_CurWndState;
                    }

                    if (null == f_TaskBar)
                    {
                        f_TaskBar = TaskBarProgress.TryCreate(this.Handle, true);
                    }
                }
                else
                {
                    if (true == f_Prefs.f_GeneralPrefs.f_MinimizeToTray)
                    {
                        if (null != f_TaskBar)
                        {
                            f_TaskBar.Dispose();
                            f_TaskBar = null;
                        }

                        this.Visible = false;
                    }
                }
            }

            base.OnResize(e);
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            if (ApplicationState.PreStart == f_AppState)
            {
                StartupHandler();
            }

            base.OnShown(e);
        }

        private void StartupHandler()
        {
            f_AppState = ApplicationState.StartingUp;

            f_KadPanel.ArrangeLayoutOnShown();
            f_SearchPanel.ArrangeLayoutOnShown();
            f_FilesPanel.ArrangeLayoutOnShown();
            f_PeersPanel.ArrangeLayoutOnShown();

            // We load the preferences as late as now, since we will
            // right away create a new persona if there is no valid
            // persona yet (and for this task we need a working GUI).

            StorageStates l_State = f_Inst.ReadFromElement(new ElmKey(Instance.c_PrefsElmKey), LoadPreferences);

            if (StorageStates.Success != l_State)
            {
                if (StorageStates.NoData != l_State)
                {
                    Util.DebugBreak();
                }

                OwnUserPersona l_NewPersona;
                // The return value is maningless, since f_AppState
                // was not set to ApplicationState.Started yet.
                bool l_Junk = KeyGenerationWaitDialog.ShowDialog(f_TextSet, this, false, out l_NewPersona);

                f_LoadedPrefs = Preferences.CreateDefault(l_NewPersona);
                f_PrevSessionWndState = FormWindowState.Normal;
                f_PrefForwardData = null;
                f_Categories = CategoryList.CreateDefault();
                f_VisualCust.Reset();

                // Create the default preview folder in the temp dir
                // if it doesn't exist.
                bool l_PrevFolderCreated = Util.EnsureDirectoryExists(f_LoadedPrefs.f_GeneralPrefs.f_PreviewFolder);
            }

            // Delay a possible call to AddToLogAndShowBalloon() until
            // after the preferences were taken over (in order to
            // initialize f_NotifyIcon first).
            bool l_ShowWarningThatSharedDirsWereChanged = false;
            bool l_WasPreviouslyConnected = (null != f_LoadedPrefs.f_GeneralPrefs.p_DatagramKeyPair ||
                                             null != f_LoadedPrefs.f_GeneralPrefs.p_StreamKeyPair);

            // Make sure the list of shared dirs has full integrity.
            // Do this after AdjustCaseSensitivity() was possibly
            // called.

            SharedDirectories l_SafeSharedDirs = f_LoadedPrefs.f_IOPrefs.f_SharedDirs.RecreateWithEnforcedIntegrity();

            if (false == f_LoadedPrefs.f_IOPrefs.f_SharedDirs.ValueEquals(l_SafeSharedDirs))
            {
                // We need to modify the shared dirs in order to get
                // rid of overlapping entries.

                IOPrefs l_NewIOPrefs;
                f_LoadedPrefs.f_IOPrefs.RecreateWithNewSharedDirs(l_SafeSharedDirs, out l_NewIOPrefs);
                f_LoadedPrefs = f_LoadedPrefs.RecreateWithNewIOPrefs(ref l_NewIOPrefs);

                l_ShowWarningThatSharedDirsWereChanged = true;
            }

            f_LoadedPrefs.f_IOPrefs.f_SharedDirs.DebugAssertIntegrity();

            // Restore the window state of the previous session.

            if (FormWindowState.Maximized == f_PrevSessionWndState)
            {
                this.WindowState = FormWindowState.Maximized;
            }

            // Load and start the content and the network manager.

            LoadAndStartContext l_Cntxt = new LoadAndStartContext();

            Util.StartThreadWithParam(c_LoadAndStartWorkerName, LoadAndStartWorker, System.Threading.ThreadPriority.Normal, l_Cntxt);

            bool l_Running = WaitMessageDialog.ShowDialog(f_TextSet, this, l_Cntxt, WaitMessageDialog.ProgressBarStyles.NoProgressBar, TextKeys.LoadingAppDataDlgTitle, TextKeys.LoadingAppDataDlgMsg);

            // Start timer.

            f_AutoSaveTime = f_Inst.GetTime().AddSeconds(c_AutoSaveIntervalSecs);
            f_Timer.Start();
            GuiGroupBox.s_MouseTimer.Start();
            f_AppState = ApplicationState.Started;

            // Apply preferences.

            UpdatePreferences(this, f_LoadedPrefs, true);
            f_LoadedPrefs = null;

            // Apply main tab preference.

            {
                MainTabs l_NewMainTab;

                if ((MainTabs)c_MainTabsCount <= f_Prefs.f_GeneralPrefs.f_PreferredMainTab)
                {
                    // In this case we will set the main tab that was
                    // last selected in the previous session.

                    if ((MainTabs)c_MainTabsCount <= f_LastSelectedMainTabOfPreviousSession)
                    {
                        // The stored value from the last session is invalid.

                        l_NewMainTab = (MainTabs)f_TabCntrl.p_SelectedPageIndex;
                    }
                    else
                    {
                        l_NewMainTab = f_LastSelectedMainTabOfPreviousSession;
                    }
                }
                else
                {
                    // The user has a fixed preferrence.

                    l_NewMainTab = f_Prefs.f_GeneralPrefs.f_PreferredMainTab;
                }

                if ((MainTabs)f_TabCntrl.p_SelectedPageIndex != l_NewMainTab)
                {
                    AdvancedTabPage l_SelPage = (AdvancedTabPage)f_TabCntrl.GetPageByIndex((uint)l_NewMainTab);
                    bool l_Selected = f_TabCntrl.TrySelectNearbyVisiblePage(l_SelPage, SelectionChangeOrigins.Programmatic);

                    this.Refresh();
                }
            }

            // Apply stored visual appearance.

            f_KadPanel.FinishLoading();
            f_SearchPanel.FinishLoading();
            f_FilesPanel.FinishLoading();
            f_PeersPanel.FinishLoading();

            // Handle search history loading (if any).

            if (false == f_Prefs.f_GeneralPrefs.f_SaveSearchHistory)
            {
                l_State = StorageStates.NoData;
            }
            else
            {
                l_State = f_Inst.ReadFromElement(new ElmKey(Instance.c_SearchCacheElmKey), f_SearchPanel.LoadSearchCache);

                if (StorageStates.Success == l_State)
                {
                    f_SearchPanel.SetSearchResultsFilter(f_Prefs.f_KadPrefs.f_KadFilter);
                }
                else if (StorageStates.NoData != l_State)
                {
                    Util.DebugBreak();
                }
            }

            if (StorageStates.Success != l_State)
            {
                f_SearchPanel.ResetSearchCache();
            }

            // Handle shell input.

            if (null != f_InitialShellCmd)
            {
                HandleShellInput(f_InitialShellCmd, InputSource.OwnCommandLine);
                f_InitialShellCmd = null;
            }

            // Manually call the timer's handler.

            f_Splitter.Enabled = true;
            ProcessGui(null, null);

            if (true == l_ShowWarningThatSharedDirsWereChanged)
            {
                AddToLogAndShowBalloon(f_TextSet[TextKeys.TitleListOfSharedDirsAutoModified], f_TextSet[TextKeys.MsgListOfSharedDirsAutoModified], true);
            }
            else if ((366 / 2) < Software.s_ReleaseInfo.GetReleaseAgeInDays())
            {
                // This release seems to older than six month old.

                AddToLogAndShowBalloon(f_TextSet[TextKeys.TitleReleaseOlderThanSixMonth], f_TextSet.Format(TextKeys.MsgReleaseOlderThanSixMonth, f_TextSet.f_AppName, Software.c_ProjectURL), true);
            }

            if (false == l_WasPreviouslyConnected)
            {
                SamBridgeTestDialog.ShowModal(f_Inst, this);
            }
        }

        private void LoadAndStartWorker(object a_Obj)
        {
            LoadAndStartContext l_Cntxt = (LoadAndStartContext)a_Obj;

            StorageStates l_ContentMngrStatus = f_Inst.f_ContentMngr.LoadFiles();

            if (StorageStates.Success != l_ContentMngrStatus &&
                StorageStates.NoData != l_ContentMngrStatus)
            {
                Util.DebugBreak();
            }

            f_Inst.f_ContentMngr.Start();
            f_Inst.f_NetworkMngr.Start();

            l_Cntxt.SetFinished();
        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            // Need to call base class implementation.

            ShutdownHandler(e);

            base.OnFormClosing(e);
        }

        private void ShutdownHandler(FormClosingEventArgs e)
        {
            if (ApplicationState.Started < f_AppState)
            {
                // Don't allow the form to close so early, but close
                // as soon as possible.

                f_TryingToClose = true;
                e.Cancel = true;

                return;
            }
            else if (ApplicationState.Started != f_AppState)
            {
                // Only begin shutting down once and ignore further/
                // recursive calls (they can happen).

                return;
            }

            this.RestoreIfMinimized(false);

            bool l_MakeHaste = (CloseReason.UserClosing != e.CloseReason);

            if (true == f_Prefs.f_GeneralPrefs.f_PromptOnExit &&
                false == l_MakeHaste &&
                false == YesNoMsgBox(f_TextSet.FormatWithAppName(TextKeys.QuestionReallyExitApp)))
            {
                e.Cancel = true;

                return;
            }

            ResetTaskBarState();

            f_AppState = ApplicationState.ShutDownInProgress;
            f_Timer.Stop();
            GuiGroupBox.s_MouseTimer.Stop();
            UpdateFormTitle(false);

            // Save preferences and search cache.

            StorageStates l_State = SavePreferences();
            l_State = SaveSearchCache();

            // Actually shut down.

            bool l_FindLeaks = (false == l_MakeHaste &&
                                true == f_Prefs.f_GeneralPrefs.f_CompactOnExit);
            GeneralShutdownContext l_ShutdownCntxt = new GeneralShutdownContext(f_Inst, l_FindLeaks,  f_Prefs.f_GeneralPrefs.f_CompactOnExit);
            l_ShutdownCntxt.BeginAsyncShutdown();
            bool l_Running = WaitMessageDialog.ShowDialog(f_TextSet, this, l_ShutdownCntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.ShutdownDlgTitle, f_TextSet.FormatWithAppName(TextKeys.ShutdownDlgMsg));

            foreach (ExtractionCookie l_Cookie in f_ExtractionCookies)
            {
                // This will delete the file.
                l_Cookie.f_TempFile.Close();
            }

            f_ExtractionCookies.Clear();

            // This tells any parts of the GUI that are busy in
            // parallel to quit as soon as possible.
            f_AppState = ApplicationState.ShutDownFinished;

            BootConfig l_BootConfig = Software.s_BootConfig;

            if (true == l_BootConfig.HasApplicationDataDirectoryChanged())
            {
                if (false == Software.DoesAnyAppDataFileExist(l_BootConfig.p_NewAppDataDir))
                {
                    using (MoveDataFilesDialog l_Dlg = new MoveDataFilesDialog(f_TextSet, this, p_DisplaySISizes))
                    {
                        l_Dlg.HandleMoving(this);
                    }
                }
                else
                {
                    // The destination dir already contains application
                    // data, so don't move anything, just save the new
                    // location.

                    if (false == l_BootConfig.TakeOverNewApplicationDataDirectoryAndSave())
                    {
                        ErrMsgBox(f_TextSet[TextKeys.MsgErrWhileWritingBootConfig]);
                    }
                    else
                    {
                        InfoMsgBox(f_TextSet.Format(TextKeys.MoveDataFilesDlgMsgDirChangeSuccess, l_BootConfig.p_AppDataDir));
                    }
                }
            }

            f_TabCntrl.e_SelectedPageChanged -= TabCntrl_SelectedPageChanged;

#if MONO
            // On Mono we need this to really quit.
            Application.ExitThread();
#endif
        }

        private void NotifyIcon_MouseClick(object sender, MouseEventArgs e)
        {
            if (MouseButtons.Left == e.Button)
            {
                RestoreIfMinimized(true);
            }
        }

        private bool IsMinimized()
        {
            return (FormWindowState.Minimized == this.WindowState);
        }

        private void RestoreIfMinimized(bool a_BringToForeground)
        {
            if (true == IsMinimized())
            {
                // The window may be invisible, if so most likely from a
                // minimize to tray operation.
                this.Visible = true;

                this.WindowState = f_WndStateBeforeMinimization;
            }

            if (true == a_BringToForeground)
            {
                this.TopMost = true;
                this.TopMost = false;
                this.BringToFront();
                this.Activate();
            }
        }

        public void ExternalFocusMainForm()
        {
            if (true == ShouldAcceptExternalCall())
            {
                FocusMainForm();
            }
        }

        private void FocusMainForm()
        {
            // This routine is usually called at the request of another
            // process.

            RestoreIfMinimized(true);
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            // Need to call base class implementation.

            if (ApplicationState.Started == f_AppState)
            {
                Keys l_Modifiers = e.Modifiers;

                if (Keys.None == l_Modifiers)
                {
                    Keys l_KeyCode = e.KeyCode;

                    if (Keys.F1 <= l_KeyCode &&
                        Keys.F6 >= l_KeyCode)
                    {
                        MainTabs l_MainTab = (MainTabs)(l_KeyCode - Keys.F1);
                        SelectMainTab(l_MainTab);

                        e.Handled = true;
                    }
                    else if (Keys.Escape == l_KeyCode)
                    {
                        if (FormWindowState.Minimized != this.WindowState)
                        {
                            this.WindowState = FormWindowState.Minimized;

                            e.Handled = true;
                        }
                    }
                }
                else if (Keys.Control == l_Modifiers &&
                         Keys.Q == e.KeyCode &&
                         true == f_Prefs.f_GeneralPrefs.f_AdditionalKeyCombinationForExiting &&
                         1 == Application.OpenForms.Count)
                {
                    Close();

                    e.Handled = true;
                }
            }

            base.OnKeyDown(e);
        }

        public string BuildDefaultSizeStr(ulong a_Bytes)
        {
            return SizeFormatter.BuildSizeStr(p_DisplaySISizes, false, false, true, a_Bytes);
        }

        public string BuildLongSizeStr(ulong a_Bytes)
        {
            return SizeFormatter.BuildSizeStr(p_DisplaySISizes, true, false, true, a_Bytes);
        }

        public void RegisterGridView(IGridView a_GridView)
        {
            f_InstancedGridViews.Add(a_GridView);
        }

        public void UnregisterGridView(IGridView a_GridView)
        {
            bool l_OK = f_InstancedGridViews.Remove(a_GridView);
        }

        public void UpdatePreferences(IWin32Window a_Owner,
                                      Preferences a_Prefs,
                                      bool a_StartingUp)
        {
            // This routine may only be used by the GUI. The caller
            // should check p_IsRunning after return, unless control
            // is directly returned to the message loop.

            PrefChangeKinds l_PrefsChangeKind = Preferences.DetermineChangeKind(f_Prefs, a_Prefs);

            if (false == a_StartingUp)
            {
                if (PrefChangeKinds.AppRestartNeededToFullyTakeOver == l_PrefsChangeKind ||
                    true == Software.s_BootConfig.IsDirty(f_TextSet))
                {
                    f_NeedRestartToTakeOverChanges = true;
                }

                if (BridgeStatus.Disconnected != f_ConnData.f_BridgeStatus &&
                    PrefChangeKinds.ReconnectNeededToFullyTakeOver == l_PrefsChangeKind)
                {
                    f_ConnData.f_ConnectionIsDirty = true;
                }

                if (false == f_NeedRestartToTakeOverChanges)
                {
                    if (true == f_ConnData.f_ConnectionIsDirty)
                    {
                        InfoMsgBox(f_TextSet[TextKeys.MsgTakeOverRequiresReconnect]);
                    }
                }
                else
                {
                    InfoMsgBox(f_TextSet[TextKeys.MsgTakeOverRequiresAppRestart]);
                }
            }

            if (PrefChangeKinds.NoChangeAtAll == l_PrefsChangeKind)
            {
                return;
            }

            // Take over changes.

            f_Prefs = a_Prefs;

            bool l_HadBarBgImages = f_FilesPanel.HasDetailsPanelBarBackgroundImages();
            bool l_BarBgImgsToggled = (f_Prefs.f_GeneralPrefs.f_UseBackgroundImagesForComplBar != l_HadBarBgImages);

            UpdatePreferencesContext l_Cntxt = new UpdatePreferencesContext(true == l_BarBgImgsToggled &&
                                                                            false == l_HadBarBgImages);

            Util.StartThreadWithParam(c_UpdatePrefsWorkerName, UpdatePreferencesWorker, System.Threading.ThreadPriority.Normal, l_Cntxt);

            // Do some GUI work in the meantime.
            UpdatePreferencesGUI(a_StartingUp);

            bool l_Running = WaitMessageDialog.ShowDialog(f_TextSet, a_Owner, l_Cntxt, WaitMessageDialog.ProgressBarStyles.NoProgressBar, TextKeys.OptionsTakeOverDlgTitle, TextKeys.OptionsTakeOverDlgMsg);

            // Finally handle background images for the completion
            // bar (if required, they were loaded in parallel).

            if (true == l_BarBgImgsToggled)
            {
                BarBackgroundImages l_BarBgImgs = l_Cntxt.GetBarBackgroundImages();
                f_FilesPanel.SetDetailsPanelBarBackgroundImages(l_BarBgImgs);
            }
        }

        private void UpdatePreferencesGUI(bool a_StartingUp)
        {
            if (false == a_StartingUp)
            {
                f_SearchPanel.SetSearchResultsFilter(f_Prefs.f_KadPrefs.f_KadFilter);
            }

            f_NotifyIcon.Visible = f_Prefs.f_GeneralPrefs.f_ShowTrayIcon;
            f_SearchPanel.SetSizeFormat(p_DisplaySISizes);
            f_OpenOutputDirMenuItem.Enabled = (null != f_Prefs.f_IOPrefs.f_SharedDirs.GetCompletionTargetPath());

            // Update grid view font.

            f_GridViewFont.Dispose();
            f_BoldGridViewFont.Dispose();

            f_GridViewFont = f_Prefs.f_GeneralPrefs.BuildGridViewFont(this.Font);

            foreach (DataGridViewCellStyle l_Style in f_FileSearchCellStyles)
            {
                l_Style.Font = f_GridViewFont;
            }

            f_FriendCellStyle.Font = f_GridViewFont;
            f_BoldGridViewFont = new Font(f_GridViewFont, FontStyle.Bold);
            f_BoldCellStyle.Font = f_BoldGridViewFont;

            foreach (IGridView l_GridView in f_InstancedGridViews)
            {
                l_GridView.TakeOverFontAndBorderStyle();
            }

            // We have to do update the icon storage using the GUI
            // thread.

            Color l_NewIconBgColor = ColorPicker.ChooseColorEasy(f_Prefs.f_GeneralPrefs.f_IconBgColor, IconStorage.s_SuggestedBgColor);

            if (true == Software.s_IconStorage.SetBackgroundColor(l_NewIconBgColor))
            {
                f_Categories.NotifyDefaultIconBgColorChanged();

                f_TabCntrl.UpdateImagesOfAllPages();
                f_FilesPanel.UpdateImagesOfAllPages();
                f_SearchPanel.UpdateImagesOfAllPages();
            }

            bool l_AdvancedMode = f_Prefs.f_GeneralPrefs.f_UIModeAdvanced;
            f_TabCntrl.AdjustPageVisibility(f_KadPanel, l_AdvancedMode, SelectionChangeOrigins.Programmatic);
            f_TabCntrl.AdjustPageVisibility(f_PeersPanel, l_AdvancedMode, SelectionChangeOrigins.Programmatic);
            f_TabCntrl.AdjustPageVisibility(f_TabCntrl.GetPageByIndex((uint)MainTabs.Log), l_AdvancedMode, SelectionChangeOrigins.Programmatic);

            // Update file associations.

            WindowsIntegration.ApplyAssociations(ref f_Prefs.f_AssocPrefs);
        }

        private void UpdatePreferencesWorker(object a_Obj)
        {
            // This routine handles update tasks that do not require
            // the GUI thread.

            UpdatePreferencesContext l_Cntxt = (UpdatePreferencesContext)a_Obj;

            GeneralAPCCookie l_NetworkCookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiUpdatePreferences, f_Prefs);
            GeneralAPCCookie l_ContentCookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiUpdatePreferences, f_Prefs);

            f_Inst.f_NetworkMngr.EnqueueAPC(l_NetworkCookie);
            f_Inst.f_ContentMngr.EnqueueAPC(l_ContentCookie);

            // Load images in the meantime, if required.

            BarBackgroundImages l_BarBgImages = null;

            // NOTE: Mono v2.10.8 seems not to be able to resize an
            // image (it throws an InvalidOperationException).
#if !MONO
            if (true == l_Cntxt.f_NeedBarBgImgs)
            {
                string l_ProgressBarBgFolder = LongPath.Combine(Software.sp_ModulePath, Software.c_ProgressBarBgFolder);
                l_BarBgImages = BarBackgroundImages.LoadFromDisk(l_ProgressBarBgFolder, f_BarDrawer.f_MaxAvailWidth);
            }
#endif

            l_NetworkCookie.WaitForCompletion();
            l_ContentCookie.WaitForCompletion();

            l_Cntxt.SetFinished(l_BarBgImages);
        }

        private bool ManageSamConnection(Time a_Time)
        {
            // Do not continue processing if false is returned.

            if (true == f_ConnData.f_ConnectionLocked)
            {
                return true;
            }

            f_ConnData.CheckForOutdatedRelease(a_Time);

            if (true == f_ConnData.p_DisableNetworkCapabilities)
            {
                f_ConnMenuFunc = ConnectMenuFunction.ManualConnect;
                f_ConnMenuItem.Text = f_TextSet[TextKeys.MainMenuConnect];
                f_ConnMenuItem.ShortcutKeys = Keys.None;

                if (BridgeStatus.Disconnected == p_BridgeStatus)
                {
                    return true;
                }

                // Close the connection.

                bool l_StillConnected;

                if (false == Disconnect(a_Time, false, out l_StillConnected))
                {
                    return false;
                }

                if (true == l_StillConnected)
                {
                    return true;
                }

                ErrMsgBox(Software.BuildObsoleteClientMessage(false));

                return p_IsRunning;
            }

            if (BridgeStatus.Disconnected == f_ConnData.f_BridgeStatus)
            {
                bool l_NeedToConnect;

                switch (f_Prefs.f_GeneralPrefs.f_ConnStartKind)
                {
                    case GeneralPrefs.ConnStartKinds.ManualStart:
                        l_NeedToConnect = f_ConnData.f_ConnectOnce;
                        break;

                    case GeneralPrefs.ConnStartKinds.ManualStartButRepair:
                        l_NeedToConnect = (true == f_ConnData.f_ConnectOnce ||
                                           (true == f_ConnData.f_ConnRepairMeaningful &&
                                            true == f_ConnData.CanAutoConnect(a_Time)));
                        break;

                    case GeneralPrefs.ConnStartKinds.AutoStart:
                        l_NeedToConnect = (true == f_ConnData.f_ConnectOnce ||
                                           true == f_ConnData.CanAutoConnect(a_Time));
                        break;

                    default: throw (new ApplicationException());
                }

                if (true == l_NeedToConnect)
                {
                    f_ConnData.f_BridgeStatus = BridgeStatus.TryingToConnect;
                    f_ConnData.f_ConnectOnce = false;
                    f_ConnData.f_ConnectionIsDirty = false;

                    f_ConnData.f_ConnCookie = new SamConnectionCookie(f_Inst.f_NetworkMngr, f_Prefs);
                    f_Inst.f_NetworkMngr.EnqueueAPC(f_ConnData.f_ConnCookie);

                    f_ConnData.f_ConnectionLocked = true;
                    {
                        f_ConnData.f_ConnCookie.GuiWaitForCompletion();
                    }
                    f_ConnData.f_ConnectionLocked = false;

                    // This check is important. We may be dead.
                    if (false == p_IsRunning)
                    {
                        return false;
                    }

                    if (f_ConnData.f_ConnCookie.p_State == CookieStates.Failure)
                    {
                        Util.DebugBreak();

                        f_ConnData.f_ConnCookie = null;
                        f_ConnData.f_BridgeStatus = BridgeStatus.Disconnected;
                    }
                }
            }
            else
            {
                BridgeStatus l_NewBridgeStatus = f_ConnData.f_ConnCookie.p_BridgeStatusVolatile;

                if (l_NewBridgeStatus != f_ConnData.f_BridgeStatus)
                {
                    if (BridgeStatus.Disconnected == l_NewBridgeStatus)
                    {
                        // The connection collapsed without our consent.

                        bool l_WasConnected = (BridgeStatus.Connected == f_ConnData.f_BridgeStatus);

                        f_ConnData.NotifyDisconnect(a_Time, l_WasConnected);
                        AdjustConnMenuItemsReceptiveness(false);

                        if (true == l_WasConnected)
                        {
                            ShowBalloonTip(f_TextSet[TextKeys.TitleSAMConnBroken], f_TextSet[TextKeys.MsgSAMConnBroken], ToolTipIcon.Error);
                        }
                    }
                    else if (BridgeStatus.Connected == l_NewBridgeStatus)
                    {
                        // Connection established. We need to propagate
                        // the key pairs.

                        I2PKeyPair l_DatagramKey = f_ConnData.f_ConnCookie.p_EffDatagramKeyPair;
                        I2PKeyPair l_StreamKey = f_ConnData.f_ConnCookie.p_EffStreamKeyPair;

                        f_Prefs.f_GeneralPrefs.SetKeyPairs(l_DatagramKey, l_StreamKey);

                        if (null != f_ActiveOptionsForm)
                        {
                            // The options dialog is shown.

                            f_ActiveOptionsForm.DeliverKeysFromNewConnection(l_DatagramKey, l_StreamKey);
                        }

                        f_ConnData.NotifyConnect();
                        AdjustConnMenuItemsReceptiveness(true);
                    }
                    else
                    {
                        // Should never happen.

                        throw (new ApplicationException());
                    }
                }
            }

            TextKeys l_TextKey;
            Keys l_ShortcutKeys;

            if (GeneralPrefs.ConnStartKinds.AutoStart == f_Prefs.f_GeneralPrefs.f_ConnStartKind)
            {
                f_ConnMenuFunc = ConnectMenuFunction.AutoReconnect;
                l_TextKey = (BridgeStatus.Disconnected == f_ConnData.f_BridgeStatus ? TextKeys.MainMenuConnect : TextKeys.MainMenuReconnect);
                l_ShortcutKeys = Keys.None;
            }
            else
            {
                // Manual connection start.

                switch (f_ConnData.f_BridgeStatus)
                {
                    case BridgeStatus.Disconnected:
                        f_ConnMenuFunc = ConnectMenuFunction.ManualConnect;
                        l_TextKey = TextKeys.MainMenuConnect;
                        l_ShortcutKeys = (Keys.Control | Keys.F11);
                        break;

                    case BridgeStatus.TryingToConnect:
                        f_ConnMenuFunc = ConnectMenuFunction.ManualDisconnect;
                        l_TextKey = TextKeys.MainMenuStopConnecting;
                        l_ShortcutKeys = (Keys.Control | Keys.F12);
                        break;

                    case BridgeStatus.Connected:
                        f_ConnMenuFunc = ConnectMenuFunction.ManualDisconnect;
                        l_TextKey = TextKeys.MainMenuDisconnect;
                        l_ShortcutKeys = (Keys.Control | Keys.F12);
                        break;

                    default: throw (new ApplicationException());
                }
            }

            f_ConnMenuItem.Text = f_TextSet[l_TextKey];
            f_ConnMenuItem.ShortcutKeys = l_ShortcutKeys;

            return true;
        }

        private void MenuConnect_Click(object sender, EventArgs e)
        {
            if (true == f_ConnData.f_ConnectionLocked)
            {
                return;
            }

            Time l_Time = f_Inst.GetTime();

            switch (f_ConnMenuFunc)
            {
                case ConnectMenuFunction.ManualDisconnect:
                    {
                        bool l_Junk;
                        if (false == Disconnect(l_Time, true, out l_Junk))
                        {
                            return;
                        }
                    }
                    break;

                case ConnectMenuFunction.ManualConnect:
                    if (BridgeStatus.Disconnected == f_ConnData.f_BridgeStatus)
                    {
                        if (true == f_ConnData.p_DisableNetworkCapabilities)
                        {
                            ErrMsgBox(Software.BuildObsoleteClientMessage(false));

                            return;
                        }

                        f_ConnData.f_ConnectOnce = true;
                    }
                    break;

                case ConnectMenuFunction.AutoReconnect:
                    {
                        bool l_StillConnected;

                        if (false == Disconnect(l_Time, true, out l_StillConnected))
                        {
                            return;
                        }

                        if (false == l_StillConnected)
                        {
                            f_ConnData.f_ConnectOnce = true;
                        }
                    }
                    break;

                default: throw (new ApplicationException());
            }

            ManageSamConnection(l_Time);
        }

        private bool Disconnect(Time a_Time,
                                bool a_PromptIfConnected,
                                out bool a_StillConnected)
        {
            // Only the Connect_Click() routine may call this routine.
            // Do not continue processing if false is returned.

            if (BridgeStatus.Disconnected == f_ConnData.f_BridgeStatus)
            {
                a_StillConnected = false;

                return true;
            }

            if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus &&
                true == a_PromptIfConnected &&
                false == YesNoMsgBox(f_TextSet[TextKeys.QuestionReallyCloseConn]))
            {
                a_StillConnected = true;

                return true;
            }

            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiKillConnection);
            f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);

            f_ConnData.f_ConnectionLocked = true;
            {
                l_Cookie.GuiWaitForCompletion();
            }
            f_ConnData.f_ConnectionLocked = false;

            // This check is important. We may be dead.
            if (false == p_IsRunning)
            {
                a_StillConnected = false;

                return false;
            }

            if (CookieStates.Failure == l_Cookie.p_State)
            {
                Util.DebugBreak();

                a_StillConnected = true;

                return true;
            }

            f_ConnData.NotifyDisconnect(a_Time, false);
            f_ShowConnSecInfoMenuItem.Enabled = false;

            a_StillConnected = false;

            return true;
        }

        private void MenuShowOptions_Click(object sender, EventArgs e)
        {
            ShowOptionsDialog(ShowOptionsDlgReasons.MenuClick);
        }

        public void ShowOptionsDialog(ShowOptionsDlgReasons a_Reason)
        {
            if (null != f_ActiveOptionsForm)
            {
                return;
            }

            using (f_ActiveOptionsForm = new OptionsDialog(f_Inst, f_ScrollStates))
            {
                f_ActiveOptionsForm.ShowDialog(this, a_Reason);
                f_ActiveOptionsForm.CopyScrollStates(f_ScrollStates);
            }

            f_ActiveOptionsForm = null;
        }

        private void MenuSamTest_Click(object sender, EventArgs e)
        {
            if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus)
            {
                InfoMsgBox(f_TextSet[TextKeys.MsgSAMBridgeConnSuccess]);

                return;
            }

            SamBridgeTestDialog.ShowModal(f_Inst, this);
        }

        private void MenuViewI2PRouterStatus_Click(object sender, EventArgs e)
        {
            using (I2PRouterStatusDialog l_Dlg = new I2PRouterStatusDialog(f_TextSet))
            {
                l_Dlg.ShowDialog();
            }
        }

        private void MenuShowConnSecInfo_Click(object sender, EventArgs e)
        {
            if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus)
            {
                InfoMsgBox(f_ConnData.f_ConnCookie.BuildTunnelInfoStr(f_TextSet));
            }
        }

        private void MenuCopyOwnStreamAddressToClipboard_Click(object sender, EventArgs e)
        {
            if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus)
            {
                I2PAddress l_Addr = f_ConnData.f_ConnCookie.p_EffStreamKeyPair.f_PublicKey;
                Util.SetClipboard(l_Addr.ToFullString());
            }
        }

        private void MenuRepublish_Click(object sender, EventArgs e)
        {
            if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus)
            {
                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiForcePublishing, null);
                f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
            }
        }

        private void AdjustConnMenuItemsReceptiveness(bool a_Set)
        {
            f_ShowConnSecInfoMenuItem.Enabled = a_Set;
            f_CopyOwnStreamAddressMenuItem.Enabled = a_Set;
            f_RepublishMenuItem.Enabled = a_Set;
        }

#region Nodes

        private byte[] ReadNodesFile(string a_FileName,
                                     out string a_ErrMsg)
        {
            byte[] l_Bytes = ReadNodesFileCore(a_FileName, out a_ErrMsg);

            if (null == l_Bytes)
            {
                a_ErrMsg = f_TextSet.BuildFileErrMsg(TextKeys.MsgNodesErrLoadingFromFile, a_FileName, a_ErrMsg);

                return null;
            }

            return l_Bytes;
        }

        private byte[] ReadNodesFileCore(string a_FileName,
                                         out string a_SubErrMsg)
        {
            // Return can be null, in which case a_ErrMsg will be set.

            Exception l_Exc;
            FHandle l_File = FHandle.OpenR(a_FileName, false, FileOptions.None, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

            if (null == l_File)
            {
                a_SubErrMsg = l_Exc.Message;

                return null;
            }

            ulong l_Size;

            if (false == l_File.GetSize(out l_Size))
            {
                a_SubErrMsg = f_TextSet[TextKeys.MsgFileErrAccessFailed];

                return null;
            }

            if ((1024 * 1024) < l_Size)
            {
                l_File.Close();

                a_SubErrMsg = f_TextSet[TextKeys.MsgFileErrFileTooLarge];

                return null;
            }

            ByteArray l_Data = new ByteArray((int)l_Size);
            bool l_Read = l_File.Read(0, ref l_Data);

            l_File.Close();

            if (false == l_Read)
            {
                a_SubErrMsg = f_TextSet[TextKeys.MsgFileErrReadFailed];

                return null;
            }

            a_SubErrMsg = null;

            return l_Data.f_Arr;
        }

        private string BuildNodeFilesFilterStr()
        {
            return BuildFilterStr(f_TextSet[TextKeys.FileDlgNodeFiles], "dat");
        }

        public string BuildCollectionFilesFilterStr()
        {
            return BuildFilterStr(f_TextSet.FormatWithAppName(TextKeys.FileDlgNBCollections), "nbc");
        }

        private string BuildFilterStr(string a_FileCaption,
                                      string a_FileExtension)
        {
            return (a_FileCaption + "|*." + a_FileExtension + "|" + f_TextSet[TextKeys.FileDlgAllFiles] + "|*.*");
        }

        private void MenuAddKadAddress_Click(object sender, EventArgs e)
        {
            bool l_OK;
            I2PAddress l_Addr;

            using (I2PAddressInputDialog l_Dlg = new I2PAddressInputDialog(f_Inst, false))
            {
                l_OK = l_Dlg.ShowDialog(this, out l_Addr);
            }

            if (true == l_OK)
            {
                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_Kad.GuiAddAddressForBootstrapping, new object[] { l_Addr });
                f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
            }
        }

        private void MenuImportNodes_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog l_Dlg = new OpenFileDialog())
            {
                l_Dlg.Filter = BuildNodeFilesFilterStr();

                if (DialogResult.OK != l_Dlg.ShowDialog(this))
                {
                    return;
                }

                string l_FileName = l_Dlg.FileName;
                string l_ErrMsg;
                byte[] l_Arr = ReadNodesFile(l_FileName, out l_ErrMsg);

                if (null == l_Arr)
                {
                    ErrMsgBox(l_ErrMsg);

                    return;
                }

                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_Kad.GuiImportNodes, l_Arr);
                if (false == f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
                {
                    return;
                }

                if (CookieStates.Success == l_Cookie.p_State)
                {
                    InfoMsgBox(f_TextSet.Format(TextKeys.MsgNodesImportWorked, l_FileName));
                }
                else
                {
                    ErrMsgBox(f_TextSet.Format(TextKeys.MsgNodesErrImportFailed, l_FileName));
                }
            }
        }

        private void MenuExportNodes_Click(object sender, EventArgs e)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_Kad.GuiExportNodes);
            if (false == f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
            {
                return;
            }

            if (CookieStates.Success != l_Cookie.p_State)
            {
                InfoMsgBox(f_TextSet[TextKeys.MsgNodesErrNothingToExport]);

                return;
            }

            using (SaveFileDialog l_Dlg = new SaveFileDialog())
            {
                l_Dlg.Filter = BuildNodeFilesFilterStr();
                l_Dlg.FileName = Software.c_NodesDat;

                if (DialogResult.OK != l_Dlg.ShowDialog(this))
                {
                    return;
                }

                string l_FileName = l_Dlg.FileName;
                byte[] l_Arr = (byte[])l_Cookie.GetReturnObject();

                try
                {
                    File.WriteAllBytes(l_FileName, l_Arr);
                }
                catch (Exception l_Exc)
                {
                    ErrMsgBox(f_TextSet.BuildFileErrMsg(TextKeys.MsgNodesErrExportFailed, l_FileName, l_Exc.Message));

                    return;
                }

                InfoMsgBox(f_TextSet.Format(TextKeys.MsgNodesExportWorked, l_FileName));
            }
        }

#endregion

        public bool ShowPickiMuleFolderDialog(bool a_AllowTempDirAlso,
                                              out string a_Path)
        {
            // Return can be null.

            using (FolderBrowserDialog l_Dlg = new FolderBrowserDialog())
            {
                string l_Msg = f_TextSet[TextKeys.PickiMuleFolderDialogMsg];

                if (true == a_AllowTempDirAlso)
                {
                    l_Msg += (' ' + f_TextSet[TextKeys.PickiMuleFolderDialogTempDirExtraMsg]);
                }

                l_Dlg.Description = l_Msg;
                l_Dlg.ShowNewFolderButton = false;

                if (DialogResult.OK == l_Dlg.ShowDialog(this))
                {
                    a_Path = l_Dlg.SelectedPath;

                    return true;
                }
            }

            a_Path = null;

            return false;
        }

        public bool ShowDontRuniMuleNowMsg()
        {
            // Returns false if the user picked cancel or if the iMule
            // process was detected.

            return (DialogResult.OK == MsgBox(f_TextSet[TextKeys.KeepiMuleClosedDuringOpMsg], MessageBoxButtons.OKCancel, MessageBoxIcon.Information) &&
                    false == CheckIfiMuleIsRunningWithErrMsg());
        }

        public bool CheckIfiMuleIsRunningWithErrMsg()
        {
            // Returns true if the iMule process was detected (in this
            // case a message box was shown).

            foreach (Process l_Proc in Process.GetProcesses())
            {
                if (true == Util.StringsEqualIgnoreCase("imule", l_Proc.ProcessName))
                {
                    ErrMsgBox(f_TextSet[TextKeys.iMuleProcessDetected]);

                    return true;
                }
            }

            return false;
        }

        public void ResetTaskBarState()
        {
            if (null != f_TaskBar)
            {
                f_TaskBar.SetNoProgress();
            }
        }

        public void SetTaskBarMarqueeEffect()
        {
            if (null != f_TaskBar)
            {
                f_TaskBar.SetMarqueeEffect();
            }
        }

        public void SetTaskBarStateFilled(TaskBarProgressKinds a_Kind)
        {
            if (null != f_TaskBar)
            {
                f_TaskBar.SetFilled(a_Kind);
            }
        }

        public void SetTaskBarStateAndValues(ref TaskBarProgressContext a_Cntxt)
        {
            if (null != f_TaskBar)
            {
                f_TaskBar.SetProgress(ref a_Cntxt);
            }
        }

#region MessageBoxes

        public void ShowOperationFailedMsg()
        {
            ErrMsgBox(f_TextSet[TextKeys.MsgOperationFailed]);
        }

        public void ShowDeletedFilesMsgBoxIfNecessary(int a_Done,
                                                      int a_Count,
                                                      bool a_Aborted)
        {
            if (false == a_Aborted &&
                a_Count > a_Done)
            {
                int l_Failed = (a_Count - a_Done);
                string l_ErrMsg = f_TextSet.Format(TextKeys.FileDeletionFailedMsgWithCounters, l_Failed, a_Count);
                this.WarnMsgBox(l_ErrMsg);
            }
        }

        public void WarnMsgBox(string a_WarnMsg)
        {
            MsgBox(a_WarnMsg, MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        public void ErrMsgBox(string a_ErrMsg)
        {
            MsgBox(a_ErrMsg, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        public void InfoMsgBox(string a_InfoMsg)
        {
            MsgBox(a_InfoMsg, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        public bool YesNoMsgBox(string a_Question)
        {
            return (DialogResult.Yes == MsgBox(a_Question, MessageBoxButtons.YesNo, MessageBoxIcon.Question));
        }

        public DialogResult AbortableYesNoMsgBox(IDialogController a_Controller,
                                                 string a_Question)
        {
            return AbortableMsgBox(a_Controller, a_Question, MessageBoxButtons.YesNo, MessageBoxIcon.Question, DialogResult.Yes);
        }

        public bool SeriousYesNoChoice(string a_Question)
        {
            return (DialogResult.Yes == MsgBox(a_Question, MessageBoxButtons.YesNo, MessageBoxIcon.Warning, DialogResult.No));
        }

        private DialogResult MsgBox(string a_Msg,
                                    MessageBoxButtons a_Buttons,
                                    MessageBoxIcon a_Icon)
        {
            return MsgBox(a_Msg, a_Buttons, a_Icon, DialogResult.None);
        }

        private DialogResult MsgBox(string a_Msg,
                                    MessageBoxButtons a_Buttons,
                                    MessageBoxIcon a_Icon,
                                    DialogResult a_SelButton)
        {
            FormCollection l_Coll = Application.OpenForms;
            int l_Count = l_Coll.Count;
            IWin32Window l_Owner = null;

            if (0 < l_Count)
            {
                for (int i = (l_Count - 1); 0 <= i; i--)
                {
                    Form l_CurForm = l_Coll[i];
                    Type l_FormType = l_CurForm.GetType();

                    if (typeof(CustomMsgBox) == l_FormType ||
                        true == l_FormType.IsSubclassOf(typeof(CustomMsgBox)))
                    {
                        continue;
                    }

                    l_Owner = l_CurForm;

                    break;
                }
            }

            if (null == l_Owner)
            {
                l_Owner = this;
            }

            return CustomMsgBox.ShowMsgBox(f_TextSet, l_Owner, f_TextSet.f_AppName, a_Msg, a_Icon, a_Buttons, a_SelButton);
        }

        private DialogResult AbortableMsgBox(IDialogController a_Controller,
                                             string a_Msg,
                                             MessageBoxButtons a_Buttons,
                                             MessageBoxIcon a_Icon,
                                             DialogResult a_SelButton)
        {
            return CustomMsgBox.ShowAbortableMsgBox(f_TextSet, this, a_Controller, f_TextSet.f_AppName, a_Msg, a_Icon, a_Buttons, a_SelButton);
        }

        private bool ShowBalloonTip(string a_Title,
                                    string a_Msg,
                                    ToolTipIcon a_Icon)
        {
            return ShowBalloonTip(a_Title, a_Msg, a_Icon, null, null);
        }

        private bool ShowBalloonTip(string a_Title,
                                    string a_Msg,
                                    ToolTipIcon a_Icon,
                                    BalloonClickHandler a_BalloonClickHandler,
                                    object a_BalloonClickContext)
        {
            // Shows the tip only if the tray icon is shown and if the
            // user allows balloon tips. a_BalloonClickHandler and the
            // context may be null.

            if (true == f_Prefs.f_GeneralPrefs.f_ShowBalloonNotifications)
            {
                f_BalloonClickHandler = a_BalloonClickHandler;
                f_BalloonClickContext = a_BalloonClickContext;

                try
                {
                    f_NotifyIcon.ShowBalloonTip(2500, a_Title, a_Msg, a_Icon);

                    return true;
                }
                catch
                {
                }
            }

            return false;
        }

        private void NotifyIcon_BalloonTipClicked(object sender, EventArgs e)
        {
            // NOTE: We cannot be sure that the balloon that was
            // clicked matches the handler we use.

            if (null != f_BalloonClickHandler)
            {
                f_BalloonClickHandler(f_BalloonClickContext);
            }
        }

#endregion

#region Links

        public void ExternalHandleShellInput(string a_Stmt)
        {
            if (true == ShouldAcceptExternalCall())
            {
                HandleShellInput(a_Stmt, InputSource.ExternalCommandLine);
            }
        }

        private void HandleShellInput(string a_Stmt,
                                      InputSource a_Source)
        {
            if (InputSource.ExternalCommandLine == a_Source)
            {
                FocusMainForm();

                if (1 < Application.OpenForms.Count)
                {
                    // We are not ready to receive input from outside
                    // now, so just drop the new request.

                    return;
                }
            }

            bool l_IsValidPath;
            bool l_IsPath = (true == LongPath.IsPathRooted(a_Stmt, out l_IsValidPath) &&
                             true == l_IsValidPath);

            if (false == l_IsPath)
            {
                ImportLink(a_Stmt, a_Source);
            }
            else
            {
                OpenCollectionFile(a_Stmt);
            }
        }

        private void OpenCollectionFile(string a_FileName)
        {
            using (OpenCollectionDialog l_Dlg = new OpenCollectionDialog(f_Inst, a_FileName))
            {
                l_Dlg.LoadAndShowDialog(this);
            }
        }

        private void ImportLink(string a_Stmt,
                                InputSource a_Source)
        {
            // The caller should check p_IsRunning after return, unless
            // control is directly returned to the message loop.

            bool l_Magnet = FileLink.IsPotentialMagnetLink(a_Stmt);
            TextKeys l_What;
            FileLink l_Link;
            string l_Err;

            if (true == l_Magnet ||
                true == FileLink.IsPotentialeD2kUri(a_Stmt, true))
            {
                FileLink.ParserResult l_Res;

                if (false == l_Magnet)
                {
                    l_What = TextKeys.LinkseD2kLink;
                    l_Res = FileLink.FromeD2kUri(a_Stmt, out l_Link);
                }
                else
                {
                    l_What = TextKeys.LinksMagnetLink;
                    l_Res = FileLink.FromMagnetLink(a_Stmt, out l_Link);
                }

                TextKeys l_ErrMsgKey;

                switch (l_Res)
                {
                    case FileLink.ParserResult.Success:
                        l_ErrMsgKey = TextKeys.Empty;
                        break;

                    case FileLink.ParserResult.NoMagnetLink:
                        l_ErrMsgKey = TextKeys.MsgLinkNotMagnet;
                        break;

                    case FileLink.ParserResult.NoeD2kLink:
                        l_ErrMsgKey = TextKeys.MsgLinkNoteD2k;
                        break;

                    case FileLink.ParserResult.SeveralFilesNotAllowed:
                        l_ErrMsgKey = TextKeys.MsgLinkSeveralFilesNotAllowed;
                        break;

                    case FileLink.ParserResult.OnlyFileAllowed:
                        l_ErrMsgKey = TextKeys.MsgLinkOnlyFileAllowed;
                        break;

                    default:
                        l_ErrMsgKey = TextKeys.MsgLinkNotAccepted;
                        break;
                }

                l_Err = (TextKeys.Empty == l_ErrMsgKey ? null : f_TextSet[l_ErrMsgKey]);
            }
            else
            {
                l_What = (InputSource.Clipboard == a_Source ? TextKeys.LinksClipboardContent : TextKeys.LinksShellInput);
                l_Link = null;
                l_Err = f_TextSet[TextKeys.LinksMsgInputNotRecognized];
            }

            if (null == l_Link)
            {
                string l_Msg = f_TextSet.Format(TextKeys.LinksMsgFailedToInterpretFormat, f_TextSet[l_What], l_Err);
                ErrMsgBox(l_Msg);

                return;
            }

            StringBuilder l_Bld;

            if (false == ContentManager.IsValidFileSize(l_Link.f_FileSize))
            {
                l_Bld = new StringBuilder(f_TextSet[TextKeys.LinksMsgFileTooLarge]);
                l_Link.AddAsString(f_Inst, l_Bld, true);

                ErrMsgBox(l_Bld.ToString());

                return;
            }

            FileID l_FileID = l_Link.f_MasterHashes.CreateFileID();
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiCanAddFileWithID, new object[2] { l_FileID.f_Val, l_Link.f_FileSize });
            if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
            {
                return;
            }

            if (CookieStates.Success != l_Cookie.p_StateVolatile)
            {
                // Should never happen.

                return;
            }

            if (false == (bool)l_Cookie.GetReturnObject())
            {
                l_Bld = new StringBuilder(f_TextSet[TextKeys.LinksMsgFileIDConflict]);
                l_Link.AddAsString(f_Inst, l_Bld, true);

                ErrMsgBox(l_Bld.ToString());

                return;
            }

            l_Bld = new StringBuilder(f_TextSet[TextKeys.LinksQuestionWantToAddFile]);
            l_Link.AddAsString(f_Inst, l_Bld, true);

            if (false == YesNoMsgBox(l_Bld.ToString()))
            {
                return;
            }

            Category l_TargetCategory = (InputSource.Clipboard == a_Source ? f_FilesPanel.GetSelectedCategory() : Category.s_Unassigned);
            l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiAddFileFromLink, new object[] { l_Link, l_TargetCategory });
            if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
            {
                return;
            }

            if (CookieStates.Success != l_Cookie.p_StateVolatile)
            {
                ErrMsgBox(f_TextSet[TextKeys.LinksMsgFailedToAdd]);
            }
        }

        private void MenuPasteLink_Click(object sender, EventArgs e)
        {
            string l_Stmt;

            try
            {
                l_Stmt = Clipboard.GetText(TextDataFormat.Text);
            }
            catch
            {
                return;
            }

            ImportLink(l_Stmt, InputSource.Clipboard);
        }

        private void MenuOpenCollection_Click(object sender, EventArgs e)
        {
            string l_FileName;

            using (OpenFileDialog l_Dlg = new OpenFileDialog())
            {
                l_Dlg.Filter = BuildCollectionFilesFilterStr();

                if (DialogResult.OK != l_Dlg.ShowDialog(this))
                {
                    return;
                }

                l_FileName = l_Dlg.FileName;
            }

            OpenCollectionFile(l_FileName);
        }

        private void MenuImportIncomplFilesFromiMule_Click(object sender, EventArgs e)
        {
            string l_Path;

            if (false == ShowPickiMuleFolderDialog(true, out l_Path) ||
                false == ShowDontRuniMuleNowMsg())
            {
                return;
            }

            using (ImportPartialDownloadsDialog l_Dlg = new ImportPartialDownloadsDialog(f_Inst, l_Path))
            {
                l_Dlg.LoadAndShowDialog(this);
            }
        }

        private void MenuGC_Click(object sender, EventArgs e)
        {
            GC.Collect();
        }

        private void MenuResetBans_Click(object sender, EventArgs e)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiResetBans);
            f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
        }

        private void MenuRefreshSharedFiles_Click(object sender, EventArgs e)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiReloadSharedFiles);
            f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
        }

        private void MenuRescanForFinishedFiles_Click(object sender, EventArgs e)
        {
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiRescanForFinishedFiles);
            f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
        }

        private void MenuOpenOutputDirectory_Click(object sender, EventArgs e)
        {
            if (null == f_Prefs)
            {
                return;
            }

            string l_Path = f_Prefs.f_IOPrefs.f_SharedDirs.GetCompletionTargetPath();

            if (null == l_Path)
            {
                ErrMsgBox("There was no output directory specified yet.");
            }
            else
            {
                bool l_Opened = NavigateToFolder(l_Path);
            }
        }

#endregion

        public void AddExtractionCookie(ExtractionCookie a_Cookie)
        {
            f_ExtractionCookies.Add(a_Cookie);
        }

        private void HandlePreviewing()
        {
            for (int i = (f_ExtractionCookies.Count - 1); 0 <= i; i--)
            {
                ExtractionCookie l_Cookie = f_ExtractionCookies[i];
                CookieStates l_State = l_Cookie.p_StateVolatile;

                if (CookieStates.Pending == l_State)
                {
                    continue;
                }

                f_ExtractionCookies.RemoveAt(i);

                string l_ErrMsg;

                if (CookieStates.Success == l_State)
                {
                    bool l_Executed = Util.ShellOpenOrExecuteFile(l_Cookie.f_TempFile.f_FullFileName);

                    if (true == l_Executed)
                    {
                        l_Cookie.f_TempFile.CancelDeletionAndClose();

                        return;
                    }

                    l_ErrMsg = f_TextSet[TextKeys.MsgPreviewProcessDidntStart];
                }
                else
                {
                    // This should happen rarely.

                    l_ErrMsg = f_TextSet[TextKeys.MsgPreviewGeneralErr];
                }

                // This will delete the file.
                l_Cookie.f_TempFile.Close();

                ErrMsgBox(l_ErrMsg);
            }
        }

        public void SelectUserContextOfPeerContext(PeerContext a_PeerCntxt)
        {
            // This routine is used by the files panel.

            if (false == f_Prefs.f_GeneralPrefs.f_UIModeAdvanced)
            {
                return;
            }

            UserContext l_UserCntxt = a_PeerCntxt.f_Clnt.p_UserCntxt;

            if (null != l_UserCntxt)
            {
                f_PeersPanel.SetAutoSelectUserContext(l_UserCntxt);
                SelectMainTab(MainTabs.GeneralPeers);
            }
        }

        private bool SelectMainTab(MainTabs a_MainTab)
        {
            if (true == f_TabCntrl.GetPageByIndex((uint)a_MainTab).p_PageVisible)
            {
                f_TabCntrl.SetSelectedPageIndex((uint)a_MainTab, SelectionChangeOrigins.UserInteraction);

                return true;
            }

            return false;
        }

        public bool ShouldTabUpdateControls(bool a_IsTabSelected,
                                            out bool a_UpdateTaskBar)
        {
            a_UpdateTaskBar = (true == a_IsTabSelected &&
                               null != this.f_TaskBar);

            if (false == IsMinimized())
            {
                return a_IsTabSelected;
            }

            // Window is minimized.

            if (true == this.f_Prefs.f_GeneralPrefs.f_MinimizeToTray)
            {
                a_UpdateTaskBar = false;
            }

            return false;
        }

        private void ProcessGui(object sender, EventArgs e)
        {
            if (true == f_TryingToClose)
            {
                f_TryingToClose = false;
                this.Close();

                if (ApplicationState.Started != f_AppState)
                {
                    return;
                }
            }

            Time l_Time = f_Inst.GetTime();

            // Update connection status.

            if (false == ManageSamConnection(l_Time))
            {
                return;
            }

            // Update speed display (only each second turn, in order
            // to keep it from changing so often).

            if (false == f_TitleUpdateFlag)
            {
                UpdateFormTitle(true);
            }

            f_TitleUpdateFlag = !f_TitleUpdateFlag;

            if (true == f_AutoSaveTime.IsEarlierThan(l_Time))
            {
                HandleAutoSaving();

                f_AutoSaveTime = l_Time.AddSeconds(c_AutoSaveIntervalSecs);
            }

            if (true == IsMinimized())
            {
                // Don't update while minimized.

                return;
            }

            // Update contacts display.

            {
                uint l_Gen = Util.VolatileRead(ref f_Input.f_ContactInfosGeneration);

                if (l_Gen != f_LastContactInfosGeneration)
                {
                    GuiContactInfo[] l_Arr = Util.VolatileRead(ref f_Input.f_ContactInfos);
                    f_KadPanel.Update(l_Arr);

                    f_LastContactInfosGeneration = l_Gen;
                }
            }

            // Update searches.

            f_SearchPanel.Process(l_Time);

            // Update files display.

            {
                FormInput.Files l_Files = Util.VolatileRead(ref f_Input.f_Files);

                if (null != l_Files &&
                    l_Files.f_Generation != f_LastFilesGeneration)
                {
                    f_FilesPanel.UpdateFiles(l_Files);

                    f_LastFilesGeneration = l_Files.f_Generation;
                }
            }

            // Update peers.

            {
                FormInput.GeneralPeers l_Peers = Util.VolatileRead(ref f_Input.f_GeneralPeers);

                if (null != l_Peers &&
                    l_Peers.f_Generation != f_LastPeersGeneration)
                {
                    f_PeersPanel.UpdatePeers(l_Peers);

                    f_LastPeersGeneration = l_Peers.f_Generation;
                }
            }

            // Update stats.

            {
                uint l_Gen = Util.VolatileRead(ref f_Input.f_StatsGeneration);

                if (l_Gen != f_LastStatsGeneration)
                {
                    string l_Stats = Util.VolatileRead(ref f_Input.f_Stats);

                    if (true == f_CopyStatsToClipboardFlag)
                    {
                        f_CopyStatsToClipboardFlag = false;
                        UpdateSubscribedNetDataTypes();

                        Util.SetClipboard(l_Stats);
                    }

                    f_StatsPanel.SetText(l_Stats);

                    f_LastStatsGeneration = l_Gen;
                }
            }

            // Finally update the log.

            UpdateLog();

            // Subscribe.

            f_Input.SetFutureNetworkUpdateTime(l_Time, 750);

            // Handle previews.

            HandlePreviewing();

            if (false == f_NewerVersionInfoIsDisplayed &&
                true == Software.s_LaterVersionAvailable)
            {
                f_NewerVersionInfoIsDisplayed = true;
                string l_Msg = f_TextSet.FormatWithAppName(TextKeys.MsgNewerVersionDetected);
                f_Menu.SetMessage(l_Msg);
                AddToLog(l_Msg);
            }
        }

        private void UpdateFormTitle(bool a_MayShowSpeed)
        {
            string l_TitleText = f_TextSet.f_AppNameWithVer;
            bool l_TrayIconUpdated = false;

            if (true == a_MayShowSpeed)
            {
                bool l_ShowOnTitle = f_Prefs.f_GeneralPrefs.f_ShowSpeedOnTitle;
                bool l_ShowTrayIcon = f_Prefs.f_GeneralPrefs.f_ShowTrayIcon;

                if (true == l_ShowOnTitle ||
                    true == l_ShowTrayIcon)
                {
                    string l_SpeedStr;

                    if (BridgeStatus.Connected == f_ConnData.f_BridgeStatus)
                    {
                        Speed l_DlSpeed;
                        Speed l_UlSpeed;
                        Speed.CopyVolatile(ref f_Input.f_GlobalDlSpeed, out l_DlSpeed);
                        Speed.CopyVolatile(ref f_Input.f_GlobalUlSpeed, out l_UlSpeed);

                        bool l_SI = p_DisplaySISizes;
                        l_SpeedStr = f_TextSet.Format(TextKeys.AppTitleSpeedFormat, l_DlSpeed.ToKiloString(l_SI), l_UlSpeed.ToKiloString(l_SI));
                    }
                    else
                    {
                        l_SpeedStr = f_TextSet[BridgeStatus.Disconnected == f_ConnData.f_BridgeStatus ? TextKeys.AppTitleNotConn : TextKeys.AppTitleConnecting];
                    }

                    if (true == l_ShowOnTitle)
                    {
                        l_TitleText = f_TextSet.Format(TextKeys.AppTitleWithSpeedFormat, l_SpeedStr, l_TitleText);
                    }

                    if (true == l_ShowTrayIcon)
                    {
                        l_TrayIconUpdated = true;
                        f_NotifyIcon.Text = (f_TextSet.f_AppName + " - " + l_SpeedStr);
                    }
                }
            }

            this.Text = l_TitleText;

            if (false == l_TrayIconUpdated)
            {
                f_NotifyIcon.Text = string.Empty;
            }
        }

        private void HandleAutoSaving()
        {
            switch (f_AutoSaveMode)
            {
                case 0:
                    {
                        GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiSaveFiles);
                        f_Inst.f_ContentMngr.EnqueueAPC(l_Cookie);
                        f_AutoSaveMode++;
                    }
                    break;

                case 1:
                    {
                        GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiSaveNetworkData);
                        f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);
                        f_AutoSaveMode++;
                    }
                    break;

                case 2:
                    {
                        StorageStates l_State = SavePreferences();
                        f_AutoSaveMode = 0;
                    }
                    break;

                default: throw (new ApplicationException());
            }
        }

        private StorageStates SaveSearchCache()
        {
            // Should only be done when the application shuts down,
            // since it can be quite expensive.

            ElmKey l_ElmKey = new ElmKey(Instance.c_SearchCacheElmKey);

            if (false == f_Prefs.f_GeneralPrefs.f_SaveSearchHistory)
            {
                return f_Inst.DeleteElement(l_ElmKey);
            }

            return f_Inst.WriteToElement(l_ElmKey, f_SearchPanel.SaveSearchCache);
        }

        private StorageStates SavePreferences()
        {
            return f_Inst.WriteToElement(new ElmKey(Instance.c_PrefsElmKey), SavePreferencesCore);
        }

        private StorageStates SavePreferencesCore(BufferIO a_Buffer)
        {
            // Should not be called directly, use SavePreferences()
            // instead.

            a_Buffer.WriteUInt32LE(c_PrefsMagic);
            // We save the window state from v1.4.0 on.
            a_Buffer.WriteByte((byte)this.WindowState);
            // We save the accepted license editions as a 16-bit wide
            // bit field from v1.4.8 on.
            a_Buffer.WriteUInt16LE((ushort)f_AcceptedLicenseEditions.AsUInt());
            // We save the version number of the highest version
            // encountered from v1.4.14 on.
            a_Buffer.WriteUInt32LE(Software.s_NewerVersion);
            // We save the last selected main tab from v1.4.21 on.
            a_Buffer.WriteByte((byte)f_TabCntrl.p_SelectedPageIndex);

            a_Buffer.SetPosition(c_PrefsHeaderSize);
            f_Prefs.ToBuffer(a_Buffer, f_PrefForwardData);

            // Save GUI appearance and related settings.

            f_Categories.ToBuffer(a_Buffer);

            // Flush visual appearance.
            f_KadPanel.PrepareSaving();
            f_SearchPanel.PrepareSaving();
            f_FilesPanel.PrepareSaving();
            f_PeersPanel.PrepareSaving();

            f_VisualCust.ToBuffer(a_Buffer);

            return StorageStates.Success;
        }

        private StorageStates LoadPreferences(BufferIO a_Buffer)
        {
            // If this routine doesn't return a success code, the state
            // of f_LoadedPrefs, f_PrefExtras and f_Categories is
            // undefined.

            if (c_PrefsHeaderSize > a_Buffer.GetRemaining() ||
                c_PrefsMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            f_PrevSessionWndState = (FormWindowState)Math.Min(a_Buffer.ReadByte(), (byte)FormWindowState.Maximized);
            f_AcceptedLicenseEditions = new BitField(a_Buffer.ReadUInt16LE());
            uint l_NewestKnownVer = a_Buffer.ReadUInt32LE();
            f_LastSelectedMainTabOfPreviousSession = (MainTabs)a_Buffer.ReadByte();

            if (Software.s_ReleaseInfo.EncodeVersion() < l_NewestKnownVer)
            {
                Software.s_LaterVersionAvailable = true;
                Software.s_NewerVersion = l_NewestKnownVer;
            }

            if (true == Software.s_SecondLicenseAccepted)
            {
                f_AcceptedLicenseEditions.SetFlag(2);
            }

            a_Buffer.SetPosition(c_PrefsHeaderSize);

            try
            {
                f_LoadedPrefs = Preferences.FromBuffer(a_Buffer, out f_PrefForwardData);

                if (null == f_LoadedPrefs)
                {
                    return StorageStates.Corruption;
                }

                f_Categories = CategoryList.FromBuffer(a_Buffer);

                if (null == f_Categories ||
                    false == f_VisualCust.FromBuffer(a_Buffer))
                {
                    return StorageStates.Corruption;
                }
            }
            catch (BufferIO.InterpretBufferException)
            {
                return StorageStates.Corruption;
            }

            if (false == a_Buffer.EndReached())
            {
                return StorageStates.Corruption;
            }

            return StorageStates.Success;
        }

        public bool NavigateToObject(string a_Path)
        {
            return WinShell.OpenFolderAndSelectItem(a_Path);
        }

        public bool NavigateToFolder(string a_Path)
        {
            bool l_Opened = Util.ShellOpenDirectory(a_Path);

            if (false == l_Opened)
            {
                string l_Msg;

                if (0 == a_Path.Trim().Length)
                {
                    l_Msg = f_TextSet[TextKeys.DirectoryFailedSinceEmptyString];
                }
                else
                {
                    l_Msg = f_TextSet.Format(TextKeys.DirectoryFailedToOpen, a_Path);
                }

                WarnMsgBox(l_Msg);
            }

            return l_Opened;
        }

        private void TabCntrl_SelectedPageChanged(AdvancedTabControl a_Sender)
        {
            UpdateSubscribedNetDataTypes();
        }

        private void UpdateSubscribedNetDataTypes()
        {
            FormInput.NetworkDataTypes l_Type;

            switch ((MainTabs)f_TabCntrl.p_SelectedPageIndex)
            {
                case MainTabs.KademliaContacts:
                    l_Type = FormInput.NetworkDataTypes.Contacts;
                    break;

                case MainTabs.GeneralPeers:
                    l_Type = FormInput.NetworkDataTypes.GeneralPeers;
                    break;

                case MainTabs.Stats:
                    l_Type = FormInput.NetworkDataTypes.Stats;
                    break;

                default:
                    l_Type = FormInput.NetworkDataTypes.None;
                    break;
            }

            if (true == f_CopyStatsToClipboardFlag)
            {
                l_Type |= FormInput.NetworkDataTypes.Stats;
            }

            f_Input.f_SubscribedNetDataTypes = l_Type;
            f_Input.SetNetworkUpdateTime(f_Inst.GetTime());
        }

#region Log

        public void AddToLog(TextKeys a_Key)
        {
            AddToLog(f_TextSet[a_Key]);
        }

        public void AddToLog(string a_Str)
        {
            // Thread safe.

            DateTime l_DateTime = f_Inst.GetTime().ToDateTime();
            string f_CompleteStr = f_TextSet.Format(TextKeys.LogEntryFormat, f_TextSet.TimeStr(l_DateTime), a_Str);
            int l_StrLen = a_Str.Length;
            bool l_EndWithLineBreak = (0 < l_StrLen &&
                                       '\n' == a_Str[l_StrLen - 1]);

            Util.MonitorEnter(f_LogStrBld);
            {
                f_LogStrBld.Append(f_CompleteStr);

                // If the input string already has an ending line
                // break, we won't add another.
                if (false == l_EndWithLineBreak)
                {
                    f_LogStrBld.AppendLine();
                }
            }
            Util.MonitorLeave(f_LogStrBld);
        }

        public void UpdateLog()
        {
            string l_LogEntries;

            Util.MonitorEnter(f_LogStrBld);
            {
                if (0 == f_LogStrBld.Length)
                {
                    l_LogEntries = null;
                }
                else
                {
                    l_LogEntries = f_LogStrBld.ToString();
                    f_LogStrBld.Length = 0;
                }
            }
            Util.MonitorLeave(f_LogStrBld);

            if (null == l_LogEntries)
            {
                return;
            }

            string l_Text = f_LogTextBox.Text;

            if (c_MaxLogChars >= (l_Text.Length + l_LogEntries.Length))
            {
                f_LogTextBox.AppendText(l_LogEntries);
                f_LogTextBox.ScrollToEnd();

                return;
            }

            // Get rid of one fourth of the log. Try to find a line
            // break for that.

            int l_Limit = (l_Text.Length / 4);
            int l_Idx = l_Text.IndexOf('\n', l_Limit);
            l_Idx = (0 > l_Idx ? l_Limit : (l_Idx + 1));

            f_LogTextBox.Text = (l_Text.Substring(l_Idx) + l_LogEntries);
            f_LogTextBox.ScrollToEnd();
        }

        private void MenuCopyStatsToClipboard_Click(object sender, EventArgs e)
        {
            // Query the stats from the backend and then copy.
            f_CopyStatsToClipboardFlag = true;
            UpdateSubscribedNetDataTypes();
        }

        private void MenuCopyLogToClipboard_Click(object sender, EventArgs e)
        {
            Util.SetClipboard(f_LogTextBox.GetTextWithPlatfromLineBreak());
        }

        private void MenuViewLicenseDialog_Click(object sender, EventArgs e)
        {
            using (LicenseDialog l_LicenseForm = new LicenseDialog(f_TextSet, LicenseDialog.DialogMode.ViewLicense))
            {
                l_LicenseForm.ShowDialog(this);
            }
        }

        private void MenuViewReleaseNotes_Click(object sender, EventArgs e)
        {
            string l_ExpectedFileName = LongPath.Combine(Software.sp_ModulePath, Software.s_ReleaseInfo.BuildReleaseNotesFileName());

            if (false == File.Exists(l_ExpectedFileName))
            {
                ShowOperationFailedMsg();

                return;
            }

            Util.ShellOpenOrExecuteFile(l_ExpectedFileName);
        }

        private void MenuShowAboutDialog_Click(object sender, EventArgs e)
        {
            using (AboutDialog l_Dlg = new AboutDialog(f_Inst))
            {
                l_Dlg.ShowDialog(this);
            }
        }

#endregion

        public void NotifyDownloadOfDuplicate(string a_OriginalFileName,
                                              string a_ExistingFullFileName)
        {
            // a_ExistingFullFileName is the full path of the existing
            // file.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            string l_Msg = f_TextSet.Format(TextKeys.MsgDlFileExistsAlready, a_OriginalFileName, a_ExistingFullFileName);
            NotifyDownload(l_Msg);
        }

        public void NotifyDownloadCompletion(string a_OriginalFileName,
                                             string a_CreatedFullFileName)
        {
            // a_CreatedFullFileName is the full path of the created
            // file. a_OriginalFileName can be different than the
            // filename portion of a_CreatedFullFileName.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            string l_SuccMsg = f_TextSet.Format(TextKeys.MsgFileDlSucceeded, a_OriginalFileName, a_CreatedFullFileName);
            NotifyDownload(l_SuccMsg);

            ShowBalloonTip(f_TextSet[TextKeys.TitleDlFinished], l_SuccMsg, ToolTipIcon.Info, OpenFolderOnBalloonClick, a_CreatedFullFileName);
        }

        private void NotifyDownload(string a_LogMsg)
        {
            AddToLog(a_LogMsg);

            if (true == f_Prefs.f_GeneralPrefs.f_BeepWhenDlIsDone)
            {
                Util.Beep();
            }
        }

        private const uint c_AlertUserAboutContinuedTroubleIntervalSecs = (10 * Time.c_SecsPerMinute);
        private Time f_EarliestMissingOutputDirAlertTime;
        private Time f_EarliestFailedToExtractDueToInsuffDiskSpaceAlertTime;
        private Time f_EarliestAvailDiskSpaceStateChangedAlertTime;
        private Time f_EarliestConnectedButWithoutKademilaContactsAlertTime;

        public void NotifyMissingOutputDirectory()
        {
            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            if (false == CanAlertAgain(ref f_EarliestMissingOutputDirAlertTime))
            {
                return;
            }

            string l_Msg = f_TextSet[TextKeys.MsgCantFinishDlDueToMissingOutDir];

            AddToLog(l_Msg);
            ShowBalloonTip(f_TextSet[TextKeys.TitleMissingOutDir], l_Msg, ToolTipIcon.Warning, OpenOptionsToDefineOutputDirOnBalloonClick, null);
        }

        public void NotifyFailedToExtractDueToInsufficientDiskSpace()
        {
            // This routine gets called whenever extraction is
            // attempted and fails. It is independet of the
            // insufficient disk space condition that affects
            // the application data files (they can be on another
            // drive).
            // Do not use a balloon in this case, since it might hide
            // the more important balloon we need to indicate download
            // suspension.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            AddToLogAndShowBalloonWithTimeCheck(ref f_EarliestFailedToExtractDueToInsuffDiskSpaceAlertTime,
                                                string.Empty,
                                                f_TextSet[TextKeys.MsgCantFinishDlDueToLackOfSpace],
                                                false);
        }

        public void NotifyResumedDownloadsAndHashingDueToAvailDiskSpace()
        {
            // This routine gets called once when the resuming
            // from a state of insufficient disk space. Log
            // the message in any case.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            Time l_Time = new Time();
            AddToLogAndShowBalloonWithTimeCheck(ref l_Time,
                                   string.Empty,
                                   f_TextSet[TextKeys.MsgDlHashingResumed],
                                   false);

            f_EarliestAvailDiskSpaceStateChangedAlertTime = new Time();
        }

        public void NotifySuspendedDownloadsAndHashingDueToInsuffDiskSpace()
        {
            // This routine gets called periodically during the
            // state of insufficient disk space.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            AddToLogAndShowBalloonWithTimeCheck(ref f_EarliestAvailDiskSpaceStateChangedAlertTime,
                                   f_TextSet[TextKeys.TitleDlHashingSuspended],
                                   f_TextSet[TextKeys.MsgDlHashingSuspendedDueToLackOfSpace],
                                   true);
        }

        public void NotifyConnectedButWithoutKademilaContacts()
        {
            // This routine gets called periodically while a
            // connection to the I2P router is up AND there is
            // not a single known node.

            if (false == ShouldAcceptExternalCall())
            {
                return;
            }

            AddToLogAndShowBalloonWithTimeCheck(ref f_EarliestConnectedButWithoutKademilaContactsAlertTime,
                                   f_TextSet[TextKeys.TitleMissingKadNodes],
                                   f_TextSet.Format(TextKeys.MsgMissingKadNodes, Software.c_NodesDat),
                                   true);
        }

        private void AddToLogAndShowBalloonWithTimeCheck(ref Time a_NextPossibleAlertTime,
                                                         string a_Caption,
                                                         string a_Message,
                                                         bool a_UseBalloon)
        {
            if (false == CanAlertAgain(ref a_NextPossibleAlertTime))
            {
                return;
            }

            AddToLogAndShowBalloon(a_Caption, a_Message, a_UseBalloon);
        }

        private void AddToLogAndShowBalloon(string a_Caption,
                                            string a_Message,
                                            bool a_UseBalloon)
        {
            AddToLog(a_Message);

            if (true == a_UseBalloon)
            {
                ShowBalloonTip(a_Caption, a_Message, ToolTipIcon.Warning);
            }
        }

        private bool CanAlertAgain(ref Time a_NextPossibleAlertTime)
        {
            Time l_Time = f_Inst.GetTime();

            if (l_Time < a_NextPossibleAlertTime)
            {
                return false;
            }

            a_NextPossibleAlertTime = l_Time.AddSeconds(c_AlertUserAboutContinuedTroubleIntervalSecs);

            return true;
        }

        private void OpenOptionsToDefineOutputDirOnBalloonClick(object a_Context)
        {
            ShowOptionsDialog(ShowOptionsDlgReasons.SetOutputDir);
        }

        private static void OpenFolderOnBalloonClick(object a_Context)
        {
            string l_Path = (string)a_Context;
            bool l_Success = (true == WinShell.OpenFolderAndSelectItem(l_Path) ||
                              true == Util.ShellOpenDirectory(l_Path));
        }

        private static readonly Bitmap s_TinyBitmap;
        private static readonly Graphics s_MeasurementGraphics;

        public static void MeasureText(string a_Text,
                                       Font a_Font,
                                       out Size a_Size)
        {
            if (0 >= a_Text.Length)
            {
                a_Size = new Size();

                return;
            }

            SizeF l_SizeF = s_MeasurementGraphics.MeasureString(a_Text, a_Font);
            a_Size = new Size((int)l_SizeF.Width + 1, (int)l_SizeF.Height + 1);
        }

        public static int MeasureTextWidth(string a_Text,
                                           Font a_Font)
        {
            Size l_Size;
            MeasureText(a_Text, a_Font, out l_Size);

            return l_Size.Width;
        }

        public static bool CreateFont(string a_FamilyName,
                                      float a_PointSize,
                                      FontStyle a_Style,
                                      out Font a_NewFont)
        {
            // Returns zero if the requested font family is not
            // installed.

            a_NewFont = new Font(a_FamilyName, a_PointSize, a_Style);

            if (a_NewFont.FontFamily.Name == a_FamilyName)
            {
                return true;
            }

            a_NewFont.Dispose();

            return false;
        }

        public static Font RecreateFont(Font a_Font,
                                        float a_MinPointSize,
                                        FontStyle a_Style)
        {
            return new Font(a_Font.FontFamily, Math.Max(a_Font.SizeInPoints, a_MinPointSize), a_Style);
        }

        public DataGridViewCellStyle CreateCellStyle(DataGridViewCellStyle a_Template,
                                                     Font a_Font,
                                                     Color a_ForeColor)
        {
            DataGridViewCellStyle l_CellStyle = a_Template.Clone();
            l_CellStyle.Font = a_Font;
            l_CellStyle.ForeColor = a_ForeColor;

            return l_CellStyle;
        }

        public static void InsertMenuSeparator(ContextMenuStrip a_Menu,
                                               int a_Idx)
        {
            a_Menu.Items.Insert(a_Idx, new ToolStripSeparator());
        }

        public static void AddMenuSeparator(ContextMenuStrip a_Menu)
        {
            a_Menu.Items.Add(new ToolStripSeparator());
        }

        public static void AddMenuSeparator(ToolStripMenuItem a_Parent)
        {
            a_Parent.DropDownItems.Add(new ToolStripSeparator());
        }

        public static ToolStripMenuItem AddMenuItem(ContextMenuStrip a_Menu,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler)
        {
            return AddMenuItem(a_Menu, a_Text, a_ClickHandler, Keys.None, IconTypes.NoIcon);
        }

        public static ToolStripMenuItem AddMenuItem(ToolStripMenuItem a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler)
        {
            return AddMenuItem(a_Parent, a_Text, a_ClickHandler, Keys.None, IconTypes.NoIcon);
        }

        public static ToolStripMenuItem AddMenuItem(ContextMenuStrip a_Menu,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    IconTypes a_IconType)
        {
            return AddMenuItem(a_Menu, a_Text, a_ClickHandler, Keys.None, a_IconType);
        }

        public static ToolStripMenuItem AddMenuItem(ToolStripMenuItem a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    IconTypes a_IconType)
        {
            return AddMenuItem(a_Parent, a_Text, a_ClickHandler, Keys.None, a_IconType);
        }

        public static ToolStripMenuItem AddMenuItem(ContextMenuStrip a_Menu,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    Keys a_ShortcutKeys,
                                                    IconTypes a_IconType)
        {
            ToolStripMenuItem l_Item = CreateMenuItem(a_Text, a_ClickHandler, a_ShortcutKeys, a_IconType);

            a_Menu.Items.Add(l_Item);

            return l_Item;
        }

        public static ToolStripMenuItem AddMenuItem(ToolStripMenuItem a_Parent,
                                                    string a_Text,
                                                    EventHandler a_ClickHandler,
                                                    Keys a_ShortcutKeys,
                                                    IconTypes a_IconType)
        {
            ToolStripMenuItem l_Item = CreateMenuItem(a_Text, a_ClickHandler, a_ShortcutKeys, a_IconType);

            a_Parent.DropDownItems.Add(l_Item);

            return l_Item;
        }

        private static ToolStripMenuItem CreateMenuItem(string a_Text,
                                                        EventHandler a_ClickHandler,
                                                        Keys a_ShortcutKeys,
                                                        IconTypes a_IconType)
        {
            ToolStripMenuItem l_New = new ToolStripMenuItem(a_Text);
            l_New.Click += a_ClickHandler;
            l_New.ShortcutKeys = a_ShortcutKeys;
            l_New.Image = Software.s_IconStorage.GetDefaultIcon(a_IconType);

            return l_New;
        }

        public static void DisposeMenuItems(ToolStrip a_Menu)
        {
            // Dispose from end to start. Each disposed item will
            // automatically leave the collection.

            for (int i = (a_Menu.Items.Count - 1); 0 <= i; i--)
            {
                // This will also dispose any subitems of an item that
                // are populating the DropDownItems property.
                a_Menu.Items[i].Dispose();
            }
        }

        public static bool HasPositiveArea(ref Rectangle a_Rect)
        {
            return (0 < a_Rect.Width &&
                    0 < a_Rect.Height);
        }

        public static void ZeroImpossibleSize(ref Rectangle a_Rect)
        {
            if (0 > a_Rect.Width ||
                0 > a_Rect.Height)
            {
                a_Rect.Width = 0;
                a_Rect.Height = 0;
            }
        }

        public class FormInput
        {
            // These fields get written to by the GUI thread
            // periodically.

            public long f_NextNetworkUpdateTime;
            public NetworkDataTypes f_SubscribedNetDataTypes;

            public long f_NextFileUpdateTime;
            public Category f_SubscribedFileCategory;

            // These fields get written to by the network thread (if
            // the GUI demands it).

            public Speed f_GlobalDlSpeed;
            public Speed f_GlobalUlSpeed;

            public uint f_ContactInfosGeneration;
            public GuiContactInfo[] f_ContactInfos;

            public GeneralPeers f_GeneralPeers;

            public uint f_StatsGeneration;
            public string f_Stats;

            // These fields get written to by the content manager
            // (if the GUI demands it).

            public Files f_Files;

            public void SetNetworkUpdateTime(Time a_Time)
            {
                // Used by the GUI.

                SetFutureNetworkUpdateTime(a_Time, 0);
            }

            public void SetFutureNetworkUpdateTime(Time a_Time,
                                                   uint a_WaitMillisecs)
            {
                // Used by the GUI.

                if (NetworkDataTypes.None == f_SubscribedNetDataTypes)
                {
                    return;
                }

                Time l_Soon = a_Time.AddMilliseconds(a_WaitMillisecs);

                if (a_Time.ToBinary() < (ulong)f_NextNetworkUpdateTime &&
                    l_Soon.ToBinary() >= (ulong)f_NextNetworkUpdateTime)
                {
                    // Do not update the time if this would make us
                    // wait longer.

                    return;
                }

                System.Threading.Interlocked.Exchange(ref f_NextNetworkUpdateTime, (long)l_Soon.ToBinary());
            }

            [Flags]
            public enum NetworkDataTypes : byte
            {
                None = 0,

                Contacts = 0x1,
                GeneralPeers = 0x2,
                Stats = 0x4,
            }

            public class GeneralPeers
            {
                public uint f_Generation;
                public uint f_ConnectedPeers;
                // f_UserCntxts can be null.
                public UserContext[] f_UserCntxts;

                public GeneralPeers(uint a_Generation,
                                    uint a_ConnectedPeers,
                                    UserContext[] a_UserCntxts)
                {
                    f_Generation = a_Generation;
                    f_ConnectedPeers = a_ConnectedPeers;
                    f_UserCntxts = a_UserCntxts;
                }
            }

            public class Files
            {
                public readonly uint f_Generation;
                public readonly Category f_Category;
                // f_FileArr can be null.
                public readonly KnownFile[] f_FileArr;

                public Files(uint a_Generation,
                             Category a_Category,
                             KnownFile[] a_FileArr)
                {
                    f_Generation = a_Generation;
                    f_Category = a_Category;
                    f_FileArr = a_FileArr;
                }
            }
        }

        public enum MainTabs : byte
        {
            KademliaContacts = 0,
            FileSearch = 1,
            Files = 2,
            GeneralPeers = 3,
            Stats = 4,
            Log = 5,
        }

        private enum ConnectMenuFunction : byte
        {
            ManualDisconnect = 0,
            ManualConnect = 1,
            AutoReconnect = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ConnData
        {
            // f_ConnectionLocked tells us if the GUI thread should
            // not modify the connection for the moment.
            public bool f_ConnectionLocked;
            // f_ConnCookie will be null while we're not connected.
            public SamConnectionCookie f_ConnCookie;
            // f_BridgeStatus indicates the bridge status as the GUI
            // thread knows it.
            public BridgeStatus f_BridgeStatus;
            public bool f_ConnRepairMeaningful;
            // f_ConnectOnce is used for manual connection starts.
            public bool f_ConnectOnce;
            // f_ConnectionDirty indicates that a new connection is
            // necessary to enforce the most recent settings.
            public bool f_ConnectionIsDirty;
            private Time f_NextAutoConnTime;

            private bool f_DisableNetworkCapabilities;
            private Time f_NextCheckForOutdatedReleaseTime;

            public bool p_DisableNetworkCapabilities
            {
                get { return f_DisableNetworkCapabilities; }
            }

            public bool CanAutoConnect(Time a_Time)
            {
                return f_NextAutoConnTime.IsEarlierThan(a_Time);
            }

            public void NotifyConnect()
            {
                f_BridgeStatus = BridgeStatus.Connected;
                f_ConnRepairMeaningful = false;
            }

            public void NotifyDisconnect(Time a_Time,
                                         bool a_ConnRepairMeaningful)
            {
                f_ConnCookie = null;
                f_BridgeStatus = BridgeStatus.Disconnected;

                if (false == a_ConnRepairMeaningful)
                {
                    // Defer the next auto connect attempt.

                    f_NextAutoConnTime = a_Time.AddSeconds(c_AutoConnectIntervalSecs);
                }
                else
                {
                    f_ConnRepairMeaningful = true;
                }
            }

            public void CheckForOutdatedRelease(Time a_Time)
            {
                if (a_Time >= f_NextCheckForOutdatedReleaseTime)
                {
                    f_NextCheckForOutdatedReleaseTime = a_Time.AddSeconds(Time.c_SecsPerQuarterHour);

                    if (false == f_DisableNetworkCapabilities &&
                        true == Software.s_ReleaseInfo.IsOneYearOld())
                    {
                        f_DisableNetworkCapabilities = true;
                    }
                }
            }
        }

        public enum ShowOptionsDlgReasons : byte
        {
            MenuClick = 0,
            EditCommentFilter = 1,
            SetOutputDir = 2,
        }

        private enum ApplicationState : byte
        {
            PreStart = 0,
            StartingUp = 1,
            Started = 2,
            ShutDownInProgress = 3,
            ShutDownFinished = 4,
        }

        private enum InputSource : byte
        {
            OwnCommandLine = 0,
            ExternalCommandLine = 1,
            Clipboard = 2,
        }

        private class UpdatePreferencesContext : ControllerContext
        {
            public readonly bool f_NeedBarBgImgs;

            private BarBackgroundImages f_BarBgImgs;

            public UpdatePreferencesContext(bool a_NeedBarBgImgs) :
                                       base(false)
            {
                f_NeedBarBgImgs = a_NeedBarBgImgs;
            }

            public void SetFinished(BarBackgroundImages a_BarBgImgs)
            {
                f_BarBgImgs = a_BarBgImgs;

                base.SetFinishedBase();
            }

            public BarBackgroundImages GetBarBackgroundImages()
            {
                // Return can be null.

                return f_BarBgImgs;
            }
        }

        private class LoadAndStartContext : ControllerContext
        {
            public LoadAndStartContext() :
                                  base(false)
            {
            }

            public void SetFinished()
            {
                base.SetFinishedBase();
            }
        }

        public class GeneralShutdownContext : AbortableControllerContext, HardMemory.ICompactProgressCollector
        {
            private const string c_ShutdownThreadName = "ShutdownWorker";
            private const byte c_FullProgress = 100;
            private const byte c_HalfProgress = (c_FullProgress / 2);

            private readonly Instance f_Inst;
            private readonly bool f_FindLeaks;
            private readonly bool f_Compact;

            private bool f_PhaseTwo;

            public GeneralShutdownContext(Instance a_Inst,
                                          bool a_FindLeaks,
                                          bool a_Compact)
            {
                f_Inst = a_Inst;
                f_FindLeaks = a_FindLeaks;
                f_Compact = a_Compact;
            }

            public void NotifyProgress(uint a_Done,
                                       uint a_Total)
            {
                // We never expect a_Total to be zero.

                uint l_Perc = ((a_Done * c_HalfProgress) / a_Total);

                if (true == f_PhaseTwo)
                {
                    l_Perc += c_HalfProgress;
                }

                base.SetPercentage(l_Perc);
            }

            public void BeginAsyncShutdown()
            {
                Util.StartThread(c_ShutdownThreadName, ShutdownWorker);
            }

            private void ShutdownWorker()
            {
                // Shut down the network manager.

                Software.UpdateShutDownState(0x1);

                NetworkManager.ShutdownCookie l_Cookie = new NetworkManager.ShutdownCookie(f_Inst.f_NetworkMngr);
                f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);
                l_Cookie.Wait();

                Software.UpdateShutDownState(0x40);

                // Shut down the content manager.

                f_Inst.f_ContentMngr.Shutdown(f_FindLeaks, f_Compact, this);

                Software.UpdateShutDownState(0x100000);

                f_PhaseTwo = true;
                base.SetPercentage(c_HalfProgress);
                f_Inst.f_Mem.Close(f_Compact, this);

                Software.UpdateShutDownState(0x200000);

                base.SetPercentage(c_FullProgress);
                base.SetFinishedBase();

                Software.UpdateShutDownState(0x400000);
            }
        }
    }
}

namespace NB
{
#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct APCQueue<T> where T : APCCookie
    {
        // Don't use the struct default constructor.

        private readonly Queue<T> f_Cookies;
        private bool f_ShutDown;

        private APCQueue(Queue<T> a_Cookies)
        {
            f_Cookies = a_Cookies;
            f_ShutDown = false;
        }

        public static void Create(out APCQueue<T> a_APCQueue)
        {
            a_APCQueue = new APCQueue<T>(new Queue<T>(1));
        }

        public void Enqueue(T a_Cookie)
        {
            // This routine may be called by any thread.

            Enter();
            {
                if (false == f_ShutDown)
                {
                    f_Cookies.Enqueue(a_Cookie);

                    Monitor.Pulse(f_Cookies);
                }
                else
                {
                    a_Cookie.SetState(CookieStates.Failure);
                }
            }
            Leave();
        }

        public void ShutDown()
        {
            Enter();
            {
                foreach (APCCookie l_Cookie in f_Cookies)
                {
                    l_Cookie.SetState(CookieStates.Failure);
                }

                f_Cookies.Clear();
                f_ShutDown = true;
            }
            Leave();
        }

        private T[] CopyCookies()
        {
            // The caller must hold the lock.

            if (0 < f_Cookies.Count)
            {
                T[] l_Arr = new T[f_Cookies.Count];

                f_Cookies.CopyTo(l_Arr, 0);
                f_Cookies.Clear();

                return l_Arr;
            }

            return null;
        }

        public T[] GetNow()
        {
            // Return can be null.

            T[] l_Arr = null;

            if (0 < f_Cookies.Count)
            {
                Enter();
                {
                    l_Arr = CopyCookies();
                }
                Leave();
            }

            return l_Arr;
        }

        public T[] WaitForInput()
        {
            return WaitForInput(Util.c_MiniSleepTimeout);
        }

        public T[] WaitForInput(int a_MillisecsTimeout)
        {
            // Return can be null.

            T[] l_Arr;

            Enter();
            {
                if (0 == f_Cookies.Count)
                {
                    Monitor.Wait(f_Cookies, a_MillisecsTimeout);
                }

                l_Arr = CopyCookies();
            }
            Leave();

            return l_Arr;
        }

        public T[] WaitForInput(ref bool a_WatchedFlag)
        {
            // Return can be null. This routine will block until data
            // is available or the watched flag becomes false.

            T[] l_Arr = null;

            Enter();
            {
                while (true == Util.VolatileRead(ref a_WatchedFlag))
                {
                    if (0 == f_Cookies.Count)
                    {
                        Monitor.Wait(f_Cookies, Util.c_MiniSleepTimeout);
                    }

                    l_Arr = CopyCookies();

                    if (null != l_Arr)
                    {
                        break;
                    }
                }
            }
            Leave();

            return l_Arr;
        }

        private void Enter()
        {
            Util.MonitorEnter(f_Cookies);
        }

        private void Leave()
        {
            Util.MonitorLeave(f_Cookies);
        }

        public static void ExecuteAndClear(ref GeneralAPCCookie[] a_Cookies)
        {
            // a_Cookies will be set to null after execution.

            foreach (GeneralAPCCookie l_Cookie in a_Cookies)
            {
                l_Cookie.f_Delegate(l_Cookie);
            }

            a_Cookies = null;
        }
    }

    class APCCookie
    {
        // APC stands for Asynchronous Procedure Call.

        private CookieStates f_State;

        public CookieStates p_State
        {
            get { return f_State; }
        }

        public CookieStates p_StateVolatile
        {
            get { return Util.VolatileRead(ref f_State); }
        }

        public void SetState(CookieStates a_State)
        {
            Util.VolatileWrite(out f_State, a_State);
        }

        public void WaitForCompletion()
        {
            while (CookieStates.Pending == Util.VolatileRead(ref f_State))
            {
                Util.MiniSleep();
            }
        }

        public bool GuiWaitForCompletion()
        {
            return GuiWaitForCompletion(ref f_State);
        }

        public static bool GuiWaitForCompletion(ref CookieStates a_State)
        {
            // Returns false if the GUI has closed in the meantime.
            // The return value does not implicate the state of the
            // cookie.

            while (CookieStates.Pending == Util.VolatileRead(ref a_State))
            {
                Application.DoEvents();
                Util.MiniSleep();
            }

            // It can happen that the GUI thread has waited for an apc
            // to complete and finds the GUI closed afterwards. In
            // this case the shortest path back to the message loop
            // should be taken in order to allow graceful termination.
            return Software.sp_MainForm.p_IsRunning;
        }
    }

    class GeneralAPCCookie : APCCookie
    {
        public delegate void APCHandler(GeneralAPCCookie a_CallCntxt);

        public readonly APCHandler f_Delegate;
        private readonly object f_ArgObj;
        private object f_RetObj;

        public GeneralAPCCookie(APCHandler a_Delegate,
                                object a_ArgObj)
        {
            f_Delegate = a_Delegate;
            f_ArgObj = a_ArgObj;
        }

        public GeneralAPCCookie(APCHandler a_Delegate) :
                           this(a_Delegate, null)
        {
        }

        public object GetArg()
        {
            return f_ArgObj;
        }

        public object[] GetArgAsObjArray()
        {
            return (object[])f_ArgObj;
        }

        public KnownFile GetArgAsKnownFile()
        {
            return (KnownFile)f_ArgObj;
        }

        public KnownFile[] GetArgAsKnownFileArray()
        {
            return (KnownFile[])f_ArgObj;
        }

        public UInt128 GetArgAsUInt128()
        {
            return (UInt128)f_ArgObj;
        }

        public byte[] GetArgAsByteArray()
        {
            return (byte[])f_ArgObj;
        }

        public void SetReturnObject(object a_RetObj)
        {
            f_RetObj = a_RetObj;
        }

        public void SetReturnObjectAndState(object a_RetObj,
                                            CookieStates a_State)
        {
            f_RetObj = a_RetObj;
            SetState(a_State);
        }

        public object GetReturnObject()
        {
            return f_RetObj;
        }
    }

    enum CookieStates : byte
    {
        Pending = 0,
        Success = 1,
        Failure = 2,
        RangeSet = 3,
        WaitingForHashes = 4,
        SearchRunning = 5,
    }

    enum StorageStates : byte
    {
        Success = 0,
        Corruption = 1,
        NoData = 2,
        TooAdvanced = 3,
        UpdateFailed = 4,
        ReadFailed = 5,
        TempFileIOError = 6,
        QueryError = 7,
    }

    class SamConnectionCookie : GeneralAPCCookie
    {
        // From GUI.
        public readonly IPEndPoint f_SamEndPoint;
        public readonly TunnelSettings f_InTunnelSettings;
        public readonly TunnelSettings f_OutTunnelSettings;
        public readonly I2PKeyPair f_ProposedDatagramKeyPair;
        public readonly I2PKeyPair f_ProposedStreamKeyPair;

        // From network manager.
        private BridgeStatus f_BridgeStatus;
        private I2PKeyPair f_EffDatagramKeyPair;
        private I2PKeyPair f_EffStreamKeyPair;

        public BridgeStatus p_BridgeStatus
        {
            get { return f_BridgeStatus; }
        }

        public BridgeStatus p_BridgeStatusVolatile
        {
            get { return Util.VolatileRead(ref f_BridgeStatus); }
        }

        public I2PKeyPair p_EffDatagramKeyPair
        {
            get { return f_EffDatagramKeyPair; }
        }

        public I2PKeyPair p_EffStreamKeyPair
        {
            get { return f_EffStreamKeyPair; }
        }

        public SamConnectionCookie(NetworkManager a_NetworkMngr,
                                   Preferences a_Prefs) :
                              base(a_NetworkMngr.GuiBuildConnection)
        {
            Time l_Time = Time.Now();
            I2PKeyPair l_DatagramKeyPair = LoadKeyPair(a_Prefs.f_GeneralPrefs.f_AutoManageDatagramKeys, GeneralPrefs.c_AutoDropDatagramKeyPairDays, a_Prefs.f_GeneralPrefs.p_DatagramKeyPair, l_Time);
            I2PKeyPair l_StreamKeyPair = LoadKeyPair(a_Prefs.f_GeneralPrefs.f_AutoManageStreamKeys, GeneralPrefs.c_AutoDropStreamKeyPairDays, a_Prefs.f_GeneralPrefs.p_StreamKeyPair, l_Time);

            f_SamEndPoint = a_Prefs.f_GeneralPrefs.f_SamEndPoint;
            f_InTunnelSettings = a_Prefs.f_GeneralPrefs.f_InboundTunnelSettings;
            f_OutTunnelSettings = a_Prefs.f_GeneralPrefs.GetEffectiveOutbountTunnelSettings();
            f_ProposedDatagramKeyPair = l_DatagramKeyPair;
            f_ProposedStreamKeyPair = l_StreamKeyPair;
            f_BridgeStatus = BridgeStatus.TryingToConnect;
        }

        public void SetEffectiveKeyPairs(I2PKeyPair a_DatagramKeyPair,
                                         I2PKeyPair a_StreamKeyPair)
        {
            f_EffDatagramKeyPair = a_DatagramKeyPair;
            f_EffStreamKeyPair = a_StreamKeyPair;
        }

        public void SetBridgeStatus(BridgeStatus a_BridgeStatus)
        {
            Util.VolatileWrite(out f_BridgeStatus, a_BridgeStatus);
        }

        public string BuildTunnelInfoStr(TextSet a_TextSet)
        {
            // Used by the GUI.

            StringBuilder l_Bld = new StringBuilder(512);

            l_Bld.AppendLine(a_TextSet[TextKeys.OptionsDlgTunnelsInboundSettings]);
            f_InTunnelSettings.ToString(a_TextSet, l_Bld);
            l_Bld.AppendLine();
            l_Bld.AppendLine(a_TextSet[TextKeys.OptionsDlgTunnelsOutboundSettings]);
            f_OutTunnelSettings.ToString(a_TextSet, l_Bld);
            l_Bld.AppendLine();
            l_Bld.AppendLine(a_TextSet[TextKeys.OptionsDlgTunnelsMsgDoNotReveal]);

            return l_Bld.ToString();
        }

        private static I2PKeyPair LoadKeyPair(bool a_AutoManage,
                                              int a_MaxAgeInDays,
                                              I2PKeyPair a_ExistingKeyPair,
                                              Time a_RefTime)
        {
            // a_ExistingKeyPair may be null. Return can be null
            // (hinting that a new key pair should be created).

            if (true == a_AutoManage &&
                null != a_ExistingKeyPair &&
                a_MaxAgeInDays <= a_ExistingKeyPair.CalcAgeInDays(a_RefTime))
            {
                // This key got too old and needs to be replaced for
                // security reasons.

                return null;
            }

            return a_ExistingKeyPair;
        }
    }

    class CategoryList : IDisposable
    {
        // c_OldCatListMagic was used until v1.4.0.
        private const uint c_OldCatListMagic = 0x4362D4E7;
        private const uint c_ExtendedCatListMagic = 0x26E440F8;
        private const byte c_MyFormatRevision = 1;

        public const byte c_MaxCustomCategories = 64;
        // Limit is in UTF-16 code units and was lifted from 32 to 64
        // in v1.4.0.
        public const byte c_MaxCategoryNameChars = 64;

        private uint f_GenericsVisibilty;
        private FlexArray<CategoryInfo> f_Customs;

        public int p_CustomCount
        {
            get { return f_Customs.p_Count; }
        }

        private CategoryList(uint a_GenericsVisibilty,
                             int a_Capacity)
        {
            f_GenericsVisibilty = a_GenericsVisibilty;
            FlexArray<CategoryInfo>.Create(out f_Customs);
        }

        public static CategoryList CreateDefault()
        {
            const uint l_DefaultVis = 0x6B;

            return new CategoryList(l_DefaultVis, 0);
        }

        public void Dispose()
        {
            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = (f_Customs.p_Count - 1); 0 <= i; i--)
            {
                l_Arr[i].Dispose();
            }

            f_Customs.Clear();
        }

        public static CategoryList FromBuffer(BufferIO a_Buffer)
        {
            uint l_Magic = a_Buffer.ReadUInt32LE();
            uint l_Revision;

            if (c_ExtendedCatListMagic == l_Magic)
            {
                l_Revision = a_Buffer.ReadByte();

                if (0 == l_Revision ||
                    c_MyFormatRevision < l_Revision)
                {
                    return null;
                }
            }
            else if (c_OldCatListMagic == l_Magic)
            {
                // We expect no revision byte.

                l_Revision = 0;
            }
            else
            {
                return null;
            }

            return FromBufferCore(a_Buffer, l_Revision);
        }

        private static CategoryList FromBufferCore(BufferIO a_Buffer,
                                                   uint a_Revision)
        {
            // The first category will always be visible.
            uint l_GenericsVisibilty = (a_Buffer.ReadUInt32LE() | 1);
            int l_Count = a_Buffer.ReadUInt16LE();

            if (c_MaxCustomCategories < l_Count)
            {
                return null;
            }

            CategoryList l_Categories = new CategoryList(l_GenericsVisibilty, l_Count);

            for (int i = 0; i < l_Count; i++)
            {
                Category l_Category = new Category(a_Buffer.ReadUInt16LE());
                string l_Name = a_Buffer.ReadUTF16LEStringEncLen();

                if (true == l_Category.IsGeneric() ||
                    false == IsNameLegal(l_Name))
                {
                    return null;
                }

                CategoryInfo[] l_Arr = l_Categories.f_Customs.p_Arr;
                for (int k = 0; k < i; k++)
                {
                    if (l_Category == l_Arr[k].f_Category ||
                        true == IsSameName(l_Name, l_Arr[k].f_Name))
                    {
                        return null;
                    }
                }

                bool l_Visible = a_Buffer.ReadBool();
                IconTypes l_IconType;
                Color l_IconColor;

                if (0 == a_Revision)
                {
                    l_IconType = IconTypes.NoIcon;
                    l_IconColor = Color.Empty;
                }
                else
                {
                    l_IconType = (IconTypes)a_Buffer.ReadByte();
                    l_IconColor = Color.FromArgb((int)a_Buffer.ReadUInt32LE());
                }

                int l_Idx = l_Categories.f_Customs.ReserveAtEnd();
                l_Arr = l_Categories.f_Customs.p_Arr;

                l_Arr[l_Idx] = new CategoryInfo(l_Category, l_Name, l_Visible);
                l_Arr[l_Idx].SetIconAttribs(l_IconType, l_IconColor);
            }

            return l_Categories;
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt32LE(c_ExtendedCatListMagic);
            a_Buffer.WriteByte(c_MyFormatRevision);
            a_Buffer.WriteUInt32LE(f_GenericsVisibilty);

            int l_Count = f_Customs.p_Count;

            a_Buffer.WriteUInt16LE((ushort)l_Count);

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0; i < l_Count; i++)
            {
                l_Arr[i].ToBuffer(a_Buffer);
            }
        }

        public CategoryInfo[] ExposeArray()
        {
            return f_Customs.p_Arr;
        }

        public void FillComboWithCustomCategories(LabeledIconComboBox<Category> a_ComboBox)
        {
            int l_Count = f_Customs.p_Count;
            LabeledIconComboBox<Category>.Item[] l_Items = new LabeledIconComboBox<Category>.Item[l_Count + 1];
            l_Items[0] = new LabeledIconComboBox<Category>.Item(string.Empty, null, Category.s_Unassigned);

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0; i < l_Count; i++)
            {
                l_Items[i + 1] = new LabeledIconComboBox<Category>.Item(l_Arr[i].f_Name, l_Arr[i].p_Icon, l_Arr[i].f_Category);
            }

            ComboBox l_Box = a_ComboBox.f_Box;

            l_Box.BeginUpdate();
            {
                l_Box.Items.Clear();
                l_Box.Items.AddRange(l_Items);
                l_Box.SelectedIndex = 0;
            }
            l_Box.EndUpdate();
        }

        public void FillMenuWithHiddenCategories(ContextMenuStripEx a_DestMenu,
                                                 string[] a_GenericCategoryNames,
                                                 string a_Format,
                                                 EventHandler a_Click,
                                                 bool a_AddSeparator)
        {
            int l_SeparatorIdx = a_DestMenu.Items.Count;

            // NOTE: We never expect the first category to be hidden.
            for (int i = 1; i < Category.c_GenericCount; i++)
            {
                if (false == IsGenericCategoryVisibleCore(i))
                {
                    string l_Text = string.Format(a_Format, a_GenericCategoryNames[i]);
                    GenericMenuItem<Category>.CreateItem(a_DestMenu, l_Text, a_Click, new Category((uint)i));
                }
            }

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0, lim = f_Customs.p_Count; i < lim; i++)
            {
                if (false == l_Arr[i].f_Visibile)
                {
                    string l_Text = string.Format(a_Format, l_Arr[i].f_Name);
                    GenericMenuItem<Category>.CreateItem(a_DestMenu, l_Text, a_Click, l_Arr[i].p_Icon, l_Arr[i].f_Category);
                }
            }

            // Add the separator if we had entries.

            if (true == a_AddSeparator &&
                a_DestMenu.Items.Count > l_SeparatorIdx)
            {
                MainForm.InsertMenuSeparator(a_DestMenu, l_SeparatorIdx);
            }
        }

        public void FillMenuWithCustomCategorySubMenu(TextSet a_TextSet,
                                                      ContextMenuStrip a_DestMenu,
                                                      TextKeys a_BaseItemTextKey,
                                                      bool a_AddItemForUnassigning,
                                                      EventHandler a_Click)
        {
            ToolStripMenuItem l_BaseItem = MainForm.AddMenuItem(a_DestMenu, a_TextSet[a_BaseItemTextKey], null, IconTypes.File);
            int l_Count = f_Customs.p_Count;

            if (0 == l_Count)
            {
                l_BaseItem.Enabled = false;

                return;
            }

            if (true == a_AddItemForUnassigning)
            {
                GenericMenuItem<Category>.CreateSubItem(l_BaseItem, a_TextSet[TextKeys.FilesTabListCntxtMenuNoCustomTab], a_Click, Category.s_Unassigned);
                MainForm.AddMenuSeparator(l_BaseItem);
            }

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0; i < l_Count; i++)
            {
                GenericMenuItem<Category>.CreateSubItem(l_BaseItem, l_Arr[i].f_Name, a_Click, l_Arr[i].p_Icon, l_Arr[i].f_Category);
            }

            return;
        }

        public ReorderItemsDialog<Category>.Item[] CreateItemsArrayForReordering(string[] a_GenericCategoryNames)
        {
            int l_CustomTabCount = f_Customs.p_Count;
            ReorderItemsDialog<Category>.Item[] l_Items = new ReorderItemsDialog<Category>.Item[Category.c_GenericCount + l_CustomTabCount];

            for (int k = 0; k < Category.c_GenericCount; k++)
            {
                bool l_Visible = (0 == k ? true : IsGenericCategoryVisibleCore(k));
                l_Items[k] = new ReorderItemsDialog<Category>.Item(a_GenericCategoryNames[k], new Category((uint)k), l_Visible, k);
            }

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0; i < l_CustomTabCount; i++)
            {
                int l_Idx = (Category.c_GenericCount + i);
                l_Items[l_Idx] = new ReorderItemsDialog<Category>.Item(l_Arr[i].f_Name, l_Arr[i].f_Category, l_Arr[i].f_Visibile, l_Idx);
            }

            return l_Items;
        }

        public void ApplyItemsArrayAfterReordering(ReorderItemsDialog<Category>.Item[] a_Items,
                                                   bool a_OrderChanged)
        {
            for (int k = 1; k < Category.c_GenericCount; k++)
            {
                AdjustVisibility(a_Items[k].f_Tag, a_Items[k].p_Visible);
            }

            int l_CustomTabCount = f_Customs.p_Count;

            if (0 >= l_CustomTabCount)
            {
                return;
            }

            int l_ItemsCount = a_Items.Length;

            if (false == a_OrderChanged)
            {
                // Just the visibility changed.

                for (int k = Category.c_GenericCount; k < l_ItemsCount; k++)
                {
                    AdjustVisibility(a_Items[k].f_Tag, a_Items[k].p_Visible);
                }

                return;
            }

            ReorderItemsDialog<Category>.Item.SortArrayByDisplayOrderIndex(a_Items);

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int k = Category.c_GenericCount; k < l_ItemsCount; k++)
            {
                bool l_Hit = false;

                int l_Idx = (k - Category.c_GenericCount);
                for (int i = l_Idx; i < l_CustomTabCount; i++)
                {
                    if (a_Items[k].f_Tag == l_Arr[i].f_Category)
                    {
                        // Update visibility.
                        l_Arr[i].f_Visibile = a_Items[k].p_Visible;

                        // Now move l_Arr[i] to l_Arr[Idx].

                        if (i != l_Idx)
                        {
                            Util.Swap(ref l_Arr[i], ref l_Arr[l_Idx]);
                        }

                        l_Hit = true;

                        break;
                    }
                }

                if (false == l_Hit)
                {
                    // We should never land here.
                    throw (new ApplicationException());
                }
            }
        }

        public bool FindCategoryByIndex(uint a_GlobalIndex,
                                        out Category a_Category)
        {
            if (Category.c_GenericCount > a_GlobalIndex)
            {
                a_Category = new Category(a_GlobalIndex);

                return true;
            }

            uint l_NonGenericIdx = (a_GlobalIndex - Category.c_GenericCount);

            if (f_Customs.p_Count > l_NonGenericIdx)
            {
                a_Category = f_Customs.p_Arr[l_NonGenericIdx].f_Category;

                return true;
            }

            a_Category = Category.s_Unassigned;

            return false;
        }

        public void PrepareListingPages(Category a_SelCategory,
                                        string[] a_GenericCategoryNames,
                                        ref AdvancedTabControl.PageListItem[] a_Items)
        {
            uint l_CustCount = (uint)f_Customs.p_Count;
            a_Items = new AdvancedTabControl.PageListItem[Category.c_GenericCount + l_CustCount];

            for (uint i = 0; i != Category.c_GenericCount; i++)
            {
                Category l_Cat = new Category(i);
                a_Items[i] = new AdvancedTabControl.PageListItem(i, GetListItemState(IsGenericCategoryVisibleCore((int)i), l_Cat == a_SelCategory), null, a_GenericCategoryNames[i]);
            }

            CategoryInfo[] l_Arr = f_Customs.p_Arr;

            for (uint i = Category.c_GenericCount, k = 0; k != l_CustCount; i++, k++)
            {
                a_Items[i] = new AdvancedTabControl.PageListItem(i, GetListItemState(l_Arr[k].f_Visibile, l_Arr[k].f_Category == a_SelCategory), l_Arr[k].p_Icon, l_Arr[k].f_Name);
            }
        }

        private static AdvancedTabControl.ListItemState GetListItemState(bool a_Visible,
                                                                         bool a_Selected)
        {
            return (false == a_Visible ? AdvancedTabControl.ListItemState.HiddenPage : (false == a_Selected ? AdvancedTabControl.ListItemState.VisiblePage : AdvancedTabControl.ListItemState.SelectedPage));
        }

        private int FindIndex(Category a_Category)
        {
            CategoryInfo[] l_Junk;
            return FindIndex(a_Category, out l_Junk);
        }

        private int FindIndex(Category a_Category,
                              out CategoryInfo[] a_Arr)
        {
            // Returns a negative value if no such category is found.

            a_Arr = f_Customs.p_Arr;

            if (true == a_Category.IsNonGeneric())
            {
                for (int i = (f_Customs.p_Count - 1); 0 <= i; i--)
                {
                    if (a_Arr[i].f_Category == a_Category)
                    {
                        return i;
                    }
                }
            }

            return (-1);
        }

        private int FindIndexByName(string a_Name,
                                    Category a_IgnoreConflict)
        {
            // a_IgnoreConflict can be an non-generic category that
            // shouldn't be tested against a_Name.

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = (f_Customs.p_Count - 1); 0 <= i; i--)
            {
                if (true == IsSameName(a_Name, l_Arr[i].f_Name) &&
                    l_Arr[i].f_Category != a_IgnoreConflict)
                {
                    return i;
                }
            }

            return (-1);
        }

        public bool FindVisibilityIndex(Category a_Category,
                                        out int a_VisibilityIndex)
        {
            // Returns true if the category has a display index. If
            // it's not visible, the return value is false. In this
            // case a_VisibilityIndex will indicate the display index
            // of the (visible) predecessor (there will always be one
            // for an invisible item).

            a_VisibilityIndex = (-1);

            for (ushort i = 0; Category.c_GenericCount > i; i++)
            {
                bool l_Visible = IsGenericCategoryVisibleCore(i);

                if (true == l_Visible)
                {
                    a_VisibilityIndex++;
                }

                if (a_Category.f_Val == i)
                {
                    return l_Visible;
                }
            }

            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = 0, lim = f_Customs.p_Count; i < lim; i++)
            {
                bool l_Visible = l_Arr[i].f_Visibile;

                if (true == l_Visible)
                {
                    a_VisibilityIndex++;
                }

                if (l_Arr[i].f_Category == a_Category)
                {
                    return l_Visible;
                }
            }

            // No such category.

            return false;
        }

        public bool HasCustomCategory(Category a_Category)
        {
            return (0 <= FindIndex(a_Category));
        }

        private Category FindUnoccupiedCategoryNum()
        {
            while (true)
            {
                uint l_Num = (Util.GetRandomUIntFast() % (ushort.MaxValue - Category.c_ReservedGenericNums + 1));
                Category l_Category = new Category(Category.c_ReservedGenericNums + l_Num);

                if (false == HasCustomCategory(l_Category))
                {
                    return l_Category;
                }
            }
        }

        public bool Add(string a_Name,
                        IconTypes a_IconType,
                        Color a_IconColor,
                        out Category a_Category)
        {
            if (c_MaxCustomCategories <= f_Customs.p_Count ||
                false == IsNameLegal(a_Name))
            {
                a_Category = new Category();

                return false;
            }

            int l_Idx = FindIndexByName(a_Name, new Category());

            if (0 <= l_Idx)
            {
                // This name is taken.

                a_Category = new Category();

                return false;
            }

            a_Category = FindUnoccupiedCategoryNum();

            l_Idx = f_Customs.ReserveAtEnd();
            CategoryInfo[] l_Arr = f_Customs.p_Arr;

            l_Arr[l_Idx] = new CategoryInfo(a_Category, a_Name, true);
            l_Arr[l_Idx].SetIconAttribs(a_IconType, a_IconColor);

            return true;
        }

        public bool Modify(Category a_Category,
                           string a_NewName,
                           IconTypes a_IconType,
                           Color a_IconColor)
        {
            if (false == IsNameLegal(a_NewName))
            {
                return false;
            }

            int l_Idx = FindIndexByName(a_NewName, a_Category);

            if (0 <= l_Idx)
            {
                // This name is taken.

                return false;
            }

            CategoryInfo[] l_Arr;
            l_Idx = FindIndex(a_Category, out l_Arr);

            if (0 > l_Idx)
            {
                // No such category.

                return false;
            }

            l_Arr[l_Idx].f_Name = a_NewName;
            l_Arr[l_Idx].SetIconAttribs(a_IconType, a_IconColor);

            return true;
        }

        public bool GetName(Category a_Category,
                            out string a_Name)
        {
            bool l_Junk0;
            Bitmap l_Junk1;
            return GetAttributes(a_Category, out a_Name, out l_Junk0, out l_Junk1);
        }

        public bool GetAttributes(Category a_Category,
                                  out string a_Name,
                                  out bool a_Visibilty,
                                  out Bitmap a_Icon)
        {
            CategoryInfo[] l_Arr;
            int l_Idx = FindIndex(a_Category, out l_Arr);

            if (0 > l_Idx)
            {
                a_Name = null;
                a_Visibilty = false;
                a_Icon = null;

                return false;
            }

            a_Name = l_Arr[l_Idx].f_Name;
            a_Visibilty = l_Arr[l_Idx].f_Visibile;
            a_Icon = l_Arr[l_Idx].p_Icon;

            return true;
        }

        public bool GetAttributes(Category a_Category,
                                  out string a_Name,
                                  out IconTypes a_IconType,
                                  out Color a_IconColor)
        {
            CategoryInfo[] l_Arr;
            int l_Idx = FindIndex(a_Category, out l_Arr);

            if (0 > l_Idx)
            {
                a_Name = null;
                a_IconType = IconTypes.NoIcon;
                a_IconColor = Color.Empty;

                return false;
            }

            a_Name = l_Arr[l_Idx].f_Name;
            l_Arr[l_Idx].GetIconAttribs(out a_IconType, out a_IconColor);

            return true;
        }

        public bool Remove(Category a_Category)
        {
            CategoryInfo[] l_Arr;
            int l_Idx = FindIndex(a_Category, out l_Arr);

            if (0 > l_Idx)
            {
                return false;
            }

            l_Arr[l_Idx].Dispose();
            f_Customs.DeleteAt(l_Idx);

            return true;
        }

        public uint GetGenericVisibleCount()
        {
            return BitMap.CountSetBits(f_GenericsVisibilty);
        }

        public bool IsMoreThanOneCustomCategoryVisible()
        {
            uint l_Count = (uint)f_Customs.p_Count;
            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            bool l_FoundOne = false;

            for (uint i = 0; i < l_Count; i++)
            {
                if (true == l_Arr[i].f_Visibile)
                {
                    if (true == l_FoundOne)
                    {
                        return true;
                    }

                    l_FoundOne = true;
                }
            }

            return false;
        }

        public bool IsCategoryVisible(Category a_Category)
        {
            // If the passed category does not exists, this routine
            // returns zero.

            if (true == a_Category.IsGeneric())
            {
                return IsGenericCategoryVisible(a_Category);
            }

            CategoryInfo[] l_Arr;
            int l_Idx = FindIndex(a_Category, out l_Arr);

            return (0 <= l_Idx &&
                    true == l_Arr[l_Idx].f_Visibile);
        }

        public bool IsGenericCategoryVisible(Category a_Category)
        {
            return IsGenericCategoryVisibleCore((int)a_Category.f_Val);
        }

        private bool IsGenericCategoryVisibleCore(int a_CategoryValue)
        {
            return (0 != (f_GenericsVisibilty & GetGenericVisibilityBit(a_CategoryValue)));
        }

        public bool Hide(Category a_Category)
        {
            return AdjustVisibility(a_Category, false);
        }

        public bool Show(Category a_Category)
        {
            return AdjustVisibility(a_Category, true);
        }

        private bool AdjustVisibility(Category a_Category,
                                      bool a_Visible)
        {
            if (true == a_Category.IsNonGeneric())
            {
                CategoryInfo[] l_Arr;
                int l_Idx = FindIndex(a_Category, out l_Arr);

                if (0 > l_Idx)
                {
                    return false;
                }

                l_Arr[l_Idx].f_Visibile = a_Visible;

                return true;
            }

            if (0 == a_Category.f_Val ||
                Category.c_GenericCount <= a_Category.f_Val)
            {
                // Disallow adjusting the visibility of the 'all'
                // category.

                return false;
            }

            uint l_Mask = GetGenericVisibilityBit((int)a_Category.f_Val);

            if (false == a_Visible)
            {
                f_GenericsVisibilty &= ~l_Mask;
            }
            else
            {
                f_GenericsVisibilty |= l_Mask;
            }

            return true;
        }

        public void AdjustIndex(Category a_Category,
                                Category a_PredecessorCategory)
        {
            if (true == a_Category.IsGeneric())
            {
                return;
            }

            int l_OldIdx = FindIndex(a_Category);

            if (0 > l_OldIdx)
            {
                return;
            }

            if (true == a_PredecessorCategory.IsGeneric())
            {
                MoveItem(l_OldIdx, 0);

                return;
            }

            int l_PredIdx = FindIndex(a_PredecessorCategory);

            MoveItem(l_OldIdx, l_OldIdx <= l_PredIdx ? l_PredIdx : (l_PredIdx + 1));
        }

        private void MoveItem(int a_OldIdx,
                              int a_NewIdx)
        {
            if (a_OldIdx == a_NewIdx)
            {
                return;
            }

            int l_EffNewIdx = (a_OldIdx < a_NewIdx ? (a_NewIdx + 1) : a_NewIdx);
            int l_EffOldIdx = (a_OldIdx < a_NewIdx ? a_OldIdx : (a_OldIdx + 1));

            f_Customs.ReserveAt(l_EffNewIdx);
            CategoryInfo[] l_Arr = f_Customs.p_Arr;

            l_Arr[l_EffNewIdx] = l_Arr[l_EffOldIdx];
            f_Customs.DeleteAt(l_EffOldIdx);
        }

        public void NotifyDefaultIconBgColorChanged()
        {
            CategoryInfo[] l_Arr = f_Customs.p_Arr;
            for (int i = (f_Customs.p_Count - 1); 0 <= i; i--)
            {
                l_Arr[i].NotifyDefaultIconBgColorChanged();
            }
        }

        private static uint GetGenericVisibilityBit(int a_CategoryValue)
        {
            return (1u << a_CategoryValue);
        }

        public static bool IsSameName(string a_A,
                                      string a_B)
        {
            return Util.StringsEqualIgnoreCase(a_A, a_B);
        }

        private static bool IsNameLegal(string a_PossibleName)
        {
            return (0 < a_PossibleName.Length &&
                    c_MaxCategoryNameChars >= a_PossibleName.Length);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct CategoryInfo : IDisposable
        {
            public readonly Category f_Category;
            public string f_Name;
            public bool f_Visibile;
            private IconTypes f_IconType;
            private Color f_IconColor;

            // f_Icon can be null.
            private Bitmap f_Icon;

            public Bitmap p_Icon
            {
                get { return f_Icon; }
            }

            public CategoryInfo(Category a_Category,
                                string a_Name,
                                bool a_Visible)
            {
                f_Category = a_Category;
                f_Name = a_Name;
                f_Visibile = a_Visible;
                f_IconType = IconTypes.NoIcon;
                f_IconColor = Color.Empty;
                f_Icon = null;
            }

            public void Dispose()
            {
                DisposeIcon();
            }

            private void DisposeIcon()
            {
                if (null != f_Icon)
                {
                    f_Icon.Dispose();
                    f_Icon = null;
                }
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteUInt16LE(f_Category.f_Val);
                a_Buffer.WriteUTF16LEStringEncLen(f_Name);
                a_Buffer.WriteBool(f_Visibile);
                a_Buffer.WriteByte((byte)f_IconType);
                a_Buffer.WriteUInt32LE((uint)f_IconColor.ToArgb());
            }

            public void GetIconAttribs(out IconTypes a_IconType,
                                       out Color a_IconColor)
            {
                a_IconType = f_IconType;
                a_IconColor = f_IconColor;
            }

            public void SetIconAttribs(IconTypes a_IconType,
                                       Color a_IconColor)
            {
                if (a_IconType == f_IconType &&
                    a_IconColor == f_IconColor)
                {
                    return;
                }

                f_IconType = a_IconType;
                f_IconColor = a_IconColor;

                UpdateIcon();
            }

            public void UpdateIcon()
            {
                if (IconTypes.NoIcon == f_IconType)
                {
                    DisposeIcon();

                    return;
                }

                if (null == f_Icon)
                {
                    f_Icon = IconStorage.BuildEmptyIcon();
                }

                Color l_EffBgColor = ColorPicker.ChooseColorEasy(f_IconColor, Software.s_IconStorage.p_BackgroundColor);
                Software.s_IconStorage.FillIcon(f_IconType, l_EffBgColor, f_Icon);
            }

            public void NotifyDefaultIconBgColorChanged()
            {
                // This routine assumes that the icon is already up to
                // date except for the default background color.

                if (IconTypes.NoIcon != f_IconType &&
                    false == ColorPicker.IsColorUsable(f_IconColor))
                {
                    // This icon uses the default color, so update.

                    UpdateIcon();
                }
            }
        }
    }

    class Preferences
    {
        public GeneralPrefs f_GeneralPrefs;
        public AssocPrefs f_AssocPrefs;
        public NetworkPrefs f_NetworkPrefs;
        public KadPrefs f_KadPrefs;
        public IOPrefs f_IOPrefs;

        private Preferences(OwnUserPersona a_OwnPersona)
        {
            GeneralPrefs.CreateDefault(a_OwnPersona, out f_GeneralPrefs);
            AssocPrefs.CreateDefault(out f_AssocPrefs);
            NetworkPrefs.CreateDefault(out f_NetworkPrefs);
            KadPrefs.CreateDefault(out f_KadPrefs);
            IOPrefs.CreateDefault(out f_IOPrefs);
        }

        public Preferences(ref GeneralPrefs a_GeneralPrefs,
                           ref AssocPrefs a_AssocPrefs,
                           ref NetworkPrefs a_NetworkPrefs,
                           ref KadPrefs a_KadPrefs,
                           ref IOPrefs a_IOPrefs)
        {
            f_GeneralPrefs = a_GeneralPrefs;
            f_AssocPrefs = a_AssocPrefs;
            f_NetworkPrefs = a_NetworkPrefs;
            f_KadPrefs = a_KadPrefs;
            f_IOPrefs = a_IOPrefs;
        }

        public Preferences RecreateWithNewIOPrefs(ref IOPrefs a_IOPrefs)
        {
            return new Preferences(ref f_GeneralPrefs, ref f_AssocPrefs, ref f_NetworkPrefs, ref f_KadPrefs, ref a_IOPrefs);
        }

        public static Preferences FromBuffer(BufferIO a_Buffer,
                                             out ForwardData[] a_ForwardData)
        {
            // Returns null on error. a_ForwardData can be null on
            // return in any case.

            GeneralPrefs l_GeneralPrefs;
            AssocPrefs l_AssocPrefs;
            NetworkPrefs l_NetworkPrefs;
            KadPrefs l_KadPrefs;
            IOPrefs l_IOPrefs;

            ForwardInformation l_ForwardInfo = new ForwardInformation(a_Buffer, null);
            uint l_Revision;

            if (// GeneralPrefs
                false == l_ForwardInfo.PrepareRead(out l_Revision) ||
                false == GeneralPrefs.FromBuffer(a_Buffer, l_Revision, out l_GeneralPrefs) ||
                // AssocPrefs
                false == l_ForwardInfo.PrepareRead(out l_Revision) ||
                false == AssocPrefs.FromBuffer(a_Buffer, l_Revision, out l_AssocPrefs) ||
                // NetworkPrefs
                false == l_ForwardInfo.PrepareRead(out l_Revision) ||
                false == NetworkPrefs.FromBuffer(a_Buffer, l_Revision, out l_NetworkPrefs) ||
                // KadPrefs
                false == l_ForwardInfo.PrepareRead(out l_Revision) ||
                false == KadPrefs.FromBuffer(a_Buffer, l_Revision, out l_KadPrefs) ||
                // IOPrefs
                false == l_ForwardInfo.PrepareRead(out l_Revision) ||
                false == IOPrefs.FromBuffer(a_Buffer, l_Revision, out l_IOPrefs) ||
                false == l_ForwardInfo.FinishLastRead())
            {
                a_ForwardData = null;

                return null;
            }

            a_ForwardData = l_ForwardInfo.GetForwardData();

            return new Preferences(ref l_GeneralPrefs, ref l_AssocPrefs, ref l_NetworkPrefs, ref l_KadPrefs, ref l_IOPrefs);
        }

        public void ToBuffer(BufferIO a_Buffer,
                             ForwardData[] a_ForwardData)
        {
            ForwardInformation l_ForwardInfo = new ForwardInformation(a_Buffer, a_ForwardData);
            // GeneralPrefs
            l_ForwardInfo.PrepareFirstWrite();
            l_ForwardInfo.FinishWrite(f_GeneralPrefs.ToBuffer(a_Buffer));
            // AssocPrefs
            l_ForwardInfo.FinishWrite(f_AssocPrefs.ToBuffer(a_Buffer));
            // NetworkPrefs
            l_ForwardInfo.FinishWrite(f_NetworkPrefs.ToBuffer(a_Buffer));
            // KadPrefs
            l_ForwardInfo.FinishWrite(f_KadPrefs.ToBuffer(a_Buffer));
            // IOPrefs
            l_ForwardInfo.FinishWrite(f_IOPrefs.ToBuffer(a_Buffer));
        }

        public static Preferences CreateDefault(OwnUserPersona a_OwnPersona)
        {
            return new Preferences(a_OwnPersona);
        }

        public static PrefChangeKinds DetermineChangeKind(Preferences a_From,
                                                          Preferences a_To)
        {
            PrefChangeKinds l_Kind = PrefChangeKinds.NoChangeAtAll;

            a_From.f_GeneralPrefs.DetermineChangeKind(ref a_To.f_GeneralPrefs, ref l_Kind);
            a_From.f_AssocPrefs.DetermineChangeKind(ref a_To.f_AssocPrefs, ref l_Kind);
            a_From.f_NetworkPrefs.DetermineChangeKind(ref a_To.f_NetworkPrefs, ref l_Kind);
            a_From.f_KadPrefs.DetermineChangeKind(ref a_To.f_KadPrefs, ref l_Kind);
            a_From.f_IOPrefs.DetermineChangeKind(ref a_To.f_IOPrefs, ref l_Kind);

            return l_Kind;
        }

        public static bool NeedsReconnOrRestart(PrefChangeKinds a_ChangeKind)
        {
            return (PrefChangeKinds.ReconnectNeededToFullyTakeOver <= a_ChangeKind);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ForwardData
        {
            public readonly byte f_Revision;
            public readonly byte[] f_ExtraData;

            public ForwardData(uint a_Revision,
                               byte[] a_ExtraData)
            {
                f_Revision = (byte)a_Revision;
                f_ExtraData = a_ExtraData;
            }

            public bool HasData()
            {
                return (null != f_ExtraData &&
                        0 < f_ExtraData.Length);
            }
        }

        private class ForwardInformation
        {
            private const byte c_FirstRevisionWithLengthInfo = 1;
            private const byte c_SubCount = 5;
            private const byte c_HeaderSize = (sizeof(byte) + sizeof(uint));
            private const uint c_MaxExtraSize = (4 * 1024 * 1024);

            private readonly ForwardData[] f_ForwardData;
            private readonly BufferIO f_Buffer;

            private int f_LastPrefIdx;
            private uint f_HeaderPos;
            private uint f_BodyPos;

            private uint f_ReadRevision;
            private bool f_HasLength;
            private uint f_Length;

            public ForwardInformation(BufferIO a_Buffer,
                                      ForwardData[] a_ForwardData)
            {
                // a_ForwardData may be null.

                f_Buffer = a_Buffer;

                if (null == a_ForwardData)
                {
                    f_ForwardData = new ForwardData[c_SubCount];
                }
                else
                {
                    f_ForwardData = a_ForwardData;
                }
            }

            public ForwardData[] GetForwardData()
            {
                // Returns can be null if there is no extra content.

                for (int i = 0; i < c_SubCount; i++)
                {
                    if (true == f_ForwardData[i].HasData())
                    {
                        return f_ForwardData;
                    }
                }

                return null;
            }

            public bool PrepareRead(out uint a_Revision)
            {
                if (0 != f_BodyPos)
                {
                    // Finish the previous read.

                    if (false == FinishRead())
                    {
                        a_Revision = 0;

                        return false;
                    }
                }

                // Read the header.

                f_HeaderPos = f_Buffer.p_Position;
                f_ReadRevision = f_Buffer.ReadByte();
                f_HasLength = (c_FirstRevisionWithLengthInfo <= f_ReadRevision);
                f_Length = (false == f_HasLength ? 0 : f_Buffer.ReadUInt32LE());

                // Header read.

                f_BodyPos = f_Buffer.p_Position;
                a_Revision = f_ReadRevision;

                return true;
            }

            public bool FinishLastRead()
            {
                return FinishRead();
            }

            private bool FinishRead()
            {
                if (true == f_HasLength)
                {
                    uint l_ReadStopPos = f_Buffer.p_Position;

                    uint l_UsedLength = (l_ReadStopPos - f_BodyPos);

                    if (l_UsedLength != f_Length)
                    {
                        if (l_UsedLength > f_Length)
                        {
                            // Corruption.

                            Util.DebugBreak();

                            return false;
                        }

                        uint l_ExtraSize = (f_Length - l_UsedLength);

                        if (c_MaxExtraSize < l_ExtraSize)
                        {
                            // Unexpected.

                            Util.DebugBreak();

                            return false;
                        }

                        f_ForwardData[f_LastPrefIdx] = new ForwardData(f_ReadRevision, f_Buffer.ReadBytes(l_ExtraSize));
                    }
                }

                f_LastPrefIdx++;

                return true;
            }

            public void PrepareFirstWrite()
            {
                PrepareWrite();
            }

            public void FinishWrite(uint a_MinRevision)
            {
                uint l_ForwardRevision = f_ForwardData[f_LastPrefIdx].f_Revision;
                uint l_EffRevision;

                if (l_ForwardRevision <= a_MinRevision)
                {
                    l_EffRevision = a_MinRevision;
                }
                else
                {
                    // We have extra data.

                    l_EffRevision = l_ForwardRevision;
                    f_Buffer.WriteBytes(f_ForwardData[f_LastPrefIdx].f_ExtraData);
                }

                uint l_WriteStopPos = f_Buffer.p_Position;

                f_Buffer.SetPosition(f_HeaderPos);
                f_Buffer.WriteByte((byte)l_EffRevision);
                f_Buffer.WriteUInt32LE(l_WriteStopPos - f_BodyPos);

                f_Buffer.SetPosition(l_WriteStopPos);
                f_LastPrefIdx++;

                if (c_SubCount > f_LastPrefIdx)
                {
                    PrepareWrite();
                }
                else
                {
                }
            }

            private void PrepareWrite()
            {
                f_HeaderPos = f_Buffer.p_Position;
                f_Buffer.JumpAhead(c_HeaderSize);
                f_BodyPos = f_Buffer.p_Position;
            }
        }
    }

    enum PrefChangeKinds : byte
    {
        NoChangeAtAll = 0,
        ApplicableWithoutReconnect = 1,
        ReconnectNeededToFullyTakeOver = 2,
        AppRestartNeededToFullyTakeOver = 3,
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct GeneralPrefs
    {
        // NOTE: Don't drop the datagram address very often, otherwise
        // the Kademlia network will suffer.
        public const ushort c_AutoDropDatagramKeyPairDays = 365;
        public const ushort c_AutoDropStreamKeyPairDays = 90;
        public const byte c_MinGridViewTextSizeRatio = 75;
        public const byte c_MaxGridViewTextSizeRatio = 250;
        public const byte c_DefaultGridViewTextSizeRatio = 125;
        private const bool c_DefaultUseBgImgsForComplBar = false;
        private const ushort c_DefaultGridViewAutoSortItemThreshold = 250;
        // This is an arbitrary limit.
        public const uint c_MaxGridViewAutoSortItemThreshold = 100000;
        private const bool c_DefaultShowPeersPanelAdvSubMenu = false;
        private const bool c_RunFilesOnDblClick = false;
        public const MainForm.MainTabs c_InvalidPreferredMainTab = (MainForm.MainTabs)byte.MaxValue;

        // GUI managed preferences.

        public readonly bool f_SaveSearchHistory;
        public readonly bool f_ShowSpeedOnTitle;
        public readonly bool f_ShowTrayIcon;
        public readonly bool f_ShowBalloonNotifications;
        public readonly bool f_MinimizeToTray;
        public readonly bool f_PromptOnExit;
        public readonly bool f_BeepWhenDlIsDone;
        public readonly bool f_DisplaySISizes;
        public readonly ProgessBarKinds f_ProgressBarKind;
        public readonly bool f_CompactOnExit;
        public readonly ConnStartKinds f_ConnStartKind;
        public readonly OwnUserPersona f_OwnPersona;
        public readonly bool f_UseInboundTunnelSettingsForOutboundTunnels;
        public readonly IPEndPoint f_SamEndPoint;
        public readonly TunnelSettings f_InboundTunnelSettings;
        public readonly TunnelSettings f_OutboundTunnelSettings;
        // Both keys can be null.
        private I2PKeyPair f_DatagramKeyPair;
        private I2PKeyPair f_StreamKeyPair;
        public readonly bool f_AutoManageDatagramKeys;
        public readonly bool f_AutoManageStreamKeys;
        // From v1.3.0 on.
        public readonly string f_PreviewFolder;
        public readonly byte f_GridViewTextSizeRatio;
        public readonly GridViewBorderStyles f_GridViewBorderStyle;
        public readonly bool f_AllowComplFilesDeletion;
        // From v1.4.0 on.
        public readonly bool f_UseBackgroundImagesForComplBar;
        public readonly uint f_GridViewAutoSortItemThreshold;
        public readonly bool f_ShowPeersPanelAdvSubMenu;
        public readonly Color f_IconBgColor;
        public readonly bool f_RunFilesOnDblClick;
        // From v1.4.9 on.
        public readonly bool f_UIModeAdvanced;
        // From v1.4.21 on.
        public readonly MainForm.MainTabs f_PreferredMainTab;
        public readonly bool f_AdditionalKeyCombinationForExiting;

        public I2PKeyPair p_DatagramKeyPair
        {
            get { return f_DatagramKeyPair; }
        }

        public I2PKeyPair p_StreamKeyPair
        {
            get { return f_StreamKeyPair; }
        }

        public GeneralPrefs(bool a_SaveSearchHistory,
                            bool a_ShowSpeedOnTitle,
                            bool a_ShowTrayIcon,
                            bool a_ShowBalloonNotifications,
                            bool a_MinimizeToTray,
                            bool a_PromptOnExit,
                            bool a_BeepWhenDlIsDone,
                            bool a_DisplaySISizes,
                            ProgessBarKinds a_ProgressBarKind,
                            bool a_CompactOnExit,
                            ConnStartKinds a_ConnStartKind,
                            OwnUserPersona a_OwnPersona,
                            bool a_UseInboundTunnelSettingsForOutboundTunnels,
                            IPEndPoint a_SamEndPoint,
                            TunnelSettings a_InboundTunnelSettings,
                            TunnelSettings a_OutboundTunnelSettings,
                            I2PKeyPair a_DatagramKeyPair,
                            I2PKeyPair a_StreamKeyPair,
                            bool a_AutoManageDatagramKeys,
                            bool a_AutoManageStreamKeys,
                            string a_PreviewFolder,
                            byte a_GridViewTextSizeRatio,
                            GridViewBorderStyles a_GridViewBorderStyle,
                            bool a_AllowComplFilesDeletion,
                            bool a_UseBackgroundImagesForComplBar,
                            uint a_GridViewAutoSortItemThreshold,
                            bool a_ShowPeersPanelAdvSubMenu,
                            Color a_IconBgColor,
                            bool a_RunFilesOnDblClick,
                            bool a_UIModeAdvanced,
                            MainForm.MainTabs a_PreferredMainTab,
                            bool a_AdditionalKeyCombinationForExiting)
        {
            f_SaveSearchHistory = a_SaveSearchHistory;
            f_ShowSpeedOnTitle = a_ShowSpeedOnTitle;
            f_ShowTrayIcon = a_ShowTrayIcon;
            f_ShowBalloonNotifications = a_ShowBalloonNotifications;
            f_MinimizeToTray = a_MinimizeToTray;
            f_PromptOnExit = a_PromptOnExit;
            f_BeepWhenDlIsDone = a_BeepWhenDlIsDone;
            f_DisplaySISizes = a_DisplaySISizes;
            f_ProgressBarKind = a_ProgressBarKind;
            f_CompactOnExit = a_CompactOnExit;
            f_ConnStartKind = a_ConnStartKind;
            f_OwnPersona = a_OwnPersona;
            f_UseInboundTunnelSettingsForOutboundTunnels = a_UseInboundTunnelSettingsForOutboundTunnels;
            f_SamEndPoint = a_SamEndPoint;
            f_InboundTunnelSettings = a_InboundTunnelSettings;
            f_OutboundTunnelSettings = a_OutboundTunnelSettings;
            f_DatagramKeyPair = a_DatagramKeyPair;
            f_StreamKeyPair = a_StreamKeyPair;
            f_AutoManageDatagramKeys = a_AutoManageDatagramKeys;
            f_AutoManageStreamKeys = a_AutoManageStreamKeys;
            f_PreviewFolder = a_PreviewFolder;
            f_GridViewTextSizeRatio = a_GridViewTextSizeRatio;
            f_GridViewBorderStyle = a_GridViewBorderStyle;
            f_AllowComplFilesDeletion = a_AllowComplFilesDeletion;
            f_UseBackgroundImagesForComplBar = a_UseBackgroundImagesForComplBar;
            f_GridViewAutoSortItemThreshold = a_GridViewAutoSortItemThreshold;
            f_ShowPeersPanelAdvSubMenu = a_ShowPeersPanelAdvSubMenu;
            f_IconBgColor = a_IconBgColor;
            f_RunFilesOnDblClick = a_RunFilesOnDblClick;
            f_UIModeAdvanced = a_UIModeAdvanced;
            f_PreferredMainTab = a_PreferredMainTab;
            f_AdditionalKeyCombinationForExiting = a_AdditionalKeyCombinationForExiting;
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      uint a_Revision,
                                      out GeneralPrefs a_Dest)
        {
            bool l_SaveSearchHistory = a_Buffer.ReadBool();
            bool l_ShowSpeedOnTitle = a_Buffer.ReadBool();
            bool l_ShowTrayIcon = a_Buffer.ReadBool();
            bool l_ShowBalloonNotifications = a_Buffer.ReadBool();
            bool l_MinimizeToTray = a_Buffer.ReadBool();
            bool l_PromptOnExit = a_Buffer.ReadBool();
            bool l_BeepWhenDlIsDone = a_Buffer.ReadBool();
            bool l_DisplaySISizes = a_Buffer.ReadBool();
            ProgessBarKinds l_ProgressBarKind = (ProgessBarKinds)a_Buffer.ReadByte();
            bool l_CompactOnExit = a_Buffer.ReadBool();
            ConnStartKinds l_ConnStartKind = (ConnStartKinds)a_Buffer.ReadByte();
            OwnUserPersona l_OwnPersona = OwnUserPersona.FromFile(a_Buffer);
            bool l_UseInboundTunnelSettingsForOutboundTunnels = a_Buffer.ReadBool();
            IPEndPoint l_SamEndPoint;
            TunnelSettings l_InboundTunnelSettings;
            TunnelSettings l_OutboundTunnelSettings;
            I2PKeyPair l_DatagramKeyPair;
            I2PKeyPair l_StreamKeyPair;

            if (ProgessBarKinds.IncludeUnverified < l_ProgressBarKind ||
                ConnStartKinds.AutoStart < l_ConnStartKind ||
                null == l_OwnPersona ||
                false == a_Buffer.ReadIPEndPoint(out l_SamEndPoint) ||
                false == TunnelSettings.FromBuffer(a_Buffer, out l_InboundTunnelSettings) ||
                false == TunnelSettings.FromBuffer(a_Buffer, out l_OutboundTunnelSettings) ||
                false == I2PKeyPairFromBuffer(a_Buffer, out l_DatagramKeyPair) ||
                false == I2PKeyPairFromBuffer(a_Buffer, out l_StreamKeyPair))
            {
                Util.DebugBreak();

                a_Dest = new GeneralPrefs();

                return false;
            }

            bool l_AutoManageDatagramKeys = a_Buffer.ReadBool();
            bool l_AutoManageStreamKeys = a_Buffer.ReadBool();
            string l_PreviewFolder;
            byte l_GridViewTextSizeRatio;
            GridViewBorderStyles l_GridViewBorderStyle;
            bool l_AllowComplFilesDeletion;

            if (1 > a_Revision)
            {
                l_PreviewFolder = Software.sp_DefaultPreviewFolder;
                l_GridViewTextSizeRatio = c_DefaultGridViewTextSizeRatio;
                l_GridViewBorderStyle = GridViewBorderStyles.NoBorder;
                l_AllowComplFilesDeletion = false;
            }
            else
            {
                l_PreviewFolder = Software.ReadAbsPathUTF16LEStringEncLen(a_Buffer);
                l_GridViewTextSizeRatio = a_Buffer.ReadByte();
                l_GridViewBorderStyle = (GridViewBorderStyles)a_Buffer.ReadByte();
                l_AllowComplFilesDeletion = a_Buffer.ReadBool();

                if (c_MinGridViewTextSizeRatio > l_GridViewTextSizeRatio ||
                    c_MaxGridViewTextSizeRatio < l_GridViewTextSizeRatio ||
                    GridViewBorderStyles.Both < l_GridViewBorderStyle)
                {
                    Util.DebugBreak();

                    a_Dest = new GeneralPrefs();

                    return false;
                }
            }

            bool l_UseBackgroundImagesForComplBar;
            uint l_GridViewAutoSortItemThreshold;
            bool l_ShowPeersPanelAdvSubMenu;
            Color l_IconBgColor;
            bool l_RunOrPreviewFileOnDblClick;

            if (2 > a_Revision)
            {
                l_UseBackgroundImagesForComplBar = c_DefaultUseBgImgsForComplBar;
                l_GridViewAutoSortItemThreshold = c_DefaultGridViewAutoSortItemThreshold;
                l_ShowPeersPanelAdvSubMenu = c_DefaultShowPeersPanelAdvSubMenu;
                l_IconBgColor = Color.Empty;
                l_RunOrPreviewFileOnDblClick = c_RunFilesOnDblClick;
            }
            else
            {
                l_UseBackgroundImagesForComplBar = a_Buffer.ReadBool();
                l_GridViewAutoSortItemThreshold = Math.Min(a_Buffer.ReadUInt32LE(), c_MaxGridViewAutoSortItemThreshold);
                l_ShowPeersPanelAdvSubMenu = a_Buffer.ReadBool();
                l_IconBgColor = Color.FromArgb((int)a_Buffer.ReadUInt32LE());
                l_RunOrPreviewFileOnDblClick = a_Buffer.ReadBool();
            }

            bool l_UIModeAdvanced;

            if (3 > a_Revision)
            {
                l_UIModeAdvanced = true;
            }
            else
            {
                l_UIModeAdvanced = a_Buffer.ReadBool();
            }

            MainForm.MainTabs l_PreferredMainTab;
            bool l_AdditionalKeyCombinationForExiting;

            if (4 > a_Revision)
            {
                l_PreferredMainTab = MainForm.MainTabs.Files;
                l_AdditionalKeyCombinationForExiting = false;
            }
            else
            {
                l_PreferredMainTab = (MainForm.MainTabs)a_Buffer.ReadByte();

                if (MainForm.c_MainTabsCount <= (uint)l_PreferredMainTab)
                {
                    l_PreferredMainTab = c_InvalidPreferredMainTab;
                }

                l_AdditionalKeyCombinationForExiting = a_Buffer.ReadBool();
            }

            a_Dest = new GeneralPrefs(l_SaveSearchHistory, l_ShowSpeedOnTitle, l_ShowTrayIcon, l_ShowBalloonNotifications, l_MinimizeToTray, l_PromptOnExit, l_BeepWhenDlIsDone, l_DisplaySISizes, l_ProgressBarKind, l_CompactOnExit, l_ConnStartKind, l_OwnPersona, l_UseInboundTunnelSettingsForOutboundTunnels, l_SamEndPoint, l_InboundTunnelSettings, l_OutboundTunnelSettings, l_DatagramKeyPair, l_StreamKeyPair, l_AutoManageDatagramKeys, l_AutoManageStreamKeys, l_PreviewFolder, l_GridViewTextSizeRatio, l_GridViewBorderStyle, l_AllowComplFilesDeletion, l_UseBackgroundImagesForComplBar, l_GridViewAutoSortItemThreshold, l_ShowPeersPanelAdvSubMenu, l_IconBgColor, l_RunOrPreviewFileOnDblClick, l_UIModeAdvanced, l_PreferredMainTab, l_AdditionalKeyCombinationForExiting);

            return true;
        }

        public uint ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteBool(f_SaveSearchHistory);
            a_Buffer.WriteBool(f_ShowSpeedOnTitle);
            a_Buffer.WriteBool(f_ShowTrayIcon);
            a_Buffer.WriteBool(f_ShowBalloonNotifications);
            a_Buffer.WriteBool(f_MinimizeToTray);
            a_Buffer.WriteBool(f_PromptOnExit);
            a_Buffer.WriteBool(f_BeepWhenDlIsDone);
            a_Buffer.WriteBool(f_DisplaySISizes);
            a_Buffer.WriteByte((byte)f_ProgressBarKind);
            a_Buffer.WriteBool(f_CompactOnExit);
            a_Buffer.WriteByte((byte)f_ConnStartKind);
            f_OwnPersona.ToBuffer(a_Buffer);
            a_Buffer.WriteBool(f_UseInboundTunnelSettingsForOutboundTunnels);
            a_Buffer.WriteIPEndPoint(f_SamEndPoint);
            f_InboundTunnelSettings.ToBuffer(a_Buffer);
            f_OutboundTunnelSettings.ToBuffer(a_Buffer);
            I2PKeyPairToBuffer(a_Buffer, f_DatagramKeyPair);
            I2PKeyPairToBuffer(a_Buffer, f_StreamKeyPair);
            a_Buffer.WriteBool(f_AutoManageDatagramKeys);
            a_Buffer.WriteBool(f_AutoManageStreamKeys);
            a_Buffer.WriteUTF16LEStringEncLen(f_PreviewFolder);
            a_Buffer.WriteByte(f_GridViewTextSizeRatio);
            a_Buffer.WriteByte((byte)f_GridViewBorderStyle);
            a_Buffer.WriteBool(f_AllowComplFilesDeletion);
            a_Buffer.WriteBool(f_UseBackgroundImagesForComplBar);
            a_Buffer.WriteUInt32LE(f_GridViewAutoSortItemThreshold);
            a_Buffer.WriteBool(f_ShowPeersPanelAdvSubMenu);
            a_Buffer.WriteUInt32LE((uint)f_IconBgColor.ToArgb());
            a_Buffer.WriteBool(f_RunFilesOnDblClick);
            a_Buffer.WriteBool(f_UIModeAdvanced);
            a_Buffer.WriteByte((byte)f_PreferredMainTab);
            a_Buffer.WriteBool(f_AdditionalKeyCombinationForExiting);

            // Increase with every change.
            return 4;
        }

        public static void CreateDefault(OwnUserPersona a_OwnPersona,
                                         out GeneralPrefs a_Dest)
        {
            IPEndPoint l_SamEndPoint = new IPEndPoint(IPAddress.Loopback, I2P.Internal.SamControlClient.c_DefaultSamPort);
            TunnelSettings l_InboundTunnelSettings;
            TunnelSettings l_OutboundTunnelSettings;
            TunnelSettings.CreateDefault(out l_InboundTunnelSettings);
            TunnelSettings.CreateDefault(out l_OutboundTunnelSettings);

            a_Dest = new GeneralPrefs(true,
                                      true,
                                      true,
                                      true,
                                      false,
                                      true,
                                      false,
                                      false,
                                      ProgessBarKinds.IncludeUnverified,
                                      true,
                                      ConnStartKinds.AutoStart,
                                      a_OwnPersona,
                                      true,
                                      l_SamEndPoint,
                                      l_InboundTunnelSettings,
                                      l_OutboundTunnelSettings,
                                      null,
                                      null,
                                      true,
                                      true,
                                      Software.sp_DefaultPreviewFolder,
                                      c_DefaultGridViewTextSizeRatio,
                                      GridViewBorderStyles.NoBorder,
                                      false,
                                      c_DefaultUseBgImgsForComplBar,
                                      c_DefaultGridViewAutoSortItemThreshold,
                                      c_DefaultShowPeersPanelAdvSubMenu,
                                      Color.Empty,
                                      c_RunFilesOnDblClick,
                                      false,
                                      MainForm.MainTabs.Files,
                                      false);
        }

        public void DetermineChangeKind(ref GeneralPrefs a_ChangeTo,
                                        ref PrefChangeKinds a_ChangeKind)
        {
            if (PrefChangeKinds.AppRestartNeededToFullyTakeOver == a_ChangeKind)
            {
                return;
            }

            TunnelSettings l_OldOutboundTunnelSettings = this.GetEffectiveOutbountTunnelSettings();
            TunnelSettings l_NewOutboundTunnelSettings = a_ChangeTo.GetEffectiveOutbountTunnelSettings();

            if (PrefChangeKinds.ReconnectNeededToFullyTakeOver == a_ChangeKind ||
                false == this.f_OwnPersona.ValueEquals(a_ChangeTo.f_OwnPersona) ||
                // NOTE: IPEndPoint.Equals() works as value comparer.
                false == this.f_SamEndPoint.Equals(a_ChangeTo.f_SamEndPoint) ||
                false == this.f_InboundTunnelSettings.ValueEquals(a_ChangeTo.f_InboundTunnelSettings) ||
                false == l_OldOutboundTunnelSettings.ValueEquals(l_NewOutboundTunnelSettings) ||
                false == IsSameI2PKeyPair(this.f_DatagramKeyPair, a_ChangeTo.f_DatagramKeyPair) ||
                false == IsSameI2PKeyPair(this.f_StreamKeyPair, a_ChangeTo.f_StreamKeyPair))
            {
                a_ChangeKind = PrefChangeKinds.ReconnectNeededToFullyTakeOver;

                return;
            }

            if (PrefChangeKinds.ApplicableWithoutReconnect == a_ChangeKind ||
                this.f_SaveSearchHistory != a_ChangeTo.f_SaveSearchHistory ||
                this.f_ShowSpeedOnTitle != a_ChangeTo.f_ShowSpeedOnTitle ||
                this.f_ShowTrayIcon != a_ChangeTo.f_ShowTrayIcon ||
                this.f_ShowBalloonNotifications != a_ChangeTo.f_ShowBalloonNotifications ||
                this.f_MinimizeToTray != a_ChangeTo.f_MinimizeToTray ||
                this.f_PromptOnExit != a_ChangeTo.f_PromptOnExit ||
                this.f_BeepWhenDlIsDone != a_ChangeTo.f_BeepWhenDlIsDone ||
                this.f_DisplaySISizes != a_ChangeTo.f_DisplaySISizes ||
                this.f_ProgressBarKind != a_ChangeTo.f_ProgressBarKind ||
                this.f_CompactOnExit != a_ChangeTo.f_CompactOnExit ||
                this.f_ConnStartKind != a_ChangeTo.f_ConnStartKind ||
                this.f_UseInboundTunnelSettingsForOutboundTunnels != a_ChangeTo.f_UseInboundTunnelSettingsForOutboundTunnels ||
                this.f_AutoManageDatagramKeys != a_ChangeTo.f_AutoManageDatagramKeys ||
                this.f_AutoManageStreamKeys != a_ChangeTo.f_AutoManageStreamKeys ||
                this.f_PreviewFolder != a_ChangeTo.f_PreviewFolder ||
                this.f_GridViewTextSizeRatio != a_ChangeTo.f_GridViewTextSizeRatio ||
                this.f_GridViewBorderStyle != a_ChangeTo.f_GridViewBorderStyle ||
                this.f_AllowComplFilesDeletion != a_ChangeTo.f_AllowComplFilesDeletion ||
                this.f_UseBackgroundImagesForComplBar != a_ChangeTo.f_UseBackgroundImagesForComplBar ||
                this.f_GridViewAutoSortItemThreshold != a_ChangeTo.f_GridViewAutoSortItemThreshold ||
                this.f_ShowPeersPanelAdvSubMenu != a_ChangeTo.f_ShowPeersPanelAdvSubMenu ||
                this.f_IconBgColor != a_ChangeTo.f_IconBgColor ||
                this.f_RunFilesOnDblClick != a_ChangeTo.f_RunFilesOnDblClick ||
                this.f_UIModeAdvanced != a_ChangeTo.f_UIModeAdvanced ||
                this.f_PreferredMainTab != a_ChangeTo.f_PreferredMainTab ||
                this.f_AdditionalKeyCombinationForExiting != a_ChangeTo.f_AdditionalKeyCombinationForExiting)
            {
                a_ChangeKind = PrefChangeKinds.ApplicableWithoutReconnect;
            }
        }

        public TunnelSettings GetEffectiveOutbountTunnelSettings()
        {
            return (false == f_UseInboundTunnelSettingsForOutboundTunnels ?
                    f_OutboundTunnelSettings : f_InboundTunnelSettings);
        }

        public void SetKeyPairs(I2PKeyPair a_DatagramKeyPair,
                                I2PKeyPair a_StreamKeyPair)
        {
            f_DatagramKeyPair = a_DatagramKeyPair;
            f_StreamKeyPair = a_StreamKeyPair;
        }

        private static bool I2PKeyPairFromBuffer(BufferIO a_Buffer,
                                                 out I2PKeyPair a_KeyPair)
        {
            // a_KeyPair can be null even on successful return.

            bool l_HasKeys = a_Buffer.ReadBool();

            if (false == l_HasKeys)
            {
                a_KeyPair = null;
            }
            else
            {
                a_KeyPair = I2PKeyPair.FromBuffer(a_Buffer);

                if (null == a_KeyPair)
                {
                    return false;
                }
            }

            return true;
        }

        private static void I2PKeyPairToBuffer(BufferIO a_Buffer,
                                               I2PKeyPair a_KeyPair)
        {
            // a_KeyPair may be null.

            if (null == a_KeyPair)
            {
                a_Buffer.WriteBool(false);
            }
            else
            {
                a_Buffer.WriteBool(true);
                a_KeyPair.ToBuffer(a_Buffer);
            }
        }

        private static bool IsSameI2PKeyPair(I2PKeyPair a_A,
                                             I2PKeyPair a_B)
        {
            // Input may be null

            if (null == a_A)
            {
                return (null == a_B);
            }

            return (null != a_B &&
                    true == a_A.ValueEquals(a_B, false));
        }

        public Font BuildGridViewFont(Font a_Template)
        {
            float l_Size = a_Template.SizeInPoints;
            return new Font(a_Template.FontFamily, l_Size * (f_GridViewTextSizeRatio / 100f), a_Template.Style);
        }

        public enum ProgessBarKinds : byte
        {
            VerifiedOnly = 0,
            IncludeUnverified = 1,
        }

        public enum ConnStartKinds : byte
        {
            ManualStart = 0,
            ManualStartButRepair = 1,
            AutoStart = 2,
        }
    }

    enum FileRemovalModes : byte
    {
        DoNothing = 0,
        MoveToRecycleBin = 1,
        DeletePermanently = 2,
    }

    class Instance
    {
        private const uint c_BufferThreshold = (2 * 1024 * 1024);

        public const ushort c_ReservedMemElmKeys = 256;
        public const byte c_PrefsElmKey = 1;
        public const byte c_NodesElmKey = 2;
        public const byte c_UserCntxtsElmKey = 3;
        public const byte c_KnownFilesElmKey = 4;
        public const byte c_SearchCacheElmKey = 5;

        public delegate StorageStates BufferIOCallback(BufferIO a_Buffer);

        private TimeKeeper f_TimeKeeper;

        public readonly Resource f_MemResource;
        public readonly HardMemory f_Mem;

        // Essentials.
        public readonly TextSet f_TextSet;
        public readonly MainForm f_MainForm;
        public readonly NetworkManager f_NetworkMngr;
        public readonly ContentManager f_ContentMngr;

        // We will init the following after the above objects were
        // created (mind the order).
        public readonly Kademlia f_Kad;
        public readonly UpDownManager f_UpDwnMngr;
        public readonly UserManager f_UserMngr;

        public Instance(TextSet a_TextSet,
                        MainForm a_MainForm,
                        Preferences a_Prefs,
                        HardMemory a_Mem,
                        HardMemory a_IncomplMem,
                        byte[] a_InitialNodes,
                        out Instance a_Inst)
        {
            // a_Inst immediately receives our address.
            a_Inst = this;
            f_TimeKeeper = new TimeKeeper(Time.Now());

            f_MemResource = Resource.CreateSMP();
            f_Mem = a_Mem;
            f_Mem.SetResource(f_MemResource, false);

            f_TextSet = a_TextSet;
            f_MainForm = a_MainForm;
            f_NetworkMngr = new NetworkManager(this, a_Prefs.f_GeneralPrefs.f_OwnPersona, ref a_Prefs.f_NetworkPrefs);
            f_ContentMngr = new ContentManager(this, a_IncomplMem, ref a_Prefs.f_IOPrefs);

            f_Kad = new Kademlia(this, ref a_Prefs.f_KadPrefs, a_InitialNodes);
            f_UserMngr = new UserManager(this);
            f_UpDwnMngr = new UpDownManager(this);
        }

        public string this[TextKeys a_Key]
        {
            get { return f_TextSet.GetString(a_Key); }
        }

        public void InvokeMainFormAsync(Delegate a_Delegate)
        {
            InvokeMainFormAsync(a_Delegate, null);
        }

        public void InvokeMainFormAsync(Delegate a_Delegate,
                                        params object[] a_Args)
        {
            f_MainForm.f_Invoker.InvokeAsync(a_Delegate, a_Args);
        }

        public Time UpdateTime()
        {
            return f_TimeKeeper.UpdateTime();
        }

        public Time GetTime()
        {
            return f_TimeKeeper.GetTime();
        }

        public StorageStates DeleteElement(ElmKey a_ElmKey)
        {
            StorageStates l_State = StorageStates.Success;

            f_MemResource.AcquireExclusive();
            {
                if (false == f_Mem.Delete(a_ElmKey))
                {
                    l_State = StorageStates.UpdateFailed;
                }
            }
            f_MemResource.ReleaseExclusive();

            return l_State;
        }

        public StorageStates WriteToElement(ElmKey a_ElmKey,
                                            BufferIOCallback a_Callback)
        {
            TemporaryFile l_TempFile = TemporaryFile.Create();

            if (null == l_TempFile)
            {
                return StorageStates.TempFileIOError;
            }

            FHandle l_DestFile = l_TempFile.ExposeFHandle();
            StorageStates l_State;

            try
            {
                FileIO l_File = FileIO.CreateWritable(l_DestFile);
                l_State = a_Callback(l_File);
                l_File.Flush();
            }
            catch (BufferIO.BufferIOException)
            {
                l_State = StorageStates.TempFileIOError;
            }

            if (StorageStates.Success == l_State)
            {
                f_MemResource.AcquireExclusive();
                {
                    if (false == f_Mem.SetAtomic(a_ElmKey, l_DestFile))
                    {
                        l_State = StorageStates.UpdateFailed;
                    }
                }
                f_MemResource.ReleaseExclusive();
            }

            l_TempFile.Close();

            return l_State;
        }

        private static StorageStates ReadFromElement(HardMemory a_Mem,
                                                     ElmKey a_ElmKey,
                                                     BufferIOCallback a_Callback)
        {
            // Loads data from an element into a buffer OR file.

            Resource f_Resource = a_Mem.p_Resource;
            TemporaryFile l_TempFile = null;

            // Begin of potential critical section.

            if (null != f_Resource)
            {
                f_Resource.AcquireShared();
            }

            StorageStates l_State;
            HardMemory.Element l_Elm = a_Mem.GetElement(a_ElmKey);

            if (null == l_Elm)
            {
                l_State = StorageStates.NoData;

                goto RELEASE_AND_EXIT;
            }

            BufferIO l_Buffer;

            if (c_BufferThreshold >= l_Elm.f_MaxSize)
            {
                ByteArray l_Data = new ByteArray((int)l_Elm.f_MaxSize);

                if (false == a_Mem.Read(l_Elm, 0, ref l_Data))
                {
                    l_State = StorageStates.ReadFailed;
                }
                else
                {
                    l_Buffer = MemoryIO.CreateReadable(l_Data.f_Arr);
                    l_State = a_Callback(l_Buffer);
                }
            }
            else
            {
                l_TempFile = TemporaryFile.Create();

                if (null == l_TempFile)
                {
                    l_State = StorageStates.TempFileIOError;

                    goto RELEASE_AND_EXIT;
                }

                FHandle l_SrcFile = l_TempFile.ExposeFHandle();

                if (false == a_Mem.CopyToFile(l_Elm, l_SrcFile, false, false))
                {
                    l_State = StorageStates.TempFileIOError;

                    goto RELEASE_AND_EXIT;
                }

                try
                {
                    l_Buffer = FileIO.CreateReadable(l_SrcFile);
                    l_State = a_Callback(l_Buffer);
                }
                catch (BufferIO.BufferIOException)
                {
                    l_State = StorageStates.TempFileIOError;
                }
            }

            RELEASE_AND_EXIT:

            if (null != f_Resource)
            {
                f_Resource.ReleaseShared();
            }

            // End of potential critical section.

            if (null != l_TempFile)
            {
                l_TempFile.Close();
            }

            return l_State;
        }

        public StorageStates ReadFromElement(ElmKey a_ElmKey,
                                             BufferIOCallback a_Callback)
        {
            return ReadFromElement(f_Mem, a_ElmKey, a_Callback);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct TimeKeeper
        {
            // Since Time.Now() requires a system call, we query and
            // store the latest time periodically and keep it ready
            // here for anyone to query without leaving user mode.

            private bool f_64Bit;
            private long f_Long;

            public TimeKeeper(Time a_InitialTime)
            {
                f_64Bit = (sizeof(ulong) == IntPtr.Size);
                f_Long = (long)a_InitialTime.ToBinary();
            }

            public Time UpdateTime()
            {
                long l_NewValue = (long)Time.Now().ToBinary();

                if (false == f_64Bit)
                {
                    // Atomic store operation.

                    System.Threading.Interlocked.Exchange(ref f_Long, l_NewValue);
                }
                else
                {
                    f_Long = l_NewValue;
                }

                return Time.FromBinary((ulong)l_NewValue);
            }

            public Time GetTime()
            {
                long l_Value;

                if (false == f_64Bit)
                {
                    // Atomic load operation.

                    l_Value = System.Threading.Interlocked.Read(ref f_Long);
                }
                else
                {
                    l_Value = f_Long;
                }

                return Time.FromBinary((ulong)l_Value);
            }
        }
    }

    class VisualCustomization
    {
        public readonly ColumnSet f_KademliaPanelColSet;
        public readonly ColumnSet f_FileSearchPanelColSet;
        public readonly ColumnSet f_FilesTabColSet;
        public readonly ColumnSet f_FilePeersPanelColSet;
        public readonly ColumnSet f_TransferPanelColSet;
        public readonly ColumnSet f_FileNamesPanelColSet;
        public readonly ColumnSet f_CommentsPanelColSet;
        public readonly ColumnSet f_SharedDirsPanelColSet;
        public readonly ColumnSet f_OpenCollectionDlgColSet;
        public readonly ColumnSet f_GeneralPeersColSet;
        public readonly ColumnSet f_ImportPartialDlsDlgColSet;
        // From 1.4.25 on.
        public readonly ColumnSet f_KadFilterPanelColSet;
        public readonly ColumnSet f_TextItemsDlgColSet;

        private readonly ColumnSet[] f_Arr;

        // f_OptionsDlgPage is not saved to disk yet.
        public OptionsDialog.PageTypes f_OptionsDlgPage = OptionsDialog.PageTypes.KadFilter;

        public VisualCustomization(TextSet a_TextSet)
        {
            bool l_CheckedBuild =
#if DEBUG
            true;
#else
            false;
#endif

            ColumnSet.Creator l_Creator;

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.KademliaTabColumnTitles, (int)KadContactColumns.Distance + 1);
            l_Creator.AddText(KadContactColumns.ClientID, true, 235);
            l_Creator.AddText(KadContactColumns.LastActivity, true, 80);
            l_Creator.AddText(KadContactColumns.Distance, true, 225);
            f_KademliaPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FileSearchTabColumnTitles, (int)SearchResultColumns.LastSeen + 1);
            l_Creator.AddText(SearchResultColumns.FileName, true, 350);
            l_Creator.AddText(SearchResultColumns.FileSize, true, 70);
            l_Creator.AddText(SearchResultColumns.Sources, true, 60);
            l_Creator.AddText(SearchResultColumns.FileType, true, 80);
            l_Creator.AddText(SearchResultColumns.FileID, true, 235);
            l_Creator.AddText(SearchResultColumns.FirstSeen, false, 120);
            l_Creator.AddText(SearchResultColumns.LastSeen, true, 120);
            f_FileSearchPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FilesTabColumnTitles, (int)FileColumns.LastReception + 1);
            l_Creator.AddText(FileColumns.FileName, true, 200);
            l_Creator.AddText(FileColumns.Size, true, 70);
            l_Creator.AddText(FileColumns.Status, true, 65);
            l_Creator.AddText(FileColumns.Completed, false, 110);
            l_Creator.AddText(FileColumns.CompletedUnchecked, true, 125);
            l_Creator.AddText(FileColumns.Remaining, false, 80);
            l_Creator.Add(FileColumns.Progress, true, 60, ColumnSet.ColumnKind.OwnerDrawn);
            l_Creator.AddText(FileColumns.ETA, true, 60);
            l_Creator.AddText(FileColumns.Sources, true, 60);
            l_Creator.AddText(FileColumns.DownSpeed, true, 75);
            l_Creator.AddText(FileColumns.UpSpeed, true, 75);
            l_Creator.AddText(FileColumns.Received, false, 80);
            l_Creator.AddText(FileColumns.Sent, true, 80);
            l_Creator.AddText(FileColumns.Ratio, true, 60);
            l_Creator.AddText(FileColumns.ULDL, false, 55);
            l_Creator.AddText(FileColumns.Availability, true, 70);
            l_Creator.AddText(FileColumns.TimeActive, false, 75);
            l_Creator.AddText(FileColumns.AddedOn, false, 120);
            l_Creator.AddText(FileColumns.LastSeenComplete, false, 120);
            l_Creator.AddText(FileColumns.LastReception, false, 120);
            f_FilesTabColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FilesTabPeersColumnTitles, (int)FilePeerColumns.SentUserTotal + 1);
            l_Creator.AddText(FilePeerColumns.UserNum, l_CheckedBuild, 70);
            l_Creator.AddText(FilePeerColumns.RemoteEndpoint, true, 185);
            l_Creator.AddText(FilePeerColumns.Direction, true, 70);
            l_Creator.AddText(FilePeerColumns.UserHash, l_CheckedBuild, 220);
            l_Creator.AddText(FilePeerColumns.UserName, true, 150);
            l_Creator.AddText(FilePeerColumns.Client, true, 100);
            l_Creator.AddText(FilePeerColumns.Authentication, true, 120);
            l_Creator.AddText(FilePeerColumns.DlStatus, true, 110);
            l_Creator.AddText(FilePeerColumns.Done, true, 60);
            l_Creator.Add(FilePeerColumns.Completeness, true, 80, ColumnSet.ColumnKind.OwnerDrawn);
            l_Creator.AddText(FilePeerColumns.DownSpeed, true, 75);
            l_Creator.AddText(FilePeerColumns.UpSpeed, true, 75);
            l_Creator.AddText(FilePeerColumns.ReceivedDuringConn, true, 90);
            l_Creator.AddText(FilePeerColumns.SentDuringConn, true, 90);
            l_Creator.AddText(FilePeerColumns.ReceivedUserTotal, false, 90);
            l_Creator.AddText(FilePeerColumns.SentUserTotal, false, 90);
            f_FilePeersPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FilesTabTransfersColumnTitles, (int)TransferColumns.Progress + 1);
            l_Creator.AddText(TransferColumns.UserNum, l_CheckedBuild, 70);
            l_Creator.AddText(TransferColumns.RemoteEndpoint, true, 185);
            l_Creator.AddText(TransferColumns.Direction, true, 70);
            l_Creator.AddText(TransferColumns.RangeOffset, true, 90);
            l_Creator.AddText(TransferColumns.RangeLength, true, 70);
            l_Creator.AddText(TransferColumns.Encoding, true, 70);
            l_Creator.Add(TransferColumns.Progress, true, 60, ColumnSet.ColumnKind.OwnerDrawn);
            f_TransferPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FilesTabFileNamesColumnTitles, (int)FileNameColumns.FileName + 1);
            l_Creator.AddText(FileNameColumns.FileName, true, 500);
            f_FileNamesPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.FilesTabCommentsColumnTitles, (int)CommentColumns.Description + 1);
            l_Creator.AddText(CommentColumns.FileName, true, 210);
            l_Creator.AddText(CommentColumns.Rating, true, 105);
            l_Creator.AddText(CommentColumns.Description, true, 280);
            f_CommentsPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.OptionsDlgSharedDirsDlgColumnTitles, (int)SharedFilesColumns.FriendsOnly + 1);
            l_Creator.AddText(SharedFilesColumns.Path, true, 195);
            l_Creator.Add(SharedFilesColumns.Recursive, true, 65, ColumnSet.ColumnKind.Bool);
            l_Creator.Add(SharedFilesColumns.FriendsOnly, true, 75, ColumnSet.ColumnKind.Bool);
            f_SharedDirsPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.OpenCollDlgColumnTitles, (int)CollectionColumns.TAICHHash + 1);
            l_Creator.Add(CollectionColumns.Add, true, 30, ColumnSet.ColumnKind.Bool);
            l_Creator.AddText(CollectionColumns.FileName, true, 150);
            l_Creator.AddText(CollectionColumns.Size, true, 70);
            l_Creator.AddText(CollectionColumns.eD2kHash, true, 240);
            l_Creator.AddText(CollectionColumns.AICHHash, true, 265);
            l_Creator.AddText(CollectionColumns.TAICHHash, true, 240);
            f_OpenCollectionDlgColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.PeersTabColumnTitles, (int)GeneralPeersColumns.Banned + 1);
            l_Creator.AddText(GeneralPeersColumns.UserNum, l_CheckedBuild, 70);
            l_Creator.AddText(GeneralPeersColumns.RemoteEndpoint, true, 185);
            l_Creator.AddText(GeneralPeersColumns.PublicKey, true, 90);
            l_Creator.AddText(GeneralPeersColumns.UserHash, l_CheckedBuild, 220);
            l_Creator.AddText(GeneralPeersColumns.UserName, true, 150);
            l_Creator.Add(GeneralPeersColumns.Friend, true, 60, ColumnSet.ColumnKind.Bool);
            l_Creator.AddText(GeneralPeersColumns.Connections, true, 80);
            l_Creator.AddText(GeneralPeersColumns.LastConnected, true, 120);
            l_Creator.AddText(GeneralPeersColumns.ReceivedFrom, true, 115);
            l_Creator.AddText(GeneralPeersColumns.SentTo, true, 115);
            l_Creator.AddText(GeneralPeersColumns.Reputation, true, 75);
            l_Creator.Add(GeneralPeersColumns.Banned, true, 60, ColumnSet.ColumnKind.Bool);
            f_GeneralPeersColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.ImportPartialDlsDlgColumnTitles, (int)ImportPartialDlsColumns.UlPriority + 1);
            l_Creator.Add(ImportPartialDlsColumns.Import, true, 30, ColumnSet.ColumnKind.Bool);
            l_Creator.AddText(ImportPartialDlsColumns.FileName, true, 150);
            l_Creator.AddText(ImportPartialDlsColumns.eD2kHash, true, 240);
            l_Creator.AddText(ImportPartialDlsColumns.SourceFilePath, false, 150);
            l_Creator.AddText(ImportPartialDlsColumns.FileSize, true, 70);
            l_Creator.AddText(ImportPartialDlsColumns.Completed, true, 70);
            l_Creator.Add(ImportPartialDlsColumns.Progress, true, 60, ColumnSet.ColumnKind.OwnerDrawn);
            l_Creator.AddText(ImportPartialDlsColumns.Fragments, false, 70);
            l_Creator.Add(ImportPartialDlsColumns.PartHashesPresent, true, 80, ColumnSet.ColumnKind.Bool);
            l_Creator.AddText(ImportPartialDlsColumns.Status, true, 75);
            l_Creator.AddText(ImportPartialDlsColumns.DlPriority, true, 65);
            l_Creator.AddText(ImportPartialDlsColumns.UlPriority, true, 65);
            f_ImportPartialDlsDlgColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.OptionsDlgKadFilterRuleColumnEnabled, (int)KadFilterPanelColumns.Title + 1);
            l_Creator.Add(KadFilterPanelColumns.Enabled, true, 65, ColumnSet.ColumnKind.Bool);
            l_Creator.AddText(KadFilterPanelColumns.Title, true, 300);
            f_KadFilterPanelColSet = l_Creator.Finish();

            l_Creator = new ColumnSet.Creator(a_TextSet, TextKeys.TextItemsDlgColumnText, (int)TextItemsDialogColumns.Text + 1);
            l_Creator.AddText(TextItemsDialogColumns.Text, true, 300);
            f_TextItemsDlgColSet = l_Creator.Finish();

            f_Arr = new ColumnSet[13]
            {
                f_KademliaPanelColSet,
                f_FileSearchPanelColSet,
                f_FilesTabColSet,
                f_FilePeersPanelColSet,
                f_TransferPanelColSet,
                f_FileNamesPanelColSet,
                f_CommentsPanelColSet,
                f_SharedDirsPanelColSet,
                f_OpenCollectionDlgColSet,
                f_GeneralPeersColSet,
                f_ImportPartialDlsDlgColSet,
                f_KadFilterPanelColSet,
                f_TextItemsDlgColSet,
            };
        }

        public void Reset()
        {
            foreach (ColumnSet l_ColSet in f_Arr)
            {
                l_ColSet.ResetAdjustments();
            }
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteByte((byte)f_Arr.Length);

            foreach (ColumnSet l_ColSet in f_Arr)
            {
                l_ColSet.AdjustmentsToBuffer(a_Buffer);
            }
        }

        public bool FromBuffer(BufferIO a_Buffer)
        {
            // This routine is backward compatible and forward
            // tolerant.

            uint l_UseableCount = Util.GetArrayLength(f_Arr);
            uint l_ProvidedCount = a_Buffer.ReadByte();

            for (uint i = 0, l_CommonCount = Math.Min(l_UseableCount, l_ProvidedCount); i != l_CommonCount; i++)
            {
                if (false == f_Arr[i].AdjustmentsFromBuffer(a_Buffer))
                {
                    return false;
                }
            }

            if (l_UseableCount == l_ProvidedCount)
            {
                return true;
            }

            if (l_UseableCount > l_ProvidedCount)
            {
                for (uint i = l_ProvidedCount; i != l_UseableCount; i++)
                {
                    f_Arr[i].ResetAdjustments();
                }
            }
            else
            {
                if (false == ColumnSet.DummyReadFromBuffer(a_Buffer, l_ProvidedCount - l_UseableCount))
                {
                    return false;
                }
            }

            return true;
        }

        public enum KadContactColumns : byte
        {
            ClientID,
            LastActivity,
            Distance,
        }

        public enum SearchResultColumns : byte
        {
            FileName,
            FileSize,
            Sources,
            FileType,
            FileID,
            FirstSeen,
            LastSeen,
        }

        public enum FileColumns : byte
        {
            FileName,
            Size,
            Status,
            Completed,
            CompletedUnchecked,
            Remaining,
            Progress,
            ETA,
            Sources,
            DownSpeed,
            UpSpeed,
            Received,
            Sent,
            Ratio,
            ULDL,
            Availability,
            TimeActive,
            AddedOn,
            LastSeenComplete,
            LastReception,
            Category,
            DlPriority,
            UlPriority,
            DlLimit,
            UlLimit,
        }

        public enum FilePeerColumns : byte
        {
            UserNum,
            RemoteEndpoint,
            Direction,
            UserHash,
            UserName,
            Client,
            Authentication,
            DlStatus,
            Done,
            Completeness,
            DownSpeed,
            UpSpeed,
            ReceivedDuringConn,
            SentDuringConn,
            ReceivedUserTotal,
            SentUserTotal,
        }

        public enum TransferColumns : byte
        {
            UserNum,
            RemoteEndpoint,
            Direction,
            RangeOffset,
            RangeLength,
            Encoding,
            Progress,
        }

        public enum FileNameColumns : byte
        {
            FileName,
        }

        public enum CommentColumns : byte
        {
            FileName,
            Rating,
            Description,
        }

        public enum SharedFilesColumns : byte
        {
            Path,
            Recursive,
            FriendsOnly,
        }

        public enum CollectionColumns : byte
        {
            Add,
            FileName,
            Size,
            eD2kHash,
            AICHHash,
            TAICHHash,
        }

        public enum GeneralPeersColumns : byte
        {
            UserNum,
            RemoteEndpoint,
            PublicKey,
            UserHash,
            UserName,
            Friend,
            Connections,
            LastConnected,
            ReceivedFrom,
            SentTo,
            Reputation,
            Banned,
        }

        public enum ImportPartialDlsColumns : byte
        {
            Import,
            FileName,
            eD2kHash,
            SourceFilePath,
            FileSize,
            Completed,
            Progress,
            Fragments,
            PartHashesPresent,
            Status,
            DlPriority,
            UlPriority,
        }

        enum KadFilterPanelColumns : byte
        {
            Enabled,
            Title,
        }

        enum TextItemsDialogColumns : byte
        {
            Text,
        }
    }
}

// End of file.
